#include "apm.h"

APM::AccuracyType APM::m_eAccuracyType = APM::atExact;

//////////////////////////////////////////////////////////////////////
// APM Class Implementation

APM::APM()
{
}

APM::~APM()
{
}

VOID APM::Init()
{
}

VOID APM::SolveLinear(DOUBLE *pdSolutions, UINT_32 &uiSolutions, DOUBLE *pdCoefs, UINT_32 uiCoefs/* = 2*/)
{
  ASSERT( pdSolutions != NULL );
  ASSERT( pdCoefs != NULL );
  ASSERT( uiCoefs == 2 );

  if ( pdCoefs[1] == 0.0 )
    {
      uiSolutions = 0;
      *pdSolutions = 0.0;
      return;
    }

  pdSolutions[0] = -pdCoefs[0] / pdCoefs[1];
  uiSolutions = 1;
}

VOID APM::SolveQuadratic(DOUBLE *pdSolutions, UINT_32 &uiSolutions, DOUBLE *pdCoefs, UINT_32 uiCoefs/* = 3*/)
{
  ASSERT( pdSolutions != NULL );
  ASSERT( pdCoefs != NULL );
  ASSERT( uiCoefs == 3 );

  if ( pdCoefs[2] == 0.0 )
    {
      pdSolutions[1] = 0.0;
      SolveLinear(pdSolutions, uiSolutions, pdCoefs);
      return;
    }
  DOUBLE dDen = 2.0*pdCoefs[2];
  ASSERT( dDen != 0.0 );

  DOUBLE dDisc = pdCoefs[1]*pdCoefs[1] - 4.0*pdCoefs[0]*pdCoefs[2];
  if ( dDisc < 0.0 )
    {
      uiSolutions = 0;
      pdSolutions[0] = -pdCoefs[1] / dDen;
      pdSolutions[1] = pdSolutions[0];
      return;
    }
  dDisc = sqrt(dDisc);

  uiSolutions = 0;
  pdSolutions[0] = ( -pdCoefs[1] + dDisc ) / dDen;
  pdSolutions[1] = ( -pdCoefs[1] - dDisc ) / dDen;
}

VOID APM::SolveCubic(DOUBLE *pdSolutions, UINT_32 &uiSolutions, DOUBLE *pdCoefs, UINT_32 uiCoefs/* = 4*/)
{
  ASSERT( pdSolutions != NULL );
  ASSERT( pdCoefs != NULL );
  ASSERT( uiCoefs == 4 );

  UINT_32 i;

  if ( pdCoefs[3] == 0.0 )
    {
      pdSolutions[2] = 0.0;
      SolveQuadratic(pdSolutions, uiSolutions, pdCoefs);
      return;
    }

  // Normalize the Cubic to 
  //   x^3 + c2 x^2 + c1 x + c0 = 0
  DOUBLE pdC[3];
  for ( i = 0; i < 3; i++ )
    {
      pdC[i] = pdCoefs[i] / pdCoefs[3];
    }

  // Transform with x = y - (c2 / 3) to 
  //   y^3 + a1 y + a0 = 0
  //   a1 = ( 3 c1 - c2^2 ) / 3
  //   a0 = ( 2 c2^3 - 9 c2 c1 + 27 c0 ) / 27
  
  DOUBLE pdA[2];
  pdA[1] = ( 3.0*pdC[1] - pdC[2]*pdC[2] ) / 3.0;
  pdA[0] = ( 2.0*pdC[2]*pdC[2]*pdC[2] - 9.0*pdC[2]*pdC[1] + 27.0*pdC[0] ) / 27.0;

  // Solutions are of the form
  //   A = ( -a0 / 2 + sqrt(a0^2 / 4 + a1^3 / 27) )^(1/3)
  //   B = -( a0 / 2 + sqrt(a0^2 / 4 + a1^3 / 27) )^(1/3)
  //   y = A + B, -(A + B)/2 + (A - B)sqrt(-3)/2, -(A + B)/2 - (A - B)sqrt(-3)/2
  DOUBLE  pdY[3];

  DOUBLE dDisc = ( pdA[0]*pdA[0]/4.0 ) + ( pdA[1]*pdA[1]*pdA[1]/27.0 );
  if ( dDisc > 0.0 )
    {
      // 1 real root, 2 complex conjugate solutions
      dDisc = sqrt(dDisc);

      uiSolutions = 1;
      pdY[0] = pow(-pdA[0]/2.0 + dDisc,1.0/3.0) - pow(pdA[0]/2.0 + dDisc,1.0/3.0);
      pdY[1] = 0.0;
      pdY[2] = 0.0;
    }
  else if ( dDisc == 0.0 )
    {
      // 3 real roots, at least one is a double root
      uiSolutions = 3;
      pdY[1] = -pow(pdA[0]/2.0,1.0/3.0);
      pdY[2] = pdY[1];
      pdY[0] = 2.0*pdY[1];
    }
  else
    {
      // dDisc < 0.0
      // 3 real distinct roots use the trigonometric solution
      DOUBLE dM = -pdA[1]/3.0;
      if ( dM <= 0.0 )
	{
	  ASSERT( FALSE );
	}
      dM = 2.0*sqrt(dM);

      ASSERT( pdA[1] != 0.0 );
      DOUBLE dCos3Theta = 3.0*pdA[0]/(dM*pdA[1]);
      if ( dCos3Theta > 1.0 )
	{
	  ASSERT( FALSE );
	  dCos3Theta = 1.0;
	}
      if ( dCos3Theta < -1.0 )
	{
	  ASSERT( FALSE );
	  dCos3Theta = -1.0;
	}

      DOUBLE dTheta = acos(dCos3Theta)/3.0;

      uiSolutions = 3;
      pdY[0] = dM*cos(dTheta);
      pdY[1] = dM*cos(dTheta + (2.0*PI/3.0));
      pdY[2] = dM*cos(dTheta + (4.0*PI/3.0));
    }

  // Transform back with x = y - (c2 / 3)
  for ( i = 0; i < uiSolutions; i++ )
    {
      pdSolutions[i] = pdY[i] - ( pdC[2] / 3.0 );
    }
  for ( ; i < 3; i++ )
    {
      pdSolutions[i] = 0.0;
    }
}
