#ifndef __APM2D_H__
#define __APM2D_H__

#include "apm.h"

//////////////////////////////////////////////////////////////////////
// APM2D Class -- Adaptive Precision Math 2D
//

class APM2D
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  enum EdgeZoneType { eztStart, eztMiddle, eztEnd, eztSize }; 
  enum CircleTestType { cttInCircle = -1, cttOnCircle = 0, cttOutCircle = 1, cttSize = 2 }; 
  enum InscribedCircleCenterType { icctNull, 
				   icctVVV, 
				   icctVVE, icctVVEAntiparallel, 
				   icctVvE, icctEvV, 
				   icctVEE, icctVEEAntiparallel,
				   icctEvE, // degenerate case where the vertex is shared
				   icctVEvE, icctVEcE, 
				   icctvEE, icctvEvE, 
				   icctEEv, icctEvEv,
				   icctEEE, 
				   icctEvEE, icctEcEE, 
				   icctEvEvE, icctEcEvE, icctEvEcE, 
				   icctEvEvEv,
				   icctInfinity,
				   icctSize };

//////////////////////////////////////////////////
// Initialization
//
public:
  APM2D();
  virtual ~APM2D();
  virtual VOID Init();
  virtual VOID Uninit() = 0;

//////////////////////////////////////////////////
// Intersection point of two edges
//
public:
  static BOOL IntersectionEE(DOUBLE *pdV_x, DOUBLE *pdV_y,
			     DOUBLE dE00_x, DOUBLE dE00_y, 
			     DOUBLE dE01_x, DOUBLE dE01_y,
			     DOUBLE dE10_x, DOUBLE dE10_y, 
			     DOUBLE dE11_x, DOUBLE dE11_y);

//////////////////////////////////////////////////
// Site closest to a Vertex
//
public:
  // Use this to determine if a point is closest to the edge or one of its endpoints
  //   result < 0  Vertex E10
  //   result = 0  body of the edge (the line)
  //   result > 0  Vertex E11
  static EdgeZoneType EdgeZone(DOUBLE dV0_x, DOUBLE dV0_y,
			       DOUBLE dE10_x, DOUBLE dE10_y, 
			       DOUBLE dE11_x, DOUBLE dE11_y);

  // Is V0 closer to S1 or S2 (edges are treated like infinite lines, use EdgeZone first!)
  //  result < 0  S1 is closer
  //  result = 0  equidistant
  //  result > 0  S2 is closer
  static INT_32 DistanceCompareV_VV(DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dV2_x, DOUBLE dV2_y);
  static INT_32 DistanceCompareV_VL(DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y);
  static INT_32 DistanceCompareV_LL(DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dE10_x, DOUBLE dE10_y, 
				    DOUBLE dE11_x, DOUBLE dE11_y,
				    DOUBLE dE20_x, DOUBLE dE20_y, 
				    DOUBLE dE21_x, DOUBLE dE21_y);

  // Edge compare to two other sites
  static INT_32 DistanceCompareEvv_VV(DOUBLE dE0A_x, DOUBLE dE0A_y,
				      DOUBLE dE0B_x, DOUBLE dE0B_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dV2_x, DOUBLE dV2_y);
  static INT_32 DistanceCompareEvv_VL(DOUBLE dE0A_x, DOUBLE dE0A_y,
				      DOUBLE dE0B_x, DOUBLE dE0B_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dE20_x, DOUBLE dE20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y);
  static INT_32 DistanceCompareEvv_LL(DOUBLE dE0A_x, DOUBLE dE0A_y,
				      DOUBLE dE0B_x, DOUBLE dE0B_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y);

  static INT_32 DistanceCompareElv_VV(DOUBLE dE00_x, DOUBLE dE00_y,
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dE0B_x, DOUBLE dE0B_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dV2_x, DOUBLE dV2_y);
  static INT_32 DistanceCompareElv_VL(DOUBLE dE00_x, DOUBLE dE00_y,
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dE0B_x, DOUBLE dE0B_y,
				      DOUBLE dV1_x, DOUBLE dV1_y, 
				      DOUBLE dE20_x, DOUBLE dE20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y);

  static INT_32 DistanceCompareL_VV(DOUBLE dE00_x, DOUBLE dE00_y,
				    DOUBLE dE01_x, DOUBLE dE01_y,
				    DOUBLE dV1_x, DOUBLE dV1_y, 
				    DOUBLE dV2_x, DOUBLE dV2_y);

//////////////////////////////////////////////////
// Orientation of 3 Vertices
//
public:
  // result > 0  CCW
  // result = 0  Colinear
  // result < 0  CW
  static DOUBLE SinTheta(DOUBLE dV2_x, DOUBLE dV2_y,
			 DOUBLE dV0_x, DOUBLE dV0_y,
			 DOUBLE dV1_x, DOUBLE dV1_y);
protected:
  static DOUBLE OrientationVVVValue(DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV1_x, DOUBLE dV1_y,
				    DOUBLE dV2_x, DOUBLE dV2_y);
public:
  static INT_32 OrientationVVV(DOUBLE dV0_x, DOUBLE dV0_y,
			       DOUBLE dV1_x, DOUBLE dV1_y,
			       DOUBLE dV2_x, DOUBLE dV2_y);

  static inline BOOL Counterclockwise(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y,
				      DOUBLE dV2_x, DOUBLE dV2_y);
  static inline BOOL Colinear(DOUBLE dV0_x, DOUBLE dV0_y,
			      DOUBLE dV1_x, DOUBLE dV1_y,
			      DOUBLE dV2_x, DOUBLE dV2_y);
  static inline BOOL Clockwise(DOUBLE dV0_x, DOUBLE dV0_y,
			       DOUBLE dV1_x, DOUBLE dV1_y,
			       DOUBLE dV2_x, DOUBLE dV2_y);

//////////////////////////////////////////////////
// Parallel Edges
//
public:
  // | E0 x E1 | > 0  
  // | E0 x E1 | = 0  Parallel or Anti-Parallel
  // | E0 x E1 | < 0  
  static INT_32 OrientationExE(DOUBLE dE00_x, DOUBLE dE00_y,
			       DOUBLE dE01_x, DOUBLE dE01_y,
			       DOUBLE dE10_x, DOUBLE dE10_y,
			       DOUBLE dE11_x, DOUBLE dE11_y);

  static inline BOOL Parallel(DOUBLE dE00_x, DOUBLE dE00_y,
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y,
			      DOUBLE dE11_x, DOUBLE dE11_y);
  
  // ( E0 dot E1 ) > 0  Same
  // ( E0 dot E1 ) = 0  Perpendicular
  // ( E0 dot E1 ) < 0  Opposite
  static INT_32 OrientationEdotE(DOUBLE dE00_x, DOUBLE dE00_y,
				 DOUBLE dE01_x, DOUBLE dE01_y,
				 DOUBLE dE10_x, DOUBLE dE10_y,
				 DOUBLE dE11_x, DOUBLE dE11_y);

//////////////////////////////////////////////////
// Inscribed Circle Center
//
public:
  static InscribedCircleCenterType InscribedCircleCenterVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dV00_x, DOUBLE dV00_y,
							    DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV01_x, DOUBLE dV01_y,
							    DOUBLE dV10_x, DOUBLE dV10_y,
							    DOUBLE dV1_x, DOUBLE dV1_y,
							    DOUBLE dV11_x, DOUBLE dV11_y,
							    DOUBLE dV20_x, DOUBLE dV20_y,
							    DOUBLE dV2_x, DOUBLE dV2_y,
							    DOUBLE dV21_x, DOUBLE dV21_y);
  static EdgeZoneType EdgeZoneVVV(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dV1_x, DOUBLE dV1_y,
				  DOUBLE dV2_x, DOUBLE dV2_y,
				  DOUBLE dE30_x, DOUBLE dE30_y, 
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleVVV_V(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y,
				      DOUBLE dV2_x, DOUBLE dV2_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleVVV_E(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y,
				      DOUBLE dV2_x, DOUBLE dV2_y,
				      DOUBLE dE30_x, DOUBLE dE30_y, 
				      DOUBLE dE31_x, DOUBLE dE31_y);

  static InscribedCircleCenterType InscribedCircleCenterVVE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dV00_x, DOUBLE dV00_y,
							    DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV01_x, DOUBLE dV01_y,
							    DOUBLE dV10_x, DOUBLE dV10_y,
							    DOUBLE dV1_x, DOUBLE dV1_y,
							    DOUBLE dV11_x, DOUBLE dV11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y);
  static EdgeZoneType EdgeZoneVVE(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dV1_x, DOUBLE dV1_y,
				  DOUBLE dE20_x, DOUBLE dE20_y,
				  DOUBLE dE21_x, DOUBLE dE21_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleVVE_V(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleVVE_E(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_x, DOUBLE dV1_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  static EdgeZoneType EdgeZoneVVEAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dV1_x, DOUBLE dV1_y,
					      DOUBLE dE20_x, DOUBLE dE20_y,
					      DOUBLE dE21_x, DOUBLE dE21_y,
					      DOUBLE dE30_x, DOUBLE dE30_y,
					      DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleVVE_VAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						  DOUBLE dV1_x, DOUBLE dV1_y,
						  DOUBLE dE20_x, DOUBLE dE20_y,
						  DOUBLE dE21_x, DOUBLE dE21_y,
						  DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleVVE_EAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						  DOUBLE dV1_x, DOUBLE dV1_y,
						  DOUBLE dE20_x, DOUBLE dE20_y,
						  DOUBLE dE21_x, DOUBLE dE21_y,
						  DOUBLE dE30_x, DOUBLE dE30_y,
						  DOUBLE dE31_x, DOUBLE dE31_y);
  static InscribedCircleCenterType InscribedCircleCenterVvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dV00_x, DOUBLE dV00_y,
							    DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV01_x, DOUBLE dV01_y,
							    DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y);
  static EdgeZoneType EdgeZoneVvE(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				  DOUBLE dE21_x, DOUBLE dE21_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleVvE_V(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleVvE_E(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dV1_E20_x, DOUBLE dV1_E20_y, 
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  static InscribedCircleCenterType InscribedCircleCenterEvV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dE00_x, DOUBLE dE00_y,
							    DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
							    DOUBLE dV20_x, DOUBLE dV20_y,
							    DOUBLE dV2_x, DOUBLE dV2_y,
							    DOUBLE dV21_x, DOUBLE dV21_y);
  static EdgeZoneType EdgeZoneEvV(DOUBLE dE00_x, DOUBLE dE00_y,
				  DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				  DOUBLE dV2_x, DOUBLE dV2_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleEvV_V(DOUBLE dE00_x, DOUBLE dE00_y,
				      DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				      DOUBLE dV2_x, DOUBLE dV2_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleEvV_E(DOUBLE dE00_x, DOUBLE dE00_y,
				      DOUBLE dE01_V1_x, DOUBLE dE01_V1_y, 
				      DOUBLE dV2_x, DOUBLE dV2_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  
  static InscribedCircleCenterType InscribedCircleCenterVEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dV00_x, DOUBLE dV00_y,
							    DOUBLE dV0_x, DOUBLE dV0_y,
							    DOUBLE dV01_x, DOUBLE dV01_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y);
  static InscribedCircleCenterType InscribedCircleCenterVEvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							     DOUBLE dV00_x, DOUBLE dV00_y,
							     DOUBLE dV0_x, DOUBLE dV0_y,
							     DOUBLE dV01_x, DOUBLE dV01_y,
							     DOUBLE dE10_x, DOUBLE dE10_y, 
							     DOUBLE dE11_E20_x, DOUBLE dE11_E20_y,
							     DOUBLE dE21_x, DOUBLE dE21_y);
  static EdgeZoneType EdgeZoneVEE(DOUBLE dV0_x, DOUBLE dV0_y,
				  DOUBLE dE10_x, DOUBLE dE10_y, 
				  DOUBLE dE11_x, DOUBLE dE11_y,
				  DOUBLE dE20_x, DOUBLE dE20_y,
				  DOUBLE dE21_x, DOUBLE dE21_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleVEE_V(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleVEE_E(DOUBLE dV0_x, DOUBLE dV0_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  static EdgeZoneType EdgeZoneVEEAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
					      DOUBLE dE10_x, DOUBLE dE10_y, 
					      DOUBLE dE11_x, DOUBLE dE11_y,
					      DOUBLE dE20_x, DOUBLE dE20_y,
					      DOUBLE dE21_x, DOUBLE dE21_y,
					      DOUBLE dE30_x, DOUBLE dE30_y,
					      DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleVEE_VAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						  DOUBLE dE10_x, DOUBLE dE10_y, 
						  DOUBLE dE11_x, DOUBLE dE11_y,
						  DOUBLE dE20_x, DOUBLE dE20_y,
						  DOUBLE dE21_x, DOUBLE dE21_y,
						  DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleVEE_EAntiparallel(DOUBLE dV0_x, DOUBLE dV0_y,
						  DOUBLE dE10_x, DOUBLE dE10_y, 
						  DOUBLE dE11_x, DOUBLE dE11_y,
						  DOUBLE dE20_x, DOUBLE dE20_y,
						  DOUBLE dE21_x, DOUBLE dE21_y,
						  DOUBLE dE30_x, DOUBLE dE30_y,
						  DOUBLE dE31_x, DOUBLE dE31_y);
  static InscribedCircleCenterType InscribedCircleCentervEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y);
  static InscribedCircleCenterType InscribedCircleCentervEvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							     DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
							     DOUBLE dE11_E20_x, DOUBLE dE11_E20_y,
							     DOUBLE dE21_x, DOUBLE dE21_y);
  static EdgeZoneType EdgeZonevEE(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
				  DOUBLE dE11_x, DOUBLE dE11_y,
				  DOUBLE dE20_x, DOUBLE dE20_y,
				  DOUBLE dE21_x, DOUBLE dE21_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCirclevEE_V(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCirclevEE_E(DOUBLE dv0_E10_x, DOUBLE dV0_E10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  static InscribedCircleCenterType InscribedCircleCenterEEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dE00_x, DOUBLE dE00_y, 
							    DOUBLE dE01_x, DOUBLE dE01_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_V2_x, DOUBLE dE11_V2_y);
  static InscribedCircleCenterType InscribedCircleCenterEvEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							     DOUBLE dE00_x, DOUBLE dE00_y, 
							     DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
							     DOUBLE dE11_V2_x, DOUBLE dE11_V2_y);
  static EdgeZoneType EdgeZoneEEv(DOUBLE dE00_x, DOUBLE dE00_y, 
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dE10_x, DOUBLE dE10_y, 
				  DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleEEv_V(DOUBLE dE00_x, DOUBLE dE00_y, 
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleEEv_E(DOUBLE dE00_x, DOUBLE dE00_y, 
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_V2_x, DOUBLE dE11_V2_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  
  static InscribedCircleCenterType InscribedCircleCenterEEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							    DOUBLE dE00_x, DOUBLE dE00_y, 
							    DOUBLE dE01_x, DOUBLE dE01_y,
							    DOUBLE dE10_x, DOUBLE dE10_y, 
							    DOUBLE dE11_x, DOUBLE dE11_y,
							    DOUBLE dE20_x, DOUBLE dE20_y, 
							    DOUBLE dE21_x, DOUBLE dE21_y);
  static InscribedCircleCenterType InscribedCircleCenterEvEE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							     DOUBLE dE00_x, DOUBLE dE00_y, 
							     DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
							     DOUBLE dE11_x, DOUBLE dE11_y,
							     DOUBLE dE20_x, DOUBLE dE20_y, 
							     DOUBLE dE21_x, DOUBLE dE21_y);
  static InscribedCircleCenterType InscribedCircleCenterEvEvE(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							      DOUBLE dE00_x, DOUBLE dE00_y, 
							      DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
							      DOUBLE dE11_E20_x, DOUBLE dE11_E20_y,
							      DOUBLE dE21_x, DOUBLE dE21_y);
  static InscribedCircleCenterType InscribedCircleCenterEvEvEv(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdR, 
							       DOUBLE dE21_E00_x, DOUBLE dE21_E00_y, 
							       DOUBLE dE01_E10_x, DOUBLE dE01_E10_y,
							       DOUBLE dE11_E20_x, DOUBLE dE11_E20_y);
  static EdgeZoneType EdgeZoneEEE(DOUBLE dE00_x, DOUBLE dE00_y, 
				  DOUBLE dE01_x, DOUBLE dE01_y,
				  DOUBLE dE10_x, DOUBLE dE10_y, 
				  DOUBLE dE11_x, DOUBLE dE11_y,
				  DOUBLE dE20_x, DOUBLE dE20_y,
				  DOUBLE dE21_x, DOUBLE dE21_y,
				  DOUBLE dE30_x, DOUBLE dE30_y,
				  DOUBLE dE31_x, DOUBLE dE31_y);
  static CircleTestType InCircleEEE_V(DOUBLE dE00_x, DOUBLE dE00_y, 
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dV3_x, DOUBLE dV3_y);
  static CircleTestType InCircleEEE_E(DOUBLE dE00_x, DOUBLE dE00_y, 
				      DOUBLE dE01_x, DOUBLE dE01_y,
				      DOUBLE dE10_x, DOUBLE dE10_y, 
				      DOUBLE dE11_x, DOUBLE dE11_y,
				      DOUBLE dE20_x, DOUBLE dE20_y,
				      DOUBLE dE21_x, DOUBLE dE21_y,
				      DOUBLE dE30_x, DOUBLE dE30_y,
				      DOUBLE dE31_x, DOUBLE dE31_y);
  
//////////////////////////////////////////////////
// InCircle Functions
//
public:
  // result > 0  InCircle
  // result = 0  OnCircle
  // result < 0  OutCircle
  static CircleTestType OrientationCircle(DOUBLE dV0_x, DOUBLE dV0_y,
					  DOUBLE dV1_x, DOUBLE dV1_y,
					  DOUBLE dV2_x, DOUBLE dV2_y,
					  DOUBLE dV3_x, DOUBLE dV3_y);
  static DOUBLE OrientationCircleValue(DOUBLE dV0_x, DOUBLE dV0_y,
				       DOUBLE dV1_x, DOUBLE dV1_y,
				       DOUBLE dV2_x, DOUBLE dV2_y,
				       DOUBLE dV3_x, DOUBLE dV3_y);
  
  static inline BOOL InCircle(DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV3_x, DOUBLE dV3_y);
  static inline BOOL OnCircle(DOUBLE dV0_x, DOUBLE dV0_y,
				DOUBLE dV1_x, DOUBLE dV1_y,
				DOUBLE dV2_x, DOUBLE dV2_y,
				DOUBLE dV3_x, DOUBLE dV3_y);
  static inline BOOL OutCircle(DOUBLE dV0_x, DOUBLE dV0_y,
				 DOUBLE dV1_x, DOUBLE dV1_y,
				 DOUBLE dV2_x, DOUBLE dV2_y,
				 DOUBLE dV3_x, DOUBLE dV3_y);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:
  static VOID DisplayMPFUNFilterCounters(FILE *fpOutput);

//////////////////////////////////////////////////
// Class Variables
//
public:
  static UINT_32 m_uiEdgeZoneMPFUNFilter;
  static UINT_32 m_uiEdgeZoneMPFUN;

  static UINT_32 m_uiDistanceCompareV_VVMPFUNFilter;
  static UINT_32 m_uiDistanceCompareV_VVMPFUN;

  static UINT_32 m_uiDistanceCompareV_VLMPFUNFilter;
  static UINT_32 m_uiDistanceCompareV_VLMPFUN;

  static UINT_32 m_uiDistanceCompareV_LLMPFUNFilter;
  static UINT_32 m_uiDistanceCompareV_LLMPFUN;

  static UINT_32 m_uiDistanceCompareEvv_VVMPFUNFilter;
  static UINT_32 m_uiDistanceCompareEvv_VVMPFUN;

  static UINT_32 m_uiDistanceCompareEvv_VLMPFUNFilter;
  static UINT_32 m_uiDistanceCompareEvv_VLMPFUN;

  static UINT_32 m_uiDistanceCompareEvv_LLMPFUNFilter;
  static UINT_32 m_uiDistanceCompareEvv_LLMPFUN;

  static UINT_32 m_uiDistanceCompareElv_VVMPFUNFilter;
  static UINT_32 m_uiDistanceCompareElv_VVMPFUN;

  static UINT_32 m_uiDistanceCompareElv_VLMPFUNFilter;
  static UINT_32 m_uiDistanceCompareElv_VLMPFUN;

  static UINT_32 m_uiDistanceCompareL_VVMPFUNFilter;
  static UINT_32 m_uiDistanceCompareL_VVMPFUN;

  static UINT_32 m_uiOrientationExEMPFUNFilter;
  static UINT_32 m_uiOrientationExEMPFUN;

  static UINT_32 m_uiOrientationEdotEMPFUNFilter;
  static UINT_32 m_uiOrientationEdotEMPFUN;

  static UINT_32 m_uiInscribedCircleCenterVVVMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterVVVMPFUN;
  static UINT_32 m_uiInscribedCircleCenterVVVMPFUNFinal;

  static UINT_32 m_uiEdgeZoneVVVMPFUNFilter;
  static UINT_32 m_uiEdgeZoneVVVMPFUN;

  static UINT_32 m_uiInCircleVVV_VMPFUNFilter;
  static UINT_32 m_uiInCircleVVV_VMPFUN;

  static UINT_32 m_uiInCircleVVV_EMPFUNFilter;
  static UINT_32 m_uiInCircleVVV_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterVVEMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterVVEMPFUN;
  static UINT_32 m_uiInscribedCircleCenterVVEMPFUNFinal;

  static UINT_32 m_uiEdgeZoneVVEMPFUNFilter;
  static UINT_32 m_uiEdgeZoneVVEMPFUN;

  static UINT_32 m_uiInCircleVVE_VMPFUNFilter;
  static UINT_32 m_uiInCircleVVE_VMPFUN;

  static UINT_32 m_uiInCircleVVE_EMPFUNFilter;
  static UINT_32 m_uiInCircleVVE_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterVVEAntiparallelMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterVVEAntiparallelMPFUN;
  static UINT_32 m_uiInscribedCircleCenterVVEAntiparallelMPFUNFinal;

  static UINT_32 m_uiEdgeZoneVVEAntiparallelMPFUNFilter;
  static UINT_32 m_uiEdgeZoneVVEAntiparallelMPFUN;

  static UINT_32 m_uiInCircleVVE_VAntiparallelMPFUNFilter;
  static UINT_32 m_uiInCircleVVE_VAntiparallelMPFUN;

  static UINT_32 m_uiInCircleVVE_EAntiparallelMPFUNFilter;
  static UINT_32 m_uiInCircleVVE_EAntiparallelMPFUN;

  static UINT_32 m_uiInscribedCircleCenterVvEMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterVvEMPFUN;
  static UINT_32 m_uiInscribedCircleCenterVvEMPFUNFinal;

  static UINT_32 m_uiEdgeZoneVvEMPFUNFilter;
  static UINT_32 m_uiEdgeZoneVvEMPFUN;

  static UINT_32 m_uiInCircleVvE_VMPFUNFilter;
  static UINT_32 m_uiInCircleVvE_VMPFUN;

  static UINT_32 m_uiInCircleVvE_EMPFUNFilter;
  static UINT_32 m_uiInCircleVvE_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterEvVMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterEvVMPFUN;
  static UINT_32 m_uiInscribedCircleCenterEvVMPFUNFinal;

  static UINT_32 m_uiEdgeZoneEvVMPFUNFilter;
  static UINT_32 m_uiEdgeZoneEvVMPFUN;

  static UINT_32 m_uiInCircleEvV_VMPFUNFilter;
  static UINT_32 m_uiInCircleEvV_VMPFUN;

  static UINT_32 m_uiInCircleEvV_EMPFUNFilter;
  static UINT_32 m_uiInCircleEvV_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterVEEMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterVEEMPFUN;
  static UINT_32 m_uiInscribedCircleCenterVEEMPFUNFinal;

  static UINT_32 m_uiEdgeZoneVEEMPFUNFilter;
  static UINT_32 m_uiEdgeZoneVEEMPFUN;

  static UINT_32 m_uiInCircleVEE_VMPFUNFilter;
  static UINT_32 m_uiInCircleVEE_VMPFUN;

  static UINT_32 m_uiInCircleVEE_EMPFUNFilter;
  static UINT_32 m_uiInCircleVEE_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterVEEAntiparallelMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterVEEAntiparallelMPFUN;
  static UINT_32 m_uiInscribedCircleCenterVEEAntiparallelMPFUNFinal;

  static UINT_32 m_uiEdgeZoneVEEAntiparallelMPFUNFilter;
  static UINT_32 m_uiEdgeZoneVEEAntiparallelMPFUN;

  static UINT_32 m_uiInCircleVEE_VAntiparallelMPFUNFilter;
  static UINT_32 m_uiInCircleVEE_VAntiparallelMPFUN;

  static UINT_32 m_uiInCircleVEE_EAntiparallelMPFUNFilter;
  static UINT_32 m_uiInCircleVEE_EAntiparallelMPFUN;

  static UINT_32 m_uiInscribedCircleCentervEEMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCentervEEMPFUN;
  static UINT_32 m_uiInscribedCircleCentervEEMPFUNFinal;

  static UINT_32 m_uiEdgeZonevEEMPFUNFilter;
  static UINT_32 m_uiEdgeZonevEEMPFUN;

  static UINT_32 m_uiInCirclevEE_VMPFUNFilter;
  static UINT_32 m_uiInCirclevEE_VMPFUN;

  static UINT_32 m_uiInCirclevEE_EMPFUNFilter;
  static UINT_32 m_uiInCirclevEE_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterEEvMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterEEvMPFUN;
  static UINT_32 m_uiInscribedCircleCenterEEvMPFUNFinal;

  static UINT_32 m_uiEdgeZoneEEvMPFUNFilter;
  static UINT_32 m_uiEdgeZoneEEvMPFUN;

  static UINT_32 m_uiInCircleEEv_VMPFUNFilter;
  static UINT_32 m_uiInCircleEEv_VMPFUN;

  static UINT_32 m_uiInCircleEEv_EMPFUNFilter;
  static UINT_32 m_uiInCircleEEv_EMPFUN;

  static UINT_32 m_uiInscribedCircleCenterEEEMPFUNFilter;
  static UINT_32 m_uiInscribedCircleCenterEEEMPFUN;
  static UINT_32 m_uiInscribedCircleCenterEEEMPFUNFinal;

  static UINT_32 m_uiEdgeZoneEEEMPFUNFilter;
  static UINT_32 m_uiEdgeZoneEEEMPFUN;

  static UINT_32 m_uiInCircleEEE_VMPFUNFilter;
  static UINT_32 m_uiInCircleEEE_VMPFUN;

  static UINT_32 m_uiInCircleEEE_EMPFUNFilter;
  static UINT_32 m_uiInCircleEEE_EMPFUN;

//////////////////////////////////////////////////
// Member Variables
//
protected:
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline BOOL APM2D::Counterclockwise(DOUBLE dV0_x, DOUBLE dV0_y,
				    DOUBLE dV1_x, DOUBLE dV1_y,
				    DOUBLE dV2_x, DOUBLE dV2_y)
{
  return (OrientationVVV(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) > 0);
}

inline BOOL APM2D::Colinear(DOUBLE dV0_x, DOUBLE dV0_y,
			    DOUBLE dV1_x, DOUBLE dV1_y,
			    DOUBLE dV2_x, DOUBLE dV2_y)
{
  return (OrientationVVV(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) == 0);
}

inline BOOL APM2D::Clockwise(DOUBLE dV0_x, DOUBLE dV0_y,
			     DOUBLE dV1_x, DOUBLE dV1_y,
			     DOUBLE dV2_x, DOUBLE dV2_y)
{
  return (OrientationVVV(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y) < 0);
}

inline BOOL APM2D::Parallel(DOUBLE dE00_x, DOUBLE dE00_y,
			    DOUBLE dE01_x, DOUBLE dE01_y,
			    DOUBLE dE10_x, DOUBLE dE10_y,
			    DOUBLE dE11_x, DOUBLE dE11_y)
{
  return (OrientationExE(dE00_x, dE00_y, dE01_x, dE01_y, 
			dE10_x, dE10_y, dE11_x, dE11_y) == 0);
}

inline BOOL APM2D::InCircle(DOUBLE dV0_x, DOUBLE dV0_y,
			    DOUBLE dV1_x, DOUBLE dV1_y,
			    DOUBLE dV2_x, DOUBLE dV2_y,
			    DOUBLE dV3_x, DOUBLE dV3_y)
{
  return (OrientationCircle(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y, dV3_x, dV3_y) > 0);
}

inline BOOL APM2D::OnCircle(DOUBLE dV0_x, DOUBLE dV0_y,
			    DOUBLE dV1_x, DOUBLE dV1_y,
			    DOUBLE dV2_x, DOUBLE dV2_y,
			    DOUBLE dV3_x, DOUBLE dV3_y)
{
  return (OrientationCircle(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y, dV3_x, dV3_y) == 0);
}

inline BOOL APM2D::OutCircle(DOUBLE dV0_x, DOUBLE dV0_y,
			     DOUBLE dV1_x, DOUBLE dV1_y,
			     DOUBLE dV2_x, DOUBLE dV2_y,
			     DOUBLE dV3_x, DOUBLE dV3_y)
{
  return (OrientationCircle(dV0_x, dV0_y, dV1_x, dV1_y, dV2_x, dV2_y, dV3_x, dV3_y) < 0);
}

#endif // __APM2D_H__
