#include "apm2d_adaptive.h"

#include "predicates.h"

//////////////////////////////////////////////////
// START: Stuff from predicates.c
//

/* On some machines, the exact arithmetic routines might be defeated by the  */
/*   use of internal extended precision floating-point registers.  Sometimes */
/*   this problem can be fixed by defining certain values to be volatile,    */
/*   thus forcing them to be stored to memory and rounded off.  This isn't   */
/*   a great solution, though, as it slows the arithmetic down.              */
/*                                                                           */
/* To try this out, write "#define INEXACT volatile" below.  Normally,       */
/*   however, INEXACT should be defined to be nothing.  ("#define INEXACT".) */

#define INEXACT                          /* Nothing */
/* #define INEXACT volatile */

#define REAL double                      /* float or double */
#define REALPRINT doubleprint
#define REALRAND doublerand
#define NARROWRAND narrowdoublerand
#define UNIFORMRAND uniformdoublerand

/* Which of the following two methods of finding the absolute values is      */
/*   fastest is compiler-dependent.  A few compilers can inline and optimize */
/*   the fabs() call; but most will incur the overhead of a function call,   */
/*   which is disastrously slow.  A faster way on IEEE machines might be to  */
/*   mask the appropriate bit, but that's difficult to do in C.              */

#define Absolute(a)  ((a) >= 0.0 ? (a) : -(a))
/* #define Absolute(a)  fabs(a) */

/* Many of the operations are broken up into two pieces, a main part that    */
/*   performs an approximate operation, and a "tail" that computes the       */
/*   roundoff error of that operation.                                       */
/*                                                                           */
/* The operations Fast_Two_Sum(), Fast_Two_Diff(), Two_Sum(), Two_Diff(),    */
/*   Split(), and Two_Product() are all implemented as described in the      */
/*   reference.  Each of these macros requires certain variables to be       */
/*   defined in the calling routine.  The variables `bvirt', `c', `abig',    */
/*   `_i', `_j', `_k', `_l', `_m', and `_n' are declared `INEXACT' because   */
/*   they store the result of an operation that may incur roundoff error.    */
/*   The input parameter `x' (or the highest numbered `x_' parameter) must   */
/*   also be declared `INEXACT'.                                             */

#define Fast_Two_Sum_Tail(a, b, x, y) \
  bvirt = x - a; \
  y = b - bvirt

#define Fast_Two_Sum(a, b, x, y) \
  x = (REAL) (a + b); \
  Fast_Two_Sum_Tail(a, b, x, y)

#define Fast_Two_Diff_Tail(a, b, x, y) \
  bvirt = a - x; \
  y = bvirt - b

#define Fast_Two_Diff(a, b, x, y) \
  x = (REAL) (a - b); \
  Fast_Two_Diff_Tail(a, b, x, y)

#define Two_Sum_Tail(a, b, x, y) \
  bvirt = (REAL) (x - a); \
  avirt = x - bvirt; \
  bround = b - bvirt; \
  around = a - avirt; \
  y = around + bround

#define Two_Sum(a, b, x, y) \
  x = (REAL) (a + b); \
  Two_Sum_Tail(a, b, x, y)

#define Two_Diff_Tail(a, b, x, y) \
  bvirt = (REAL) (a - x); \
  avirt = x + bvirt; \
  bround = bvirt - b; \
  around = a - avirt; \
  y = around + bround

#define Two_Diff(a, b, x, y) \
  x = (REAL) (a - b); \
  Two_Diff_Tail(a, b, x, y)

#define Split(a, ahi, alo) \
  c = (REAL) (splitter * a); \
  abig = (REAL) (c - a); \
  ahi = c - abig; \
  alo = a - ahi

#define Two_Product_Tail(a, b, x, y) \
  Split(a, ahi, alo); \
  Split(b, bhi, blo); \
  err1 = x - (ahi * bhi); \
  err2 = err1 - (alo * bhi); \
  err3 = err2 - (ahi * blo); \
  y = (alo * blo) - err3

#define Two_Product(a, b, x, y) \
  x = (REAL) (a * b); \
  Two_Product_Tail(a, b, x, y)

/* Two_Product_Presplit() is Two_Product() where one of the inputs has       */
/*   already been split.  Avoids redundant splitting.                        */

#define Two_Product_Presplit(a, b, bhi, blo, x, y) \
  x = (REAL) (a * b); \
  Split(a, ahi, alo); \
  err1 = x - (ahi * bhi); \
  err2 = err1 - (alo * bhi); \
  err3 = err2 - (ahi * blo); \
  y = (alo * blo) - err3

/* Two_Product_2Presplit() is Two_Product() where both of the inputs have    */
/*   already been split.  Avoids redundant splitting.                        */

#define Two_Product_2Presplit(a, ahi, alo, b, bhi, blo, x, y) \
  x = (REAL) (a * b); \
  err1 = x - (ahi * bhi); \
  err2 = err1 - (alo * bhi); \
  err3 = err2 - (ahi * blo); \
  y = (alo * blo) - err3

/* Square() can be done more quickly than Two_Product().                     */

#define Square_Tail(a, x, y) \
  Split(a, ahi, alo); \
  err1 = x - (ahi * ahi); \
  err3 = err1 - ((ahi + ahi) * alo); \
  y = (alo * alo) - err3

#define Square(a, x, y) \
  x = (REAL) (a * a); \
  Square_Tail(a, x, y)

/* Macros for summing expansions of various fixed lengths.  These are all    */
/*   unrolled versions of Expansion_Sum().                                   */

#define Two_One_Sum(a1, a0, b, x2, x1, x0) \
  Two_Sum(a0, b , _i, x0); \
  Two_Sum(a1, _i, x2, x1)

#define Two_One_Diff(a1, a0, b, x2, x1, x0) \
  Two_Diff(a0, b , _i, x0); \
  Two_Sum( a1, _i, x2, x1)

#define Two_Two_Sum(a1, a0, b1, b0, x3, x2, x1, x0) \
  Two_One_Sum(a1, a0, b0, _j, _0, x0); \
  Two_One_Sum(_j, _0, b1, x3, x2, x1)

#define Two_Two_Diff(a1, a0, b1, b0, x3, x2, x1, x0) \
  Two_One_Diff(a1, a0, b0, _j, _0, x0); \
  Two_One_Diff(_j, _0, b1, x3, x2, x1)

#define Four_One_Sum(a3, a2, a1, a0, b, x4, x3, x2, x1, x0) \
  Two_One_Sum(a1, a0, b , _j, x1, x0); \
  Two_One_Sum(a3, a2, _j, x4, x3, x2)

#define Four_Two_Sum(a3, a2, a1, a0, b1, b0, x5, x4, x3, x2, x1, x0) \
  Four_One_Sum(a3, a2, a1, a0, b0, _k, _2, _1, _0, x0); \
  Four_One_Sum(_k, _2, _1, _0, b1, x5, x4, x3, x2, x1)

#define Four_Four_Sum(a3, a2, a1, a0, b4, b3, b1, b0, x7, x6, x5, x4, x3, x2, \
                      x1, x0) \
  Four_Two_Sum(a3, a2, a1, a0, b1, b0, _l, _2, _1, _0, x1, x0); \
  Four_Two_Sum(_l, _2, _1, _0, b4, b3, x7, x6, x5, x4, x3, x2)

#define Eight_One_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b, x8, x7, x6, x5, x4, \
                      x3, x2, x1, x0) \
  Four_One_Sum(a3, a2, a1, a0, b , _j, x3, x2, x1, x0); \
  Four_One_Sum(a7, a6, a5, a4, _j, x8, x7, x6, x5, x4)

#define Eight_Two_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b1, b0, x9, x8, x7, \
                      x6, x5, x4, x3, x2, x1, x0) \
  Eight_One_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b0, _k, _6, _5, _4, _3, _2, \
                _1, _0, x0); \
  Eight_One_Sum(_k, _6, _5, _4, _3, _2, _1, _0, b1, x9, x8, x7, x6, x5, x4, \
                x3, x2, x1)

#define Eight_Four_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b4, b3, b1, b0, x11, \
                       x10, x9, x8, x7, x6, x5, x4, x3, x2, x1, x0) \
  Eight_Two_Sum(a7, a6, a5, a4, a3, a2, a1, a0, b1, b0, _l, _6, _5, _4, _3, \
                _2, _1, _0, x1, x0); \
  Eight_Two_Sum(_l, _6, _5, _4, _3, _2, _1, _0, b4, b3, x11, x10, x9, x8, \
                x7, x6, x5, x4, x3, x2)

/* Macros for multiplying expansions of various fixed lengths.               */

#define Two_One_Product(a1, a0, b, x3, x2, x1, x0) \
  Split(b, bhi, blo); \
  Two_Product_Presplit(a0, b, bhi, blo, _i, x0); \
  Two_Product_Presplit(a1, b, bhi, blo, _j, _0); \
  Two_Sum(_i, _0, _k, x1); \
  Fast_Two_Sum(_j, _k, x3, x2)

#define Four_One_Product(a3, a2, a1, a0, b, x7, x6, x5, x4, x3, x2, x1, x0) \
  Split(b, bhi, blo); \
  Two_Product_Presplit(a0, b, bhi, blo, _i, x0); \
  Two_Product_Presplit(a1, b, bhi, blo, _j, _0); \
  Two_Sum(_i, _0, _k, x1); \
  Fast_Two_Sum(_j, _k, _i, x2); \
  Two_Product_Presplit(a2, b, bhi, blo, _j, _0); \
  Two_Sum(_i, _0, _k, x3); \
  Fast_Two_Sum(_j, _k, _i, x4); \
  Two_Product_Presplit(a3, b, bhi, blo, _j, _0); \
  Two_Sum(_i, _0, _k, x5); \
  Fast_Two_Sum(_j, _k, x7, x6)

#define Two_Two_Product(a1, a0, b1, b0, x7, x6, x5, x4, x3, x2, x1, x0) \
  Split(a0, a0hi, a0lo); \
  Split(b0, bhi, blo); \
  Two_Product_2Presplit(a0, a0hi, a0lo, b0, bhi, blo, _i, x0); \
  Split(a1, a1hi, a1lo); \
  Two_Product_2Presplit(a1, a1hi, a1lo, b0, bhi, blo, _j, _0); \
  Two_Sum(_i, _0, _k, _1); \
  Fast_Two_Sum(_j, _k, _l, _2); \
  Split(b1, bhi, blo); \
  Two_Product_2Presplit(a0, a0hi, a0lo, b1, bhi, blo, _i, _0); \
  Two_Sum(_1, _0, _k, x1); \
  Two_Sum(_2, _k, _j, _1); \
  Two_Sum(_l, _j, _m, _2); \
  Two_Product_2Presplit(a1, a1hi, a1lo, b1, bhi, blo, _j, _0); \
  Two_Sum(_i, _0, _n, _0); \
  Two_Sum(_1, _0, _i, x2); \
  Two_Sum(_2, _i, _k, _1); \
  Two_Sum(_m, _k, _l, _2); \
  Two_Sum(_j, _n, _k, _0); \
  Two_Sum(_1, _0, _j, x3); \
  Two_Sum(_2, _j, _i, _1); \
  Two_Sum(_l, _i, _m, _2); \
  Two_Sum(_1, _k, _i, x4); \
  Two_Sum(_2, _i, _k, x5); \
  Two_Sum(_m, _k, x7, x6)

/* An expansion of length two can be squared more quickly than finding the   */
/*   product of two different expansions of length two, and the result is    */
/*   guaranteed to have no more than six (rather than eight) components.     */

#define Two_Square(a1, a0, x5, x4, x3, x2, x1, x0) \
  Square(a0, _j, x0); \
  _0 = a0 + a0; \
  Two_Product(a1, _0, _k, _1); \
  Two_One_Sum(_k, _1, _j, _l, _2, x1); \
  Square(a1, _j, _1); \
  Two_Two_Sum(_j, _1, _l, _2, x5, x4, x3, x2)

extern "C" {

extern DOUBLE splitter;
extern DOUBLE resulterrbound;
extern DOUBLE ccwerrboundA;
extern DOUBLE ccwerrboundB;
extern DOUBLE ccwerrboundC;

extern DOUBLE estimate(int elen, DOUBLE *e);
extern int fast_expansion_sum(int elen, DOUBLE *e, int flen, DOUBLE *f, DOUBLE *h);
extern int fast_expansion_sum_zeroelim(int elen, DOUBLE *e, int flen, DOUBLE *f, DOUBLE *h);

extern int grow_expansion(int elen, DOUBLE *e, DOUBLE b, DOUBLE *h);
extern int grow_expansion_zeroelim(int elen, DOUBLE *e, DOUBLE b, DOUBLE *h);

extern int scale_expansion(int elen, DOUBLE *e, DOUBLE b, DOUBLE *h);
extern int scale_expansion_zeroelim(int elen, DOUBLE *e, DOUBLE b, DOUBLE *h);
}

//
// END  : Stuff from predicates.c
//////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// APM2D Adaptive Implementation


//////////////////////////////////////////////////
// expansion_product_zeroelim
//   pdTemp and pdResult must have size = 2*(iA * iB)
//
static 
int expansion_product_zeroelim(int iA, DOUBLE *pdA, int iB, DOUBLE *pdB, 
			       DOUBLE *pdTemp, 
			       DOUBLE *pdResult)
{
  int i, j;
  DOUBLE *pd;
  DOUBLE pdProduct[2];
  int iTemp, iResult;

  //////////////////////////////////////////////////
  // Variables created by macro-calls
  //INEXACT DOUBLE bvirt;
  //DOUBLE avirt, bround, around;
  INEXACT DOUBLE c;
  INEXACT DOUBLE abig;
  DOUBLE ahi, alo, bhi, blo;
  DOUBLE err1, err2, err3;
  //INEXACT DOUBLE _i, _j;
  //DOUBLE _0;
  // Variables created by macro-calls
  //////////////////////////////////////////////////

  if ( iA > iB )
    {
      // swap operands
      i = iA;
      iA = iB;
      iB = i;

      pd = pdA;
      pdA = pdB;
      pdB = pd;
    }
  
  iTemp = 1;
  iResult = 1;
  pdResult[0] = 0.0;
  for ( i = 0; i < iA; i++ )
    {
      for ( j = 0; j < iB; j++ )
	{
	  Two_Product(pdA[i], pdB[j], pdProduct[1], pdProduct[0]);
	  iTemp = grow_expansion_zeroelim(iResult, 
					  pdResult,
					  pdProduct[0],
					  pdTemp);
	  iResult = grow_expansion_zeroelim(iTemp, 
					    pdTemp,
					    pdProduct[1],
					    pdResult);
	}
    }

  return iResult;
}

//////////////////////////////////////////////////
// expansion_square_zeroelim
//   pdTemp and pdResult must have size = (iA * (iA + 1))
//
static 
int expansion_square_zeroelim(int iA, DOUBLE *pdA, 
			      DOUBLE *pdTemp, 
			      DOUBLE *pdResult)
{
  int i, j;
  DOUBLE pdProduct[2];
  int iTemp, iResult;

  //////////////////////////////////////////////////
  // Variables created by macro-calls
  //INEXACT DOUBLE bvirt;
  //DOUBLE avirt, bround, around;
  INEXACT DOUBLE c;
  INEXACT DOUBLE abig;
  DOUBLE ahi, alo, bhi, blo;
  DOUBLE err1, err2, err3;
  //INEXACT DOUBLE _i, _j;
  //DOUBLE _0;
  // Variables created by macro-calls
  //////////////////////////////////////////////////

  iTemp = 1;
  iResult = 1;
  pdResult[0] = 0.0;
  for ( i = 0; i < iA; i++ )
    {
      Square(pdA[i], pdProduct[1], pdProduct[0]);
      iTemp = grow_expansion_zeroelim(iResult, 
				      pdResult,
				      pdProduct[0],
				      pdTemp);
      iResult = grow_expansion_zeroelim(iTemp, 
					pdTemp,
					pdProduct[1],
					pdResult);
      for ( j = i+1; j < iA; j++ )
	{
	  Two_Product(pdA[i], pdA[j], pdProduct[1], pdProduct[0]);
	  pdProduct[0] = 2*pdProduct[0];
	  pdProduct[1] = 2*pdProduct[1];
	  iTemp = grow_expansion_zeroelim(iResult, 
					  pdResult,
					  pdProduct[0],
					  pdTemp);
	  iResult = grow_expansion_zeroelim(iTemp, 
					    pdTemp,
					    pdProduct[1],
					    pdResult);
	}
    }

  return iResult;
}

// Adaptive State Classes

//////////////////////////////
// Cross Product Calculation
class ASExE {
public:
  ASExE(DOUBLE dE00_x, DOUBLE dE00_y,
	DOUBLE dE01_x, DOUBLE dE01_y,
	DOUBLE dE10_x, DOUBLE dE10_y,
	DOUBLE dE11_x, DOUBLE dE11_y);
  ~ASExE();

public:
  VOID Compute(int iApproxMin = 1, int iApproxMax = 4);

public:
  int m_iApprox;
  DOUBLE m_dE00_x; DOUBLE m_dE00_y;
  DOUBLE m_dE01_x; DOUBLE m_dE01_y;
  DOUBLE m_dE10_x; DOUBLE m_dE10_y;
  DOUBLE m_dE11_x; DOUBLE m_dE11_y;
  DOUBLE m_ppdE0[2][2];
  DOUBLE m_ppdE1[2][2];
  DOUBLE m_ppdE0xE1Left[4][2];
  DOUBLE m_ppdE0xE1Right[4][2];
  DOUBLE m_pdE0xE1B[4];
  DOUBLE m_pdE0xE1[16]; int m_iE0xE1;
  DOUBLE m_dE0xE1; DOUBLE m_dE0xE1Mag;
};

ASExE::ASExE(DOUBLE dE00_x, DOUBLE dE00_y,
	     DOUBLE dE01_x, DOUBLE dE01_y,
	     DOUBLE dE10_x, DOUBLE dE10_y,
	     DOUBLE dE11_x, DOUBLE dE11_y)
{
  m_iApprox = 0;
  m_dE00_x = dE00_x;
  m_dE00_y = dE00_y;
  m_dE01_x = dE01_x;
  m_dE01_y = dE01_y;
  m_dE10_x = dE10_x;
  m_dE10_y = dE10_y;
  m_dE11_x = dE11_x;
  m_dE11_y = dE11_y;
  m_iE0xE1 = 0;
}

ASExE::~ASExE()
{
}

VOID ASExE::Compute(int iApproxMin/* = 1*/, int iApproxMax/* = 4*/)
{
  DOUBLE pdU[4];
  DOUBLE pdC1[8];
  DOUBLE pdC2[12];
  DOUBLE dErrorBound;
  int iC1, iC2;

  //////////////////////////////////////////////////
  // Variables created by macro-calls
  INEXACT DOUBLE bvirt;
  DOUBLE avirt, bround, around;
  INEXACT DOUBLE c;
  INEXACT DOUBLE abig;
  DOUBLE ahi, alo, bhi, blo;
  DOUBLE err1, err2, err3;
  INEXACT DOUBLE _i, _j;
  DOUBLE _0;
  // Variables created by macro-calls
  //////////////////////////////////////////////////

  switch ( m_iApprox )
    {
    case 0:
      m_iApprox = 1;

      m_ppdE0[0][1] = m_dE01_x - m_dE00_x;
      m_ppdE0[1][1] = m_dE01_y - m_dE00_y;
      m_ppdE1[0][1] = m_dE11_x - m_dE10_x;
      m_ppdE1[1][1] = m_dE11_y - m_dE10_y;

      m_ppdE0xE1Left[3][1] = m_ppdE0[0][1]*m_ppdE1[1][1];
      m_ppdE0xE1Right[3][1] = m_ppdE0[1][1]*m_ppdE1[0][1];
      m_dE0xE1 = m_ppdE0xE1Left[3][1] - m_ppdE0xE1Right[3][1];
      
      m_dE0xE1Mag = 0.0;
      if ( m_ppdE0xE1Left[3][1] > 0.0 )
	{
	  if ( m_ppdE0xE1Right[3][1] <= 0.0 )
	    {
	      m_dE0xE1Mag = m_ppdE0xE1Left[3][1] - m_ppdE0xE1Right[3][1];
	      if ( m_iApprox >= iApproxMin )
		{
		  return;
		}
	    }  
	  else
	    {
	      m_dE0xE1Mag = m_ppdE0xE1Left[3][1] + m_ppdE0xE1Right[3][1];
	    }
	}
      else if ( m_ppdE0xE1Left[3][1] < 0.0 )
	{
	  if ( m_ppdE0xE1Right[3][1] >= 0.0 )
	    {
	      m_dE0xE1Mag = -m_ppdE0xE1Left[3][1] + m_ppdE0xE1Right[3][1];
	      if ( m_iApprox >= iApproxMin )
		{
		  return;
		}
	    }
	  else
	    {
	      m_dE0xE1Mag = -m_ppdE0xE1Left[3][1] - m_ppdE0xE1Right[3][1];
	    }
	}
      else
	{
	  if ( m_ppdE0xE1Right[3][1] >= 0.0 )
	    {
	      m_dE0xE1Mag = m_ppdE0xE1Right[3][1];
	    }
	  else
	    {
	      m_dE0xE1Mag = -m_ppdE0xE1Right[3][1];
	    }
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}
      
      dErrorBound = ccwerrboundA * m_dE0xE1Mag;
      if ( (m_dE0xE1 >= dErrorBound) || (-m_dE0xE1 >= dErrorBound) ) 
	{
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}
    case 1:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 2;

      Two_Product(m_ppdE0[0][1], m_ppdE1[1][1], m_ppdE0xE1Left[3][1], m_ppdE0xE1Left[3][0]);
      Two_Product(m_ppdE0[1][1], m_ppdE1[0][1], m_ppdE0xE1Right[3][1], m_ppdE0xE1Right[3][0]);
      Two_Two_Diff(m_ppdE0xE1Left[3][1], m_ppdE0xE1Left[3][0], m_ppdE0xE1Right[3][1], m_ppdE0xE1Right[3][0],
		   m_pdE0xE1B[3], m_pdE0xE1B[2], m_pdE0xE1B[1], m_pdE0xE1B[0]);
      
      m_dE0xE1 = estimate(4, m_pdE0xE1B);
      dErrorBound = ccwerrboundB * m_dE0xE1Mag;
      if ( (m_dE0xE1 >= dErrorBound) || (-m_dE0xE1 >= dErrorBound) ) 
	{
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}

    case 2:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 3;

      Two_Diff_Tail(m_dE01_x, m_dE00_x, m_ppdE0[0][1], m_ppdE0[0][0]);
      Two_Diff_Tail(m_dE01_y, m_dE00_y, m_ppdE0[1][1], m_ppdE0[1][0]);
      Two_Diff_Tail(m_dE11_x, m_dE10_x, m_ppdE1[0][1], m_ppdE1[0][0]);
      Two_Diff_Tail(m_dE11_y, m_dE10_y, m_ppdE1[1][1], m_ppdE1[1][0]);
      
      if ( (m_ppdE0[0][0] == 0.0) && (m_ppdE0[1][0] == 0.0) &&
	   (m_ppdE1[0][0] == 0.0) && (m_ppdE1[1][0] == 0.0) ) 
	{
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}
	  
      dErrorBound = ccwerrboundC * m_dE0xE1Mag + resulterrbound * Absolute(m_dE0xE1);
      m_dE0xE1 += ( (m_ppdE0[0][1] * m_ppdE1[1][0] + m_ppdE1[1][1] * m_ppdE0[0][0])
		  - (m_ppdE0[1][1] * m_ppdE1[0][0] + m_ppdE1[0][1] * m_ppdE0[1][0]) );
      if ((m_dE0xE1 >= dErrorBound) || (-m_dE0xE1 >= dErrorBound)) {
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
      }

    case 3:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 4;

      Two_Product(m_ppdE0[0][0], m_ppdE1[1][1], m_ppdE0xE1Left[2][1], m_ppdE0xE1Left[2][0]);
      Two_Product(m_ppdE0[1][0], m_ppdE1[0][1], m_ppdE0xE1Right[2][1], m_ppdE0xE1Right[2][0]);
      Two_Two_Diff(m_ppdE0xE1Left[2][1], m_ppdE0xE1Left[2][0], m_ppdE0xE1Right[2][1], m_ppdE0xE1Right[2][0],
		   pdU[3], pdU[2], pdU[1], pdU[0]);
      iC1 = fast_expansion_sum_zeroelim(4, m_pdE0xE1B, 4, pdU, pdC1);
      
      Two_Product(m_ppdE0[0][1], m_ppdE1[1][0], m_ppdE0xE1Left[1][1], m_ppdE0xE1Left[1][0]);
      Two_Product(m_ppdE0[1][1], m_ppdE1[0][0], m_ppdE0xE1Right[1][1], m_ppdE0xE1Right[1][0]);
      Two_Two_Diff(m_ppdE0xE1Left[1][1], m_ppdE0xE1Left[1][0], m_ppdE0xE1Right[1][1], m_ppdE0xE1Right[1][0],
		   pdU[3], pdU[2], pdU[1], pdU[0]);
      iC2 = fast_expansion_sum_zeroelim(iC1, pdC1, 4, pdU, pdC2);
      
      Two_Product(m_ppdE0[0][0], m_ppdE1[1][0], m_ppdE0xE1Left[0][1], m_ppdE0xE1Left[0][0]);
      Two_Product(m_ppdE0[1][0], m_ppdE1[0][0], m_ppdE0xE1Right[0][1], m_ppdE0xE1Right[0][0]);
      Two_Two_Diff(m_ppdE0xE1Left[0][1], m_ppdE0xE1Left[0][0], m_ppdE0xE1Right[0][1], m_ppdE0xE1Right[0][0],
		   pdU[3], pdU[2], pdU[1], pdU[0]);
      m_iE0xE1 = fast_expansion_sum_zeroelim(iC2, pdC2, 4, pdU, m_pdE0xE1);
      
      m_dE0xE1 = m_pdE0xE1[m_iE0xE1 - 1];

    case 4:
      break;
    default:
      ASSERT( FALSE );
    }
}

//////////////////////////////
// Dot Product Calculation
class ASEdotE {
public:
  ASEdotE(DOUBLE dE00_x, DOUBLE dE00_y,
	DOUBLE dE01_x, DOUBLE dE01_y,
	DOUBLE dE10_x, DOUBLE dE10_y,
	DOUBLE dE11_x, DOUBLE dE11_y);
  ~ASEdotE();

public:
  VOID Compute(int iApproxMin = 1, int iApproxMax = 4);

public:
  int m_iApprox;
  DOUBLE m_dE00_x; DOUBLE m_dE00_y;
  DOUBLE m_dE01_x; DOUBLE m_dE01_y;
  DOUBLE m_dE10_x; DOUBLE m_dE10_y;
  DOUBLE m_dE11_x; DOUBLE m_dE11_y;
  DOUBLE m_ppdE0[2][2];
  DOUBLE m_ppdE1[2][2];
  DOUBLE m_ppdE0dotE1Left[4][2];
  DOUBLE m_ppdE0dotE1Right[4][2];
  DOUBLE m_pdE0dotE1B[4];
  DOUBLE m_pdE0dotE1[16]; int m_iE0dotE1;
  DOUBLE m_dE0dotE1; DOUBLE m_dE0dotE1Mag;
};

ASEdotE::ASEdotE(DOUBLE dE00_x, DOUBLE dE00_y,
		 DOUBLE dE01_x, DOUBLE dE01_y,
		 DOUBLE dE10_x, DOUBLE dE10_y,
		 DOUBLE dE11_x, DOUBLE dE11_y)
{
  m_iApprox = 0;
  m_dE00_x = dE00_x;
  m_dE00_y = dE00_y;
  m_dE01_x = dE01_x;
  m_dE01_y = dE01_y;
  m_dE10_x = dE10_x;
  m_dE10_y = dE10_y;
  m_dE11_x = dE11_x;
  m_dE11_y = dE11_y;
  m_iE0dotE1 = 0;
}

ASEdotE::~ASEdotE()
{
}

//static
VOID ASEdotE::Compute(int iApproxMin/* = 1*/, int iApproxMax/* = 4*/)
{
  DOUBLE pdU[4];
  DOUBLE pdC1[8];
  DOUBLE pdC2[12];
  DOUBLE dErrorBound;
  int iC1, iC2;

  //////////////////////////////////////////////////
  // Variables created by macro-calls
  INEXACT DOUBLE bvirt;
  DOUBLE avirt, bround, around;
  INEXACT DOUBLE c;
  INEXACT DOUBLE abig;
  DOUBLE ahi, alo, bhi, blo;
  DOUBLE err1, err2, err3;
  INEXACT DOUBLE _i, _j;
  DOUBLE _0;
  // Variables created by macro-calls
  //////////////////////////////////////////////////

  switch ( m_iApprox )
    {
    case 0:
      m_iApprox = 1;

      m_ppdE0[0][1] = m_dE01_x - m_dE00_x;
      m_ppdE0[1][1] = m_dE01_y - m_dE00_y;
      m_ppdE1[0][1] = m_dE11_x - m_dE10_x;
      m_ppdE1[1][1] = m_dE11_y - m_dE10_y;

      m_ppdE0dotE1Left[3][1] = m_ppdE0[0][1]*m_ppdE1[0][1];
      m_ppdE0dotE1Right[3][1] = m_ppdE0[1][1]*m_ppdE1[1][1];
      m_dE0dotE1 = m_ppdE0dotE1Left[3][1] + m_ppdE0dotE1Right[3][1];
      
      m_dE0dotE1Mag = 0.0;
      if ( m_ppdE0dotE1Left[3][1] > 0.0 )
	{
	  if ( m_ppdE0dotE1Right[3][1] >= 0.0 )
	    {
	      m_dE0dotE1Mag = m_ppdE0dotE1Left[3][1] + m_ppdE0dotE1Right[3][1];
	      if ( m_iApprox >= iApproxMin )
		{
		  return;
		}
	    }  
	  else
	    {
	      m_dE0dotE1Mag = m_ppdE0dotE1Left[3][1] - m_ppdE0dotE1Right[3][1];
	    }
	}
      else if ( m_ppdE0dotE1Left[3][1] < 0.0 )
	{
	  if ( m_ppdE0dotE1Right[3][1] <= 0.0 )
	    {
	      m_dE0dotE1Mag = -m_ppdE0dotE1Left[3][1] - m_ppdE0dotE1Right[3][1];
	      if ( m_iApprox >= iApproxMin )
		{
		  return;
		}
	    }
	  else
	    {
	      m_dE0dotE1Mag = -m_ppdE0dotE1Left[3][1] + m_ppdE0dotE1Right[3][1];
	    }
	}
      else
	{
	  if ( m_ppdE0dotE1Right[3][1] >= 0.0 )
	    {
	      m_dE0dotE1Mag = m_ppdE0dotE1Right[3][1];
	    }
	  else
	    {
	      m_dE0dotE1Mag = -m_ppdE0dotE1Right[3][1];
	    }
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}
      
      dErrorBound = ccwerrboundA * m_dE0dotE1Mag;
      if ( (m_dE0dotE1 >= dErrorBound) || (-m_dE0dotE1 >= dErrorBound) ) 
	{
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}
    case 1:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 2;

      Two_Product(m_ppdE0[0][1], m_ppdE1[0][1], m_ppdE0dotE1Left[3][1], m_ppdE0dotE1Left[3][0]);
      Two_Product(m_ppdE0[1][1], m_ppdE1[1][1], m_ppdE0dotE1Right[3][1], m_ppdE0dotE1Right[3][0]);
      Two_Two_Sum(m_ppdE0dotE1Left[3][1], m_ppdE0dotE1Left[3][0], m_ppdE0dotE1Right[3][1], m_ppdE0dotE1Right[3][0],
		   m_pdE0dotE1B[3], m_pdE0dotE1B[2], m_pdE0dotE1B[1], m_pdE0dotE1B[0]);
      
      m_dE0dotE1 = estimate(4, m_pdE0dotE1B);
      dErrorBound = ccwerrboundB * m_dE0dotE1Mag;
      if ( (m_dE0dotE1 >= dErrorBound) || (-m_dE0dotE1 >= dErrorBound) ) 
	{
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}

    case 2:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 3;

      Two_Diff_Tail(m_dE01_x, m_dE00_x, m_ppdE0[0][1], m_ppdE0[0][0]);
      Two_Diff_Tail(m_dE01_y, m_dE00_y, m_ppdE0[1][1], m_ppdE0[1][0]);
      Two_Diff_Tail(m_dE11_x, m_dE10_x, m_ppdE1[0][1], m_ppdE1[0][0]);
      Two_Diff_Tail(m_dE11_y, m_dE10_y, m_ppdE1[1][1], m_ppdE1[1][0]);
      
      if ( (m_ppdE0[0][0] == 0.0) && (m_ppdE0[1][0] == 0.0) &&
	   (m_ppdE1[0][0] == 0.0) && (m_ppdE1[1][0] == 0.0) ) 
	{
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
	}
	  
      dErrorBound = ccwerrboundC * m_dE0dotE1Mag + resulterrbound * Absolute(m_dE0dotE1);
      m_dE0dotE1 += (m_ppdE0[0][1] * m_ppdE1[0][0] + m_ppdE1[0][1] * m_ppdE0[0][0])
	+ (m_ppdE0[1][1] * m_ppdE1[1][0] + m_ppdE1[1][1] * m_ppdE0[1][0]);
      if ((m_dE0dotE1 >= dErrorBound) || (-m_dE0dotE1 >= dErrorBound)) {
	  if ( m_iApprox >= iApproxMin )
	    {
	      return;
	    }
      }

    case 3:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 4;

      Two_Product(m_ppdE0[0][0], m_ppdE1[1][1], m_ppdE0dotE1Left[2][1], m_ppdE0dotE1Left[2][0]);
      Two_Product(m_ppdE0[1][0], m_ppdE1[0][1], m_ppdE0dotE1Right[2][1], m_ppdE0dotE1Right[2][0]);
      Two_Two_Sum(m_ppdE0dotE1Left[2][1], m_ppdE0dotE1Left[2][0], m_ppdE0dotE1Right[2][1], m_ppdE0dotE1Right[2][0],
		   pdU[3], pdU[2], pdU[1], pdU[0]);
      iC1 = fast_expansion_sum_zeroelim(4, m_pdE0dotE1B, 4, pdU, pdC1);
      
      Two_Product(m_ppdE0[0][1], m_ppdE1[1][0], m_ppdE0dotE1Left[1][1], m_ppdE0dotE1Left[1][0]);
      Two_Product(m_ppdE0[1][1], m_ppdE1[0][0], m_ppdE0dotE1Right[1][1], m_ppdE0dotE1Right[1][0]);
      Two_Two_Sum(m_ppdE0dotE1Left[1][1], m_ppdE0dotE1Left[1][0], m_ppdE0dotE1Right[1][1], m_ppdE0dotE1Right[1][0],
		   pdU[3], pdU[2], pdU[1], pdU[0]);
      iC2 = fast_expansion_sum_zeroelim(iC1, pdC1, 4, pdU, pdC2);
      
      Two_Product(m_ppdE0[0][0], m_ppdE1[1][0], m_ppdE0dotE1Left[0][1], m_ppdE0dotE1Left[0][0]);
      Two_Product(m_ppdE0[1][0], m_ppdE1[0][0], m_ppdE0dotE1Right[0][1], m_ppdE0dotE1Right[0][0]);
      Two_Two_Sum(m_ppdE0dotE1Left[0][1], m_ppdE0dotE1Left[0][0], m_ppdE0dotE1Right[0][1], m_ppdE0dotE1Right[0][0],
		   pdU[3], pdU[2], pdU[1], pdU[0]);
      m_iE0dotE1 = fast_expansion_sum_zeroelim(iC2, pdC2, 4, pdU, m_pdE0dotE1);
      
      m_dE0dotE1 = m_pdE0dotE1[m_iE0dotE1 - 1];

    case 4:
      break;
    default:
      ASSERT( FALSE );
    }
}

//////////////////////////////
// Edge Magnitude2 Calculation
class ASEmag2 {
public:
  ASEmag2(DOUBLE dE00_x, DOUBLE dE00_y,
	  DOUBLE dE01_x, DOUBLE dE01_y);
  ~ASEmag2();

public:
  VOID Compute(int iApproxMin = 1, int iApproxMax = 2);

public:
  int m_iApprox;
  DOUBLE m_dE00_x; DOUBLE m_dE00_y;
  DOUBLE m_dE01_x; DOUBLE m_dE01_y;
  DOUBLE m_ppdE0[2][2];
  DOUBLE m_ppdE0mag2Left[3][2];
  DOUBLE m_ppdE0mag2Right[3][2];
  DOUBLE m_pdE0mag2B[4];
  DOUBLE m_pdE0mag2[12]; int m_iE0mag2;
  DOUBLE m_dE0mag2; DOUBLE m_dE0mag2Mag;
};

ASEmag2::ASEmag2(DOUBLE dE00_x, DOUBLE dE00_y,
		 DOUBLE dE01_x, DOUBLE dE01_y)
{
  m_iApprox = 0;
  m_dE00_x = dE00_x;
  m_dE00_y = dE00_y;
  m_dE01_x = dE01_x;
  m_dE01_y = dE01_y;
  m_iE0mag2 = 0;
}

ASEmag2::~ASEmag2()
{
}

//static
VOID ASEmag2::Compute(int iApproxMin/* = 1*/, int iApproxMax/* = 2*/)
{
  DOUBLE pdU[4];
  DOUBLE pdC[8];
  int iC;

  //////////////////////////////////////////////////
  // Variables created by macro-calls
  INEXACT DOUBLE bvirt;
  DOUBLE avirt, bround, around;
  INEXACT DOUBLE c;
  INEXACT DOUBLE abig;
  DOUBLE ahi, alo, bhi, blo;
  DOUBLE err1, err2, err3;
  INEXACT DOUBLE _i, _j, _k, _l;
  DOUBLE _0, _1, _2;
  // Variables created by macro-calls
  //////////////////////////////////////////////////

  switch ( m_iApprox )
    {
    case 0:
      m_iApprox = 1;

      m_ppdE0[0][1] = m_dE01_x - m_dE00_x;
      m_ppdE0[1][1] = m_dE01_y - m_dE00_y;

      m_ppdE0mag2Left[2][1] = m_ppdE0[0][1]*m_ppdE0[0][1];
      m_ppdE0mag2Right[2][1] = m_ppdE0[1][1]*m_ppdE0[1][1];

      ASSERT( m_ppdE0mag2Left[2][1] >= 0.0 );
      ASSERT( m_ppdE0mag2Right[2][1] >= 0.0 );

      m_dE0mag2 = m_ppdE0mag2Left[2][1] + m_ppdE0mag2Right[2][1];
      m_dE0mag2Mag = m_dE0mag2;
      if ( m_iApprox >= iApproxMin )
	{
	  return;
	}

    case 1:
      if ( m_iApprox >= iApproxMax )
	{
	  return;
	}
      m_iApprox = 2;

      Two_Diff_Tail(m_dE01_x, m_dE00_x, m_ppdE0[0][1], m_ppdE0[0][0]);
      Two_Diff_Tail(m_dE01_y, m_dE00_y, m_ppdE0[1][1], m_ppdE0[1][0]);
      
      //if ( (m_ppdE0[0][0] == 0.0) && (m_ppdE0[1][0] == 0.0) )
	//{
	  //Square_Tail(m_ppdE0[0][1], m_ppdE0mag2Left[2][1], m_ppdE0mag2Left[2][0]);
	  //Square_Tail(m_ppdE0[1][1], m_ppdE0mag2Right[2][1], m_ppdE0mag2Right[2][0]);
	  //Two_Two_Sum(m_ppdE0mag2Left[2][1], m_ppdE0mag2Left[2][0], 
		      //m_ppdE0mag2Right[2][1], m_ppdE0mag2Right[2][0],
		      //m_pdE0mag2B[3], m_pdE0mag2B[2], m_pdE0mag2B[1], m_pdE0mag2B[0]);
	  //m_iE0mag2 = grow_expansion_zeroelim(4, m_pdE0mag2B, 0.0, m_pdE0mag2);
	//}
      //else
	//{
	  Two_Square(m_ppdE0[0][1], m_ppdE0[0][0], 
		     m_ppdE0mag2Left[2][1], m_ppdE0mag2Left[2][0],
		     m_ppdE0mag2Left[1][1], m_ppdE0mag2Left[1][0], 
		     m_ppdE0mag2Left[0][1], m_ppdE0mag2Left[0][0]);
	  Two_Square(m_ppdE0[1][1], m_ppdE0[1][0], 
		     m_ppdE0mag2Right[2][1], m_ppdE0mag2Right[2][0],
		     m_ppdE0mag2Right[1][1], m_ppdE0mag2Right[1][0], 
		     m_ppdE0mag2Right[0][1], m_ppdE0mag2Right[0][0]);
	  
	  Two_Two_Sum(m_ppdE0mag2Left[2][1], m_ppdE0mag2Left[2][0], 
		      m_ppdE0mag2Right[2][1], m_ppdE0mag2Right[2][0],
		      m_pdE0mag2B[3], m_pdE0mag2B[2], m_pdE0mag2B[1], m_pdE0mag2B[0]);
	  
	  Two_Two_Sum(m_ppdE0mag2Left[1][1], m_ppdE0mag2Left[1][0], 
		      m_ppdE0mag2Right[1][1], m_ppdE0mag2Right[1][0],
		      pdU[3], pdU[2], pdU[1], pdU[0]);
	  iC = fast_expansion_sum_zeroelim(4, m_pdE0mag2B, 4, pdU, pdC);
	  
	  Two_Two_Sum(m_ppdE0mag2Left[0][1], m_ppdE0mag2Left[0][0], 
		      m_ppdE0mag2Right[0][1], m_ppdE0mag2Right[0][0],
		      pdU[3], pdU[2], pdU[1], pdU[0]);
	  m_iE0mag2 = fast_expansion_sum_zeroelim(iC, pdC, 4, pdU, m_pdE0mag2);
        //}
      m_dE0mag2 = m_pdE0mag2[m_iE0mag2 - 1];
    case 4:
      break;
    default:
      ASSERT( FALSE );
    }
}

//static
BOOL IntersectionEEAdaptive(DOUBLE *pdV_x, DOUBLE *pdV_y,
			    DOUBLE dE00_x, DOUBLE dE00_y, 
			    DOUBLE dE01_x, DOUBLE dE01_y,
			    DOUBLE dE10_x, DOUBLE dE10_y, 
			    DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASSERT( pdV_x != NULL && pdV_y != NULL );
  // Set to zero for the case of no intersection
  *pdV_x = 0.0;
  *pdV_y = 0.0;
  
  //////////////////////////////
  // Parallel or colinear check
  DOUBLE dE0xE1;
  dE0xE1 = OrientationExEAdaptive(dE00_x, dE00_y, 
				  dE01_x, dE01_y,
				  dE10_x, dE10_y, 
				  dE11_x, dE11_y);
  if ( dE0xE1 == 0.0 )
    {
      // Parallel or colinear edges
      return FALSE;
    }
  BOOL bPositive;
  bPositive = (dE0xE1 > 0.0);

  //////////////////////////////
  // Check I vs E00 half plane
  DOUBLE dE10xE11_E00;
  dE10xE11_E00 = OrientationExEAdaptive(dE00_x, dE00_y, 
					dE10_x, dE10_y,
					dE00_x, dE00_y, 
					dE11_x, dE11_y);
  if ( !bPositive )
    {
      if ( 0.0 < dE10xE11_E00 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 > dE10xE11_E00 )
	{
	  return FALSE;
	}
    }

  //////////////////////////////
  // Check I vs E01 half plane
  DOUBLE dE10xE11_E01;
  dE10xE11_E01 = OrientationExEAdaptive(dE01_x, dE01_y, 
					dE10_x, dE10_y,
					dE01_x, dE01_y, 
					dE11_x, dE11_y);
  if ( !bPositive )
    {
      if ( 0.0 > dE10xE11_E01 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 < dE10xE11_E01 )
	{
	  return FALSE;
	}
    }

  //////////////////////////////
  // Check I vs E10 half plane
  DOUBLE dE00xE01_E10;
  dE00xE01_E10 = OrientationExEAdaptive(dE10_x, dE10_y, 
					dE00_x, dE00_y,
					dE10_x, dE10_y, 
					dE01_x, dE01_y);
  if ( bPositive )
    {
      if ( 0.0 < dE00xE01_E10 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 > dE00xE01_E10 )
	{
	  return FALSE;
	}
    }

  //////////////////////////////
  // Check I vs E11 half plane
  DOUBLE dE00xE01_E11;
  dE00xE01_E11 = OrientationExEAdaptive(dE11_x, dE11_y, 
					dE00_x, dE00_y,
					dE11_x, dE11_y, 
					dE01_x, dE01_y);
  if ( bPositive )
    {
      if ( 0.0 > dE00xE01_E11 )
	{
	  return FALSE;
	}
    }
  else
    {
      if ( 0.0 < dE00xE01_E11 )
	{
	  return FALSE;
	}
    }

  *pdV_x = dE00_x + dE10xE11_E00*(dE01_x - dE00_x)/dE0xE1;
  *pdV_y = dE00_y + dE10xE11_E00*(dE01_y - dE00_y)/dE0xE1;
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Distance Comparison Functions

//static
APM2D::EdgeZoneType EdgeZoneAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
				     DOUBLE dE10_x, DOUBLE dE10_y, 
				     DOUBLE dE11_x, DOUBLE dE11_y)
{
  APM2D::EdgeZoneType eResult;
  DOUBLE d;

  eResult = APM2D::eztMiddle;

  d = OrientationEdotEAdaptive(dV0_x, dV0_y, dE10_x, dE10_y, 
			       dE10_x, dE10_y, dE11_x, dE11_y);
  if ( d > 0.0 )
    {
      eResult = APM2D::eztStart;
    }
  else
    {
      d = OrientationEdotEAdaptive(dV0_x, dV0_y, dE11_x, dE11_y, 
				   dE10_x, dE10_y, dE11_x, dE11_y);
      if ( d < 0.0 )
	{
	  eResult = APM2D::eztEnd;
	}
      else
	{
	  eResult = APM2D::eztMiddle;
	}
    }

  return eResult;
}

//static 
INT_32 DistanceCompareV_VVAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
				   DOUBLE dV1_x, DOUBLE dV1_y, 
				   DOUBLE dV2_x, DOUBLE dV2_y)
{
  INT_32 iResult;
  DOUBLE pdNegV2dotV2[16];
  DOUBLE pdDiff[32];
  int iDiff, i;  
  DOUBLE dDiff, dDiffMag;
  DOUBLE dErrorBound;

  iResult = 0;

  ASEdotE asV1dotV1(dV0_x, dV0_y,
		    dV1_x, dV1_y, 
		    dV0_x, dV0_y,
		    dV1_x, dV1_y);
  ASEdotE asV2dotV2(dV0_x, dV0_y,
		    dV2_x, dV2_y, 
		    dV0_x, dV0_y,
		    dV2_x, dV2_y);
  
  //////////////////////////////
  // Floating point filter
  asV1dotV1.Compute(1, 1);
  asV2dotV2.Compute(1, 1);

  dDiff = asV1dotV1.m_dE0dotE1 - asV2dotV2.m_dE0dotE1;
  dDiffMag = asV1dotV1.m_dE0dotE1Mag + asV2dotV2.m_dE0dotE1Mag;

  // I DON'T KNOW THE REAL ERROR BOUND YET
  dErrorBound = 2 * ccwerrboundA * dDiffMag;
  if ( (dDiff >= dErrorBound) || (-dDiff >= dErrorBound) )  
    {
      goto end;
    }

  //////////////////////////////
  // Force the complete evaluation of the dot products
  asV1dotV1.Compute(4, 4);
  asV2dotV2.Compute(4, 4);

  // V1dotV1 - V2dotV2
  for ( i = 0; i < asV2dotV2.m_iE0dotE1; i++ )
    {
      pdNegV2dotV2[i] = -asV2dotV2.m_pdE0dotE1[i];
    }
  iDiff = fast_expansion_sum_zeroelim(asV1dotV1.m_iE0dotE1, asV1dotV1.m_pdE0dotE1, 
				      asV2dotV2.m_iE0dotE1, pdNegV2dotV2, 
				      pdDiff);
  dDiff = pdDiff[iDiff - 1];
  
 end:
  if ( dDiff == 0.0 )
    {
      iResult = 0;
    }
  else if ( dDiff > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( dDiff < 0.0 );
      iResult = -1;
    }

  return iResult;
}

//#define USE_OLD_VERSION_DistanceCompareV_VLAdaptive
#ifdef USE_OLD_VERSION_DistanceCompareV_VLAdaptive

//static 
INT_32 DistanceCompareV_VLAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
				   DOUBLE dV1_x, DOUBLE dV1_y, 
				   DOUBLE dE20_x, DOUBLE dE20_y, 
				   DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  DOUBLE dErrorBound;
  int i;  

  iResult = 0;

  int iLeft;
  DOUBLE pdTempLeft[512];
  DOUBLE pdLeft[512];
  DOUBLE dLeft, dLeftMag;

  int iRight;
  DOUBLE pdNegRight[512];
  DOUBLE pdRight[512];
  DOUBLE dRight, dRightMag;

  int iDet;  
  DOUBLE pdDet[1024];
  DOUBLE dDet, dDetMag;

  ASEdotE asV1dotV1(dV0_x, dV0_y,
		    dV1_x, dV1_y, 
		    dV0_x, dV0_y,
		    dV1_x, dV1_y);
  ASEdotE asE2dotE2(dE20_x, dE20_y,
		    dE21_x, dE21_y, 
		    dE20_x, dE20_y,
		    dE21_x, dE21_y);
  ASExE asE20xE21(dV0_x, dV0_y,
		  dE20_x, dE20_y, 
		  dV0_x, dV0_y,
		  dE21_x, dE21_y);
  
  //////////////////////////////
  // Floating point filter
  //   Det = V1dotV1*E2dotE2 - E20xE21*E20xE21

  asV1dotV1.Compute(1, 1);
  asE2dotE2.Compute(1, 1);
  asE20xE21.Compute(1, 1);

  dLeft = asV1dotV1.m_dE0dotE1 * asE2dotE2.m_dE0dotE1;
  dLeftMag = asV1dotV1.m_dE0dotE1Mag * asE2dotE2.m_dE0dotE1Mag;

  dRight = asE20xE21.m_dE0xE1 * asE20xE21.m_dE0xE1;
  dRightMag = asE20xE21.m_dE0xE1Mag * asE20xE21.m_dE0xE1Mag;

  dDet = dLeft - dRight;
  dDetMag = dLeftMag + dRightMag;

  // I DON'T KNOW THE REAL ERROR BOUND YET
  dErrorBound = 4 * ccwerrboundA * dDetMag;
  if ( (dDet >= dErrorBound) || (-dDet >= dErrorBound) )  
    {
      goto end;
    }

  //////////////////////////////
  // Force the complete evaluation of the products
  //   Det = V1dotV1*E2dotE2 - E20xE21*E20xE21
  asV1dotV1.Compute(4, 4);
  asE2dotE2.Compute(4, 4);
  asE20xE21.Compute(4, 4);

  // Left  = V1dotV1 * E2dotE2
  iLeft = expansion_product_zeroelim(asV1dotV1.m_iE0dotE1, asV1dotV1.m_pdE0dotE1, 
				     asE2dotE2.m_iE0dotE1, asE2dotE2.m_pdE0dotE1, 
				     pdTempLeft, pdLeft);
  
  // Right = E20xE21 * E20xE21
  iRight = expansion_product_zeroelim(asE20xE21.m_iE0xE1, asE20xE21.m_pdE0xE1, 
				      asE20xE21.m_iE0xE1, asE20xE21.m_pdE0xE1, 
				      pdNegRight, pdRight);

  // Det = V1dotV1*E2dotE2 - E20xE21*E20xE21
  for ( i = 0; i < iRight; i++ )
    {
      pdNegRight[i] = -pdRight[i];
    }
  iDet = fast_expansion_sum_zeroelim(iLeft, pdLeft, 
				     iRight, pdNegRight, 
				     pdDet);
  dDet = pdDet[iDet - 1];
  
 end:
  if ( dDet == 0.0 )
    {
      iResult = 0;
    }
  else if ( dDet > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( dDet < 0.0 );
      iResult = -1;
    }

  return iResult;
}

#else // USE_OLD_VERSION_DistanceCompareV_VLAdaptive

//static 
INT_32 DistanceCompareV_VLAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
				   DOUBLE dV1_x, DOUBLE dV1_y, 
				   DOUBLE dE20_x, DOUBLE dE20_y, 
				   DOUBLE dE21_x, DOUBLE dE21_y)
{
  INT_32 iResult;
  DOUBLE dErrorBound;
  int i;  

  iResult = 0;

  const int ciLeftMax = 2*(12*12);
  int iLeft;
  DOUBLE pdTempLeft[ciLeftMax];
  DOUBLE pdLeft[ciLeftMax];
  DOUBLE dLeft, dLeftMag;

  const int ciRightMax = (16*(16+1));
  int iRight;
  DOUBLE pdNegRight[ciRightMax];
  DOUBLE pdRight[ciRightMax];
  DOUBLE dRight, dRightMag;

  const int ciDetMax = ciLeftMax + ciRightMax;
  int iDet;  
  DOUBLE pdDet[ciDetMax];
  DOUBLE dDet, dDetMag;

  ASEmag2 asV1mag2(dV0_x, dV0_y,
		   dV1_x, dV1_y);
  ASEmag2 asE2mag2(dE20_x, dE20_y,
		   dE21_x, dE21_y);
  ASExE asE20xE21(dV0_x, dV0_y,
		  dE20_x, dE20_y, 
		  dV0_x, dV0_y,
		  dE21_x, dE21_y);
  
  //////////////////////////////
  // Floating point filter
  //   Det = V1mag2*E2mag2 - E20xE21*E20xE21

  asV1mag2.Compute(1, 1);
  asE2mag2.Compute(1, 1);
  asE20xE21.Compute(1, 1);

  dLeft = asV1mag2.m_dE0mag2 * asE2mag2.m_dE0mag2;
  dLeftMag = asV1mag2.m_dE0mag2Mag * asE2mag2.m_dE0mag2Mag;

  dRight = asE20xE21.m_dE0xE1 * asE20xE21.m_dE0xE1;
  dRightMag = asE20xE21.m_dE0xE1Mag * asE20xE21.m_dE0xE1Mag;

  dDet = dLeft - dRight;
  dDetMag = dLeftMag + dRightMag;

  // I DON'T KNOW THE REAL ERROR BOUND YET
  dErrorBound = 4 * ccwerrboundA * dDetMag;
  if ( (dDet >= dErrorBound) || (-dDet >= dErrorBound) )  
    {
      goto end;
    }

  //////////////////////////////
  // Force the complete evaluation of the products
  //   Det = V1mag2*E2mag2 - E20xE21*E20xE21
  asV1mag2.Compute(2, 2);
  asE2mag2.Compute(2, 2);
  asE20xE21.Compute(4, 4);

  // Left  = V1mag2 * E2mag2
  iLeft = expansion_product_zeroelim(asV1mag2.m_iE0mag2, asV1mag2.m_pdE0mag2, 
				     asE2mag2.m_iE0mag2, asE2mag2.m_pdE0mag2, 
				     pdTempLeft, pdLeft);
  
  // Right = E20xE21 * E20xE21
  iRight = expansion_square_zeroelim(asE20xE21.m_iE0xE1, asE20xE21.m_pdE0xE1, 
				     pdNegRight, pdRight);

  // Det = V1mag2*E2mag2 - E20xE21*E20xE21
  for ( i = 0; i < iRight; i++ )
    {
      pdNegRight[i] = -pdRight[i];
    }
  iDet = fast_expansion_sum_zeroelim(iLeft, pdLeft, 
				     iRight, pdNegRight, 
				     pdDet);
  dDet = pdDet[iDet - 1];
  
 end:
  if ( dDet == 0.0 )
    {
      iResult = 0;
    }
  else if ( dDet > 0.0 )
    {
      iResult = 1;
    }
  else
    {
      ASSERT( dDet < 0.0 );
      iResult = -1;
    }

  return iResult;
}

#endif // USE_OLD_VERSION_DistanceCompareV_VLAdaptive

//////////////////////////////////////////////////////////////////////
// EE Orientaion Functions

//static
DOUBLE OrientationExEAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
			      DOUBLE dE01_x, DOUBLE dE01_y,
			      DOUBLE dE10_x, DOUBLE dE10_y,
			      DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASExE asE0xE1(dE00_x, dE00_y,
		dE01_x, dE01_y,
		dE10_x, dE10_y,
		dE11_x, dE11_y);
  asE0xE1.Compute();
  return asE0xE1.m_dE0xE1;
}
  
//static
DOUBLE OrientationEdotEAdaptive(DOUBLE dE00_x, DOUBLE dE00_y,
				DOUBLE dE01_x, DOUBLE dE01_y,
				DOUBLE dE10_x, DOUBLE dE10_y,
				DOUBLE dE11_x, DOUBLE dE11_y)
{
  ASEdotE asE0dotE1(dE00_x, dE00_y,
		    dE01_x, dE01_y,
		    dE10_x, dE10_y,
		    dE11_x, dE11_y);
  asE0dotE1.Compute();
  return asE0dotE1.m_dE0dotE1;
}

//////////////////////////////////////////////////////////////////////
// VVV Functions

//static
DOUBLE OrientationVVVValueAdaptive(DOUBLE dV0_x, DOUBLE dV0_y,
				   DOUBLE dV1_x, DOUBLE dV1_y,
				   DOUBLE dV2_x, DOUBLE dV2_y)
{
  return OrientationExEAdaptive(dV2_x, dV2_y,
				dV0_x, dV0_y,
			        dV2_x, dV2_y,
			        dV1_x, dV1_y);
}

