#include "apm3d.h"

//////////////////////////////////////////////////
// APM3D Implementation
//

static inline DOUBLE Cross(DOUBLE dV0_x, DOUBLE dV0_y, 
			   DOUBLE dV1_x, DOUBLE dV1_y)
{
  return ( (dV0_x*dV1_y) - (dV0_y*dV1_x) );
}

static inline DOUBLE Dot(DOUBLE dV0_x, DOUBLE dV0_y, 
			 DOUBLE dV1_x, DOUBLE dV1_y)
{
  return ( (dV0_x*dV1_x) + (dV0_y*dV1_y) );
}

static inline DOUBLE Dot(DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z, 
			 DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z)
{
  return ( (dV0_x*dV1_x) + (dV0_y*dV1_y) + (dV0_z*dV1_z) );
}

//////////////////////////////////////////////////
// Inscribed Sphere Center
//

DOUBLE APM3D::OrientationValueVVVV(DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
				   DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
				   DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
				   DOUBLE dV3_x, DOUBLE dV3_y, DOUBLE dV3_z)
{
  DOUBLE ppdV[4][3];
  ppdV[0][0] = 0.0;
  ppdV[0][1] = 0.0;
  ppdV[0][2] = 0.0;
  ppdV[1][0] = dV1_x - dV0_x;
  ppdV[1][1] = dV1_y - dV0_y;
  ppdV[1][2] = dV1_z - dV0_z;
  ppdV[2][0] = dV2_x - dV0_x;
  ppdV[2][1] = dV2_y - dV0_y;
  ppdV[2][2] = dV2_z - dV0_z;
  ppdV[3][0] = dV3_x - dV0_x;
  ppdV[3][1] = dV3_y - dV0_y;
  ppdV[3][2] = dV3_z - dV0_z;

  DOUBLE dDet = ( ( ppdV[1][0]*ppdV[2][1]*ppdV[3][2] +
		    ppdV[1][1]*ppdV[2][2]*ppdV[3][0] +
		    ppdV[1][2]*ppdV[2][0]*ppdV[3][1] ) - 
		  ( ppdV[1][2]*ppdV[2][1]*ppdV[3][0] +
		    ppdV[1][0]*ppdV[2][2]*ppdV[3][1] +
		    ppdV[1][1]*ppdV[2][0]*ppdV[3][2] ) );

  return dDet;
}


//////////////////////////////////////////////////
// Inscribed Sphere Center
//

BOOL APM3D::InscribedCircleCenterVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdC_z, DOUBLE *pdR, 
				     DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
				     DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
				     DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdC_z != NULL && pdR != NULL );

  DOUBLE ppdV[4][3], pdC[3];
  ppdV[0][0] = 0.0;
  ppdV[0][1] = 0.0;
  ppdV[0][2] = 0.0;
  ppdV[1][0] = dV1_x - dV0_x;
  ppdV[1][1] = dV1_y - dV0_y;
  ppdV[1][2] = dV1_z - dV0_z;
  ppdV[2][0] = dV2_x - dV0_x;
  ppdV[2][1] = dV2_y - dV0_y;
  ppdV[2][2] = dV2_z - dV0_z;
  ppdV[3][0] = dV2_x - dV1_x;
  ppdV[3][1] = dV2_y - dV1_y;
  ppdV[3][2] = dV2_z - dV1_z;

  DOUBLE dV1dotV1 = Dot(ppdV[1][0], ppdV[1][1], ppdV[1][2],
			ppdV[1][0], ppdV[1][1], ppdV[1][2]);
  DOUBLE dV2dotV2 = Dot(ppdV[2][0], ppdV[2][1], ppdV[2][2],
			ppdV[2][0], ppdV[2][1], ppdV[2][2]);
  DOUBLE dV0dotV0 = Dot(ppdV[3][0], ppdV[3][1], ppdV[3][2],
			ppdV[3][0], ppdV[3][1], ppdV[3][2]);
  DOUBLE pdE[3];
  pdE[0] = sqrt(dV0dotV0);
  pdE[1] = sqrt(dV1dotV1);
  pdE[2] = sqrt(dV2dotV2);
  DOUBLE pdV1dotV1[3];
  pdV1dotV1[0] = Dot(ppdV[1][1], ppdV[1][2],
		     ppdV[1][1], ppdV[1][2]);
  pdV1dotV1[1] = Dot(ppdV[1][2], ppdV[1][0],
		     ppdV[1][2], ppdV[1][0]);
  pdV1dotV1[2] = Dot(ppdV[1][0], ppdV[1][1],
		     ppdV[1][0], ppdV[1][1]);
  DOUBLE pdV2dotV2[3];
  pdV2dotV2[0] = Dot(ppdV[2][1], ppdV[2][2],
		     ppdV[2][1], ppdV[2][2]);
  pdV2dotV2[1] = Dot(ppdV[2][2], ppdV[2][0],
		     ppdV[2][2], ppdV[2][0]);
  pdV2dotV2[2] = Dot(ppdV[2][0], ppdV[2][1],
		     ppdV[2][0], ppdV[2][1]);
  DOUBLE pdV1dotV2[3];
  pdV1dotV2[0] = Dot(ppdV[1][1], ppdV[1][2],
		     ppdV[2][1], ppdV[2][2]);
  pdV1dotV2[1] = Dot(ppdV[1][2], ppdV[1][0],
		     ppdV[2][2], ppdV[2][0]);
  pdV1dotV2[2] = Dot(ppdV[1][0], ppdV[1][1],
		     ppdV[2][0], ppdV[2][1]);
  DOUBLE pdV1xV2[3];
  pdV1xV2[0] = Cross(ppdV[1][1], ppdV[1][2],
		     ppdV[2][1], ppdV[2][2]);
  pdV1xV2[1] = Cross(ppdV[1][2], ppdV[1][0],
		     ppdV[2][2], ppdV[2][0]);
  pdV1xV2[2] = Cross(ppdV[1][0], ppdV[1][1],
		     ppdV[2][0], ppdV[2][1]);
  DOUBLE dV1xV2dotV1xV2 = Dot(pdV1xV2[0], pdV1xV2[1], pdV1xV2[2],
			      pdV1xV2[0], pdV1xV2[1], pdV1xV2[2]);

  DOUBLE dDen = 2.0*dV1xV2dotV1xV2;
  if ( dDen == 0.0 )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdC_z = 0.0;
      *pdR = INFINITY;
      return FALSE;
    }

  UINT_32 i;
  DOUBLE dR;
  DOUBLE dS, dRHeron;
  dR = 0.0;
  dS = 0.0;
  for ( i = 0; i < 3; i++ )
    {
      pdC[i] = ( (dV1dotV1*pdV2dotV2[i] - dV2dotV2*pdV1dotV2[i])*ppdV[1][i] +
		 (dV2dotV2*pdV1dotV1[i] - dV1dotV1*pdV1dotV2[i])*ppdV[2][i] ) / dDen;
      dR += ( pdC[i] * pdC[i] );
      dS += pdE[i];
    }
  dR = sqrt(dR);
  dS /= 2.0;
  dRHeron = dS*(dS - pdE[0])*(dS - pdE[1])*(dS - pdE[2]);
  if ( dRHeron < 0.0 )
    {
      ASSERT( FALSE );
    }
  dRHeron = sqrt(dRHeron);
  dRHeron = pdE[0]*pdE[1]*pdE[2]/(4.0*dRHeron);
  fprintf(stderr, "  Circle R = %f == %f = R Heron\n", dR, dRHeron);

  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdC_z = dV0_z + pdC[2];
  *pdR = dR;

  return TRUE;
}

//////////////////////////////////////////////////
// Inscribed Sphere Center
//

BOOL APM3D::InscribedSphereCenterVVVV(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdC_z, DOUBLE *pdR, 
				      DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
				      DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
				      DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
				      DOUBLE dV3_x, DOUBLE dV3_y, DOUBLE dV3_z)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdC_z != NULL && pdR != NULL );

  DOUBLE ppdV[4][3], pdC[3];
  ppdV[0][0] = 0.0;
  ppdV[0][1] = 0.0;
  ppdV[0][2] = 0.0;
  ppdV[1][0] = dV1_x - dV0_x;
  ppdV[1][1] = dV1_y - dV0_y;
  ppdV[1][2] = dV1_z - dV0_z;
  ppdV[2][0] = dV2_x - dV0_x;
  ppdV[2][1] = dV2_y - dV0_y;
  ppdV[2][2] = dV2_z - dV0_z;
  ppdV[3][0] = dV3_x - dV0_x;
  ppdV[3][1] = dV3_y - dV0_y;
  ppdV[3][2] = dV3_z - dV0_z;

  DOUBLE dDet = ( ( ppdV[1][0]*ppdV[2][1]*ppdV[3][2] +
		    ppdV[1][1]*ppdV[2][2]*ppdV[3][0] +
		    ppdV[1][2]*ppdV[2][0]*ppdV[3][1] ) - 
		  ( ppdV[1][2]*ppdV[2][1]*ppdV[3][0] +
		    ppdV[1][0]*ppdV[2][2]*ppdV[3][1] +
		    ppdV[1][1]*ppdV[2][0]*ppdV[3][2] ) );
  DOUBLE dDen = 2.0*dDet;
  if ( dDen == 0.0 )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdC_z = 0.0;
      *pdR = INFINITY;
      return FALSE;
    }

  DOUBLE dV1dotV1 = Dot(ppdV[1][0], ppdV[1][1], ppdV[1][2],
			ppdV[1][0], ppdV[1][1], ppdV[1][2]);
  DOUBLE dV2dotV2 = Dot(ppdV[2][0], ppdV[2][1], ppdV[2][2],
			ppdV[2][0], ppdV[2][1], ppdV[2][2]);
  DOUBLE dV3dotV3 = Dot(ppdV[3][0], ppdV[3][1], ppdV[3][2],
			ppdV[3][0], ppdV[3][1], ppdV[3][2]);
  DOUBLE pdV1xV2[3];
  pdV1xV2[0] = Cross(ppdV[1][1], ppdV[1][2],
		     ppdV[2][1], ppdV[2][2]);
  pdV1xV2[1] = Cross(ppdV[1][2], ppdV[1][0],
		     ppdV[2][2], ppdV[2][0]);
  pdV1xV2[2] = Cross(ppdV[1][0], ppdV[1][1],
		     ppdV[2][0], ppdV[2][1]);
  DOUBLE pdV2xV3[3];
  pdV2xV3[0] = Cross(ppdV[2][1], ppdV[2][2],
		     ppdV[3][1], ppdV[3][2]);
  pdV2xV3[1] = Cross(ppdV[2][2], ppdV[2][0],
		     ppdV[3][2], ppdV[3][0]);
  pdV2xV3[2] = Cross(ppdV[2][0], ppdV[2][1],
		     ppdV[3][0], ppdV[3][1]);
  DOUBLE pdV3xV1[3];
  pdV3xV1[0] = Cross(ppdV[3][1], ppdV[3][2],
		     ppdV[1][1], ppdV[1][2]);
  pdV3xV1[1] = Cross(ppdV[3][2], ppdV[3][0],
		     ppdV[1][2], ppdV[1][0]);
  pdV3xV1[2] = Cross(ppdV[3][0], ppdV[3][1],
		     ppdV[1][0], ppdV[1][1]);

  UINT_32 i;
  DOUBLE dR;
  dR = 0.0;
  for ( i = 0; i < 3; i++ )
    {
      pdC[i] = ( dV1dotV1*pdV2xV3[i] +
		 dV2dotV2*pdV3xV1[i] +
		 dV3dotV3*pdV1xV2[i] ) / dDen;
      dR += ( pdC[i] * pdC[i] );
    }
  dR = sqrt(dR);

  *pdC_x = dV0_x + pdC[0];
  *pdC_y = dV0_y + pdC[1];
  *pdC_z = dV0_z + pdC[2];
  *pdR = dR;

  return TRUE;
}

BOOL APM3D::InscribedSphereCenterVVVSA(DOUBLE *pdC_x, DOUBLE *pdC_y, DOUBLE *pdC_z, DOUBLE *pdR, 
				       DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
				       DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
				       DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
				       DOUBLE dSolidAngle)
{
  ASSERT( pdC_x != NULL && pdC_y != NULL && pdC_z != NULL && pdR != NULL );

  if ( dSolidAngle == 0.0 )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdC_z = 0.0;
      *pdR = INFINITY;
      return FALSE;
    }

  BOOL b;
  UINT_32 i, j, k;
  DOUBLE ppdV[4][3], pdC[3];
  ppdV[0][0] = 0.0;
  ppdV[0][1] = 0.0;
  ppdV[0][2] = 0.0;
  ppdV[1][0] = dV1_x - dV0_x;
  ppdV[1][1] = dV1_y - dV0_y;
  ppdV[1][2] = dV1_z - dV0_z;
  ppdV[2][0] = dV2_x - dV0_x;
  ppdV[2][1] = dV2_y - dV0_y;
  ppdV[2][2] = dV2_z - dV0_z;
  ppdV[3][0] = dV2_x - dV1_x;
  ppdV[3][1] = dV2_y - dV1_y;
  ppdV[3][2] = dV2_z - dV1_z;

  DOUBLE pdCircleC[3], dCircleR;
  b = InscribedCircleCenterVVV(&pdCircleC[0], &pdCircleC[1], &pdCircleC[2], &dCircleR,
			       dV0_x, dV0_y, dV0_z,
			       dV1_x, dV1_y, dV1_z,
			       dV2_x, dV2_y, dV2_z);
  if ( !b )
    {
      *pdC_x = 0.0;
      *pdC_y = 0.0;
      *pdC_z = 0.0;
      *pdR = INFINITY;
      return FALSE;
    }
  DOUBLE dCircleR2 = dCircleR*dCircleR;
  DOUBLE dCircleR3 = dCircleR2*dCircleR;

  DOUBLE pdE[3];
  DOUBLE pdE2[3];
  for ( i = 0; i < 3; i++ )
    {
      pdE2[i] = Dot(ppdV[i+1][0], ppdV[i+1][1], ppdV[i+1][2],
		    ppdV[i+1][0], ppdV[i+1][1], ppdV[i+1][2]);
      pdE[i] = sqrt(pdE2[i]);
    }

  DOUBLE dK = cos(PI*(4.0*fabs(dSolidAngle) + 1.0));

  DOUBLE pdDisc[3];
  for ( i = 0; i < 3; i++ )
    {
      pdDisc[i] = 4.0 * dCircleR2 - pdE2[i];
      if ( pdDisc[i] < 0.0 )
	{
	  ASSERT( FALSE );
	  pdDisc[i] = 0.0;
	}
      pdDisc[i] = sqrt(pdDisc[i]);
    }

  /*
  //////////////////////////////////////////////////
  // Compute the coefficients of the cubic equation
  DOUBLE pdCoefs[4];
  pdCoefs[0] = dCircleR3*pdE2[0]*pdE2[1]*pdE2[2]*(dK - 1.0);

  pdCoefs[1] = 0.0;
  for ( i = 0; i < 3; i++ )
    {
      j = ( i + 1 ) % 3;
      k = ( i + 2 ) % 3;
      pdCoefs[1] += pdE[j]*pdE[k]*(pdE2[i]*pdDisc[i] - 2.0*pdE[j]*pdE[k]*dCircleR);
    }
  pdCoefs[1] += -2.0*dCircleR*(pdE2[1]*pdE2[2] + pdE2[2]*pdE2[0] + pdE2[0]*pdE2[1])*dK;
  pdCoefs[1] *= 2.0*dCircleR2;

  pdCoefs[2] = 0.0;
  for ( i = 0; i < 3; i++ )
    {
      j = ( i + 1 ) % 3;
      k = ( i + 2 ) % 3;
      pdCoefs[2] += pdE[j]*pdE[k]*(pdE[j]*pdE[k] + 2.0*dCircleR*pdDisc[i] - pdDisc[j]*pdDisc[k]);
    }
  pdCoefs[2] += 4.0*dCircleR2*(pdE2[0] + pdE2[1] + pdE2[2])*dK;
  pdCoefs[2] *= 4.0*dCircleR;

  pdCoefs[3] = pdDisc[0]*pdDisc[1]*pdDisc[2];
  for ( i = 0; i < 3; i++ )
    {
      j = ( i + 1 ) % 3;
      k = ( i + 2 ) % 3;
      pdCoefs[3] += -pdE[j]*pdE[k]*pdDisc[i];
    }
  pdCoefs[3] += -8.0*dCircleR3*dK;
  pdCoefs[3] *= 8.0;
  */

  //////////////////////////////////////////////////
  // Compute the coefficients of the cubic equation
  DOUBLE pdCoefs[4];
  pdCoefs[0] = pdE2[0]*pdE2[1]*pdE2[2]*(1.0 - dK);

  pdCoefs[1] = 0.0;
  for ( i = 0; i < 3; i++ )
    {
      j = ( i + 1 ) % 3;
      k = ( i + 2 ) % 3;
      pdCoefs[1] += pdE2[j]*pdE2[i];
    }
  pdCoefs[1] *= 2.0*dK;
  for ( i = 0; i < 3; i++ )
    {
      pdCoefs[1] += pdE2[i]*pdE2[i];
    }
  pdCoefs[1] *= 2.0;

  pdCoefs[2] = 0.0;
  for ( i = 0; i < 3; i++ )
    {
      pdCoefs[2] += pdE2[i];
    }
  pdCoefs[2] *= -16.0*(1.0 + dK);

  pdCoefs[3] = 64.0*(1.0 + dK);

  //////////////////////////////////////////////////
  // Solve cubic equation to get radius squared
  DOUBLE pdSphereR2[3], dSphereR;
  UINT_32 uiSphereR2s, uiSphereR;
  
  APM::SolveCubic(pdSphereR2, uiSphereR2s, pdCoefs);

  uiSphereR = 3;
  for ( i = 0; i < uiSphereR2s; i++ )
    {
      dSphereR = pdSphereR2[i];
      if ( dSphereR > 0.0 )
	{
	  uiSphereR = i;
	  dSphereR = sqrt(dSphereR);
	  break;
	}
    }
  if ( uiSphereR == 3 )
    {
      return FALSE;
    }

  //////////////////////////////////////////////////
  // Given a sphere radius generate the sphere center

  DOUBLE pdV1xV2[3];
  pdV1xV2[0] = Cross(ppdV[1][1], ppdV[1][2],
		     ppdV[2][1], ppdV[2][2]);
  pdV1xV2[1] = Cross(ppdV[1][2], ppdV[1][0],
		     ppdV[2][2], ppdV[2][0]);
  pdV1xV2[2] = Cross(ppdV[1][0], ppdV[1][1],
		     ppdV[2][0], ppdV[2][1]);

  DOUBLE dMag = Dot(pdV1xV2[0], pdV1xV2[1], pdV1xV2[2],
		    pdV1xV2[0], pdV1xV2[1], pdV1xV2[2]);
  ASSERT( dMag > 0.0 );
  dMag = sqrt(dMag);
  for ( i = 0; i < 3; i++ )
    {
      pdV1xV2[i] /= dMag;
    }

  fprintf(stderr, "  Circle Center (%f, %f, %f) r = %f R = %f\n", 
	  pdCircleC[0], pdCircleC[1], pdCircleC[2], dCircleR, dSphereR);

  DOUBLE dH = dSphereR*dSphereR - dCircleR2;
  if ( dH < 0.0 )
    {
      ASSERT( FALSE );
      dH = 0.0;
    }
  dH = sqrt(dH);

  if ( dSolidAngle > 0.0 )
    {
      dH = -dH;
    }

  for ( i = 0; i < 3; i++ )
    {
      pdC[i] = pdCircleC[i] + dH*pdV1xV2[i];
    }
  
  *pdC_x = pdC[0];
  *pdC_y = pdC[1];
  *pdC_z = pdC[2];
  *pdR = dSphereR;

  return TRUE;
}

BOOL APM3D::SolidAngle(DOUBLE *pdSolidAngle, 
		       DOUBLE dV0_x, DOUBLE dV0_y, DOUBLE dV0_z,
		       DOUBLE dV1_x, DOUBLE dV1_y, DOUBLE dV1_z,
		       DOUBLE dV2_x, DOUBLE dV2_y, DOUBLE dV2_z,
		       DOUBLE dSphereR)
{
  ASSERT( pdSolidAngle != NULL );

  if ( dSphereR == 0.0 )
    {
      *pdSolidAngle = 0.0;
      return FALSE;
    }

  UINT_32 i, j, k;
  DOUBLE ppdV[4][3];
  ppdV[0][0] = 0.0;
  ppdV[0][1] = 0.0;
  ppdV[0][2] = 0.0;
  ppdV[1][0] = dV1_x - dV0_x;
  ppdV[1][1] = dV1_y - dV0_y;
  ppdV[1][2] = dV1_z - dV0_z;
  ppdV[2][0] = dV2_x - dV0_x;
  ppdV[2][1] = dV2_y - dV0_y;
  ppdV[2][2] = dV2_z - dV0_z;
  ppdV[3][0] = dV2_x - dV1_x;
  ppdV[3][1] = dV2_y - dV1_y;
  ppdV[3][2] = dV2_z - dV1_z;

  DOUBLE pdE[3];
  DOUBLE pdE2[3];
  for ( i = 0; i < 3; i++ )
    {
      pdE2[i] = Dot(ppdV[i+1][0], ppdV[i+1][1], ppdV[i+1][2],
		    ppdV[i+1][0], ppdV[i+1][1], ppdV[i+1][2]);
      pdE[i] = sqrt(pdE2[i]);
    }

  DOUBLE dR = fabs(dSphereR);
  DOUBLE dR2 = dR*dR;
  DOUBLE pdDisc[3];
  for ( i = 0; i < 3; i++ )
    {
      pdDisc[i] = 4.0 * dR2 - pdE2[i];
      if ( pdDisc[i] < 0.0 )
	{
	  ASSERT( FALSE );
	  pdDisc[i] = 0.0;
	}
      pdDisc[i] = sqrt(pdDisc[i]);
    }

  //////////////////////////////////////////////////
  // Compute the cos
  DOUBLE pdCos[3];
  DOUBLE pdAngle[3];
  DOUBLE dSin, dDisc;
  //DOUBLE dDen;

  dDisc = ( dR2*(2.0*(pdE2[0]*pdE2[1] + pdE2[1]*pdE2[2] + pdE2[2]*pdE2[0]) -
		 (pdE2[0]*pdE2[0] + pdE2[1]*pdE2[1] + pdE2[2]*pdE2[2])) -
	    (pdE2[0]*pdE2[1]*pdE2[2]) );
  if ( dDisc < 0.0 )
    {
      ASSERT( FALSE );
      dDisc = 0.0;
    }
  dSin = 2.0*dR*sqrt(dDisc);
  for ( i = 0; i < 3; i++ )
    {
      j = ( i + 1 ) % 3;
      k = ( i + 2 ) % 3;
      /*
      dDen = pdE[j]*pdE[k]*pdDisc[j]*pdDisc[k];
      if ( dDen == 0.0 )
	{
	  ASSERT( FALSE );
	  return FALSE;
	}
      pdAngle[i] = asin(dSin/dDen);
      */
      pdCos[i] = (2.0*dR2*(-pdE2[i] + pdE2[j] + pdE2[k]) - pdE2[j]*pdE2[k]);
      pdAngle[i] = atan2(dSin, pdCos[i]);

      ASSERT( 0.0 <= pdAngle[i] && pdAngle[i] <= PI );
   }

  DOUBLE dSolidAngle = ( pdAngle[0] + pdAngle[1] + pdAngle[2] - PI ) / ( 4.0*PI );

  if ( dSphereR < 0.0 )
    {
      dSolidAngle = -dSolidAngle;
    }
  *pdSolidAngle = dSolidAngle;

  return TRUE;
}

