#include "gif.h"

#define SLOP_SIZE 256
#define DEBUG 0

const char *gcppcGIFTypeTags[GIF_TYPE_MAX] = {"GIF87a", "GIF89a"};

// GIF loading code

// modified to work in windows
#ifdef WIN32
typedef int boolean;
#else
typedef unsigned char BOOL;
#endif

#define NEXTBYTE (*gpbData++)
#define EXTENSION     0x21
#define IMAGESEP      0x2c
#define TRAILER       0x3b
#define INTERLACEMASK 0x40
#define COLORMAPMASK  0x80

FILE *fp;

int BitOffset = 0,		/* Bit Offset of next code */
    XC = 0, YC = 0,		/* Output X and Y coords of current pixel */
    Pass = 0,			/* Used by output routine if interlaced pic */
    OutCount = 0,		/* Decompressor output 'stack count' */
    RWidth, RHeight,		/* screen dimensions */
    Width, Height,		/* image dimensions */
    LeftOfs, TopOfs,		/* image offset */
    BitsPerPixel,		/* Bits per pixel, read from GIF header */
    BytesPerScanline,		/* bytes per scanline in output raster */
    ColorMapSize,		/* number of colors */
    Background,			/* background color */
    CodeSize,			/* Code size, read from GIF header */
    InitCodeSize,		/* Starting code size, used during Clear */
    Code,			/* Value returned by ReadCode */
    MaxCode,			/* limiting value for current code size */
    ClearCode,			/* GIF clear code */
    EOFCode,			/* GIF end-of-information code */
    CurCode, OldCode, InCode,	/* Decompressor variables */
    FirstFree,			/* First free code, generated per GIF spec */
    FreeCode,			/* Decompressor,next free slot in hash table */
    FinChar,			/* Decompressor variable */
    BitMask,			/* AND mask for data size */
    ReadMask,			/* Code AND mask for current code size */
    Misc;                       /* miscellaneous bits (interlace, local cmap)*/


BOOL gbInterlace, gbHasColormap;

BYTE *gpbRawGIF;              // The heap array to hold it, raw
BYTE *gpbRaster;	      // The raster data stream, unblocked
BYTE *pic8;

    /* The hash table used by the decompressor */
int gpiPrefix[4096];
int gpiSuffix[4096];

    /* An output array used by the decompressor */
int gpiOutCode[4097];

int giGIFType = 0;

/*
static int EGApalette[16][3] = {
  {0,0,0},       {0,0,128},     {0,128,0},     {0,128,128}, 
  {128,0,0},     {128,0,128},   {128,128,0},   {200,200,200},
  {100,100,100}, {100,100,255}, {100,255,100}, {100,255,255},
  {255,100,100}, {255,100,255}, {255,255,100}, {255,255,255} };
*/
/*
static BYTE gpbEGApalette[16][3] = 
{
  {0,0,0},       {0,0,128},     {0,128,0},     {0,128,128}, 
  {128,0,0},     {128,0,128},   {128,128,0},   {200,200,200},
  {100,100,100}, {100,100,255}, {100,255,100}, {100,255,255},
  {255,100,100}, {255,100,255}, {255,255,100}, {255,255,255} 
};
*/
  
static ColorEntry gpceEGApalette[16] = 
{
  {  0,  0,  0}, {  0,  0,128}, {  0,128,  0}, {  0,128,128}, 
  {128,  0,  0}, {128,  0,128}, {128,128,  0}, {200,200,200},
  {100,100,100}, {100,100,255}, {100,255,100}, {100,255,255},
  {255,100,100}, {255,100,255}, {255,255,100}, {255,255,255} 
};

static int readImage(BitMap *pBitMap);
static int readCode();
static void doInterlace(int Index);
static int gifError(BitMap *pBitMap, char *st);
static void gifWarning(char *st);

int giFileSize;
char *gpcBaseName;

BYTE *gpbData;

//////////////////////////////////////////////////
// LoadGIF -- Loads and Decompresses a GIF File
//
int LoadGIF(char *pcFile, FILE *fpFile, BitMap *pBitMap)
{
  /* returns '1' if successful */

  register BYTE ch, ch1, *pbOrig;
  register int i, block;
  int iLength, aspect, gotimage;

  assert( pcFile != NULL );
  assert( fpFile != NULL );
  assert( pBitMap != NULL );

  /* initialize variables */
  BitOffset = XC = YC = Pass = OutCount = gotimage = 0;
  gpbRawGIF = gpbRaster = pic8 = NULL;
  giGIFType = 0;

  pBitMap->pvBuffer  = NULL;
  pBitMap->pcComment = NULL;

  gpcBaseName = BaseName(pcFile);

  /* find the size of the file */
  fseek(fpFile, 0L, SEEK_END);
  giFileSize = ftell(fpFile);
  fseek(fpFile, 0L, SEEK_SET);
  
  // the "+SLOP_SIZE" 's are so we can read truncated GIF files without fear of 
  // segmentation violation
  gpbRawGIF = new BYTE[giFileSize+SLOP_SIZE];
  if ( gpbRawGIF == NULL )
    {
      return( gifError(pBitMap, "not enough memory to read gif file") );
    }
  memset(gpbRawGIF, 0, (giFileSize+SLOP_SIZE)*sizeof(BYTE));
  gpbData = gpbRawGIF;
  
  gpbRaster = new BYTE[giFileSize+SLOP_SIZE];
  if ( gpbRaster == NULL )
    {
      return( gifError(pBitMap, "not enough memory to read gif file") );
    }
  memset(gpbRaster, 0, (giFileSize+SLOP_SIZE)*sizeof(BYTE));
  
  if ( fread(gpbData, (size_t)giFileSize, (size_t)1, fpFile) != 1 )
    { 
      return( gifError(pBitMap, "GIF data read failed") );
    }

  pbOrig = gpbData;

  for ( giGIFType = 0; giGIFType < GIF_TYPE_MAX; giGIFType++ )
    {
      iLength = strlen(gcppcGIFTypeTags[giGIFType]);
      if ( strncmp((char*) gpbData, gcppcGIFTypeTags[giGIFType], 
		   (size_t)iLength) ==0 ) 
	{
	  gpbData += iLength;
	  break;
	}
    }
  if ( giGIFType == GIF_TYPE_MAX )
    {
      return( gifError(pBitMap, "not a GIF file"));
    }

  
  // Get variables from the GIF screen descriptor
  
  ch = NEXTBYTE;
  RWidth = ch + 0x100 * NEXTBYTE;	/* screen dimensions... not used. */
  ch = NEXTBYTE;
  RHeight = ch + 0x100 * NEXTBYTE;
  
  ch = NEXTBYTE;
  gbHasColormap = ((ch & COLORMAPMASK) ? TRUE : FALSE);
  
  BitsPerPixel = (ch & 7) + 1;
  ColorMapSize = 1 << BitsPerPixel;
  BitMask = ColorMapSize - 1;
  
  Background = NEXTBYTE;		/* background color... not used. */
  
  aspect = NEXTBYTE;
  if (aspect) {
    if (!giGIFType) return(gifError(pBitMap,"corrupt GIF file (screen descriptor)"));
  }
  
  
  // Read in global colormap.
  
  if ( gbHasColormap )
    {
      for ( i = 0; i < ColorMapSize; i++ )
	{
	  pBitMap->pColorMap[i].pbComp[R] = NEXTBYTE;
	  pBitMap->pColorMap[i].pbComp[G] = NEXTBYTE;
	  pBitMap->pColorMap[i].pbComp[B] = NEXTBYTE;
	}
    }
  else 
    {  
      // no colormap in GIF file 
      // put std EGA palette (repeated 16 times) into colormap, for lack of
      // anything better to do

    /*
      for (i=0; i<256; i++)
        {
	  pBitMap->pColorMap[i].pbComp[R] = gpbEGApalette[i&15][0];
	  pBitMap->pColorMap[i].pbComp[G] = gpbEGApalette[i&15][0];
	  pBitMap->pColorMap[i].pbComp[B] = gpbEGApalette[i&15][0];
	}
    */
      for (i=0; i<256; i++) 
	{
	  pBitMap->pColorMap[i] = gpceEGApalette[i&15];
	}
  }

  /* possible things at this point are:
   *   an application extension block
   *   a comment extension block
   *   an (optional) graphic control extension block
   *       followed by either an image
   *	   or a plaintext extension
   */

  while ( TRUE ) 
    {
      block = NEXTBYTE;
      
      if (DEBUG) fprintf(stderr,"LoadGIF: ");
      
      if ( block == EXTENSION ) // parse extension blocks
	{  
	  int i, fn, blocksize, aspnum, aspden;
	
	  /* read extension block */
	  fn = NEXTBYTE;
	  
	  if (DEBUG) fprintf(stderr,"GIF extension type 0x%02x\n", fn);
	  
	  if (fn == 'R')                  /* GIF87 aspect extension */
	    {
	      int sbsize;
	      
	      blocksize = NEXTBYTE;
	      if (blocksize == 2) 
		{
		  aspnum = NEXTBYTE;
		  aspden = NEXTBYTE;
		  if (aspden>0 && aspnum>0)
		    { 
		    }
		  else 
		    { 
		      aspnum = aspden = 1; 
		    }
	      
		  if (DEBUG) 
		    {
		      fprintf(stderr,"GIF87 aspect extension: %d:%d =\n\n", 
				 aspnum, aspden);
		    }
		}
	      else 
		{
		  for ( i = 0; i < blocksize; i++ ) 
		    {
		      NEXTBYTE;
		    }
		}
	      
	      while ( (sbsize=NEXTBYTE) > 0 ) 
		{  /* eat any following data subblocks */
		  for ( i = 0; i < sbsize; i++ ) 
		    {
		      NEXTBYTE;
		    }
		}
	    }
	  else if (fn == 0xFE)  // Comment Extension
	    {
	      int   ch, j, sbsize, cmtlen;
	      BYTE *ptr1, *cmt, *cmt1, *sp;
	      
	      cmtlen = 0;
	      ptr1 = gpbData;      // remember start of comments
	  
	      // figure out length of comment
	      do {
		sbsize = NEXTBYTE;
		cmtlen += sbsize;
		for (j=0; j<sbsize; j++) ch = NEXTBYTE;
	      } while (sbsize);


	      if (cmtlen>0)  // build into one un-blocked comment
		{
		  cmt = new BYTE[cmtlen + 1];
		  if (!cmt) 
		    {
		      gifWarning("couldn't allocate space for comments\n");
		    }
		  else 
		    {
		      sp = cmt;
		      do {
			sbsize = (*ptr1++);
			for (j=0; j<sbsize; j++, sp++, ptr1++) *sp = *ptr1;
		      } while (sbsize);
		      *sp = '\0';
		      
		      if (pBitMap->pcComment) // have to strcat onto old comments */
			{
			  cmt1 = new BYTE[strlen(pBitMap->pcComment) + cmtlen + 2];
			  if (!cmt1) {
			    gifWarning("couldn't allocate space for comments\n");
			    delete []cmt;
			  }
			  else {
			    strcpy((char *) cmt1, (char *) pBitMap->pcComment);
			    strcat((char *) cmt1, (char *) "\n");
			    strcat((char *) cmt1, (char *) cmt);
			    delete [](pBitMap->pcComment);
			    delete []cmt;
			    pBitMap->pcComment = (char *) cmt1;
			  }
			}
		      else
			{
			  pBitMap->pcComment = (char *) cmt;
			}
		    }  // if (cmt)
		}  // if (cmtlen>0)
	    }

	  else if (fn == 0x01) // PlainText Extension 
	    {  
	      int j,sbsize,ch;
	      int tgLeft, tgTop, tgWidth, tgHeight, cWidth, cHeight, fg, bg;
      
	      SetISTR(ISTR_INFO, "%s:  %s", gpcBaseName, 
		      "PlainText extension found in GIF file.  Ignored.");

	      sbsize   = NEXTBYTE;
	      tgLeft   = NEXTBYTE;  tgLeft   += (NEXTBYTE)<<8;
	      tgTop    = NEXTBYTE;  tgTop    += (NEXTBYTE)<<8;
	      tgWidth  = NEXTBYTE;  tgWidth  += (NEXTBYTE)<<8;
	      tgHeight = NEXTBYTE;  tgHeight += (NEXTBYTE)<<8;
	      cWidth   = NEXTBYTE;
	      cHeight  = NEXTBYTE;
	      fg       = NEXTBYTE;
	      bg       = NEXTBYTE;
	      for ( i = 12; i < sbsize; i++ ) // read rest of first subblock
		{
		  NEXTBYTE;
		}
	      if (DEBUG) 
		{
		  fprintf(stderr,
			  "PlainText: tgrid=%d,%d %dx%d  cell=%dx%d  col=%d,%d\n",
			  tgLeft, tgTop, tgWidth, tgHeight, 
			  cWidth, cHeight, fg, bg);
		}
		  
	      /* read (and ignore) data sub-blocks */
	      do {
		j = 0;
		sbsize = NEXTBYTE;
		while (j<sbsize) 
		  {
		    ch = NEXTBYTE;  j++;
		    if (DEBUG) fprintf(stderr,"%c", ch);
		  }
	      } while (sbsize);
	      if (DEBUG) 
		{
		  fprintf(stderr,"\n\n");
		}
	    }
	  else if (fn == 0xF9) // Graphic Control Extension
	    {
	      int j, sbsize;
	      
	      if (DEBUG) fprintf(stderr,"Graphic Control extension\n\n");
	      
	      SetISTR(ISTR_INFO, "%s:  %s", gpcBaseName, 
		      "Graphic Control Extension in GIF file.  Ignored.");
	
	      // read (and ignore) data sub-blocks
	      do {
		j = 0; sbsize = NEXTBYTE;
		while (j<sbsize) 
		  { 
		    NEXTBYTE;  
		    j++; 
		  }
	      } while (sbsize);
	    }
      

	  else if (fn == 0xFF) // Application Extension
	    {
	      int j, sbsize;

	      if (DEBUG) fprintf(stderr,"Application extension\n\n");
	      
	      /* read (and ignore) data sub-blocks */
	      do {
		j = 0; sbsize = NEXTBYTE;
		while (j<sbsize)
		  { 
		    NEXTBYTE;  
		    j++; 
		  }
	      } while (sbsize);
	    }
	  else // unknown extension
	    {
	      int j, sbsize;
	      
	      if (DEBUG) fprintf(stderr,"unknown GIF extension 0x%02x\n\n", fn);
	      
	      SetISTR(ISTR_INFO,
		      "%s:  Unknown extension 0x%02x in GIF file.  Ignored.",
		      gpcBaseName, fn);
	
	      /* read (and ignore) data sub-blocks */
	      do {
		j = 0; sbsize = NEXTBYTE;
		while (j<sbsize) 
		  { 
		    NEXTBYTE;  
		    j++; 
		  }
	      } while (sbsize);
	    }
	}


      else if (block == IMAGESEP) 
	{
	  if (DEBUG) fprintf(stderr,"imagesep (got=%d)  ",gotimage);
	  if (DEBUG) fprintf(stderr,"  at start: offset=0x%lx\n",gpbData-gpbRawGIF);
	  
	  if (gotimage) /* just skip over remaining images */
	    {
	      int i,misc,ch,ch1;
	    
	      /* skip image header */
	      NEXTBYTE;  NEXTBYTE;  /* left position */
	      NEXTBYTE;  NEXTBYTE;  /* top position */
	      NEXTBYTE;  NEXTBYTE;  /* width */
	      NEXTBYTE;  NEXTBYTE;  /* height */
	      misc = NEXTBYTE;      /* misc. bits */
	    
	      if (misc & 0x80) // image has local colormap.  skip it */
		{
		  for (i=0; i< 1 << ((misc&7)+1);  i++)
		    {
		      NEXTBYTE;  NEXTBYTE;  NEXTBYTE;
		    }
		}

	      NEXTBYTE;       // minimum code size
	      
	      // skip image data sub-blocks
	      do {
		ch = ch1 = NEXTBYTE;
		while (ch--) NEXTBYTE;
		if ((gpbData - gpbRawGIF) > giFileSize) break;      /* EOF */
	      } while(ch1);
	    }
	  else if (readImage(pBitMap)) 
	    {
	      gotimage = 1;
	    }
	  if (DEBUG) 
	    {
	      fprintf(stderr,"  at end:   gpbData=0x%lx\n",gpbData-gpbRawGIF);
	    }
	}
      else if (block == TRAILER) // stop reading blocks */
	{
	  if (DEBUG) fprintf(stderr,"trailer");
	  break;
	}

      else {      /* unknown block type */
	char str[128];
	
	if (DEBUG) fprintf(stderr,"block type 0x%02x  ", block);
	
	/* don't mention bad block if file was trunc'd, as it's all bogus */
	if ((gpbData - pbOrig) < giFileSize) {
	  sprintf(str, "Unknown block type (0x%02x) at offset 0x%lx",
		block, (gpbData - pbOrig) - 1);
	  
	  if (!gotimage) return gifError(pBitMap, str);
	  else gifWarning(str);
	}
	
	break;
    }

      if (DEBUG) fprintf(stderr,"\n");
  }
  
  delete []gpbRawGIF;  
  gpbRawGIF = NULL;
  delete []gpbRaster;  
  gpbRaster = NULL;

  if (!gotimage)
    { 
      return( gifError(pBitMap, "no image data found in GIF file") );
    }

  return 1;
}


/********************************************/
static int readImage(BitMap *pBitMap)
{
  register BYTE ch, ch1, *ptr1, *picptr;
  int           i, npixels, maxpixels;

  npixels = maxpixels = 0;

  /* read in values from the image descriptor */
  
  ch = NEXTBYTE;
  LeftOfs = ch + 0x100 * NEXTBYTE;
  ch = NEXTBYTE;
  TopOfs  = ch + 0x100 * NEXTBYTE;
  ch = NEXTBYTE;
  Width   = ch + 0x100 * NEXTBYTE;
  ch = NEXTBYTE;
  Height  = ch + 0x100 * NEXTBYTE;

  Misc = NEXTBYTE;
  gbInterlace = ((Misc & INTERLACEMASK) ? TRUE : FALSE);

  if (Misc & 0x80) 
    {
      for ( i = 0; i< 1 << ((Misc&7)+1); i++ ) 
	{
	  pBitMap->pColorMap[i].pbComp[R] = NEXTBYTE;
	  pBitMap->pColorMap[i].pbComp[G] = NEXTBYTE;
	  pBitMap->pColorMap[i].pbComp[B] = NEXTBYTE;
	}
  }


  if (!gbHasColormap && !(Misc&0x80)) {
    /* no global or local colormap */
    SetISTR(ISTR_WARNING, "%s:  %s", gpcBaseName, 
	    "No colormap in this GIF file.  Assuming EGA colors.");
  }
    

  
  /* Start reading the raster data. First we get the intial code size
   * and compute decompressor constant values, based on this code size.
   */
  
  CodeSize = NEXTBYTE;

  ClearCode = (1 << CodeSize);
  EOFCode = ClearCode + 1;
  FreeCode = FirstFree = ClearCode + 2;
  
  /* The GIF spec has it that the code size is the code size used to
   * compute the above values is the code size given in the file, but the
   * code size used in compression/decompression is the code size given in
   * the file plus one. (thus the ++).
   */
  
  CodeSize++;
  InitCodeSize = CodeSize;
  MaxCode = (1 << CodeSize);
  ReadMask = MaxCode - 1;
  


  /* UNBLOCK:
   * Read the raster data.  Here we just transpose it from the GIF array
   * to the Raster array, turning it from a series of blocks into one long
   * data stream, which makes life much easier for readCode().
   */
  
  ptr1 = gpbRaster;
  do {
    ch = ch1 = NEXTBYTE;
    while (ch--) { *ptr1 = NEXTBYTE; ptr1++; }
    if ((gpbData - gpbRawGIF) > giFileSize) {
      SetISTR(ISTR_WARNING,"%s:  %s", gpcBaseName,
	      "This GIF file seems to be truncated.  Winging it.");
      break;
    }
  } while(ch1);




  if (DEBUG) 
    {
      fprintf(stderr,"xv: LoadGIF() - picture is %dx%d, %d bits, %sinterlaced\n",
	      Width, Height, BitsPerPixel, gbInterlace ? "" : "non-");
    }
  

  /* Allocate the 'pic' */
  maxpixels = Width*Height;
  picptr = pic8 = new BYTE[maxpixels];
  if ( pic8 == NULL ) 
    {
      return( gifError(pBitMap, "couldn't allocate 'pic8'") );
    }

  pBitMap->pvBuffer = new ColorEntry[maxpixels];
  if ( pBitMap->pvBuffer == NULL ) 
    {
      return( gifError(pBitMap, "couldn't allocate buffer") );
    }
  
  /* Decompress the file, continuing until you see the GIF EOF code.
   * One obvious enhancement is to add checking for corrupt files here.
   */
  
  Code = readCode();
  while (Code != EOFCode) {
    /* Clear code sets everything back to its initial value, then reads the
     * immediately subsequent code as uncompressed data.
     */

    if (Code == ClearCode) {
      CodeSize = InitCodeSize;
      MaxCode = (1 << CodeSize);
      ReadMask = MaxCode - 1;
      FreeCode = FirstFree;
      Code = readCode();
      CurCode = OldCode = Code;
      FinChar = CurCode & BitMask;
      if (!gbInterlace) *picptr++ = FinChar;
         else doInterlace(FinChar);
      npixels++;
    }
    else {
      /* If not a clear code, must be data: save same as CurCode and InCode */

      /* if we're at maxcode and didn't get a clear, stop loading */
      if (FreeCode>=4096) { /* printf("freecode blew up\n"); */
			    break; }

      CurCode = InCode = Code;
      
      /* If greater or equal to FreeCode, not in the hash table yet;
       * repeat the last character decoded
       */
      
      if (CurCode >= FreeCode) {
	CurCode = OldCode;
	if (OutCount > 4096) {  /* printf("outcount1 blew up\n"); */ break; }
	gpiOutCode[OutCount++] = FinChar;
      }
      
      /* Unless this code is raw data, pursue the chain pointed to by CurCode
       * through the hash table to its end; each code in the chain puts its
       * associated output code on the output queue.
       */
      
      while (CurCode > BitMask) {
	if (OutCount > 4096) break;   /* corrupt file */
	gpiOutCode[OutCount++] = gpiSuffix[CurCode];
	CurCode = gpiPrefix[CurCode];
      }
      
      if (OutCount > 4096) { /* printf("outcount blew up\n"); */ break; }
      
      /* The last code in the chain is treated as raw data. */
      
      FinChar = CurCode & BitMask;
      gpiOutCode[OutCount++] = FinChar;
      
      /* Now we put the data out to the Output routine.
       * It's been stacked LIFO, so deal with it that way...
       */

      /* safety thing:  prevent exceeding range of 'pic8' */
      if (npixels + OutCount > maxpixels) OutCount = maxpixels-npixels;
	
      npixels += OutCount;
      if (!gbInterlace) for (i=OutCount-1; i>=0; i--) *picptr++ = gpiOutCode[i];
                else  for (i=OutCount-1; i>=0; i--) doInterlace(gpiOutCode[i]);
      OutCount = 0;

      /* Build the hash table on-the-fly. No table is stored in the file. */
      
      gpiPrefix[FreeCode] = OldCode;
      gpiSuffix[FreeCode] = FinChar;
      OldCode = InCode;
      
      /* Point to the next slot in the table.  If we exceed the current
       * MaxCode value, increment the code size unless it's already 12.  If it
       * is, do nothing: the next code decompressed better be CLEAR
       */
      
      FreeCode++;
      if (FreeCode >= MaxCode) {
	if (CodeSize < 12) {
	  CodeSize++;
	  MaxCode *= 2;
	  ReadMask = (1 << CodeSize) - 1;
	}
      }
    }
    Code = readCode();
    if (npixels >= maxpixels) break;
  }
/*  
  if (npixels != maxpixels) {
    SetISTR(ISTR_WARNING,"%s:  %s", gpcBaseName,
	    "This GIF file seems to be truncated.  Winging it.");
    if (!gbInterlace)  // clear->EOBuffer
      xvbzero((char *) pic8+npixels, (size_t) (maxpixels-npixels));
  }
*/
  if (npixels != maxpixels) {
    gifError(pBitMap, "This GIF file seems to be truncated");
  }

  // fill in the BitMap structure

  pBitMap->iWidth       = Width;           
  pBitMap->iHeight      = Height;
  pBitMap->iBitsPerComp = 8;
  pBitMap->iComps       = 3;

  // pBitMap->pcComment gets handled in main LoadGIF() block-reader

  for ( i = 0; i < maxpixels; i++ )
    {
      ((ColorEntry*)(pBitMap->pvBuffer))[i] = pBitMap->pColorMap[pic8[i]];
    }
  

  delete []pic8;
  pic8 = NULL;

  return 1;
}



/* Fetch the next code from the raster data stream.  The codes can be
 * any length from 3 to 12 bits, packed into 8-bit bytes, so we have to
 * maintain our location in the Raster array as a BIT Offset.  We compute
 * the byte Offset into the raster array by dividing this by 8, pick up
 * three bytes, compute the bit Offset into our 24-bit chunk, shift to
 * bring the desired code to the bottom, then mask it off and return it. 
 */
static int readCode()
{
  int RawCode, ByteOffset;
  
  ByteOffset = BitOffset / 8;
  RawCode = gpbRaster[ByteOffset] + (gpbRaster[ByteOffset + 1] << 8);
  if (CodeSize >= 8)
    RawCode += ( ((int) gpbRaster[ByteOffset + 2]) << 16);
  RawCode >>= (BitOffset % 8);
  BitOffset += CodeSize;

  return(RawCode & ReadMask);
}


/***************************/
static void doInterlace(int Index)
{
  static BYTE *ptr = NULL;
  static int   oldYC = -1;
  
  if (oldYC != YC) {  ptr = pic8 + YC * Width;  oldYC = YC; }
  
  if (YC<Height)
    *ptr++ = Index;
  
  /* Update the X-coordinate, and if it overflows, update the Y-coordinate */
  
  if (++XC == Width) {
    
    /* deal with the interlace as described in the GIF
     * spec.  Put the decoded scan line out to the screen if we haven't gone
     * past the bottom of it
     */
    
    XC = 0;
    
    switch (Pass) {
    case 0:
      YC += 8;
      if (YC >= Height) { Pass++; YC = 4; }
      break;
      
    case 1:
      YC += 8;
      if (YC >= Height) { Pass++; YC = 2; }
      break;
      
    case 2:
      YC += 4;
      if (YC >= Height) { Pass++; YC = 1; }
      break;
      
    case 3:
      YC += 2;  break;
      
    default:
      break;
    }
  }
}


      
/*****************************/
static int gifError(BitMap *pBitMap, char *st)
{
  gifWarning(st);

  if ( gpbRawGIF != NULL ) 
    {
      delete []gpbRawGIF;
      gpbRawGIF = NULL;
    }

  if ( gpbRaster != NULL )
    {
      delete []gpbRaster;
      gpbRaster = NULL;
    }

  if ( pic8 != NULL ) 
    {
      delete []pic8;
      pic8 = NULL;
    }
  if ( pBitMap->pvBuffer != NULL )
    {
      delete [](pBitMap->pvBuffer);
      pBitMap->pvBuffer = NULL;
    }
  if ( pBitMap->pcComment != NULL ) 
    {
      delete [](pBitMap->pcComment);
      pBitMap->pcComment = NULL;
    }

  return 0;
}


/*****************************/
static void gifWarning(char *st)
{
  SetISTR(ISTR_WARNING,"%s:  %s", gpcBaseName, st);
}


