#include "heap.h"

//////////////////////////////////////////////////////////////////////
// CHeapEntry Class
//

CHeapEntry::CHeapEntry()
{
  m_uiIndex = 0;

  m_fKey = -INFINITY;
  m_pvData = NULL;
}

CHeapEntry::~CHeapEntry()
{
  Uninit();
}

VOID CHeapEntry::Init()
{
  m_uiIndex = 0;

  m_fKey = -INFINITY;
  m_pvData = NULL;
}

VOID CHeapEntry::Init(UINT_32 uiIndex)
{
  Init();

  m_uiIndex = uiIndex;
}

VOID CHeapEntry::Init(UINT_32 uiIndex, FLOAT fKey, VOID *pvData)
{
  Init(uiIndex);

  m_fKey = fKey;
  m_pvData = pvData;
}

VOID CHeapEntry::Uninit()
{
  m_uiIndex = 0;

  m_fKey = -INFINITY;
  m_pvData = NULL;
}

BOOL CHeapEntry::Initialized()
{
  return TRUE;
}

BOOL CHeapEntry::Valid()
{
  return ( m_fKey == -INFINITY && m_pvData == NULL ) ||
         ( m_pvData != NULL );
}

VOID CHeapEntry::Display()
{
  printf("(%ld: %3.3f, %p)", m_uiIndex, m_fKey, m_pvData);
}

//////////////////////////////////////////////////////////////////////
// CHeap Class
//

static const UINT_32 cbBiggestFirst = TRUE;
UINT_32 CHeap::m_uiObjectCount = 0;
UINT_32 CHeap::m_uiNewedEntries = 0;
UINT_32 CHeap::m_uiAllocatedEntries = 0;
CListInt *CHeap::m_plistFreeEntries = NULL;

CHeap::CHeap()
{
  if ( m_uiObjectCount == 0 )
    {
      m_plistFreeEntries = new CListInt;
      ASSERT( m_plistFreeEntries != NULL );
      m_plistFreeEntries->Init();
    }
  m_uiObjectCount++;

  m_bBiggestFirst = cbBiggestFirst;
  m_uiSizeMax = 0;

  m_ppCHeapEntry = NULL;
}

CHeap::~CHeap()
{
  Uninit();

  m_uiObjectCount--;
  if ( m_uiObjectCount == 0 )
    {
      // Clean up our free list of Entries
      ASSERT( m_uiAllocatedEntries == 0 );
      if ( m_plistFreeEntries != NULL )
	{
	  CHeapEntry *pCHeapEntry;

	  for ( pCHeapEntry = (CHeapEntry*)m_plistFreeEntries->RemoveFirst();
		pCHeapEntry != NULL;
		pCHeapEntry = (CHeapEntry*)m_plistFreeEntries->RemoveFirst() )
	    {
	      delete pCHeapEntry;
	    }

	  delete m_plistFreeEntries;
	  m_plistFreeEntries = NULL;
	}
    }
}

VOID CHeap::Init(UINT_32 uiSizeMax/* = 1*/)
{

  ASSERT( uiSizeMax > 0 );
  
  m_bBiggestFirst = cbBiggestFirst;

  m_uiSize = 0;
  m_uiSizeMax = uiSizeMax;

  m_ppCHeapEntry = new CHeapEntry*[m_uiSizeMax + 1];
  ASSERT( m_ppCHeapEntry != NULL );
  memset(m_ppCHeapEntry, 0, (m_uiSizeMax+1)*sizeof(CHeapEntry*));
}

VOID CHeap::Uninit()
{
  RemoveAll();

  m_bBiggestFirst = cbBiggestFirst;

  m_uiSize = 0;
  m_uiSizeMax = 0;
  
  if ( m_ppCHeapEntry != NULL )
    {
      delete[] m_ppCHeapEntry;
      m_ppCHeapEntry = NULL;
    }
}

BOOL CHeap::Initialized()
{
  return m_ppCHeapEntry != NULL;
}

BOOL CHeap::Valid()
{
  return Initialized();
}

CHeapEntry *CHeap::Insert(FLOAT fKey, VOID *pvData)
{
  CHeapEntry *pCHeapEntry;

  for ( UINT_32 i = 1; i <= m_uiSize; i++ )
    {
      ASSERT( m_ppCHeapEntry[i]->GetData() != pvData );
    }

  pCHeapEntry = AllocateCHeapEntry();
  pCHeapEntry->Init(m_uiSize, fKey, pvData);

  m_ppCHeapEntry[m_uiSize] = pCHeapEntry;

  HeapifyUp(m_uiSize);

  return pCHeapEntry;
}

VOID *CHeap::PeekTop(FLOAT *pfKey/* = NULL*/)
{
  if ( m_uiSize == 0 )
    {
      return NULL;
    }

  if ( pfKey != NULL )
    {
      *pfKey = m_ppCHeapEntry[1]->GetKey();
    }

  return m_ppCHeapEntry[1]->GetData();
}

VOID CHeap::RemoveAll()
{
  CHeapEntry *pCHeapEntry;

  while ( m_uiSize > 0 )
    {
      pCHeapEntry = m_ppCHeapEntry[m_uiSize];
      m_ppCHeapEntry[m_uiSize] = NULL;
      DeallocateCHeapEntry(pCHeapEntry);
    }
}

VOID *CHeap::RemoveTop(FLOAT *pfKey/* = NULL*/)
{
  if ( m_uiSize == 0 )
    {
      return NULL;
    }

  return Remove(m_ppCHeapEntry[1], pfKey);
}

VOID *CHeap::RemoveLast(FLOAT *pfKey/* = NULL*/)
{
  if ( m_uiSize == 0 )
    {
      return NULL;
    }

  return Remove(m_ppCHeapEntry[m_uiSize], pfKey);
}

VOID *CHeap::Remove(CHeapEntry *pCHeapEntry, FLOAT *pfKey/* = NULL*/)
{
  UINT_32 uiIndex;
  VOID *pvData;

  ASSERT( pCHeapEntry != NULL );

  uiIndex = pCHeapEntry->GetIndex();
  ASSERT( m_ppCHeapEntry[uiIndex] == pCHeapEntry );

  pvData = pCHeapEntry->GetData();
  if ( pfKey != NULL )
    {
      *pfKey = pCHeapEntry->GetKey();
    }

  if ( uiIndex != m_uiSize )
    {
      // Swap the last array element into its position
      SetEntry(uiIndex, m_ppCHeapEntry[m_uiSize]);
      m_ppCHeapEntry[m_uiSize] = NULL;

      DeallocateCHeapEntry(pCHeapEntry);

      // Reheapify
      HeapifyDown(uiIndex);
    }
  else
    {
      DeallocateCHeapEntry(pCHeapEntry);
    }

  return pvData;
}

VOID CHeap::Update(CHeapEntry *pCHeapEntry, FLOAT fKeyNew)
{
  UINT_32 uiIndex;
  FLOAT fKeyOld;

  ASSERT( pCHeapEntry != NULL );

  uiIndex = pCHeapEntry->GetIndex();
  ASSERT( m_ppCHeapEntry[uiIndex] == pCHeapEntry );

  fKeyOld = pCHeapEntry->GetKey();
  pCHeapEntry->SetKey(fKeyNew);

  if ( fKeyNew >= fKeyOld )
    {
      HeapifyUp(uiIndex);
    }
  else
    {
      HeapifyDown(uiIndex);
    }
}

VOID CHeap::Swap(UINT_32 uiIndex0, UINT_32 uiIndex1)
{
  CHeapEntry *pCHeapEntry;

  ASSERT( uiIndex0 <= m_uiSize );
  ASSERT( uiIndex1 <= m_uiSize );
  ASSERT( m_ppCHeapEntry[uiIndex0]->GetIndex() == uiIndex0 );
  ASSERT( m_ppCHeapEntry[uiIndex1]->GetIndex() == uiIndex1 );

  pCHeapEntry = m_ppCHeapEntry[uiIndex0];
  m_ppCHeapEntry[uiIndex0] = m_ppCHeapEntry[uiIndex1];
  m_ppCHeapEntry[uiIndex1] = pCHeapEntry;

  m_ppCHeapEntry[uiIndex0]->SetIndex(uiIndex0);
  m_ppCHeapEntry[uiIndex1]->SetIndex(uiIndex1);
}

VOID CHeap::HeapifyUp(UINT_32 uiIndex)
{
  UINT_32 uiParentIndex;

  if ( CheckRoot(uiIndex) )
    {
      return;
    }

  uiParentIndex = GetParent(uiIndex);

  if ( ( m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiIndex]->GetKey() > 
	   m_ppCHeapEntry[uiParentIndex]->GetKey() ) ) ||
       ( !m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiIndex]->GetKey() < 
	   m_ppCHeapEntry[uiParentIndex]->GetKey() ) ) )
    {
      Swap(uiParentIndex, uiIndex);

      HeapifyUp(uiParentIndex);
    }
}

VOID CHeap::HeapifyDown(UINT_32 uiIndex)
{
  UINT_32 uiChildIndex, uiLeftChildIndex, uiRightChildIndex;

  uiLeftChildIndex = GetLeftChild(uiIndex);
  uiRightChildIndex = GetRightChild(uiIndex);

  if ( uiLeftChildIndex > m_uiSize )
    {
      // This is a leaf node
      return;
    }
  else if ( uiRightChildIndex > m_uiSize )
    {
      // There is only a left child
      uiChildIndex = uiLeftChildIndex;
    }
  else
    {
      // There are two children so chose the maximum one
      if ( ( m_bBiggestFirst && 
	     ( m_ppCHeapEntry[uiRightChildIndex]->GetKey() > 
	       m_ppCHeapEntry[uiLeftChildIndex]->GetKey() ) ) ||
	   ( !m_bBiggestFirst && 
	     ( m_ppCHeapEntry[uiRightChildIndex]->GetKey() < 
	       m_ppCHeapEntry[uiLeftChildIndex]->GetKey() ) ) )
	{
	  uiChildIndex = uiRightChildIndex;
	}
      else
	{
	  uiChildIndex = uiLeftChildIndex;
	}
    }

  if ( ( m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiChildIndex]->GetKey() > 
	   m_ppCHeapEntry[uiIndex]->GetKey() ) )
       ( !m_bBiggestFirst && 
	 ( m_ppCHeapEntry[uiChildIndex]->GetKey() < 
	   m_ppCHeapEntry[uiIndex]->GetKey() ) ) )
    {
      Swap(uiIndex, uiChildIndex);

      HeapifyDown(uiChildIndex);
    }
}

CHeapEntry *CHeap::AllocateCHeapEntry()
{
  CHeapEntry *pCHeapEntry;

  ASSERT( m_plistFreeEntries != NULL );
  ASSERT( m_uiSize <= m_uiSizeMax );
  ASSERT( m_ppCHeapEntry != NULL );

  if ( m_uiSize == m_uiSizeMax )
    {
      CHeapEntry **ppCHeapEntry;
      UINT_32 uiSizeMax;

      uiSizeMax = m_uiSizeMax;
      ppCHeapEntry = m_ppCHeapEntry;

      // Allocate a new array
      m_uiSizeMax = 2*(m_uiSizeMax + 1) - 1;
      m_ppCHeapEntry = new CHeapEntry*[m_uiSizeMax+1];
      ASSERT( m_ppCHeapEntry != NULL );
      memset(m_ppCHeapEntry, 0, (m_uiSizeMax+1)*sizeof(CHeapEntry*));

      // Copy the current Heap into the new array
      memcpy(m_ppCHeapEntry, ppCHeapEntry, (uiSizeMax+1)*sizeof(CHeapEntry*));

      // Deallocate the old array
      delete[] ppCHeapEntry;
    }

  m_uiSize++;

  // Allocate the Entry
  pCHeapEntry = (CHeapEntry*)m_plistFreeEntries->RemoveFirst();
  if ( pCHeapEntry == NULL )
    {
      ASSERT( m_uiNewedEntries == m_uiAllocatedEntries );
      m_uiNewedEntries++;
      pCHeapEntry = new CHeapEntry;
      ASSERT( pCHeapEntry != NULL );
    }
  ASSERT( m_uiAllocatedEntries < m_uiNewedEntries );

  m_uiAllocatedEntries++;

  ASSERT( m_uiSize <= m_uiSizeMax );

  return pCHeapEntry;
}

VOID CHeap::DeallocateCHeapEntry(CHeapEntry *pCHeapEntry)
{
  ASSERT( pCHeapEntry != NULL );
  ASSERT( m_uiSize > 0 );

  m_uiSize--;

  ASSERT( m_plistFreeEntries != NULL );
  
  ASSERT( m_uiAllocatedEntries > 0 );

  m_plistFreeEntries->InsertFirst(pCHeapEntry, 0);
  m_uiAllocatedEntries--;
  ASSERT( m_uiAllocatedEntries < m_uiNewedEntries );
}

VOID CHeap::Display()
{
  UINT_32 i;

  printf("\nHeap: size = %ld max = %ld\n", m_uiSize, m_uiSizeMax); 

  for ( i = 1; i <= m_uiSize; i++ )
    {
      m_ppCHeapEntry[i]->Display();

      printf(" ");
      if ( (i & (i+1)) == 0x0 )
	{
	  printf("\n");
	}
    }

  printf("\n");
}
