#ifndef __INTERVAL_H__
#define __INTERVAL_H__

#include "genlib.h"

//#define TIMING_ONLY
#ifdef TIMING_ONLY

#define ROUND_UP(e) e
#define ROUND_DOWN(e) e

#define PUSH_ROUND_UP()
#define PUSH_ROUND_DOWN()
#define SET_ROUND_UP()
#define SET_ROUND_DOWN()
#define POP_ROUND()

#define ROUND_TO_PLUS_INFINITY 0

#else // TIMING_ONLY

#ifdef __OS_IRIX__

#include <sys/fpu.h>
extern "C" {
extern int swapRM(int x);
}

#define ROUND_UP(e) \
Interval::m_iRM = swapRM(ROUND_TO_PLUS_INFINITY); \
e; \
swapRM(Interval::m_iRM)
#define ROUND_DOWN(e) \
Interval::m_iRM = swapRM(ROUND_TO_MINUS_INFINITY); \
e; \
swapRM(Interval::m_iRM)

#define PUSH_ROUND_UP() \
Interval::m_iRM = swapRM(ROUND_TO_PLUS_INFINITY)
#define PUSH_ROUND_DOWN() \
Interval::m_iRM = swapRM(ROUND_TO_MINUS_INFINITY)
#define SET_ROUND_UP() \
swapRM(ROUND_TO_PLUS_INFINITY)
#define SET_ROUND_DOWN() \
swapRM(ROUND_TO_MINUS_INFINITY)
#define POP_ROUND() \
swapRM(Interval::m_iRM)

#else // __OS_IRIX__

#ifdef __OS_LINUX__

#include <fpu_control.h>

#define ROUND_TO_PLUS_INFINITY _FPU_RC_UP
#define ROUND_TO_MINUS_INFINITY _FPU_RC_DOWN

#define ROUND_UP(e) \
_FPU_GETCW(Interval::m_iRM); \
Interval::m_iRMTmp=_FPU_RC_UP; \
_FPU_SETCW(Interval::m_iRMTmp); \
e; \
_FPU_SETCW(Interval::m_iRM)
#define ROUND_DOWN(e) \
_FPU_GETCW(Interval::m_iRM); \
Interval::m_iRMTmp=_FPU_RC_DOWN; \
_FPU_SETCW(Interval::m_iRMTmp); \
e; \
_FPU_SETCW(Interval::m_iRM)

#define PUSH_ROUND_UP() \
_FPU_GETCW(Interval::m_iRM); \
Interval::m_iRMTmp=_FPU_RC_UP; \
_FPU_SETCW(Interval::m_iRMTmp)
#define PUSH_ROUND_DOWN() \
_FPU_GETCW(Interval::m_iRM); \
Interval::m_iRMTmp=_FPU_RC_DOWN; \
_FPU_SETCW(Interval::m_iRMTmp)
#define SET_ROUND_UP() \
Interval::m_iRMTmp=_FPU_RC_UP; \
_FPU_SETCW(Interval::m_iRMTmp)
#define SET_ROUND_DOWN() \
Interval::m_iRMTmp=_FPU_RC_DOWN; \
_FPU_SETCW(Interval::m_iRMTmp)
#define POP_ROUND() \
_FPU_SETCW(Interval::m_iRM)

#else // __OS_LINUX__

#define ROUND_UP(e) e
#define ROUND_DOWN(e) e

#define PUSH_ROUND_UP()
#define PUSH_ROUND_DOWN()
#define SET_ROUND_UP()
#define SET_ROUND_DOWN()
#define POP_ROUND()

#define ROUND_TO_PLUS_INFINITY 0

#endif // __OS_LINUX__

#endif // __OS_IRIX__



#endif // TIMING_ONLY

//////////////////////////////////////////////////////////////////////
// Interval Class -- Interval Math
//

class Interval;

inline Interval sqrt(const Interval &val);
inline double dble(const Interval &val);

class Interval
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend inline Interval sqrt(const Interval &val);
  friend inline double dble(const Interval &val);

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  //enum AccuracyType { atGlobal, atLocal, atExact, atMPFUN, atMPFUNNormalized, atSize };

//////////////////////////////////////////////////
// Initialization
//
public:
  inline Interval();
  inline Interval(double d);
  inline Interval(const Interval &val);
  inline ~Interval();

//////////////////////////////////////////////////
// Operator Functions
//
public:
  void ConservativeBound(double d);

  inline Interval &operator=(double d);
  inline Interval &operator=(const Interval &val);
  inline Interval &operator+=(double d);
  inline Interval &operator+=(const Interval &val);
  inline Interval &operator-=(double d);
  inline Interval &operator-=(const Interval &val);
  inline Interval &operator*=(double d);
  inline Interval &operator*=(const Interval &val);
  // Division only works for interval that don't contain zero!!!
  inline Interval &operator/=(double d);
  inline Interval &operator/=(const Interval &val);

  inline Interval operator+(double d) const;
  inline Interval operator+(const Interval &val) const;
  inline Interval operator-(double d) const;
  inline Interval operator-(const Interval &val) const;
  inline Interval operator-() const;
  inline Interval operator*(double d) const;
  inline Interval operator*(const Interval &val) const;
  // Division only works for interval that don't contain zero!!!
  inline Interval operator/(double d) const;
  inline Interval operator/(const Interval &val) const;

  // Returns
  //  0  => FALSE
  //  1  => TRUE
  //  -1 => Uncertain
  inline int operator<(const Interval &val) const;
  inline int operator>(const Interval &val) const;
  inline int operator<=(const Interval &val) const;
  inline int operator>=(const Interval &val) const;
  inline int operator!=(const Interval &val) const;
  inline int operator==(const Interval &val) const;
  inline int operator<(double d) const;
  inline int operator>(double d) const;
  inline int operator<=(double d) const;
  inline int operator>=(double d) const;
  inline int operator!=(double d) const;
  inline int operator==(double d) const;

  void Display(FILE *fpOutput = stdout);

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static int m_iRM; // swap space for current rounding mode
  static int m_iRMTmp; // swap space for current rounding mode

//////////////////////////////////////////////////
// Member Variables
//
protected:
  double m_dMin;
  double m_dMax;
};

//////////////////////////////////////////////////////////////////////
// Inlined Member Functions
//

inline Interval::Interval()
{
}

inline Interval::Interval(double d)
{
  m_dMin = d;
  m_dMax = d;
}

inline Interval::Interval(const Interval &val)
{
  m_dMin = val.m_dMin;
  m_dMax = val.m_dMax;
}

inline Interval::~Interval()
{
}

inline Interval &Interval::operator=(double d)
{
  m_dMin = d;
  m_dMax = d;

  return *this;
}

inline Interval &Interval::operator=(const Interval &val)
{
  m_dMin = val.m_dMin;
  m_dMax = val.m_dMax;

  return *this;
}

inline Interval &Interval::operator+=(double d)
{
  PUSH_ROUND_DOWN();
  m_dMin = m_dMin + d;
  SET_ROUND_UP();
  m_dMax = m_dMax + d;
  POP_ROUND();

  return *this;
}

inline Interval &Interval::operator+=(const Interval &val)
{
  PUSH_ROUND_DOWN();
  m_dMin = m_dMin + val.m_dMin;
  SET_ROUND_UP();
  m_dMax = m_dMax + val.m_dMax;
  POP_ROUND();

  return *this;
}

inline Interval &Interval::operator-=(double d)
{
  PUSH_ROUND_DOWN();
  m_dMin = m_dMin - d;
  SET_ROUND_UP();
  m_dMax = m_dMax - d;
  POP_ROUND();

  return *this;
}

inline Interval &Interval::operator-=(const Interval &val)
{
  PUSH_ROUND_DOWN();
  m_dMin = m_dMin - val.m_dMax;
  SET_ROUND_UP();
  m_dMax = m_dMax - val.m_dMin;
  POP_ROUND();

  return *this;
}

inline Interval &Interval::operator*=(double d)
{
  double dMin, dMax;
  dMin = m_dMin;
  dMax = m_dMax;

  if ( dMin >= 0.0 )
    {
      // dMin >= 0.0 && dMax >= 0.0
      if ( d >= 0.0 )
	{
	  // val.m_dMin >= 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * d;
	  SET_ROUND_UP();
	  m_dMax = dMax * d;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * d;
	  SET_ROUND_UP();
	  m_dMax = dMin * d;
	  POP_ROUND();
	}
    }
  else if ( dMax < 0.0 )
    {
      // dMin < 0.0 && dMax < 0.0
      if ( d >= 0.0 )
	{
	  // val.m_dMin >= 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * d;
	  SET_ROUND_UP();
	  m_dMax = dMax * d;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * d;
	  SET_ROUND_UP();
	  m_dMax = dMin * d;
	  POP_ROUND();
	}
    }
  else
    {
      // dMin < 0.0 && dMax >= 0.0
      if ( d >= 0.0 )
	{
	  // val.m_dMin >= 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * d;
	  SET_ROUND_UP();
	  m_dMax = dMax * d;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * d;
	  SET_ROUND_UP();
	  m_dMax = dMin * d;
	  POP_ROUND();
	}
    }

  return *this;
}

inline Interval &Interval::operator*=(const Interval &val)
{
  double dMin, dMax;
  dMin = m_dMin;
  dMax = m_dMax;

  if ( dMin >= 0.0 )
    {
      // dMin >= 0.0 && dMax >= 0.0
      if ( val.m_dMin >= 0.0 )
	{
	  // val.m_dMin >= 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMax * val.m_dMax;
	  POP_ROUND();
	}
      else if ( val.m_dMax < 0.0 )
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMin * val.m_dMax;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin < 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMax * val.m_dMax;
	  POP_ROUND();
	}
    }
  else if ( dMax < 0.0 )
    {
      // dMin < 0.0 && dMax < 0.0
      if ( val.m_dMin >= 0.0 )
	{
	  // val.m_dMin >= 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMax * val.m_dMin;
	  POP_ROUND();
	}
      else if ( val.m_dMax < 0.0 )
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMin * val.m_dMin;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin < 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMax * val.m_dMin;
	  POP_ROUND();
	}
    }
  else
    {
      // dMin < 0.0 && dMax >= 0.0
      if ( val.m_dMin >= 0.0 )
	{
	  // val.m_dMin >= 0.0 && val.m_dMax >= 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMax * val.m_dMax;
	  POP_ROUND();
	}
      else if ( val.m_dMax < 0.0 )
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax * val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMin * val.m_dMin;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin < 0.0 && val.m_dMax >= 0.0
	  double dTmp;

	  PUSH_ROUND_DOWN();
	  m_dMin = dMin * val.m_dMax;
	  dTmp = dMax * val.m_dMin;
	  if ( dTmp < m_dMin )
	    {
	      m_dMin = dTmp;
	    }
	  SET_ROUND_UP();
	  m_dMax = dMax * val.m_dMax;
	  dTmp = dMin * val.m_dMin;
	  if ( dTmp > m_dMax )
	    {
	      m_dMax = dTmp;
	    }
	  POP_ROUND();
	}
    }

  return *this;
}

inline Interval &Interval::operator/=(double d)
{
  Interval val(d);

  *this /= val;

  return *this;
}

inline Interval &Interval::operator/=(const Interval &val)
{
  double dMin, dMax;
  dMin = m_dMin;
  dMax = m_dMax;

  if ( dMin >= 0.0 )
    {
      // dMin >= 0.0 && dMax >= 0.0
      if ( val.m_dMin > 0.0 )
	{
	  // val.m_dMin > 0.0 && val.m_dMax > 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin / val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMax / val.m_dMin;
	  POP_ROUND();
	}
      else if ( val.m_dMax < 0.0 )
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax / val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMin / val.m_dMin;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin <= 0.0 && val.m_dMax >= 0.0
	  // The interval contains zero
	  ASSERT( FALSE );
	}
    }
  else if ( dMax < 0.0 )
    {
      // dMin < 0.0 && dMax < 0.0
      if ( val.m_dMin > 0.0 )
	{
	  // val.m_dMin > 0.0 && val.m_dMax > 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin / val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMax / val.m_dMax;
	  POP_ROUND();
	}
      else if ( val.m_dMax < 0.0 )
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax / val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMin / val.m_dMax;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin <= 0.0 && val.m_dMax >= 0.0
	  // The interval contains zero
	  ASSERT( FALSE );
	}
    }
  else
    {
      // dMin < 0.0 && dMax >= 0.0
      if ( val.m_dMin > 0.0 )
	{
	  // val.m_dMin > 0.0 && val.m_dMax > 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMin / val.m_dMin;
	  SET_ROUND_UP();
	  m_dMax = dMax / val.m_dMin;
	  POP_ROUND();
	}
      else if ( val.m_dMax < 0.0 )
	{
	  // val.m_dMin < 0.0 && val.m_dMax < 0.0
	  PUSH_ROUND_DOWN();
	  m_dMin = dMax / val.m_dMax;
	  SET_ROUND_UP();
	  m_dMax = dMin / val.m_dMax;
	  POP_ROUND();
	}
      else
	{
	  // val.m_dMin <= 0.0 && val.m_dMax >= 0.0
	  // The interval contains zero
	  ASSERT( FALSE );
	}
    }

  return *this;
}

inline Interval Interval::operator+(double d) const
{
  Interval valResult(*this);

  valResult += d;

  return valResult;
}

inline Interval Interval::operator+(const Interval &val) const
{
  Interval valResult(*this);

  valResult += val;

  return valResult;
}

inline Interval Interval::operator-(double d) const
{
  Interval valResult(*this);

  valResult -= d;

  return valResult;
}

inline Interval Interval::operator-(const Interval &val) const
{
  Interval valResult(*this);

  valResult -= val;

  return valResult;
}

inline Interval Interval::operator-() const
{
  Interval valResult;

  valResult.m_dMin = -m_dMax;
  valResult.m_dMax = -m_dMin;

  return valResult;
}

inline Interval Interval::operator*(double d) const
{
  Interval valResult(*this);

  valResult *= d;

  return valResult;
}

inline Interval Interval::operator*(const Interval &val) const
{
  Interval valResult(*this);
  valResult *= val;

  return valResult;
}

inline Interval Interval::operator/(double d) const
{
  Interval valResult(*this);

  valResult /= d;

  return valResult;
}

inline Interval Interval::operator/(const Interval &val) const
{
  Interval valResult(*this);
  valResult /= val;

  return valResult;
}

inline int Interval::operator<(double d) const
{
  int iResult;

  iResult = -1;

  if ( m_dMax < d )
    {
      iResult = 1;
    }
  else if ( m_dMin >= d )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator<(const Interval &val) const
{
  int iResult;

  iResult = -1;

  if ( m_dMax < val.m_dMin )
    {
      iResult = 1;
    }
  else if ( m_dMin >= val.m_dMax )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator>(double d) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin > d )
    {
      iResult = 1;
    }
  else if ( m_dMax <= d )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator>(const Interval &val) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin > val.m_dMax )
    {
      iResult = 1;
    }
  else if ( m_dMax <= val.m_dMin )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator<=(double d) const
{
  int iResult;

  iResult = -1;

  if ( m_dMax <= d )
    {
      iResult = 1;
    }
  else if ( m_dMin > d )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator<=(const Interval &val) const
{
  int iResult;

  iResult = -1;

  if ( m_dMax <= val.m_dMin )
    {
      iResult = 1;
    }
  else if ( m_dMin > val.m_dMax )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator>=(double d) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin >= d )
    {
      iResult = 1;
    }
  else if ( m_dMax < d )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator>=(const Interval &val) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin >= val.m_dMax )
    {
      iResult = 1;
    }
  else if ( m_dMax < val.m_dMin )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator!=(double d) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin > d ||
       m_dMax < d )
    {
      iResult = 1;
    }
  else if ( m_dMin == d &&
	    m_dMax == d )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator!=(const Interval &val) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin > val.m_dMax ||
       m_dMax < val.m_dMin )
    {
      iResult = 1;
    }
  else if ( m_dMin == m_dMax &&
	    m_dMin == val.m_dMin &&
	    m_dMax == val.m_dMax )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator==(double d) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin == d &&
       m_dMax == d )
    {
      iResult = 1;
    }
  else if ( m_dMin > d ||
	    m_dMax < d )
    {
      iResult = 0;
    }

  return iResult;
}

inline int Interval::operator==(const Interval &val) const
{
  int iResult;

  iResult = -1;

  if ( m_dMin == m_dMax &&
       m_dMin == val.m_dMin &&
       m_dMax == val.m_dMax )
    {
      iResult = 1;
    }
  else if ( m_dMin > val.m_dMax ||
	    m_dMax < val.m_dMin )
    {
      iResult = 0;
    }

  return iResult;
}

inline Interval sqrt(const Interval &val)
{
  Interval valResult;

  PUSH_ROUND_DOWN();
  valResult.m_dMin = sqrt(val.m_dMin);
  SET_ROUND_UP();
  valResult.m_dMax = sqrt(val.m_dMax);
  POP_ROUND();

  return valResult;
}

inline double dble(const Interval &val)
{
  return (val.m_dMin + val.m_dMax) / 2.0;
}

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline Interval operator+(double d, const Interval &val)
{
  return ( val + d );
}

inline Interval operator-(double d, const Interval &val)
{
  Interval valA(d);

  return ( valA - val );
}

inline Interval operator*(double d, const Interval &val)
{
  return ( val * d );
}

inline Interval operator/(double d, const Interval &val)
{
  Interval valA(d);

  return ( valA / val );
}

inline int operator<(double d, const Interval &val)
{
  return ( val > d );
}

inline int operator>(double d, const Interval &val)
{
  return ( val < d );
}

inline int operator<=(double d, const Interval &val)
{
  return ( val >= d );
}

inline int operator>=(double d, const Interval &val)
{
  return ( val <= d );
}

inline int operator==(double d, const Interval &val)
{
  return ( val == d );
}

inline int operator!=(double d, const Interval &val)
{
  return ( val != d );
}

#endif // __INTERVAL_H__
