#include "lpvector.h"

LPVector::LPVector()
{
  m_uiDim = 0;
  m_uiStride = 1;
  m_pfComps = NULL;

  m_bAllocated = FALSE;
}

LPVector::~LPVector()
{
  Uninit();
}

VOID LPVector::Init(UINT_32 uiDim)
{
  ASSERT( !m_bAllocated );
  ASSERT( uiDim > 0 );

  m_uiDim = uiDim;
  m_uiStride = 1;

  m_pfComps = new FLOAT[m_uiDim];
  ASSERT( m_pfComps != NULL );
  m_bAllocated = TRUE;

  Null();
}

VOID LPVector::Init(UINT_32 uiDim, UINT_32 uiStride, FLOAT *pfComps)
{
  ASSERT( !m_bAllocated );
  ASSERT( uiDim > 0 );
  ASSERT( uiStride > 0 );
  ASSERT( pfComps != NULL );

  m_uiDim = uiDim;
  m_uiStride = uiStride;

  m_pfComps = pfComps;
}

VOID LPVector::Init(UINT_32 uiDim, FLOAT *pfComps)
{
  Init(uiDim, 1, pfComps);
}

VOID LPVector::Uninit()
{
  if ( m_bAllocated )
    {
      if ( m_pfComps != NULL )
	{
	  delete m_pfComps;
	}
    }

  m_uiDim = 0;
  m_pfComps = NULL;
  m_bAllocated = FALSE;
}

#ifdef __LPVECTOR4D_FAST_IMPLEMENTATION__

FLOAT LPVector::Magnitude() const
{
  return (FLOAT)sqrt((*this).Dot(*this));
}

VOID LPVector::Null()
{
  UINT_32 i, j;

  ASSERT( Valid() );

  j = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] = 0.0;
      j += m_uiStride;
    }
}

BOOL LPVector::Normalize()
{
  FLOAT fMag;
  UINT_32 i, j;

  ASSERT( Valid() );

  fMag = Magnitude();

  if ( fMag <= 0.0f )
    {
      return FALSE;
    }

  j = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] /= fMag;
      j += m_uiStride;
    }

  return TRUE;
}

BOOL LPVector::CompletePlaneEquation(const LPVector &v)
{
  LPVector w;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim > 1 );
  ASSERT( m_uiDim == v.GetDimension() );

  w.Init(m_uiDim-1, m_uiStride, m_pfComps);
  if ( !w.Normalize() )
    {
      return FALSE;
    }

  (*this)[m_uiDim-1] = -((*this).Dot(v));

  return TRUE;
}

LPVector &LPVector::operator=(const LPVector &v)
{
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] = v.m_pfComps[k];
      j += m_uiStride;
      k += v.GetStride();
    }

  return *this;
}

LPVector &LPVector::operator+=(const LPVector &v)
{
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] += v.m_pfComps[k];
      j += m_uiStride;
      k += v.GetStride();
    }

  return *this;
}

LPVector &LPVector::operator-=(const LPVector &v)
{
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] -= v.m_pfComps[k];
      j += m_uiStride;
      k += v.GetStride();
    }

  return *this;
}

LPVector &LPVector::operator*=(const FLOAT f)
{
  UINT_32 i, j;

  ASSERT( Valid() );

  j = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] = m_pfComps[j] * f;
      j += m_uiStride;
    }

  return *this;
}

LPVector &LPVector::operator/=(const FLOAT f)
{
  UINT_32 i, j;

  ASSERT( Valid() );
  ASSERT( f != 0.0 );

  j = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] = m_pfComps[j] / f;
      j += m_uiStride;
    }

  return *this;
}

BOOL LPVector::operator==(const LPVector &v) const
{
  BOOL bEqual;
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  bEqual = TRUE;
  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      bEqual = ( bEqual && 
		 ( (m_pfComps[j] > (v.m_pfComps[k] - EPSILON)) && 
		   (m_pfComps[j] < (v.m_pfComps[k] + EPSILON)) ) );
      j += m_uiStride;
      k += v.GetStride();
    }

  return bEqual;
}

BOOL LPVector::operator!=(const LPVector &v) const
{
  return !(*this == v);
}

FLOAT LPVector::Dot(const LPVector &v) const
{
  FLOAT fResult;
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  fResult = 0.0;
  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      fResult += m_pfComps[j] * v.m_pfComps[k];
      j += m_uiStride;
      k += v.GetStride();
    }

  return fResult;
}

VOID LPVector::Display() const
{
  UINT_32 i, j;

  ASSERT( Valid() );

  printf("< ");
  j = 0;
  for ( i = 0; i < m_uiDim; i++)
    {
      printf("%3.3f ", m_pfComps[j]);
      //printf("%.30f ", m_pfComps[j]);
      j += m_uiStride;
    }
  printf(">\n");
}

#else // __LPVECTOR4D_FAST_IMPLEMENTATION__

FLOAT LPVector::Magnitude() const
{
  return sqrt((*this).Dot(*this));
}

VOID LPVector::Null()
{
  UINT_32 i;

  ASSERT( Valid() );

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[i] = 0.0;
    }
}

BOOL LPVector::Normalize()
{
  FLOAT fMag;
  UINT_32 i;

  ASSERT( Valid() );

  fMag = Magnitude();

  if ( fMag <= 0.0f )
    {
      return FALSE;
    }

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[j] /= fMag;
    }

  return TRUE;
}

BOOL LPVector::CompletePlaneEquation(const LPVector &v)
{
  LPVector w;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim > 1 );
  ASSERT( m_uiDim == v.GetDimension() );

  w.Init(m_uiDim-1, m_uiStride, m_pfComps);
  if ( !w.Normalize() )
    {
      return FALSE;
    }

  (*this)[m_uiDim-1] = -((*this).Dot(v));

  return TRUE;
}

LPVector &LPVector::operator=(const LPVector &v)
{
  UINT_32 i;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[i] = v(i);
    }

  return *this;
}

LPVector &LPVector::operator+=(const LPVector &v)
{
  UINT_32 i;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[i] = (*this)(i) + v(i);
    }

  return *this;
}

LPVector &LPVector::operator-=(const LPVector &v)
{
  UINT_32 i;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[i] = (*this)(i) - v(i);
    }

  return *this;
}

LPVector &LPVector::operator*=(const FLOAT f)
{
  UINT_32 i;

  ASSERT( Valid() );

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[i] = (*this)(i) * f;
    }

  return *this;
}

LPVector &LPVector::operator/=(const FLOAT f)
{
  UINT_32 i;

  ASSERT( Valid() );

  for ( i = 0; i < m_uiDim; i++ )
    {
      (*this)[i] = (*this)(i) / f;
    }

  return *this;
}

BOOL LPVector::operator==(const LPVector &v) const
{
  BOOL bEqual;
  UINT_32 i;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  bEqual = TRUE;
  for ( i = 0; i < m_uiDim; i++ )
    {
      bEqual = ( bEqual && 
		 ( ((*this)(i) > (v(i) - EPSILON)) && 
		   ((*this)(i) < (v(i) + EPSILON)) ) );
    }

  return bEqual;
}

BOOL LPVector::operator!=(const LPVector &v) const
{
  return !(*this == v);
}

FLOAT LPVector::Dot(const LPVector &v) const
{
  FLOAT fResult;
  UINT_32 i;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim == v.GetDimension() );

  fResult = 0.0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      fResult += (*this)(i) * v(i); 
    }

  return fResult;
}

VOID LPVector::Display() const
{
  UINT_32 i;

  ASSERT( Valid() );

  printf("< ");
  for ( i = 0; i < m_uiDim; i++)
    {
      printf("%3.3f ", (*this)(i));
    }
  printf(">\n");
}

#endif // __LPVECTOR4D_FAST_IMPLEMENTATION__
