#ifndef __LPVECTOR4D_H__
#define __LPVECTOR4D_H__

#include "genlib.h"
//#include "mathlibx.h"

#define __LPVECTOR4D_FAST_IMPLEMENTATION__

//////////////////////////////////////////////////
// Operations Supported
//
// LPVector u, v, w;   // Vectors
// FLOAT f;          // Scalars
// BOOL b;           // Booleans
//
// f = v[0];       // LHS Index Operation
// v[0] = f;       // LHS Index Operation
// f = v(0);       // RHS Index Operation
// v(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// u = v;
// u += v;
// u -= v;
// u *= f;         // Scalar Multiplication
// u /= f;         // Scalar Division
// u = -v;
// u = v + w;
// u = v - w;
// f = v.Dot(w);   // Dot Product
// u = v * f;      // Scalar Multiplication
// u = f * v;      // Scalar Multiplication
// u = v / f;      // Scalar Division
// u = f / v;      // Scalar Inversion
//
// Logic
// b = (u == v); // Quasi Equal allows for floating point inaccuracy
// b = (u != v); // Quasi Not Equal allows for floating point inaccuracy

//////////////////////////////////////////////////////////////////////
// LPVector Class
//

class LPVector
{
public:
  LPVector();
  ~LPVector();
  VOID Init(const UINT_32 uiDim);
  VOID Init(const UINT_32 uiDim, const UINT_32 uiStride, FLOAT *pfComps);
  VOID Init(const UINT_32 uiDim, FLOAT *pfComps);
  VOID Uninit();

public:
  inline BOOL Initialized() const;
  inline BOOL Valid() const;
  inline VOID Set(const UINT_32 uiDim, const UINT_32 uiStride, 
		  const FLOAT *pfComps);
  inline VOID Set(const UINT_32 uiDim, const FLOAT *pfComps);
  inline VOID Set(const LPVector &v);
  inline VOID Get(const UINT_32 uiDim, const UINT_32 uiStride, 
		  FLOAT *pfComps) const;
  inline VOID Get(const UINT_32 uiDim, FLOAT *pfComps) const;
  inline UINT_32 GetDimension() const;

protected:
  inline UINT_32 GetStride() const;

public:
  FLOAT Magnitude() const;
  VOID Null();
  BOOL Normalize();
  BOOL CompletePlaneEquation(const LPVector &v);

public:
  inline FLOAT &operator[](const UINT_32 uiIndex);
  inline FLOAT operator()(const UINT_32 uiIndex) const;

  LPVector &operator=(const LPVector &v);
  LPVector &operator+=(const LPVector &v);
  LPVector &operator-=(const LPVector &v);
  LPVector &operator*=(const FLOAT f);
  LPVector &operator/=(const FLOAT f);

  BOOL operator==(const LPVector &v) const;
  BOOL operator!=(const LPVector &v) const;

  FLOAT Dot(const LPVector &v) const;

public:
  VOID Display() const;

protected:
  UINT_32 m_uiDim;
  UINT_32 m_uiStride;
  FLOAT *m_pfComps;

  BOOL m_bAllocated;
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

inline LPVector operator-(const LPVector &v);
inline LPVector operator+(const LPVector &v, const LPVector &w);
inline LPVector operator-(const LPVector &v, const LPVector &w);
inline LPVector operator*(const LPVector &v, const FLOAT f);
inline LPVector operator*(const FLOAT f, const LPVector &v);
inline LPVector operator/(const LPVector &v, const FLOAT f);

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline BOOL LPVector::Initialized() const
{
  return ( m_uiDim > 0 &&
	   m_uiStride > 0 &&
	   m_pfComps != NULL );
} 

inline BOOL LPVector::Valid() const
{
  return Initialized();
} 

inline VOID LPVector::Set(const UINT_32 uiDim, const UINT_32 uiStride, 
			  const FLOAT *pfComps)
{
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( pfComps != NULL );
  ASSERT( uiDim >= m_uiDim );
  ASSERT( uiStride > 0 );

  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] = pfComps[k];
      j += m_uiStride;
      k += uiStride;
    }
}

inline VOID LPVector::Set(const UINT_32 uiDim, const FLOAT *pfComps)
{
  Set(uiDim, 1, pfComps);
}

inline VOID LPVector::Set(const LPVector &v)
{
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( v.Valid() );
  ASSERT( m_uiDim <= v.GetDimension() );

  j = 0;
  k = 0;
  for ( i = 0; i < m_uiDim; i++ )
    {
      m_pfComps[j] = v.m_pfComps[k];
      j += m_uiStride;
      k += v.GetStride();
    }
}

inline VOID LPVector::Get(const UINT_32 uiDim, const UINT_32 uiStride, 
			  FLOAT *pfComps) const
{
  UINT_32 i, j, k;

  ASSERT( Valid() );
  ASSERT( pfComps != NULL );
  ASSERT( uiDim <= m_uiDim );
  ASSERT( uiStride > 0 );

  j = 0;
  k = 0;
  for ( i = 0; i < uiDim; i++ )
    {
      pfComps[k] = m_pfComps[j];
      j += m_uiStride;
      k += uiStride;
    }
}

inline VOID LPVector::Get(const UINT_32 uiDim, FLOAT *pfComps) const
{
  Get(uiDim, 1, pfComps);
}

inline UINT_32 LPVector::GetDimension() const
{
  ASSERT( Valid() );

  return m_uiDim;
}

inline UINT_32 LPVector::GetStride() const
{
  ASSERT( Valid() );

  return m_uiStride;
}

inline FLOAT &LPVector::operator[](const UINT_32 uiIndex)
{
  ASSERT( Valid() );
  ASSERT( uiIndex < m_uiDim );

  return m_pfComps[m_uiStride*uiIndex];
}

inline FLOAT LPVector::operator()(const UINT_32 uiIndex) const
{
  ASSERT( Valid() );
  ASSERT( uiIndex < m_uiDim );

  return m_pfComps[m_uiStride*uiIndex];
}

//////////////////////////////////////////////////////////////////////
// External Operator Implementations
//

inline LPVector operator-(const LPVector &v)
{
  LPVector w;

  ASSERT( v.Valid() );

  w.Init(v.GetDimension());
  w -= v;

  return w;
}

inline LPVector operator+(const LPVector &v, const LPVector &w)
{
  LPVector u;

  ASSERT( v.Valid() );
  ASSERT( w.Valid() );
  ASSERT( v.GetDimension() == w.GetDimension() );

  u.Init(v.GetDimension());
  u = v;
  u += w;

  return u;
}

inline LPVector operator-(const LPVector &v, const LPVector &w)
{
  LPVector u;

  ASSERT( v.Valid() );
  ASSERT( w.Valid() );
  ASSERT( v.GetDimension() == w.GetDimension() );

  u.Init(v.GetDimension());
  u = v;
  u -= w;

  return u;
}

inline LPVector operator*(const LPVector &v, const FLOAT f)
{
  LPVector u;

  ASSERT( v.Valid() );

  u.Init(v.GetDimension());
  u = v;
  u *= f;

  return u;
}

inline LPVector operator*(const FLOAT f, const LPVector &v)
{
  LPVector u;

  ASSERT( v.Valid() );

  u.Init(v.GetDimension());
  u = v;
  u *= f;

  return u;
}

inline LPVector operator/(const LPVector &v, const FLOAT f)
{
  LPVector u;

  ASSERT( v.Valid() );
  ASSERT( f != 0.0 );

  u.Init(v.GetDimension());
  u = v;
  u /= f;

  return u;
}

#endif // __LPVECTOR4D_H__

