#ifndef __MATRIX4D_H__
#define __MATRIX4D_H__

#include "genlib.h"
#include "vector.h"
class Quater;


// Column Order Elements
// Based on OpenGL Description
//
//            4 x 4        =                1 x 16
//
// M = | m0  m4  m8  m12 | = { m0, m1, m2,  m3,  m4,  m5,  m6,  m7, 
//     | m1  m5  m9  m13 |     m8, m9, m10, m11, m12, m13, m14, m15 }
//     | m2  m6  m10 m14 |
//     | m3  m7  m11 m15 |

//////////////////////////////////////////////////
// Operations Supported
//
// Matrix mA, mB, mC;// Matrices
// Vector u, v, w;   // Vectors
// FLOAT f;          // Scalars
// BOOL b;           // Booleans
//
// f = v[0];       // LHS Index Operation
// v[0] = f;       // LHS Index Operation
// f = v(0);       // RHS Index Operation
// v(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// u = v;
// u += v;
// u -= v;
// u *= f;         // Scalar Multiplication
// u = -v;
// u = v + w;
// u = v - w;
// u = v * w;      // Cross Product
// u = v.Cross(w); // Cross Product
// f = v.Dot(w);   // Dot Product
// u = v * f;      // Scalar Multiplication
// u = f * v;      // Scalar Multiplication
// u = v / f;      // Scalar Division
// u = f / v;      // Scalar Inversion
//

class Matrix
{
public:
  MATHLIB_API Matrix();
  MATHLIB_API ~Matrix();

public:
  MATHLIB_API inline UINT_32 Index(UINT_32 uiRow, UINT_32 uiCol) const;
  MATHLIB_API inline VOID Set(const MATRIX4D m);
  MATHLIB_API inline VOID Get(MATRIX4D m) const;
  MATHLIB_API inline const FLOAT *Get() const;
  MATHLIB_API inline VOID Set(UINT_32 uiRow, UINT_32 uiCol, FLOAT f);
  MATHLIB_API inline FLOAT Get(UINT_32 uiRow, UINT_32 uiCol) const;

  // Initialization Methods
public:
  MATHLIB_API VOID SetNull();
  MATHLIB_API VOID SetIdentity();
  MATHLIB_API VOID SetCross(const Vector &v);
  MATHLIB_API VOID SetTranslate(const Vector &v);
  MATHLIB_API VOID SetTwist(const Vector &vV, const Vector &vW, DOUBLE dTheta);
  MATHLIB_API VOID SetTwistInverse(const Matrix &m);
  MATHLIB_API VOID SetRotate(const Vector &vW, DOUBLE dTheta);
  MATHLIB_API VOID SetRotate(const Vector &vEx, const Vector &vEy, const Vector &vEz);
  MATHLIB_API VOID SetRotate(const Quater &q);
  MATHLIB_API VOID SetRotateX(DOUBLE dTheta);
  MATHLIB_API VOID SetRotateY(DOUBLE dTheta);
  MATHLIB_API VOID SetRotateZ(DOUBLE dTheta);
  MATHLIB_API VOID SetRotateZ(FLOAT sint, FLOAT cost);
  MATHLIB_API VOID SetScale(const Vector &v);
  MATHLIB_API VOID SetScale(const FLOAT f);
  MATHLIB_API VOID SetShearXY(const Vector &v);
  MATHLIB_API VOID SetShearYZ(const Vector &v);
  MATHLIB_API VOID SetShearXZ(const Vector &v);
  MATHLIB_API VOID SetPerspectiveWarp(const FLOAT fZmin);
  MATHLIB_API VOID SetPerspectiveWarpInverse(const FLOAT fZmin);
  MATHLIB_API VOID SetViewPort(const Vector &vVPMin, const Vector &vVPMax, 
		   const Vector &vWWMin, const Vector &vWWMax,
		   FLOAT fDYDX = 1.0);
  // Transform from a World Window (WW) to a screne ViewPort (VP)

  // Decomposition Methods
public:
  MATHLIB_API FLOAT Trace() const;
  MATHLIB_API FLOAT Determinant() const;

  // Operations on this Matrix
public:
  MATHLIB_API VOID Transpose();
  MATHLIB_API VOID Invert();

  // Operations which create new Matrices
public:  
  MATHLIB_API Matrix &operator=(const Matrix& m);
  MATHLIB_API Matrix &operator+=(const Matrix& m);
  MATHLIB_API Matrix &operator-=(const Matrix& m);
  MATHLIB_API Matrix &operator*=(const FLOAT f);
  MATHLIB_API Matrix &operator/=(const FLOAT f);

  MATHLIB_API BOOL operator==(const Matrix &m) const;
  MATHLIB_API BOOL operator!=(const Matrix &m) const;

  MATHLIB_API Matrix operator*(const Matrix& m) const;
  MATHLIB_API Vector operator*(const Vector& v) const;
  MATHLIB_API Point operator*(const Point& pt) const;

public:
  MATHLIB_API VOID Display() const;

protected:
  MATRIX4D m_m; // Column Order Elements
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

MATHLIB_API inline Matrix operator+(const Matrix& mA, const Matrix& mB);
MATHLIB_API inline Matrix operator-(const Matrix& mA, const Matrix& mB);
MATHLIB_API inline Matrix operator*(const Matrix& m, const FLOAT f);
MATHLIB_API inline Matrix operator*(const FLOAT f, const Matrix& m);
MATHLIB_API inline Matrix operator/(const Matrix& m, const FLOAT f);

// Row Operations -- used for transforming normals
MATHLIB_API Vector operator*(const Vector& v, const Matrix &m);


//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

MATHLIB_API inline UINT_32 Matrix::Index(UINT_32 uiRow, UINT_32 uiCol) const
{
  return 4*uiCol + uiRow;
}

MATHLIB_API inline VOID Matrix::Set(const MATRIX4D m)
{
  memcpy(m_m, m, sizeof(MATRIX4D));
}

MATHLIB_API inline VOID Matrix::Get(MATRIX4D m) const
{
  memcpy(m, m_m, sizeof(MATRIX4D));
}

MATHLIB_API inline const FLOAT *Matrix::Get() const
{
  return m_m;
}

MATHLIB_API inline VOID Matrix::Set(UINT_32 uiRow, UINT_32 uiCol, FLOAT f)
{
  ASSERT( uiRow < 4 && uiCol < 4 );
  m_m[Index(uiRow, uiCol)] = f;
}

MATHLIB_API inline FLOAT Matrix::Get(UINT_32 uiRow, UINT_32 uiCol) const
{
  ASSERT( uiRow < 4 && uiCol < 4 );
  return m_m[Index(uiRow, uiCol)];
}

//////////////////////////////////////////////////////////////////////
// Implementation of External Functions
//

MATHLIB_API inline Matrix operator+(const Matrix& mA, const Matrix& mB)
{
  Matrix mResult;

  mResult = mA;
  mResult += mB;

  return mResult;
}

MATHLIB_API inline Matrix operator-(const Matrix& mA, const Matrix& mB)
{
  Matrix mResult;

  mResult = mA;
  mResult -= mB;

  return mResult;
}

MATHLIB_API inline Matrix operator*(const Matrix& m, const FLOAT f)
{
  Matrix mResult;

  mResult = m;
  mResult *= f;

  return mResult;
}

MATHLIB_API inline Matrix operator*(const FLOAT f, const Matrix& m)
{
  Matrix mResult;

  mResult = m;
  mResult *= f;

  return mResult;
}

MATHLIB_API inline Matrix operator/(const Matrix& m, const FLOAT f)
{
  Matrix mResult;

  mResult = m;
  mResult /= f;

  return mResult;
}

#endif // __MATRIX4D_H__










