#ifndef __QEEDGE_H__
#define __QEEDGE_H__

#include "qemem.h"

#include "qevertex.h"
#include "qefacet.h"

class QEWedge;

//////////////////////////////////////////////////////////////////////
// QEEdge Class
//

class QEEdge
{
//////////////////////////////////////////////////
// Friends
//
public:

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  enum VertexType { vtStart, vtSplit, vtMerge, vtEnd, vtLeft, vtRight, vtSize };

//////////////////////////////////////////////////
// Initialization
//
public:
  QEEdge();
  virtual ~QEEdge();
  virtual VOID Init();
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID Preprocess();

  QEEdge::VertexType CalculateVertexType();
  FLOAT CalculateScanX(FLOAT fY);

  Point CalculateMidpointCatmullClark();

protected:

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline QEVertex *GetVertex();
  inline VOID SetVertex(QEVertex *pvx);
  inline QEWedge *GetWedge();
  inline VOID SetWedge(QEWedge *pwe);
  inline QEFacet *GetFacet();
  inline VOID SetFacet(QEFacet *pfa);
  inline QEEdge *GetEdge();
  inline VOID SetEdge(QEEdge *pe);

  inline QEVertex *GetNeighborVertex();
  inline VOID SetNeighborVertex(QEVertex *pvx);
  inline QEFacet *GetNeighborFacet();
  inline VOID SetNeighborFacet(QEFacet *pfa);

  inline QEEdge *GetEdgePrev(UINT_32 uiDual);
  inline VOID SetEdgePrev(UINT_32 uiDual, QEEdge *pe);
  inline QEEdge *GetEdgeNext(UINT_32 uiDual);
  inline VOID SetEdgeNext(UINT_32 uiDual, QEEdge *pe);

  inline QEEdge *GetEdge(UINT_32 uiDual, UINT_32 uiIndex);
  inline VOID SetEdge(UINT_32 uiDual, UINT_32 uiIndex, QEEdge *pe);

  inline FLOAT GetKnotInterval();
  inline VOID SetKnotInterval(FLOAT fKnotInterval);

  inline VertexType GetVertexType();
  inline VOID SetVertexType(VertexType eVertexType);

  inline QEEdge *GetForwardEdge();
  inline VOID SetForwardEdge(QEEdge *pe);
  inline QEVertex *GetForwardVertex();
  inline VOID SetForwardVertex(QEVertex *pvx);

  inline CHeapEntry<VOID*, FLOAT> *GetHeapEntry();
  inline VOID SetHeapEntry(CHeapEntry<VOID*, FLOAT> *phe);

  inline QEEdge *GetPrevLink();
  inline VOID SetPrevLink(QEEdge *pe);
  inline QEEdge *GetNextLink();
  inline VOID SetNextLink(QEEdge *pe);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // Dual space or primal space (i.e. vertex or facet)
  //   0 = Primal Space or Facet
  //   1 = Dual Space or Vertex
  // Prev or Next
  //   0 = Prev
  //   1 = Next
  QEVertex *m_pQEVertex;
  QEWedge *m_pQEWedge;
  QEFacet *m_pQEFacet;
  QEEdge *m_pQEEdge;

  QEEdge *m_pppQEEdge[2][2];

  // Subdivision
  FLOAT m_fKnotInterval;

  // Y Monotone Variables
  VertexType m_eVertexType;

  // Forwarding Pointers
  QEEdge *m_peForward;
  QEVertex *m_pvxForward;

  // Heap for selective processes
  CHeapEntry<VOID*, FLOAT> *m_phe;			 

  // Listing in a QEMesh or on a Free List
  QEEdge *m_peLink[2];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline QEVertex *QEEdge::GetVertex()
{
  return m_pQEVertex;
}

inline VOID QEEdge::SetVertex(QEVertex *pvx)
{
  m_pQEVertex = pvx;
  m_pQEVertex->SetEdge(this);
}

inline QEWedge *QEEdge::GetWedge()
{
  return m_pQEWedge;
}

inline VOID QEEdge::SetWedge(QEWedge *pwe)
{
  m_pQEWedge = pwe;
}

inline QEFacet *QEEdge::GetFacet()
{
  return m_pQEFacet;
}

inline VOID QEEdge::SetFacet(QEFacet *pfa)
{
  m_pQEFacet = pfa;
  m_pQEFacet->SetEdge(this);
}

inline QEEdge *QEEdge::GetEdge()
{
  return m_pQEEdge;
}

inline VOID QEEdge::SetEdge(QEEdge *pe)
{
  m_pQEEdge = pe;
}

inline QEVertex *QEEdge::GetNeighborVertex()
{
  return m_pQEEdge->GetVertex();
}

inline VOID QEEdge::SetNeighborVertex(QEVertex *pvx)
{
  m_pQEEdge->SetVertex(pvx);
}

inline QEFacet *QEEdge::GetNeighborFacet()
{
  return m_pQEEdge->GetFacet();
}

inline VOID QEEdge::SetNeighborFacet(QEFacet *pfa)
{
  m_pQEEdge->SetFacet(pfa);
}

inline QEEdge *QEEdge::GetEdgePrev(UINT_32 uiDual)
{
  return GetEdge(uiDual, 0);
}

inline VOID QEEdge::SetEdgePrev(UINT_32 uiDual, QEEdge *pe)
{
  SetEdge(uiDual, 0, pe);
}

inline QEEdge *QEEdge::GetEdgeNext(UINT_32 uiDual)
{
  return GetEdge(uiDual, 1);
}

inline VOID QEEdge::SetEdgeNext(UINT_32 uiDual, QEEdge *pe)
{
  SetEdge(uiDual, 1, pe);
}

inline QEEdge *QEEdge::GetEdge(UINT_32 uiDual, UINT_32 uiIndex)
{
  ASSERT( uiDual < 2 && uiIndex < 2 );
  return m_pppQEEdge[uiDual][uiIndex];
}

inline VOID QEEdge::SetEdge(UINT_32 uiDual, UINT_32 uiIndex, QEEdge *pe)
{
  ASSERT( uiDual < 2 && uiIndex < 2 );
  m_pppQEEdge[uiDual][uiIndex] = pe;
}

inline FLOAT QEEdge::GetKnotInterval()
{
  return m_fKnotInterval;
}

inline VOID QEEdge::SetKnotInterval(FLOAT fKnotInterval)
{
  m_fKnotInterval = fKnotInterval;
}

inline QEEdge::VertexType QEEdge::GetVertexType()
{
  return m_eVertexType;
}

inline VOID QEEdge::SetVertexType(QEEdge::VertexType eVertexType)
{
  ASSERT( eVertexType < vtSize );
  m_eVertexType = eVertexType;
}

inline QEEdge *QEEdge::GetForwardEdge()
{
  return m_peForward;
}

inline VOID QEEdge::SetForwardEdge(QEEdge *pe)
{
  m_peForward = pe;
}

inline QEVertex *QEEdge::GetForwardVertex()
{
  return m_pvxForward;
}

inline VOID QEEdge::SetForwardVertex(QEVertex *pvx)
{
  m_pvxForward = pvx;
}

inline CHeapEntry<VOID*, FLOAT> *QEEdge::GetHeapEntry()
{
  return m_phe;
}

inline VOID QEEdge::SetHeapEntry(CHeapEntry<VOID*, FLOAT> *phe)
{
  m_phe = phe;
}

inline QEEdge *QEEdge::GetPrevLink()
{
  return m_peLink[0];
}

inline VOID QEEdge::SetPrevLink(QEEdge *pe)
{
  m_peLink[0] = pe;
}

inline QEEdge *QEEdge::GetNextLink()
{
  return m_peLink[1];
}

inline VOID QEEdge::SetNextLink(QEEdge *pe)
{
  m_peLink[1] = pe;
}

#endif // __QEEDGE_H__

