#include "qefacet.h"

#include "qevertex.h"
#include "qeedge.h"
#include "qewedge.h"
#include "qemesh.h"

//////////////////////////////////////////////////////////////////////
// QEFacet Class
//

QEFacet::QEFacet()
{
  UINT_32 i;

  m_uiID = 0;
  m_uiGeneration = 0;
  m_uiBirth = 0;

  m_pQEEdge = NULL;

  m_pQEFacetNode = NULL;

  m_heapQEVertex.Init();
  m_pCHeapEntry = NULL;

  m_pfaForward = NULL;
  m_pvxForward = NULL;
  m_peForward = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pphe[i] = NULL;
      m_pfaLink[i] = NULL;
    }

  m_pCSLIDESurface = NULL;
}

QEFacet::~QEFacet()
{
  Uninit();
}

VOID QEFacet::Init()
{
  UINT_32 i;

  m_uiID = 0;
  m_uiGeneration = 0;
  m_uiBirth = 0;

  m_pQEEdge = NULL;

  m_ptCentroid.Null();
  m_vNormal.Null();

  m_pQEFacetNode = NULL;

  ASSERT( m_heapQEVertex.GetSize() == 0 );
  ASSERT( m_pCHeapEntry == NULL );

  m_pfaForward = NULL;
  m_pvxForward = NULL;
  m_peForward = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pphe[i] = NULL;
      ASSERT( m_pfaLink[i] == NULL );
    }

  m_pCSLIDESurface = NULL;
}

VOID QEFacet::Uninit()
{
  UINT_32 i;

  m_uiID = 0;
  m_uiGeneration = 0;
  m_uiBirth = 0;

  m_pQEEdge = NULL;

  m_pQEFacetNode = NULL;

  ASSERT( m_heapQEVertex.GetSize() == 0 );
  ASSERT( m_pCHeapEntry == NULL );

  m_pfaForward = NULL;
  m_pvxForward = NULL;
  m_peForward = NULL;

  for ( i = 0; i < 2; i++ )
    {
      m_pphe[i] = NULL;
      m_pfaLink[i] = NULL;
    }
}

BOOL QEFacet::Initialized()
{
  return TRUE;
}

BOOL QEFacet::Valid()
{
  return TRUE;
}

VOID QEFacet::Preprocess()
{
  CalculateCentroid();
  CalculateNormal();
  
  m_ptForward = CalculateCircumcenter();
  //m_ptForward = CalculatePointCatmullClark();
}

VOID QEFacet::PreprocessLimit()
{
  m_ptCentroid = CalculateLimitCentroid();
  m_vNormal = CalculateLimitNormal(m_ptCentroid);
}

BOOL QEFacet::CheckOutside(QEVertex *pvx)
{
  ASSERT( pvx != NULL );

  return CalculateDistance(pvx->GetPoint()) > 0.0;
}

FLOAT QEFacet::CalculateDistance(const Point &pt)
{
  Vector v;
  
  v = pt;
  v[W] = 1.0;
  
  return v.Dot(m_vNormal);
}

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

VOID QEFacet::CalculateNormal()
{
  QEFacetIter iter;
  QEVertex *pQEVertex;
  Vector ptStart, ptPrev, ptCurr;

  m_vNormal.Null();

  iter.Init(this);

  pQEVertex = iter.PeekFirstVertex(); 
  if ( pQEVertex == NULL )
    {
      fprintf(stderr, "QEFacet Warning: Null normal because no points for facet\n");
      return;
    }
  ptStart = pQEVertex->GetPoint() - m_ptCentroid;
  ptPrev = ptStart;

  for ( pQEVertex = iter.PeekNextVertex(); 
	pQEVertex != NULL;
	pQEVertex = iter.PeekNextVertex() )
    {
      ptCurr = pQEVertex->GetPoint() - m_ptCentroid;

      m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !m_vNormal.Normalize() )
    {
      fprintf(stderr, "QEFacet Warning: Null normal because of zero area facet\n");
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  m_vNormal.CompletePlaneEquation(m_ptCentroid);
}

VOID QEFacet::CalculateCentroid()
{
  QEFacetIter iter;
  UINT_32 i;
  QEVertex *pQEVertex;

  iter.Init(this);

  m_ptCentroid.Null();
  m_ptCentroid[W] = 0.0;

  i = 0;
  for ( pQEVertex = iter.PeekFirstVertex();
	pQEVertex != NULL;
	pQEVertex = iter.PeekNextVertex() )
    {
      m_ptCentroid = m_ptCentroid + pQEVertex->GetPoint();
      i++;
    }

  if ( i > 0 )
    {
      m_ptCentroid = m_ptCentroid / (FLOAT)i;
    }

  m_ptCentroid[W] = 1.0;
}

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

Vector QEFacet::CalculateLimitNormal(const Point &ptCentroid)
{
  QEFacetIter iter;
  QEVertex *pQEVertex;
  Vector ptStart, ptPrev, ptCurr;
  Vector vNormal;

  vNormal.Null();

  iter.Init(this);

  pQEVertex = iter.PeekFirstVertex(); 
  if ( pQEVertex == NULL )
    {
      fprintf(stderr, "QEFacet Warning: Null normal because no points for limit facet\n");
      return vNormal;
    }
  ptStart = pQEVertex->GetLimitPosition() - ptCentroid;
  ptPrev = ptStart;

  for ( pQEVertex = iter.PeekNextVertex(); 
	pQEVertex != NULL;
	pQEVertex = iter.PeekNextVertex() )
    {
      ptCurr = pQEVertex->GetLimitPosition() - ptCentroid;

      vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vNormal.Normalize() )
    {
      fprintf(stderr, "QEFacet Warning: Null normal because of zero area limit facet\n");
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vNormal.CompletePlaneEquation(ptCentroid);

  return vNormal;
}

Point QEFacet::CalculateLimitCentroid()
{
  QEFacetIter iter;
  UINT_32 i;
  QEVertex *pQEVertex;
  Point ptCentroid;

  iter.Init(this);

  ptCentroid.Null();
  ptCentroid[W] = 0.0;

  i = 0;
  for ( pQEVertex = iter.PeekFirstVertex();
	pQEVertex != NULL;
	pQEVertex = iter.PeekNextVertex() )
    {
      ptCentroid = ptCentroid + pQEVertex->GetLimitPosition();
      i++;
    }

  if ( i > 0 )
    {
      ptCentroid = ptCentroid / (FLOAT)i;
    }

  ptCentroid[W] = 1.0;

  return ptCentroid;
}

VOID QEFacet::CalculateLimitCentroidAndNormal(Point &ptCentroid, Vector &vNormal)
{
  ptCentroid = CalculateLimitCentroid();
  vNormal = CalculateLimitNormal(ptCentroid);
}

VOID QEFacet::CalculateBound()
{
}

Point QEFacet::CalculateCircumcenter()
{
  QEFacetIter iter;
  QEVertex *ppvx[3], *pvx;
  UINT_32 i;
  Point pt;

  iter.Init(this);
  for ( i = 0, pvx = iter.PeekFirstVertex();
	i < 3 && pvx != NULL;
	i++, pvx = iter.PeekNextVertex() )
    {
      ppvx[i] = pvx;
    }

  if ( i == 3 )
    {
      pt = QEMesh::CalculateCircumcenter(ppvx[0]->GetPoint(), 
					 ppvx[1]->GetPoint(), 
					 ppvx[2]->GetPoint());
    }
  else
    {
      pt.Null();
    }

  return pt;
}

Point QEFacet::CalculatePointCatmullClark()
{
  Point pt;
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  FLOAT pfWeight[2], fWeight, fWeights;

  // Count the sides
  fWeights = 0.0;
  pt.Null();

  iterQEFacet.Init(this);
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pfWeight[0] = 0.0;
      pfWeight[1] = 0.0;

      peCurr = pe->GetEdge();
      // d i+1,i,0
      pfWeight[0] += peCurr->GetKnotInterval();
      // d i+1,i,2
      pfWeight[0] += peCurr->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();
      // d i+1,i,-2
      pfWeight[0] += peCurr->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();

      peCurr = peCurr->GetEdgePrev(1)->GetEdge();
      // d i+2,i+1,0
      pfWeight[1] += peCurr->GetKnotInterval();
      // d i+2,i+1,2
      pfWeight[1] += peCurr->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();
      // d i+2,i+1,-2
      pfWeight[1] += peCurr->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();
 
      peCurr = pe->GetEdgePrev(0);
      // d i-1,i,0
      pfWeight[1] += peCurr->GetKnotInterval();
      // d i-1,i,-2
      pfWeight[1] += peCurr->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();
      // d i-1,i,2
      pfWeight[1] += peCurr->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();

      peCurr = peCurr->GetEdgePrev(0);
      // d i-2,i-1,0
      pfWeight[0] += peCurr->GetKnotInterval();
      // d i-2,i-1,-2
      pfWeight[0] += peCurr->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();
      // d i-2,i-1,2
      pfWeight[0] += peCurr->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();
 
      fWeight = pfWeight[0] * pfWeight[1];
      pt += fWeight * pe->GetVertex()->GetPoint();
      fWeights += fWeight;
    }
  
  pt /= fWeights;
  pt[3] = 1.0;

  return pt;
}

UINT_32 QEFacet::CalculateBoundaryEdges()
{
  QEFacetIter iter;
  QEEdge *pe;
  UINT_32 uiBoundaryEdges;

  iter.Init(this);

  uiBoundaryEdges = 0;
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      if ( pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
	{
	  uiBoundaryEdges++;
	}
    }

  return uiBoundaryEdges;
}

FLOAT QEFacet::CalculateCDTSubdivideDistance(const Point &pt)
{
  QEFacetIter iter;
  QEEdge *pe;
  Vector v, w;
  FLOAT fDistance;

  iter.Init(this);

  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      if ( pe->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
	{
	  v = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint(); 
	  v.Set(-v(1), v(0), 0.0);
	  if ( !v.Normalize() )
	    {
	      v = pe->GetVertex()->GetPoint() - pt;
	      fDistance = v.Magnitude();
	    }
	  else
	    {
	      v.CompletePlaneEquation(pe->GetVertex()->GetPoint());
	      w = pt;
	      w[3] = 1.0;
	      fDistance = v.Dot(w);
	    }

	  return fDistance;
	}
    }

  ASSERT( FALSE );
  return 0.0;
}

/*
VOID QEFacet::Render()
{
  RenderGeometry();
}

VOID QEFacet::RenderGeometry()
{
  QEFacetIter iter;
  QEEdge *pe;

  gilNormal3fv(m_vNormal.Get());
  gilBegin(GL_POLYGON);
  //gilBegin(GL_LINE_LOOP);
  {
    iter.Init(this);
    
    for ( pe = iter.PeekFirstEdge();
	  pe != NULL;
	  pe = iter.PeekNextEdge() )
      {
	pe->GetVertex()->Render(pe->GetWedge(), TRUE);
      }
  }
  gilEnd();
}

VOID QEFacet::RenderNormals()
{
  Vector vNormal;
  Point pt;

  vNormal = m_vNormal;
  vNormal[3] = 0.0;

  gilBegin(GL_LINES);
  {    
    gilVertex3fv(m_ptCentroid.Get());
    pt = m_ptCentroid + vNormal;
    gilVertex3fv(pt.Get());
  }
  gilEnd();
}
*/
VOID QEFacet::Display()
{
  QEFacetIter iter;
  QEEdge *pe;
  QEVertex *pvxCurr;

  fprintf(stderr, "    START: Facet %p: ", this);
  m_vNormal.Display(stderr);

  fprintf(stderr, "      START: Vertices\n");
  iter.Init(this);
  
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      pvxCurr = pe->GetVertex();
      fprintf(stderr, "    Edge %p: ", pe);
      pvxCurr->Display();
    }
  fprintf(stderr, "      END:   Vertices\n");

  fprintf(stderr, "    END  : Facet %p: ", this);
  m_vNormal.Display(stderr);
}

QEFacet::FacetType QEFacet::GetType()
{
  return ftFacet;
}

//////////////////////////////////////////////////////////////////////
// QEFacet Class
//

QEFacetIter::QEFacetIter()
{
  m_pQEFacet = NULL;
  m_peHead = NULL;
  m_peCurr = NULL;
}

QEFacetIter::~QEFacetIter()
{
  Uninit();
}

VOID QEFacetIter::Init(QEFacet *pQEFacet)
{
  ASSERT( pQEFacet != NULL );

  m_pQEFacet = pQEFacet;

  m_peHead = NULL;
  m_peCurr = NULL;
}

VOID QEFacetIter::Uninit()
{
  m_pQEFacet = NULL;
  m_peHead = NULL;
  m_peCurr = NULL;
}

BOOL QEFacetIter::Initialized()
{
  return m_pQEFacet != NULL;
}

BOOL QEFacetIter::Valid()
{
  return Initialized();
}

QEVertex *QEFacetIter::PeekFirstVertex()
{
  PeekFirstEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetVertex();
}

QEVertex *QEFacetIter::PeekNextVertex()
{
  PeekNextEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetVertex();
}

QEVertex *QEFacetIter::PeekNextWrapVertex()
{
  PeekNextWrapEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetVertex();
}

QEVertex *QEFacetIter::PeekLastVertex()
{
  PeekLastEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetVertex();
}

QEVertex *QEFacetIter::PeekPrevVertex()
{
  PeekPrevEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetVertex();
}

QEVertex *QEFacetIter::PeekPrevWrapVertex()
{
  PeekPrevWrapEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetVertex();
}

QEEdge *QEFacetIter::PeekFirstEdge()
{
  m_peHead = m_pQEFacet->m_pQEEdge;
  m_peCurr = m_peHead;

  return m_peCurr;
}

QEEdge *QEFacetIter::PeekNextEdge()
{
  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  m_peCurr = m_peCurr->GetEdgeNext(0);
  if ( m_peCurr == m_peHead )
    {
      m_peCurr = NULL;
    }

  return m_peCurr;
}

QEEdge *QEFacetIter::PeekNextWrapEdge()
{
  m_peCurr = m_peCurr->GetEdgeNext(0);

  return m_peCurr;
}

QEEdge *QEFacetIter::PeekLastEdge()
{
  m_peHead = m_pQEFacet->m_pQEEdge;
  m_peCurr = m_peHead->GetEdgePrev(0);

  return m_peCurr;
}

QEEdge *QEFacetIter::PeekPrevEdge()
{
  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  if ( m_peCurr == m_peHead )
    {
      m_peCurr = NULL;
      return NULL;
    }
  m_peCurr = m_peCurr->GetEdgePrev(0);

  return m_peCurr;
}

QEEdge *QEFacetIter::PeekPrevWrapEdge()
{
  m_peCurr = m_peCurr->GetEdgePrev(0);

  return m_peCurr;
}

QEWedge *QEFacetIter::PeekFirstWedge()
{
  PeekFirstEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetWedge();
}

QEWedge *QEFacetIter::PeekNextWedge()
{
  PeekNextEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetWedge();
}

QEWedge *QEFacetIter::PeekNextWrapWedge()
{
  PeekNextWrapEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetWedge();
}

QEWedge *QEFacetIter::PeekLastWedge()
{
  PeekLastEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetWedge();
}

QEWedge *QEFacetIter::PeekPrevWedge()
{
  PeekPrevEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetWedge();
}

QEWedge *QEFacetIter::PeekPrevWrapWedge()
{
  PeekPrevWrapEdge();

  if ( m_peCurr == NULL )
    {
      return NULL;
    }

  return m_peCurr->GetWedge();
}

