#ifndef __QEFACET_H__
#define __QEFACET_H__

#include "datalib.h"
#include "qemem.h"

class CSLIDESurface;

class QEVertex;
class QEEdge;
class QEWedge;

class QEFacetNode;

class QEFacetIter;

//////////////////////////////////////////////////////////////////////
// QEFacet Class
//

class QEFacet
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend class QEFacetIter;

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  enum FacetType { ftFacet, ftTriangle, ftSize };

//////////////////////////////////////////////////
// Initialization
//
public:
  QEFacet();
  virtual ~QEFacet();
  virtual VOID Init();
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID Preprocess();
  VOID PreprocessLimit();
  BOOL CheckOutside(QEVertex *pvx);
  FLOAT CalculateDistance(const Point &pt);

  Point CalculateCircumcenter();
  Point CalculatePointCatmullClark();

  UINT_32 CalculateBoundaryEdges();
  FLOAT CalculateCDTSubdivideDistance(const Point &pt);

  VOID CalculateLimitCentroidAndNormal(Point &ptCentroid, Vector &vNormal);

protected:
  VOID CalculateNormal();
  VOID CalculateCentroid();
  Vector CalculateLimitNormal(const Point &ptCentroid);
  Point CalculateLimitCentroid();
  virtual VOID CalculateBound();


//////////////////////////////////////////////////
// Rendering
//
public:
//  virtual VOID Render();
//  virtual VOID RenderGeometry();
//  virtual VOID RenderNormals();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID Display();

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline UINT_32 GetID();
  inline VOID SetID(UINT_32 uiID);
  inline UINT_32 GetGeneration();
  inline VOID SetGeneration(UINT_32 uiGeneration);
  inline UINT_32 GetBirth();
  inline VOID SetBirth(UINT_32 uiBirth);

  virtual FacetType GetType();
  inline QEEdge *GetEdge();
  inline VOID SetEdge(QEEdge *pe);
  inline const Vector &GetNormal();
  inline const Point &GetCentroid();
  inline const Point &GetForwardPoint();

  // Incremental Delaunay
  inline QEFacetNode *GetFacetNode();
  inline VOID SetFacetNode(QEFacetNode *pfn);

  // Convex Hull
  inline VOID AddOutsideVertex(FLOAT fDistance, QEVertex *pvx);
  inline QEVertex *PeekFurthestOutsideVertex(FLOAT *pfDistance = NULL);
  inline QEVertex *RemoveFurthestOutsideVertex(FLOAT *pfDistance = NULL);
  inline QEVertex *RemoveOutsideVertex(FLOAT *pfDistance = NULL);

  inline CHeapEntry<QEFacet*, FLOAT> *GetPriority();
  inline VOID SetPriority(CHeapEntry<QEFacet*, FLOAT> *pCHeapEntry);

  // Forward
  inline QEFacet *GetForwardFacet();
  inline VOID SetForwardFacet(QEFacet *pfa);
  inline QEVertex *GetForwardVertex();
  inline VOID SetForwardVertex(QEVertex *pvx);
  inline QEEdge *GetForwardEdge();
  inline VOID SetForwardEdge(QEEdge *pe);

  // Heap
  inline CHeapEntry<VOID*, FLOAT> *GetHeapEntry();
  inline VOID SetHeapEntry(CHeapEntry<VOID*, FLOAT> *phe);
  inline CHeapEntry<VOID*, FLOAT> *GetHeapEntryReverse();
  inline VOID SetHeapEntryReverse(CHeapEntry<VOID*, FLOAT> *phe);

  // Linked List
  inline QEFacet *GetPrevLink();
  inline VOID SetPrevLink(QEFacet *pfa);
  inline QEFacet *GetNextLink();
  inline VOID SetNextLink(QEFacet *pfa);

  inline CSLIDESurface *GetSurface();
  inline VOID SetSurface(CSLIDESurface *pCSLIDESurface);

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  UINT_32 m_uiID;
  // Pack the generation info into a single word
  UINT_32 m_uiGeneration : 16;
  UINT_32 m_uiBirth      : 16;
  CSLIDESurface *m_pCSLIDESurface;

  QEEdge *m_pQEEdge;

  Point m_ptCentroid;
  Vector m_vNormal;

  // Incremental Delaunay Triangulation
  QEFacetNode *m_pQEFacetNode;

  // Convex Hull Variables
  CHeap<QEVertex*, FLOAT> m_heapQEVertex;
  CHeapEntry<QEFacet*, FLOAT> *m_pCHeapEntry;

  // Voronoi or Subdivision Variables
  Point m_ptForward;

  // Forwarding Variables
  QEFacet *m_pfaForward;
  QEVertex *m_pvxForward;
  QEEdge *m_peForward;

  // Heap for selective processes
  CHeapEntry<VOID*, FLOAT> *m_pphe[2];			 

  // Listing in a QEMesh or on a Free List
  QEFacet *m_pfaLink[2];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline UINT_32 QEFacet::GetID()
{
  return m_uiID;
}

inline VOID QEFacet::SetID(UINT_32 uiID)
{
  m_uiID = uiID;
}

inline UINT_32 QEFacet::GetGeneration()
{
  return m_uiGeneration;
}

inline VOID QEFacet::SetGeneration(UINT_32 uiGeneration)
{
  m_uiGeneration = uiGeneration;
}

inline UINT_32 QEFacet::GetBirth()
{
  return m_uiBirth;
}

inline VOID QEFacet::SetBirth(UINT_32 uiBirth)
{
  m_uiBirth = uiBirth;
}

inline QEEdge *QEFacet::GetEdge()
{
  return m_pQEEdge;
}

inline VOID QEFacet::SetEdge(QEEdge *pe)
{
  m_pQEEdge = pe;
}

inline const Vector &QEFacet::GetNormal()
{
  return m_vNormal;
}

inline const Point &QEFacet::GetCentroid()
{
  return m_ptCentroid;
}

inline const Point &QEFacet::GetForwardPoint()
{
  return m_ptForward;
}

inline QEFacetNode *QEFacet::GetFacetNode()
{
  return m_pQEFacetNode;
}

inline VOID QEFacet::SetFacetNode(QEFacetNode *pfn)
{
  m_pQEFacetNode = pfn;
}

inline VOID QEFacet::AddOutsideVertex(FLOAT fDistance, QEVertex *pvx)
{
  ASSERT( fDistance > 0.0 );
  ASSERT( pvx != NULL );

  m_heapQEVertex.Insert(fDistance, pvx);
}

inline QEVertex *QEFacet::PeekFurthestOutsideVertex(FLOAT *pfDistance/* = NULL*/)
{
  return m_heapQEVertex.PeekTop(pfDistance);
}

inline QEVertex *QEFacet::RemoveFurthestOutsideVertex(FLOAT *pfDistance/* = NULL*/)
{
  return m_heapQEVertex.RemoveTop(pfDistance);
}

inline QEVertex *QEFacet::RemoveOutsideVertex(FLOAT *pfDistance/* = NULL*/)
{
  return m_heapQEVertex.RemoveLast(pfDistance);
}

inline CHeapEntry<QEFacet*, FLOAT> *QEFacet::GetPriority()
{
  return m_pCHeapEntry;
}

inline VOID QEFacet::SetPriority(CHeapEntry<QEFacet*, FLOAT> *pCHeapEntry)
{
  m_pCHeapEntry = pCHeapEntry;
}

inline QEFacet *QEFacet::GetForwardFacet()
{
  return m_pfaForward;
}

inline VOID QEFacet::SetForwardFacet(QEFacet *pfa)
{
  m_pfaForward = pfa;
}

inline QEVertex *QEFacet::GetForwardVertex()
{
  return m_pvxForward;
}

inline VOID QEFacet::SetForwardVertex(QEVertex *pvx)
{
  m_pvxForward = pvx;
}

inline QEEdge *QEFacet::GetForwardEdge()
{
  return m_peForward;
}

inline VOID QEFacet::SetForwardEdge(QEEdge *pe)
{
  m_peForward = pe;
}

inline CHeapEntry<VOID*, FLOAT> *QEFacet::GetHeapEntry()
{
  return m_pphe[0];
}

inline VOID QEFacet::SetHeapEntry(CHeapEntry<VOID*, FLOAT> *phe)
{
  m_pphe[0] = phe;
}

inline CHeapEntry<VOID*, FLOAT> *QEFacet::GetHeapEntryReverse()
{
  return m_pphe[1];
}

inline VOID QEFacet::SetHeapEntryReverse(CHeapEntry<VOID*, FLOAT> *phe)
{
  m_pphe[1] = phe;
}

inline QEFacet *QEFacet::GetPrevLink()
{
  return m_pfaLink[0];
}

inline VOID QEFacet::SetPrevLink(QEFacet *pfa)
{
  m_pfaLink[0] = pfa;
}

inline QEFacet *QEFacet::GetNextLink()
{
  return m_pfaLink[1];
}

inline VOID QEFacet::SetNextLink(QEFacet *pfa)
{
  m_pfaLink[1] = pfa;
}

inline CSLIDESurface *QEFacet::GetSurface()
{
  return m_pCSLIDESurface;
}

inline VOID QEFacet::SetSurface(CSLIDESurface *pCSLIDESurface)
{
  m_pCSLIDESurface = pCSLIDESurface;
}

//////////////////////////////////////////////////////////////////////
// QEFacetIter Class
//

class QEFacetIter
{
//////////////////////////////////////////////////
// Initialization
//
public:
  QEFacetIter();
  ~QEFacetIter();
  VOID Init(QEFacet *pQEFacet);
  VOID Uninit();

  BOOL Initialized();

  BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:
  QEVertex *PeekFirstVertex();
  QEVertex *PeekNextVertex();
  QEVertex *PeekNextWrapVertex();
  QEVertex *PeekLastVertex();
  QEVertex *PeekPrevVertex();
  QEVertex *PeekPrevWrapVertex();

  QEEdge *PeekFirstEdge();
  QEEdge *PeekNextEdge();
  QEEdge *PeekNextWrapEdge();
  QEEdge *PeekLastEdge();
  QEEdge *PeekPrevEdge();
  QEEdge *PeekPrevWrapEdge();

  QEWedge *PeekFirstWedge();
  QEWedge *PeekNextWedge();
  QEWedge *PeekNextWrapWedge();
  QEWedge *PeekLastWedge();
  QEWedge *PeekPrevWedge();
  QEWedge *PeekPrevWrapWedge();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  QEFacet *m_pQEFacet;

  QEEdge *m_peHead;
  QEEdge *m_peCurr;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif // __QEFACET_H__
