#include "qemesh.h"

//////////////////////////////////////////////////
// QEMesh Subdivision Geometric based

QEMesh *QEMesh::SubdivideGeometricTesselate()
{
  return SubdivideGeometric(stTesselate);
}

QEMesh *QEMesh::SubdivideGeometricBiarc()
{
  return SubdivideGeometric(stSphere);
}

///////////////////////////////////////////////////
// QEMesh::SubdivideGeometric                    //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0  => v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
QEMesh *QEMesh::SubdivideGeometric(SubdivisionType eSubdivisionType)
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew, *pfaCenter;
  QEEdge *pe, *peNew;
  Point pt;
  Vector v;
  FLOAT fSharpness;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      // Create the Vertex Vertex
      pt = CalculateVertexSubdivisionGeometric(eSubdivisionType, pvx, v);
      pvxNew = pme->CreateVertex(pt);
      ASSERT( pvxNew != NULL );
      pvxNew->SetNormal(v);
      pvxNew->SetSharpVertexType(pvx->GetSharpVertexType());
      pvx->SetForwardVertex(pvxNew);

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  ASSERT( pe->GetForwardEdge() == NULL );

	  ASSERT( pe->GetKnotInterval() == pe->GetEdge()->GetKnotInterval() );
	  fSharpness = pe->GetKnotInterval();
	  if ( fSharpness != INFINITY )
	    {
	      fSharpness = MAX(0.0, fSharpness - 1.0);
	    }

	  peNew = pme->CreateEdge();
	  peNew->SetFacet(GetNullFacet());
	  peNew->GetEdge()->SetFacet(GetNullFacet());
	  peNew->SetVertex(pvx->GetForwardVertex());
	  peNew->SetKnotInterval(fSharpness);
	  peNew->GetEdge()->SetKnotInterval(fSharpness);
	  pe->SetForwardEdge(peNew);

	  if ( pe->GetEdgePrev(1)->GetForwardEdge() != NULL )
	    {
	      Splice(1, pe->GetEdgePrev(1)->GetForwardEdge(), pe->GetForwardEdge());
	    }

	  if ( pe->GetEdge()->GetForwardEdge() == NULL )
	    {
	      pt = CalculateEdgeSubdivisionGeometric(eSubdivisionType, pe, v);
	      pt[3] = 1.0;
	      pvxNew = pme->CreateVertex(pt);
	      ASSERT( pvxNew != NULL );
	      pvxNew->SetNormal(v);
	      if ( pe->GetFacet() == GetNullFacet() ||
		   pe->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  pvxNew->SetSharpVertexType(svtBoundaryCrease);
		}
	      else if ( pe->GetKnotInterval() == INFINITY )
		{
		  pvxNew->SetSharpVertexType(svtRegularCrease);
		}
	      else
		{
		  pvxNew->SetSharpVertexType(svtSmooth);
		}
	      pe->GetForwardEdge()->GetEdge()->SetVertex(pvxNew);
	    }
	  else
	    {
	      peNew->GetEdge()->SetVertex(pe->GetEdge()->GetForwardEdge()->GetEdge()->GetVertex());	      
	      Splice(1, peNew->GetEdge(), pe->GetEdge()->GetForwardEdge()->GetEdge());
	    }
	}
    }

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      // Create the facet-facet
      pfaCenter = pme->CreateFacet();

      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( pe->GetForwardEdge() != NULL );

	  // Cut off the corner
	  pfaNew = pme->CreateFacet();

	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(0.0);
	  peNew->GetEdge()->SetKnotInterval(0.0);
	  peNew->SetFacet(pfaNew);
	  peNew->GetEdge()->SetFacet(pfaCenter);

	  pe->GetForwardEdge()->SetFacet(pfaNew);
	  pe->GetForwardEdge()->GetEdgePrev(0)->SetFacet(pfaNew);

	  peNew->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetForwardEdge()->GetEdgePrev(0)->GetVertex());

	  Splice(1, peNew, pe->GetForwardEdge()->GetEdge()->GetEdgePrev(1));
	  Splice(1, peNew->GetEdge(), pe->GetForwardEdge()->GetEdgePrev(0));
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      ASSERT( pe->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge()->GetFacet() != pme->GetNullFacet() );

      pe->GetForwardEdge()->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge());

      pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge()->GetEdgeNext(0));
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge()->GetEdgeNext(0));
    }

  pme->Preprocess();
  //pme->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();

  return pme;
}

Point QEMesh::CalculateVertexSubdivisionGeometric(SubdivisionType eSubdivisionType, QEVertex *pvx, Vector &vNormal)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateVertexSubdivisionGeometricTesselate(pvx, vNormal);
      break;
    case stSphere:
      pt = CalculateVertexSubdivisionGeometricBiarc(pvx, vNormal);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionGeometric(SubdivisionType eSubdivisionType, QEEdge *pe, Vector &vNormal)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateEdgeSubdivisionGeometricTesselate(pe, vNormal);
      break;
    case stSphere:
      pt = CalculateEdgeSubdivisionGeometricBiarc(pe, vNormal);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

//////////////////////////////////////////////////////////////////////
// Tesselate

Point QEMesh::CalculateVertexSubdivisionGeometricTesselate(QEVertex *pvx, Vector &vNormal)
{
  Point pt;

  pt = pvx->GetPoint();
  vNormal = pvx->GetNormal();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionGeometricTesselate(QEEdge *pe, Vector &vNormal)
{
  QEEdge *ppe[2];
  Point pt;
  Point ppt[2];
  Vector pvN[2];
  UINT_32 i;

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  for ( i = 0; i < 2; i++ )
    {
      ppt[i] = ppe[i]->GetVertex()->GetPoint();
      pvN[i] = ppe[i]->GetVertex()->GetNormal();
    }

  vNormal = pvN[0] + pvN[1];
  vNormal.Normalize();

  pt = ( ppt[0] + ppt[1] ) / 2.0;
  pt[3] = 1.0;

  return pt;
}

//////////////////////////////////////////////////////////////////////
// Biarc interpolation

Point QEMesh::CalculateVertexSubdivisionGeometricBiarc(QEVertex *pvx, Vector &vNormal)
{
  Point pt;

  pt = pvx->GetPoint();
  vNormal = pvx->GetNormal();

  return pt;
}

//#define DEBUG_BIARCS

Point QEMesh::CalculateEdgeSubdivisionGeometricBiarc(QEEdge *pe, Vector &vNormal)
{
  QEEdge *ppe[2], *peCurr;
  Point pt;
  Point ppt[2];
  Vector pvT[2];
  Vector pvNSurf[2], pvNProj[2];
  Vector v, vE, vX, vY, vZ;
  FLOAT fMag;
  UINT_32 i, j, uiN;

#ifdef DEBUG_BIARCS
  fprintf(stderr, "START: CalculateEdgeSubdivisionGeometricBiarc\n");
#endif // DEBUG_BIARCS

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  for ( i = 0; i < 2; i++ )
    {
      ppt[i] = ppe[i]->GetVertex()->GetPoint();
      pvNSurf[i] = ppe[i]->GetVertex()->GetNormal();
      pvNSurf[i][3] = 0.0;

      // Compute tangent vector
      peCurr = ppe[i];
      uiN = 0;
      do
	{
	  uiN++;
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != ppe[i] );
      
      peCurr = ppe[i];
      pvT[i].Null();
      j = 0;
      do
	{
	  pvT[i] += cos(2.0*PI*j/uiN)*(peCurr->GetEdge()->GetVertex()->GetPoint() - ppt[i]);
	  j++;
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != ppe[i] );

      pvT[i].Normalize();
    }

#ifdef DEBUG_BIARCS
  fprintf(stderr, "  V0 = ");
  ppt[0].Display();
  fprintf(stderr, "  V1 = ");
  ppt[1].Display();
  fprintf(stderr, "  NSurf0 = ");
  pvNSurf[0].Display();
  fprintf(stderr, "  NSurf1 = ");
  pvNSurf[1].Display();
#endif // DEBUG_BIARCS

  vNormal = pvNSurf[0] + pvNSurf[1];
  vNormal.Normalize();

  pt = ( ppt[0] + ppt[1] ) / 2.0;
  pt[3] = 1.0;

  vE = ppt[1] - ppt[0];
  fMag = vE.Magnitude();
  vX = vE;
  if ( !vX.Normalize() )
    {
      return pt;
    }

  // Normalize the surface normals and 
  // project them into the plane perpendicular to the edge
  for ( i = 0; i < 2; i++ )
    {
      if ( !pvNSurf[i].Normalize() )
	{
	  return pt;
	}
      pvNProj[i] = pvNSurf[i] - ((vX.Dot(pvNSurf[i]))*vX);
      pvNProj[i].Normalize();
    }

  // Average the normals to get a vY vector
  vY = pvNProj[0] + pvNProj[1];
  if ( !vY.Normalize() )
    {
      return pt;
    }

  vZ = vX.Cross(vY);
  if ( !vZ.Normalize() )
    {
      return pt;
    }

#ifdef DEBUG_BIARCS
  fprintf(stderr, "\n");
  fprintf(stderr, "  X = ");
  vX.Display();
  fprintf(stderr, "  Y = ");
  vY.Display();
  fprintf(stderr, "  Z = ");
  vZ.Display();
#endif // DEBUG_BIARCS

  // Project the surface normals into the plane perpendicular to vZ
  for ( i = 0; i < 2; i++ )
    {
      pvNProj[i] = pvNSurf[i] - ((vZ.Dot(pvNSurf[i]))*vZ);
      if ( !pvNProj[i].Normalize() )
	{
	  return pt;
	}
    }

#ifdef DEBUG_BIARCS
  fprintf(stderr, "\n");
  fprintf(stderr, "  NProj0 = ");
  pvNProj[0].Display();
  fprintf(stderr, "  NProj1 = ");
  pvNProj[1].Display();
#endif // DEBUG_BIARCS

  //////////////////////////////////////////////////
  // Compute the Biarc midpoint

  FLOAT fSin, fCos;
  FLOAT fA, fDisc, pfH[4];
  Vector pvN[2], pvV[2], ppvC[4][2], ppvChord[4][2], pvM[4], ppvNM[4][2];
  FLOAT ppfN[2][2];
  FLOAT ppfR[4][2];
  FLOAT fRMin, fRMax, fDen, fHMax;
  FLOAT fSign;
  UINT_32 uiSolutions, uiMax;

  pvV[0].Set(-1.0, 0.0, 0.0);
  pvV[1].Set(1.0, 0.0, 0.0);

  pvN[0].Set(pvNProj[0].Dot(vX), pvNProj[0].Dot(vY), 0.0);
  pvN[1].Set(pvNProj[1].Dot(vX), pvNProj[1].Dot(vY), 0.0);

  fCos = MIN(MAX(pvN[0].Dot(pvN[1]),-1.0), 1.0);
  fSin = MIN(MAX(pvN[0](0)*pvN[1](1) - pvN[0](1)*pvN[1](0),-1.0), 1.0);
  if ( fSin == 0.0 )
    {
      return pt;
    }

  uiSolutions = 0;

  fA = 1.0 + fCos;
  fDisc = 2.0*fA;
  if ( fDisc >= 0.0 )
    {
      pfH[uiSolutions] = (fA - sqrt(fDisc)) / fSin;
      uiSolutions++;
      pfH[uiSolutions] = (fA + sqrt(fDisc)) / fSin;
      uiSolutions++;
    }

  fA = 1.0 + fCos;
  fDisc = 2.0*fA;
  if ( fDisc >= 0.0 )
    {
      pfH[uiSolutions] = (fA - sqrt(fDisc)) / fSin;
      uiSolutions++;
      pfH[uiSolutions] = (fA + sqrt(fDisc)) / fSin;
      uiSolutions++;
    }

  if ( uiSolutions == 0 )
    {
      return pt;
    }

#ifdef DEBUG_BIARCS
  fprintf(stderr, "\n");
  fprintf(stderr, "  V0 = ");
  pvV[0].Display();
  fprintf(stderr, "  V1 = ");
  pvV[1].Display();
  fprintf(stderr, "  N0 = ");
  pvN[0].Display();
  fprintf(stderr, "  N1 = ");
  pvN[1].Display();
#endif // DEBUG_BIARCS

  for ( i = 0; i < uiSolutions; i++ )
    {
      pvM[i].Set(0.0, pfH[i], 0.0);
      for ( j = 0; j < 2; j++ )
	{
	  ppvC[i][j].Set((pvN[j](0)*((pfH[i]*pfH[i] - 1.0)/2.0) + pvN[j](1)*pvV[j](0)*pfH[i]),
			 (pvN[j](1)*((pfH[i]*pfH[i] + 1.0)/2.0)), 
			 0.0);
	  ppvChord[i][j] = pvM[i] - pvV[j];
	  fDen = ppvChord[i][j].Dot(pvN[j]);
	  if ( fDen == 0.0 )
	    {	      
	      ppfR[i][j] = INFINITY;
	    }
	  else
	    {
	      ppfR[i][j] = (1.0 + pfH[i]*pfH[i])/fDen;
	      ppvC[i][j] /= fDen;
	    }
	  ppfR[i][j] = fabs(ppfR[i][j]);
	  ppvNM[i][j] = pvM[i] - ppvC[i][j];
	  ppvNM[i][j].Normalize();
	}

#ifdef DEBUG_BIARCS
      fprintf(stderr, "  M[%ld] = ", i);
      pvM[i].Display();
      fprintf(stderr, "  R[%ld][%ld] = %f\n", i, 0, ppfR[i][0]);
      fprintf(stderr, "  R[%ld][%ld] = %f\n", i, 1, ppfR[i][1]);
      fprintf(stderr, "  C[%ld][%ld] = ", i, 0);
      ppvC[i][0].Display();
      fprintf(stderr, "  C[%ld][%ld] = ", i, 1);
      ppvC[i][1].Display();
      fprintf(stderr, "  NM[%ld][%ld] = ", i, 0);
      ppvNM[i][0].Display();
      fprintf(stderr, "  NM[%ld][%ld] = ", i, 1);
      ppvNM[i][1].Display();
#endif // DEBUG_BIARCS
    }

  UINT_32 puiSolution[4], uiPossibleSolutions;
  UINT_32 puiConcaveUp[2];
  Vector vNMax;
  FLOAT fDot, fH;

  const FLOAT cfNormalEpsilon = 1e-6;
  if ( (fabs(pvN[0](0)) <= cfNormalEpsilon) && 
       (fabs(pvN[1](0)) <= cfNormalEpsilon) )
    {
      return pt;
    }
  else if ( fabs(pvN[0](0)) <= cfNormalEpsilon )
    {
      puiConcaveUp[1] = (pvN[1](0) < 0.0);
      puiConcaveUp[0] = !puiConcaveUp[1];
    }
  else if ( fabs(pvN[1](0)) <= cfNormalEpsilon )
    {
      puiConcaveUp[0] = (pvN[0](0) > 0.0);
      puiConcaveUp[1] = !puiConcaveUp[0];
    }
  else
    {
      puiConcaveUp[0] = (pvN[0](0) > 0.0);
      puiConcaveUp[1] = (pvN[1](0) < 0.0);
    }

  uiPossibleSolutions = 0;  
  if ( puiConcaveUp[0] == puiConcaveUp[1] )
    {
      if ( puiConcaveUp[0] )
	{
	  for ( i = 0; i < uiSolutions; i++ )
	    {
	      fDot = ppvNM[i][0].Dot(ppvNM[i][1]);
	      if ( fDot > 0.0 && pvM[i](1) <= 0.0 )
		{
		  if ( uiPossibleSolutions == 0 )
		    {
		      fHMax = pvM[i](1);
		      uiMax = i;
		      vNMax = -ppvNM[i][0];
		    }
		  else if ( puiConcaveUp[0] )
		    {
		      if ( fHMax < pvM[i](1) )
			{
			  fHMax = pvM[i](1);
			  uiMax = i;
			  vNMax = -ppvNM[i][0];
			}
		    }
		  puiSolution[uiPossibleSolutions] = i;
		  uiPossibleSolutions++;
		}
	    }
	}
      else
	{
	  for ( i = 0; i < uiSolutions; i++ )
	    {
	      fDot = ppvNM[i][0].Dot(ppvNM[i][1]);
	      if ( fDot > 0.0 && pvM[i](1) >= 0.0 )
		{
		  if ( uiPossibleSolutions == 0 )
		    {
		      fHMax = pvM[i](1);
		      uiMax = i;
		      vNMax = ppvNM[i][0];
		    }
		  else
		    {
		      if ( fHMax > pvM[i](1) )
			{
			  fHMax = pvM[i](1);
			  uiMax = i;
			  vNMax = ppvNM[i][0];
			}
		    }
		  puiSolution[uiPossibleSolutions] = i;
		  uiPossibleSolutions++;
		}
	    }
	}
    }
  else
    {
      for ( i = 0; i < uiSolutions; i++ )
	{
	  fH = MIN(fabs((ppvC[i][0](1)+ppfR[i][0])-(ppvC[i][1](1)-ppfR[i][1])),
		   fabs((ppvC[i][0](1)-ppfR[i][0])-(ppvC[i][1](1)+ppfR[i][1])));

	  if ( uiPossibleSolutions == 0 )
	    {
	      fHMax = fH;
	      uiMax = i;
	      vNMax = ppvNM[i][0];
	    }
	  else
	    {
	      if ( fHMax > fH )
		{
		  fHMax = fH;
		  uiMax = i;
		  vNMax = ppvNM[i][0];
		}
	    }
	  puiSolution[uiPossibleSolutions] = i;
	  uiPossibleSolutions++;
	}

      if ( vNMax(1) < 0.0 )
	{
	  vNMax = -vNMax;
	}
    }

  if ( uiPossibleSolutions == 0 )
    {
      return pt;
    }

#ifdef DEBUG_BIARCS
  fprintf(stderr, "\n  Chose %ld\n", uiMax);
#endif // DEBUG_BIARCS
  
  //vNormal = ( (pvN[0][0]*(pfH[uiMax]*pfH[uiMax] - 1.0) - 2.0*pvN[0][1]*pfH[uiMax])*vX +
  //	      (-pvN[0][1]*(pfH[uiMax]*pfH[uiMax] - 1.0) - 2.0*pvN[0][0]*pfH[uiMax])*vY );
  vNormal = (vNMax(0)*vX) + (vNMax(1)*vY);
  vNormal.Normalize();

  pt += ((pfH[uiMax]*fMag/2.0)*vY);
  pt[3] = 1.0;

#ifdef DEBUG_BIARCS
  fprintf(stderr, "  E = ");
  pt.Display();
  fprintf(stderr, "  N = ");
  vNormal.Display();
#endif // DEBUG_BIARCS

#ifdef DEBUG_BIARCS
  fprintf(stderr, "END: CalculateEdgeSubdivisionGeometricBiarc\n");
#endif // DEBUG_BIARCS

  return pt;
}






