#include "qemesh.h"

QEMesh *QEMesh::SubdividePentagon()
{
  return SubdividePentagon(stCatmullClark);
}

QEMesh *QEMesh::SubdividePentagonTesselate()
{
  return SubdividePentagon(stTesselate);
}

///////////////////////////////////////////////////
// QEMesh::SubdividePentagon                     //
//                                               //
//       Old Quad          Subdivided Facets     //
//                                               //
//  v3--------------v2     v6------v5------v4    //
//   |               |      |       |       |    //
//   |               |      |      v10      |    //
//   |               |      |     /   \     |    //
//   |               |  => v7--v11     v9--v3    //
//   |               |      |     \   /     |    //
//   |               |      |      v8       |    //
//   |               |      |       |       |    //
//  v0--------------v1     v0------v1------v2    //
//                                               //
QEMesh *QEMesh::SubdividePentagon(SubdivisionType eSubdivisionType)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaNew, *pfaCenter;
  QEEdge *pe, *peNew, *peCurr, *ppe[2];
  QEVertex *pvx, *pvxNew;
  FLOAT fSharpness;
  UINT_32 i;

  Point pt;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      //////////////////////////////////////////////////
      // Construct the border edges
      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( ( pe->GetEdge()->GetForwardEdge() == NULL &&
		    pe->GetForwardEdge() == NULL ) ||
		  ( pe->GetEdge()->GetForwardEdge() != NULL &&
		    pe->GetForwardEdge() != NULL ) );

	  if ( pe->GetForwardEdge() == NULL )
	    {
	      //////////////////////////////////////////////////
	      // Split the edge

	      // Create the midpoint
	      pt = CalculateEdgeSubdivisionPentagon(eSubdivisionType, pe);
	      pvxNew = pme->CreateVertex(pt);
	      ASSERT( pvxNew != NULL );

	      // Create the two subdivided edges
	      peNew = pme->CreateEdge();
	      ASSERT( peNew != NULL );
	      pe->SetForwardEdge(peNew);

	      peCurr = pme->CreateEdge();
	      ASSERT( peCurr != NULL );
	      pe->GetEdge()->SetForwardEdge(peCurr);

	      fSharpness = pe->GetKnotInterval();
	      if ( fSharpness != INFINITY )
		{
		  fSharpness--;
		}
	      if ( fSharpness < 0.0 )
		{
		  fSharpness = 0.0;
		}
	      peNew->SetKnotInterval(fSharpness);
	      peNew->GetEdge()->SetKnotInterval(fSharpness);
	      peCurr->SetKnotInterval(fSharpness);
	      peCurr->GetEdge()->SetKnotInterval(fSharpness);

	      peNew->GetEdge()->SetVertex(pvxNew);
	      peCurr->GetEdge()->SetVertex(pvxNew);

	      Splice(1, peNew->GetEdge(), peCurr->GetEdge());

	      //////////////////////////////////////////////////
	      // Deal with the end point Vertices
	      ppe[0] = pe;
	      ppe[1] = pe->GetEdge();
	      for ( i = 0; i < 2; i++ )
		{
		  pvx = ppe[i]->GetVertex();
		  if ( pvx->GetForwardVertex() == NULL )
		    {
		      pt = CalculateVertexSubdivisionPentagon(eSubdivisionType, pvx);
		      pvxNew = pme->CreateVertex(pt);
		      ASSERT( pvxNew != NULL );
		      pvx->SetForwardVertex(pvxNew);

		      ppe[i]->GetForwardEdge()->SetVertex(pvxNew);
		    }
		  else
		    {
		      for ( peCurr = ppe[i]->GetEdgePrev(1);
			    peCurr != ppe[i];
			    peCurr = peCurr->GetEdgePrev(1) )
			{
			  if ( peCurr->GetForwardEdge() != NULL )
			    {
			      ppe[i]->GetForwardEdge()->SetVertex(peCurr->GetForwardEdge()->GetVertex());
			      Splice(1, ppe[i]->GetForwardEdge(), peCurr->GetForwardEdge());
			      break;
			    }
			}
		      ASSERT( peCurr != ppe[i] );
		    }
		}	      
	    }
	}

      //////////////////////////////////////////////////
      // Create edge perpendiculars
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( pe->GetEdge()->GetForwardEdge() != NULL &&
		  pe->GetForwardEdge() != NULL );

	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(0.0);
	  peNew->GetEdge()->SetKnotInterval(0.0);

	  // Create the facet-edge point
	  pt = CalculateFacetSubdivisionPentagon(eSubdivisionType, pe);
	  pvxNew = pme->CreateVertex(pt);
	  ASSERT( pvxNew != NULL );

	  peNew->GetEdge()->SetVertex(pvxNew);
	  peNew->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());

	  Splice(1, peNew, pe->GetForwardEdge()->GetEdge()->GetEdgePrev(1));
	}

      //////////////////////////////////////////////////
      // Seal off and create the facets
      pfaCenter = pme->CreateFacet();
      pfa->SetForwardFacet(pfaCenter);

      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( pe->GetEdge()->GetForwardEdge() != NULL &&
		  pe->GetForwardEdge() != NULL );

	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(0.0);
	  peNew->GetEdge()->SetKnotInterval(0.0);

	  // Create the facet
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );

	  peNew->GetEdge()->SetFacet(pfaCenter);
	  peNew->SetFacet(pfaNew);
	  pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pfaNew);
	  pe->GetForwardEdge()->SetFacet(pfaNew);
	  pe->GetForwardEdge()->GetEdgePrev(0)->SetFacet(pfaNew);
	  pe->GetForwardEdge()->GetEdgePrev(0)->GetEdgePrev(0)->SetFacet(pfaNew);

	  peNew->SetVertex(pe->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetForwardEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex());

	  Splice(1, peNew, pe->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetEdgePrev(1));
	  Splice(1, peNew->GetEdge(), pe->GetForwardEdge()->GetEdgePrev(0)->GetEdgePrev(0));
	}
    }

  pme->Preprocess();

  return pme;
}

Point QEMesh::CalculateFacetSubdivisionPentagon(SubdivisionType eSubdivisionType, QEEdge *pe)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateFacetSubdivisionPentagonTesselate(pe);
      break;
    case stCatmullClark:
      pt = CalculateFacetSubdivisionPentagon(pe);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionPentagon(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateVertexSubdivisionTesselate(pvx);
      break;
    case stCatmullClark:
      pt = CalculateVertexSubdivisionPentagon(pvx);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionPentagon(SubdivisionType eSubdivisionType, QEEdge *pe)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateEdgeSubdivisionTesselate(pe);
      break;
    case stCatmullClark:
      pt = CalculateEdgeSubdivisionPentagon(pe);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionPentagonTesselate(QEEdge *pe)
{
  FLOAT fR, fE;
  Point pt, ptE, ptF;

  ptF = CalculateFacetSubdivisionTesselate(pe->GetFacet());
  ptE = CalculateEdgeSubdivisionTesselate(pe);

  //  fE = 1.0;
  //  fR = fE / (2.0 * sin(PI/5.0));

  fE = 1.0;
  fR = 1.25;

  pt = ( fE * ptE + fR * ptF ) / ( fE + fR );
  pt[3] = 1.0;
 
  return pt;
}

Point QEMesh::CalculateFacetSubdivisionPentagon(QEEdge *pe)
{
  Point pt;

  pt = CalculateFacetSubdivisionPentagonTesselate(pe);
  Vector v;
  v = pt;
  v[3] = 0.0;
  v.Normalize();
  pt = v;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionPentagon(QEVertex *pvx)
{
  Point pt;

  pt = CalculateVertexSubdivisionTesselate(pvx);
  Vector v;
  v = pt;
  v[3] = 0.0;
  v.Normalize();
  pt = v;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionPentagon(QEEdge *pe)
{
  Point pt;

  pt = CalculateEdgeSubdivisionTesselate(pe);
  Vector v;
  v = pt;
  v[3] = 0.0;
  v.Normalize();
  pt = v;
  pt[3] = 1.0;

  return pt;
}




