#include "qemesh.h"
#include "apm3d.h"

//////////////////////////////////////////////////////////////////////
// Sqrt(3) Subdivision

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsSqrt3()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;
  Point pt;
  Vector vNormal;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexLimitSubdivisionSqrt3(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = CalculateVertexNormalSubdivisionSqrt3(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
    }
}

QEMesh *QEMesh::SubdivideSqrt3()
{
  return SubdivideSqrt3(stLoop);
}

QEMesh *QEMesh::SubdivideSqrt3SphereBarycenter()
{
  return SubdivideSqrt3(stSphere);
}

QEMesh *QEMesh::SubdivideSqrt3SphereCenter()
{
  return SubdivideSqrt3(stButterfly);
}

QEMesh *QEMesh::SubdivideSqrt3SphereRadius()
{
  return SubdivideSqrt3(stZorin);
}

QEMesh *QEMesh::SubdivideSqrt3SphereCapArea()
{
  return SubdivideSqrt3(stDooSabin);
}

QEMesh *QEMesh::SubdivideSqrt3SphereTriangleArea()
{
  return SubdivideSqrt3(stCatmullClark);
}

QEMesh *QEMesh::SubdivideSqrt3SphereTangent()
{
  return SubdivideSqrt3(stCornerRounding);
}

///////////////////////////////////////////////////
// QEMesh::SubdivideSqrt3                        //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1------e1------f0    //
//           |              |\            / |    //
//           |              |  e5      e4   |    //
//           |              |    \    /     |    //
//  e2------vV------e0  => e2      vV      e0    //
//           |              |    /    \     |    //
//           |              |  e6      e7   |    //
//           |              |/            \ |    //
//  f2      e3      f3     f2------e3------f3    //
//                                               //
QEMesh *QEMesh::SubdivideSqrt3(SubdivisionType eSubdivisionType)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvxNew;
  QEFacet *pfaNew;
  QEEdge *pe, *peCurr, *peNew;

  Point pt;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      peNew = pme->CreateEdge();
      ASSERT( peNew != NULL );

      pe->SetForwardEdge(peNew);

      // The facet vertex
      if ( pe->GetFacet()->GetForwardVertex() != NULL )
	{
	  for ( peCurr = pe->GetEdgePrev(0);
		peCurr->GetForwardEdge() == NULL;
		peCurr = peCurr->GetEdgePrev(0) )
	    {
	      ASSERT( peCurr != pe );
	    }

	  Splice(1, peNew, peCurr->GetForwardEdge());

	  peNew->SetVertex(pe->GetFacet()->GetForwardVertex());
	}
      else
	{
	  pt = CalculateFacetSubdivisionSqrt3(eSubdivisionType, pe->GetFacet());
	  pvxNew = pme->CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pe->GetFacet()->SetForwardVertex(pvxNew);

	  peNew->SetVertex(pvxNew);
	}
      
      // The vertex vertex
      if ( pe->GetVertex()->GetForwardVertex() != NULL )
	{
	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr->GetForwardEdge() == NULL;
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      ASSERT( peCurr != pe );
	    }

	  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetEdge());

	  peNew->GetEdge()->SetVertex(pe->GetVertex()->GetForwardVertex());
	}
      else
	{
	  pt = CalculateVertexSubdivisionSqrt3(eSubdivisionType, pe->GetVertex());
	  pvxNew = pme->CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pe->GetVertex()->SetForwardVertex(pvxNew);

	  peNew->GetEdge()->SetVertex(pvxNew);
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() != NULL )
	{
	  peNew = pme->CreateEdge();
	  ASSERT( peNew != NULL );

	  peNew->SetVertex(pe->GetForwardEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetForwardEdge()->GetVertex());

	  Splice(1, peNew, pe->GetForwardEdge());
	  Splice(1, peNew->GetEdge(), pe->GetEdge()->GetForwardEdge());

	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );
	  peCurr = peNew;
	  do
	    {
	      peCurr->SetFacet(pfaNew);
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( peCurr != peNew );

	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );
	  peCurr = peNew->GetEdge();
	  do
	    {
	      peCurr->SetFacet(pfaNew);
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( peCurr != peNew->GetEdge() );

	  pe->SetForwardEdge(NULL);
	  pe->GetEdge()->SetForwardEdge(NULL);
	}
    }
  pme->Preprocess();

  return pme;
}

Point QEMesh::CalculateFacetSubdivisionSqrt3(SubdivisionType eSubdivisionType, QEFacet *pfa)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stLoop:
      pt = CalculateFacetSubdivisionSqrt3(pfa);
      break;
    case stSphere:
      pt = CalculateFacetSubdivisionSqrt3SphereBarycenter(pfa);
      break;
    case stButterfly:
      pt = CalculateFacetSubdivisionSqrt3SphereCenter(pfa);
      break;
    case stZorin:
      pt = CalculateFacetSubdivisionSqrt3SphereRadius(pfa);
      break;
    case stDooSabin:
      pt = CalculateFacetSubdivisionSqrt3SphereCapArea(pfa);
      break;
    case stCatmullClark:
      pt = CalculateFacetSubdivisionSqrt3SphereTriangleArea(pfa);
      break;
    case stCornerRounding:
      pt = CalculateFacetSubdivisionSqrt3SphereTangent(pfa);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stLoop:
      pt = CalculateVertexSubdivisionSqrt3(pvx);
      break;
    case stSphere:
      pt = CalculateVertexSubdivisionSqrt3SphereBarycenter(pvx);
      break;
    case stButterfly:
      pt = CalculateVertexSubdivisionSqrt3SphereCenter(pvx);
      break;
    case stZorin:
      pt = CalculateVertexSubdivisionSqrt3SphereRadius(pvx);
      break;
    case stDooSabin:
      pt = CalculateVertexSubdivisionSqrt3SphereCapArea(pvx);
      break;
    case stCatmullClark:
      pt = CalculateVertexSubdivisionSqrt3SphereTriangleArea(pvx);
      break;
    case stCornerRounding:
      pt = CalculateVertexSubdivisionSqrt3SphereTangent(pvx);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionSqrt3(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN;

  pt.Null();
  iterQEFacet.Init(pfa);
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      pt += pvx->GetPoint();
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  FLOAT fAlpha;
  UINT_32 uiN;

  pt.Null();
  iterQEVertex.Init(pvx);
  for ( uiN = 0, pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	uiN++, pe = iterQEVertex.PeekNextEdge() )
    {
      pt += pe->GetEdge()->GetVertex()->GetPoint();
    }

  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;

  pt = ( ((1.0 - fAlpha)*pvx->GetPoint()) + (fAlpha/(FLOAT)uiN)*pt );
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionSqrt3(QEVertex *pvx)
{
  ASSERT( pvx != NULL );
  
  QEVertexIter iter;
  QEEdge *pe;
  Point pt;
  UINT_32 uiNeighbors;
  FLOAT fAlpha, fBeta;

  pt.Null();
  uiNeighbors = 0;
  iter.Init(pvx);
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      pt += pe->GetEdge()->GetVertex()->GetPoint();
      uiNeighbors++;
    }
  pt /= (FLOAT)uiNeighbors;

  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiNeighbors)) / 9.0;
  fBeta = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiNeighbors)) / 9.0;

  pt[3] = 1.0;

  return pt;  
}

Vector QEMesh::CalculateVertexNormalSubdivisionSqrt3(QEVertex *pvx)
{
  Vector vNormal;

  vNormal.Set(0.0, 0.0, 1.0);

  return vNormal;
}

#define DEBUG_SQRT_3_SPHERE
Point QEMesh::CalculateFacetSubdivisionSqrt3SphereBarycenter(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  Point pt, ptFacetCenter, ptSphereCenter, ptBarycenter, pptTetra[4];
  Vector v, pvTetra[4], vCenter;
  DOUBLE ppdV[4][3];
  DOUBLE pdC[3];
  DOUBLE dDen, dRadius;
  UINT_32 uiN, i, j;
  BOOL b;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "START: CalculateFacetSubdivisionSqrt3SphereBarycenter %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  iterQEFacet.Init(pfa);
  ptFacetCenter.Null();
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ptFacetCenter += pvx->GetPoint();
    }
  ptFacetCenter /= uiN;
  ptFacetCenter[3] = 1.0;

  pt.Null();
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pptTetra[0] = pe->GetVertex()->GetPoint();
      pptTetra[1] = pe->GetEdge()->GetVertex()->GetPoint();

      pptTetra[2].Null();
      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[2] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[2] /= i;
      pptTetra[2][3] = 1.0;

      pptTetra[3].Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[3] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[3] /= i;
      pptTetra[3][3] = 1.0;

      //////////////////////////////////////////////////
      // Calculate the sphere center
      for ( i = 0; i < 4; i++ )
	{
	  for ( j = 0; j < 3; j++ )
	    {
	      ppdV[i][j] = pptTetra[i](j);
	    }
	}

      dDen = APM3D::OrientationValueVVVV(ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					 ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					 ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					 ppdV[2][0], ppdV[2][1], ppdV[2][2]);

      if ( fabs(dDen) <= 1e-6 )
	{
	  ptBarycenter = ptFacetCenter;
	}
      else
	{
	  b = APM3D::InscribedSphereCenterVVVV(&pdC[0], &pdC[1], &pdC[2], &dRadius, 
					       ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					       ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					       ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					       ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );

	  ptSphereCenter.Set(pdC[0], pdC[1], pdC[2]);

#ifdef DEBUG_SQRT_3_SPHERE
	  fprintf(stderr, "  Sphere Center: ");
	  ptSphereCenter.Display(stderr);
	  fprintf(stderr, "    Radius = %f\n", dRadius);
#endif // DEBUG_SQRT_3_SPHERE
	  v = ptFacetCenter - ptSphereCenter;
	  v[3] = 0.0;
	  if ( !v.Normalize() )
	    {
	      v = (pptTetra[1] - pptTetra[0]).Cross(pptTetra[2] - pptTetra[0]);
 	      if ( dDen < 0.0 )
		{
		  v = -v;
		}
	      if ( !v.Normalize() )
		{
		  ptBarycenter = ptFacetCenter;
		}
	      else
		{
		  v *= dRadius;
		  ptBarycenter = ptSphereCenter + v;
		}
	    }
	  else
	    {
	      v *= dRadius;
	      ptBarycenter = ptSphereCenter + v;
	    }
	}

      pt += ptBarycenter;
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "END  : CalculateFacetSubdivisionSqrt3SphereBarycenter %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SphereBarycenter(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionSqrt3SphereCenter(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  Point pt, ptFacetCenter, ptSphereCenter, ptBarycenter, pptTetra[4];
  Point ptCircleCenter;
  DOUBLE dCircleRadius;
  Vector v, pvTetra[4], vCenter;
  DOUBLE ppdV[4][3];
  DOUBLE pdC[3];
  DOUBLE dDen, dRadius;
  UINT_32 uiN, i, j;
  BOOL b;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "START: CalculateFacetSubdivisionSqrt3SphereCenter %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  iterQEFacet.Init(pfa);
  ptFacetCenter.Null();
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ptFacetCenter += pvx->GetPoint();
    }
  ptFacetCenter /= uiN;
  ptFacetCenter[3] = 1.0;

  pt.Null();
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pptTetra[0] = pe->GetVertex()->GetPoint();
      pptTetra[1] = pe->GetEdge()->GetVertex()->GetPoint();

      pptTetra[2].Null();
      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[2] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[2] /= i;
      pptTetra[2][3] = 1.0;

      pptTetra[3].Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[3] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[3] /= i;
      pptTetra[3][3] = 1.0;

      //////////////////////////////////////////////////
      // Calculate the sphere center
      for ( i = 0; i < 4; i++ )
	{
	  for ( j = 0; j < 3; j++ )
	    {
	      ppdV[i][j] = pptTetra[i](j);
	    }
	}

      b = APM3D::InscribedCircleCenterVVV(&pdC[0], &pdC[1], &pdC[2], &dCircleRadius, 
					  ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					  ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					  ppdV[2][0], ppdV[2][1], ppdV[2][2]);
      ASSERT( b );
      ptCircleCenter.Set(pdC[0], pdC[1], pdC[2]);

      dDen = APM3D::OrientationValueVVVV(ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					 ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					 ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					 ppdV[2][0], ppdV[2][1], ppdV[2][2]);

      if ( fabs(dDen) <= 1e-6 )
	{
	  pt += ptFacetCenter;
	}
      else
	{
	  b = APM3D::InscribedSphereCenterVVVV(&pdC[0], &pdC[1], &pdC[2], &dRadius, 
					       ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					       ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					       ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					       ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );

	  ptSphereCenter.Set(pdC[0], pdC[1], pdC[2]);
	  v = ptSphereCenter - ptCircleCenter;
          v.Normalize();

#ifdef DEBUG_SQRT_3_SPHERE
	  fprintf(stderr, "  Sphere Center: ");
	  ptSphereCenter.Display(stderr);
	  fprintf(stderr, "    Radius = %f\n", dRadius);
	  fprintf(stderr, "    Axis: ");
	  v.Display(stderr);
#endif // DEBUG_SQRT_3_SPHERE

	  pt += ptSphereCenter;
	}
    }
  pt /= uiN;
  pt[3] = 1.0;

  ptSphereCenter = pt;
  v = pptTetra[0] - ptSphereCenter;
  dRadius = v.Magnitude();
#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "  Average Sphere Center: ");
  ptSphereCenter.Display(stderr);
  fprintf(stderr, "    Resulting Radius = %f\n", dRadius);
#endif // DEBUG_SQRT_3_SPHERE
  
  v = ptFacetCenter - ptSphereCenter;
  v[3] = 0.0;
  if ( !v.Normalize() )
    {
      v = (pptTetra[1] - pptTetra[0]).Cross(pptTetra[2] - pptTetra[0]);
      if ( dDen < 0.0 )
	{
	  v = -v;
		}
      if ( !v.Normalize() )
	{
	  ptBarycenter = ptFacetCenter;
	}
      else
	{
	  v *= dRadius;
	  ptBarycenter = ptSphereCenter + v;
	}
    }
  else
    {
      v *= dRadius;
	      ptBarycenter = ptSphereCenter + v;
    }

  pt = ptBarycenter;
  pt[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "END  : CalculateFacetSubdivisionSqrt3SphereCenter %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SphereCenter(QEVertex *pvx)
{
  return CalculateVertexSubdivisionSqrt3SphereBarycenter(pvx);
}

Point QEMesh::CalculateFacetSubdivisionSqrt3SphereRadius(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  Point pt, ptFacetCenter, ptSphereCenter, ptBarycenter, pptTetra[4];
  Point ptCircleCenter;
  DOUBLE dCircleRadius;
  Vector v, vAxis, pvTetra[4], vCenter;
  DOUBLE ppdV[4][3];
  DOUBLE pdC[3];
  DOUBLE dDen, dRadius, dSphereRadius, dMag;
  UINT_32 uiN, i, j;
  BOOL b;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "START: CalculateFacetSubdivisionSqrt3SphereRadius %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  iterQEFacet.Init(pfa);
  ptFacetCenter.Null();
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ptFacetCenter += pvx->GetPoint();
    }
  ptFacetCenter /= uiN;
  ptFacetCenter[3] = 1.0;

  dSphereRadius = 0.0;
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pptTetra[0] = pe->GetVertex()->GetPoint();
      pptTetra[1] = pe->GetEdge()->GetVertex()->GetPoint();

      pptTetra[2].Null();
      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[2] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[2] /= i;
      pptTetra[2][3] = 1.0;

      pptTetra[3].Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[3] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[3] /= i;
      pptTetra[3][3] = 1.0;

      //////////////////////////////////////////////////
      // Calculate the sphere center
      for ( i = 0; i < 4; i++ )
	{
	  for ( j = 0; j < 3; j++ )
	    {
	      ppdV[i][j] = pptTetra[i](j);
	    }
	}

      dDen = APM3D::OrientationValueVVVV(ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					 ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					 ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					 ppdV[2][0], ppdV[2][1], ppdV[2][2]);

      if ( fabs(dDen) <= 1e-6 )
	{
	  dSphereRadius += 0.0;
	}
      else
	{
	  b = APM3D::InscribedSphereCenterVVVV(&pdC[0], &pdC[1], &pdC[2], &dRadius, 
					       ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					       ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					       ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					       ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );
	  if ( dDen < 0.0 )
	    {
	      dRadius = -dRadius;
	    }
	  dSphereRadius += 1.0/dRadius;
	}
    }
  if ( dSphereRadius == 0.0 )
    {
      return ptFacetCenter;
    }
  dSphereRadius = uiN/dSphereRadius;
  dRadius = fabs(dSphereRadius);

  b = APM3D::InscribedCircleCenterVVV(&pdC[0], &pdC[1], &pdC[2], &dCircleRadius, 
				      ppdV[0][0], ppdV[0][1], ppdV[0][2], 
				      ppdV[1][0], ppdV[1][1], ppdV[1][2], 
				      ppdV[2][0], ppdV[2][1], ppdV[2][2]);
  ASSERT( b );
  ptCircleCenter.Set(pdC[0], pdC[1], pdC[2]);
  vAxis = (pptTetra[0] - pptTetra[2]).Cross(pptTetra[1] - pptTetra[2]);
  vAxis.Normalize();

  dMag = (dSphereRadius*dSphereRadius) - (dCircleRadius*dCircleRadius);
  ASSERT( dMag > 0.0 );
  dMag = sqrt(dMag);
  if ( dSphereRadius > 0.0 )
    {
      dMag = -dMag;
    }
  ptSphereCenter = ptCircleCenter + (dMag*vAxis);
  ptSphereCenter[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "  Last Circle Center: ");
  ptCircleCenter.Display(stderr);
  fprintf(stderr, "  Axis: ");
  vAxis.Display(stderr);
  fprintf(stderr, "  Magnitude = %f\n", dMag);
  fprintf(stderr, "  Average Sphere Center: ");
  ptSphereCenter.Display(stderr);
  fprintf(stderr, "    Resulting Radius = %f\n", dSphereRadius);
#endif // DEBUG_SQRT_3_SPHERE

  v = ptFacetCenter - ptSphereCenter;
  v[3] = 0.0;
  if ( !v.Normalize() )
    {
      v = (pptTetra[1] - pptTetra[0]).Cross(pptTetra[2] - pptTetra[0]);
      if ( dDen < 0.0 )
	{
	  v = -v;
		}
      if ( !v.Normalize() )
	{
	  ptBarycenter = ptFacetCenter;
	}
      else
	{
	  v *= dRadius;
	  ptBarycenter = ptSphereCenter + v;
	}
    }
  else
    {
      v *= dRadius;
	      ptBarycenter = ptSphereCenter + v;
    }

  pt = ptBarycenter;
  pt[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "END  : CalculateFacetSubdivisionSqrt3SphereRadius %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SphereRadius(QEVertex *pvx)
{
  return CalculateVertexSubdivisionSqrt3SphereBarycenter(pvx);
}

Point QEMesh::CalculateFacetSubdivisionSqrt3SphereCapArea(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  Point pt, ptFacetCenter, ptSphereCenter, ptBarycenter, pptTetra[4];
  Point ptCircleCenter;
  DOUBLE dCircleRadius;
  Vector v, vAxis, pvTetra[4], vCenter;
  DOUBLE ppdV[4][3];
  DOUBLE pdC[3];
  DOUBLE dDen, dRadius, dSphereRadius, dMag;
  DOUBLE dDisc, dArea, dAvgArea;
  UINT_32 uiN, i, j;
  BOOL b;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "START: CalculateFacetSubdivisionSqrt3SphereCapArea %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  iterQEFacet.Init(pfa);
  ptFacetCenter.Null();
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ptFacetCenter += pvx->GetPoint();
    }
  ptFacetCenter /= uiN;
  ptFacetCenter[3] = 1.0;

  dAvgArea = 0.0;
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pptTetra[0] = pe->GetVertex()->GetPoint();
      pptTetra[1] = pe->GetEdge()->GetVertex()->GetPoint();

      pptTetra[2].Null();
      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[2] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[2] /= i;
      pptTetra[2][3] = 1.0;

      pptTetra[3].Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[3] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[3] /= i;
      pptTetra[3][3] = 1.0;

      //////////////////////////////////////////////////
      // Calculate the sphere center
      for ( i = 0; i < 4; i++ )
	{
	  for ( j = 0; j < 3; j++ )
	    {
	      ppdV[i][j] = pptTetra[i](j);
	    }
	}

      dDen = APM3D::OrientationValueVVVV(ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					 ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					 ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					 ppdV[2][0], ppdV[2][1], ppdV[2][2]);

      if ( fabs(dDen) <= 1e-6 )
	{
	  dAvgArea += 0.0;
	}
      else
	{
	  b = APM3D::InscribedCircleCenterVVV(&pdC[0], &pdC[1], &pdC[2], &dCircleRadius, 
					      ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					      ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					      ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );
	  ptCircleCenter.Set(pdC[0], pdC[1], pdC[2]);

	  b = APM3D::InscribedSphereCenterVVVV(&pdC[0], &pdC[1], &pdC[2], &dSphereRadius, 
					       ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					       ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					       ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					       ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );
	  ptSphereCenter.Set(pdC[0], pdC[1], pdC[2]);
	  ASSERT( dSphereRadius > 0.0 );

	  dDisc = (dSphereRadius*dSphereRadius) - (dCircleRadius*dCircleRadius);
	  if ( dDisc < 0.0 )
	    {
	      dDisc = 0.0;
	    }
	  dDisc = sqrt(dDisc);

	  dArea = (dSphereRadius - dDisc) / (2.0*dSphereRadius);
	  if ( dDen < 0.0 )
	    {
	      dArea = -dArea;
	    }
	  dAvgArea += dArea;
	}
    }
  dAvgArea /= uiN;
  if ( dAvgArea == 0.0 )
    {
      return ptFacetCenter;
    }
  dArea = fabs(dAvgArea);
  ASSERT( 0.0 < dArea && dArea < 1.0 );
  
  dDisc = dArea*(1.0 - dArea);
  dRadius = dCircleRadius / (2.0*sqrt(dDisc));
  dSphereRadius = dRadius;
  if ( dAvgArea < 0.0 )
    {
      dSphereRadius = -dSphereRadius;
    }

  vAxis = (pptTetra[0] - pptTetra[2]).Cross(pptTetra[1] - pptTetra[2]);
  vAxis.Normalize();

  dMag = (dSphereRadius*dSphereRadius) - (dCircleRadius*dCircleRadius);
  ASSERT( dMag > 0.0 );
  dMag = sqrt(dMag);
  if ( dSphereRadius > 0.0 )
    {
      dMag = -dMag;
    }
  ptSphereCenter = ptCircleCenter + (dMag*vAxis);
  ptSphereCenter[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "  Avg Area = %f\n", dAvgArea);
  fprintf(stderr, "  Last Circle Center: ");
  ptCircleCenter.Display(stderr);
  fprintf(stderr, "  Axis: ");
  vAxis.Display(stderr);
  fprintf(stderr, "  Magnitude = %f\n", dMag);
  fprintf(stderr, "  Average Sphere Center: ");
  ptSphereCenter.Display(stderr);
  fprintf(stderr, "    Resulting Radius = %f\n", dSphereRadius);
#endif // DEBUG_SQRT_3_SPHERE

  v = ptFacetCenter - ptSphereCenter;
  v[3] = 0.0;
  if ( !v.Normalize() )
    {
      v = (pptTetra[1] - pptTetra[0]).Cross(pptTetra[2] - pptTetra[0]);
      if ( dDen < 0.0 )
	{
	  v = -v;
		}
      if ( !v.Normalize() )
	{
	  ptBarycenter = ptFacetCenter;
	}
      else
	{
	  v *= dRadius;
	  ptBarycenter = ptSphereCenter + v;
	}
    }
  else
    {
      v *= dRadius;
	      ptBarycenter = ptSphereCenter + v;
    }

  pt = ptBarycenter;
  pt[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "END  : CalculateFacetSubdivisionSqrt3SphereCapArea %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SphereCapArea(QEVertex *pvx)
{
  return CalculateVertexSubdivisionSqrt3SphereBarycenter(pvx);
}

Point QEMesh::CalculateFacetSubdivisionSqrt3SphereTriangleArea(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  Point pt, ptFacetCenter, ptSphereCenter, ptBarycenter, pptTetra[4];
  Point ptCircleCenter;
  DOUBLE dCircleRadius;
  Vector v, vAxis, pvTetra[4], vCenter;
  DOUBLE ppdV[4][3];
  DOUBLE pdC[3];
  DOUBLE dDen, dRadius, dSphereRadius, dMag;
  DOUBLE dDisc, dArea, dAvgArea;
  UINT_32 uiN, i, j;
  BOOL b;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "START: CalculateFacetSubdivisionSqrt3SphereTriangleArea %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  iterQEFacet.Init(pfa);
  ptFacetCenter.Null();
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ptFacetCenter += pvx->GetPoint();
    }
  ptFacetCenter /= uiN;
  ptFacetCenter[3] = 1.0;

  dAvgArea = 0.0;
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pptTetra[0] = pe->GetVertex()->GetPoint();
      pptTetra[1] = pe->GetEdge()->GetVertex()->GetPoint();

      pptTetra[2].Null();
      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[2] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[2] /= i;
      pptTetra[2][3] = 1.0;

      pptTetra[3].Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[3] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[3] /= i;
      pptTetra[3][3] = 1.0;

      //////////////////////////////////////////////////
      // Calculate the sphere center
      for ( i = 0; i < 4; i++ )
	{
	  for ( j = 0; j < 3; j++ )
	    {
	      ppdV[i][j] = pptTetra[i](j);
	    }
	}

      dDen = APM3D::OrientationValueVVVV(ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					 ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					 ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					 ppdV[2][0], ppdV[2][1], ppdV[2][2]);

      if ( fabs(dDen) <= 1e-6 )
	{
	  dAvgArea += 0.0;
	}
      else
	{
	  b = APM3D::InscribedSphereCenterVVVV(&pdC[0], &pdC[1], &pdC[2], &dSphereRadius, 
					       ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					       ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					       ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					       ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );
	  ptSphereCenter.Set(pdC[0], pdC[1], pdC[2]);
	  ASSERT( dSphereRadius > 0.0 );
	  if ( dDen < 0.0 )
	    {
	      dSphereRadius = -dSphereRadius;
	    }

	  b = APM3D::SolidAngle(&dArea, 
				ppdV[0][0], ppdV[0][1], ppdV[0][2], 
				ppdV[1][0], ppdV[1][1], ppdV[1][2], 
				ppdV[2][0], ppdV[2][1], ppdV[2][2],
				dSphereRadius);
	  ASSERT( b );

	  dAvgArea += dArea;
	}
    }
  dAvgArea /= uiN;
  if ( dAvgArea == 0.0 )
    {
      return ptFacetCenter;
    }
  dArea = fabs(dAvgArea);
  ASSERT( 0.0 <= dArea && dArea <= 0.5 );
  
  // Compute the average sphere center
  b = APM3D::InscribedSphereCenterVVVSA(&pdC[0], &pdC[1], &pdC[2], &dSphereRadius, 
					ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					ppdV[2][0], ppdV[2][1], ppdV[2][2],
					dAvgArea);
  ASSERT( b );
  ptSphereCenter.Set(pdC[0], pdC[1], pdC[2]);
  ASSERT( dSphereRadius > 0.0 );
  dRadius = dSphereRadius;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "  Avg Area = %f\n", dAvgArea);
  fprintf(stderr, "  Average Sphere Center: ");
  ptSphereCenter.Display(stderr);
  fprintf(stderr, "    Resulting Radius = %f\n", dSphereRadius);
#endif // DEBUG_SQRT_3_SPHERE

  v = ptFacetCenter - ptSphereCenter;
  v[3] = 0.0;
  if ( !v.Normalize() )
    {
      v = (pptTetra[1] - pptTetra[0]).Cross(pptTetra[2] - pptTetra[0]);
      if ( dDen < 0.0 )
	{
	  v = -v;
		}
      if ( !v.Normalize() )
	{
	  ptBarycenter = ptFacetCenter;
	}
      else
	{
	  v *= dRadius;
	  ptBarycenter = ptSphereCenter + v;
	}
    }
  else
    {
      v *= dRadius;
	      ptBarycenter = ptSphereCenter + v;
    }

  pt = ptBarycenter;
  pt[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "  Average Point: ");
  pt.Display(stderr);
  fprintf(stderr, "END  : CalculateFacetSubdivisionSqrt3SphereTriangleArea %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SphereTriangleArea(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionSqrt3SphereTangent(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  QEVertex *pvx;
  Point pt, ptFacetCenter, ptSphereCenter, ptBarycenter, pptTetra[4];
  Point ptCircleCenter;
  DOUBLE dCircleRadius;
  Vector v, vAxis, pvTetra[4], vCenter;
  DOUBLE ppdV[4][3];
  DOUBLE pdC[3];
  DOUBLE dDen, dRadius, dSphereRadius, dMag;
  DOUBLE dDisc, dTheta, dAvgTheta, dSinTheta;
  UINT_32 uiN, i, j;
  BOOL b;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "START: CalculateFacetSubdivisionSqrt3SphereTangent %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  iterQEFacet.Init(pfa);
  ptFacetCenter.Null();
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ptFacetCenter += pvx->GetPoint();
    }
  ptFacetCenter /= uiN;
  ptFacetCenter[3] = 1.0;

  dAvgTheta = 0.0;
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pptTetra[0] = pe->GetVertex()->GetPoint();
      pptTetra[1] = pe->GetEdge()->GetVertex()->GetPoint();

      pptTetra[2].Null();
      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe;
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[2] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[2] /= i;
      pptTetra[2][3] = 1.0;

      pptTetra[3].Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge();
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pptTetra[3] += peCurr->GetVertex()->GetPoint();
	}
      pptTetra[3] /= i;
      pptTetra[3][3] = 1.0;

      //////////////////////////////////////////////////
      // Calculate the sphere center
      for ( i = 0; i < 4; i++ )
	{
	  for ( j = 0; j < 3; j++ )
	    {
	      ppdV[i][j] = pptTetra[i](j);
	    }
	}

      dDen = APM3D::OrientationValueVVVV(ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					 ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					 ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					 ppdV[2][0], ppdV[2][1], ppdV[2][2]);

      if ( fabs(dDen) <= 1e-6 )
	{
	  dAvgTheta += 0.0;
	}
      else
	{
	  b = APM3D::InscribedCircleCenterVVV(&pdC[0], &pdC[1], &pdC[2], &dCircleRadius, 
					      ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					      ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					      ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );
	  ptCircleCenter.Set(pdC[0], pdC[1], pdC[2]);

	  b = APM3D::InscribedSphereCenterVVVV(&pdC[0], &pdC[1], &pdC[2], &dSphereRadius, 
					       ppdV[3][0], ppdV[3][1], ppdV[3][2], 
					       ppdV[0][0], ppdV[0][1], ppdV[0][2], 
					       ppdV[1][0], ppdV[1][1], ppdV[1][2], 
					       ppdV[2][0], ppdV[2][1], ppdV[2][2]);
	  ASSERT( b );
	  ptSphereCenter.Set(pdC[0], pdC[1], pdC[2]);
	  ASSERT( dSphereRadius > 0.0 );

	  dSinTheta = dCircleRadius/dSphereRadius;
	  dSinTheta = MIN(MAX(dSinTheta, -1.0), 1.0);
	  dTheta = asin(dSinTheta);

	  if ( dDen < 0.0 )
	    {
	      dTheta = -dTheta;
	    }
	  dAvgTheta += dTheta;
	}
    }
  dAvgTheta /= uiN;
  if ( dAvgTheta == 0.0 )
    {
      return ptFacetCenter;
    }
  dTheta = fabs(dAvgTheta);
  ASSERT( 0.0 <= dTheta && dTheta <= PI/2.0 );

  vAxis = (pptTetra[0] - pptTetra[2]).Cross(pptTetra[1] - pptTetra[2]);
  vAxis.Normalize();

  dSphereRadius = dCircleRadius/sin(dTheta);
  dRadius = fabs(dSphereRadius);
  dMag = dCircleRadius/tan(dTheta);
  if ( dAvgTheta > 0.0 )
    {
      dMag = -dMag;
    }
  ptSphereCenter = ptCircleCenter + (dMag*vAxis);
  ptSphereCenter[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "  Avg Theta = %f\n", dAvgTheta);
  fprintf(stderr, "  Last Circle Center: ");
  ptCircleCenter.Display(stderr);
  fprintf(stderr, "  Axis: ");
  vAxis.Display(stderr);
  fprintf(stderr, "  Magnitude = %f\n", dMag);
  fprintf(stderr, "  Average Sphere Center: ");
  ptSphereCenter.Display(stderr);
  fprintf(stderr, "    Resulting Radius = %f\n", dSphereRadius);
#endif // DEBUG_SQRT_3_SPHERE

  v = ptFacetCenter - ptSphereCenter;
  v[3] = 0.0;
  if ( !v.Normalize() )
    {
      v = (pptTetra[1] - pptTetra[0]).Cross(pptTetra[2] - pptTetra[0]);
      if ( dDen < 0.0 )
	{
	  v = -v;
		}
      if ( !v.Normalize() )
	{
	  ptBarycenter = ptFacetCenter;
	}
      else
	{
	  v *= dRadius;
	  ptBarycenter = ptSphereCenter + v;
	}
    }
  else
    {
      v *= dRadius;
	      ptBarycenter = ptSphereCenter + v;
    }

  pt = ptBarycenter;
  pt[3] = 1.0;

#ifdef DEBUG_SQRT_3_SPHERE
  fprintf(stderr, "END  : CalculateFacetSubdivisionSqrt3SphereTangent %p\n", pfa);
#endif // DEBUG_SQRT_3_SPHERE

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SphereTangent(QEVertex *pvx)
{
  return CalculateVertexSubdivisionSqrt3SphereBarycenter(pvx);
}

//////////////////////////////////////////////////////////////////////
// Selective

QEMesh *QEMesh::SubdivideSqrt3Selective(QEMesh::ErrorMetricType eErrorMetric,
					FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertex *pvx;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;
  Vector v;

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideSqrt3Selective(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);

      pvx->SetGeneration(0);
      v.Null();
      pvx->SetTexture(v);
    }

  pme->ReSubdivideSqrt3Selective(eErrorMetric, fTolerance, uiFacetsMax);

  pme->Preprocess();

  return pme;
}

VOID QEMesh::ReSubdivideSqrt3Selective(QEMesh::ErrorMetricType /*eErrorMetric*/,
				       FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peStart;
  FLOAT fError;
  //Vector v;
  UINT_32 uiN;

  uiN = 0;
  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      uiN++;
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  pvxCurr = NULL;
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideSqrt3Selective(pvx);

      // Update the current vertex
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvx);
      m_pheapSelective->Update(phe, fError);

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{

	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }

	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  if ( pvxCurr->GetHeapEntry() == NULL )
	    {   
	      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvxCurr);
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }


	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }

	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdgeNext(0)->GetEdge()->GetVertex();
	  if ( pvxCurr->GetHeapEntry() == NULL )
	    {   
	      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvxCurr);
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt3Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }
	}
    }
  
  if ( uiN > 0 )
    {
      Preprocess();
    }
}

///////////////////////////////////////////////////
// QEMesh::SubdivideSqrt3Selective               //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1              f0    //
//           |               \            /      //
//           |                 e1      e0        //
//           |                   \    /          //
//  e2------vV------e0  =>         vV            //
//           |                   /    \          //
//           |                 e2      e3        //
//           |               /            \      //
//  f2      e3      f3     f2              f3    //
//                                               //
//                                               //
VOID QEMesh::SubdivideSqrt3Selective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEVertex *pvxNew;
  QEEdge *pe, *peCurr, *peNext, *peSpoke, *peNew, *peStart;
  QEFacet *pfaNew;
  Point ptNew, pt;
  //Point ptCenter;
  Vector v;

  v.Null();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Create facet centers and connect them to the central vertex
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  break;
	}
    }

  if ( pe != NULL )
    {
      peStart = pe;
      do
	{
	  peNext = pe->GetEdgeNext(1);

	  // Create facet centers and connect them to the central vertex
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
	  if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  if ( peCurr->GetVertex()->GetGeneration() > pvx->GetGeneration() )
		    {
		      break;
		    }
		}

	      if ( peCurr == pe )
		{
		  // Case 1
		  v = pvx->GetTexture();
		  pt = CalculateFacetSubdivisionSqrt3Selective(pe->GetFacet());
		  pvxNew = CreateVertex(pt);
		  ASSERT( pvxNew != NULL );
		  pvxNew->SetGeneration(pvx->GetGeneration()+1);
		  pvxNew->SetTexture(v);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );

		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(pvxNew);

		  Splice(1, peSpoke, pe);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );

		  peNew->SetVertex(peSpoke->GetEdgePrev(0)->GetVertex());
		  peNew->GetEdge()->SetVertex(pvxNew);
		  peNew->GetEdge()->SetFacet(pfaNew);
		  peNew->SetFacet(pe->GetFacet());

		  Splice(1, peNew, peSpoke->GetEdgePrev(0));
		  Splice(1, peNew->GetEdge(), peSpoke->GetEdge());

		  peSpoke->SetFacet(pfaNew);
		  peSpoke->GetEdgePrev(0)->SetFacet(pfaNew);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );

		  peNew->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(pvxNew);
		  peNew->SetFacet(pfaNew);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());

		  Splice(1, peNew, pe->GetEdgeNext(0));
		  Splice(1, peNew->GetEdge(), peSpoke->GetEdge());

		  peSpoke->GetEdge()->SetFacet(pfaNew);
		  pe->SetFacet(pfaNew);
		}
	      else if ( peCurr == pe->GetEdgePrev(0) )
		{
		  if ( peCurr == pe->GetEdgeNext(0)->GetEdgeNext(0) )
		    {
		      // Case 2
		    }
		  else
		    {
		      // Case 3
		      pfaNew = CreateFacet();
		      ASSERT( pfaNew != NULL );

		      peNew = CreateEdge();
		      ASSERT( peNew != NULL );

		      peNew->SetVertex(pe->GetEdge()->GetVertex());
		      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		      peNew->SetFacet(pfaNew);
		      peNew->GetEdge()->SetFacet(pe->GetFacet());

		      Splice(1, peNew, pe->GetEdgeNext(0));
		      Splice(1, peNew->GetEdge(), peCurr);

		      peCurr->SetFacet(pfaNew);
		      pe->SetFacet(pfaNew);
		    }
		}
	      else if ( peCurr == pe->GetEdgePrev(0)->GetEdgePrev(0) &&
			peCurr == pe->GetEdgeNext(0)->GetEdgeNext(0) )
		{
		  // Case 4
		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(pe->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(pfaNew);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());
	      
		  Splice(1, peNew, pe);
		  Splice(1, peNew->GetEdge(), peCurr);

		  peCurr->SetFacet(pfaNew);
		  peCurr->GetEdgeNext(0)->SetFacet(pfaNew);
		}
	      else if ( peCurr == pe->GetEdgeNext(0)->GetEdgeNext(0) )
		{
		  // Case 5
		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );
	      
		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(peCurr->GetVertex());
		  peSpoke->GetEdge()->SetFacet(pfaNew);
	      
		  Splice(1, peSpoke, pe);
		  Splice(1, peSpoke->GetEdge(), peCurr);

		  pe->SetFacet(pfaNew);
		  pe->GetEdgeNext(0)->SetFacet(pfaNew);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(peSpoke->GetEdgePrev(0)->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(peCurr->GetFacet());
		  peCurr->SetFacet(peCurr->GetFacet());
		  peNew->GetEdge()->SetFacet(pfaNew);
	      
		  Splice(1, peNew, pe);
		  Splice(1, peNew->GetEdge(), peCurr);

		  peSpoke->SetFacet(pfaNew);
		  peSpoke->GetEdgePrev(0)->SetFacet(pfaNew);

		}
	      else if ( peCurr == pe->GetEdgePrev(0)->GetEdgePrev(0) )
		{
		  // Case 6
		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(pfaNew);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());
	      
		  Splice(1, peNew, pe->GetEdgeNext(0));
		  Splice(1, peNew->GetEdge(), peCurr);

		  pe->SetFacet(pfaNew);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );
	      
		  peSpoke->GetEdge()->SetFacet(pfaNew);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(peCurr->GetVertex());
		  peSpoke->SetFacet(pfaNew);
	      
		  Splice(1, peSpoke, pe);
		  Splice(1, peSpoke->GetEdge(), peCurr);

		  peCurr->SetFacet(pfaNew);
		  peCurr->GetEdgeNext(0)->SetFacet(pfaNew);
		}
	      else
		{
		  // Case 7
		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(pfaNew);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());
	      
		  Splice(1, peNew, pe->GetEdgeNext(0));
		  Splice(1, peNew->GetEdge(), peCurr);

		  pe->SetFacet(pfaNew);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );
	      
		  peSpoke->GetEdge()->SetFacet(pfaNew);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(peCurr->GetVertex());
		  peSpoke->SetFacet(pfaNew);
	      
		  Splice(1, peSpoke, pe);
		  Splice(1, peSpoke->GetEdge(), peCurr);

		  peSpoke->GetEdgePrev(0)->SetFacet(pfaNew);

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->GetEdge()->SetFacet(pfaNew);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );

		  peNew->SetVertex(peSpoke->GetEdgePrev(0)->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(pfaNew);
	      
		  Splice(1, peNew, peSpoke->GetEdgePrev(0));
		  Splice(1, peNew->GetEdge(), peCurr);

		  for ( peNew = peCurr->GetEdgeNext(0);
			peNew != peCurr;
			peNew = peNew->GetEdgeNext(0) )
		    {
		      peNew->SetFacet(pfaNew);
		    }
		  peCurr->SetFacet(pfaNew);
		}
	    }
	  else
	    {
	      ASSERT( peNext->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
	      if ( peNext->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
		{
		  if ( pe->GetEdgeNext(0) != pe->GetEdgePrev(0)->GetEdgePrev(0) )
		    {
		      // case 8
		      peCurr = pe->GetEdgeNext(0);

		      pfaNew = CreateFacet();
		      ASSERT( pfaNew != NULL );

		      peNew = CreateEdge();
		      ASSERT( peNew != NULL );
	      
		      peNew->SetVertex(pe->GetEdgePrev(0)->GetVertex());
		      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		      peNew->SetFacet(pe->GetFacet());
		      peNew->GetEdge()->SetFacet(pfaNew);
	      
		      Splice(1, peNew, pe->GetEdgePrev(0));
		      Splice(1, peNew->GetEdge(), peCurr);

		      pe->SetFacet(pfaNew);
		      pe->GetEdgePrev(0)->SetFacet(pfaNew);		      
		    }
		  else
		    {
		      // case 9
		    }
		}
	      else
		{
		  //case 10
		}
	    }

	  pe = peNext;
	}
      while ( pe != peStart );
    }

  // Rotate the edges
  for ( peStart = pvx->GetEdge();
	peStart->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration();
	peStart = peStart->GetEdgeNext(1) )
    {
    }
  pe = peStart;
  do
    {
      peCurr = pe->GetEdgeNext(1);
      ASSERT( peCurr->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( peCurr->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  DelaunayFlip(peCurr);
	}

      pe = pe->GetEdgeNext(1);
    }
  while ( pe != peStart );

  v = pvx->GetPoint();
  pvx->SetTexture(v);
  ptNew = CalculateVertexSubdivisionSqrt3Selective(pvx);
  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()+1);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideSqrt3Selective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3Selective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  FLOAT fAlpha;
  UINT_32 uiN;

  pt.Null();
  iterQEVertex.Init(pvx);
  for ( uiN = 0, pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	uiN++, pe = iterQEVertex.PeekNextEdge() )
    {
      if ( pe->GetEdgeNext(0)->GetEdgePrev(1)->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  pt += pe->GetEdgeNext(0)->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();
	}
      else
	{
	  pt += pe->GetEdgeNext(0)->GetEdgePrev(1)->GetEdge()->GetVertex()->GetTexture();
	}
    }

  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;

  pt = ( ((1.0 - fAlpha)*pvx->GetPoint()) + (fAlpha/(FLOAT)uiN)*pt );
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionSqrt3Selective(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxStart;
  Point pt;
  UINT_32 uiN;

  pt.Null();
  iterQEFacet.Init(pfa);
  pvxStart = iterQEFacet.PeekFirstVertex();
  for ( uiN = 0, pvx = pvxStart;
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ASSERT( pvxStart->GetGeneration() == pvx->GetGeneration() );
      pt += pvx->GetPoint();
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

QEMesh *QEMesh::SubdivideSqrt3SelectiveFacet(QEMesh::ErrorMetricType eErrorMetric,
					     FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  QEVertex *pvx;
  QEFacet *pfa;
  QEEdge *pe;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError, fAlpha, fBeta;
  //Vector v;
  Point pt;
  UINT_32 uiN;

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      fError = pme->CalculateFacetErrorSubdivideSqrt3SelectiveFacet(eErrorMetric, pfa);
      phe = pme->m_pheapSelective->Insert(fError, pfa);
      pfa->SetHeapEntry(phe);

      pfa->SetGeneration(0);
    }

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetGeneration(0);
    }

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      uiN = 0;
      pt.Null();
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pt += pe->GetEdge()->GetVertex()->GetPoint();
	  uiN++;
	}
      pvx->SetValence(uiN);
      fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;
      fBeta = 3.0*fAlpha/(1.0 + 3.0*fAlpha);
      pt *= fBeta/(FLOAT)uiN;

      pt += (1.0 - fBeta)*pvx->GetPoint();
      pt[3] = 1.0;
      pvx->SetLimitPosition(pt);
    }

  pme->ReSubdivideSqrt3SelectiveFacet(eErrorMetric, fTolerance, uiFacetsMax);

  pme->Preprocess();

  return pme;
}

VOID QEMesh::ReSubdivideSqrt3SelectiveFacet(QEMesh::ErrorMetricType eErrorMetric,
					    FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEVertexIter iterQEVertex;
  QEFacet *pfa, *pfaCurr;
  QEVertex *pvx;
  FLOAT fError;

  for ( pfa = (QEFacet*)(m_pheapSelective->PeekTop(&fError));
	pfa != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pfa = (QEFacet*)(m_pheapSelective->PeekTop(&fError)) )
    {
      pvx = pfa->GetForwardVertex();
      ASSERT( pvx != NULL );

      iterQEVertex.Init(pvx);
      pfa = iterQEVertex.PeekFirstFacet();
      for ( pfaCurr = iterQEVertex.PeekNextFacet();
	    pfaCurr != NULL;
	    pfaCurr = iterQEVertex.PeekNextFacet() )
	{
	  if ( pfaCurr->GetGeneration() < pfa->GetGeneration() )
	    {
	      pfa = pfaCurr;
	    }
	}

      SubdivideSqrt3SelectiveFacet(eErrorMetric, pfa);
    }
}

///////////////////////////////////////////////////
// QEMesh::SubdivideSqrt3SelectiveFacet          //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1------e1------f0    //
//           |              |\            / |    //
//           |              |  e5      e4   |    //
//           |              |    \    /     |    //
//  e2------vV------e0  => e2      vV      e0    //
//           |              |    /    \     |    //
//           |              |  e6      e7   |    //
//           |              |/            \ |    //
//  f2      e3      f3     f2------e3------f3    //
//                                               //
VOID QEMesh::SubdivideSqrt3SelectiveFacet(QEMesh::ErrorMetricType eErrorMetric, 
					  QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvxNew;
  QEEdge *peStart, *pe, *peNext;
  //Point pt;

  if ( !(pfa->GetGeneration() & 0x1) )
    {
      // Balance condition
      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( pfa->GetGeneration() < pe->GetEdge()->GetFacet()->GetGeneration() + 2 &&
		  pfa->GetGeneration() + 2 > pe->GetEdge()->GetFacet()->GetGeneration() );
	  if ( pfa->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() + 1 )
	    {
	      SubdivideSqrt3SelectiveFacet(eErrorMetric, pe->GetEdge()->GetFacet());
	    }
	}

      pvxNew = SubdivideSqrt3SelectiveFacetSplit(eErrorMetric, pfa);

      peStart = pvxNew->GetEdge();
      pe = peStart;
      do
	{
	  peNext = pe->GetEdgeNext(1);
	  if ( pe->GetFacet()->GetGeneration() ==
	       pe->GetFacet()->GetEdge()->GetEdge()->GetFacet()->GetGeneration() )
	    {
	      SubdivideSqrt3SelectiveFacetSwap(eErrorMetric, pe->GetEdgeNext(0));
	    }
	  pe = peNext;
	}
      while ( pe != peStart );
    }
  else
    {
      if ( pfa->GetGeneration() - 2 == pfa->GetEdge()->GetEdge()->GetFacet()->GetGeneration() )
	{
	  SubdivideSqrt3SelectiveFacet(eErrorMetric, pfa->GetEdge()->GetEdge()->GetFacet());
	}
      // Triggers edge swap
      SubdivideSqrt3SelectiveFacet(eErrorMetric, pfa->GetEdge()->GetEdge()->GetFacet());
    }
}

QEVertex *QEMesh::SubdivideSqrt3SelectiveFacetSplit(QEMesh::ErrorMetricType eErrorMetric,
						    QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peStart, *peNew, *peCurr;
  QEVertex *pvxNew;
  QEFacet *pfaNew, *pfaCurr;
  Point pt, ptCurr;
  //Vector v;
  UINT_32 uiN;
  FLOAT fAlpha, fBeta;

  QEVertexIter iterQEVertex;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;

  pt = CalculateFacetSubdivisionSqrt3SelectiveFacet(pfa);
  pvxNew = CreateVertex(pt);
  ASSERT( pvxNew != NULL );
  pvxNew->SetGeneration(pfa->GetGeneration()/2 + 1);

  uiN = 0;
  pt.Null();
  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pt += CalculateVertexSubdivisionSqrt3SelectiveFacetGeneration(pe->GetVertex(), pfa->GetGeneration()/2 + 1);

      ASSERT( pe->GetEdge()->GetFacet()->GetGeneration() == pfa->GetGeneration() ||
	      pe->GetEdge()->GetFacet()->GetGeneration() == pfa->GetGeneration() + 1 );
      if ( pe->GetEdge()->GetFacet()->GetGeneration() == pfa->GetGeneration() )
	{
	  ptCurr = CalculateFacetSubdivisionSqrt3SelectiveFacet(pe->GetEdge()->GetFacet());
	  pt += ptCurr;
	}
      else
	{
	  pt += CalculateVertexSubdivisionSqrt3SelectiveFacetGeneration(pe->GetEdge()->GetEdgePrev(0)->GetVertex(), pfa->GetGeneration()/2 + 1);
	}
      uiN += 2;
    }
  pvxNew->SetValence(uiN);
  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;
  fBeta = 3.0*fAlpha/(1.0 + 3.0*fAlpha);
  pt *= fBeta/(FLOAT)uiN;
  
  pt += (1.0 - fBeta)*pvxNew->GetPoint();
  pt[3] = 1.0;
  pvxNew->SetLimitPosition(pt);
  
  peStart = pfa->GetEdge();
  pe = peStart;

  peNew = CreateEdge();
  ASSERT( peNew != NULL );

  peNew->SetVertex(pvxNew);
  peNew->GetEdge()->SetVertex(pe->GetVertex());
  peNew->SetFacet(pe->GetFacet());

  // Maintain Mate Pointer
  pe->SetFacet(pe->GetFacet());

  Splice(1, pe, peNew->GetEdge());

  for ( pe = pe->GetEdgeNext(0);
	pe != peStart->GetEdgeNext(1);
	pe = pe->GetEdgeNext(0) )
    {
        pfaNew = CreateFacet();
	ASSERT( pfaNew != NULL );
	pfaNew->SetGeneration(pfa->GetGeneration()+1);

        peNew = CreateEdge();
	ASSERT( peNew != NULL );

	peNew->SetVertex(pvxNew);
	peNew->GetEdge()->SetVertex(pe->GetVertex());
	peNew->SetFacet(pfaNew);
	peNew->GetEdge()->SetFacet(pe->GetEdgePrev(0)->GetFacet());
	
	// Maintain Mate Pointer
	pe->SetFacet(pfaNew);
	pe->GetEdgePrev(0)->SetFacet(pe->GetEdgePrev(0)->GetFacet());

	Splice(1, pe->GetEdgePrev(0)->GetEdgePrev(0), peNew);
	Splice(1, pe, peNew->GetEdge());
    }

  peStart->GetEdgeNext(1)->SetFacet(peStart->GetEdgeNext(1)->GetEdgePrev(0)->GetFacet());

  // Maintain Mate Pointer
  peStart->GetEdgeNext(1)->GetEdgePrev(0)->SetFacet(peStart->GetEdgeNext(1)->GetEdgePrev(0)->GetFacet());

  pfa->SetGeneration(pfa->GetGeneration()+1);

  //////////////////////////////////////////////////
  // Update the Heap

  iterQEVertex.Init(pvxNew);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      for ( peCurr = pe->GetEdge();
	    peCurr != pe->GetEdge()->GetEdgePrev(1);
	    peCurr = peCurr->GetEdgeNext(1) )
	{
	  // Insert the new vertices
	  pfaCurr = peCurr->GetFacet();
	  pfaCurr->PreprocessLimit();
	  if ( pfaCurr->GetHeapEntry() == NULL )
	    {   
	      fError = CalculateFacetErrorSubdivideSqrt3SelectiveFacet(eErrorMetric, 
								       pfaCurr);
	      phe = m_pheapSelective->Insert(fError, pfaCurr);
	      pfaCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      phe = pfaCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateFacetErrorSubdivideSqrt3SelectiveFacet(eErrorMetric, 
								       pfaCurr);
	      m_pheapSelective->Update(phe, fError);
	    }
	}
    }

  return pvxNew;
}

VOID QEMesh::SubdivideSqrt3SelectiveFacetSwap(QEMesh::ErrorMetricType eErrorMetric, QEEdge *pe)
{
  QEVertex *ppvx[4];
  QEEdge *peCurr;
  QEFacet *pfaCurr;
  UINT_32 i;

  QEVertexIter iterQEVertex;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;

  peCurr = pe->GetEdgePrev(0);
  pe->GetEdge()->GetVertex()->SetEdge(pe->GetEdge()->GetEdgeNext(1));
  pe->GetEdge()->SetVertex(peCurr->GetVertex());
  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));
  Splice(1, peCurr, pe->GetEdge());

  peCurr = pe->GetEdgePrev(1)->GetEdgeNext(0);
  pe->GetVertex()->SetEdge(pe->GetEdgeNext(1));
  pe->SetVertex(peCurr->GetVertex());
  Splice(1, pe, pe->GetEdgePrev(1));
  Splice(1, peCurr, pe);

  pe->GetFacet()->SetGeneration(pe->GetFacet()->GetGeneration() + 1);
  peCurr = pe;
  do
    {
      peCurr->SetFacet(pe->GetFacet());
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe );

  pe->GetEdge()->GetFacet()->SetGeneration(pe->GetEdge()->GetFacet()->GetGeneration() + 1);
  peCurr = pe->GetEdge();
  do
    {
      peCurr->SetFacet(pe->GetEdge()->GetFacet());
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe->GetEdge() );

  //////////////////////////////////////////////////
  // Update the Heap

  ppvx[0] = pe->GetEdgePrev(0)->GetVertex();
  ppvx[1] = pe->GetVertex();
  ppvx[2] = pe->GetEdge()->GetEdgePrev(0)->GetVertex();
  ppvx[3] = pe->GetEdge()->GetVertex();

  for ( i = 0; i < 4; i++ )
    {
      iterQEVertex.Init(ppvx[i]);
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  // Insert the new vertices
	  pfaCurr = peCurr->GetFacet();
	  if ( pfaCurr->GetHeapEntry() == NULL )
	    {   
	      fError = CalculateFacetErrorSubdivideSqrt3SelectiveFacet(eErrorMetric, 
								       pfaCurr);
	      phe = m_pheapSelective->Insert(fError, pfaCurr);
	      pfaCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      phe = pfaCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateFacetErrorSubdivideSqrt3SelectiveFacet(eErrorMetric, 
								       pfaCurr);
	      m_pheapSelective->Update(phe, fError);
	    }
	}
    }

  pe->GetFacet()->PreprocessLimit();
  pe->GetEdge()->GetFacet()->PreprocessLimit();
}

/*
FLOAT QEMesh::CalculateFacetErrorSubdivideSqrt3SelectiveFacet(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  QEVertex *pvx;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid.Null();
  uiN = 0;
  iterQEFacet.Init(pfa);
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      ptCentroid += pvx->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pvx = iterQEFacet.PeekFirstVertex();
  ASSERT( pvx != NULL );

  ptStart = pvx->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pvx = iterQEFacet.PeekNextVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      ptCurr = pvx->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  fErrorMax = 0.0;
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  v = pe->GetEdge()->GetVertex()->GetPoint();
	  v[3] = 1.0;
	  fError = fabs(vPlane.Dot(v));
	  
	  fErrorMax = MAX(fErrorMax, fError);
	}
    }

  return fErrorMax;
}
*/

FLOAT QEMesh::CalculateFacetErrorSubdivideSqrt3SelectiveFacet(QEMesh::ErrorMetricType eErrorMetric, QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxMax;
  FLOAT fError, fErrorMax;

  fErrorMax = -1.0;
  pvxMax = NULL;
  iterQEFacet.Init(pfa);
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      fError = CalculateVertexErrorSubdivideSqrt3SelectiveFacet(eErrorMetric, 
								pfa->GetGeneration()/2, pvx);
      if ( fError > fErrorMax )
	{
	  fErrorMax = fError;
	  pvxMax = pvx;
	}
    }
  ASSERT( pvxMax != NULL );
  pfa->SetForwardVertex(pvxMax);

  return fErrorMax;
}

FLOAT QEMesh::CalculateVertexErrorSubdivideSqrt3SelectiveFacet(ErrorMetricType eErrorMetric,
							       UINT_32 uiGeneration,
							       QEVertex *pvx)
{
  FLOAT fError;

  fError = 0.0;
  switch ( eErrorMetric )
    {
    case emtBestFitPlane:
      {
	fError = CalculateVertexErrorSubdivideSqrt3SelectiveFacetBestFitPlane(pvx);
      }
      break;
    case emtLimitDistance:
      {
	fError = CalculateVertexErrorSubdivideSqrt3SelectiveFacetLimitDistance(uiGeneration, pvx);
      }
      break;
    default:
      {
	ASSERT( FALSE );
      }
      break;
    }

  return fError;
}

FLOAT QEMesh::CalculateVertexErrorSubdivideSqrt3SelectiveFacetBestFitPlane(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

FLOAT QEMesh::CalculateVertexErrorSubdivideSqrt3SelectiveFacetLimitDistance(UINT_32 uiGeneration, 
									    QEVertex *pvx)
{
  FLOAT fError;
  Vector u, v;

  u = CalculateVertexSubdivisionSqrt3SelectiveFacetGeneration(pvx, uiGeneration);
  v = pvx->GetLimitPosition();
  
  u[3] = 0.0;
  v[3] = 0.0;

  v -= u;
  fError = v.Magnitude();

  return fError;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SelectiveFacet(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt3SelectiveFacetGeneration(QEVertex *pvx, UINT_32 uiGeneration)
{
  Point pt;
  UINT_32 uiN;
  FLOAT fAlpha, fGamma;

  uiGeneration -= pvx->GetGeneration();

  uiN = pvx->GetValence();
  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;
  fGamma = pow(((2.0/3.0) - fAlpha), uiGeneration);

  pt = ((1.0 - fGamma)*pvx->GetLimitPosition()) + (fGamma*pvx->GetPoint());
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionSqrt3SelectiveFacet(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN, uiGeneration;

  uiGeneration = pfa->GetGeneration()/2;

  uiN = 0;
  pt.Null();
  iterQEFacet.Init(pfa);
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      uiN++;
      pt += CalculateVertexSubdivisionSqrt3SelectiveFacetGeneration(pvx, uiGeneration);
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

