#include "qerender.h"
#include "qemesh.h"
#include "gillib.h"
#include "bound.h"

static VOID RenderGeometry(QEFacet *pfa, BOOL bFaceted, BOOL bLimitSurface,
			   BOOL bCalculateNormals, BOOL bSelective);
static VOID RenderNormals(QEFacet *pfa, BOOL bLimitSurface);

static VOID RenderGeometry(QEEdge *pe, QEWedge *pwe, BOOL bFaceted, BOOL bLimitSurface,
			   BOOL bCalculateNormals, BOOL bSelective);
static VOID RenderNormals(QEVertex *pvx, BOOL bLimitSurface,
			  BOOL bCalculateNormals, BOOL bSelective);

//static VOID Render(QEWedge *pwe);

static VOID RenderLineSegment(const Point &pt0, const Point &pt1, BOOL bDrawHeight);
static VOID RenderParabolaAdaptive(const Point &ptV0, 
				   const Point &ptE10, const Point &ptE11, 
				   const Point &pt0, const Point &pt1,
				   UINT_32 uiLevel, UINT_32 uiLevelMax,
				   FLOAT fEpsilon, BOOL bDrawHeight);
static VOID RenderHyperbolaAdaptive(const Point &ptV0, const Point &ptV1, 
				    const Point &pt0, const Point &pt1,
				    UINT_32 uiLevel, UINT_32 uiLevelMax,
				    FLOAT fEpsilon, BOOL bDrawHeight);

//////////////////////////////////////////////////////////////////////
// QEMesh Rendering Implementation
//

const FLOAT cfNormalLength = 0.1;

//#define COUNT_TRIANGLES
#ifdef COUNT_TRIANGLES
static UINT_32 s_uiTrianglesRendered = 0;
#endif // COUNT_TRIANGLES

VOID RenderGeometry(QEMesh *pme, BOOL bFaceted/* = TRUE*/, BOOL bLimitSurface/* = FALSE*/, 
		    BOOL bCalculateNormals/* = FALSE*/, BOOL bSelective/*= FALSE*/)
{
  ASSERT( pme != NULL );

  QEMeshIter iterQEMesh;
  QEFacet *pfa;

  //if ( !bFaceted && bLimitSurface )
  //  {
  //    fprintf(stderr, "START: Mesh\n");
  //  }

#ifdef COUNT_TRIANGLES
  s_uiTrianglesRendered = 0;
#endif // COUNT_TRIANGLES

  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      ::RenderGeometry(pfa, bFaceted, bLimitSurface, bCalculateNormals, bSelective);
    }

#ifdef COUNT_TRIANGLES
  fprintf(stderr, "START: RenderMesh %p\n", pme);
  fprintf(stderr, "  Triangles = %ld\n", s_uiTrianglesRendered);  
  pme->DisplayEuler();
  fprintf(stderr, "END  : RenderMesh %p\n", pme);
#endif // COUNT_TRIANGLES

  //if ( !bFaceted && bLimitSurface )
  //  {
  //    fprintf(stderr, "END  : Mesh\n");
  //  }
}

VOID RenderNormals(QEMesh *pme, BOOL bFaceted/* = TRUE*/, BOOL bLimitSurface/* = FALSE*/, 
		   BOOL bCalculateNormals/* = FALSE*/, BOOL bSelective/*= FALSE*/)
{
  ASSERT( pme != NULL );

  QEMeshIter iterQEMesh;
  QEFacet *pfa;
  QEVertex *pvx;

  iterQEMesh.Init(pme);
  if ( bFaceted )
    {
      for ( pfa = iterQEMesh.PeekFirstFacet();
	    pfa != NULL;
	    pfa = iterQEMesh.PeekNextFacet() )
	{
	  ::RenderNormals(pfa, bLimitSurface);
	}
    }
  else
    {
      for ( pvx = iterQEMesh.PeekFirstVertex();
	    pvx != NULL;
	    pvx = iterQEMesh.PeekNextVertex() )
	{
	  ::RenderNormals(pvx, bLimitSurface, bCalculateNormals, bSelective);
	}
    }
}

static VOID RenderGeometry(QEFacet *pfa, BOOL bFaceted, BOOL bLimitSurface, 
			   BOOL bCalculateNormals, BOOL bSelective)
{
  ASSERT( pfa != NULL );

  QEFacetIter iter;
  QEEdge *pe;
#ifdef COUNT_TRIANGLES
  UINT_32 uiVertices;

  uiVertices = 0;
#endif // COUNT_TRIANGLES

  //if ( !bFaceted && bLimitSurface )
  //  {
  //    fprintf(stderr, "  START: Triangle Fan: ");
  //    pfa->GetNormal().Display();
  //  }

  glNormal3fv(pfa->GetNormal().Get());
  glBegin(GL_POLYGON);
  //glBegin(GL_TRIANGLE_FAN);
  {
    iter.Init(pfa);
    
    for ( pe = iter.PeekFirstEdge();
	  pe != NULL;
	  pe = iter.PeekNextEdge() )
      {
	RenderGeometry(pe, pe->GetWedge(), bFaceted, bLimitSurface, 
		       bCalculateNormals, bSelective);
#ifdef COUNT_TRIANGLES
	uiVertices++;
#endif // COUNT_TRIANGLES
      }
  }
  glEnd();

#ifdef COUNT_TRIANGLES
  ASSERT( uiVertices > 2 );
  s_uiTrianglesRendered += (uiVertices-2);
#endif // COUNT_TRIANGLES

  //if ( !bFaceted && bLimitSurface )
  //  {
  //    fprintf(stderr, "  END  : Triangle Fan: ");
  //    pfa->GetNormal().Display();
  //  }
}

static VOID RenderNormals(QEFacet *pfa, BOOL bLimitSurface)
{
  ASSERT( pfa != NULL );

  Vector vNormal;
  Point pt;

  if ( bLimitSurface )
    {
      pfa->CalculateLimitCentroidAndNormal(pt, vNormal);
    }
  else
    {
      pt = pfa->GetCentroid();
      vNormal = pfa->GetNormal();
    }
  vNormal[3] = 0.0;
  vNormal *= cfNormalLength;

  glBegin(GL_LINES);
  {    
    glVertex3fv(pt.Get());
    pt += vNormal;
    glVertex3fv(pt.Get());
  }
  glEnd();
}

static VOID RenderGeometry(QEEdge *pe, QEWedge *pwe, BOOL bFaceted, BOOL bLimitSurface,
			   BOOL bCalculateNormals, BOOL bSelective)
{
  QEVertex *pvx;
  QEEdge *peCurr;
  Vector vNormal;

  //if ( !bFaceted )
  //  {
  //    pwe = pvx->GetWedge();
  //  }
  //pwe->Render();

  pvx = pe->GetVertex();

  if ( bLimitSurface )
    {
      if ( !bFaceted )
	{
	  vNormal = pvx->GetNormal();
	  if ( bCalculateNormals )
	    {
	      // Check if this vertex is on a sharp crease of a subdivision surface 
	      switch ( pvx->GetSharpVertexType() )
		{
		case QE::svtRegularCrease:
		case QE::svtNonregularCrease: 
		case QE::svtCorner:
		  //fprintf(stderr, "  calculating normal ...\n");

		  peCurr = pe;
		  do
		    {
		      if ( peCurr->GetKnotInterval() != 0 )
			{
			  if ( bSelective )
			    {
			      vNormal = QEMesh::CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(peCurr);
			    }
			  else
			    {
			      vNormal = QEMesh::CalculateVertexNormalSubdivisionSharpLoop(peCurr);
			    }
			  break;
			}
		      peCurr = peCurr->GetEdgePrev(1);
		    }
		  while ( peCurr != pe );
		  break;
		default:
		  break;
		}
	    }
	  //fprintf(stderr, "    normal: ");
	  //vNormal.Display();
	  glNormal3fv(vNormal.Get());
	  //fprintf(stderr, "    vertex: ");
	  //pvx->GetLimitPosition().Display();
	}
      glVertex3fv(pvx->GetLimitPosition().Get());
    }
  else
    {
      if ( !bFaceted )
	{
	  vNormal = pvx->CalculateAveragedFacetNormal();
	  glNormal3fv(vNormal.Get());
	}
      glVertex3fv(pvx->GetPoint().Get());
    }
}

static VOID RenderNormals(QEVertex *pvx, BOOL bLimitSurface,
			  BOOL bCalculateNormals, BOOL bSelective)
{
  ASSERT( pvx != NULL );

  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vNormal;
  Point pt, ptEnd;

  if ( bLimitSurface )
    {
      pt = pvx->GetLimitPosition();
      vNormal = pvx->GetNormal();
      if ( bCalculateNormals )
	{
	  switch ( pvx->GetSharpVertexType() )
	    {
	    case QE::svtRegularCrease:
	    case QE::svtNonregularCrease: 
	    case QE::svtCorner:
	      iterQEVertex.Init(pvx);
	      for ( pe = iterQEVertex.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEVertex.PeekNextEdge() )
		{
		  if ( pe->GetKnotInterval() != 0 )
		    {
		      if ( bSelective )
			{
			  vNormal = QEMesh::CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(pe);
			}
		      else
			{
			  if ( pe->GetFacet() == QEMesh::GetNullFacet() )
			    {
			      pe = pe->GetEdgeNext(1);
			    }
			  vNormal = QEMesh::CalculateVertexNormalSubdivisionSharpLoop(pe);
			}

		      vNormal[3] = 0.0;
		      vNormal *= cfNormalLength;
		      
		      glBegin(GL_LINES);
		      {    
			glVertex3fv(pt.Get());
			ptEnd = pt + vNormal;
			glVertex3fv(ptEnd.Get());
		      }
		      glEnd();
		    }
		}
	      break;
	    default:
	      vNormal[3] = 0.0;
	      vNormal *= cfNormalLength;
	      
	      glBegin(GL_LINES);
	      {    
		glVertex3fv(pt.Get());
		pt += vNormal;
		glVertex3fv(pt.Get());
	      }
	      glEnd();
	      break;
	    }
	}
      else
	{
	  vNormal[3] = 0.0;
	  vNormal *= cfNormalLength;
	  
	  glBegin(GL_LINES);
	  {    
	    glVertex3fv(pt.Get());
	    pt += vNormal;
	    glVertex3fv(pt.Get());
	  }
	  glEnd();
	}
    }
  else if ( bCalculateNormals )
    {
      pt = pvx->GetPoint();
      vNormal = pvx->GetNormal();

      vNormal[3] = 0.0;
      vNormal *= cfNormalLength;
      
      glBegin(GL_LINES);
      {    
	glVertex3fv(pt.Get());
	pt += vNormal;
	glVertex3fv(pt.Get());
      }
      glEnd();
    }
  else
    {
      pt = pvx->GetPoint();
      vNormal = pvx->CalculateAveragedFacetNormal();

      vNormal[3] = 0.0;
      vNormal *= cfNormalLength;
      
      glBegin(GL_LINES);
      {    
	glVertex3fv(pt.Get());
	pt += vNormal;
	glVertex3fv(pt.Get());
      }
      glEnd();
    }
}

/*
static VOID Render(QEWedge *pwe)
{
  ASSERT( pwe != NULL );

  glTexCoord2fv(pwe->GetTexture().Get());
  glNormal3fv(pwe->GetNormal().Get());
}
*/

//////////////////////////////////////////////////////////////////////
// Voronoi and Offset Rendering
//

static Vector HSVtoRGB(const Vector &cvHSV);

static Vector HSVtoRGB(const Vector &cvHSV)
{
  Vector vHSV;
  Vector vRGB;
  FLOAT fFloor, fF, fV, fP, fQ, fT;
  UINT_32 i;

  vHSV = cvHSV;

  if ( vHSV(1) <= 0.0 ) 
    {
      vRGB.Set(vHSV(2), vHSV(2), vHSV(2));
    } 
  else 
    {
      while ( vHSV(0) >= 360.0 ) 
	{
	   vHSV[0] -= 360.0;
	}

      while ( vHSV(0) < 0.0 )
	{
	  vHSV[0] -= 360.0;
	}

      vHSV[0] /= 60.0;
      fFloor = floor(vHSV(0));
      i = (UINT_32)fFloor;
      fF = vHSV(0) - fFloor;
	
      fV = vHSV(2);
      fP = vHSV(2) * (1.0 - vHSV(1));
      fQ = vHSV(2) * (1.0 - (vHSV(1) * fF));
      fT = vHSV(2) * (1.0 - (vHSV(1) * (1.0 - fF)));

      switch ( i )
	{
	case 0:
	  { 
	    vRGB.Set(fV, fT, fP);
	    break;
	  }
	case 1:
	  { 
	    vRGB.Set(fQ, fV, fP);
	    break;
	  }
	case 2:
	  { 
	    vRGB.Set(fP, fV, fT);
	    break;
	  }
	case 3:
	  { 
	    vRGB.Set(fP, fQ, fV);
	    break;
	  }
	case 4:
	  { 
	    vRGB.Set(fT, fP, fV);
	    break;
	  }
	case 5:
	  { 
	    vRGB.Set(fV, fP, fQ);
	    break;
	  }
	default:
	  {
	    ASSERT( FALSE );
	  }
	}
    }

    return vRGB;
}

VOID RenderVoronoi(QEMesh *pmeDelaunay, QEMesh *pmeVoronoi, QE::VoronoiType eType)
{
  ASSERT( pmeDelaunay != NULL );

  QEMeshIter iterQEMesh;
  QEEdge *pe;
  QEVertex *pvx;
  Point pt, ptMidpoint;
  Vector v, u;

  switch ( eType )
    {
    case QE::vtDelaunay:
    case QE::vtDivideAndConquer:
    case QE::vtDivideAndConquerAlternating:
      {
	glBegin(GL_LINES);
	{
	  ASSERT( pmeVoronoi != NULL );
	  iterQEMesh.Init(pmeVoronoi);
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      if ( pe->GetFacet() != QEMesh::GetNullFacet() &&
		   pe->GetEdge()->GetFacet() != QEMesh::GetNullFacet() )
		{
		  if ( pe->GetEdgePrev(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() &&
		       pe->GetEdgeNext(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
		    {
		      ptMidpoint = ( pe->GetFacet()->GetForwardVertex()->GetPoint() + 
				     pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint() ) / 2.0;
		      v = pe->GetVertex()->GetPoint();
		      ASSERT( v[3] == 0.0 );
		      pt = ptMidpoint + v;
		      glVertex3fv(pt.Get());

		      v = pe->GetEdge()->GetVertex()->GetPoint();
		      ASSERT( v[3] == 0.0 );
		      pt = ptMidpoint + v;
		      glVertex3fv(pt.Get());
		    }
		  else
		    {
		      pt = pe->GetVertex()->GetPoint();
		      if ( pe->GetEdgePrev(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
			{
			  v = pt;
			  ASSERT( v[3] == 0.0 );
			  v *= 100.0;
			  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetFacet() != QEMesh::GetNullFacet() );
			  pt = pe->GetEdge()->GetVertex()->GetPoint();
			  pt += v;
			}
		      glVertex3fv(pt.Get());
		      
		      pt = pe->GetEdge()->GetVertex()->GetPoint();
		      if ( pe->GetEdgeNext(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
			{
			  v = pt;
			  ASSERT( v[3] == 0.0 );
			  v *= 100.0;
			  ASSERT( pe->GetEdgePrev(0)->GetEdge()->GetFacet() != QEMesh::GetNullFacet() );
			  pt = pe->GetVertex()->GetPoint();
			  pt += v;
			}
		      glVertex3fv(pt.Get());
		    }
		}
	    }
	}
	glEnd();
      }
      break;

    case QE::vtCones:
      {
	UINT_32 uiSlices;

	uiSlices = 32;

	UINT_32 uiVertices;
	CBound bound;
	Point ptApex;
	Vector vHSV, vRGB;
	FLOAT fRadius;
	FLOAT fTheta, fThetaDelta;
	FLOAT fHue, fHueDelta;
	UINT_32 i;

	// Count Vertices and Compute Bounding Box
	uiVertices = 0;
	iterQEMesh.Init(pmeDelaunay);
	for ( pvx = iterQEMesh.PeekFirstVertex();
	      pvx != NULL;
	      pvx = iterQEMesh.PeekNextVertex() )
	  {
	    bound.Insert(pvx->GetPoint());
	    uiVertices++;
	  }

	if ( uiVertices > 0 )
	  {
	    fRadius = 2.0 * bound.GetRadius();
	    fRadius *= 2.0 / sqrt(uiVertices);
	    if ( fRadius == 0.0 )
	      {
		fRadius = 1.0;
	      }
	    
	    fHueDelta = 360.0/(FLOAT)uiVertices;
	    fThetaDelta = 2.0*PI/(FLOAT)uiSlices;
	    
	    glPolygonMode(GL_FRONT, GL_FILL);
	    fHue = 0.0;
	    for ( pvx = iterQEMesh.PeekFirstVertex();
		  pvx != NULL;
		  pvx = iterQEMesh.PeekNextVertex() )
	      {
		ptApex = pvx->GetPoint();
		
		vHSV.Set(fHue, 0.75, 1.0);
		vRGB = HSVtoRGB(vHSV);

		glColor3fv(vRGB.Get());
		glBegin(GL_TRIANGLE_FAN);
		{
		  glVertex3fv(ptApex.Get());

		  fTheta = 0.0;
		  for ( i = 0; i <= uiSlices; i++ )
		    {
		      v.Set(fRadius*cos(fTheta), fRadius*sin(fTheta), -fRadius);
		      pt = ptApex + v;
		      glVertex3fv(pt.Get());

		      fTheta += fThetaDelta;
		    }
		}
		glEnd();

		fHue += fHueDelta;
	      }
	  }
	break;
      }

    case QE::vtPlanes:
      {
	UINT_32 uiVertices;
	CBound bound;
	Point ptApex;
	Vector vHSV, vRGB;
	FLOAT fRadius, fZ;
	FLOAT fHue, fHueDelta;

	// Count Vertices and Compute Bounding Box
	uiVertices = 0;
	iterQEMesh.Init(pmeDelaunay);
	for ( pvx = iterQEMesh.PeekFirstVertex();
	      pvx != NULL;
	      pvx = iterQEMesh.PeekNextVertex() )
	  {
	    pt = pvx->GetPoint();
	    pt[2] = pt(0)*pt(0) + pt(1)*pt(1);
	    bound.Insert(pt);
	    uiVertices++;
	  }

	if ( uiVertices > 0 )
	  {
	    pt = bound.Get(CBound::bbtMax);
	    fZ = -pt(2);
	    fRadius = 2.0 * bound.GetRadius();
	    //fZ = -fRadius*fRadius;
	    //fRadius = sqrt(fRadius*fRadius*(fRadius*fRadius + 1.0));
	    fRadius *= 2.0 / sqrt(uiVertices);
	    if ( fRadius == 0.0 )
	      {
		fRadius = 1.0;
	      }
	    
	    fHueDelta = 360.0/(FLOAT)uiVertices;
	    
	    glPolygonMode(GL_FRONT, GL_FILL);
	    fHue = 0.0;
	    for ( pvx = iterQEMesh.PeekFirstVertex();
		  pvx != NULL;
		  pvx = iterQEMesh.PeekNextVertex() )
	      {
		ptApex = pvx->GetPoint();
		ptApex[2] = ptApex(0)*ptApex(0) + ptApex(1)*ptApex(1) + fZ;

		vHSV.Set(fHue, 0.75, 1.0);
		vRGB = HSVtoRGB(vHSV);

		u.Set(1.0, 0.0, 2.0*ptApex(0));
		u.Normalize();
		u *= fRadius;
		v.Set(0.0, 1.0, 2.0*ptApex(1));
		v.Normalize();
		v *= fRadius;

		glColor3fv(vRGB.Get());
		glBegin(GL_POLYGON);
		{
		  pt = ptApex + (u + v);
		  glVertex3fv(pt.Get());

		  pt = ptApex + (-u + v);
		  glVertex3fv(pt.Get());

		  pt = ptApex + (-u - v);
		  glVertex3fv(pt.Get());

		  pt = ptApex + (u - v);
		  glVertex3fv(pt.Get());
		}
		glEnd();

		fHue += fHueDelta;
	      }
	  }
	break;
      }

    default:
      {
	ASSERT( FALSE );
	break;
      }
    }      
}

VOID RenderVoronoiPolygon(QEMesh *pmeCDT, QEMesh *pmeVoronoi, QE::VoronoiType eType, 
			  UINT_32 uiSubdivisionLevelMax, FLOAT fTolerance, 
			  BOOL bDrawHeight)
{
  ASSERT( pmeCDT != NULL );

  QEMeshIter iterQEMesh;
  QEEdge *pe;
  QEVertex *pvx;
  Point ppt[2];
  Point pt, ptMidpoint;
  FLOAT fEpsilon;
  Vector v, u;

  switch ( eType )
    {
    case QE::vtDelaunay:
    case QE::vtPlanes:
      {
	RenderGeometry(pmeVoronoi);
      }
      break;

    case QE::vtDivideAndConquer:
    case QE::vtDivideAndConquerAlternating:
      {
	glBegin(GL_LINES);
	{
	  ASSERT( pmeVoronoi != NULL );
	  iterQEMesh.Init(pmeVoronoi);
	  for ( pe = iterQEMesh.PeekFirstEdge();
		pe != NULL;
		pe = iterQEMesh.PeekNextEdge() )
	    {
	      if ( pe->GetFacet() != QEMesh::GetNullFacet() &&
		   pe->GetEdge()->GetFacet() != QEMesh::GetNullFacet() &&
		   !( pe->GetFacet()->GetForwardVertex() != NULL &&
		      pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL &&
		      pe->GetFacet()->GetForwardVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex() ) )
		{
		  if ( pe->GetEdgePrev(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() &&
		       pe->GetEdgeNext(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
		    {
		      if ( pe->GetFacet()->GetForwardVertex() != NULL &&
			   pe->GetEdge()->GetFacet()->GetForwardVertex() != NULL )
			{
			  ptMidpoint = ( pe->GetFacet()->GetForwardVertex()->GetPoint() + 
					 pe->GetEdge()->GetFacet()->GetForwardVertex()->GetPoint() ) / 2.0;
			}
		      else if ( pe->GetFacet()->GetForwardEdge() != NULL &&
				pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL )
			{
			  ptMidpoint = ( pe->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint() + 
					 pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint() + 
					 pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint() + 
					 pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint() ) / 4.0;
			}
		      else if ( pe->GetFacet()->GetForwardEdge() != NULL )
			{
			  ASSERT( (pe->GetFacet()->GetForwardEdge()->GetVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex()) ||
				  (pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() == pe->GetEdge()->GetFacet()->GetForwardVertex()) );

			  ptMidpoint = ( pe->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint() + 
					 pe->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint() ) / 2.0;
			}
		      else 
			{
			  ASSERT( pe->GetEdge()->GetFacet()->GetForwardEdge() != NULL );
			  ASSERT( (pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex() == pe->GetFacet()->GetForwardVertex()) ||
				  (pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex() == pe->GetFacet()->GetForwardVertex()) );

			  ptMidpoint = ( pe->GetEdge()->GetFacet()->GetForwardEdge()->GetVertex()->GetPoint() + 
					 pe->GetEdge()->GetFacet()->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint() ) / 2.0;
			}

		      ptMidpoint[3] = 1.0;

		      v = pe->GetVertex()->GetPoint();
		      ASSERT( v[3] == 0.0 );
		      pt = ptMidpoint + v;
		      if ( !bDrawHeight )
			{
			  pt[2] = 0.0;
			}
		      glVertex3fv(pt.Get());

		      v = pe->GetEdge()->GetVertex()->GetPoint();
		      ASSERT( v[3] == 0.0 );
		      pt = ptMidpoint + v;
		      if ( !bDrawHeight )
			{
			  pt[2] = 0.0;
			}
		      glVertex3fv(pt.Get());
		    }
		  else
		    {
		      QEEdge *ppe[2], *ppeZone[2], *peSwap;
		      QEVertex *ppvxZone[2], *pvxSwap;
		      UINT_32 i, uiVertices, uiEdges;

		      fEpsilon = 0.001;

		      ppe[0] = pe;
		      ppe[1] = pe->GetEdge();

		      ppvxZone[0] = NULL;
		      ppvxZone[1] = NULL;

		      ppeZone[0] = NULL;
		      ppeZone[1] = NULL;

		      uiVertices = 0;
		      uiEdges = 0;
		      for ( i = 0; i < 2; i++ )
			{
			  if ( ppe[i]->GetFacet()->GetForwardVertex() != NULL )
			    {
			      ppvxZone[i] = ppe[i]->GetFacet()->GetForwardVertex();
			      uiVertices++;
			    }
			  else
			    {
			      ASSERT( ppe[i]->GetFacet()->GetForwardEdge() != NULL );
			      ppeZone[i] = ppe[i]->GetFacet()->GetForwardEdge();
			      uiEdges++;
			    }
			}

		      switch ( uiVertices )
			{
			case 0:
			  break;
			case 1:
			  if ( ppvxZone[0] == NULL )
			    {
			      peSwap = ppe[0];
			      ppe[0] = ppe[1];
			      ppe[1] = peSwap;

			      peSwap = ppeZone[0];
			      ppeZone[0] = ppeZone[1];
			      ppeZone[1] = peSwap;

			      pvxSwap = ppvxZone[0];
			      ppvxZone[0] = ppvxZone[1];
			      ppvxZone[1] = pvxSwap;
			    }
			  break;
			case 2:
			  break;
			default:
			  ASSERT( FALSE );
			  break;
			}

		      for ( i = 0; i < 2; i++ )
			{
			  ppt[i] = ppe[i]->GetVertex()->GetPoint();
			  if ( ppe[i]->GetEdgePrev(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
			    {
			      v = ppt[i];
			      if ( v[3] != 0.0 )
				{
				  fprintf(stderr, "START: QERender ERROR\n");
				  fprintf(stderr, "  Vertex   %p: ( %f, %f, %f, %f )\n",
					  ppe[i]->GetVertex(), v(0), v(1), v(2), v(3));
				  fprintf(stderr, "  Neighbor %p: ( %f, %f, %f, %f )\n",
					  ppe[i]->GetEdge()->GetVertex(), 
					  ppe[i]->GetEdge()->GetVertex()->GetPoint()(0),
					  ppe[i]->GetEdge()->GetVertex()->GetPoint()(1),
					  ppe[i]->GetEdge()->GetVertex()->GetPoint()(2),
					  ppe[i]->GetEdge()->GetVertex()->GetPoint()(3));
				  fprintf(stderr, "END  : QERender ERROR\n");
				}
			      //ASSERT( v[3] == 0.0 );
			      v *= 100.0;
			      ASSERT( ppe[i]->GetEdgeNext(0)->GetEdge()->GetFacet() != QEMesh::GetNullFacet() );
			      ppt[i] = ppe[i]->GetEdge()->GetVertex()->GetPoint();
			      ppt[i] += v;

			      switch ( uiVertices )
				{
				case 0:
				  break;
				case 1:
				  break;
				case 2:
				  {
				    v = ppvxZone[0]->GetPoint() - ppt[i];
				    v[2] = 0.0;
				    v[3] = 0.0;
				    if ( ppt[i][2] < 0.0 )
				      {
					ppt[i][2] = -v.Magnitude();
				      }
				    else
				      {
					ppt[i][2] = v.Magnitude();
				      }
				  }
				  break;
				default:
				  ASSERT( FALSE );
				  break;
				}
			    }
			}

		      switch ( uiVertices )
			{
			case 0:
			  {
			    ::RenderLineSegment(ppt[0], ppt[1], bDrawHeight);
			  }
			  break;
			case 1:
			  {
			    if ( ppe[0]->GetEdgePrev(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() ||
				 ppe[1]->GetEdgePrev(0)->GetEdge()->GetFacet() == QEMesh::GetNullFacet() )
			      {
				::RenderLineSegment(ppt[0], ppt[1], bDrawHeight);
			      }
			    else if ( ppvxZone[0] == ppeZone[1]->GetVertex() ||
				      ppvxZone[0] == ppeZone[1]->GetEdge()->GetVertex() )
			      {
				::RenderLineSegment(ppt[0], ppt[1], bDrawHeight);
			      }
			    else
			      {
				::RenderParabolaAdaptive(ppvxZone[0]->GetPoint(),
							 ppeZone[1]->GetVertex()->GetPoint(),
							 ppeZone[1]->GetEdge()->GetVertex()->GetPoint(),
							 ppt[0], ppt[1],
							 0, uiSubdivisionLevelMax, 
							 fTolerance, bDrawHeight);
			      }
			  }
			  break;
			case 2:
			  {
			    ::RenderHyperbolaAdaptive(ppvxZone[0]->GetPoint(),
						      ppvxZone[1]->GetPoint(),
						      ppt[0], ppt[1],
						      0, uiSubdivisionLevelMax, 
						      fTolerance, bDrawHeight);
			  }
			  break;
			default:
			  ASSERT( FALSE );
			  break;
			}
		    }
		}
	    }
	}
	glEnd();
      }
      break;

    case QE::vtCones:
      {
	UINT_32 uiSlices;

	uiSlices = 32;

	UINT_32 uiVertices;
	CBound bound;
	Point ptApex;
	Vector vHSV, vRGB;
	FLOAT fRadius;
	FLOAT fTheta, fThetaDelta;
	FLOAT fHue, fHueDelta;
	UINT_32 i;

	Point ppt[2];
	Vector vOrtho;

	// Count Vertices and Compute Bounding Box
	uiVertices = 0;
	iterQEMesh.Init(pmeCDT);
	for ( pvx = iterQEMesh.PeekFirstVertex();
	      pvx != NULL;
	      pvx = iterQEMesh.PeekNextVertex() )
	  {
	    bound.Insert(pvx->GetPoint());
	    uiVertices++;
	  }

	if ( uiVertices > 0 )
	  {
	    fRadius = 2.0 * bound.GetRadius();
	    fRadius *= 2.0 / sqrt(uiVertices);
	    if ( fRadius == 0.0 )
	      {
		fRadius = 1.0;
	      }
	    
	    fHueDelta = 360.0/(2.0*(FLOAT)uiVertices);
	    fThetaDelta = 2.0*PI/(FLOAT)uiSlices;
	    
	    glPolygonMode(GL_FRONT, GL_FILL);
	    fHue = 0.0;
	    for ( pvx = iterQEMesh.PeekFirstVertex();
		  pvx != NULL;
		  pvx = iterQEMesh.PeekNextVertex() )
	      {
		ptApex = pvx->GetPoint();
		
		vHSV.Set(fHue, 0.75, 1.0);
		vRGB = HSVtoRGB(vHSV);

		glColor3fv(vRGB.Get());
		glBegin(GL_TRIANGLE_FAN);
		{
		  glVertex3fv(ptApex.Get());

		  fTheta = 0.0;
		  for ( i = 0; i <= uiSlices; i++ )
		    {
		      v.Set(fRadius*cos(fTheta), fRadius*sin(fTheta), -fRadius);
		      pt = ptApex + v;
		      glVertex3fv(pt.Get());

		      fTheta += fThetaDelta;
		    }
		}
		glEnd();

		fHue += fHueDelta;
	      }

	    for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
		  pe != NULL;
		  pe = iterQEMesh.PeekNextEdgeBoundary() )
	      {
		vHSV.Set(fHue, 0.75, 1.0);
		vRGB = HSVtoRGB(vHSV);

		glColor3fv(vRGB.Get());
		glBegin(GL_TRIANGLE_STRIP);
		{
		  ppt[0] = pe->GetEdge()->GetVertex()->GetPoint();
		  ppt[1] = pe->GetVertex()->GetPoint();

		  v = ppt[1] - ppt[0];
		  vOrtho[0] = -v(1);
		  vOrtho[1] = v(0);
		  vOrtho.Normalize();

		  vOrtho *= fRadius;
		  v = vOrtho;
		  v[2] = -fRadius;

		  pt = ppt[1] + v;
		  glVertex3fv(pt.Get());
		  pt = ppt[0] + v;
		  glVertex3fv(pt.Get());

		  pt = ppt[1];
		  glVertex3fv(pt.Get());
		  pt = ppt[0];
		  glVertex3fv(pt.Get());

		  v = -vOrtho;
		  v[2] = -fRadius;

		  pt = ppt[1] + v;
		  glVertex3fv(pt.Get());
		  pt = ppt[0] + v;
		  glVertex3fv(pt.Get());
		}
		glEnd();

		fHue += fHueDelta;
	      }
	  }
      }
      break;


    default:
      {
	ASSERT( FALSE );
      }
      break;
    }      
}

// The number of line segments you would like to use to approximate a unit circle
static const UINT_32 s_cuiThetaSlicesMax = 30;
static const FLOAT s_cfThetaDelta = DEGTORAD(2.0);

VOID RenderOffsetContour(QEMesh *pmeOffset, BOOL bDrawHeight)
{
  ASSERT( pmeOffset != NULL );

  QEMeshIter iterQEMesh;
  QEEdge *pe;
  Point ptCenter, ppt[2], pt;
  Vector pv[2], v, vAxis;
  Matrix mRot;
  FLOAT fDot, fMag, fTheta, fThetaDelta, fR, fEpsilon;
  UINT_32 i, uiSlices;
  //UINT_32 uiSlicesMin;

  fEpsilon = 0.001;

  glBegin(GL_LINES);
  {
    iterQEMesh.Init(pmeOffset);
    for ( pe = iterQEMesh.PeekFirstEdge();
	  pe != NULL;
	  pe = iterQEMesh.PeekNextEdge() )
      {
	if ( pe->GetFacet() != QEMesh::GetNullFacet() )
	  {
	    if ( pe->GetForwardEdge() == NULL )
	      {
		ppt[0] = pe->GetVertex()->GetPoint();
		ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();		
		if ( !bDrawHeight )
		  {
		    ppt[0][2] = 0.0;
		    ppt[1][2] = 0.0;
		  }
		glVertex3fv(ppt[0].Get());	
		glVertex3fv(ppt[1].Get());	
	      }
	    else
	      {
		//////////////////////////////////////////////////
		// Circular Arc

		ptCenter = pe->GetForwardEdge()->GetVertex()->GetPoint();
		ppt[0] = pe->GetVertex()->GetPoint();
		ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
		if ( !bDrawHeight )
		  {
		    ptCenter[2] = 0.0;
		    ppt[0][2] = 0.0;
		    ppt[1][2] = 0.0;
		  }
		pv[0] = ppt[0] - ptCenter;
		pv[0][2] = 0.0;
		pv[0][3] = 0.0;
		fR = pv[0].Magnitude();

		pv[1] = ppt[1] - ptCenter;
		pv[1][2] = 0.0;
		pv[1][3] = 0.0;

		fDot = pv[0].Dot(pv[1]);
		vAxis = pv[0] * pv[1];
		fMag = vAxis.Magnitude();
		vAxis.Normalize();

		fTheta = atan2(fMag, fDot);

		fThetaDelta = DEGTORAD(10.0);
		if ( fR != 0.0 )
		  {
		    fR = 1.0 - (fEpsilon/fR);
		    if ( fR > 0.0 )
		      {
			fR = 2.0*acos(fR);
			if ( fThetaDelta > fR )
			  {
			    fThetaDelta = fR;
			  }
		      }
		  }
		uiSlices = (UINT_32)(1.0 + fTheta/fThetaDelta);
		uiSlices = MIN(uiSlices, s_cuiThetaSlicesMax);
		//printf("Circle Slices = %ld\n", uiSlices);
		
		/*
		fThetaDelta = fTheta;
		if ( fThetaDelta < 0.0 )
		  {
		    fThetaDelta += 2.0*PI;
		    if ( fThetaDelta < 0.0 )
		      {
			fThetaDelta = 0.0;
		      }
		  }
		uiSlicesMin = (UINT_32)(1.0 + fThetaDelta/(2.0*PI));
		fThetaDelta = fThetaDelta * fR;
		uiSlices = (UINT_32)(1.0 + fThetaDelta/s_cfThetaDelta);

		uiSlices = MIN(MAX(uiSlices, uiSlicesMin), s_cuiThetaSlicesMax);	    
		//printf("Circle Slices = %ld\n", uiSlices);
		*/

		fThetaDelta = fTheta / (FLOAT)uiSlices;

		mRot.SetRotate(vAxis, fThetaDelta);

		v = pv[0];

		v[2] = ppt[0][2]; 

		pt = ptCenter + v;
		for ( i = 0; i < uiSlices; i++ )
		  {
		    glVertex3fv(pt.Get());	

		    v = mRot * v;

		    pt = ptCenter + v;
		    glVertex3fv(pt.Get());	
		  }
	      }
	  }
      }
  }
  glEnd();
}

static VOID RenderLineSegment(const Point &pt0, const Point &pt1, BOOL bDrawHeight)
{
  Point pt;

  pt = pt0;
  if ( !bDrawHeight )
    {
      pt[2] = 0.0;
    }
  glVertex3fv(pt.Get());
  
  pt = pt1;
  if ( !bDrawHeight )
    {
      pt[2] = 0.0;
    }
  glVertex3fv(pt.Get());      
}

static VOID RenderParabolaAdaptive(const Point &ptV0, 
				   const Point &ptE10, const Point &ptE11, 
				   const Point &pt0, const Point &pt1,
				   UINT_32 uiLevel, UINT_32 uiLevelMax,
				   FLOAT fEpsilon, BOOL bDrawHeight)
{
  Vector ppvV[2][2], pvE[2];
  Vector pvT[2], vT;
  Vector v, u;
  FLOAT fZ, fE1dotE1, fE10xE11, fTdotT;
  FLOAT fPlusOrMinus, fDisc;
  FLOAT fError;
  Point pt;

  //fprintf(stderr, "Parabola: Level = %ld / %ld\n", uiLevel, uiLevelMax);

  if ( uiLevel >= uiLevelMax )
    {
      ::RenderLineSegment(pt0, pt1, bDrawHeight);
      return;
    }
  /*
  fprintf(stderr, "START: RenderParabolaAdaptive\n");
  fprintf(stderr, "  V0: ( %.3f, %.3f, %.3f )\n", ptV0(0), ptV0(1), ptV0(2));
  fprintf(stderr, "  E1: ( %.3f, %.3f, %.3f ) => ( %.3f, %.3f, %.3f )\n", 
	  ptE10(0), ptE10(1), ptE10(2),
	  ptE11(0), ptE11(1), ptE11(2));
  fprintf(stderr, "  T0: ( %.3f, %.3f, %.3f )\n", pt0(0), pt0(1), pt0(2));
  fprintf(stderr, "  T1: ( %.3f, %.3f, %.3f )\n", pt1(0), pt1(1), pt1(2));
  fprintf(stderr, "END  : RenderParabolaAdaptive\n");
  */
  ppvV[0][0].Null();
  ppvV[0][1].Null();
  ppvV[1][0] = ptE10 - ptV0;
  ppvV[1][1] = ptE11 - ptV0;

  pvE[0].Null();
  pvE[1] = ptE11 - ptE10;

  pvT[0] = pt0 - ptV0;
  pvT[1] = pt1 - ptV0;
  
  vT = pt1 - pt0;

  fPlusOrMinus = (pvT[0][0]*pvE[1][0]) + (pvT[0][1]*pvE[1][1]); 
  fZ = (pvT[1][0]*pvE[1][0]) + (pvT[1][1]*pvE[1][1]);
  if ( fabs(fZ) > fabs(fPlusOrMinus) )
    {
      fPlusOrMinus = fZ;
    }
  if ( fPlusOrMinus < 0.0 )
    {
      fPlusOrMinus = -1.0;
    }
  else
    {
      fPlusOrMinus = 1.0;
    }

  fE1dotE1 = (pvE[1][0]*pvE[1][0]) + (pvE[1][1]*pvE[1][1]);
  fE10xE11 = (ppvV[1][0][0]*ppvV[1][1][1]) - (ppvV[1][0][1]*ppvV[1][1][0]);
  fTdotT = (vT[0]*vT[0]) + (vT[1]*vT[1]) + (vT[2]*vT[2]);

  // Pick a distance for a middle point
  fZ = fE10xE11*(fTdotT - vT[2]*vT[2]) / (2.0*sqrt(fE1dotE1)*(fTdotT - 2.0*vT[2]*vT[2]));
  pt[2] = fZ;
  if ( pvT[0][2] < 0.0 || pvT[1][2] < 0.0 )
    {
      pt[2] = -pt[2];
    }

  // Compute the X,Y values at that distance
  fDisc = (2.0*sqrt(fE1dotE1)*fZ - fE10xE11)*fE10xE11;
  if ( fDisc < 0.0 )
    {
      fDisc = 0.0;
    }

  pt[0] = ( ptV0(0) + 
	    ( -(sqrt(fE1dotE1)*fZ - fE10xE11)*pvE[1][1] +
	      fPlusOrMinus*pvE[1][0]*sqrt(fDisc) ) / fE1dotE1 );
  pt[1] = ( ptV0(1) + 
	    ( (sqrt(fE1dotE1)*fZ - fE10xE11)*pvE[1][0] +
	      fPlusOrMinus*pvE[1][1]*sqrt(fDisc) ) / fE1dotE1 );

  // Compute error for new middle point
  v = pt - pt0;
  u = v*vT;
  fError = u.Magnitude() / fTdotT;

  if ( fError > fEpsilon )
    {
      ::RenderParabolaAdaptive(ptV0, ptE10, ptE11, 
			       pt0, pt,
			       uiLevel + 1, uiLevelMax,
			       fEpsilon, bDrawHeight);
      ::RenderParabolaAdaptive(ptV0, ptE10, ptE11, 
			       pt, pt1,
			       uiLevel + 1, uiLevelMax,
			       fEpsilon, bDrawHeight);
    }
  else
    {
      ::RenderLineSegment(pt0, pt1, bDrawHeight);
    }
}

static VOID RenderHyperbolaAdaptive(const Point &ptV0, const Point &ptV1, 
				    const Point &pt0, const Point &pt1,
				    UINT_32 uiLevel, UINT_32 uiLevelMax,
				    FLOAT fEpsilon, BOOL bDrawHeight)
{
  Vector pvV[2];
  Vector pvT[2], vT;
  Vector v, u;
  FLOAT fZ, fV1dotV1, fTdotT;
  FLOAT fPlusOrMinus, fDisc;
  FLOAT fError;
  Point pt;

  //fprintf(stderr, "Hyperbola: Level = %ld / %ld\n", uiLevel, uiLevelMax);

  if ( uiLevel >= uiLevelMax || !bDrawHeight )
    {
      ::RenderLineSegment(pt0, pt1, bDrawHeight);
      return;
    }

  /*
  fprintf(stderr, "START: RenderHyperbolaAdaptive\n");
  fprintf(stderr, "  V0: ( %.3f, %.3f, %.3f)\n", ptV0(0), ptV0(1), ptV0(2));
  fprintf(stderr, "  V1: ( %.3f, %.3f, %.3f)\n", ptV1(0), ptV1(1), ptV1(2));
  fprintf(stderr, "  T0: ( %.3f, %.3f, %.3f)\n", pt0(0), pt0(1), pt0(2));
  fprintf(stderr, "  T1: ( %.3f, %.3f, %.3f)\n", pt1(0), pt1(1), pt1(2));
  fprintf(stderr, "END  : RenderHyperbolaAdaptive\n");
  */

  pvV[0].Null();
  pvV[1] = ptV1 - ptV0;

  pvT[0] = pt0 - ptV0;
  pvT[1] = pt1 - ptV0;
  
  vT = pt1 - pt0;

  fPlusOrMinus = (pvT[0][1]*pvV[1][0]) - (pvT[0][0]*pvV[1][1]); 
  fZ = (pvT[1][1]*pvV[1][0]) - (pvT[1][0]*pvV[1][1]);
  if ( fabs(fZ) > fabs(fPlusOrMinus) )
    {
      fPlusOrMinus = fZ;
    }
  if ( fPlusOrMinus < 0.0 )
    {
      fPlusOrMinus = -1.0;
    }
  else
    {
      fPlusOrMinus = 1.0;
    }

  fV1dotV1 = (pvV[1][0]*pvV[1][0]) + (pvV[1][1]*pvV[1][1]);
  fTdotT = (vT[0]*vT[0]) + (vT[1]*vT[1]) + (vT[2]*vT[2]);

  // Pick a distance for a middle point
  fZ = fV1dotV1*(fTdotT - vT[2]*vT[2]) / (4.0*(fTdotT - 2.0*vT[2]*vT[2]));
  fZ = sqrt(fZ);
  if ( pvT[0][2] < 0.0 || pvT[1][2] < 0.0 )
    {
      fZ = -fZ;
    }
  pt[2] = fZ;
  fZ = fabs(fZ);

  // Compute the X,Y values at that distance
  fDisc = (4.0*fZ*fZ - fV1dotV1)/fV1dotV1;
  if ( fDisc < 0.0 )
    {
      fDisc = 0.0;
    }

  pt[0] = ( ptV0(0) + pvV[1][0]/2.0 - fPlusOrMinus*pvV[1][1]*sqrt(fDisc)/2.0 );
  pt[1] = ( ptV0(1) + pvV[1][1]/2.0 + fPlusOrMinus*pvV[1][0]*sqrt(fDisc)/2.0 );

  // Compute error for new middle point
  v = pt - pt0;
  u = v*vT;
  fError = u.Magnitude() / fTdotT;

  if ( fError > fEpsilon )
    {
      ::RenderHyperbolaAdaptive(ptV0, ptV1, 
			       pt0, pt,
			       uiLevel + 1, uiLevelMax,
			       fEpsilon, bDrawHeight);
      ::RenderHyperbolaAdaptive(ptV0, ptV1, 
			       pt, pt1,
			       uiLevel + 1, uiLevelMax,
			       fEpsilon, bDrawHeight);
    }
  else
    {
      ::RenderLineSegment(pt0, pt1, bDrawHeight);
    }
}
