#ifndef __QEVERTEX_H__
#define __QEVERTEX_H__

#include "qemem.h"
#include "qe.h"
#include "apm2d.h"

class QEEdge;
class QEWedge;
class QEFacet;

class QEVertexIter;

//////////////////////////////////////////////////////////////////////
// QEVertex Class
//

class QEVertex
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend class QEVertexIter;

//////////////////////////////////////////////////
// Enumerated Types
//
public:
  enum ComparisonType { ctXYZ, ctYXZ, ctSize };

//////////////////////////////////////////////////
// Initialization
//
public:
  QEVertex();
  virtual ~QEVertex();
  virtual VOID Init(const Point &pt);
  virtual VOID Uninit();

  virtual BOOL Initialized();
  virtual BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID Preprocess();
  Vector CalculateAveragedFacetNormal();

protected:

//////////////////////////////////////////////////
// Rendering
//
public:

//////////////////////////////////////////////////
// Output Functions
//
public:
  VOID Display();

//////////////////////////////////////////////////
// Ordering Operators
//
public:
  QEVertex &operator=(const QEVertex &vx);
  BOOL operator<(const QEVertex &vx) const;
  BOOL operator>(const QEVertex &vx) const;

  BOOL operator==(const QEVertex &vx) const;
  BOOL operator!=(const QEVertex &vx) const;

  BOOL operator<=(const QEVertex &vx) const;
  BOOL operator>=(const QEVertex &vx) const;

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline const Point &GetPoint() const;
  inline VOID SetPoint(const Point &pt);
  inline const Point &GetLimitPosition() const;
  inline VOID SetLimitPosition(const Point &ptLimit);
  inline const Vector &GetNormal() const;
  inline VOID SetNormal(const Vector &vNormal);
  inline const Vector &GetTexture() const;
  inline VOID SetTexture(const Vector &vTexture);

  inline QEWedge *GetWedge();
  inline VOID SetWedge(QEWedge *pw);
  inline QEEdge *GetEdge();
  inline VOID SetEdge(QEEdge *pe);

  inline UINT_32 GetID();
  inline VOID SetID(UINT_32 uiID);
  inline UINT_32 GetValence() const;
  inline VOID SetValence(UINT_32 uiValence);
  inline UINT_32 GetGeneration() const;
  inline VOID SetGeneration(UINT_32 uiGeneration);
  inline UINT_32 GetBirth();
  inline VOID SetBirth(UINT_32 uiBirth);
  inline QE::SharpVertexType GetSharpVertexType();
  inline VOID SetSharpVertexType(QE::SharpVertexType eType);

  inline APM2D::InscribedCircleCenterType GetInscribedCircleCenterType();
  inline VOID SetInscribedCircleCenterType(APM2D::InscribedCircleCenterType eICCT);

  // Forward
  inline QEVertex *GetForwardVertex();
  inline VOID SetForwardVertex(QEVertex *pvx);
  inline QEFacet *GetForwardFacet();
  inline VOID SetForwardFacet(QEFacet *pfa);

  // Heap
  inline CHeapEntry<VOID*, FLOAT> *GetHeapEntry();
  inline VOID SetHeapEntry(CHeapEntry<VOID*, FLOAT> *phe);
  inline CHeapEntry<VOID*, FLOAT> *GetHeapEntryReverse();
  inline VOID SetHeapEntryReverse(CHeapEntry<VOID*, FLOAT> *phe);

  // Linked List
  inline QEVertex *GetPrevLink();
  inline VOID SetPrevLink(QEVertex *pvx);
  inline QEVertex *GetNextLink();
  inline VOID SetNextLink(QEVertex *pvx);

//////////////////////////////////////////////////
// Class Functions
//
public:
  static inline ComparisonType GetComparisonType();
  static inline VOID SetComparisonType(ComparisonType eComparison);
  static inline VOID SetComparisonTypeDefault();

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static ComparisonType m_eComparison;

//////////////////////////////////////////////////
// Member Variables
//
protected:
  Point m_pt;
  Point m_ptLimit;
  Vector m_vNormal;
  Vector m_vTexture;

  QEWedge *m_pQEWedge;
  QEEdge *m_pQEEdge;

  UINT_32 m_uiID;
  // Generation info packed in a single word (32 bits)
  UINT_32 m_uiValence          : 7;
  UINT_32 m_uiGeneration       : 11;
  UINT_32 m_uiBirth            : 11;
  UINT_32 m_uiSharpVertexType  : 3;

  // Voronoi Stuff
  APM2D::InscribedCircleCenterType m_eICCT;

  // Forwarding Variables
  QEVertex *m_pvxForward;
  QEFacet *m_pfaForward;

  // Heap for selective processes
  CHeapEntry<VOID*, FLOAT> *m_pphe[2];
			  
  // Listing in a QEMesh or on a Free List
  QEVertex *m_pvxLink[2];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline const Point &QEVertex::GetPoint() const
{
  return m_pt;
}

inline VOID QEVertex::SetPoint(const Point &pt)
{
  m_pt = pt;
}

inline const Point &QEVertex::GetLimitPosition() const
{
  return m_ptLimit;
}

inline VOID QEVertex::SetLimitPosition(const Point &ptLimit)
{
  m_ptLimit = ptLimit;
}

inline const Vector &QEVertex::GetNormal() const
{
  return m_vNormal;
}

inline VOID QEVertex::SetNormal(const Vector &vNormal)
{
  m_vNormal = vNormal;
}

inline const Vector &QEVertex::GetTexture() const
{
  return m_vTexture;
}

inline VOID QEVertex::SetTexture(const Vector &vTexture)
{
  m_vTexture = vTexture;
}

inline QEWedge *QEVertex::GetWedge()
{
  return m_pQEWedge;
}

inline VOID QEVertex::SetWedge(QEWedge *pw)
{
  m_pQEWedge = pw;
}

inline QEEdge *QEVertex::GetEdge()
{
  return m_pQEEdge;
}

inline VOID QEVertex::SetEdge(QEEdge *pe)
{
  m_pQEEdge = pe;
}

inline UINT_32 QEVertex::GetID()
{
  return m_uiID;
}

inline VOID QEVertex::SetID(UINT_32 uiID)
{
  m_uiID = uiID;
}

inline UINT_32 QEVertex::GetValence() const
{
  return m_uiValence;
}

inline VOID QEVertex::SetValence(UINT_32 uiValence)
{
  m_uiValence = uiValence;
}

inline UINT_32 QEVertex::GetGeneration() const
{
  return m_uiGeneration;
}

inline VOID QEVertex::SetGeneration(UINT_32 uiGeneration)
{
  m_uiGeneration = uiGeneration;
}

inline UINT_32 QEVertex::GetBirth()
{
  return m_uiBirth;
}

inline VOID QEVertex::SetBirth(UINT_32 uiBirth)
{
  m_uiBirth = uiBirth;
}

inline QE::SharpVertexType QEVertex::GetSharpVertexType()
{
  return (QE::SharpVertexType)m_uiSharpVertexType;
}

inline VOID QEVertex::SetSharpVertexType(QE::SharpVertexType eType)
{
  m_uiSharpVertexType = (UINT_32)eType;
}

inline APM2D::InscribedCircleCenterType QEVertex::GetInscribedCircleCenterType()
{
  return m_eICCT;
}

inline VOID QEVertex::SetInscribedCircleCenterType(APM2D::InscribedCircleCenterType eICCT)
{
  m_eICCT = eICCT;
}

inline QEVertex *QEVertex::GetForwardVertex()
{
  return m_pvxForward;
}

inline VOID QEVertex::SetForwardVertex(QEVertex *pvx)
{
  m_pvxForward = pvx;
}

inline QEFacet *QEVertex::GetForwardFacet()
{
  return m_pfaForward;
}

inline VOID QEVertex::SetForwardFacet(QEFacet *pfa)
{
  m_pfaForward = pfa;
}

inline CHeapEntry<VOID*, FLOAT> *QEVertex::GetHeapEntry()
{
  return m_pphe[0];
}

inline VOID QEVertex::SetHeapEntry(CHeapEntry<VOID*, FLOAT> *phe)
{
  m_pphe[0] = phe;
}

inline CHeapEntry<VOID*, FLOAT> *QEVertex::GetHeapEntryReverse()
{
  return m_pphe[1];
}

inline VOID QEVertex::SetHeapEntryReverse(CHeapEntry<VOID*, FLOAT> *phe)
{
  m_pphe[1] = phe;
}

inline QEVertex *QEVertex::GetPrevLink()
{
  return m_pvxLink[0];
}

inline VOID QEVertex::SetPrevLink(QEVertex *pvx)
{
  m_pvxLink[0] = pvx;
}

inline QEVertex *QEVertex::GetNextLink()
{
  return m_pvxLink[1];
}

inline VOID QEVertex::SetNextLink(QEVertex *pvx)
{
  m_pvxLink[1] = pvx;
}

inline QEVertex::ComparisonType QEVertex::GetComparisonType()
{
  return m_eComparison;
}

inline VOID QEVertex::SetComparisonType(QEVertex::ComparisonType eComparison)
{
  ASSERT( eComparison < ctSize );
  m_eComparison = eComparison;
}

inline VOID QEVertex::SetComparisonTypeDefault()
{
  m_eComparison = ctYXZ;
}

//////////////////////////////////////////////////////////////////////
// List Functions
//

inline VOID DisplayValue(const QEVertex &vx) 
{ 
  printf("<%.3f, %.3f, %.3f>", vx.GetPoint()(X), vx.GetPoint()(Y), vx.GetPoint()(Z));
}

inline VOID SetNull(QEVertex &vx) 
{ 
  Point pt;

  pt.Null();
  vx.SetPoint(pt);
}

inline VOID SetMaxValue(QEVertex &vx)
{ 
  Point pt;

  pt.Set(INFINITY, INFINITY, INFINITY);
  vx.SetPoint(pt);
}

inline VOID SetMinValue(QEVertex &vx)
{ 
  Point pt;

  pt.Set(-INFINITY, -INFINITY, -INFINITY);
  vx.SetPoint(pt);
}

inline UINT_32 Hash(const QEVertex &vx) 
{ 
  Point pt;
  FLOAT f;
  UINT_32 i, uiHash;

  pt = vx.GetPoint();
  uiHash = 0;
  for ( i = 0; i < 3; i++ )
    {
      f = pt(i);
      uiHash += *((UINT_32*)&f);
    }

  return uiHash; 
}

//////////////////////////////////////////////////////////////////////
// QEVertexIter Class
//

class QEVertexIter
{
//////////////////////////////////////////////////
// Initialization
//
public:
  QEVertexIter();
  ~QEVertexIter();
  VOID Init(QEVertex *pQEVertex);
  VOID Uninit();

  BOOL Initialized();

  BOOL Valid();

//////////////////////////////////////////////////
// Iteration Functions
//
public:
  QEFacet *PeekFirstFacet();
  QEFacet *PeekNextFacet();
  QEFacet *PeekNextWrapFacet();
  QEFacet *PeekLastFacet();
  QEFacet *PeekPrevFacet();
  QEFacet *PeekPrevWrapFacet();

  QEEdge *PeekFirstEdge();
  QEEdge *PeekNextEdge();
  QEEdge *PeekNextWrapEdge();
  QEEdge *PeekLastEdge();
  QEEdge *PeekPrevEdge();
  QEEdge *PeekPrevWrapEdge();

//////////////////////////////////////////////////
// Accessor Functions
//
public:

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  QEVertex *m_pQEVertex;

  QEEdge *m_peHead;
  QEEdge *m_peCurr;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

#endif // __QEVERTEX_H__



