#include "scanner.h"
#include <stdarg.h>

//////////////////////////////////////////////////////////////////////////////
// CFileFrame Class
//

CFileFrame::CFileFrame()
{
  m_pcName = NULL;
  m_fpFile = NULL;
  m_yyBufferState = NULL;
  m_uiLine = 0;
  m_uiColumn = 0;
}

CFileFrame::~CFileFrame()
{
  Uninit();
}

VOID CFileFrame::Init(CHAR *pcName, FILE *fpFile, YY_BUFFER_STATE yyBufferState,
		      UINT_32 uiLine, UINT_32 uiColumn)
{
  ASSERT( pcName != NULL && fpFile != NULL );
  ASSERT( uiLine > 0 && uiColumn > 0);

  m_pcName = pcName;
  m_fpFile = fpFile;
  m_yyBufferState = yyBufferState;
  m_uiLine = uiLine;
  m_uiColumn = uiColumn;
}

VOID CFileFrame::Uninit()
{
  if ( m_pcName == NULL )
    {
      delete m_pcName;
      m_pcName = NULL;
    }

  if ( m_fpFile != NULL )
    {
      if ( m_fpFile != stdin )
	{
	  fclose(m_fpFile);
	}
      m_fpFile = NULL;
    }
  
  m_uiLine = 0;
  m_uiColumn = 0;
}

//////////////////////////////////////////////////////////////////////////////
// CScanner Class
//

CScanner::CScanner()
{
  m_fpError = NULL;

  m_uiErrorCount = 0;
  m_uiWarningCount = 0;

  // Comments
  m_uiCommentLevel = 0;
  m_uiCommentStartLine = 0;
  m_uiCommentStartColumn = 0;

  // Strings
  m_pcStringBuffer[0] = '\0';
  m_pcStringBufferPos = m_pcStringBuffer;
  m_uiStringStartLine = 0;
  m_uiStringStartColumn = 0;

  // Tcl Blocks
  m_pcTclBlockBuffer[0] = '\0';
  m_pcTclBlockBufferPos = m_pcTclBlockBuffer;
  m_uiTclBlockLevel = 0;
  m_uiTclBlockStartLine = 0;
  m_uiTclBlockStartColumn = 0;

  // FileNames
  m_pcFileNameBuffer[0] = '\0';
  m_pcFileNameBufferPos = m_pcFileNameBuffer;
  m_uiFileNameStartLine = 0;
  m_uiFileNameStartColumn = 0;

  m_pyylloc = NULL;

  m_pfyy_BeginInitial = NULL;
  m_pfyy_GetCurrentBuffer = NULL;
  m_pfyy_CreateBuffer = NULL;
  m_pfyy_DeleteBuffer = NULL;
  m_pfyy_SwitchToBuffer = NULL;
  m_pfyy_SetValueInt = NULL;
  m_pfyy_SetValueFloat = NULL;
  m_pfyy_SetValueString = NULL;

  // Search Paths
  m_uiDirectories = 0;
  m_ppcDirectories = NULL;
}

CScanner::~CScanner()
{
  Uninit();
}

VOID CScanner::Init(CHAR *pcFile, CHAR **ppcDirectories, UINT_32 uiDirectories, 
		    YYLTYPE *pyylloc, int *pyyleng,
		    YY_BEGIN_INITIAL_TYPE *pfyy_BeginInitial,
		    YY_GET_CURRENT_BUFFER_TYPE *pfyy_GetCurrentBuffer,
		    YY_CREATE_BUFFER_TYPE *pfyy_CreateBuffer,
		    YY_DELETE_BUFFER_TYPE *pfyy_DeleteBuffer,
		    YY_SWITCH_TO_BUFFER_TYPE *pfyy_SwitchToBuffer,
		    YY_SET_VALUE_INT_TYPE *pfyy_SetValueInt,
		    YY_SET_VALUE_FLOAT_TYPE *pfyy_SetValueFloat,
		    YY_SET_VALUE_STRING_TYPE *pfyy_SetValueString)
{
  UINT_32 i;

  ASSERT( pcFile != NULL );
  ASSERT( (ppcDirectories == NULL && uiDirectories == 0) ||
	  (ppcDirectories != NULL && uiDirectories > 0) );
  ASSERT( pyylloc != NULL && pyyleng != NULL &&
	  pfyy_BeginInitial != NULL &&
	  pfyy_GetCurrentBuffer != NULL &&
	  pfyy_CreateBuffer != NULL &&
	  pfyy_DeleteBuffer != NULL &&
	  pfyy_SwitchToBuffer != NULL &&
	  pfyy_SetValueInt != NULL &&
	  pfyy_SetValueFloat != NULL &&
	  pfyy_SetValueString != NULL );

  m_pyylloc = pyylloc;
  m_pyylloc->first_line = 1;
  m_pyylloc->first_column = 1;

  m_pyyleng = pyyleng;

  m_pfyy_BeginInitial = pfyy_BeginInitial;
  m_pfyy_GetCurrentBuffer = pfyy_GetCurrentBuffer;
  m_pfyy_CreateBuffer = pfyy_CreateBuffer;
  m_pfyy_DeleteBuffer = pfyy_DeleteBuffer;
  m_pfyy_SwitchToBuffer = pfyy_SwitchToBuffer;
  m_pfyy_SetValueInt = pfyy_SetValueInt;
  m_pfyy_SetValueFloat = pfyy_SetValueFloat;
  m_pfyy_SetValueString = pfyy_SetValueString;

  m_fpError = stderr;

  m_hashIdentifierTable.Init(3);
  m_hashStringTable.Init(3);
  m_hashFileNameTable.Init(3);
  m_hashOpenFileTable.Init(3);
  m_listOpenFileStack.Init();

  m_uiErrorCount = 0;
  m_uiWarningCount = 0;

  // Comments
  m_uiCommentLevel = 0;
  m_uiCommentStartLine = 0;
  m_uiCommentStartColumn = 0;

  // Strings
  m_pcStringBuffer[0] = '\0';
  m_pcStringBufferPos = m_pcStringBuffer; 
  m_uiStringStartLine = 0;
  m_uiStringStartColumn = 0;

  // Tcl Blocks
  m_pcTclBlockBuffer[0] = '\0';
  m_pcTclBlockBufferPos = m_pcTclBlockBuffer;
  m_uiTclBlockLevel = 0;
  m_uiTclBlockStartLine = 0;
  m_uiTclBlockStartColumn = 0;

  // FileNames
  m_pcFileNameBuffer[0] = '\0';
  m_pcFileNameBufferPos = m_pcFileNameBuffer;
  m_uiFileNameStartLine = 0;
  m_uiFileNameStartColumn = 0;

  // Search Path
  m_uiDirectories = uiDirectories;
  m_ppcDirectories = NULL;
  if ( ppcDirectories != NULL )
    {
      m_ppcDirectories = new CHAR*[m_uiDirectories];
      ASSERT( m_ppcDirectories != NULL );
      for ( i = 0; i < m_uiDirectories; i++ )
	{
	  m_ppcDirectories[i] = new CHAR[strlen(ppcDirectories[i]) + 1];
	  ASSERT( m_ppcDirectories[i] != NULL );
	  strcpy(m_ppcDirectories[i], ppcDirectories[i]);
	}
    }

  PushFile(pcFile);
}

VOID CScanner::Uninit()
{
  UINT_32 i;
  CHAR *pc;

  for ( pc = (CHAR*)(m_hashIdentifierTable.RemoveFirst()); pc != NULL;
	pc = (CHAR*)(m_hashIdentifierTable.RemoveFirst()) )
    {
      delete pc;
    }

  for ( pc = (CHAR*)(m_hashStringTable.RemoveFirst()); pc != NULL;
	pc = (CHAR*)(m_hashStringTable.RemoveFirst()) )
    {
      delete pc;
    }

  if ( m_ppcDirectories != NULL )
    {
      for ( i = 0; i < m_uiDirectories; i++ )
	{
	  if ( m_ppcDirectories[i] != NULL )
	    {
	      delete []m_ppcDirectories[i];
	      m_ppcDirectories[i] = NULL;
	    }
	}

      delete []m_ppcDirectories;
      m_ppcDirectories = NULL;
    }
  m_uiDirectories = 0;
}

VOID CScanner::ScanInteger(CHAR *pcToken)
{
  ASSERT( pcToken != NULL );

  m_pfyy_SetValueInt(atoi(pcToken));

  ConsumeToken();
}

VOID CScanner::ScanFloat(CHAR *pcToken)
{
  ASSERT( pcToken != NULL );

  m_pfyy_SetValueFloat((FLOAT)atof(pcToken));

  ConsumeToken();
}

VOID CScanner::ScanChar(CHAR *pcBuffer)
{
  CHAR *pc;
  CHAR c;

  ASSERT( pcBuffer != NULL );
  ASSERT( strlen(pcBuffer) >= 3 && strlen(pcBuffer) <= 6 );
  ASSERT( pcBuffer[0] == '\'' );

  c = '\0';

  pc = pcBuffer;

  // Advance past the '\'' character
  pc++;

  if ( pc[0] == '\\' )
    {
      // Advance past the '\'' character
      pc++;

      // It is an escape sequence
      if ( pc[0] >= '0' && pc[0] <= '7' )
	{
	  // Octal constant
	  // This will cover '\0'
	  CHAR *pcExtra;
	  UINT_32 uiValue;
	  
	  pcExtra = NULL;
	  uiValue = strtoul(pc, &pcExtra, 8);
	  ASSERT( pcExtra == NULL || (pcExtra[0] == '\'' && pcExtra[1] == '\0') );
	  if ( uiValue > 255 )
	    {
	      HandleError(m_pyylloc->first_line, m_pyylloc->first_column, 
			  "Bad escape sequence \"%s\"", pcBuffer);
	    }
	  else
	    {
	      c = (CHAR)uiValue;
	    }
	}
      else if ( pc[0] == 'x' && pc[1] != '\0' )
	{
	  // Hexidecimal constant
	  CHAR *pcExtra;
	  UINT_32 uiValue;
	  
	  // Advance past the 'x'
	  pc++;
	  
	  pcExtra = NULL;
	  uiValue = strtoul(pc, &pcExtra, 16);
	  ASSERT( pcExtra == NULL || (pcExtra[0] == '\'' && pcExtra[1] == '\0') );
	  if ( uiValue > 255 )
	    {
	      HandleError(m_pyylloc->first_line, m_pyylloc->first_column, 
			  "Bad escape sequence \"%s\"", pcBuffer);
	    }
	  else
	    {
	      c = (CHAR)uiValue;
	    }
	}
      else
	{
	  switch ( pc[0] )
	    {
	    case '\"':
	      c = '\"';
	      break;
	      
	    case '\'':
	      c = '\'';
	      break;
	      
	    case 'a':
	      c = '\a';
	      break;
	      
	    case 'b':
	      c = '\b';
	      break;
	      
	    case 'f':
	      c = '\f';
	      break;
	      
	    case 'n':
	      c = '\n';
	      break;
	      
	    case 'r':
	      c = '\r';
	      break;
	  
	    case 't':
	      c = '\t';
	      break;
	      
	    case 'v':
	      c = '\v';
	      break;
	      
	    default:
	      // This cover '\?' and '\\' and any other character
	      c = pc[0];
	      break;
	    }
	  ASSERT( pc[1] == '\'' && pc[2] == '\0' );
	}
    }
  else
    {
      // It is a normal character
      ASSERT( strlen(pcBuffer) == 3 );

      c = pc[0];
      ASSERT( pc[1] == '\'' && pc[2] == '\0' );
    }

  m_pfyy_SetValueInt(c);

  ConsumeToken();
}

VOID CScanner::ScanString()
{
  CHAR *pc;

  pc = (CHAR*)m_hashStringTable.Peek(m_pcStringBuffer);
  if ( pc == NULL )
    {
      pc = new CHAR[strlen(m_pcStringBuffer)+1];
      ASSERT( pc != NULL );
      strcpy(pc, m_pcStringBuffer);
      m_hashStringTable.Insert(pc, pc);
    }

  m_pfyy_SetValueString(pc);
}

VOID CScanner::ScanTclBlock()
{
  // Tcl Blocks are really just strings so I am going to 
  // store them in the string table.  Noone is doing look ups
  // on Strings or TclBlocks ever anyway.

  CHAR *pc;

  pc = (CHAR*)m_hashStringTable.Peek(m_pcTclBlockBuffer);
  if ( pc == NULL )
    {
      pc = new CHAR[strlen(m_pcTclBlockBuffer)+1];
      ASSERT( pc != NULL );
      strcpy(pc, m_pcTclBlockBuffer);
      m_hashStringTable.Insert(pc, pc);
    }

  m_pfyy_SetValueString(pc);
}

CHAR *CScanner::ScanFileName()
{
/*
  CHAR *pc;

  pc = (CHAR*)m_hashFileNameTable.Peek(m_pcFileNameBuffer);
  if ( pc == NULL )
    {
      pc = new CHAR[strlen(m_pcFileNameBuffer)+1];
      ASSERT( pc != NULL );
      strcpy(pc, m_pcFileNameBuffer);
      m_hashFileNameTable.Insert(pc, pc);
    }

  return pc;
*/
  return m_pcFileNameBuffer;
}

VOID CScanner::ScanIdentifier(CHAR *pcToken)
{
  CHAR *pc;

  ASSERT( pcToken != NULL );

  pc = (CHAR*)m_hashIdentifierTable.Peek(pcToken);
  if ( pc == NULL )
    {
      pc = new CHAR[strlen(pcToken)+1];
      ASSERT( pc != NULL );
      strcpy(pc, pcToken);
      m_hashIdentifierTable.Insert(pc, pc);
    }
    

  //pc = FindIdentifier (pcToken, FALSE);

  m_pfyy_SetValueString(pc);

  ConsumeToken();
}

VOID CScanner::InsertSystemIdentifier(const CHAR *cpcIdentifier)
{
  ASSERT( m_hashIdentifierTable.Peek(cpcIdentifier) == NULL );

  m_hashIdentifierTable.Insert((VOID*)cpcIdentifier, cpcIdentifier);
}

CHAR *CScanner::FindIdentifier (CHAR *cpcIdentifier, BOOL bDeleteStr)
{
  CHAR *pc;

  ASSERT( cpcIdentifier != NULL );

  pc = (CHAR*)m_hashIdentifierTable.Peek(cpcIdentifier);
  if ( pc == NULL )
    {
      pc = new CHAR[strlen(cpcIdentifier)+1];
      ASSERT( pc != NULL );
      strcpy(pc, cpcIdentifier);
      m_hashIdentifierTable.Insert(pc, pc);
    }

  if (bDeleteStr)  // free memory if requested
    delete cpcIdentifier;

  return pc;
}

VOID CScanner::PushFile(CHAR *pcFileName)
{
  UINT_32 i;
  CHAR *pcName;
  FILE *fpFile;
  CFileFrame *pCFileFrame;
  YY_BUFFER_STATE yyBufferState;
  static CHAR pcFullFileNameBuffer[400];

  ASSERT( pcFileName != NULL );

  strcpy(pcFullFileNameBuffer, pcFileName);

  // Open the File
  if ( pcFullFileNameBuffer[0] == '\0' ) 
    {
      fpFile = stdin;
    } 
  else 
    {
      fpFile = fopen(pcFullFileNameBuffer, "r");
      // Check the search path for the file
      for ( i = 0; fpFile == NULL && i < m_uiDirectories; i++ )
	{
	  sprintf(pcFullFileNameBuffer, "%s/%s", m_ppcDirectories[i], pcFileName);
	  fpFile = fopen(pcFullFileNameBuffer, "r");
	}
    }

  if ( fpFile == NULL )
    {
      HandleError(m_pyylloc->first_line, m_pyylloc->first_column, "Cannot open file: \"%s\"", pcFileName);
      return;
    }

  // Check for Circular Dependencies
  if ( m_hashOpenFileTable.Peek(pcFullFileNameBuffer) != NULL )
    {
      if ( fpFile != stdin )
	{
	  fclose(fpFile);
	}
      HandleError(m_pyylloc->first_line, m_pyylloc->first_column, "Circular dependency for file: \"%s\"", pcFileName);
      return;
    }

  // Check if we have already sourced this file
  pcName = (CHAR*)m_hashFileNameTable.Peek(pcFullFileNameBuffer);
  if ( pcName == NULL )
    {
      // We have not, so store its name so that we check to see if it is
      pcName = new CHAR[strlen(pcFullFileNameBuffer)+1];
      ASSERT( pcName != NULL );
      strcpy(pcName, pcFullFileNameBuffer);
      m_hashFileNameTable.Insert(pcName, pcName);
    }
  else
    {
      // We are not rereading files in SLIDE
      HandleWarning(m_pyylloc->first_line, m_pyylloc->first_column, "File \"%s\" has already been read, so skipping.\n", pcName);
      if ( fpFile != stdin )
	{
	  fclose(fpFile);
	}
      return;
    }

  // Store the Open File Pointer in a Hash Table to make sure there are no 
  // circular dependencies
  m_hashOpenFileTable.Insert(pcName, pcName);

  // Push the File Name onto the open file stack
  pCFileFrame = new CFileFrame;
  ASSERT( pCFileFrame != NULL );
  pCFileFrame->Init(pcName, fpFile, m_pfyy_GetCurrentBuffer(),
		    m_pyylloc->first_line, m_pyylloc->first_column); 

  m_listOpenFileStack.InsertFirst(pCFileFrame, 0);

  yyBufferState = m_pfyy_CreateBuffer(fpFile);
  ASSERT( yyBufferState != NULL );
  m_pfyy_SwitchToBuffer(yyBufferState);
  m_pyylloc->text = pcName;
  m_pyylloc->first_line = 1;
  m_pyylloc->first_column = 1;
}

BOOL CScanner::PopFile()
{
  CFileFrame *pCFileFrame, *pCFileFramePeek;
  BOOL bDone;

  // Remove the current File Frame
  pCFileFrame = (CFileFrame*)(m_listOpenFileStack.RemoveFirst());
  ASSERT( pCFileFrame != NULL );

  // Check if this was an included file
  pCFileFramePeek = (CFileFrame*)(m_listOpenFileStack.PeekFirst());

  if ( m_uiCommentStartLine != 0 )
    {
      HandleError(m_uiCommentStartLine, m_uiCommentStartColumn, 
		  "Unterminated comment.");
      m_uiCommentLevel = 0;
      m_uiCommentStartLine = 0;
      m_uiCommentStartColumn = 0;
      m_pfyy_BeginInitial();
    }

  if ( m_uiStringStartLine != 0 )
    {
      HandleError(m_uiStringStartLine, m_uiStringStartColumn, 
	          "Unterminated string.");
      m_uiStringStartLine = 0;
      m_uiStringStartColumn = 0;
      m_pfyy_BeginInitial();
    }

  if ( m_uiTclBlockStartLine != 0 )
    {
      HandleError(m_uiTclBlockStartLine, m_uiTclBlockStartColumn, 
	          "Unterminated Tcl block.");
      m_uiTclBlockLevel = 0;
      m_uiTclBlockStartLine = 0;
      m_uiTclBlockStartColumn = 0;
      m_pfyy_BeginInitial();
    }

  if ( m_uiFileNameStartLine != 0 )
    {
      HandleError(m_uiFileNameStartLine, m_uiFileNameStartColumn, 
	          "Unterminated file name in include statement.");
      m_uiFileNameStartLine = 0;
      m_uiFileNameStartColumn = 0;
      m_pfyy_BeginInitial();
    }

  if ( pCFileFramePeek == NULL )
    {
      // We are done!!!
      m_pyylloc->text = NULL;

      bDone = TRUE;
    }
  else
    {  
      m_pyylloc->first_line = pCFileFrame->m_uiLine;
      m_pyylloc->first_column = pCFileFrame->m_uiColumn;

      ASSERT( pCFileFrame->m_yyBufferState != NULL );
      m_pfyy_DeleteBuffer(m_pfyy_GetCurrentBuffer());
      m_pfyy_SwitchToBuffer(pCFileFrame->m_yyBufferState);

      m_pyylloc->text = pCFileFramePeek->m_pcName;

      bDone = FALSE;
    }

  m_hashOpenFileTable.Remove(pCFileFrame->m_pcName);
  delete pCFileFrame;
  pCFileFrame = NULL;

  return bDone;
}

VOID CScanner::HandleError(UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcFormat, ...)
{
  va_list vaArgs;

  m_uiErrorCount++;
  fprintf(m_fpError, "%s: %ld: %ld: Syntax Error: ", 
	  (m_pyylloc->text == NULL) ? ("") : (m_pyylloc->text), 
	  (uiLine == 0) ? (m_pyylloc->first_line) : (uiLine), 
	  (uiColumn == 0) ? (m_pyylloc->first_column) : (uiColumn));

  // Now print the message sent by the program
  va_start(vaArgs, pcFormat);
  vfprintf(m_fpError, pcFormat, vaArgs);
  va_end(vaArgs);

  fprintf(m_fpError, "\n"); 
}

VOID CScanner::HandleWarning(UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcFormat, ...)
{
  va_list vaArgs;

  m_uiWarningCount++;
  fprintf(m_fpError, "%s: %ld: %ld: Syntax Warning: ", 
	  (m_pyylloc->text == NULL) ? ("") : (m_pyylloc->text), 
	  (uiLine == 0) ? (m_pyylloc->first_line) : (uiLine), 
	  (uiColumn == 0) ? (m_pyylloc->first_column) : (uiColumn));

  // Now print the message sent by the program
  va_start(vaArgs, pcFormat);
  vfprintf(m_fpError, pcFormat, vaArgs);
  va_end(vaArgs);

  fprintf(m_fpError, "\n"); 
}





