#include "slidebsplinecurve.h"
#include "slidesweep.h"
#include "slidesweepmorph.h"
#include "slidecrosssection.h"
#include "slidepoint.h"
#include "slidecontrolpoint.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "bsplinemath.h"

static const UINT_32 cuiCurveType = 0; // Bspline
static const UINT_32 cuiOrder = 4;
static const UINT_32 cuiKnots = 8;

static const UINT_32 cuiSlices = 5;

static const UINT_32 cuiDrawControls = 0;
static const UINT_32 cuiCurveClosed = 0;
static const FLOAT cfNormalLength = 1.0;
static const FLOAT cfNormalRatio = 1.0 / 6.0;

//////////////////////////////////////////////////////////////////////
// CSLIDEBSplineCurve Class
//

CSLIDEBSplineCurve::CSLIDEBSplineCurve()
{
  // Sweep
  m_bControlWarpDynamic = TRUE;
  m_pControlWarpScale = NULL;
  m_pControlWarpRotate = NULL;
  m_pWarpScale = NULL;
  m_pWarpRotate = NULL;

  // Curve
  m_fNormalLength = cfNormalLength;
  m_dyniDrawControls.SetValue(cuiDrawControls);
  m_dyniSlices.SetValue(cuiSlices);
  m_dyniCurveClosed.SetValue(cuiCurveClosed);
  m_uiSlices = cuiSlices;

  m_uiCurveType = cuiCurveType;
  m_bKnotsChanged = FALSE;
  m_bControlPointsDynamic = TRUE;

  m_uiOrderPrev = cuiOrder;
  m_uiOrder = cuiOrder;
  m_uiKnots = cuiKnots;
  m_uiControlPoints = cuiKnots - cuiOrder;
  m_uiControlPointsPrev = m_uiControlPoints;

  m_pfKnot = NULL;
  m_plistKnot = NULL;
  m_pControlPoint = NULL;
  m_plistControlPointIDs = NULL;
  m_plistCSLIDEControlPoint = new CListInt();

  m_pPoint = NULL;
  m_pvNormal = NULL;
  m_pvBinormal = NULL;
  m_pvTangent = NULL;
  m_pvNormalTorsionReduced = NULL;
  m_pvBinormalTorsionReduced = NULL;
  m_pvTangentTorsionReduced = NULL;
  m_pmSweepFrenetFrame = NULL;

  m_uiDerivatives = 0;
  m_uiDerivativeControlPoints = 0;
  m_ppDerivativeTable = NULL;
  m_ppfDeBoorTable = NULL;
}

CSLIDEBSplineCurve::~CSLIDEBSplineCurve()
{
}

VOID CSLIDEBSplineCurve::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDESweepPath::Init(pcFile, uiLine, uiColumn, pcID);

  // Sweep
  m_bControlWarpDynamic = TRUE;
  m_listControlWarpScale.Init();
  m_listControlWarpRotate.Init();

  // Curve
  m_fNormalLength = cfNormalLength;
  m_dyniDrawControls.SetValue(cuiDrawControls);
  m_dyniSlices.SetValue(cuiSlices);
  m_dyniCurveClosed.SetValue(cuiCurveClosed);
  m_uiSlices = cuiSlices;

  m_uiCurveType = cuiCurveType;
  m_bKnotsChanged = FALSE;
  m_bControlPointsDynamic = TRUE;

  m_uiOrderPrev = cuiOrder;
  m_uiOrder = cuiOrder;
  m_uiKnots = cuiKnots;
  m_uiControlPoints = cuiKnots - cuiOrder;
  m_uiControlPointsPrev = m_uiControlPoints;

  ReallocateKnots();
  ReallocateControlPoints();
  ReallocateControlWarps();

  InitKnots();
  InitControlPoints();
  m_listCSLIDEPoint.Init();
  m_plistCSLIDEControlPoint->Init();
  InitControlWarps();

  ReallocateWarps();
  ReallocatePoints();
  ReallocateNormals();
  ReallocateBinormals();
  ReallocateTangents();
  ReallocateNormalsTorsionReduced();
  ReallocateBinormalsTorsionReduced();
  ReallocateTangentsTorsionReduced();
  ReallocateSweepFrenetFrames();

  m_uiDerivatives = 4;
  m_uiDerivativeControlPoints = 0;
  ReallocateDerivativeTable();
  ReallocateDeBoorTable();

  // Initialize
  CalculateDerivativeTable(m_uiControlPoints, m_uiDerivatives, m_uiOrder,
			   m_pfKnot, m_pControlPoint, m_ppDerivativeTable);
  CalculateCurve();
  CalculateWarp();
  CalculateFrenetFrames();
  CalculateTorsion();
}

VOID CSLIDEBSplineCurve::Uninit()
{
  CSLIDEAtomNode::Uninit();

  if( m_plistControlPointIDs != NULL )
    {
      delete m_plistControlPointIDs;
      m_plistControlPointIDs = NULL;
    }

  if( m_plistKnot != NULL )
    {
      delete m_plistKnot;
      m_plistKnot = NULL;
    }

  DeallocateKnots();
  DeallocateControlPoints();
  DeallocateControlWarps();

  DeallocateWarps();
  DeallocatePoints();
  DeallocateNormals();
  DeallocateBinormals();
  DeallocateTangents();
  DeallocateNormalsTorsionReduced();
  DeallocateBinormalsTorsionReduced();
  DeallocateTangentsTorsionReduced();
  DeallocateSweepFrenetFrames();
  
  DeallocateDerivativeTable();
  DeallocateDeBoorTable();
}

BOOL CSLIDEBSplineCurve::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEBSplineCurve::InitKnots()
{
  UINT_32 i;

  if( m_uiCurveType == 0 ) // BSpline
    {
      for(i = 0; i < m_uiKnots; i++)
	{
	  m_pfKnot[i] = i;
	}
    }
  else // Bezier
    {
      for(i = 0; i < m_uiKnots; i++)
	{
	  m_pfKnot[i] = !(i < m_uiKnots/2); 
	}
    }
}

VOID CSLIDEBSplineCurve::InitControlPoints()
{
  // Default Control Points
  m_pControlPoint[0].Set(-1.5,  0,   0);
  m_pControlPoint[1].Set(-0.5,  1.0, 0);
  m_pControlPoint[2].Set( 0.5,  1.0, 0);
  m_pControlPoint[3].Set( 1.5,  0,   0);
}

VOID CSLIDEBSplineCurve::InitControlWarps()
{
  UINT_32 i;

  for( i = 0; i < m_uiControlPoints; i++ )
    {
      m_pControlWarpScale[i].Set(1.0, 1.0, 1.0);
      m_pControlWarpRotate[i].Set(0.0, 0.0, 0.0);
    }
}

VOID CSLIDEBSplineCurve::DeallocateKnots()
{
  if ( m_pfKnot != NULL )
    {
      delete[] m_pfKnot;
      m_pfKnot = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateKnots()
{
  if ( m_pfKnot != NULL )
    {
      delete[] m_pfKnot;
      m_pfKnot = NULL;
    }

  m_pfKnot = new FLOAT[m_uiKnots];
  ASSERT( m_pfKnot != NULL );
}

VOID CSLIDEBSplineCurve::DeallocateControlPoints()
{
  if ( m_pControlPoint != NULL )
    {
      delete[] m_pControlPoint;
      m_pControlPoint = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateControlPoints()
{
  if( m_pControlPoint != NULL )
    {
      DeallocateControlPoints();
    }

  m_pControlPoint = new Point[m_uiControlPoints];
  ASSERT(m_pControlPoint != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateControlWarps()
{
  if ( m_pControlWarpScale != NULL )
    {
      delete[] m_pControlWarpScale;
      m_pControlWarpScale = NULL;
    }

  if ( m_pControlWarpRotate != NULL )
    {
      delete[] m_pControlWarpRotate;
      m_pControlWarpRotate = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateControlWarps()
{
  if( (m_pControlWarpScale != NULL) || (m_pControlWarpRotate != NULL) )
    {
      DeallocateControlWarps();
    }

  m_pControlWarpScale = new Point[m_uiControlPoints];
  ASSERT(m_pControlWarpScale != NULL);

  m_pControlWarpRotate = new Point[m_uiControlPoints];
  ASSERT(m_pControlWarpRotate != NULL);

  InitControlWarps();
}

VOID CSLIDEBSplineCurve::DeallocateWarps()
{
  if ( m_pWarpScale != NULL )
    {
      delete[] m_pWarpScale;
      m_pWarpScale = NULL;
    }

  if ( m_pWarpRotate != NULL )
    {
      delete[] m_pWarpRotate;
      m_pWarpRotate = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateWarps()
{
  if( (m_pWarpScale != NULL) || (m_pWarpRotate != NULL) )
    {
      DeallocateWarps();
    }

  m_pWarpScale = new Point[m_uiSlices+1];
  ASSERT(m_pWarpScale != NULL);

  m_pWarpRotate = new Point[m_uiSlices+1];
  ASSERT(m_pWarpRotate != NULL);
}

VOID CSLIDEBSplineCurve::DeallocatePoints()
{
  if ( m_pPoint != NULL )
    {
      delete[] m_pPoint;
      m_pPoint = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocatePoints()
{
  if( m_pPoint != NULL )
    {
      DeallocatePoints();
    }

  m_pPoint = new Point[m_uiSlices+1];
  ASSERT(m_pPoint != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateNormals()
{
  if ( m_pvNormal != NULL )
    {
      delete[] m_pvNormal;
      m_pvNormal = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateNormals()
{
  if( m_pvNormal != NULL )
    {
      DeallocateNormals();
    }

  m_pvNormal = new Vector[m_uiSlices+1];
  ASSERT(m_pvNormal != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateBinormals()
{
  if ( m_pvBinormal != NULL )
    {
      delete[] m_pvBinormal;
      m_pvBinormal = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateBinormals()
{
  if( m_pvBinormal != NULL )
    {
      DeallocateBinormals();
    }

  m_pvBinormal = new Vector[m_uiSlices+1];
  ASSERT(m_pvBinormal != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateTangents()
{
  if ( m_pvTangent != NULL )
    {
      delete[] m_pvTangent;
      m_pvTangent = NULL;
    }
}


VOID CSLIDEBSplineCurve::ReallocateTangents()
{
  if( m_pvTangent != NULL )
    {
      DeallocateTangents();
    }

  m_pvTangent = new Vector[m_uiSlices+1];
  ASSERT(m_pvTangent != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateNormalsTorsionReduced()
{
  if ( m_pvNormalTorsionReduced != NULL )
    {
      delete[] m_pvNormalTorsionReduced;
      m_pvNormalTorsionReduced = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateNormalsTorsionReduced()
{
  if( m_pvNormalTorsionReduced != NULL )
    {
      DeallocateNormalsTorsionReduced();
    }

  m_pvNormalTorsionReduced = new Vector[m_uiSlices+1];
  ASSERT(m_pvNormalTorsionReduced != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateBinormalsTorsionReduced()
{
  if ( m_pvBinormalTorsionReduced != NULL )
    {
      delete[] m_pvBinormalTorsionReduced;
      m_pvBinormalTorsionReduced = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateBinormalsTorsionReduced()
{
  if( m_pvBinormalTorsionReduced != NULL )
    {
      DeallocateBinormalsTorsionReduced();
    }

  m_pvBinormalTorsionReduced = new Vector[m_uiSlices+1];
  ASSERT(m_pvBinormalTorsionReduced != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateTangentsTorsionReduced()
{
  if ( m_pvTangentTorsionReduced != NULL )
    {
      delete[] m_pvTangentTorsionReduced;
      m_pvTangentTorsionReduced = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateTangentsTorsionReduced()
{
  if( m_pvTangentTorsionReduced != NULL )
    {
      DeallocateTangentsTorsionReduced();
    }

  m_pvTangentTorsionReduced = new Vector[m_uiSlices+1];
  ASSERT(m_pvTangentTorsionReduced != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateSweepFrenetFrames()
{
  if ( m_pmSweepFrenetFrame != NULL )
    {
      delete[] m_pmSweepFrenetFrame;
      m_pmSweepFrenetFrame = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateSweepFrenetFrames()
{
  if( m_pmSweepFrenetFrame != NULL )
    {
      DeallocateSweepFrenetFrames();
    }

  m_pmSweepFrenetFrame = new Matrix[m_uiSlices+1];
  ASSERT(m_pmSweepFrenetFrame != NULL);
}

VOID CSLIDEBSplineCurve::DeallocateDerivativeTable()
{
 UINT_32 i;

  if ( m_ppDerivativeTable != NULL )
    {
      for ( i = 0; i < m_uiDerivativeControlPoints; i++ )
	{
	  delete[] m_ppDerivativeTable[i];
	}
      delete[] m_ppDerivativeTable;
      m_ppDerivativeTable = NULL;
      m_uiDerivativeControlPoints = 0;
    }
}

VOID CSLIDEBSplineCurve::ReallocateDerivativeTable()
{
  UINT_32 i;

  if( m_ppDerivativeTable != NULL )
    {
      DeallocateDerivativeTable();
    }

  ASSERT( m_uiDerivativeControlPoints == 0 );
  ASSERT( m_ppDerivativeTable == NULL );

  m_uiDerivativeControlPoints = m_uiControlPoints;
  m_ppDerivativeTable = new Point*[m_uiDerivativeControlPoints];
  ASSERT( m_ppDerivativeTable != NULL );

  for ( i = 0; i < m_uiDerivativeControlPoints; i++ )
    {
      m_ppDerivativeTable[i] = new Point[m_uiDerivatives]; 
      ASSERT( m_ppDerivativeTable[i] != NULL );
    }
}

VOID CSLIDEBSplineCurve::DeallocateDeBoorTable()
{
  UINT_32 i;

  if ( m_ppfDeBoorTable != NULL )
    {
      for ( i = 0; i < (2*m_uiOrderPrev - 1); i++ )
	{
	  delete[] m_ppfDeBoorTable[i];
	}
      delete[] m_ppfDeBoorTable;
      m_ppfDeBoorTable = NULL;
    }
}

VOID CSLIDEBSplineCurve::ReallocateDeBoorTable()
{
  UINT_32 i;

  if( m_ppfDeBoorTable != NULL )
    {
      DeallocateDeBoorTable();
    }

  m_uiOrderPrev = m_uiOrder;

  m_ppfDeBoorTable = new FLOAT*[2*m_uiOrder - 1];
  ASSERT( m_ppfDeBoorTable != NULL );

  for ( i = 0; i < (2*m_uiOrder - 1); i++ )
    {
      m_ppfDeBoorTable[i] = new FLOAT[m_uiOrder];
      ASSERT( m_ppfDeBoorTable[i] != NULL );
    }
}

BOOL CSLIDEBSplineCurve::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  UINT_32 i;
  DynFloat *pdynfKnot;
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint, *pCSLIDEControlPointHolder;

  bCorrect = TRUE;

  if ( !( m_uiOrder >= 3 ) )
    {
      if( m_uiCurveType == 0 ) // BSpline
	HandleError("order(%ld) must be >= 3", m_uiOrder );
      else
	HandleError("number of control points(%ld) must be >= 3", m_uiOrder );
      bCorrect = FALSE;
    }  

  // Check Control Point References
  if( m_plistControlPointIDs != NULL )
    {
      m_uiControlPoints = m_plistControlPointIDs->GetLength();

      if( m_plistControlPointIDs->GetLength() != cuiKnots-cuiOrder )
	{
	  ReallocateControlPoints();
	  ReallocateControlWarps();
	}
      
      iter.Init(m_plistControlPointIDs);	  
      for ( pcID = (CHAR*)(iter.PeekFirst());
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()) )
	{
	  pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
	  pCSLIDEControlPoint = pCSLIDEWorld->PeekControlPoint(pcID);
	  
	  if( pCSLIDEPoint != NULL )
	    {
	      pCSLIDEControlPointHolder = new CSLIDEControlPoint();
	      pCSLIDEControlPointHolder->SetPointID(pcID);
	      
	      m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
	      m_plistCSLIDEControlPoint->InsertLast(pCSLIDEControlPointHolder, 0);
	      
	      m_listControlWarpScale.InsertLast(pCSLIDEControlPointHolder->GetScale(), 0);
	      m_listControlWarpRotate.InsertLast(pCSLIDEControlPointHolder->GetRotate(), 0);
	    }
	  else if( pCSLIDEControlPoint != NULL )
	    {
	      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pCSLIDEControlPoint->GetPointID());
	      
	      if ( pCSLIDEPoint == NULL )
		{
		  HandleError("Undefined %s \"%s\" in %s \"%s\"", 
			      gcppcSLIDEReservedWords[sliderwtPoint], 
			      pCSLIDEControlPoint->GetPointID(),
			      gcppcSLIDEReservedWords[sliderwtControlPoint], 
			      pcID);
		}
	      m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
	      m_plistCSLIDEControlPoint->InsertLast(pCSLIDEControlPoint, 0);
	      
	      m_listControlWarpScale.InsertLast( pCSLIDEControlPoint->GetScale(), 0 );
	      m_listControlWarpRotate.InsertLast( pCSLIDEControlPoint->GetRotate(), 0 );
	    }
	  else
	    {
	      HandleError("Undefined %s \"%s\"", 
			  gcppcSLIDEReservedWords[sliderwtPoint], pcID);
	      HandleError("Undefined %s \"%s\"", 
			  gcppcSLIDEReservedWords[sliderwtControlPoint], pcID);
	      bCorrect = FALSE;
	    }
	}
    }

  if( m_uiCurveType == 0 ) // BSpline
    {
      if( m_uiOrder > m_uiControlPoints )
	{
	  HandleError("order (%ld) must be <= #controlpoints (%ld)", 
		      m_uiOrder, m_uiControlPoints);
	  bCorrect = FALSE;
	}
  
      m_uiKnots = m_uiOrder + m_uiControlPoints;

      if( m_plistKnot != NULL ) 
	{
	  if( m_plistKnot->GetLength() != m_uiKnots )
	    {
	      HandleError("Wrong number of knots specified in knotlist: expecting %ld",
			  m_uiKnots);
	      bCorrect = FALSE;
	    }
	  else
	    {
	      ReallocateKnots();
	      
	      iter.Init(m_plistKnot);
	      for ( pdynfKnot = (DynFloat*)(iter.PeekFirst()), i = 0;
		    pdynfKnot != NULL;
		    pdynfKnot = (DynFloat*)(iter.PeekNext()), i++ )
		{
		  m_pfKnot[i] = pdynfKnot->GetValue();
		}
	    }
	}
      else
	{
	  ReallocateKnots();
	  InitKnots();
	}
    }
  else // Bezier
    {
      m_uiOrder = m_plistControlPointIDs->GetLength();

      m_uiKnots = 2*m_uiOrder;
      ReallocateKnots();
      InitKnots();

      m_uiControlPoints = m_uiOrder;
      m_bKnotsChanged = TRUE;
    }

  if( (1 > m_uiSlices) || (m_uiSlices > 1000)  )
    {
      HandleError("slices must be between 1 and 1000" );
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDEBSplineCurve::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dyniSlices.GetValue() >= 0 ) );
}

VOID CSLIDEBSplineCurve::CheckSweepMorphCrossSectionDynamic()
{
  CListIterInt iter;
  CSLIDEControlPoint *pCSLIDEControlPoint;

  iter.Init(m_plistCSLIDEControlPoint);
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
    {
      if( pCSLIDEControlPoint->GetCrossSectionDynamic() )
	{
	  m_bSweepMorphCrossSectionDynamic = TRUE;
	  break;
	}
    }
}

VOID CSLIDEBSplineCurve::CheckDynamic()
{
  CSLIDESweepPath::CheckDynamic();

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;

  if( !GetDynamic() )
    {
      SetDynamic(m_dyniDrawControls.GetDynamic() ||
		 m_dyniSlices.GetDynamic() ||
		 m_dyniCurveClosed.GetDynamic() ||
		 m_dyniMinimizeTorsion.GetDynamic() ||
		 m_dyniClosed.GetDynamic() ||
		 m_dynfAzimuth.GetDynamic() ||
		 m_dynfTwist.GetDynamic() ||
		 m_dyniSymmetry.GetDynamic() );
    }

  // Check if control point - points are dynamic
  if( !GetDynamic() && (m_plistControlPointIDs != NULL) )
    {
      iter.Init(&m_listCSLIDEPoint);
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  if( pCSLIDEPoint->GetBoundDynamic() )
	    {
	      SetDynamic(TRUE);
	      break;
	    }
	}     
    }

  // Check if control point - control points are dynamic (scale, warp)
  if( !GetDynamic() )
    {
      iter.Init(m_plistCSLIDEControlPoint);
      for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
	    pCSLIDEControlPoint != NULL;
	    pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
	{
	  if( pCSLIDEControlPoint->GetBoundDynamic() )
	    {
	      SetDynamic(TRUE);
	      break;
	    }
	}
    }
}

VOID CSLIDEBSplineCurve::PreprocessGraphLocal()
{
  DynVector *pdynv;
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  BOOL bBoundEmpty;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  bBoundEmpty = FALSE;
  bBoundDynamic = FALSE;
  bGeometryDynamic = FALSE;
  bNormalsDynamic = FALSE;
  
  // My points (path points and controlpoints)
  m_bControlPointsDynamic = FALSE;
  if( m_plistControlPointIDs != NULL )
    {
      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;
      bGeometryDynamic = FALSE;
      bNormalsDynamic = FALSE;
      
      iter.Init(&m_listCSLIDEPoint);
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  bBoundEmpty = FALSE;
	  bBoundDynamic = bBoundDynamic || pCSLIDEPoint->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || pCSLIDEPoint->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic || pCSLIDEPoint->GetNormalDynamic();

	  if(  pCSLIDEPoint->GetGeometryDynamic() )
	    {
	      m_bControlPointsDynamic = TRUE;
	    }
	}

      // sweepmorph crosssection
      if( m_listCSLIDESweepMorphParent.GetLength() > 0 )
	{
	  iter.Init(m_plistCSLIDEControlPoint);
	  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
		pCSLIDEControlPoint != NULL;
		pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
	    {
	      pCSLIDEControlPoint->PreprocessGraphLocalCrossSection();

	      bBoundEmpty = FALSE;
	      bBoundDynamic =bBoundDynamic || pCSLIDEControlPoint->GetCrossSectionDynamic();
	      bNormalsDynamic = bNormalsDynamic || pCSLIDEControlPoint->GetCrossSectionDynamic();
	      bGeometryDynamic = bGeometryDynamic || pCSLIDEControlPoint->GetCrossSectionDynamic();
	    }
	}
    }

  // Control Points scale and rotate
  m_bControlWarpDynamic = FALSE;
  iter.Init(&m_listControlWarpScale);
  for ( pdynv = (DynVector*)(iter.PeekFirst());
	pdynv != NULL;
	pdynv = (DynVector*)(iter.PeekNext()) )
    {
      bBoundEmpty = FALSE;
      bBoundDynamic = bBoundDynamic || pdynv->GetDynamic();
      bGeometryDynamic = bGeometryDynamic || pdynv->GetDynamic();
      bNormalsDynamic = bNormalsDynamic || pdynv->GetDynamic();
      
      if( pdynv->GetDynamic() )
	{
	  m_bControlWarpDynamic = TRUE;
	}
    }

  iter.Init(&m_listControlWarpRotate);
  for ( pdynv = (DynVector*)(iter.PeekFirst());
	pdynv != NULL;
	pdynv = (DynVector*)(iter.PeekNext()) )
    {    
      bBoundEmpty = FALSE;
      bBoundDynamic = bBoundDynamic || pdynv->GetDynamic();
      bGeometryDynamic = bGeometryDynamic || pdynv->GetDynamic();
      bNormalsDynamic = bNormalsDynamic || pdynv->GetDynamic();

      if( pdynv->GetDynamic() )
	{
	  m_bControlWarpDynamic = TRUE;
	}
    }

  bRealSpaceDynamic = ( m_dyniDrawControls.GetDynamic() );
  bParameterSpaceDynamic = ( m_dyniSlices.GetDynamic() ||
			     m_dyniCurveClosed.GetDynamic() ||
			     m_dyniMinimizeTorsion.GetDynamic() ||
			     m_dyniClosed.GetDynamic() ||
			     m_dynfAzimuth.GetDynamic() ||
			     m_dynfTwist.GetDynamic() ||
			     m_dyniSymmetry.GetDynamic() );


  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bGeometryDynamic ||
		     ((m_pCSLIDESurface != NULL) && m_pCSLIDESurface->GetDynamic()) );
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || 
		    bParameterSpaceDynamic || 
		    bNormalsDynamic );
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic ||
		   bBoundDynamic);
  SetBoundEmpty(bBoundEmpty);
  SetBoundDynamic(bRealSpaceDynamic || 
		  bParameterSpaceDynamic ||
		  bBoundDynamic);

  ////////////////////////////////////////////////////////////////////////////////////
  CSLIDESweep *pCSLIDESweep;
  CSLIDESweepMorph *pCSLIDESweepMorph;
  CSLIDECrossSection *pCSLIDECrossSection;

  iter.Init(&m_listCSLIDECrossSectionParent);  
  for ( pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekFirst(); 
	pCSLIDECrossSection != NULL;
	pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDECrossSection->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDECrossSection->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDECrossSection->SetNormalsDynamic(TRUE);
    }

  iter.Init(&m_listCSLIDESweepParent);  
  for ( pCSLIDESweep = (CSLIDESweep*)iter.PeekFirst(); 
	pCSLIDESweep != NULL;
	pCSLIDESweep = (CSLIDESweep*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDESweep->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDESweep->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDESweep->SetNormalsDynamic(TRUE);
    }

  iter.Init(&m_listCSLIDESweepMorphParent);  
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst(); 
	pCSLIDESweepMorph != NULL;
	pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext() )
    {
      if( GetBoundDynamic() )
	pCSLIDESweepMorph->SetBoundDynamic(TRUE);

      if( GetGeometryDynamic() )
	pCSLIDESweepMorph->SetGeometryDynamic(TRUE);

      if( GetNormalsDynamic() )
	pCSLIDESweepMorph->SetNormalsDynamic(TRUE);
    }
}

VOID CSLIDEBSplineCurve::UpdateCrossSectionParents()
{
  CListIterInt iter;
  CSLIDECrossSection *pCSLIDECrossSection;

  iter.Init(&m_listCSLIDECrossSectionParent);
  for ( pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekFirst(); 
	pCSLIDECrossSection != NULL;
	pCSLIDECrossSection = (CSLIDECrossSection*)iter.PeekNext() )
    {
      pCSLIDECrossSection->UpdateCrossSectionPoints(m_uiSlices+1, m_pPoint);
    }
}

VOID CSLIDEBSplineCurve::UpdateSweepPathParents()
{
  CListIterInt iter;
  CSLIDESweep *pCSLIDESweep;
  CSLIDESweepMorph *pCSLIDESweepMorph;
  
  iter.Init(&m_listCSLIDESweepParent);
  for ( pCSLIDESweep = (CSLIDESweep*)iter.PeekFirst(); 
	pCSLIDESweep != NULL;
	pCSLIDESweep = (CSLIDESweep*)iter.PeekNext() )
    {
      pCSLIDESweep->UpdatePathPoints();
    }

  iter.Init(&m_listCSLIDESweepMorphParent);
  for ( pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekFirst(); 
	pCSLIDESweepMorph != NULL;
	pCSLIDESweepMorph = (CSLIDESweepMorph*)iter.PeekNext() )
    {
      pCSLIDESweepMorph->UpdatePathPoints();
    }
}

VOID CSLIDEBSplineCurve::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  UINT_32 i;
  FLOAT f;
  Point pt;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bControlPointsChanged, bTorsionChanged;
  BOOL bCrossSectionChanged, bCrossSectionSlicesChanged, bCurveClosedChanged;
  BOOL bWarpChanged, bSlicesChanged, bOrderChanged, bNumControlPointsChanged;

  bCrossSectionChanged = FALSE;
  bCrossSectionSlicesChanged = FALSE;
  bSlicesChanged = FALSE;
  bControlPointsChanged = FALSE;
  bWarpChanged = FALSE;
  bTorsionChanged = FALSE;
  bOrderChanged = FALSE;
  bNumControlPointsChanged = FALSE;
  bCurveClosedChanged = FALSE;

  if( (m_plistControlPointIDs != NULL) && (m_bControlPointsDynamic == TRUE) )
    {
      iter.Init(&m_listCSLIDEPoint);
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(), i = 0; 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(), i++ )
	{
	  if( m_pControlPoint[i] != pCSLIDEPoint->GetPoint() )
	    {
	      m_pControlPoint[i] = pCSLIDEPoint->GetPoint();
	      bControlPointsChanged = TRUE;
	    }
	}
    }

  // So normal lengths are right for sweeps using this curve
  if( m_fNormalLength == cfNormalLength )
    {
      CalculateBound();
    }

  if( m_bControlWarpDynamic )
    {
      FLOAT fx, fy, fz, fSx, fSy, fSz, fRx, fRy, fRz;
      DynVector *pdynv;
      
      iter.Init(&m_listControlWarpScale);
      for ( pdynv = (DynVector*)(iter.PeekFirst()), i = 0;
	    pdynv != NULL;
	    pdynv = (DynVector*)(iter.PeekNext()), i++ )
	{  
	  fSx = m_pControlWarpScale[i][0];
	  fSy = m_pControlWarpScale[i][1];
	  fSz = m_pControlWarpScale[i][2];

	  pdynv->Update();

	  fx = (pdynv->Get())[X];
	  fy = (pdynv->Get())[Y];
	  fz = (pdynv->Get())[Z];

	  if( (fSx != fx) || (fSy != fy) || (fSz != fz) )
	    {
	      m_pControlWarpScale[i].Set( fx, fy, fz );
	      bWarpChanged = TRUE;
	    }
	}
      
      iter.Init(&m_listControlWarpRotate);
      for ( pdynv = (DynVector*)(iter.PeekFirst()), i = 0;
	    pdynv != NULL;
	    pdynv = (DynVector*)(iter.PeekNext()), i++ )
	{  
	  fRx = m_pControlWarpRotate[i][0];
	  fRy = m_pControlWarpRotate[i][1];
	  fRz = m_pControlWarpRotate[i][2];

	  pdynv->Update();

	  fx = (pdynv->Get())[X];
	  fy = (pdynv->Get())[Y];
	  fz = (pdynv->Get())[Z];

	  if( (fRx != fx) || (fRy != fy) || (fRz != fz) )
	    {
	      m_pControlWarpRotate[i].Set( DEGTORAD(fx), DEGTORAD(fy), DEGTORAD(fz) );
	      bWarpChanged = TRUE;
	    }
	}
    }
  
  m_dyniDrawControls.Update();

  i = m_dyniMinimizeTorsion.GetValue();
  m_dyniMinimizeTorsion.Update();
  bTorsionChanged = (i != (UINT_32)m_dyniMinimizeTorsion.GetValue());

  f = m_dynfAzimuth.GetValue();
  m_dynfAzimuth.Update();
  bTorsionChanged = (f != (FLOAT)m_dynfAzimuth.GetValue()) || bTorsionChanged;

  f = m_dynfTwist.GetValue();
  m_dynfTwist.Update();
  bTorsionChanged = (f != (FLOAT)m_dynfTwist.GetValue()) || bTorsionChanged;

  i = m_dyniSymmetry.GetValue();
  m_dyniSymmetry.Update();
  bTorsionChanged = (i != (UINT_32)m_dyniSymmetry.GetValue()) || bTorsionChanged;

  m_dyniSlices.Update();
  if ( (1 > m_dyniSlices.GetValue() || (m_dyniSlices.GetValue() > 1000)) )
    {
      HandleError("slices must be between 1 and 1000" );
      m_dyniSlices.SetValue(cuiSlices);
    }
  bSlicesChanged = (m_uiSlices != (UINT_32)m_dyniSlices.GetValue());
  m_uiSlices = m_dyniSlices.GetValue();

  //bCurveClosedChanged = (0 != m_dyniCurveClosed.GetValue());
  bCurveClosedChanged = m_dyniCurveClosed.Update();
  //bCurveClosedChanged = (bCurveClosedChanged != 
  //		 (0 != (UINT_32)m_dyniCurveClosed.GetValue()) );

  if( (m_uiCurveType == 0) && bCurveClosedChanged )
    {
      UINT_32 i;
      CListIterInt iter;
      CSLIDEPoint *pCSLIDEPoint;

      if( 0 == (UINT_32)m_dyniCurveClosed.GetValue()) // open
	{
	  m_uiKnots -= (m_uiOrder - 1);
	  m_uiControlPoints = m_uiKnots - m_uiOrder;

	  ReallocateKnots();
	  ReallocateControlPoints();
	  ReallocateControlWarps();

	  InitKnots();

	  iter.Init(&m_listCSLIDEPoint);
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(), i = 0; 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(), i++ )
	    {
	      m_pControlPoint[i] = pCSLIDEPoint->GetPoint();
	    }
	}
      else // closed
	{
	  m_uiKnots += m_uiOrder - 1;
	  m_uiControlPoints = m_uiKnots - m_uiOrder;

	  ReallocateKnots();
	  ReallocateControlPoints();
	  ReallocateControlWarps();
	  InitKnots();

	  iter.Init(&m_listCSLIDEPoint);
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(), i = 0; 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(), i++ )
	    {
	      m_pControlPoint[i] = pCSLIDEPoint->GetPoint();

	      if( i < m_uiOrder-1 )
		{
		  m_pControlPoint[m_uiControlPoints-(m_uiOrder-1)+i] = 
		    pCSLIDEPoint->GetPoint();
		}
	    }
	}
    }

  // Check to see if the derivatives have to be recalculated
  //   Order not dynamic - happens when user changes default settings
  bOrderChanged = m_uiOrder != m_uiOrderPrev; 
  bNumControlPointsChanged = m_uiControlPoints != m_uiControlPointsPrev;
  if( bOrderChanged || bNumControlPointsChanged || bCurveClosedChanged ) 
    {                   
      m_uiControlPointsPrev = m_uiControlPoints;

      ReallocateDerivativeTable(); 
      ReallocateDeBoorTable();
    }

  if( bOrderChanged || bNumControlPointsChanged || 
      bControlPointsChanged || m_bKnotsChanged || bCurveClosedChanged )
    {
      CalculateDerivativeTable(m_uiControlPoints, m_uiDerivatives, m_uiOrder,
			       m_pfKnot, m_pControlPoint, m_ppDerivativeTable);
    }

  if( bSlicesChanged )
    {      
      ReallocateWarps();
      ReallocatePoints();
      ReallocateNormals();
      ReallocateBinormals();
      ReallocateTangents();
      ReallocateNormalsTorsionReduced();
      ReallocateBinormalsTorsionReduced();
      ReallocateTangentsTorsionReduced();
      ReallocateSweepFrenetFrames();
    }

  if( bSlicesChanged || bOrderChanged || bNumControlPointsChanged || m_bKnotsChanged ||
      bCurveClosedChanged )
    {
      m_bKnotsChanged = FALSE;

      CalculateCurve();
      CalculateFrenetFrames();
      CalculateWarp();
      CalculateTorsion();

      UpdateCrossSectionParents();
      UpdateSweepPathParents();
    }
  else if( bControlPointsChanged )
    {
      CalculateCurve();
      CalculateFrenetFrames();
      CalculateTorsion();

      UpdateCrossSectionParents();
      UpdateSweepPathParents();
    }
  else if( bWarpChanged )
    {
      CalculateWarp();
      CalculateTorsion();

      UpdateSweepPathParents();
    }
  else if( bTorsionChanged )
    {
      CalculateTorsion();

      UpdateSweepPathParents();
    }
}

VOID CSLIDEBSplineCurve::CalculateBound()
{
  UINT_32 uiPoint;
  CBound bound;

  for( uiPoint = 0; uiPoint < m_uiControlPoints; uiPoint++)
    {
      bound.Insert(m_pControlPoint[uiPoint]);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

UINT_32 CSLIDEBSplineCurve::FindKnotInterval(FLOAT t)
{
  UINT_32 uiDelta, uiMax;

  uiMax = m_uiKnots - m_uiOrder;

  // knot[d] <= t < knot[d+1]. 
  ASSERT( m_uiOrder > 1 );
  for ( uiDelta = m_uiOrder-1; uiDelta < uiMax; uiDelta++ )
    {
      if ( t < m_pfKnot[uiDelta+1] ) // <==> (t >= pfKnot[uiDelta])
	{
	  break;
	}
    }

  return uiDelta;
}

UINT_32 CSLIDEBSplineCurve::FindBlossomKnotInterval(FLOAT t)
{
  UINT_32 uiBlossomKnotIndex;

  uiBlossomKnotIndex = FindKnotInterval(t) - (m_uiOrder - 1) + 1;
  ASSERT( (1 <= uiBlossomKnotIndex) && 
	  (uiBlossomKnotIndex <= m_uiControlPoints - (m_uiOrder - 1) + 1) );

  return uiBlossomKnotIndex;
}

VOID CSLIDEBSplineCurve::PointAt(Point &pt, FLOAT t)
{
  UINT_32 knotIndex;

  // Use blossoming to evaluate t on this curve
  knotIndex = FindBlossomKnotInterval(t);
  BlossomAt(pt, m_uiOrder, &m_pControlPoint[knotIndex-1], &m_pfKnot[knotIndex], t);
}

VOID CSLIDEBSplineCurve::CalculateCurve()
{
  UINT_32 uiSlice, knotIndex;
  FLOAT fU, fUDelta;
  Point pt;

  fU = GetTu();
  fUDelta = (GetTuDelta() - EPSILON) / m_uiSlices;

  for( uiSlice = 0; uiSlice <= m_uiSlices; uiSlice++ )
    {
      knotIndex = FindBlossomKnotInterval(fU);
      BlossomAt(pt, m_uiOrder, &m_pControlPoint[knotIndex-1], &m_pfKnot[knotIndex], fU);
      m_pPoint[uiSlice] = pt;

      fU += fUDelta;
    }
}

VOID CSLIDEBSplineCurve::CalculateWarp()
{
  UINT_32 uiSlice, knotIndex;
  FLOAT fU, fUDelta;
  Point pt;
  
  fU = GetTu();
  fUDelta = (GetTuDelta() - EPSILON) / m_uiSlices;

  for( uiSlice = 0; uiSlice <= m_uiSlices; uiSlice++ )
    {
      knotIndex = FindBlossomKnotInterval(fU);  

      BlossomAt(pt, m_uiOrder, &m_pControlWarpScale[knotIndex-1], &m_pfKnot[knotIndex], fU);
      m_pWarpScale[uiSlice] = pt;

      BlossomAt(pt, m_uiOrder, &m_pControlWarpRotate[knotIndex-1], &m_pfKnot[knotIndex],fU);
      m_pWarpRotate[uiSlice] = pt;
      
      fU += fUDelta;
    }
}

VOID CSLIDEBSplineCurve::FrenetAt(Vector &vN, Vector &vB, Vector &vT, FLOAT fU)
{
  UINT_32 uiDelta;
  Vector v, vP0, vP1, vP2, vP3;
  FLOAT t;
  BOOL bP1Zero, bP2Zero, bBZero;
 
  ASSERT( Valid() );

  t = MAX(fU, GetTu());
  t = MIN(fU, GetTu() + GetTuDelta() - EPSILON);      

  uiDelta = FindKnotInterval(t);

  CalculateDeBoorTable(m_uiOrder, m_pfKnot, m_ppfDeBoorTable, t, uiDelta);
  DerivativeAt(vP0, m_uiOrder, m_ppDerivativeTable, m_ppfDeBoorTable, 0, t, uiDelta);
  DerivativeAt(vP1, m_uiOrder, m_ppDerivativeTable, m_ppfDeBoorTable, 1, t, uiDelta);
  DerivativeAt(vP2, m_uiOrder, m_ppDerivativeTable, m_ppfDeBoorTable, 2, t, uiDelta);
  //  DerivativeAt(vP3, m_uiOrder, m_ppDerivativeTable, m_ppfDeBoorTable, 3, t, uiDelta);

  vP1.Normalize();
  vP2.Normalize();

  vT = vP1;     
  vB = vT * vP2;
  vN = vB * vT; 

  vT.Normalize();
  vB.Normalize();
  vN.Normalize();

  bP1Zero = (vP1.Magnitude() <= EPSILON);
  bP2Zero = (vP2.Magnitude() <= EPSILON);
  bBZero  = (vB.Magnitude()  <= EPSILON);


  if( bP1Zero )
    {
      vT.Set(0, 0, 1);	      
      vN.Set(1, 0, 0);	      
      vB.Set(0, 1, 0);	      

      printf( "Warning: First derivative is zero length.\n");
      printf( "         Setting tanget=(0 0 1), normal=(1 0 0), binormal=(0 1 0)\n" );
    }
  else if( bP2Zero || bBZero )
    {
      v.Set(0, 1, 0);
      vB = v - (v.Dot(vT)) * vT;
      vB.Normalize();

      // Check if v1 = (0 1 0) then pick (1 0 0)
      if( vB.Magnitude() <= EPSILON )
	{
	  v.Set(1, 0, 0);
	  vB = v - (v.Dot(vT)) * vT;
	}

      vN = vB * vT; 

      vT.Normalize();
      vB.Normalize();
      vN.Normalize();

      printf( "Warning: Second derivative is zero length.\n");
      printf( "         Setting binormal=(%.3f %.3f %.3f), normal=(%.3f %.3f %.3f).\n", 
	      vB[X], vB[Y], vB[Z], vN[X], vN[Y], vN[Z]);
    }

  ASSERT( vT.Magnitude() > EPSILON );
  ASSERT( vB.Magnitude() > EPSILON );
  ASSERT( vN.Magnitude() > EPSILON );

  /*
  // Calculate Torsion and Curvature
  fTorsion = (vP1.Cross(vP2)).Magnitude();
  fTorsion = fTorsion * fTorsion;
  if( fTorsion != 0.0 ) 
    {
      fTorsion = (((vP1.Cross(vP2)).Dot(vP3)))/fTorsion;
    }

  fCurvature = vP1.Magnitude();
  fCurvature = fCurvature * fCurvature * fCurvature;
  if( fCurvature != 0.0 )
    {		
      fCurvature = (vP1.Cross(vP2)).Magnitude() / fCurvature;
    }
    */
}

VOID CSLIDEBSplineCurve::CalculateFrenetFrames()
{
  UINT_32 uiSlice;
  FLOAT fU, fUDelta;
  
  fUDelta = (GetTuDelta() - EPSILON) / m_uiSlices;
  fU = GetTu();
      
  for( uiSlice = 0; uiSlice <= m_uiSlices; uiSlice++ )
    {
      FrenetAt(m_pvNormal[uiSlice], m_pvBinormal[uiSlice], m_pvTangent[uiSlice], fU);
      fU += fUDelta;
    }
}

VOID CSLIDEBSplineCurve::CalculateTorsion()
{
  UINT_32 uiSlice;
  DOUBLE dTheta, dSymmetryAngle, dPhi, t;
  Point  p1, p2, c1, c2;
  Vector v, v1, v2, vT1, vT2, vB1, vB2, vN1, vN2, vPlane;
  Matrix m, m1, m2, m3, m4, m5;
  MATRIX4D mm;

  for( uiSlice = 0; uiSlice < m_uiSlices + 1; uiSlice++ )
    {
      m_pvNormalTorsionReduced[uiSlice] = m_pvNormal[uiSlice];
      m_pvBinormalTorsionReduced[uiSlice] = m_pvBinormal[uiSlice];
      m_pvTangentTorsionReduced[uiSlice] = m_pvTangent[uiSlice];
    }

  // Minimize Torsion
  if( m_dyniMinimizeTorsion.GetValue() != 0 )
    {
      // better mintorsion scheme
      dPhi = 0.0;
      
      for( uiSlice = 0; uiSlice < m_uiSlices; uiSlice++ )
	{
	  p1 = m_pPoint[uiSlice];
	  p2 = m_pPoint[uiSlice+1];

	  vT2 = m_pvTangentTorsionReduced[uiSlice+1];
	  vN1 = m_pvNormalTorsionReduced[uiSlice];
	  vN2 = m_pvNormalTorsionReduced[uiSlice+1];

	  v  = p2 - p1;
	  vN1.Normalize();
	  c1 = p1 + vN1;

	  vPlane = vT2;
	  vPlane.CompletePlaneEquation(p2);

	  t = -((vPlane[X]*c1[X] + vPlane[Y]*c1[Y] + vPlane[Z]*c1[Z] + vPlane[W]) /
		(vPlane[X]*v[X] + vPlane[Y]*v[Y] + vPlane[Z]*v[Z]));

	  c2 = c1 + t*v;
	  v2 = c2 - p2;
	  
	  dTheta = AngleBetween( vN2, v2 );
	  v = vN2 * v2;
	  if( v.Dot(vT2) < 0.0 )
	    {
	      dTheta *= -1.0;
	    }
	  m.SetRotate( vT2, dTheta );

	  m_pvNormalTorsionReduced[uiSlice+1] = m * m_pvNormalTorsionReduced[uiSlice+1];
	  m_pvBinormalTorsionReduced[uiSlice+1] = m * m_pvBinormalTorsionReduced[uiSlice+1];

	  dPhi += dTheta;
	}

      /*
	// Original min torsion scheme
      dPhi = 0.0;

      for( uiSlice = 0; uiSlice < m_uiSlices; uiSlice++ )
	{
	  vT1 = m_pvTangent[uiSlice];
	  vT2 = m_pvTangent[uiSlice+1];
	  vB1 = m_pvBinormalTorsionReduced[uiSlice];
	  vB2 = m_pvBinormal[uiSlice+1];
	  
	  // Align Tangents  1->2
	  v = vT1*vT2;
	  dTheta = AngleBetween( vT1, vT2 );
	  m.SetRotate( v, dTheta );
	  vB1 = m * vB1;
	  
	  // Find angle between binormals 2->1
	  v = vB2 * vB1;
	  dTheta = AngleBetween( vB1, vB2 );
	  if( v.Dot(vT2) < 0.0 )
	    {
	      dTheta *= -1.0;
	    }
	  m.SetRotate( vT2, dTheta );
	  
	  m_pvNormalTorsionReduced[uiSlice+1] = m * m_pvNormal[uiSlice+1];
	  m_pvBinormalTorsionReduced[uiSlice+1] = m * m_pvBinormal[uiSlice+1];

	  dPhi += dTheta;
	}
      printf( "curve \"%s\" has an intrisic twist of %f degrees\n", m_pcGUIID, RADTODEG(dPhi));
      */
    }

  // Closed Curve - Calculate overall torsion
  if( m_dyniClosed.GetValue() != 0 )
    {
      vT1 = m_pvTangentTorsionReduced[0];
      vT2 = m_pvTangentTorsionReduced[m_uiSlices];
      vB1 = m_pvBinormalTorsionReduced[0];
      vB2 = m_pvBinormalTorsionReduced[m_uiSlices];

      // Align Tangents 1->2
      v = vT1*vT2;
      dTheta = AngleBetween( vT1, vT2 );
      m.SetRotate( v, dTheta );
      vB1 = m * vB1;

      // Find angle between binormals 2->1
      v = vB2 * vB1;
      dTheta = AngleBetween( vB1, vB2 );

      // Use Symmetry to close with minimal twisting
      DOUBLE ds = m_dyniSymmetry.GetValue();
      if( ds < 1.0 )  // After CheckSemantics sym => 0 ???
	 {
	   ds = 1.0;
	 }

      dSymmetryAngle = 2.0 * PI / ds;
      dTheta = fmod(dTheta, dSymmetryAngle);

      if( dTheta != 0.0 )
	{
	  if( dTheta > (dSymmetryAngle/2.0) )
	    {
	      dTheta = dTheta - dSymmetryAngle;
	    }
	  
	  if( v.Dot(vT2) < 0.0 ) // make sure orientation is correct
	    {
	      dTheta *= -1.0;
	    }
	  
	  //printf( "curve \"%s\" closed by a total rotation of %f degrees\n", m_pcGUIID, RADTODEG(dTheta));
	  dTheta = dTheta / (DOUBLE)m_uiSlices;
	  
	  for( uiSlice = 1; uiSlice < m_uiSlices + 1; uiSlice++ )
	    {
	      m.SetIdentity();
	      m.SetRotate( m_pvTangentTorsionReduced[uiSlice], (DOUBLE)uiSlice*dTheta);
	      
	      m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
	      m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
	    }
	}
    }

  // Azimuth
  if( m_dynfAzimuth.GetValue() != 0.0 )
    {
      dTheta = DEGTORAD(m_dynfAzimuth.GetValue());

      for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++ )
	{
	  m.SetRotate( m_pvTangentTorsionReduced[uiSlice], dTheta );
	  
	  m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
	  m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
	}
    }

  // Twist
  if( m_dynfTwist.GetValue() != 0.0 )
    {
      dTheta = DEGTORAD(m_dynfTwist.GetValue());
      dTheta = dTheta / (DOUBLE)m_uiSlices;

      for( uiSlice = 1; uiSlice < m_uiSlices + 1; uiSlice++ )
	{
	  m.SetIdentity();
	  m.SetRotate( m_pvTangentTorsionReduced[uiSlice], (DOUBLE)uiSlice*dTheta);

	  m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
	  m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
	}
    }

  // User specified rotation
  for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++ )
    {
      m.SetRotate( m_pvNormalTorsionReduced[uiSlice], m_pWarpRotate[uiSlice][X] );
      m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
      m_pvTangentTorsionReduced[uiSlice] = m * m_pvTangentTorsionReduced[uiSlice];

      m.SetRotate( m_pvBinormalTorsionReduced[uiSlice], m_pWarpRotate[uiSlice][Y] );
      m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
      m_pvTangentTorsionReduced[uiSlice] = m * m_pvTangentTorsionReduced[uiSlice];

      m.SetRotate( m_pvTangentTorsionReduced[uiSlice], m_pWarpRotate[uiSlice][Z] );      
      m_pvNormalTorsionReduced[uiSlice] = m * m_pvNormalTorsionReduced[uiSlice];
      m_pvBinormalTorsionReduced[uiSlice] = m * m_pvBinormalTorsionReduced[uiSlice];
    }


  // Set Sweep Frenet Frames
  for( uiSlice = 0; uiSlice < m_uiSlices+1; uiSlice++)
    {
      // Set Change of Basis Matrix
      mm[0]  = m_pvNormalTorsionReduced[uiSlice][X];
      mm[1]  = m_pvNormalTorsionReduced[uiSlice][Y];
      mm[2]  = m_pvNormalTorsionReduced[uiSlice][Z];
      mm[3]  = m_pvNormalTorsionReduced[uiSlice][W];
      mm[4]  = m_pvBinormalTorsionReduced[uiSlice][X];
      mm[5]  = m_pvBinormalTorsionReduced[uiSlice][Y];
      mm[6]  = m_pvBinormalTorsionReduced[uiSlice][Z];
      mm[7]  = m_pvBinormalTorsionReduced[uiSlice][W];
      mm[8]  = m_pvTangentTorsionReduced[uiSlice][X];
      mm[9]  = m_pvTangentTorsionReduced[uiSlice][Y];
      mm[10] = m_pvTangentTorsionReduced[uiSlice][Z];
      mm[11] = m_pvTangentTorsionReduced[uiSlice][W];
      mm[12] = 0;
      mm[13] = 0;
      mm[14] = 0;
      mm[15] = 1;
      m1.Set(mm);

      // Scale
      v.Set(m_pWarpScale[uiSlice][X], m_pWarpScale[uiSlice][Y], m_pWarpScale[uiSlice][Z]);
      m2.SetScale(v);
      
      m_pmSweepFrenetFrame[uiSlice] = m1 * m2;
    }
}


////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDEBSplineCurve::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 i;
  
  PushAuxillaryGeometryAttributes();
  {
    glPushAttrib(GL_LINE_BIT | GL_COLOR_BUFFER_BIT);
    {
      gilLineWidth(2.0);

      //printf( "\nBSpline Points\n" );
      gilBegin(GL_LINE_STRIP);
      {
	for( i = 0; i < m_uiSlices + 1; i++ )
	  {
	    gilVertex3fv(m_pPoint[i].Get());
	    //printf( "(%.3f %.3f %.3f)\n", m_pPoint[i][X], m_pPoint[i][Y], m_pPoint[i][Z]);
	  }
      }
      gilEnd();

      if( m_dyniDrawControls.GetValue() != 0 )
	{
	  RenderControls();
	}
    }
    glPopAttrib();
  }
  PopAuxillaryGeometryAttributes();
}


VOID CSLIDEBSplineCurve::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  RenderGeometrySmooth();
}

VOID CSLIDEBSplineCurve::RenderNormalsSmooth()
{
  UINT_32 i;
  Point pt;
  
  // Render Frenet Frame
  // Tangent
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(0, 0, 1);  // Blue
    
    gilBegin(GL_LINES);
    {
      for(i = 0; i < m_uiSlices + 1; i++ ) 
	{
	  gilVertex3fv(m_pPoint[i].Get());
	  pt = m_pPoint[i] + m_pvTangentTorsionReduced[i] * m_fNormalLength;
	  gilVertex3fv(pt.Get());
	}
    }
    gilEnd();
    
    // Normal
    gilColor3f(1, 0, 0);  // Red
    
    gilBegin(GL_LINES);
    {
      for(i = 0; i < m_uiSlices + 1; i++ ) 
	{
	  gilVertex3fv(m_pPoint[i].Get());
	  pt = m_pPoint[i] + m_pvNormalTorsionReduced[i] * m_fNormalLength;
	  gilVertex3fv(pt.Get());
	}
    }
    gilEnd();
    
    // Binormal
    gilColor3f(0, 1, 0);  // Green
    
    gilBegin(GL_LINES);
    {
      for(i = 0; i < m_uiSlices + 1; i++ ) 
	{
	  gilVertex3fv(m_pPoint[i].Get());
	  pt = m_pPoint[i] + m_pvBinormalTorsionReduced[i] * m_fNormalLength;
	  gilVertex3fv(pt.Get());
	}
    }
    gilEnd();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEBSplineCurve::RenderNormalsFlat()
{
  RenderNormalsSmooth();
}

VOID CSLIDEBSplineCurve::RenderControls()
{
  UINT_32 i;
  
  PushAuxillaryGeometryAttributes();
  {
    gilLineWidth(2.0);

    // Edges
    // Draw the control edges in U
    gilColor3f(1, 1, 0);
    
    gilBegin(GL_LINE_STRIP);
    {
      for(i = 0; i < m_uiControlPoints; i++ ) 
	{
	  gilVertex3fv(m_pControlPoint[i].Get());
	}
    }
    gilEnd();
    
    // Points
    gilPointSize(3.0);
    
    gilBegin(GL_POINTS);
    {
      for(i = 0; i < m_uiControlPoints; i++ ) 
	{
	  gilVertex3fv(m_pControlPoint[i].Get());
	}
    }
    gilEnd();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEBSplineCurve::WriteSLIDE(FILE *fpOutputFile)
{
  UINT_32 i;
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  if( m_uiCurveType == 0 ) // BSpline
    {
      fprintf(fpOutputFile, "%s %s\n", gcppcSLIDEReservedWords[sliderwtBSplineCurve], 
	      m_pcGUIID);
    }
  else // Bezier
    {
      fprintf(fpOutputFile, "%s %s\n", gcppcSLIDEReservedWords[sliderwtBezierCurve], 
	      m_pcGUIID);
    }
  WriteSLIDEFlags(fpOutputFile, 0);

  if( m_uiCurveType == 0 ) // BSpline
    {
      fprintf(fpOutputFile, "  %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtOrder],
	      m_uiOrder);
      fprintf(fpOutputFile, "  %s %ld\n",
	      gcppcSLIDEReservedWords[sliderwtKnots],
	      m_uiKnots);
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtKnotList]);
      
      for( i = 0; i < m_uiKnots; i++ )
	{
	  fprintf(fpOutputFile, " %.3f", m_pfKnot[i]);
	}
      fprintf(fpOutputFile, " )\n");
    }

  if( m_plistControlPointIDs != NULL )
    {
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      iter.Init(m_plistControlPointIDs);
      
      for ( pcID = (CHAR*)(iter.PeekFirst()), i = 0;
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()), i++ )
	{
	  fprintf(fpOutputFile, " %s", pcID );
	}
    }
  else
    {
      const FLOAT* pt;

      fprintf(fpOutputFile, "# Default Control Points:\n");
      fprintf(fpOutputFile, "# %s (\n#\t",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      for( i = 0; i < m_uiControlPoints; i++ )
	{
	  pt = m_pControlPoint[i].Get();
	  fprintf(fpOutputFile, "(%.3f %.3f %.3f)", pt[X], pt[Y], pt[Z] );
	}
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtSlices],
	  m_uiSlices);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtClosed],
	  m_dyniCurveClosed.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawControls],
	  m_dyniDrawControls.GetValue());

  if( m_uiCurveType == 0 ) // BSpline
    {  
      fprintf(fpOutputFile, "%s\n", 
	      gcppcSLIDEReservedWords[sliderwtEndBSplineCurve]);
    }
  else // Bezier
    {
      fprintf(fpOutputFile, "%s\n", 
	      gcppcSLIDEReservedWords[sliderwtEndBezierCurve]);
    }
}

VOID CSLIDEBSplineCurve::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
}

VOID CSLIDEBSplineCurve::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE BSplineCurve: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEBSplineCurve::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE BSplineCurve: %s", m_pcGUIID);
}

VOID CSLIDEBSplineCurve::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
					   CSLIDESurface *pCSLIDESurface)
{
}

VOID CSLIDEBSplineCurve::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  /*
  UINT_32 i;
  Point pt;

  for( i = 0; i < m_uiSlices + 1; i++ )
    {
     pt = mTransform * m_pPoint[i];

     // SLIDE
     fprintf(fpOutputFile, "point p%s_s%ld (%f %f %f) endpoint \n", m_pcID, i,
	     pt[X], pt[Y], pt[Z]);

     // Matlab
     //printf( "p%s_s%ld = [%f %f];\n", m_pcID, i, m_pPoint[i][X], m_pPoint[i][Y]);
   }
   */ 
}







