#ifndef __SLIDEBSPLINECURVE_H__
#define __SLIDEBSPLINECURVE_H__

#include "slidesweeppath.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEBSplineCurve Class
//

class CSLIDEBSplineCurve : public CSLIDESweepPath
{
//////////////////////////////////////////////////
// Friends
//
public:
  friend class CSLIDECrossSection;
  friend class CSLIDEWormwarp;

//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEBSplineCurve();
  virtual ~CSLIDEBSplineCurve();
  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual VOID InitKnots();
  virtual VOID InitControlPoints();
  virtual VOID InitControlWarps();
  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);
  virtual BOOL Valid();

  virtual inline VOID SetType(UINT_32 uiType);

  virtual inline VOID SetOrder(DynInt *pdyniOrder);
  virtual inline VOID SetKnotList(CListInt *plistKnot);
  virtual inline VOID SetControlPointList(CListInt *plistControlPoint);
  virtual inline VOID SetBezierControlPointList(CListInt *plistControlPoint);

  virtual inline VOID SetCurveClosed(DynInt *pdyniClosed);
  virtual inline VOID SetSlices(DynInt *pdyniSlices);
  virtual inline VOID SetDrawControls(DynInt *pdyniDrawControls); 

protected:

  virtual VOID ReallocateKnots();
  virtual VOID DeallocateKnots();
  virtual VOID ReallocateControlPoints();
  virtual VOID DeallocateControlPoints();
  virtual VOID ReallocateControlWarps();
  virtual VOID DeallocateControlWarps();

  virtual VOID ReallocateWarps();
  virtual VOID DeallocateWarps();
  virtual VOID ReallocatePoints();
  virtual VOID DeallocatePoints();
  virtual VOID ReallocateNormals();
  virtual VOID DeallocateNormals();
  virtual VOID ReallocateBinormals();
  virtual VOID DeallocateBinormals();
  virtual VOID ReallocateTangents();
  virtual VOID DeallocateTangents();
  virtual VOID ReallocateNormalsTorsionReduced();
  virtual VOID DeallocateNormalsTorsionReduced();
  virtual VOID ReallocateBinormalsTorsionReduced();
  virtual VOID DeallocateBinormalsTorsionReduced();
  virtual VOID ReallocateTangentsTorsionReduced();
  virtual VOID DeallocateTangentsTorsionReduced();
  virtual VOID ReallocateSweepFrenetFrames();
  virtual VOID DeallocateSweepFrenetFrames();

  virtual VOID ReallocateDerivativeTable();
  virtual VOID DeallocateDerivativeTable();
  virtual VOID ReallocateDeBoorTable();
  virtual VOID DeallocateDeBoorTable();

//////////////////////////////////////////////////
// Geometric Computations
//
protected:
  virtual VOID CheckSweepMorphCrossSectionDynamic();

  virtual VOID CheckDynamic();
  virtual VOID PreprocessGraphLocal();
  virtual VOID UpdateState();
  virtual VOID CalculateBound();

  VOID UpdateCrossSectionParents();
  VOID UpdateSweepPathParents();

protected:
  virtual UINT_32 FindKnotInterval(FLOAT t);
  virtual UINT_32 FindBlossomKnotInterval(FLOAT t);

  virtual VOID PointAt(Point &pt, FLOAT t);
  virtual VOID FrenetAt(Vector &vN, Vector &vB, Vector &vT, FLOAT fU);

  virtual VOID CalculateCurve();
  virtual VOID CalculateWarp();
  virtual VOID CalculateFrenetFrames();
  virtual VOID CalculateTorsion();

//////////////////////////////////////////////////
// Rendering
//
protected:

  virtual VOID RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp = NULL);
  virtual VOID RenderNormalsSmooth();
  virtual VOID RenderNormalsFlat();
  virtual VOID RenderControls();

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

  virtual VOID WriteRIBGeometry(FILE *fpOutputFile,
				const Matrix &mGproj_obj,
				CBound *pCBoundClip, 
				SLIDEShadingType eShading,
				SLIDELODType eLOD);
  virtual VOID WriteRIBPushSLIDEName(FILE *fpOutputFile);
  virtual VOID WriteRIBPopSLIDEName(FILE *fpOutputFile);

  virtual VOID WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				 CSLIDESurface *pCSLIDESurface);
  virtual VOID WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  virtual inline CHAR *GetTypeString();
  virtual inline SLIDEReservedWordType GetType();

  virtual inline FLOAT GetTu();
  virtual inline FLOAT GetTuDelta();
  inline FLOAT* GetKnot();
  inline UINT_32 GetKnots();

  virtual inline UINT_32 GetControlPoints();
  virtual inline Matrix *GetFrenetMatrix(); 

  virtual inline CListInt *GetControlPointList();
  virtual inline CListInt *GetControlPointIDList();

//////////////////////////////////////////////////
// Class Variables
//
protected:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  // Sweep
  BOOL m_bControlWarpDynamic;
  CListInt m_listControlWarpScale;  // DynVectors
  CListInt m_listControlWarpRotate; // DynVectors
  Point *m_pControlWarpScale;       // [Sx, Sy, Sz]
  Point *m_pControlWarpRotate;      // [Rx, Ry, Rz]
  Point *m_pWarpScale;
  Point *m_pWarpRotate;

  // Curve
  FLOAT m_fNormalLength;
  DynInt m_dyniDrawControls;
  DynInt m_dyniCurveClosed;

  UINT_32 m_uiCurveType; // [bspline, bezier]
  BOOL m_bKnotsChanged;
  BOOL m_bControlPointsDynamic;

  UINT_32 m_uiOrderPrev;
  UINT_32 m_uiOrder;
  UINT_32 m_uiKnots;
  FLOAT *m_pfKnot;
  CListInt *m_plistKnot;

  UINT_32 m_uiControlPoints;
  UINT_32 m_uiControlPointsPrev;
  Point *m_pControlPoint;
  CListInt *m_plistControlPointIDs;
  CListInt m_listCSLIDEPoint;
  CListInt *m_plistCSLIDEControlPoint;

  Vector *m_pvNormal;
  Vector *m_pvBinormal;
  Vector *m_pvTangent;
  Vector *m_pvNormalTorsionReduced;
  Vector *m_pvBinormalTorsionReduced;
  Vector *m_pvTangentTorsionReduced;
  Matrix *m_pmSweepFrenetFrame;

  // Curve Derivatives 
  UINT_32 m_uiDerivatives;
  UINT_32 m_uiDerivativeControlPoints;
  Point **m_ppDerivativeTable;
  FLOAT **m_ppfDeBoorTable;
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEBSplineCurve::SetType(UINT_32 uiType)
{
  m_uiCurveType = uiType;
}

inline VOID CSLIDEBSplineCurve::SetOrder(DynInt *pdyniOrder)
{
  ASSERT( pdyniOrder->GetValue() >= 0 );
  m_uiOrder = (UINT_32)(pdyniOrder->GetValue());
}

inline VOID CSLIDEBSplineCurve::SetKnotList(CListInt *plistKnot)
{
  m_plistKnot = plistKnot;
}

inline VOID CSLIDEBSplineCurve::SetControlPointList(CListInt *plistControlPoint)
{
  m_plistControlPointIDs = plistControlPoint;
}

inline VOID CSLIDEBSplineCurve::SetBezierControlPointList(CListInt *plistControlPoint)
{
  m_plistControlPointIDs = plistControlPoint;
  m_uiOrder = plistControlPoint->GetLength();
}

inline VOID CSLIDEBSplineCurve::SetCurveClosed(DynInt *pdyniClosed)
{
  m_dyniCurveClosed.Set(pdyniClosed);
}

inline VOID CSLIDEBSplineCurve::SetSlices(DynInt *pdyniSlices)
{
  m_dyniSlices.Set(pdyniSlices);
}

inline VOID CSLIDEBSplineCurve::SetDrawControls(DynInt *pdyniDrawControls)
{
  m_dyniDrawControls.Set(pdyniDrawControls);
}

inline CHAR *CSLIDEBSplineCurve::GetTypeString()
{
  if( m_uiCurveType == 0 ) 
    return gcppcSLIDEReservedWords[sliderwtBSplineCurve];
  else
    return gcppcSLIDEReservedWords[sliderwtBezierCurve];
}

inline SLIDEReservedWordType CSLIDEBSplineCurve::GetType()
{
  return sliderwtBSplineCurve;
}

inline FLOAT CSLIDEBSplineCurve::GetTu() 
{
  return m_pfKnot[m_uiOrder-1];
}

inline FLOAT CSLIDEBSplineCurve::GetTuDelta() 
{
  return m_pfKnot[m_uiKnots - m_uiOrder] - m_pfKnot[m_uiOrder-1];
}

inline FLOAT* CSLIDEBSplineCurve::GetKnot()
{
  return m_pfKnot;
}

inline UINT_32 CSLIDEBSplineCurve::GetKnots()
{
  return m_uiKnots;
}

inline UINT_32 CSLIDEBSplineCurve::GetControlPoints() 
{
  return m_uiControlPoints;
}

inline Matrix *CSLIDEBSplineCurve::GetFrenetMatrix() 
{
  return m_pmSweepFrenetFrame;
}

inline CListInt *CSLIDEBSplineCurve::GetControlPointList()
{
  return m_plistCSLIDEControlPoint;
}

inline CListInt *CSLIDEBSplineCurve::GetControlPointIDList()
{
  return m_plistControlPointIDs;
}

#endif // __SLIDEBSPLINECURVE_H__
