#include "slidebsplinepatch.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "qemesh.h"
#include "qevertex.h"
#include "qewrite.h"
#include "stl.h"
#include "bsplinemath.h"

static const UINT_32 cuiUSlices = 5;
static const UINT_32 cuiVSlices = 5;
static const UINT_32 cuiUOrder = 4;
static const UINT_32 cuiUKnots = 8;
static const UINT_32 cuiVOrder = 4;
static const UINT_32 cuiVKnots = 8;
static const UINT_32 cuiDrawControls = 0;
static const FLOAT cfNormalLength = 1.0;
static const FLOAT cfNormalRatio = 1.0 / 10.0;
static const FLOAT cfOffset = 0.0;
static const UINT_32 cuiOffsetConnected = 0;

//////////////////////////////////////////////////////////////////////
// CSLIDEBSplinePatch Class
//

CSLIDEBSplinePatch::CSLIDEBSplinePatch()
{
  m_fNormalLength = cfNormalLength;

  m_bControlPointsDynamic = TRUE;
  m_bSlicesChanged = FALSE;

  m_dyniDrawControls.SetValue(cuiDrawControls);
  m_dyniUSlices.SetValue(cuiUSlices);
  m_dyniVSlices.SetValue(cuiVSlices);
  m_uiUSlices = cuiUSlices;
  m_uiVSlices = cuiVSlices;
  m_uiUSlicesPrev = cuiUSlices;

  m_uiUOrderPrev = cuiUOrder;
  m_uiUOrder = cuiUOrder;
  m_uiUKnots = cuiUKnots;
  m_uiVOrderPrev = cuiVOrder;
  m_uiVOrder = cuiVOrder;
  m_uiVKnots = cuiVKnots;
  m_uiUControlPoints = cuiUKnots - cuiUOrder;
  m_uiVControlPoints = cuiVKnots - cuiVOrder;
  m_uiUControlPointsPrev = m_uiUControlPoints;

  m_pfUKnot = NULL;
  m_plistUKnot = NULL;
  m_pfVKnot = NULL;
  m_plistVKnot = NULL;

  m_ppControlPoint = NULL;
  m_plistControlPointIDs = NULL;

  m_ppSurfacePoint = NULL;
  m_ppvSurfaceNormal = NULL;

  m_uiCurveDerivatives = 0;
  m_uiCurveDerivativeControlPoints = 0;
  m_pDerivativeCurve = NULL;
  m_ppCurveDerivativeTable = NULL;
  m_ppfCurveDeBoorTable = NULL;

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);

  m_dynfOffset.SetValue(cfOffset);
  m_dyniOffsetConnected.SetValue(cuiOffsetConnected);
}

CSLIDEBSplinePatch::~CSLIDEBSplinePatch()
{
}

VOID CSLIDEBSplinePatch::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_fNormalLength = cfNormalLength;

  m_bControlPointsDynamic = TRUE;

  m_dyniDrawControls.SetValue(cuiDrawControls);
  m_dyniUSlices.SetValue(cuiUSlices);
  m_dyniVSlices.SetValue(cuiVSlices);
  m_uiUSlices = cuiUSlices;
  m_uiVSlices = cuiVSlices;
  m_uiUSlicesPrev = cuiUSlices;

  m_uiUOrderPrev = cuiUOrder;
  m_uiUOrder = cuiUOrder;
  m_uiUKnots = cuiUKnots;
  m_uiVOrderPrev = cuiVOrder;
  m_uiVOrder = cuiVOrder;
  m_uiVKnots = cuiVKnots;
  m_uiUControlPoints = cuiUKnots - cuiUOrder;
  m_uiVControlPoints = cuiVKnots - cuiVOrder;
  m_uiUControlPointsPrev = m_uiUControlPoints;

  ReallocateUKnots();
  ReallocateVKnots();
  ReallocateControlPoints();

  InitKnots();
  InitControlPoints();

  m_listCSLIDEControlPoint.Init();

  ReallocateSurfacePoints();
  ReallocateSurfaceNormals();

  m_uiCurveDerivatives = 2;
  m_uiCurveDerivativeControlPoints = 0;
  ReallocateDerivativeCurve();
  ReallocateCurveDerivativeTable();
  ReallocateCurveDeBoorTable();

  CalculateSurface();
  CalculateNormals();

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);

  m_dynfOffset.SetValue(cfOffset);
  m_dyniOffsetConnected.SetValue(cuiOffsetConnected);
}

VOID CSLIDEBSplinePatch::Uninit()
{
  CSLIDEAtomNode::Uninit();

  if( m_plistControlPointIDs != NULL )
    {
      delete m_plistControlPointIDs;
      m_plistControlPointIDs = NULL;
    }

  if( m_plistUKnot != NULL )
    {
      delete m_plistUKnot;
      m_plistUKnot = NULL;
    }

  if( m_plistVKnot != NULL )
    {
      delete m_plistVKnot;
      m_plistVKnot = NULL;
    }

   DeallocateUKnots();
   DeallocateVKnots();
   DeallocateControlPoints();
   DeallocateSurfacePoints();
   DeallocateSurfaceNormals();

   DeallocateDerivativeCurve();
   DeallocateCurveDerivativeTable();
   DeallocateCurveDeBoorTable();
}

BOOL CSLIDEBSplinePatch::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDEBSplinePatch::InitKnots()
{
  UINT_32 i;

  for( i = 0; i < m_uiUKnots; i++ )
    {
      m_pfUKnot[i] = i;
    }

  for( i = 0; i < m_uiVKnots; i++ )
    {
      m_pfVKnot[i] = i;
    }
}

VOID CSLIDEBSplinePatch::InitControlPoints()
{
  // Default Control Points
  m_ppControlPoint[0][3].Set(-1.5,  1.5, -1.0);
  m_ppControlPoint[1][3].Set(-0.5,  1.5,  0);
  m_ppControlPoint[2][3].Set( 0.5,  1.5,  0);
  m_ppControlPoint[3][3].Set( 1.5,  1.5, -1.0);

  m_ppControlPoint[0][2].Set(-1.5,  0.5,  0);
  m_ppControlPoint[1][2].Set(-0.5,  0.5,  1.0);
  m_ppControlPoint[2][2].Set( 0.5,  0.5,  1.0);
  m_ppControlPoint[3][2].Set( 1.5,  0.5,  0);

  m_ppControlPoint[0][1].Set(-1.5, -0.5,  0);
  m_ppControlPoint[1][1].Set(-0.5, -0.5,  1.0);
  m_ppControlPoint[2][1].Set( 0.5, -0.5,  1.0);
  m_ppControlPoint[3][1].Set( 1.5, -0.5,  0);

  m_ppControlPoint[0][0].Set(-1.5, -1.5, -1.0);
  m_ppControlPoint[1][0].Set(-0.5, -1.5,  0);
  m_ppControlPoint[2][0].Set( 0.5, -1.5,  0);
  m_ppControlPoint[3][0].Set( 1.5, -1.5, -1.0);
}


VOID CSLIDEBSplinePatch::DeallocateUKnots()
{
  if ( m_pfUKnot != NULL )
    {
      delete[] m_pfUKnot;
      m_pfUKnot = NULL;
    }
}

VOID CSLIDEBSplinePatch::DeallocateVKnots()
{ 
  if ( m_pfVKnot != NULL )
    {
      delete[] m_pfVKnot;
      m_pfVKnot = NULL;
    }
}

VOID CSLIDEBSplinePatch::ReallocateUKnots()
{
  FLOAT *pfKnot;

  if ( m_pfUKnot != NULL )
    {
      delete[] m_pfUKnot;
      m_pfUKnot = NULL;
    }

  pfKnot = new FLOAT[m_uiUKnots];
  ASSERT( pfKnot != NULL );
  
  m_pfUKnot = pfKnot;
}

VOID CSLIDEBSplinePatch::ReallocateVKnots()
{ 
  FLOAT *pfKnot;

  if ( m_pfVKnot != NULL )
    {
      delete[] m_pfVKnot;
      m_pfVKnot = NULL;
    }

  pfKnot = new FLOAT[m_uiVKnots];
  ASSERT( pfKnot != NULL );

  m_pfVKnot = pfKnot;
}

VOID CSLIDEBSplinePatch::DeallocateControlPoints()
{
  UINT_32 i;
  
  if ( m_ppControlPoint != NULL )
    {
      for( i = 0; i < m_uiUControlPointsPrev; i++ ) 
	{
	  delete[] m_ppControlPoint[i];
	}
      delete[] m_ppControlPoint;
      m_ppControlPoint = NULL;
    }
}

VOID CSLIDEBSplinePatch::ReallocateControlPoints()
{
  UINT_32 i;

  if( m_ppControlPoint != NULL )
    {
      DeallocateControlPoints();
    }

  m_uiUControlPointsPrev = m_uiUControlPoints;
  m_ppControlPoint = new Point*[m_uiUControlPoints];
  ASSERT(m_ppControlPoint != NULL);
  
  for( i = 0; i < m_uiUControlPoints; i++ ) 
    {
      m_ppControlPoint[i] = new Point[m_uiVControlPoints];
      ASSERT(m_ppControlPoint[i] != NULL);
    }
}

VOID CSLIDEBSplinePatch::DeallocateSurfacePoints()
{
  UINT_32 i;
  
  if ( m_ppSurfacePoint != NULL )
    {
      for( i = 0; i < m_uiUSlicesPrev + 1; i++ ) 
	{
	  delete[] m_ppSurfacePoint[i];
	}
      delete[] m_ppSurfacePoint;
      m_ppSurfacePoint = NULL;
    }
}

VOID CSLIDEBSplinePatch::ReallocateSurfacePoints()
{
  UINT_32 i;

  if( m_ppSurfacePoint != NULL )
    {
      DeallocateSurfacePoints();
    }

  m_ppSurfacePoint = new Point*[m_uiUSlices+1];
  ASSERT(m_ppSurfacePoint != NULL);

  for(i = 0; i < m_uiUSlices + 1; i++ ) 
    {
      m_ppSurfacePoint[i] = new Point[m_uiVSlices+1];
      ASSERT(m_ppSurfacePoint[i] != NULL);
    }
}

VOID CSLIDEBSplinePatch::DeallocateSurfaceNormals()
{
  UINT_32 i;
  
  if ( m_ppvSurfaceNormal != NULL )
    {
      for( i = 0; i < m_uiUSlicesPrev + 1; i++ ) 
	{
	  delete[] m_ppvSurfaceNormal[i];
	}
      delete[] m_ppvSurfaceNormal;
      m_ppvSurfaceNormal = NULL;
    }
}

VOID CSLIDEBSplinePatch::ReallocateSurfaceNormals()
{
  UINT_32 i;

  if( m_ppvSurfaceNormal != NULL )
    {
      DeallocateSurfaceNormals();
    }

  m_ppvSurfaceNormal = new Vector*[m_uiUSlices+1];
  ASSERT(m_ppvSurfaceNormal != NULL);

  for(i = 0; i < m_uiUSlices + 1; i++ ) 
    {
      m_ppvSurfaceNormal[i] = new Vector[m_uiVSlices+1];
      ASSERT(m_ppvSurfaceNormal[i] != NULL);
    }
}


VOID CSLIDEBSplinePatch::DeallocateDerivativeCurve()
{
  delete[] m_pDerivativeCurve;
}

VOID CSLIDEBSplinePatch::ReallocateDerivativeCurve()
{

  if( m_pDerivativeCurve != NULL )
    {
      DeallocateDerivativeCurve();
    }

 if( m_uiUOrder > m_uiVOrder )
    {
      m_pDerivativeCurve = new Point[m_uiUOrder];
    }
  else
    {
      m_pDerivativeCurve = new Point[m_uiVOrder];
    }
}

VOID CSLIDEBSplinePatch::DeallocateCurveDerivativeTable()
{
 UINT_32 i;

  if ( m_ppCurveDerivativeTable != NULL )
    {
      for ( i = 0; i < m_uiCurveDerivativeControlPoints; i++ )
	{
	  delete[] m_ppCurveDerivativeTable[i];
	}
      delete[] m_ppCurveDerivativeTable;
      m_ppCurveDerivativeTable = NULL;
      m_uiCurveDerivativeControlPoints = 0;
    }
}

VOID CSLIDEBSplinePatch::ReallocateCurveDerivativeTable()
{
  UINT_32 i, uiControlPoints;

  if( m_ppCurveDerivativeTable != NULL )
    {
      DeallocateCurveDerivativeTable();
    }

  if( m_uiUControlPoints > m_uiVControlPoints )
    {
      uiControlPoints = m_uiUControlPoints;
    }
  else
    {
      uiControlPoints = m_uiVControlPoints;
    }

  ASSERT( m_uiCurveDerivativeControlPoints == 0 );
  ASSERT( m_ppCurveDerivativeTable == NULL );

  m_uiCurveDerivativeControlPoints = uiControlPoints;
  m_ppCurveDerivativeTable = new Point*[m_uiCurveDerivativeControlPoints];
  ASSERT( m_ppCurveDerivativeTable != NULL );

  for ( i = 0; i < m_uiCurveDerivativeControlPoints; i++ )
    {
      m_ppCurveDerivativeTable[i] = new Point[m_uiCurveDerivatives]; 
      ASSERT( m_ppCurveDerivativeTable[i] != NULL );
    }
}

VOID CSLIDEBSplinePatch::DeallocateCurveDeBoorTable()
{
  UINT_32 i, uiOrder;

  if ( m_ppfCurveDeBoorTable != NULL )
    {

      if( m_uiUOrder > m_uiVOrder )
	{
	  uiOrder = m_uiUOrderPrev;
	}
      else
	{
	  uiOrder = m_uiVOrderPrev;
	}

      for ( i = 0; i < (2*uiOrder - 1); i++ )
	{
	  delete[] m_ppfCurveDeBoorTable[i];
	}
      delete[] m_ppfCurveDeBoorTable;
      m_ppfCurveDeBoorTable = NULL;
    }
}

VOID CSLIDEBSplinePatch::ReallocateCurveDeBoorTable()
{
  UINT_32 i, uiOrder;

  if( m_ppfCurveDeBoorTable != NULL )
    {
      DeallocateCurveDeBoorTable();
    }

  if( m_uiUOrder > m_uiVOrder )
    {
      uiOrder = m_uiUOrder;
    }
  else
    {
      uiOrder = m_uiVOrder;
    }

  m_uiUOrderPrev = m_uiUOrder;
  m_uiVOrderPrev = m_uiVOrder;

  m_ppfCurveDeBoorTable = new FLOAT*[2*uiOrder - 1];
  ASSERT( m_ppfCurveDeBoorTable != NULL );

  for ( i = 0; i < (2*uiOrder - 1); i++ )
    {
      m_ppfCurveDeBoorTable[i] = new FLOAT[uiOrder];
      ASSERT( m_ppfCurveDeBoorTable[i] != NULL );
    }
}

BOOL CSLIDEBSplinePatch::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  UINT_32 i;
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;
  DynFloat *pdynfKnot;

  bCorrect = TRUE;

  if ( !( m_uiUOrder > 1 ) )
    {
      HandleError("uorder(%ld) must be > 1", m_uiUOrder );
      bCorrect = FALSE;
    }

  if ( !( m_uiUKnots - m_uiUOrder >= m_uiUOrder ) )
    {
      HandleError("uknots(%ld) must be >= 2*uorder(%ld)", m_uiUKnots, m_uiUOrder );
      bCorrect = FALSE;
    }

  if( m_plistUKnot != NULL ) 
    {
      if( m_plistUKnot->GetLength() != m_uiUKnots )
	{
	  HandleError("Wrong number of uknots specified in uknotlist: expecting %ld",
		      m_uiUKnots);
	  bCorrect = FALSE;
	}
      else
	{
	  ReallocateUKnots();

	  iter.Init(m_plistUKnot);	  
	  for ( pdynfKnot = (DynFloat*)(iter.PeekFirst()), i = 0;
		pdynfKnot != NULL;
		pdynfKnot = (DynFloat*)(iter.PeekNext()), i++ )
	    {      
	      ASSERT( i < m_uiUKnots );
	      m_pfUKnot[i] = pdynfKnot->GetValue();
	    }

	}
    }

  if ( !( m_uiVOrder > 1 ) )
    {
      HandleError("vorder(%ld) must be > 1 ", m_uiVOrder );
      bCorrect = FALSE;
    }

  if ( !( m_uiVKnots - m_uiVOrder >= m_uiVOrder ) )
    {
      HandleError("vknots(%ld) must be >= 2*vorder(%ld)", m_uiVKnots, m_uiVOrder );
      bCorrect = FALSE;
    }

  if( m_plistVKnot != NULL )
    {
      if( m_plistVKnot->GetLength() != m_uiVKnots )
	{
	  HandleError("Wrong number of vknots specified in vknotlist: expecting %ld",
		      m_uiVKnots);
	  bCorrect = FALSE;
	}
      else
	{
	  ReallocateVKnots();
	  iter.Init(m_plistVKnot);
	  
	  for ( pdynfKnot = (DynFloat*)(iter.PeekFirst()), i = 0;
		pdynfKnot != NULL;
		pdynfKnot = (DynFloat*)(iter.PeekNext()), i++ )
	    {
	      ASSERT( i < m_uiVKnots );
	      m_pfVKnot[i] = pdynfKnot->GetValue();
	    }
	}
    }

  m_uiUControlPoints = m_uiUKnots - m_uiUOrder;
  m_uiVControlPoints = m_uiVKnots - m_uiVOrder;

  // Check Point References
  if( m_plistControlPointIDs != NULL )
    {
      if( m_plistControlPointIDs->GetLength() != (m_uiUControlPoints*m_uiVControlPoints) )
	{
	  HandleError("Wrong number of controlpoints: expecting %ld", 
		      (m_uiUControlPoints*m_uiVControlPoints));
	  bCorrect = FALSE;
	}
      else
	{
	  ReallocateControlPoints();
	  iter.Init(m_plistControlPointIDs);
	  
	  for ( pcID = (CHAR*)(iter.PeekFirst());
		pcID != NULL;
		pcID = (CHAR*)(iter.PeekNext()) )
	    {
	      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
	      if ( pCSLIDEPoint == NULL )
		{
		  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint],
			      pcID);
		  bCorrect = FALSE;
		}
	      m_listCSLIDEControlPoint.InsertLast(pCSLIDEPoint, 0);
	    }
	}
    }
  else
    {
      if( 16 < (m_uiUKnots-m_uiUOrder)*(m_uiVKnots-m_uiVOrder))
	{
	  HandleError("Wrong number of control points specified:\n\t\t expecting %ld u controlpoints by %ld v controlpoints", (m_uiUKnots-m_uiUOrder), (m_uiVKnots-m_uiVOrder) );
	  bCorrect = FALSE;
	}
    }

  if( (1 > m_uiUSlices) || (m_uiUSlices > 1000)  )
    {
      HandleError("uslices must be between 1 and 1000" );
      bCorrect = FALSE;
    }

  if( (1 > m_uiVSlices) || (m_uiVSlices > 1000) )
    {
      HandleError("vslices must be between 1 and 1000" );
      bCorrect = FALSE;
    }

  ReallocateSurfacePoints();
  ReallocateSurfaceNormals();
  ReallocateDerivativeCurve();
  ReallocateCurveDerivativeTable();
  ReallocateCurveDeBoorTable();

  return bCorrect;
}

BOOL CSLIDEBSplinePatch::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dyniUSlices.GetValue() >= 0 ) &&
	   ( m_dyniVSlices.GetValue() >= 0 ) &&
	   ( m_dyniDrawControls.GetValue() >= 0 ) );
}

VOID CSLIDEBSplinePatch::GetSide(CListInt *plistPoint, CListInt *plistNormal, 
				 UINT_32 uiSide, const Matrix &mTransform, 
				 const Matrix &mInverse, BOOL bReverse)
{
  ASSERT(uiSide < 4);

  UINT_32 i, j;
  Matrix m;
  Point pt;
  Point *ppt;
  Vector vNormal;
  Vector *pvNormal;

  if( (uiSide == 0) || (uiSide == 2) )
    {
      if( uiSide == 0 )
	{
	  j = 0;
	}
      else
	{
	  j = m_uiVSlices;
	}

      if( bReverse == 0 )
	{
	  for ( i = 0; i <= m_uiUSlices; i++ )
	    {
	      pt.Set(m_ppSurfacePoint[i][j].Get());
	      pt = mTransform * pt;
	      ppt = new Point;
	      ppt->Set(pt.Get());
	      plistPoint->InsertLast(ppt, 0);
	      
	      vNormal.Set(m_ppvSurfaceNormal[i][j].Get());
	      m = mInverse;
	      m.Transpose();
	      vNormal = m * vNormal;
	      if( !vNormal.Normalize() )
		{
		  vNormal.Set(0.0, 0.0, 1.0);
		}
	      pvNormal = new Vector;
	      pvNormal->Set(vNormal.Get());
	      plistNormal->InsertLast(pvNormal, 0);
	    }
	}
      else
	{
	  for ( i = 0; i <= m_uiUSlices; i++ )
	    {
	      pt.Set(m_ppSurfacePoint[m_uiUSlices-i][j].Get());
	      pt = mTransform * pt;
	      ppt = new Point;
	      ppt->Set(pt.Get());
	      plistPoint->InsertLast(ppt, 0);
	      
	      vNormal.Set(m_ppvSurfaceNormal[m_uiUSlices-i][j].Get());
	      m = mInverse;
	      m.Transpose();
	      vNormal = m * vNormal;
	      if( !vNormal.Normalize() )
		{
		  vNormal.Set(0.0, 0.0, 1.0);
		}
	      pvNormal = new Vector;
	      pvNormal->Set(vNormal.Get());
	      plistNormal->InsertLast(pvNormal, 0);
	    }
	}
    }
  else // (uiSide == 1) || (uiSide == 3)
    {
      if( uiSide == 1 )
	{
	  i = m_uiUSlices;
	}
      else
	{
	  i = 0;
	}

      if( bReverse == 0 )
	{
	  for ( j = 0; j <= m_uiVSlices; j++ )
	    {
	      pt.Set(m_ppSurfacePoint[i][j].Get());
	      pt = mTransform * pt;
	      ppt = new Point;
	      ppt->Set(pt.Get());
	      plistPoint->InsertLast(ppt, 0);
	      
	      vNormal.Set(m_ppvSurfaceNormal[i][j].Get());
	      m = mInverse;
	      m.Transpose();
	      vNormal = m * vNormal;
	      if( !vNormal.Normalize() )
		{
		  vNormal.Set(0.0, 0.0, 1.0);
		}
	      pvNormal = new Vector;
	      pvNormal->Set(vNormal.Get());
	      plistNormal->InsertLast(pvNormal, 0);
	    }
	}
      else
	{
	  for ( j = 0; j <= m_uiVSlices; j++ )
	    {
	      pt.Set(m_ppSurfacePoint[i][m_uiVSlices-j].Get());
	      pt = mTransform * pt;
	      ppt = new Point;
	      ppt->Set(pt.Get());
	      plistPoint->InsertLast(ppt, 0);
	      
	      vNormal.Set(m_ppvSurfaceNormal[i][m_uiVSlices-j].Get());
	      m = mInverse;
	      m.Transpose();
	      vNormal = m * vNormal;
	      if( !vNormal.Normalize() )
		{
		  vNormal.Set(0.0, 0.0, 1.0);
		}
	      pvNormal = new Vector;
	      pvNormal->Set(vNormal.Get());
	      plistNormal->InsertLast(pvNormal, 0);
	    }
	}
    }
}

VOID CSLIDEBSplinePatch::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  if( !GetDynamic() )
    {
      SetDynamic(m_dyniDrawControls.GetDynamic() ||
		 m_dyniUSlices.GetDynamic() ||
		 m_dyniVSlices.GetDynamic() || 
		 m_pvTexture[0].GetDynamic() ||
		 m_pvTexture[1].GetDynamic() ||
		 m_pvTexture[2].GetDynamic() ||
		 m_pvTexture[3].GetDynamic() ||
		 m_dynfOffset.GetDynamic() ||
		 m_dyniOffsetConnected.GetDynamic() );
    }

  if( !GetDynamic() )
    {
      CListIterInt iter;
      CSLIDEPoint *pCSLIDEPoint;
      BOOL bBoundDynamic;
      
      bBoundDynamic = FALSE;
      
      if( m_plistControlPointIDs != NULL )
	{
	  iter.Init(&m_listCSLIDEControlPoint);
	  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
		pCSLIDEPoint != NULL;
		pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	    {
	      bBoundDynamic = bBoundDynamic || pCSLIDEPoint->GetBoundDynamic();
	    }
	}     
      SetDynamic( bBoundDynamic );
    }
}

VOID CSLIDEBSplinePatch::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bBoundEmpty;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  if( m_plistControlPointIDs != NULL )
    {
      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;
      bGeometryDynamic = FALSE;
      bNormalsDynamic = FALSE;
      
      iter.Init(&m_listCSLIDEControlPoint);
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  bBoundEmpty = FALSE;
	  bBoundDynamic = bBoundDynamic || pCSLIDEPoint->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || pCSLIDEPoint->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic || pCSLIDEPoint->GetNormalDynamic();
	}
    }
  else
    {
      // Used Default patch
      bBoundEmpty = FALSE;
      bBoundDynamic = FALSE;
      bGeometryDynamic = FALSE;
      bNormalsDynamic = FALSE;
    }

  bRealSpaceDynamic = ( m_dyniDrawControls.GetDynamic() ||
			m_dynfOffset.GetDynamic() ||
			m_dyniOffsetConnected.GetDynamic() );
  bParameterSpaceDynamic = ( m_dyniUSlices.GetDynamic() ||
			     m_dyniVSlices.GetDynamic() );
  bTextureSpaceDynamic = ( m_pvTexture[0].GetDynamic() ||
			   m_pvTexture[1].GetDynamic() ||
			   m_pvTexture[2].GetDynamic() ||
			   m_pvTexture[3].GetDynamic() );

  m_bControlPointsDynamic = bGeometryDynamic;
  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bGeometryDynamic ||
		     m_pCSLIDESurface->GetDynamic() ||
		     bTextureSpaceDynamic);
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || 
		    bParameterSpaceDynamic || 
		    bNormalsDynamic );
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic);
  SetBoundEmpty(bBoundEmpty);
  SetBoundDynamic(bRealSpaceDynamic || 
		  bBoundDynamic);
}

VOID CSLIDEBSplinePatch::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  UINT_32 i, j;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bControlPointsChanged;

  m_bSlicesChanged = FALSE;
  bControlPointsChanged = FALSE;

  // m_bControlPointsDynamic = TRUE on first call to UpdateState
  if( (m_plistControlPointIDs != NULL) && (m_bControlPointsDynamic == TRUE) )
    {
      iter.Init(&m_listCSLIDEControlPoint);

      i = 0;
      j = m_uiVControlPoints-1;

      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  if( m_ppControlPoint[i][j] != pCSLIDEPoint->GetPoint() )
	    {
	      m_ppControlPoint[i][j] = pCSLIDEPoint->GetPoint();
	      bControlPointsChanged = TRUE;
	    }
	  
	  if( i == m_uiUControlPoints - 1)
	    {
	      i = 0;
	      j--;
	    }
	  else
	    {
	      i++;
	    }
	}
    }

  m_dyniDrawControls.Update();

  for ( i = 0; i < 4; i++ )
    {
      m_pvTexture[i].Update();
    }

  m_dynfOffset.Update();
  m_dyniOffsetConnected.Update();

  m_dyniUSlices.Update();
  if ( (1 > m_dyniUSlices.GetValue() || (m_dyniUSlices.GetValue() > 1000)) )
    {
      HandleError("uslices must be between 1 and 1000" );
      m_dyniUSlices.SetValue(cuiUSlices);
    }
  m_dyniVSlices.Update();
  if ( (1 > m_dyniVSlices.GetValue() || (m_dyniVSlices.GetValue() > 1000)) )
    {
      HandleError("vslices must be between 1 and 1000" );
      m_dyniVSlices.SetValue(cuiVSlices);
    }

  // Slices are the only dynamic variable that can change memory needed
  if( ((UINT_32)m_dyniUSlices.GetValue() != m_uiUSlices) || 
      ((UINT_32)m_dyniVSlices.GetValue() != m_uiVSlices) )
    {
      m_bSlicesChanged = TRUE;

      m_uiUSlicesPrev = m_uiUSlices;
      m_uiUSlices = m_dyniUSlices.GetValue();
      m_uiVSlices = m_dyniVSlices.GetValue();
      
      ReallocateSurfacePoints();
      ReallocateSurfaceNormals();
      ReallocateDerivativeCurve();
      ReallocateCurveDerivativeTable();
      ReallocateCurveDeBoorTable();

      CalculateSurface();
      CalculateNormals();
    }
  else if ( bControlPointsChanged )
    {
      CalculateSurface();
      CalculateNormals();
    }
}

VOID CSLIDEBSplinePatch::CalculateBound()
{
  UINT_32 i, j;
  CBound bound;

  for( i = 0; i < m_uiUControlPoints; i++ )
    {
      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  bound.Insert(m_ppControlPoint[i][j]);
	}
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }
}

#ifdef __USE_CDA__
VOID CSLIDEBSplinePatch::CalculateCDAVertices(CListInt *plistCDAVertex, 
					      const Matrix &mGroot_obj)
{

}
#endif // __USE_CDA__

UINT_32 CSLIDEBSplinePatch::FindKnotInterval(FLOAT t, BOOL bUDirection)
{
  UINT_32 uiDelta, uiMin, uiMax, uiOrder;
  FLOAT *pfKnot; 

  uiMin = 0;

  if( bUDirection ) {   // Look in the u-direction 
    uiMax = m_uiUKnots - m_uiUOrder;
    uiOrder = m_uiUOrder;
    pfKnot = m_pfUKnot;
  } else {              // Look in the v-direction
    uiMax = m_uiVKnots - m_uiVOrder;
    uiOrder = m_uiVOrder;
    pfKnot = m_pfVKnot;
  }

  // knot[d] <= t < knot[d+1]. 
  ASSERT( uiOrder > 1 );
  for ( uiDelta = uiOrder-1; uiDelta < uiMax; uiDelta++ )
    {
      if ( t < pfKnot[uiDelta+1] ) // <==> (t >= pfKnot[uiDelta])
	{
	  break;
	}
    }

  return uiDelta;
}

UINT_32 CSLIDEBSplinePatch::FindBlossomKnotInterval(FLOAT t, BOOL bUDirection)
{
  UINT_32 uiBlossomKnotIndex;

  if( bUDirection )
    {
      uiBlossomKnotIndex = FindKnotInterval(t, bUDirection) - (m_uiUOrder - 1) + 1;
      ASSERT( (1 <= uiBlossomKnotIndex) && 
	      (uiBlossomKnotIndex <= m_uiUControlPoints - (m_uiUOrder - 1) + 1) );
    }
  else
    {
      uiBlossomKnotIndex = FindKnotInterval(t, bUDirection) - (m_uiVOrder - 1) + 1;
      ASSERT( (1 <= uiBlossomKnotIndex) && 
	      (uiBlossomKnotIndex <= m_uiVControlPoints - (m_uiVOrder - 1) + 1) );
    }

  return uiBlossomKnotIndex;
}

VOID CSLIDEBSplinePatch::PointAt(Point &pt, FLOAT fU, FLOAT fV)
{
  UINT_32 knotIndex;

  Point *pControlCurvePoint;

  pControlCurvePoint = new Point[m_uiUControlPoints];

  // First find the control points for the curve in the u-direction for tv = v
  ControlCurveAt(pControlCurvePoint, fV, TRUE);

  // Use blossoming to evaluate u on this curve
  knotIndex = FindBlossomKnotInterval(fU, TRUE);
  BlossomAt(pt, m_uiUOrder, &pControlCurvePoint[knotIndex-1], &m_pfUKnot[knotIndex], fU);

  delete []pControlCurvePoint;
}

VOID CSLIDEBSplinePatch::ControlCurveAt(Point *pt, FLOAT t, BOOL bUDirection)
{
  UINT_32 uiOrder, uiOtherControlPoints, i, j, knotIndex;
  FLOAT *pfKnot;
  Point *pControlPoint;

  if( bUDirection )               
    {                             // tv = t
      uiOrder = m_uiVOrder;
      pfKnot = m_pfVKnot;
      uiOtherControlPoints = m_uiUControlPoints;
      pControlPoint = new Point[m_uiVControlPoints];
    }
  else
    {                             // tu = t
      uiOrder = m_uiUOrder;
      pfKnot = m_pfUKnot;         
      uiOtherControlPoints = m_uiVControlPoints;  
      pControlPoint = new Point[m_uiUControlPoints];
    }

  for( i = 0; i < uiOtherControlPoints; i++ ) 
    {
      if( bUDirection )
	{
	  for( j = 0; j < m_uiVControlPoints; j++ )
	    {
	      pControlPoint[j] = m_ppControlPoint[i][j];
	    }
	}
      else 
	{
	  for( j = 0; j < m_uiUControlPoints; j++ )
	    {
	      pControlPoint[j] = m_ppControlPoint[j][i];
	    }
	}

      knotIndex = FindBlossomKnotInterval(t, !bUDirection);
      BlossomAt(pt[i], uiOrder, &pControlPoint[knotIndex-1], &pfKnot[knotIndex], t);
    }
  
  if( bUDirection ) 
    {
      delete[] pControlPoint;
    }
}

VOID CSLIDEBSplinePatch::CalculateSurface()
{
  UINT_32 uiUSlice, uiVSlice, knotIndex;
  FLOAT fU, fV, fUDelta, fVDelta;
  Point pt;
  Point *pControlCurveVertex;

  fV = GetTv();
  fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;
  fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;
  pControlCurveVertex = new Point[m_uiUControlPoints];

  for( uiVSlice = 0; uiVSlice <= m_uiVSlices; uiVSlice++ )
    {
      ControlCurveAt(pControlCurveVertex, fV, TRUE);
      fU = GetTu();

      for( uiUSlice = 0; uiUSlice <= m_uiUSlices; uiUSlice++ )
	{
	  knotIndex = FindBlossomKnotInterval(fU, TRUE);
	  BlossomAt(pt, m_uiUOrder, &pControlCurveVertex[knotIndex-1], 
		    &m_pfUKnot[knotIndex], fU);
	  m_ppSurfacePoint[uiUSlice][uiVSlice] = pt;

	  fU += fUDelta;
	}
      
      fV += fVDelta;
    }
  delete[] pControlCurveVertex;
}

VOID CSLIDEBSplinePatch::CurveTangentAt(Vector &vT, FLOAT fU, FLOAT fV, BOOL bUDirection)
{
  UINT_32 uiDelta, uiOrder, uiControlPoints;
  Vector pv[2];
  FLOAT *pfKnot;
  FLOAT t;
 
  ASSERT( Valid() );

  if( bUDirection )
    {
      t = MAX(fU, GetTu());
      t = MIN(fU, GetTu() + GetTuDelta() - EPSILON);      
      uiOrder = m_uiUOrder;
      pfKnot = m_pfUKnot;
      uiControlPoints = m_uiUControlPoints;
    }
  else
    {
      t = MAX(fV, GetTv());
      t = MIN(fV, GetTv() + GetTvDelta() - EPSILON);
      uiOrder = m_uiVOrder;
      pfKnot = m_pfVKnot;
      uiControlPoints = m_uiVControlPoints;
    }

  uiDelta = FindKnotInterval(t, bUDirection);

  CalculateDerivativeTable(uiControlPoints, m_uiCurveDerivatives, uiOrder, pfKnot, 
			   m_pDerivativeCurve, m_ppCurveDerivativeTable);
  CalculateDeBoorTable(uiOrder, pfKnot, m_ppfCurveDeBoorTable, t, uiDelta);
  DerivativesAt(pv, uiOrder, m_ppCurveDerivativeTable, m_ppfCurveDeBoorTable, 
		0, 1, t, uiDelta);

  vT = pv[1];
  vT.Normalize();
}

VOID CSLIDEBSplinePatch::NormalAt(Vector &vN, FLOAT fU, FLOAT fV)
{
  Vector vVT, vUT;
  Point pt;

  ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
  CurveTangentAt( vUT, fU, fV, TRUE );

  ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
  CurveTangentAt( vVT, fU, fV, FALSE );    

  vN = vUT.Cross(vVT);
  vN.Normalize();
}

VOID CSLIDEBSplinePatch::CalculateNormals()
{
  UINT_32 uiUSlice, uiVSlice;
  FLOAT fU, fV, fUDelta, fVDelta;
  Vector vNormal, vUT, vVT;
  
  fV = GetTv();
  fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;
  fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;
  
  for( uiVSlice = 0; uiVSlice <= m_uiVSlices; uiVSlice++ )
    {
      fU = GetTu();
      
      for( uiUSlice = 0; uiUSlice <= m_uiUSlices; uiUSlice++ )
	{
	  NormalAt(vNormal, fU, fV);

	  if( vNormal.Magnitude() <= EPSILON )
	    {
	      printf( "Zero length normal because of degenerate triangle; Setting normal to (0 0 1)\n" );
	      vNormal.Set(0, 0, 1, 1);	      
	    }

	  m_ppvSurfaceNormal[uiUSlice][uiVSlice] = vNormal;
	  
	  fU += fUDelta;
	}
      fV += fVDelta;
    }
}

VOID CSLIDEBSplinePatch::CalculateQEMesh()
{
  QEVertex* pqev;
  CObjectList<QEVertex> listQEVertex;

  UINT_32 i, j;
   
  ASSERT(m_pQEMesh == NULL);
  m_pQEMesh=AllocateQEMesh();
  ASSERT(m_pQEMesh != NULL);
  m_pQEMesh->Init();
  
  listQEVertex.Init();
  
  for( i = 0; i < m_uiUSlices; i++ )
    {
      listQEVertex.RemoveAll();
      //gilBegin(GL_TRIANGLE_STRIP);
      {
	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pqev->Init(m_ppSurfacePoint[i][j]);
	    listQEVertex.InsertLast(pqev);
	    
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pqev->Init(m_ppSurfacePoint[i+1][j]);
	    listQEVertex.InsertLast(pqev);
	    
	  }
      }
      m_pQEMesh->MergeTriangleStrip(&listQEVertex);
      //gilEnd();
    }
}

VOID CSLIDEBSplinePatch::CalculateLeafQEMesh()
{
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDEBSplinePatch::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

VOID CSLIDEBSplinePatch::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  if( m_dyniDrawControls.GetValue() != 0 )
    {
      RenderControls();
    }

  if( m_dynfOffset.GetValue() != 0.0 )
    {
      RenderOffsetGeometrySmooth();
      return;
    }

  UINT_32 i, j;
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvU[2][2];
  Vector pvV[2];
  
  pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
  pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
  
  ppvU[0][0] = m_pvTexture[3];
  ppvU[0][1] = m_pvTexture[0];
  ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
  ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
  
  for( i = 0; i < m_uiUSlices; i++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	pvV[0] = ppvU[0][0];
	pvV[1] = ppvU[1][0];
	
	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	    gilNormal3fv(m_ppvSurfaceNormal[i][j].Get());
	    gilVertex3fv(m_ppSurfacePoint[i][j].Get());
	    
	    gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	    gilNormal3fv(m_ppvSurfaceNormal[i+1][j].Get());
	    gilVertex3fv(m_ppSurfacePoint[i+1][j].Get());
	    
	    pvV[0] += pvVDelta[0];
	    pvV[1] += pvVDelta[1];
	  }
      }
      gilEnd();
      
      ppvU[0][0] += pvUDelta[0];
      ppvU[0][1] += pvUDelta[1];
      ppvU[1][0] += pvUDelta[0];
      ppvU[1][1] += pvUDelta[1];
    }
}

VOID CSLIDEBSplinePatch::RenderOffsetGeometrySmooth()
{
  UINT_32 i, j;
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvU[2][2];
  Vector pvV[2];
  Vector vNormal;
  Point pt;
  FLOAT fOffset;

  fOffset = m_dynfOffset.GetValue();
  
  pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
  pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
  
  // Correct the orientation
  if( fOffset < 0.0 )
    {
      // Original Surface
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    for( j = 0; j < m_uiVSlices+1; j++ )
	      {
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv(m_ppvSurfaceNormal[i][j].Get());
		gilVertex3fv(m_ppSurfacePoint[i][j].Get());
		
		gilTexCoord2f(pvV[1][0], pvV[1][1]);	
		gilNormal3fv(m_ppvSurfaceNormal[i+1][j].Get());
		gilVertex3fv(m_ppSurfacePoint[i+1][j].Get());
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
      
      // Offset Surface with orienation reversed
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    for( j = 0; j < m_uiVSlices+1; j++ )
	      {
		vNormal = m_ppvSurfaceNormal[i+1][j];
		ASSERT( vNormal.Normalize() );
		pt = m_ppSurfacePoint[i+1][j] + vNormal*fOffset;		
		gilTexCoord2f(pvV[1][0], pvV[1][1]);
		gilNormal3fv((-vNormal).Get());
		gilVertex3fv(pt.Get());
		
		vNormal = m_ppvSurfaceNormal[i][j];
		ASSERT( vNormal.Normalize() );
		pt = m_ppSurfacePoint[i][j] + vNormal*fOffset;		
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv((-vNormal).Get());
		gilVertex3fv(pt.Get());
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }
  else
    {
      // Original Surface with orienation reversed
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    for( j = 0; j < m_uiVSlices+1; j++ )
	      {
		gilTexCoord2f(pvV[1][0], pvV[1][1]);	
		gilNormal3fv((-m_ppvSurfaceNormal[i+1][j]).Get());
		gilVertex3fv(m_ppSurfacePoint[i+1][j].Get());		    
		
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv((-m_ppvSurfaceNormal[i][j]).Get());
		gilVertex3fv(m_ppSurfacePoint[i][j].Get());
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
      
      // Offset Surface
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    for( j = 0; j < m_uiVSlices+1; j++ )
	      {
		vNormal = m_ppvSurfaceNormal[i][j];
		ASSERT( vNormal.Normalize() );
		pt = m_ppSurfacePoint[i][j] + vNormal*fOffset;		
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv(pt.Get());
		
		vNormal = m_ppvSurfaceNormal[i+1][j];
		ASSERT( vNormal.Normalize() );
		pt = m_ppSurfacePoint[i+1][j] + vNormal*fOffset;		
		gilTexCoord2f(pvV[1][0], pvV[1][1]);
		gilNormal3fv(vNormal.Get());
		gilVertex3fv(pt.Get());
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }
      
  // Connect the surface and offset surface
  if( m_dyniOffsetConnected.GetValue() != 0 ) 
    {
      RenderOffsetGeometryConnectedSmooth();
    }
}

VOID CSLIDEBSplinePatch::RenderOffsetGeometryConnectedSmooth()
{
  UINT_32 i, j;
  Point p1, p2;
  Vector vNormal;
  FLOAT fOffset, fV, fU, fUDelta, fVDelta;

  fOffset = m_dynfOffset.GetValue();

  // Correct the orientation
  if( fOffset < 0.0 )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = 0;
	fV = GetTv();
	fU = GetTu();
	fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;

	for( i = 0; i < m_uiUSlices+1; i++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;	    
	    ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
	    CurveTangentAt( vNormal, fU, fV, FALSE ); 
	    ASSERT(vNormal.Normalize());

	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p1.Get());
	    
	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p2.Get());

	    fU += fUDelta;
	  }
      }
      gilEnd();

      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = m_uiVSlices;
	fV = GetTv() + GetTvDelta();
	fU = GetTu();
	fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;

	for( i = 0; i < m_uiUSlices+1; i++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
	    CurveTangentAt( vNormal, fU, fV, FALSE ); 
	    ASSERT(vNormal.Normalize());
	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p2.Get());
	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p1.Get());
	    
	    fU += fUDelta;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = 0;
	fV = GetTv();
	fU = GetTu();
	fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;

	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
	    CurveTangentAt( vNormal, fU, fV, TRUE );
	    ASSERT(vNormal.Normalize());

	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p2.Get());

	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p1.Get());

	    fV += fVDelta;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = m_uiUSlices;
	fV = GetTv();
	fU = GetTu() + GetTuDelta();
	fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;
	    
	for( j = 0; j < m_uiVSlices+1; j++ ) 
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
	    CurveTangentAt( vNormal, fU, fV, TRUE );
	    ASSERT(vNormal.Normalize());

	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p1.Get());
	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p2.Get());

	    fV += fVDelta;
	  }
      }
      gilEnd();
    }
  else
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = 0;
	fV = GetTv();
	fU = GetTu();
	fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;

	for( i = 0; i < m_uiUSlices+1; i++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;	    
	    ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
	    CurveTangentAt( vNormal, fU, fV, FALSE ); 
	    ASSERT(vNormal.Normalize());

	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p2.Get());
	    
	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p1.Get());

	    fU += fUDelta;
	  }
      }
      gilEnd();

      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = m_uiVSlices;
	fV = GetTv() + GetTvDelta();
	fU = GetTu();
	fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;

	for( i = 0; i < m_uiUSlices+1; i++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
	    CurveTangentAt( vNormal, fU, fV, FALSE ); 
	    ASSERT(vNormal.Normalize());
	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p1.Get());
	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p2.Get());
	    
	    fU += fUDelta;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = 0;
	fV = GetTv();
	fU = GetTu();
	fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;

	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
	    CurveTangentAt( vNormal, fU, fV, TRUE );
	    ASSERT(vNormal.Normalize());

	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p1.Get());

	    gilNormal3fv((-vNormal).Get());
	    gilVertex3fv(p2.Get());

	    fV += fVDelta;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = m_uiUSlices;
	fV = GetTv();
	fU = GetTu() + GetTuDelta();
	fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;
	    
	for( j = 0; j < m_uiVSlices+1; j++ ) 
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
	    CurveTangentAt( vNormal, fU, fV, TRUE );
	    ASSERT(vNormal.Normalize());

	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p2.Get());
	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p1.Get());

	    fV += fVDelta;
	  }
      }
      gilEnd();
    }
}

VOID CSLIDEBSplinePatch::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  if( m_dyniDrawControls.GetValue() != 0 )
    {
      RenderControls();
    }

  if( m_dynfOffset.GetValue() != 0.0 )
    {
      RenderOffsetGeometryFlat();
      return;
    }

  UINT_32 i, j;
  Point p1, p2, p3, p4;
  Vector vNormal;
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvU[2][2];
  Vector pvV[2];
  
  pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
  pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
  
  ppvU[0][0] = m_pvTexture[3];
  ppvU[0][1] = m_pvTexture[0];
  ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
  ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
  
  for( i = 0; i < m_uiUSlices; i++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	pvV[0] = ppvU[0][0];
	pvV[1] = ppvU[1][0];
	
	p1 = m_ppSurfacePoint[i][0];
	p2 = m_ppSurfacePoint[i+1][0];
	
	gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	gilVertex3fv(p1.Get());
	gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	gilVertex3fv(p2.Get());
	
	pvV[0] += pvVDelta[0];
	pvV[1] += pvVDelta[1];
	
	for( j = 1; j < m_uiVSlices+1; j++)	
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = m_ppSurfacePoint[i+1][j];

	    vNormal = (p2 - p1) * (p3 - p1);
	    gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p2 - p3) * (p4 - p3);
	    gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	    
	    pvV[0] += pvVDelta[0];
	    pvV[1] += pvVDelta[1];
	  }
      }
      gilEnd();
      
      ppvU[0][0] += pvUDelta[0];
      ppvU[0][1] += pvUDelta[1];
      ppvU[1][0] += pvUDelta[0];
      ppvU[1][1] += pvUDelta[1];
    }
}

VOID CSLIDEBSplinePatch::RenderOffsetGeometryFlat()
{
  UINT_32 i, j;
  Point p1, p2, p3, p4, pt;
  Vector vNormal, v1, v2, v3, v4;
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvU[2][2];
  Vector pvV[2];
  FLOAT fOffset;

  fOffset = m_dynfOffset.GetValue();

  pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
  pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
  
  // Correct the orientation
  if( fOffset < 0.0 )
    {
      // Original Surface
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    p1 = m_ppSurfacePoint[i][0];
	    p2 = m_ppSurfacePoint[i+1][0];
	    
	    gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	    gilVertex3fv(p1.Get());
	    gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	    gilVertex3fv(p2.Get());
	    
	    pvV[0] += pvVDelta[0];
	    pvV[1] += pvVDelta[1];
	    
	    for( j = 1; j < m_uiVSlices+1; j++)	
	      {
		p3 = m_ppSurfacePoint[i][j];
		p4 = m_ppSurfacePoint[i+1][j];

		vNormal = (p2 - p1) * (p3 - p1);
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv(p3.Get());
		
		vNormal = (p2 - p3) * (p4 - p3);
		gilTexCoord2f(pvV[1][0], pvV[1][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv(p4.Get());
		
		p1 = p3;
		p2 = p4;
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}

      // Offset Surface with orienation reversed
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    p1 = m_ppSurfacePoint[i][0];
	    v1 = m_ppvSurfaceNormal[i][0];
	    p2 = m_ppSurfacePoint[i+1][0];
	    v2 = m_ppvSurfaceNormal[i+1][0];
	    
	    gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	    gilVertex3fv((p2 + v2*fOffset).Get());
	    gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	    gilVertex3fv((p1 + v1*fOffset).Get());
	    
	    pvV[0] += pvVDelta[0];
	    pvV[1] += pvVDelta[1];
	    
	    for( j = 1; j < m_uiVSlices+1; j++)	
	      {
		p3 = m_ppSurfacePoint[i][j];
		v3 = m_ppvSurfaceNormal[i][j];
		p4 = m_ppSurfacePoint[i+1][j];
		v4 = m_ppvSurfaceNormal[i+1][j];

		vNormal = (p2 - p3) * (p4 - p3);
		gilTexCoord2f(pvV[1][0], pvV[1][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv((p4 + v4*fOffset).Get());

		vNormal = (p2 - p1) * (p3 - p1);
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv((p3 + v3*fOffset).Get());
				
		p1 = p3;
		v1 = v3;
		p2 = p4;
		v2 = v4;
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }
  else
    {
      // Original Surface with orientation reversed
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    p1 = m_ppSurfacePoint[i][0];
	    p2 = m_ppSurfacePoint[i+1][0];
	    
	    gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	    gilVertex3fv(p2.Get());
	    gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	    gilVertex3fv(p1.Get());
	    
	    pvV[0] += pvVDelta[0];
	    pvV[1] += pvVDelta[1];
	    
	    for( j = 1; j < m_uiVSlices+1; j++)	
	      {
		p3 = m_ppSurfacePoint[i][j];
		p4 = m_ppSurfacePoint[i+1][j];

		vNormal = (p2 - p3) * (p4 - p3);
		gilTexCoord2f(pvV[1][0], pvV[1][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv(p4.Get());

		vNormal = (p2 - p1) * (p3 - p1);
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv(p3.Get());
				
		p1 = p3;
		p2 = p4;
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}

      // Offset Surface
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  gilBegin(GL_TRIANGLE_STRIP);
	  {
	    pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	    pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	    pvV[0] = ppvU[0][0];
	    pvV[1] = ppvU[1][0];
	    
	    p1 = m_ppSurfacePoint[i][0];
	    v1 = m_ppvSurfaceNormal[i][0];
	    p2 = m_ppSurfacePoint[i+1][0];
	    v2 = m_ppvSurfaceNormal[i+1][0];
	    
	    gilTexCoord2f(pvV[0][0], pvV[0][1]);	
	    gilVertex3fv((p1 + v1*fOffset).Get());
	    gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	    gilVertex3fv((p2 + v2*fOffset).Get());
	    
	    pvV[0] += pvVDelta[0];
	    pvV[1] += pvVDelta[1];
	    
	    for( j = 1; j < m_uiVSlices+1; j++)	
	      {
		p3 = m_ppSurfacePoint[i][j];
		v3 = m_ppvSurfaceNormal[i][j];
		p4 = m_ppSurfacePoint[i+1][j];
		v4 = m_ppvSurfaceNormal[i+1][j];

		vNormal = (p2 - p1) * (p3 - p1);
		gilTexCoord2f(pvV[0][0], pvV[0][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv((p3 + v3*fOffset).Get());
		
		vNormal = (p2 - p3) * (p4 - p3);
		gilTexCoord2f(pvV[1][0], pvV[1][1]);	
		gilNormal3fv(vNormal.Get());
		gilVertex3fv((p4 + v4*fOffset).Get());
		
		p1 = p3;
		v1 = v3;
		p2 = p4;
		v2 = v4;
		
		pvV[0] += pvVDelta[0];
		pvV[1] += pvVDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }

  // Connect the surface and offset surface
  if( m_dyniOffsetConnected.GetValue() != 0 ) 
    {
      RenderOffsetGeometryConnectedFlat();
    }
}

VOID CSLIDEBSplinePatch::RenderOffsetGeometryConnectedFlat()
{
  UINT_32 i, j;
  Point p1, p2, p3, p4;
  Vector vNormal;
  FLOAT fOffset;

  fOffset = m_dynfOffset.GetValue();

  // Correct the orientation
  if( fOffset < 0.0 )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = 0;
	p1 = m_ppSurfacePoint[0][j];
	p2 = p1 + m_ppvSurfaceNormal[0][j]*fOffset;
	
	gilVertex3fv(p1.Get());
	gilVertex3fv(p2.Get());
	
	for( i = 1; i < m_uiUSlices+1; i++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p2-p1)*(p3-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p4-p2)*(p3-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();

      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = m_uiVSlices;
	p1 = m_ppSurfacePoint[0][j];
	p2 = p1 + m_ppvSurfaceNormal[0][j]*fOffset;
	
	gilVertex3fv(p2.Get());
	gilVertex3fv(p1.Get());
	
	for( i = 1; i < m_uiUSlices+1; i++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p1-p2)*(p4-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    vNormal = (p3-p1)*(p4-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = 0;
	p1 = m_ppSurfacePoint[i][0];
	p2 = p1 + m_ppvSurfaceNormal[i][0]*fOffset;
	
	gilVertex3fv(p2.Get());
	gilVertex3fv(p1.Get());
	
	for( j = 1; j < m_uiVSlices+1; j++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;

	    vNormal = (p1-p2)*(p4-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    vNormal = (p3-p1)*(p4-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = m_uiUSlices;
	p1 = m_ppSurfacePoint[i][0];
	p2 = p1 + m_ppvSurfaceNormal[i][0]*fOffset;
	
	gilVertex3fv(p1.Get());
	gilVertex3fv(p2.Get());
	    
	for( j = 1; j < m_uiVSlices+1; j++ ) 
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p2-p1)*(p3-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p4-p2)*(p3-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
    }
  else
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = 0;
	p1 = m_ppSurfacePoint[0][j];
	p2 = p1 + m_ppvSurfaceNormal[0][j]*fOffset;
	
	gilVertex3fv(p2.Get());
	gilVertex3fv(p1.Get());
	
	for( i = 1; i < m_uiUSlices+1; i++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p1-p2)*(p4-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    vNormal = (p3-p1)*(p4-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	j = m_uiVSlices;
	p1 = m_ppSurfacePoint[0][j];
	p2 = p1 + m_ppvSurfaceNormal[0][j]*fOffset;
	
	gilVertex3fv(p1.Get());
	gilVertex3fv(p2.Get());
	
	for( i = 1; i < m_uiUSlices+1; i++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p2-p1)*(p3-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p4-p2)*(p3-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = 0;
	p1 = m_ppSurfacePoint[i][0];
	p2 = p1 + m_ppvSurfaceNormal[i][0]*fOffset;
	
	gilVertex3fv(p1.Get());
	gilVertex3fv(p2.Get());
	
	for( j = 1; j < m_uiVSlices+1; j++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p2-p1)*(p3-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p4-p2)*(p3-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	i = m_uiUSlices;
	p1 = m_ppSurfacePoint[i][0];
	p2 = p1 + m_ppvSurfaceNormal[i][0]*fOffset;
	
	gilVertex3fv(p2.Get());
	gilVertex3fv(p1.Get());
	
	for( j = 1; j < m_uiVSlices+1; j++ )
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = p3 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    vNormal = (p1-p2)*(p4-p2);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    vNormal = (p3-p1)*(p4-p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    p1 = p3;
	    p2 = p4;
	  }
      }
      gilEnd();
    }
}

VOID CSLIDEBSplinePatch::RenderNormalsSmooth()
{
  if( m_dynfOffset.GetValue() != 0.0 )
    {
      RenderOffsetNormalsSmooth();
      return;
    }

  UINT_32 i, j;
  Point pt, p1, p2, p3;
  
  gilBegin(GL_LINES);
  {
    for(i = 0; i < m_uiUSlices + 1; i++ ) 
      {
	for(j = 0; j < m_uiVSlices + 1; j++ )
	  {
	    gilVertex3fv(m_ppSurfacePoint[i][j].Get());
	    pt = m_ppSurfacePoint[i][j] + m_ppvSurfaceNormal[i][j] * m_fNormalLength;
	    gilVertex3fv(pt.Get());
	  }
      }
      }
  gilEnd();
}

VOID CSLIDEBSplinePatch::RenderOffsetNormalsSmooth()
{
  UINT_32 i, j;
  Point pt, p1, p2, p3;
  Vector vNormal;
  FLOAT fOffset, fNormalLength;
  
  fOffset = m_dynfOffset.GetValue();

  // Correct the orientation
  if( fOffset < 0.0 )
    {
      fNormalLength = m_fNormalLength;
    }
  else
    {
      fNormalLength = -m_fNormalLength;
    }

  gilBegin(GL_LINES);
  {
    for(i = 0; i < m_uiUSlices + 1; i++ ) 
      {
	for(j = 0; j < m_uiVSlices + 1; j++ )
	  {
	    pt = m_ppSurfacePoint[i][j];
	    vNormal = m_ppvSurfaceNormal[i][j];
	    
	    // Original Surface		
	    gilVertex3fv(pt.Get());
	    gilVertex3fv((pt + vNormal * fNormalLength).Get());
	    
	    // Offset Surface with orienation reversed
	    pt = m_ppSurfacePoint[i][j] + vNormal*fOffset;
	    gilVertex3fv(pt.Get());
	    gilVertex3fv((pt + vNormal * -fNormalLength).Get());
	  }
      }
  }
  gilEnd();

  // Connect the surface and offset surface
  if( m_dyniOffsetConnected.GetValue() != 0 ) 
    {
      Point p1, p2;
      FLOAT fV, fU, fUDelta, fVDelta;
      
      gilBegin(GL_LINES);
      {
	j = 0;
	fV = GetTv();
	fU = GetTu();
	fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;
	
	for( i = 0; i < m_uiUSlices+1; i++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
	    CurveTangentAt( vNormal, fU, fV, FALSE ); 
	    ASSERT(vNormal.Normalize());
	    
	    gilVertex3fv(p1.Get());
	    gilVertex3fv((p1 + vNormal * -m_fNormalLength).Get());
	    
	    gilVertex3fv(p2.Get());
	    gilVertex3fv((p2 + vNormal * -m_fNormalLength).Get());
	    
	    fU += fUDelta;
	  }
	gilEnd();
      }
      
      gilBegin(GL_LINES);
      {
	j = m_uiVSlices;
	fV = GetTv() + GetTvDelta();
	fU = GetTu();
	fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;
	
	for( i = 0; i < m_uiUSlices+1; i++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
	    CurveTangentAt( vNormal, fU, fV, FALSE ); 
	    ASSERT(vNormal.Normalize());
	    
	    gilVertex3fv(p1.Get());
	    gilVertex3fv((p1 + vNormal * m_fNormalLength).Get());
	    
	    gilVertex3fv(p2.Get());
	    gilVertex3fv((p2 + vNormal * m_fNormalLength).Get());
	    
	    fU += fUDelta;
	  }
	gilEnd();
      }
      
      gilBegin(GL_LINES);
      {
	i = 0;
	fV = GetTv();
	fU = GetTu();
	fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;
	
	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
	    
	    ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
	    CurveTangentAt( vNormal, fU, fV, TRUE );
	    ASSERT(vNormal.Normalize());
	    
	    gilVertex3fv(p1.Get());
	    gilVertex3fv((p1 + vNormal * -m_fNormalLength).Get());
	    
	    gilVertex3fv(p2.Get());
	    gilVertex3fv((p2 + vNormal * -m_fNormalLength).Get());
	    
	    fV += fVDelta;
	  }
	gilEnd();
      }
      
      gilBegin(GL_LINES);
      {
	i = m_uiUSlices;
	fV = GetTv();
	fU = GetTu() + GetTuDelta();
	fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;

	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = p1 + m_ppvSurfaceNormal[i][j]*fOffset;
		
	    ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
	    CurveTangentAt( vNormal, fU, fV, TRUE );
	    ASSERT(vNormal.Normalize());
	    
	    gilVertex3fv(p1.Get());
	    gilVertex3fv((p1 + vNormal * m_fNormalLength).Get());
	    
	    gilVertex3fv(p2.Get());
	    gilVertex3fv((p2 + vNormal * m_fNormalLength).Get());
	    
	    fV += fVDelta;
	  }
	gilEnd();
      }
    }   
}

VOID CSLIDEBSplinePatch::RenderNormalsFlat()
{
  if( m_dynfOffset.GetValue() != 0.0 )
    {
      RenderOffsetNormalsFlat();
      return;
    }

  UINT_32 i, j;
  Point pt, p1, p2, p3, p4;
  Vector v;
  
  gilBegin( GL_LINES );
  {
    for(i = 0; i < m_uiUSlices; i++ ) 
      {
	for(j = 0; j < m_uiVSlices; j++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = m_ppSurfacePoint[i+1][j];
	    p3 = m_ppSurfacePoint[i][j+1];	  
	    p4 = m_ppSurfacePoint[i+1][j+1];
	    
	    pt = (p1 + p2 + p3) / 3.0;
	    gilVertex3fv(pt.Get());
	    
	    v = (p2 - p1) * (p3 - p1);
	    v.Normalize();
	    pt = pt + v * m_fNormalLength;
	    gilVertex3fv(pt.Get());
	    
	    ///
	    
	    pt = (p2 + p3 + p4) / 3.0;
	    gilVertex3fv(pt.Get());
	    
	    v = (p2 - p3) * (p4 - p3);
	    v.Normalize();
	    pt = pt + v * m_fNormalLength;
	    gilVertex3fv(pt.Get());
	  }
      }
  }
  gilEnd();
}

VOID CSLIDEBSplinePatch::RenderOffsetNormalsFlat()
{
  UINT_32 i, j;
  Point pt, p1, p2, p3, p4;
  Vector v;
  FLOAT fOffset, fNormalLength;
  
  fOffset = m_dynfOffset.GetValue();

  if( fOffset < 0.0 )
    {
      fNormalLength = m_fNormalLength;
    }
  else
    {
      fNormalLength = -m_fNormalLength;
    }
  
  gilBegin( GL_LINES );
  {
    for(i = 0; i < m_uiUSlices; i++ ) 
      {
	for(j = 0; j < m_uiVSlices; j++ )
	  {
	    // Original Surface	
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = m_ppSurfacePoint[i+1][j];
	    p3 = m_ppSurfacePoint[i][j+1];
	    p4 = m_ppSurfacePoint[i+1][j+1];
	    
	    pt = (p1 + p2 + p3) / 3.0;
	    gilVertex3fv(pt.Get());		
	    v = (p2 - p1) * (p3 - p1);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
		
	    pt = (p2 + p3 + p4) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p2 - p3) * (p4 - p3);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    // Offset Surface with orienation reversed
	    p1 += m_ppvSurfaceNormal[i][j]*fOffset;
	    p2 += m_ppvSurfaceNormal[i+1][j]*fOffset;
	    p3 += m_ppvSurfaceNormal[i][j+1]*fOffset;
	    p4 += m_ppvSurfaceNormal[i+1][j+1]*fOffset;
	    
	    pt = (p1 + p2 + p3) / 3.0;
	    gilVertex3fv(pt.Get());		
	    v = (p2 - p1) * (p3 - p1);
	    v.Normalize();
	    gilVertex3fv((pt + v * -fNormalLength).Get());
	    
	    pt = (p2 + p3 + p4) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p2 - p3) * (p4 - p3);
	    v.Normalize();
	    gilVertex3fv((pt + v * -fNormalLength).Get());
	  }
      }
  }
  gilEnd();
  
  // Connect the surface and offset surface
  if( m_dyniOffsetConnected.GetValue() != 0 ) 
    {
      UINT_32 i1, i2, j1, j2;
      Point p11, p12, p13, p14, p21, p22, p23, p24;
      
      gilBegin(GL_LINES);
      {
	j1 = 0;
	p11 = m_ppSurfacePoint[0][j1];
	p12 = p11 + m_ppvSurfaceNormal[0][j1]*fOffset;
	
	j2 = m_uiVSlices;
	p21 = m_ppSurfacePoint[0][j2];
	p22 = p21 + m_ppvSurfaceNormal[0][j2]*fOffset;
	
	for( i = 1; i < m_uiUSlices+1; i++ )
	  {
	    p13 = m_ppSurfacePoint[i][j1];
	    p14 = p13 + m_ppvSurfaceNormal[i][j1]*fOffset;
	    
	    pt = (p11 + p12 + p13) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p12-p11)*(p13-p11);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    pt = (p12 + p13 + p14) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p14-p12)*(p13-p12);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    p11 = p13;
	    p12 = p14;
	    
	    ///
	    
	    p23 = m_ppSurfacePoint[i][j2];
	    p24 = p23 + m_ppvSurfaceNormal[i][j2]*fOffset;
	    
	    pt = (p21 + p22 + p23) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p21-p22)*(p24-p22);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    pt = (p22 + p23 + p24) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p23-p21)*(p24-p21);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    p21 = p23;
	    p22 = p24;
	  }
      }
      gilEnd();
      
      gilBegin(GL_LINES);
      {
	i1 = 0;
	p11 = m_ppSurfacePoint[i1][0];
	p12 = p11 + m_ppvSurfaceNormal[i1][0]*fOffset;
	
	i2 = m_uiUSlices;
	p21 = m_ppSurfacePoint[i2][0];
	p22 = p21 + m_ppvSurfaceNormal[i2][0]*fOffset;
	
	for( j = 1; j < m_uiVSlices+1; j++ )
	  {
	    p13 = m_ppSurfacePoint[i1][j];
	    p14 = p13 + m_ppvSurfaceNormal[i1][j]*fOffset;
	    
	    pt = (p11 + p12 + p13) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p11-p12)*(p14-p12);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());

	    pt = (p12 + p13 + p14) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p13-p11)*(p14-p11);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    p11 = p13;
	    p12 = p14;
	    
	    ///
	    
	    p23 = m_ppSurfacePoint[i2][j];
	    p24 = p23 + m_ppvSurfaceNormal[i2][j]*fOffset;
	    
	    pt = (p21 + p22 + p23) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p22-p21)*(p23-p21);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    pt = (p22 + p23 + p24) / 3.0;
	    gilVertex3fv(pt.Get());
	    v = (p24-p22)*(p23-p22);
	    v.Normalize();
	    gilVertex3fv((pt + v * fNormalLength).Get());
	    
	    p21 = p23;
	    p22 = p24;
	  }
      }
      gilEnd();
    }
}

VOID CSLIDEBSplinePatch::RenderControls()
{
  UINT_32 i, j;

  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 1.0, 0.0);

    // Edges
    // Draw the control edges in U
    for(i = 0; i < m_uiUControlPoints; i++ ) 
      {
	gilBegin(GL_LINE_STRIP);
	{
	  for (j = 0; j < m_uiVControlPoints; j++ )
	    {
	      gilVertex3fv(m_ppControlPoint[i][j].Get());
	    }
	}
	gilEnd();
      }
    // Draw the control edges in V
    for (j = 0; j < m_uiVControlPoints; j++ )
      {
	gilBegin(GL_LINE_STRIP);
	{
	  for(i = 0; i < m_uiUControlPoints; i++ ) 
	    {
	      gilVertex3fv(m_ppControlPoint[i][j].Get());
	    }
      }
	gilEnd();
      }
    
    // Points
    gilBegin(GL_POINTS);
    {
      for(i = 0; i < m_uiUControlPoints; i++ ) 
	{
	  for (j = 0; j < m_uiVControlPoints; j++ )
	    {
	      gilVertex3fv(m_ppControlPoint[i][j].Get());
	    }
	}
    }
    gilEnd();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEBSplinePatch::WriteSLIDE(FILE *fpOutputFile)
{
  UINT_32 i, j;
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtBSplinePatch], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUOrder],
	  m_uiUOrder);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUKnots],
	  m_uiUKnots);
  fprintf(fpOutputFile, "  %s (",
	  gcppcSLIDEReservedWords[sliderwtUKnotList]);

  for( i = 0; i < m_uiUKnots; i++ )
    {
      fprintf(fpOutputFile, " %.3f", m_pfUKnot[i]);
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVOrder],
	  m_uiVOrder);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVKnots],
	  m_uiVKnots);
  fprintf(fpOutputFile, "  %s (",
	  gcppcSLIDEReservedWords[sliderwtVKnotList]);

  for( i = 0; i < m_uiVKnots; i++ )
    {
      fprintf(fpOutputFile, " %.3f", m_pfVKnot[i]);
    }
  fprintf(fpOutputFile, " )\n");


  if( m_plistControlPointIDs != NULL )
    {
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      iter.Init(m_plistControlPointIDs);
      
      for ( pcID = (CHAR*)(iter.PeekFirst()), i = 0;
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()) )
	{
	  if( i == m_uiUControlPoints )
	    {
	      fprintf(fpOutputFile, "\n\t\t    " );
	      i = 0;
	    }
	  fprintf(fpOutputFile, " %s", pcID );
	  i++;
	}
    }
  else
    {
      const FLOAT* pt;

      fprintf(fpOutputFile, "# Default Control Points:\n");
      fprintf(fpOutputFile, "# %s (\n#\t",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
	    {
	      pt = m_ppControlPoint[i][m_uiVControlPoints-1-j].Get();
	      fprintf(fpOutputFile, "(%.3f %.3f %.3f)", pt[X], pt[Y], pt[Z] );

	      if( (j != m_uiVControlPoints-1) && ( i == m_uiUControlPoints-1) )
		{
		  fprintf(fpOutputFile, "\n#\t" );
		}
	    }
	}
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUSlices],
	  m_uiUSlices);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVSlices],
	  m_uiVSlices);
  fprintf(fpOutputFile, "  %s (%.3f %.3f) (%.3f %.3f) (%.3f %.3f) (%.3f %.3f)\n",
	  gcppcSLIDEReservedWords[sliderwtTexture],
	  m_pvTexture[0][X], m_pvTexture[0][Y],
	  m_pvTexture[1][X], m_pvTexture[1][Y],
	  m_pvTexture[2][X], m_pvTexture[2][Y],
	  m_pvTexture[3][X], m_pvTexture[3][Y]);
  fprintf(fpOutputFile, "  %s %.3f %ld\n",
	  gcppcSLIDEReservedWords[sliderwtOffset],
	  m_dynfOffset.GetValue(), m_dyniOffsetConnected.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawControls],
	  m_dyniDrawControls.GetValue());
  
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndBSplinePatch]);
}

VOID CSLIDEBSplinePatch::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  UINT_32 i, j;

  if( (m_uiUOrder == 4) && (m_uiVOrder == 4) )
    {
      const FLOAT *pt;
      char *pointIds = new char[800];
      char *thisPoint = new char[50];

      RiArchiveRecord(RI_COMMENT, "\nBasis \"b-spline\" 1 \"b-spline\" 1\n");

      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
      	    {
	      pt = m_ppControlPoint[i][j].Get();
	      sprintf( thisPoint, " %.6f %.6f %.6f", pt[X], pt[Y], pt[Z] );
	      strcat( pointIds, thisPoint );
	    }
	}

      RiArchiveRecord(RI_COMMENT, "\nPatch \"bicubic\" \"P\" [%s ] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f]\n", 
      		      pointIds,
		      m_pvTexture[3][X], m_pvTexture[3][Y], 
		      m_pvTexture[2][X], m_pvTexture[2][Y],
		      m_pvTexture[0][X], m_pvTexture[0][Y], 
		      m_pvTexture[1][X], m_pvTexture[1][Y]);  

      delete[] pointIds;
      delete[] thisPoint;
    }
  else 
    {
      const FLOAT *p1, *p2, *p3, *p4, *v1, *v2, *v3, *v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvU[2][2];
      Vector pvV[2];

      pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
      pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
      
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	  pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	  pvV[0] = ppvU[0][0];
	  pvV[1] = ppvU[1][0];

	  for( j = 0; j < m_uiVSlices; j++)
	    {
	      p1 = m_ppSurfacePoint[i][j].Get();
	      p2 = m_ppSurfacePoint[i+1][j].Get();
	      p3 = m_ppSurfacePoint[i][j+1].Get();
	      p4 = m_ppSurfacePoint[i+1][j+1].Get();
	      
	      if( (eShading == slideshtSLF_GOURAUD) || (eShading == slideshtSLF_PHONG) )
		{
		  v1 = m_ppvSurfaceNormal[i][j].Get();
		  v2 = m_ppvSurfaceNormal[i+1][j].Get();
		  v3 = m_ppvSurfaceNormal[i][j+1].Get();
		  v4 = m_ppvSurfaceNormal[i+1][j+1].Get();

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], p2[X], p2[Y], p2[Z], p3[X], p3[Y], p3[Z],
				  v1[X], v1[Y], v1[Z], v2[X], v2[Y], v2[Z], v3[X], v3[Y], v3[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1], 
				  (pvV[0] + pvVDelta[0])[0], (pvV[0] + pvVDelta[0])[1]);

		  pvV[0] += pvVDelta[0];

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], p2[X], p2[Y], p2[Z], p4[X], p4[Y], p4[Z],
				  v3[X], v3[Y], v3[Z], v2[X], v2[Y], v2[Z], v4[X], v4[Y], v4[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1],
				  (pvV[1] + pvVDelta[1])[0], (pvV[1] + pvVDelta[1])[1]);
		  pvV[1] += pvVDelta[1];
		}
	      else
		{
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], p2[X], p2[Y], p2[Z], p3[X], p3[Y], p3[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1], 
				  (pvV[0] + pvVDelta[0])[0], (pvV[0] + pvVDelta[0])[1]);
		  pvV[0] += pvVDelta[0];

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], p2[X], p2[Y], p2[Z], p4[X], p4[Y], p4[Z],
				   pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1],
				  (pvV[1] + pvVDelta[1])[0], (pvV[1] + pvVDelta[1])[1]);
		  pvV[1] += pvVDelta[1];
		}
	    }
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }
}

VOID CSLIDEBSplinePatch::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE BSplinePatch: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEBSplinePatch::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE BSplinePatch: %s", m_pcGUIID);
}

VOID CSLIDEBSplinePatch::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
					   CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: BSPLINEPATCH\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDEBSplinePatch::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  if( m_dynfOffset.GetValue() != 0.0 )
    {
      fprintf(fpOutputFile, "CSLIDEBSplinePatch::WriteSTLGeometry needs to be written for offset surfaces\n" );
    }
  else if ( mTransform.Determinant() < 0 ) // Need to Mirror triangles
    {
      UINT_32 i, j;
      Point p1, p2, p3, p4;
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  p1 = m_ppSurfacePoint[i][0];
	  p2 = m_ppSurfacePoint[i+1][0];
	  
	  p1 = mTransform * p1;
	  p2 = mTransform * p2;
	  
	  for( j = 1; j < m_uiVSlices+1; j++)	
	    {
	      p3 = m_ppSurfacePoint[i][j];
	      p4 = m_ppSurfacePoint[i+1][j];
	      
	      p3 = mTransform * p3;
	      p4 = mTransform * p4;
	      
	      WriteSTLTriangle(fpOutputFile, p3, p2, p1);
	      WriteSTLTriangle(fpOutputFile, p2, p3, p4);

	      p1 = p3;
	      p2 = p4;
	    }
	}
    }    
  else // Don't mirror triangles
    {
      UINT_32 i, j;
      Point p1, p2, p3, p4;
      Vector vNormal;

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  p1 = m_ppSurfacePoint[i][0];
	  p2 = m_ppSurfacePoint[i+1][0];

	  p1 = mTransform * p1;
	  p2 = mTransform * p2;
	
	  for( j = 1; j < m_uiVSlices+1; j++)	
	    {
	      p3 = m_ppSurfacePoint[i][j];
	      p4 = m_ppSurfacePoint[i+1][j];

	      p3 = mTransform * p3;
	      p4 = mTransform * p4;

	      WriteSTLTriangle(fpOutputFile, p1, p2, p3);
	      WriteSTLTriangle(fpOutputFile, p4, p3, p2);

	      p1 = p3;
	      p2 = p4;
	    }
	}
    }
}







