#include "bound.h"
#include "slidecamera.h"

// Default Values
static const FLOAT cfNear = (FLOAT)-0.02;
static const FLOAT cfFar = (FLOAT)-100.0;

//////////////////////////////////////////////////////////////////////
// CSLIDECamera Class
//

CSLIDECamera::CSLIDECamera()
{
  m_dyniProjection.SetValue(slidepjtSLF_PARALLEL);
  m_vFrustumMin.Null();
  m_vFrustumMax.Null();
}

CSLIDECamera::~CSLIDECamera()
{
}

VOID CSLIDECamera::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniProjection.SetValue(slidepjtSLF_PARALLEL);
  m_vFrustumMin.Set((FLOAT)-1.0, (FLOAT)-1.0, cfFar);
  m_vFrustumMax.Set((FLOAT)1.0, (FLOAT)1.0, cfNear);

  m_fVRPDepth = -1.0f; /* cfFar + cfNear / 2.0; */
}

VOID CSLIDECamera::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDECamera::Initialized()
{
  return CSLIDEPrimitive::Initialized();
}

BOOL CSLIDECamera::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  return CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld);
}

BOOL CSLIDECamera::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( 0 <= m_dyniProjection.GetValue() &&
	   m_dyniProjection.GetValue() < slidepjtSize);
}

CSLIDECamera *CSLIDECamera::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  CHAR *pcID;
  CSLIDECamera *pCSLIDECamera;

  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  pCSLIDECamera = this;

  pcID = (CHAR*)piterIDs->PeekNext();
  if ( pcID == NULL )
    {
      pcID = (CHAR*)piterIDs->PeekFirst();
      if ( pcID != GetID() )
	{
	  HandleError("Bad %s path: %s id \"%s\" not specified", 
		      gcppcSLIDEReservedWords[sliderwtCamera],
		      gcppcSLIDEReservedWords[sliderwtCamera], GetID());
	  pCSLIDECamera = NULL;
	}
    }
  else if ( pcID != GetID() )
    {
      HandleError("Bad %s path: path id \"%s\" != %s id \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtCamera], pcID,
		  gcppcSLIDEReservedWords[sliderwtCamera], GetID());
      pCSLIDECamera = NULL;
    }

  return pCSLIDECamera;
}

CSLIDELight *CSLIDECamera::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtCamera], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDECamera::FindAreaLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtCamera], GetID());

  return NULL;
}

VOID CSLIDECamera::CalculateZoom(Matrix &mGzoom_vrc,
                                 Matrix &mGvrc_zoom,
                                 const Matrix &mGvrc_obj,
                                 const CBound &bObject)
{
  Vector vBMin, vBMax;
  FLOAT fBDepth;
  Vector vFMin, vFMax;

  /* Initialize */
  mGzoom_vrc.SetIdentity();
  mGvrc_zoom.SetIdentity();

  {
    CBound bFocus;

    /* Calculate axis aligned bound of object */
    bFocus.Null();
    bFocus.Insert(mGvrc_obj, bObject);

    /* Calculate bound's near face */
    vBMax.Set(bFocus.Get(CBound::bbtMax)(X), 
      bFocus.Get(CBound::bbtMax)(Y), bFocus.Get(CBound::bbtMax)(Z));
    vBMin.Set(bFocus.Get(CBound::bbtMin)(X), 
      bFocus.Get(CBound::bbtMin)(Y), bFocus.Get(CBound::bbtMax)(Z));
    fBDepth = bFocus.Get(CBound::bbtMax)(Z) - bFocus.Get(CBound::bbtMin)(Z);
  }

  {
    /* Calculate frustum bound where object will be placed */
    /* Try 5% back from ZMax */
    vFMax.Set(m_vFrustumMax.Get());
    vFMin.Set(m_vFrustumMin.Get());
    vFMax[Z] = vFMin[Z] = vFMax(Z) - 0.05f * (vFMax(Z) - vFMin(Z));

    switch ( m_dyniProjection.GetValue() )
    {
      case slidepjtSLF_PARALLEL:
        { /* Nothing new to do */ }
        break;
      case slidepjtSLF_PERSPECTIVE:
        {
          /* Scale area of frustum to Z depth */
          vFMax[X] = - vFMax(Z) * vFMax(X);
          vFMax[Y] = - vFMax(Z) * vFMax(Y);
          vFMin[X] = - vFMin(Z) * vFMin(X);
          vFMin[Y] = - vFMin(Z) * vFMin(Y);
        }
        break;
      default:
        {}
        break;
    } // End switch
  }

  /* Translate near face of bound to camera origin */
  {
    Vector vTrans;
    Matrix mTrans;

    vTrans = (vBMin + vBMax) / 2;
    mTrans.SetTranslate(-vTrans);
    mGzoom_vrc = mTrans * mGzoom_vrc;
    mTrans.SetTranslate(vTrans);
    mGvrc_zoom = mGvrc_zoom * mTrans;
  }

  /* Scale such that the bound fits in 95% of frustum */
  {
    FLOAT fScaleX, fScaleY, fScale;
    Matrix mScale;

    fScaleX = (vFMax(X)-vFMin(X))/(vBMax(X)-vBMin(X));
    fScaleY = (vFMax(Y)-vFMin(Y))/(vBMax(Y)-vBMin(Y));
    fScale = .95f * MIN( fScaleX, fScaleY );

    mScale.SetScale(fScale);
    mGzoom_vrc = mScale * mGzoom_vrc;
    mScale.SetScale(1/fScale);
    mGvrc_zoom = mGvrc_zoom * mScale;

    fBDepth = fBDepth * fScale;
  }

  /* Translate the face to frustum */
  {
    Vector vTrans;
    Matrix mTrans;

    vTrans = (vFMin + vFMax) / 2;
    mTrans.SetTranslate(vTrans);
    mGzoom_vrc = mTrans * mGzoom_vrc;
    mTrans.SetTranslate(-vTrans);
    mGvrc_zoom = mGvrc_zoom * mTrans;
    m_fVRPDepth = vTrans(Z) - (fBDepth/2.0);
    if (m_vFrustumMin(Z) > m_fVRPDepth)
      {
	m_fVRPDepth = (m_vFrustumMax(Z) + m_vFrustumMin(Z)) / 2.0;
      }
  }
}

VOID CSLIDECamera::CheckDynamic()
{
  SetDynamic(m_dyniProjection.GetDynamic() ||
	     m_vFrustumMin.GetDynamic() ||
	     m_vFrustumMax.GetDynamic());
}

VOID CSLIDECamera::UpdateState()
{
  m_dyniProjection.Update();
  if ( !( 0 <= m_dyniProjection.GetValue() &&
	  m_dyniProjection.GetValue() < slidepjtSize ) )
    {
      HandleError("Illegal Projection flag value:  %ld", 
		  m_dyniProjection.GetValue());
      m_dyniProjection.SetValue(slidepjtSLF_PARALLEL);
    }

  m_vFrustumMin.Update();
  m_vFrustumMax.Update();

  if (m_vFrustumMax.GetDynamic() ||
      m_vFrustumMin.GetDynamic() )
    {
      //      m_fVRPDepth = -1.0f; 
/*      m_fVRPDepth = m_vFrustumMax(Z) + m_vFrustumMin(Z) / 2.0; */
    }

  CalculateProjectionMatrix();
}

VOID CSLIDECamera::CalculateProjectionMatrix()
{
  Vector vDOP, vScale;
  Matrix mShear, mScale;
  Matrix mShearInv, mScaleInv;

  m_mGproj_vrc.SetIdentity();
  m_mGvrc_proj.SetIdentity();

  vDOP = (m_vFrustumMin + m_vFrustumMax) / 2;
  vDOP[Z] = 1.0;
  mShear.SetShearXY(vDOP);
  vDOP[Z] = -1.0;
  mShearInv.SetShearXY(vDOP);

  switch ( m_dyniProjection.GetValue() )
    {
    case slidepjtSLF_PARALLEL:
      {
	Vector vTrans;
	Matrix mTrans, mTransInv;

	vTrans.Set(0.0, 0.0, -m_vFrustumMax[Z]);
	mTrans.SetTranslate(vTrans);
	mTransInv.SetTranslate(-vTrans);

	vScale = (m_vFrustumMax - m_vFrustumMin);
	vScale[X] = vScale[X] / (FLOAT)2.0;
	vScale[Y] = vScale[Y] / (FLOAT)2.0;
	mScaleInv.SetScale(vScale);
	vScale = 1.0 / vScale;
	mScale.SetScale(vScale);

	m_mGproj_vrc = mScale * mTrans * mShear;
	m_mGvrc_proj = mShearInv * mTransInv * mScaleInv;	
      }
      break;
    case slidepjtSLF_PERSPECTIVE:
      {
	Matrix mWarp, mWarpInv;
	FLOAT fZfront;

	vScale = (m_vFrustumMax - m_vFrustumMin) / 2.0;
	vScale[Z] = 1.0;
	vScale = vScale * (-m_vFrustumMin[Z]);
	mScaleInv.SetScale(vScale);
	vScale = 1.0 / vScale;
	mScale.SetScale(vScale);

	ASSERT( m_vFrustumMin[Z] < 0 );
	ASSERT( m_vFrustumMax[Z] < 0 );
	fZfront = -m_vFrustumMax[Z] / m_vFrustumMin[Z];
	mWarp.SetPerspectiveWarp(fZfront);
	mWarpInv.SetPerspectiveWarpInverse(fZfront);

	m_mGproj_vrc = mWarp * mScale * mShear;
	m_mGvrc_proj = mShearInv * mScaleInv * mWarpInv;		
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    };
}

VOID CSLIDECamera::ApplyProjectionMatrix(BOOL bSelectionMode, FLOAT x, FLOAT y)
{
  gilMatrixMode(GL_PROJECTION);
  gilLoadIdentity();

  if( bSelectionMode ) 
    {
      GLint viewport[4];
      gilGetIntegerv(GL_VIEWPORT, viewport);
      gluPickMatrix((GLdouble)x, (GLdouble)y, 1.0, 1.0, viewport);
    }

/*  
  Vector vDOP, vScale;
  Matrix mShear, mScale;
  Matrix mShearInv, mScaleInv;

  m_mGproj_vrc.SetIdentity();
  m_mGvrc_proj.SetIdentity();

  vDOP = (m_vFrustumMin + m_vFrustumMax) / 2;
  vDOP[Z] = 1.0;
  mShear.SetShearXY(vDOP);
  vDOP[Z] = -1.0;
  mShearInv.SetShearXY(vDOP);
*/

  switch ( m_dyniProjection.GetValue() )
    {
    case slidepjtSLF_PARALLEL:
      {
	Vector vDOP, vMin, vMax;
	Matrix mShear;

	vDOP = (m_vFrustumMin + m_vFrustumMax) / 2;
	vDOP[Z] = 1.0;
	mShear.SetShearXY(vDOP);

	//gilMultMatrixf(mShear.Get());

	/*
	vMin = m_vFrustumMin - vDOP;
	vMax = m_vFrustumMax - vDOP;

	// Using glOrtho with Oblique Shearing
	gilOrtho(vMin[X], vMax[X],
		 vMin[Y], vMax[Y],
		 -m_vFrustumMax[Z], -m_vFrustumMin[Z]);
	*/

	Vector vScale, vTrans;
	Matrix mScale, mTrans, mGproj_vrc;

	vTrans.Set((FLOAT)0.0, (FLOAT)0.0, (m_vFrustumMin[Z] + m_vFrustumMax[Z]) / (FLOAT)-2.0);
	mTrans.SetTranslate(vTrans);

	vScale = (m_vFrustumMax - m_vFrustumMin) / 2.0;
	//vScale[X] = vScale[X] / 2.0;
	//vScale[Y] = vScale[Y] / 2.0;
	//mScaleInv.SetScale(vScale);
	vScale = 1.0 / vScale;
	vScale[Z] = -vScale[Z]; // OpenGL wants the world turned around
	mScale.SetScale(vScale);

	mGproj_vrc = mScale * mTrans * mShear;

	gilMultMatrixf(mGproj_vrc.Get());

	/*
	// No Oblique Projections
	gilOrtho(m_vFrustumMin[X], m_vFrustumMax[X],
		 m_vFrustumMin[Y], m_vFrustumMax[Y],
		 -m_vFrustumMax[Z], -m_vFrustumMin[Z]);
	*/

	/*
	//gilOrtho(-1, 1, -1, 1, 0, 1);

	Vector vTrans;
	Matrix mTrans, mTransInv;

	vTrans.Set(0.0, 0.0, -m_vFrustumMax[Z]);
	mTrans.SetTranslate(vTrans);
	mTransInv.SetTranslate(-vTrans);

	vScale = (m_vFrustumMax - m_vFrustumMin);
	vScale[X] = vScale[X] / 2.0;
	vScale[Y] = vScale[Y] / 2.0;
	mScaleInv.SetScale(vScale);
	vScale = 1.0 / vScale;
	mScale.SetScale(vScale);

	m_mGproj_vrc = mScale * mTrans * mShear;
	m_mGvrc_proj = mShearInv * mTransInv * mScaleInv;	
	*/
      }
      break;
    case slidepjtSLF_PERSPECTIVE:
      {
	gilFrustum(-m_vFrustumMax[Z]*m_vFrustumMin[X], 
		   -m_vFrustumMax[Z]*m_vFrustumMax[X],
		   -m_vFrustumMax[Z]*m_vFrustumMin[Y], 
		   -m_vFrustumMax[Z]*m_vFrustumMax[Y],
		   -m_vFrustumMax[Z], -m_vFrustumMin[Z]);
	//gilOrtho(-1, 1, -1, 1, 0, 1);
	/*
	Matrix mWarp, mWarpInv;
	FLOAT fZfront;

	vScale = (m_vFrustumMax - m_vFrustumMin) / 2.0;
	vScale[Z] = 1.0;
	vScale = vScale * (-m_vFrustumMin[Z]);
	mScaleInv.SetScale(vScale);
	vScale = 1.0 / vScale;
	mScale.SetScale(vScale);

	ASSERT( m_vFrustumMin[Z] < 0 );
	ASSERT( m_vFrustumMax[Z] < 0 );
	fZfront = -m_vFrustumMax[Z] / m_vFrustumMin[Z];
	mWarp.SetPerspectiveWarp(fZfront);
	mWarpInv.SetPerspectiveWarpInverse(fZfront);

	m_mGproj_vrc = mWarp * mScale * mShear;
	m_mGvrc_proj = mShearInv * mScaleInv * mWarpInv;		
	*/
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    };

}

VOID CSLIDECamera::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n",
	  gcppcSLIDEReservedWords[sliderwtCamera], m_pcGUIID);
  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtProjection], 
	  gcppcSLIDEProjectionReservedWords[m_dyniProjection.GetValue()]
	  ); 
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtFrustum], 
	  m_vFrustumMin[X], m_vFrustumMin[Y], m_vFrustumMin[Z],
	  m_vFrustumMax[X], m_vFrustumMax[Y], m_vFrustumMax[Z]);

  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n",
	  gcppcSLIDEReservedWords[sliderwtEndCamera]);
}

VOID CSLIDECamera::WriteRIB(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  // Set up the correct type of projection

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE CAMERA: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);

  switch ( m_dyniProjection.GetValue() )
    {
    case slidepjtSLF_PARALLEL:
      {
	RiProjection(RI_ORTHOGRAPHIC, RI_NULL);
      }
      break;
    case slidepjtSLF_PERSPECTIVE:
      {
	RiProjection(RI_PERSPECTIVE, RI_NULL);
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  RiScreenWindow(m_vFrustumMin[0], m_vFrustumMax[0], 
		 m_vFrustumMin[1], m_vFrustumMax[1]);
  RiClipping(-m_vFrustumMax[2], -m_vFrustumMin[2]);

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE CAMERA: %s", m_pcGUIID);
}

VOID CSLIDECamera::WritePOV(FILE *fpOutputFile, Matrix mTransform, const Vector &vVPSize)
{
  ASSERT( fpOutputFile != NULL );

  FLOAT fPx, fPy;
  FLOAT fAngle;
  Vector vDir, vRight, vUp;
  Vector vResultDir, vResultRight, vResultUp;
  Point ptLoc, ptResultLoc;

  fPx = (m_vFrustumMax[X]-m_vFrustumMin[X])/2;
  fPy = (m_vFrustumMax[Y]-m_vFrustumMin[Y])/2;
  fAngle = RADTODEG(atan(fPx));

  ptLoc.Set(0, 0, 0);
  vDir.Set(m_vFrustumMax[X]-fPx, m_vFrustumMax[Y]-fPy, -1);
  vRight.Set(2*fPx*(vVPSize.Get())[X], 0, 0);
  vUp.Set(0, 2*fPy*(vVPSize.Get())[Y], 0);

  ptResultLoc = mTransform * ptLoc;
  vResultDir = mTransform * vDir;
  vResultRight = mTransform * vRight;
  vResultUp = mTransform * vUp;  

  fprintf(fpOutputFile, "camera { \n");

  switch ( m_dyniProjection.GetValue() )
    {
    case slidepjtSLF_PARALLEL:
      {
	fprintf(fpOutputFile, " orthographic\n");
      }
      break;
    case slidepjtSLF_PERSPECTIVE:
      {
	fprintf(fpOutputFile, " perspective\n");
      }
      break;
    default:
      ASSERT( FALSE );
      break;
    }
  fprintf(fpOutputFile, " location  < %.3f, %.3f, %.3f >\n",
	  ptResultLoc[X], ptResultLoc[Y], ptResultLoc[Z]);
  fprintf(fpOutputFile, " direction < %.3f, %.3f, %.3f >\n",
	  vResultDir[X], vResultDir[Y], vResultDir[Z]);
  fprintf(fpOutputFile, " right     < %.3f, %.3f, %.3f >\n",
	  vResultRight[X], vResultRight[Y], vResultRight[Z]);
  fprintf(fpOutputFile, " up        < %.3f, %.3f, %.3f >\n",
	  vResultUp[X], vResultUp[Y], vResultUp[Z]);
  fprintf(fpOutputFile, " angle     %.3f \n", 2*fAngle);
  fprintf(fpOutputFile, "}\n\n");
}

VOID CSLIDECamera::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "camera_%s", m_pcGUIID);
}

VOID CSLIDECamera::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = diamond ];\n", GetGUIID());
}

VOID CSLIDECamera::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

//////////////////////////////////////////////////////////////////////
// CSLIDECameraPath Class
//

CSLIDECameraPath::CSLIDECameraPath()
{
}

CSLIDECameraPath::~CSLIDECameraPath()
{
}

CSLIDENode *CSLIDECameraPath::FindNode(CSLIDENode *pRoot, CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( pRoot != NULL );
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  return pRoot->FindCamera(piterIDs, plistCSLIDEInstance);
}

