#include "slidecyclide.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "qemesh.h"
#include "qevertex.h"
#include "qewrite.h"
#include "stl.h"

static const FLOAT cfXMajorA = 0.5;
static const FLOAT cfYMinorB = 0.4;
static const FLOAT cfStringLength = 0.8;
static const FLOAT cfPhiMin = DEGTORAD(0.0);
static const FLOAT cfPhiMax = DEGTORAD(360.0);
static const FLOAT cfThetaMin = DEGTORAD(0.0);
static const FLOAT cfThetaMax = DEGTORAD(360.0);
static const UINT_32 cuiPhiSlices = 10;
static const UINT_32 cuiThetaSlices = 10;
static const UINT_32 cuiDrawEllipse = 0;
static const UINT_32 cuiDrawHyperbola = 0;

//////////////////////////////////////////////////////////////////////
// CSLIDECyclide Class
//

CSLIDECyclide::CSLIDECyclide()
{
  m_dynfXMajorA.SetValue(cfXMajorA);
  m_dynfYMinorB.SetValue(cfYMinorB);
  m_dynfStringLength.SetValue(cfStringLength);
  m_dynfPhiMin.SetValue(cfPhiMin);
  m_dynfPhiMax.SetValue(cfPhiMax);
  m_dynfThetaMin.SetValue(cfThetaMin);
  m_dynfThetaMax.SetValue(cfThetaMax);
  m_dyniPhiSlices.SetValue(cuiPhiSlices);
  m_dyniThetaSlices.SetValue(cuiThetaSlices);
  m_dyniDrawEllipse.SetValue(cuiDrawEllipse);
  m_dyniDrawHyperbola.SetValue(cuiDrawHyperbola);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

CSLIDECyclide::~CSLIDECyclide()
{
}

VOID CSLIDECyclide::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dynfXMajorA.SetValue(cfXMajorA);
  m_dynfYMinorB.SetValue(cfYMinorB);
  m_dynfStringLength.SetValue(cfStringLength);
  m_dynfPhiMin.SetValue(cfPhiMin);
  m_dynfPhiMax.SetValue(cfPhiMax);
  m_dynfThetaMin.SetValue(cfThetaMin);
  m_dynfThetaMax.SetValue(cfThetaMax);
  m_dyniPhiSlices.SetValue(cuiPhiSlices);
  m_dyniThetaSlices.SetValue(cuiThetaSlices);
  m_dyniDrawEllipse.SetValue(cuiDrawEllipse);
  m_dyniDrawHyperbola.SetValue(cuiDrawHyperbola);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

VOID CSLIDECyclide::Uninit()
{
  CSLIDEAtomNode::Uninit();
}

BOOL CSLIDECyclide::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDECyclide::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;

  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dynfYMinorB.GetDynamic() && !m_dynfXMajorA.GetDynamic() &&
       !( m_dynfYMinorB.GetValue() <= m_dynfXMajorA.GetValue() ) )
    {
      HandleError("a (%.3f) must be <= b (%.3f)",
		  m_dynfYMinorB.GetValue(), m_dynfXMajorA.GetValue());
      m_dynfXMajorA.SetValue(cfXMajorA);
      bCorrect = FALSE;
    }

  if ( !m_dynfYMinorB.GetDynamic() && 
       !( 0.0 <= m_dynfYMinorB.GetValue() ) )
    {
      HandleError("b (%.3f) must be > 0.0", m_dynfYMinorB.GetValue());
      m_dynfYMinorB.SetValue(cfYMinorB);
      bCorrect = FALSE;
    }

  if ( !m_dynfStringLength.GetDynamic() && 
       !( 0.0 <= m_dynfStringLength.GetValue() ) )
    {
      HandleError("stringlength (%.3f) must be >= 0.0", m_dynfStringLength.GetValue());
      m_dynfStringLength.SetValue(cfStringLength);
      bCorrect = FALSE;
    }

  if ( !m_dynfPhiMin.GetDynamic() &&
       !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfPhiMin.GetValue() && 
	  m_dynfPhiMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
    {
      HandleError("Illegal phimin value:  %.3f <= phimin(%.3f) <= %.3f", 
		  0.0,
		  RADTODEG(m_dynfPhiMin.GetValue()),
		  720.0);
      m_dynfPhiMin.SetValue(cfPhiMin);
      bCorrect = FALSE;
    }

  if ( !m_dynfPhiMax.GetDynamic() && 
       !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfPhiMax.GetValue() && 
	  m_dynfPhiMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
    {
      HandleError("Illegal phimax value:  %.3f <= phimax(%.3f) <= %.3f", 
		  0.0,
		  RADTODEG(m_dynfPhiMax.GetValue()),
		  720.0);
      m_dynfPhiMax.SetValue(cfPhiMax);
      bCorrect = FALSE;
    }

  if ( !m_dynfThetaMin.GetDynamic() &&
       !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfThetaMin.GetValue() && 
	  m_dynfThetaMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
    {
      HandleError("Illegal thetamin value:  %.3f <= thetamin(%.3f) <= %.3f", 
		  0.0,
		  RADTODEG(m_dynfThetaMin.GetValue()),
		  720.0);
      m_dynfThetaMin.SetValue(cfThetaMin);
      bCorrect = FALSE;
    }

  if ( !m_dynfThetaMax.GetDynamic() &&
       !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfThetaMax.GetValue() && 
	  m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
    {
      HandleError("Illegal thetamax value:  %.3f <= thetamax(%.3f) <= %.3f", 
		  0.0,
		  RADTODEG(m_dynfThetaMax.GetValue()),
		  720.0);
      m_dynfThetaMax.SetValue(cfThetaMax);
      bCorrect = FALSE;
    }

  if ( !m_dyniPhiSlices.GetDynamic() && 
       !( 0 <= m_dyniPhiSlices.GetValue() ) )
    {
      HandleError("phislices (%ld) must be >= 0", m_dyniPhiSlices.GetValue());
      m_dyniPhiSlices.SetValue(cuiPhiSlices);
      bCorrect = FALSE;
    }

  if ( !m_dyniThetaSlices.GetDynamic() &&
       !( 0 <= m_dyniThetaSlices.GetValue() ) )
    {
      HandleError("thetaslices (%ld) must be >= 0", m_dyniThetaSlices.GetValue());
      m_dyniThetaSlices.SetValue(cuiThetaSlices);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDECyclide::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dynfYMinorB.GetValue() <= m_dynfXMajorA.GetValue() ) &&
	   ( 0.0 < m_dynfYMinorB.GetValue() ) &&
	   ( 0.0 < m_dynfStringLength.GetValue() ) &&
	   ( (FLOAT)(DEGTORAD(0.0)) <= m_dynfPhiMin.GetValue() && 
	     m_dynfPhiMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) &&
	   ( (FLOAT)(DEGTORAD(0.0)) <= m_dynfPhiMax.GetValue() && 
	     m_dynfPhiMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) &&
	   ( (FLOAT)(DEGTORAD(0.0)) <= m_dynfThetaMin.GetValue() && 
	     m_dynfThetaMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) &&
	   ( (FLOAT)(DEGTORAD(0.0)) <= m_dynfThetaMax.GetValue() && 
	     m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) &&
	   ( 0 < m_dyniPhiSlices.GetValue() ) &&
	   ( 0 < m_dyniThetaSlices.GetValue() ) );
}

VOID CSLIDECyclide::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			    const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse)
{
  ASSERT(uiSide < 4);

  FLOAT fNormalDir;                        // cyclide variables
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta, fPhi;
  Vector vPQ;
  Point pt, pZ;
  Point *ppt;
  Vector vNormal;
  Vector *pvNormal;

  // Cyclide
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;    
  fTheta = fThetaMin;

  if( (uiSide == 0) || (uiSide == 2) )
    {
      if( uiSide == 0 ) 
	{
	  fTheta = fThetaMin;
	}
      else
	{
	  fTheta = fThetaMax;
	}

      if( bReverse == 0 )
	{
	  fPhi = fPhiMin;
	}
      else
	{
	  fPhi = fPhiMax;
	  fPhiDelta *= -1.0;
	}

      for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	{
	  // p->q
	  CalculateZ( fTheta, fPhi, pZ, vPQ );

	  pt = pZ;
	  pt = mTransform * pt;	  
	  ppt = new Point;
	  ppt->Set(pt.Get());	  
	  plistPoint->InsertLast(ppt, 0);

	  vNormal = fNormalDir * vPQ;
	  vNormal = fNormalDir * vNormal;
	  vNormal.Normalize();
	  pvNormal = new Vector;
	  pvNormal->Set(vNormal.Get());
	  plistNormal->InsertLast(pvNormal, 0);
	  
	  fPhi += fPhiDelta;
	}
    }
  else // (uiSide == 1) || (uiSide == 3)  )
    {
      if( uiSide == 1 )
	{
	  fPhi = fPhiMax;
	}
      else
	{
	  fPhi = fPhiMin;
	}

      if( bReverse == 0 )
	{
	  fTheta = fThetaMin;
	}
      else
	{
	  fTheta = fThetaMax;
	  fThetaDelta *= -1.0;
	}

      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  // p->q
	  CalculateZ( fTheta, fPhi, pZ, vPQ );

	  pt = pZ;
	  pt = mTransform * pt;
	  ppt = new Point;
	  ppt->Set(pt.Get());
	  plistPoint->InsertLast(ppt, 0);

	  vNormal = fNormalDir * vPQ;
	  vNormal = fNormalDir * vNormal;
	  vNormal.Normalize();
	  pvNormal = new Vector;
	  pvNormal->Set(vNormal.Get());
	  plistNormal->InsertLast(pvNormal, 0);
	  
	  fTheta += fThetaDelta;
	}
    }
}

VOID CSLIDECyclide::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  SetDynamic(GetDynamic() ||
	     m_dynfXMajorA.GetDynamic() ||
	     m_dynfYMinorB.GetDynamic() ||
	     m_dynfStringLength.GetDynamic() ||
	     m_dynfPhiMin.GetDynamic() ||
	     m_dynfPhiMax.GetDynamic() ||
	     m_dynfThetaMin.GetDynamic() ||
	     m_dynfThetaMax.GetDynamic() ||
	     m_dyniPhiSlices.GetDynamic() ||
	     m_dyniThetaSlices.GetDynamic() ||
             m_dyniDrawEllipse.GetDynamic() ||
             m_dyniDrawHyperbola.GetDynamic() ||
	     m_pvTexture[0].GetDynamic() ||
	     m_pvTexture[1].GetDynamic() ||
	     m_pvTexture[2].GetDynamic() ||
	     m_pvTexture[3].GetDynamic());
}

VOID CSLIDECyclide::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  
  bRealSpaceDynamic = ( m_dynfXMajorA.GetDynamic() ||
			m_dynfYMinorB.GetDynamic() ||
			m_dynfStringLength.GetDynamic() ||
			m_dynfPhiMin.GetDynamic() ||
			m_dynfPhiMax.GetDynamic() ||
			m_dynfThetaMin.GetDynamic() ||
			m_dynfThetaMax.GetDynamic());
  bParameterSpaceDynamic = ( m_dyniPhiSlices.GetDynamic() ||
			     m_dyniThetaSlices.GetDynamic() );
  bTextureSpaceDynamic = ( m_pvTexture[0].GetDynamic() ||
			   m_pvTexture[1].GetDynamic() ||
			   m_pvTexture[2].GetDynamic() ||
			   m_pvTexture[3].GetDynamic() );

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || bParameterSpaceDynamic || bTextureSpaceDynamic);
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || bParameterSpaceDynamic);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic);
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic);
}

VOID CSLIDECyclide::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  UINT_32 i;

  m_dynfXMajorA.Update();
  if ( !( m_dynfYMinorB.GetValue() <= m_dynfXMajorA.GetValue() ) )
    {
      HandleError("a (%.3f) must be <= b (%.3f)",
		  m_dynfYMinorB.GetValue(), m_dynfXMajorA.GetValue());
      m_dynfXMajorA.SetValue(cfXMajorA);
    }

  m_dynfYMinorB.Update();
  if ( !( 0.0 < m_dynfYMinorB.GetValue() ) )
    {
      HandleError("b (%.3f) must be > 0.0",  m_dynfYMinorB.GetValue());
      m_dynfYMinorB.SetValue(cfYMinorB);
    }

  m_dynfStringLength.Update();
  if ( !( 0.0 <= m_dynfStringLength.GetValue() ) )
    {
      HandleError("stringlength (%.3f) must be >= 0.0", m_dynfStringLength.GetValue());
      m_dynfStringLength.SetValue(cfStringLength);
    }
  
  if ( m_dynfPhiMin.GetDynamic() )
    {
      m_dynfPhiMin.Update();
      m_dynfPhiMin.SetValue(DEGTORAD(m_dynfPhiMin.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfPhiMin.GetValue() && 
	      m_dynfPhiMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
	{
	  HandleError("Illegal phimin value:  %.3f <= phimin(%.3f) <= %.3f", 
		      0.0,
		      RADTODEG(m_dynfPhiMin.GetValue()),
		      720.0);
	  m_dynfPhiMin.SetValue(cfPhiMin);
	}
    }

  if ( m_dynfPhiMax.GetDynamic() )
    {
      m_dynfPhiMax.Update();
      m_dynfPhiMax.SetValue(DEGTORAD(m_dynfPhiMax.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfPhiMax.GetValue() && 
	      m_dynfPhiMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
	{
	  HandleError("Illegal phimax value:  %.3f <= phimax(%.3f) <= %.3f", 
		      0.0,
		      RADTODEG(m_dynfPhiMax.GetValue()),
		      720.0);
	  m_dynfPhiMax.SetValue(cfPhiMax);
	}
    }

  if ( m_dynfThetaMin.GetDynamic() )
    {
      m_dynfThetaMin.Update();
      m_dynfThetaMin.SetValue(DEGTORAD(m_dynfThetaMin.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfThetaMin.GetValue() && 
	      m_dynfThetaMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
	{
	  HandleError("Illegal thetamin value:  %.3f <= thetamin(%.3f) <= %.3f", 
		      0.0,
		      RADTODEG(m_dynfThetaMin.GetValue()),
		      720.0);
	  m_dynfThetaMin.SetValue(cfThetaMin);
	}
    }

    if ( m_dynfThetaMax.GetDynamic() )
    {
      m_dynfThetaMax.Update();
      m_dynfThetaMax.SetValue(DEGTORAD(m_dynfThetaMax.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(0.0)) <= m_dynfThetaMax.GetValue() && 
	      m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
	{
	  HandleError("Illegal thetamax value:  %.3f <= thetamax(%.3f) <= %.3f", 
		      0.0,
		      RADTODEG(m_dynfThetaMax.GetValue()),
		      720.0);
	  m_dynfThetaMax.SetValue(cfThetaMax);
	}
    }

  m_dyniPhiSlices.Update();
  if ( !( 0 < m_dyniPhiSlices.GetValue() ) )
    {
      HandleError("phislices (%ld) must be >= 0", m_dyniPhiSlices.GetValue());
      m_dyniPhiSlices.SetValue(cuiPhiSlices);
    }

  m_dyniThetaSlices.Update();
  if ( !( 0 < m_dyniThetaSlices.GetValue() ) )
    {
      HandleError("thetaslices (%ld) must be >= 0", m_dyniThetaSlices.GetValue());
      m_dyniThetaSlices.SetValue(cuiThetaSlices);
    }

  m_dyniDrawEllipse.Update();
  m_dyniDrawHyperbola.Update();

  for ( i = 0; i < 4; i++ )
    {
      m_pvTexture[i].Update();
    }
}

VOID CSLIDECyclide::CalculateBound()
{
  CBound bound;
  FLOAT fTemp;

  FLOAT fA, fB, fE;
  FLOAT fThetaMin, fThetaMax, fThetaDelta;
  FLOAT fPhiMin, fPhiMax, fPhiDelta;
  UINT_32 uiPhiSlices, uiThetaSlices;
  FLOAT fXMinTheta, fXMaxTheta, fYMinTheta, fYMaxTheta;
  FLOAT fInnerXYPhi, fOuterXYPhi, fZMinPhi, fZMaxPhi;
  Vector vPQ;
  Point pZ;
  
  fA = m_dynfXMajorA.GetValue();
  fB = m_dynfYMinorB.GetValue();
  fE = (FLOAT)sqrt( (fA*fA) - (fB*fB) ); 
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  // 0 <= ThetaMin <= ThetaMax <= 720  and  ThetaMax - ThetaMin <= 360
  if( fThetaMin > fThetaMax ) {
      fTemp = fThetaMin;
      fThetaMin = fThetaMax;
      fThetaMax = fTemp;
  }
  if( (fThetaMax - fThetaMin) > DEGTORAD(360) ) {
      fThetaMin = 0;
      fThetaMax = DEGTORAD(360);
  }

  // 0 <= PhiMin <= PhiMax <= 720   and    PhiMax - PhiMin <= 360
  if( fPhiMin > fPhiMax ) {
      fTemp = fPhiMin;
      fPhiMin = fPhiMax;
      fPhiMax = fTemp;
  }
  if( (fPhiMax - fPhiMin) > DEGTORAD(360) ) {
      fPhiMin = 0;
      fPhiMax = DEGTORAD(360);
  }

  // Find the extreme Phi values in the X and Y direction
  // absolute min at 0, absolute max at 180
  if( ((fPhiMin <= 0) && (0 <= fPhiMax)) ||
      ((fPhiMin <= DEGTORAD(360))  && (DEGTORAD(360) <= fPhiMax)) ||
      ((fPhiMin <= DEGTORAD(720))  && (DEGTORAD(720) <= fPhiMax)) )
      fInnerXYPhi = 0;            
  else if ( cos(fPhiMin) > cos(fPhiMax) )
      fInnerXYPhi = fPhiMin;
  else
      fInnerXYPhi = fPhiMax;

  if( ((fPhiMin <= DEGTORAD(180))  && (DEGTORAD(180) <= fPhiMax)) ||
      ((fPhiMin <= DEGTORAD(540))  && (DEGTORAD(540) <= fPhiMax)) )
      fOuterXYPhi = DEGTORAD(180);
  else if ( cos(fPhiMin) < cos(fPhiMax) )
      fOuterXYPhi = fPhiMin;
  else
      fOuterXYPhi = fPhiMax;

  // Y:  absolute max at theta =  acos(e/a cos(outer phi)),
  //     absolute min at theta = -acos(e/a cos(outer phi))
  fYMaxTheta = (FLOAT)acos(fE/fA*cos( fOuterXYPhi )); 
  fYMinTheta = DEGTORAD(360) - fYMaxTheta;     

  if( !( ((fThetaMin < fYMinTheta) && (fYMinTheta <= fThetaMax)) ||
	 ((fThetaMin < ((DEGTORAD(360) + fYMinTheta))) &&
	  ((DEGTORAD(360) + fYMinTheta) <= fThetaMax)) ) ) {   // no absMin

      if( fabs(fYMinTheta - fThetaMin) < fabs(fYMinTheta - fThetaMax) )
	  fYMinTheta = fThetaMin;
      else
	  fYMinTheta = fThetaMax;
  }

  if( !( ((fThetaMin < fYMaxTheta) && (fYMaxTheta <= fThetaMax)) ||
	 ((fThetaMin < ((DEGTORAD(360) + fYMaxTheta))) &&
	  ((DEGTORAD(360) + fYMaxTheta) <= fThetaMax)) ) ) {   // no absMax

      if( fabs(fYMaxTheta - fThetaMin) < fabs(fYMaxTheta - fThetaMax) )
	  fYMaxTheta = fThetaMin;
      else
	  fYMaxTheta = fThetaMax;
  }

  CalculateZ( fYMinTheta, fInnerXYPhi, pZ, vPQ );
  bound.Insert(pZ);
  CalculateZ( fYMinTheta, fOuterXYPhi, pZ, vPQ );
  bound.Insert(pZ);
  CalculateZ( fYMaxTheta, fInnerXYPhi, pZ, vPQ );
  bound.Insert(pZ);
  CalculateZ( fYMaxTheta, fOuterXYPhi, pZ, vPQ );
  bound.Insert(pZ);
  
  // X: absolute min at theta = 180, absolute max at theta = 0   
  if( ((fThetaMin <= DEGTORAD(180)) && (DEGTORAD(180) <= fThetaMax)) ||
      ((fThetaMin <= DEGTORAD(540)) && (DEGTORAD(540) <= fThetaMax)) )
      fXMinTheta = DEGTORAD(180);
  else if ( cos(fThetaMin) < cos(fThetaMax) )
      fXMinTheta = fThetaMin;
  else
      fXMinTheta = fThetaMax;

  if( ((fThetaMin <= 0) && (0 <= fThetaMax)) ||
      ((fThetaMin <= DEGTORAD(360)) && (DEGTORAD(360) <= fThetaMax)) ||
      ((fThetaMin <= DEGTORAD(720)) && (DEGTORAD(720) <= fThetaMax)) )
      fXMaxTheta = 0;
  else if( cos(fThetaMin) > cos(fThetaMax) )
      fXMaxTheta = fThetaMin;
  else
      fXMaxTheta = fThetaMax;
        
  CalculateZ( fXMaxTheta, fInnerXYPhi, pZ, vPQ );
  bound.Insert( pZ );
  CalculateZ( fXMaxTheta, fOuterXYPhi, pZ, vPQ );
  bound.Insert( pZ );
  CalculateZ( fXMinTheta, fInnerXYPhi, pZ, vPQ );
  bound.Insert( pZ );
  CalculateZ( fXMinTheta, fOuterXYPhi, pZ, vPQ );
  bound.Insert( pZ );

  // Z: absolute max at fXMinTheta, phi =  acos( e/a cos(XMinTheta) )
  //    absolute min at fXMinTheta, phi = -acos( e/a cos(XMinTheta) )
  fZMaxPhi = (FLOAT)acos( fE/fA*cos( fXMinTheta ) );
  fZMinPhi = DEGTORAD(360) - fZMaxPhi;
  
  if( !( ((fPhiMin <= fZMinPhi) && (fZMinPhi <= fPhiMax)) ||
	 ((fPhiMin <= (DEGTORAD(360) + fZMinPhi)) &&
	  ((DEGTORAD(360) + fZMinPhi) <= fPhiMax)) ) ) { // no absMin

      if ( fabs(fZMinPhi - fPhiMin) < fabs(fZMinPhi - fPhiMax) )
	  fZMinPhi = fPhiMin;
      else
	  fZMinPhi = fPhiMax;
  }
  
  if( !( ((fPhiMin <= fZMaxPhi) && (fZMaxPhi <= fPhiMax)) ||
	 ((fPhiMin <= (DEGTORAD(360) + fZMaxPhi)) &&
	  ((DEGTORAD(360) + fZMaxPhi) <= fPhiMax)) ) ) {  // no absMax
      
      if ( fabs(fZMaxPhi - fPhiMin) < fabs(fZMaxPhi - fPhiMax) )
	  fZMaxPhi = fPhiMin;
      else
	  fZMaxPhi = fPhiMax;
  }

  CalculateZ( fXMinTheta, fZMaxPhi, pZ, vPQ );
  bound.Insert( pZ );
  CalculateZ( fXMinTheta, fZMinPhi, pZ, vPQ );
  bound.Insert( pZ );

  if( fA == fB )
    {
      pZ[Z] = -pZ[Z];
      bound.Insert( pZ );
    }
    
  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDECyclide::CalculateCDAVertices(CListInt *plistCDAVertex, 
				       const Matrix &mGroot_obj)
{

}
#endif // __USE_CDA__

VOID CSLIDECyclide::CalculateZ( FLOAT fTheta, FLOAT fPhi, Point &pZ, Vector &vPQ)
{
    FLOAT fA, fB, fE, fL, fLP;
    FLOAT fPx, fPy, fPz, fQx, fQy, fQz; 
    Point pP;
    
    fA = m_dynfXMajorA.GetValue();
    fB = m_dynfYMinorB.GetValue();
    fE = (FLOAT)sqrt( (fA*fA) - (fB*fB) );     
    fL = m_dynfStringLength.GetValue();

    fLP = fL - fA - fE* (FLOAT)cos( fTheta );
    fPx = fA* (FLOAT)cos( fTheta );
    fPy = fB* (FLOAT)sin( fTheta );
    fPz = 0;
    pP.Set(fPx, fPy, fPz);

    // check for division by zero
    if( fPhi == (float)(PI/2) ) 
      {
	ASSERT ( fA != 0.0 );
	vPQ.Set( fE/fA, 0.0, fB/fA );
      } 
    else if ( fPhi == (float)(3*PI/2) ) 
      {
	ASSERT ( fA != 0.0 );
	vPQ.Set( fE/fA, 0.0, -fB/fA );
      } 
    else 
      {
	ASSERT ( cos(fPhi) != 0.0 );
      
	// Q
	fQx = fE/(FLOAT)cos( fPhi );
	fQy = 0;
	fQz = fB*(FLOAT)tan( fPhi );
	
	// PQ
	vPQ.Set( fQx-fPx, fQy-fPy, fQz-fPz);
	vPQ.Normalize();
	
	if( ((DEGTORAD(90) <= fPhi) && (fPhi <= DEGTORAD(270))) ) 
	  {
	    vPQ *= -1; 	    // flip normal
	  }
      }
    
    //vPQ *= fLP; // fLP is zero sometimes
    ASSERT( !((vPQ[0] == 0.0) && (vPQ[1] == 0.0) && (vPQ[2] == 0.0)) );
    
    pZ = pP + fLP*vPQ;
}

VOID CSLIDECyclide::CalculateQEMesh()
{
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta0, fTheta1, fPhi;
  Vector vPQ0, vPQ1;
  Point pZ0, pZ1;
  QEVertex* pqev;
  CObjectList<QEVertex> listQEVertex;
  
  //dont worry about drawing the elipse and hyperbola since we're just doing the mesh...
  
  // Cyclide
  ASSERT(m_pQEMesh == NULL);
  m_pQEMesh=AllocateQEMesh();
  ASSERT(m_pQEMesh != NULL);
  m_pQEMesh->Init();
  
  listQEVertex.Init();
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;    
  fTheta0 = fThetaMin;
  fTheta1 = fTheta0 + fThetaDelta;
 
  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
    {
      listQEVertex.RemoveAll();
      //gilBegin(GL_TRIANGLE_STRIP);
      {
	fPhi = fPhiMin;
	
	for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	  {
	    // p1->q
	    CalculateZ( fTheta1, fPhi, pZ1, vPQ1 );
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pqev->Init(pZ1);
	    listQEVertex.InsertLast(pqev);
	    
	    // p0->q
	    CalculateZ( fTheta0, fPhi, pZ0, vPQ0 );
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pqev->Init(pZ0);
	    listQEVertex.InsertLast(pqev);
	    
	    fPhi += fPhiDelta;
	  }
      }
      m_pQEMesh->MergeTriangleStrip(&listQEVertex);
      //gilEnd();
      
      fTheta0 = fTheta1;
      fTheta1 += fThetaDelta;
    }
}

VOID CSLIDECyclide::CalculateLeafQEMesh()
{
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDECyclide::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

VOID CSLIDECyclide::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  FLOAT fX, fY, fZ, fAngle;
  FLOAT fA, fB, fE;                        // ellipse and hyperbola variables      
  FLOAT fNormalDir;                        // cyclide variables
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta0, fTheta1, fPhi;
  Vector vPQ0, vPQ1;
  Point pZ0, pZ1;
  
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  
  fA = m_dynfXMajorA.GetValue();
  fB = m_dynfYMinorB.GetValue();
  fE = (FLOAT)sqrt( (fA*fA) - (fB*fB) );     
  
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);
    
    // Ellipse
    if( m_dyniDrawEllipse.GetValue() != 0 ) {
      gilBegin(GL_LINE_LOOP);
      {
	for( int i = 0; i < 360; i++ ){
	  fAngle = (FLOAT)(i*PI/180);
	  fX = fA * (FLOAT)cos( fAngle );
	  fY = fB * (FLOAT)sin( fAngle );
	  gilVertex3f( fX, fY, 0 );
	}
      }
      gilEnd();
      
      gilBegin(GL_POINTS);
      {
	gilVertex3f( fE, 0, 0 );
	gilVertex3f( -fE, 0, 0 );
      }
      gilEnd();
    }
    
    // Hyperbola
    if( m_dyniDrawHyperbola.GetValue() != 0 ) {
      gilBegin(GL_LINE_STRIP);
      {
	for( int i = -89; i < 90; i++ ){
	  fX = fE/(FLOAT)cos( DEGTORAD(i) );
	  fZ = fB * (FLOAT)tan( DEGTORAD(i) );
	  gilVertex3f( fX, 0, fZ );
	}
      }
      gilEnd();
      gilBegin(GL_LINE_STRIP);
      {
	for( int i = 91; i < 270; i++ ){
	  fX = fE/(FLOAT)cos( DEGTORAD(i) );
	  fZ = fB * (FLOAT)tan( DEGTORAD(i) );
	  gilVertex3f( fX, 0, fZ );
	}
      }
      gilEnd();
    }
  }
  PopAuxillaryGeometryAttributes();
  
  // Cyclide
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();

  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiPhiSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiPhiSlices;
  
  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;    
  fTheta0 = fThetaMin;
  fTheta1 = fTheta0 + fThetaDelta;
  
  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	fPhi = fPhiMin;
	
	pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
	pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
	pvU[0] = ppvV[0][0];
	pvU[1] = ppvV[1][0];
	
	for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	  {
	    // p1->q
	    CalculateZ( fTheta1, fPhi, pZ1, vPQ1 );
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);
	    gilNormal3fv( (fNormalDir * vPQ1).Get() );
	    gilVertex3fv( pZ1.Get() );
	    
	    // p0->q
	    CalculateZ( fTheta0, fPhi, pZ0, vPQ0 );
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);
	    gilNormal3fv( (fNormalDir * vPQ0).Get() );
	    gilVertex3fv( pZ0.Get() );
	    
	    fPhi += fPhiDelta;
	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	  }
      }
      gilEnd();
      
      fTheta0 = fTheta1;
      fTheta1 += fThetaDelta;
      
      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    }
}

VOID CSLIDECyclide::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  FLOAT fX, fY, fZ, fAngle;
  FLOAT fA, fB, fE;                            // ellipse and hyperbola variables
  
  FLOAT fNormalDir;                             // cyclide variables
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta0, fTheta1, fPhi;
  Vector vP0Q0, vP1Q0, vP0Q1, vNormal;
  Point pZ0, pZ1, pZ2, pZ3;
  
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  
  fA = m_dynfXMajorA.GetValue();
  fB = m_dynfYMinorB.GetValue();
  fE = (FLOAT)sqrt( (fA*fA) - (fB*fB) );     
  
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);
    
    // Ellipse
    if( m_dyniDrawEllipse.GetValue() != 0 ) {
      gilBegin(GL_LINE_LOOP);
      {
	for( int i = 0; i < 360; i++ ){
	  fAngle = (FLOAT)(i*PI/180);
	  fX = fA * (FLOAT)cos( fAngle );
	  fY = fB * (FLOAT)sin( fAngle );
	  gilVertex3f( fX, fY, 0 );
	}
      }
      gilEnd();
      
      gilBegin(GL_POINTS);
      {
	gilVertex3f( fE, 0, 0 );
	gilVertex3f( -fE, 0, 0 );
      }
      gilEnd();
    }
    
    // Hyperbola
    if( m_dyniDrawHyperbola.GetValue() != 0 ) {
      gilBegin(GL_LINE_STRIP);
      {
	for( int i = -89; i < 90; i++ ){
	  fX = fE/(FLOAT)cos( DEGTORAD(i) );
	  fZ = fB * (FLOAT)tan( DEGTORAD(i) );
	  gilVertex3f( fX, 0, fZ );
	}
      }
      gilEnd();
      gilBegin(GL_LINE_STRIP);
      {
	for( int i = 91; i < 270; i++ ){
	  fX = fE/(FLOAT)cos( DEGTORAD(i) );
	  fZ = fB * (FLOAT)tan( DEGTORAD(i) );
	  gilVertex3f( fX, 0, fZ );
	}
	    }
      gilEnd();
    }
  }
  PopAuxillaryGeometryAttributes();
  
  // Cyclide
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  fNormalDir = -1.0;
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  fTheta0 = fThetaMin;
  fTheta1 = fTheta0 + fThetaDelta;
  
  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiPhiSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiPhiSlices;
  
  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];

  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	fPhi = fPhiMin;
	
	pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
	pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
	pvU[0] = ppvV[0][0];
	pvU[1] = ppvV[1][0];
	
	// Must seed the for loop with the left most two points
	CalculateZ( fTheta0, fPhi, pZ0, vP0Q0 );
	CalculateZ( fTheta1, fPhi, pZ1, vP1Q0 );
	CalculateZ( fTheta0, fPhi+fPhiDelta, pZ2, vP0Q1 );
	
	vNormal = ( (pZ1-pZ0) * (pZ2-pZ0) );
	vNormal.Normalize();
	gilNormal3fv( (fNormalDir * vNormal).Get() );
	
	gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	gilVertex3fv( pZ1.Get() );
	gilTexCoord2f(pvU[0][0], pvU[0][1]);	
	gilVertex3fv( pZ0.Get() );
	
	fPhi += fPhiDelta;
	pvU[0] += pvUDelta[0];
	pvU[1] += pvUDelta[1];
	
	for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
	  {
	    CalculateZ( fTheta0, fPhi, pZ2, vP0Q0 );
	    CalculateZ( fTheta1, fPhi, pZ3, vP1Q0 );
	    
	    vNormal = ( (pZ2-pZ3) * (pZ1-pZ3) );
	    vNormal.Normalize();
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);	
	    gilNormal3fv( (fNormalDir * vNormal).Get() );
	    gilVertex3fv( pZ3.Get() );
	    
	    vNormal = ( (pZ1-pZ0) * (pZ2-pZ0) );
	    vNormal.Normalize();
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);
	    gilNormal3fv( (fNormalDir * vNormal).Get() );
	    gilVertex3fv( pZ2.Get() );
	    
	    pZ0 = pZ2;
	    pZ1 = pZ3;
	    fPhi += fPhiDelta;
	    
	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	  }
      }
      gilEnd();
      
      fTheta0 = fTheta1;
      fTheta1 += fThetaDelta;
      
      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    }
}

VOID CSLIDECyclide::RenderNormalsSmooth()
{
    UINT_32 uiPhiSlices, uiThetaSlices;
    UINT_32 uiPhiSlice, uiThetaSlice;
    FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
    FLOAT fPhiDelta, fThetaDelta;
    FLOAT fTheta, fPhi, fNormalDir;
    Vector vPQ, vNormal;
    Point pZ;
      
    uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
    uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
    fPhiMin = m_dynfPhiMin.GetValue();
    fPhiMax = m_dynfPhiMax.GetValue();
    fThetaMin = m_dynfThetaMin.GetValue();
    fThetaMax = m_dynfThetaMax.GetValue();

    if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
	 ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
	fNormalDir = -1.0;
    }
    else
    {
	fNormalDir = 1.0;
    }
	
    fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
    fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
    fTheta = fThetaMin;

    gilBegin(GL_LINES);
    {
	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	    fPhi = fPhiMin;
	    for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	    {
		CalculateZ( fTheta, fPhi, pZ, vPQ );
		gilVertex3fv( pZ.Get() );
		
		vPQ.Normalize();
		vPQ *= fNormalDir;
		pZ += vPQ;
		gilVertex3fv( pZ.Get() );
		
		fPhi += fPhiDelta;
	    }
	    fTheta += fThetaDelta;
	}
    }
    gilEnd();
}

VOID CSLIDECyclide::RenderNormalsFlat()
{
    UINT_32 uiPhiSlices, uiThetaSlices;
    UINT_32 uiPhiSlice, uiThetaSlice;
    FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
    FLOAT fPhiDelta, fThetaDelta;
    FLOAT fTheta0, fTheta1, fPhi;
    Vector vPQ, vNormal;
    Point pZ0, pZ1, pZ2, pZ3, pt;
      
    uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
    uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
    fPhiMin = m_dynfPhiMin.GetValue();
    fPhiMax = m_dynfPhiMax.GetValue();
    fThetaMin = m_dynfThetaMin.GetValue();
    fThetaMax = m_dynfThetaMax.GetValue();

    fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
    fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
    
    fTheta0 = fThetaMin;
    fTheta1 = fTheta0 + fThetaDelta;

    gilBegin(GL_LINES);
    {
	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	{	    
	    fPhi = fPhiMin;
	    
	    // Must seed the for loop with the left most two points
	    CalculateZ( fTheta0, fPhi, pZ0, vPQ );
	    CalculateZ( fTheta1, fPhi, pZ1, vPQ );

	    fPhi += fPhiDelta;
	    
	    for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	    {
		CalculateZ( fTheta0, fPhi, pZ2, vPQ );
		CalculateZ( fTheta1, fPhi, pZ3, vPQ );

		pt = (pZ0 + pZ1 + pZ2)/3.0;
		gilVertex3fv( pt.Get() );
		vNormal = ( (pZ1-pZ0) * (pZ2-pZ0) );
		vNormal.Normalize();
		pt -= vNormal;
		gilVertex3fv( pt.Get() );

		pt = (pZ1 + pZ2 + pZ3)/3.0;
		gilVertex3fv( pt.Get() );
		vNormal = ( (pZ2-pZ3) * (pZ1-pZ3) );
		vNormal.Normalize();
		pt -= vNormal; 
		gilVertex3fv( pt.Get() );
		
		pZ0 = pZ2;
		pZ1 = pZ3;
		fPhi += fPhiDelta;
	    }
	    fTheta0 = fTheta1;
	    fTheta1 += fThetaDelta;
	}
    }
    gilEnd();
}

VOID CSLIDECyclide::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtCyclide], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtMajorRadius],
	  m_dynfXMajorA.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtMinorRadius],
	  m_dynfYMinorB.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtStringLength],
	  m_dynfStringLength.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtPhiMin],
	  RADTODEG(m_dynfPhiMin.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtPhiMax],
	  RADTODEG(m_dynfPhiMax.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtThetaMin],
	  RADTODEG(m_dynfThetaMin.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtThetaMax],
	  RADTODEG(m_dynfThetaMax.GetValue()));
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtPhiSlices],
	  m_dyniPhiSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtThetaSlices],
	  m_dyniThetaSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawEllipse],
	  m_dyniDrawEllipse.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawHyperbola],
	  m_dyniDrawHyperbola.GetValue());

  
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndCyclide]);
}

VOID CSLIDECyclide::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  FLOAT fNormalDir;                        // cyclide variables
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta0, fTheta1, fPhi;
  Vector vPQ0, vPQ1, vPQ2, vPQ3;
  Point pZ0, pZ1, pZ2, pZ3;
  
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  FLOAT pfTexture0[2], pfTexture1[2];
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  
  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiPhiSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiPhiSlices;
  
  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;    
  fTheta0 = fThetaMin;
  fTheta1 = fTheta0 + fThetaDelta;
  
  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
    {
      fPhi = fPhiMin;
      
      pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
      pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
      pvU[0] = ppvV[0][0];
      pvU[1] = ppvV[1][0];

      // p1->q
      CalculateZ( fTheta1, fPhi, pZ1, vPQ1 );
      vPQ1 *= fNormalDir;
      //gilTexCoord2f(pvU[1][0], pvU[1][1]);
      pfTexture1[0] = pvU[1][0];
      pfTexture1[1] = pvU[1][1];
      
      // p0->q
      CalculateZ( fTheta0, fPhi, pZ0, vPQ0 );
      vPQ0 *= fNormalDir;
      //gilTexCoord2f(pvU[0][0], pvU[0][1]);
      pfTexture0[0] = pvU[0][0];
      pfTexture0[1] = pvU[0][1];

      fPhi += fPhiDelta;
      pvU[0] += pvUDelta[0];
      pvU[1] += pvUDelta[1];

      for ( uiPhiSlice = 1; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	{
	  // p1->q
	  CalculateZ( fTheta1, fPhi, pZ3, vPQ3 );
	  vPQ3 *= fNormalDir;
	  //gilTexCoord2f(pvU[1][0], pvU[1][1]);
	  
	  // p0->q
	  CalculateZ( fTheta0, fPhi, pZ2, vPQ2 );
	  vPQ2 *= fNormalDir;
	  //gilTexCoord2f(pvU[0][0], pvU[0][1]);
	  
	  if( (pZ3 != pZ1) && (pZ3 != pZ0) && (pZ1 != pZ0) )
	    {
	      RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f] \"N\" [%.3f %.3f %.3f  %.3f %.3f %.3f %.3f %.3f %.3f] \"st\" [%.3f %.3f %.3f %.3f %.3f %.3f]",
			      pZ1[X], pZ1[Y], pZ1[Z], pZ0[X], pZ0[Y], pZ0[Z], pZ3[X], pZ3[Y], pZ3[Z],
			      vPQ1[X], vPQ1[Y], vPQ1[Z], vPQ0[X], vPQ0[Y], vPQ0[Z], vPQ3[X], vPQ3[Y], vPQ3[Z],
			      pfTexture1[0], pfTexture1[1], pfTexture0[0], pfTexture0[1], pvU[1][0], pvU[1][1]);
	    }

	  if( (pZ3 != pZ2) && (pZ3 != pZ0) && (pZ2 != pZ0) )
	    {
	      RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f] \"N\" [%.3f %.3f %.3f  %.3f %.3f %.3f %.3f %.3f %.3f] \"st\" [%.3f %.3f %.3f %.3f %.3f %.3f]",
			      pZ2[X], pZ2[Y], pZ2[Z], pZ3[X], pZ3[Y], pZ3[Z], pZ0[X], pZ0[Y], pZ0[Z],
			      vPQ2[X], vPQ2[Y], vPQ2[Z], vPQ3[X], vPQ3[Y], vPQ3[Z], vPQ0[X], vPQ0[Y], vPQ0[Z],
			      pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], pfTexture0[0], pfTexture0[1]);
	    }

	  pZ0 = pZ2;
	  pZ1 = pZ3;
	  vPQ0 = vPQ2;
	  vPQ1 = vPQ3;
	  pfTexture0[0] = pvU[0][0];
	  pfTexture0[1] = pvU[0][1];
	  pfTexture1[0] = pvU[1][0];
	  pfTexture1[1] = pvU[1][1];

	  fPhi += fPhiDelta;
	  pvU[0] += pvUDelta[0];
	  pvU[1] += pvUDelta[1];
	}
	
      fTheta0 = fTheta1;
      fTheta1 += fThetaDelta;
      
      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    }
}

VOID CSLIDECyclide::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE Cyclide: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDECyclide::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE Cyclide: %s", m_pcGUIID);
}

VOID CSLIDECyclide::WritePOVGeometry(FILE *fpOutputFile, 
				     CListInt *plistTransformIDs,
				     CListInt *plistColors) 
{

  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;

  FLOAT fNormalDir;                        // cyclide variables
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta0, fTheta1, fPhi0, fPhi1;
  CHAR *pcTransformID;
  CHAR *pcColorID;
  Vector vPQ00, vPQ01, vPQ10, vPQ11;
  Point pZ00, pZ01, pZ10, pZ11;
        
  // Cyclide
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;    
  fTheta0 = fThetaMin;
  fTheta1 = fTheta0 + fThetaDelta;
  
  fprintf(fpOutputFile, "mesh { \n");

  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
    {

      fPhi0 = fPhiMin;
      fPhi1 = fPhi0 + fPhiDelta;

      for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	{
	  CalculateZ(fTheta0, fPhi0, pZ00, vPQ00 );
	  CalculateZ(fTheta0, fPhi1, pZ01, vPQ01 );
	  CalculateZ(fTheta1, fPhi0, pZ10, vPQ10 );
	  CalculateZ(fTheta1, fPhi1, pZ11, vPQ11 );

	  fprintf(fpOutputFile, "triangle { ");
	  fprintf(fpOutputFile, "<%f, %f, %f>,", (pZ00.Get())[X], (pZ00.Get())[Y], (pZ00.Get())[Z]);
	  fprintf(fpOutputFile, "<%f, %f, %f>,", (pZ10.Get())[X], (pZ10.Get())[Y], (pZ10.Get())[Z]);
	  fprintf(fpOutputFile, "<%f, %f, %f> ", (pZ01.Get())[X], (pZ01.Get())[Y], (pZ01.Get())[Z]);
	  fprintf(fpOutputFile, "}\n");
	  
	  fprintf(fpOutputFile, "triangle { ");
	  fprintf(fpOutputFile, "<%f, %f, %f>,", (pZ10.Get())[X], (pZ10.Get())[Y], (pZ10.Get())[Z]);
	  fprintf(fpOutputFile, "<%f, %f, %f>,", (pZ11.Get())[X], (pZ11.Get())[Y], (pZ11.Get())[Z]);
	  fprintf(fpOutputFile, "<%f, %f, %f> ", (pZ01.Get())[X], (pZ01.Get())[Y], (pZ01.Get())[Z]);
	  fprintf(fpOutputFile, "}\n");
  
	  fPhi0 = fPhi1;
	  fPhi1 = fPhi1+fPhiDelta;
	}
      
      fTheta0 = fTheta1;
      fTheta1 += fThetaDelta;
    }
  iter.Init(plistTransformIDs);
  for ( pcTransformID = (CHAR*)iter.PeekFirst(); 
	pcTransformID != NULL;
	pcTransformID = (CHAR*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, " transform %s\n", pcTransformID);
    }

  //color
  ASSERT(m_pCSLIDESurface != NULL);

  if (!(strcmp(GetSurfaceID(), "SLF_INHERIT")))
    {
      iter.Init(plistColors);
      pcColorID = (CHAR*)iter.PeekFirst();
      fprintf(fpOutputFile, " pigment { %s }\n", pcColorID);
    }
  else
    {
      fprintf(fpOutputFile, " pigment {");
      m_pCSLIDESurface->WritePOV(fpOutputFile);
      fprintf(fpOutputFile, " }\n");
    }

  fprintf(fpOutputFile, "}\n\n");
}

VOID CSLIDECyclide::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				      CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: CYCLIDE\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDECyclide::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  FLOAT fNormalDir;                             // cyclide variables
  UINT_32 uiPhiSlices, uiThetaSlices, uiPhiSlice, uiThetaSlice;
  FLOAT fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fTheta0, fTheta1, fPhi;
  Vector vP0Q0, vP1Q0, vP0Q1;
  Point pZ0, pZ1, pZ2, pZ3;

  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = m_dynfThetaMin.GetValue();
  fThetaMax = m_dynfThetaMax.GetValue();
  fNormalDir = -1.0;  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  fTheta0 = fThetaMin;
  fTheta1 = fTheta0 + fThetaDelta;

  if( mTransform.Determinant() > 0 ) // Don't Mirror
    {
      for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	{
	  fPhi = fPhiMin;
	  
	  // Must seed the for loop with the left most two points
	  CalculateZ( fTheta0, fPhi, pZ0, vP0Q0 );
	  CalculateZ( fTheta1, fPhi, pZ1, vP1Q0 );
	  
	  pZ0 = mTransform * pZ0;
	  pZ1 = mTransform * pZ1;
	  
	  fPhi += fPhiDelta;
	  
	  for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
	    {
	      CalculateZ( fTheta0, fPhi, pZ2, vP0Q0 );
	      CalculateZ( fTheta1, fPhi, pZ3, vP1Q0 );
	      
	      pZ2 = mTransform * pZ2;
	      pZ3 = mTransform * pZ3;
	      
	      WriteSTLTriangle(fpOutputFile, pZ1, pZ0, pZ3);
	      WriteSTLTriangle(fpOutputFile, pZ0, pZ2, pZ3);

	      pZ0 = pZ2;
	      pZ1 = pZ3;
	      fPhi += fPhiDelta;
	    }
	  fTheta0 = fTheta1;
	  fTheta1 += fThetaDelta;
	}
    }
  else // Mirror
    {
      for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	{
	  fPhi = fPhiMin;
	  
	  // Must seed the for loop with the left most two points
	  CalculateZ( fTheta0, fPhi, pZ0, vP0Q0 );
	  CalculateZ( fTheta1, fPhi, pZ1, vP1Q0 );
	  
	  pZ0 = mTransform * pZ0;
	  pZ1 = mTransform * pZ1;
	  
	  fPhi += fPhiDelta;
	  
	  for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
	    {
	      CalculateZ( fTheta0, fPhi, pZ2, vP0Q0 );
	      CalculateZ( fTheta1, fPhi, pZ3, vP1Q0 );
	      
	      pZ2 = mTransform * pZ2;
	      pZ3 = mTransform * pZ3;

	      WriteSTLTriangle(fpOutputFile, pZ3, pZ0, pZ1);
	      WriteSTLTriangle(fpOutputFile, pZ3, pZ2, pZ0);

	      pZ0 = pZ2;
	      pZ1 = pZ3;
	      fPhi += fPhiDelta;
	    }
	  fTheta0 = fTheta1;
	  fTheta1 += fThetaDelta;
	}
    }
}

VOID CSLIDECyclide::GetParameters(Tcl_Interp *pInterp)
{
  INT_32 iCode;
  CHAR pcBuffer[1000];
  CHAR pcSurface[35];

  //pInterp = CSLIDEWorld::GetInterp();

  if( !strcmp(m_pCSLIDESurface->GetGUIID(), "SLF_INHERIT" ) )
      sprintf(pcSurface, "inherit");
  else 
      sprintf(pcSurface, m_pCSLIDESurface->GetGUIID());

  sprintf(pcBuffer, "Editor:SetParameterValues cyclide {{majorradius %.3f} {minorradius %.3f} {stringlength %.3f} {thetamin %.3f} {thetamax %.3f} {thetaslices %ld} {phimin %.3f} {phimax %.3f} {phislices %ld} {solid %s} {surface %s} {shading %s} {lod %s} {ribbegin %s} {ribend %s}}",
	  m_dynfXMajorA.GetValue(),
	  m_dynfYMinorB.GetValue(),
	  m_dynfStringLength.GetValue(),
	  RADTODEG(m_dynfThetaMin.GetValue()),
	  RADTODEG(m_dynfThetaMax.GetValue()),
	  m_dyniThetaSlices.GetValue(),
	  RADTODEG(m_dynfPhiMin.GetValue()),
	  RADTODEG(m_dynfPhiMax.GetValue()),
	  m_dyniPhiSlices.GetValue(),
	  gcppcSLIDEGUISolidReservedWords[m_dyniSolid.GetValue()],
	  pcSurface,
	  gcppcSLIDEGUIShadingReservedWords[m_dyniShading.GetValue()],
	  gcppcSLIDEGUILODReservedWords[m_dyniLOD.GetValue()],
	  m_pcRIBBegin,
	  m_pcRIBEnd);
	  
  iCode = Tcl_Eval(pInterp, pcBuffer);
  //ASSERT( iCode == TCL_OK );
  if ( iCode != TCL_OK )
    {
      printf( "%s\n", pcBuffer );
      printf("Error %s \n", pInterp->result);
    }
}

UINT_32 CSLIDECyclide::SetParameter(CSLIDEWorld *pCSLIDEWorld, 
				 SLIDEReservedWordType uiParameter, CHAR **argv)
{
  switch(uiParameter) 
    {
    case sliderwtMajorRadius:
      m_dynfXMajorA.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtMinorRadius:
      m_dynfYMinorB.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtStringLength:
      m_dynfStringLength.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtThetaMin:
      m_dynfThetaMin.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtThetaMax:
      m_dynfThetaMax.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtThetaSlices:
      m_dyniThetaSlices.SetValue(atoi(argv[0]));
      return 1;
      break;
    case sliderwtPhiMin:
      m_dynfPhiMin.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtPhiMax:
      m_dynfPhiMax.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtPhiSlices:
      m_dyniPhiSlices.SetValue(atoi(argv[0]));
      return 1;
      break;
    default:
      return CSLIDEAtomNode::SetParameter(pCSLIDEWorld, uiParameter, argv);
      break;
    }
}
