#include "slidecylinder.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "qemesh.h"
#include "qevertex.h"
#include "qewrite.h"
#include "stl.h"

static const FLOAT cfRadius = 1.0;
static const FLOAT cfZMin = 0.0;
static const FLOAT cfZMax = 1.0;
static const FLOAT cfThetaMax = DEGTORAD(360.0);
static const UINT_32 cuiZSlices = 1;
static const UINT_32 cuiThetaSlices = 8;
static const UINT_32 cuiEndCap = 0;
static const UINT_32 cuiBeginCap = 0;

//////////////////////////////////////////////////////////////////////
// CSLIDECylinder Class
//

CSLIDECylinder::CSLIDECylinder()
{
  m_dynfRadius.SetValue(cfRadius);
  m_dynfZMin.SetValue(cfZMin);
  m_dynfZMax.SetValue(cfZMax);
  m_dynfThetaMax.SetValue(cfThetaMax);
  m_dyniZSlices.SetValue(cuiZSlices);
  m_dyniThetaSlices.SetValue(cuiThetaSlices);
  m_dyniEndCap.SetValue(cuiEndCap);
  m_dyniBeginCap.SetValue(cuiBeginCap);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

CSLIDECylinder::~CSLIDECylinder()
{
}

VOID CSLIDECylinder::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dynfRadius.SetValue(cfRadius);
  m_dynfZMin.SetValue(cfZMin);
  m_dynfZMax.SetValue(cfZMax);
  m_dynfThetaMax.SetValue(cfThetaMax);
  m_dyniZSlices.SetValue(cuiZSlices);
  m_dyniThetaSlices.SetValue(cuiThetaSlices);
  m_dyniEndCap.SetValue(cuiEndCap);
  m_dyniBeginCap.SetValue(cuiBeginCap);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

VOID CSLIDECylinder::Uninit()
{
  CSLIDEAtomNode::Uninit();
}

BOOL CSLIDECylinder::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDECylinder::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;

  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dynfRadius.GetDynamic() &&
       !( 0.0 < m_dynfRadius.GetValue() ) )
    {
      HandleError("Illegal radius value:  0.0 < %.3f", 
		  m_dynfRadius.GetValue());
      m_dynfRadius.SetValue(cfRadius);
      bCorrect = FALSE;
    }

  // zmin and zmax can be any value
  /*
  if ( !m_dynfZMin.GetDynamic() &&
       !( 0.0 <= m_dynfZMin.GetValue() && m_dynfZMin.GetValue() <= 1.0) )
    {
      HandleError("Illegal zmin value:  0.0 <= %.3f <= 1.0", m_dynfZMin.GetValue());
      m_dynfZMin.SetValue(cfZMin);
      bCorrect = FALSE;
    }

  if ( !m_dynfZMax.GetDynamic() &&
       !( 0.0 <= m_dynfZMax.GetValue() && m_dynfZMax.GetValue() <= 1.0) )
    {
      HandleError("Illegal zmax value:  0.0 <= %.3f <= 1.0", m_dynfZMax.GetValue());
      m_dynfZMax.SetValue(cfZMax);
      bCorrect = FALSE;
    }
    */

  if ( !m_dynfThetaMax.GetDynamic() &&
       !( (FLOAT)(DEGTORAD(-360.0)) <= m_dynfThetaMax.GetValue() && 
	  m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(360.0)) ) )
    {
      HandleError("Illegal thetamax value:  %.3f <= %.3f <= %.3f", 
		  -360.0,
		  RADTODEG(m_dynfThetaMax.GetValue()),
		  360.0);
      m_dynfThetaMax.SetValue(cfThetaMax);
      bCorrect = FALSE;
    }

  if ( !m_dyniZSlices.GetDynamic() && 
       !( 0 < m_dyniZSlices.GetValue() ) )
    {
      HandleError("Illegal zslices value:  %ld", 
		  m_dyniZSlices.GetValue());
      m_dyniZSlices.SetValue(cuiZSlices);
      bCorrect = FALSE;
    }

  if ( !m_dyniThetaSlices.GetDynamic() &&
       !( 0 < m_dyniThetaSlices.GetValue() ) )
    {
      HandleError("Illegal thetaslices value:  %ld", 
		  m_dyniThetaSlices.GetValue());
      m_dyniThetaSlices.SetValue(cuiThetaSlices);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDECylinder::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0.0 < m_dynfRadius.GetValue() ) &&
	   ( (FLOAT)(DEGTORAD(-360.0)) <= m_dynfThetaMax.GetValue() && 
	     m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(360.0)) ) &&
	   ( 0 < m_dyniZSlices.GetValue() ) &&
	   ( 0 < m_dyniThetaSlices.GetValue() ) );
}

VOID CSLIDECylinder::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			     const Matrix &mTransform, const Matrix &mInverse, 
			     BOOL bReverse)
{
  ASSERT(uiSide < 4);

  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ, fTheta, fNormalDir;
  Matrix m;
  Point pt;
  Point *ppt;
  Vector vNormal;
  Vector *pvNormal;
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  if ( ( (fZMin > fZMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fZMin > fZMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  if( (uiSide == 0) || (uiSide == 2) )
    {
      if( uiSide == 0 )
	{
	  fTheta = fThetaMin;
	}
      else
	{
	  fTheta = fThetaMax;
	}

      if( bReverse == 0 )
	{
	  fZ = fZMin;
	}
      else
	{
	  fZ = fZMax;
	  fZDelta *= -1.0;
	}

      for ( uiZSlice = 0; uiZSlice <= uiZSlices; uiZSlice++ )
	{
	  pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ);
	  pt = mTransform * pt;
	  ppt = new Point;
	  ppt->Set(pt.Get());
	  plistPoint->InsertLast(ppt, 0);

	  vNormal.Set((FLOAT)cos(fTheta), (FLOAT)sin(fTheta), 0.0);
	  vNormal = fNormalDir * vNormal;
	  m = mInverse;
	  m.Transpose();
	  vNormal = m * vNormal;
	  ASSERT(vNormal.Normalize());
	  pvNormal = new Vector;
	  pvNormal->Set(vNormal.Get());
	  plistNormal->InsertLast(pvNormal, 0);

	  fZ += fZDelta;
	}
    }
  else // (uiSide == 1) || (uiSide == 3)
    {
      if( uiSide == 1 )
	{
	  fZ = fZMin;
	}
      else
	{
	  fZ = fZMax;
	}

      if( bReverse == 0 )
	{
	  fTheta = fThetaMin;
	}
      else
	{
	  fTheta = fThetaMin;
	  fThetaDelta *= -1.0;
	}

      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ);
	  pt = mTransform * pt;
	  ppt = new Point;
	  ppt->Set(pt.Get());
	  plistPoint->InsertLast(ppt, 0);

	  vNormal.Set((FLOAT)cos(fTheta), (FLOAT)sin(fTheta), 0.0);
	  vNormal = fNormalDir * vNormal;
	  m = mInverse;
	  m.Transpose();
	  vNormal = m * vNormal;
	  ASSERT(vNormal.Normalize());
	  pvNormal = new Vector;
	  pvNormal->Set(vNormal.Get());
	  plistNormal->InsertLast(pvNormal, 0);

	  fTheta += fThetaDelta;
	}
    }
}

VOID CSLIDECylinder::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  // Are there variables which I must call TCL to update?
  SetDynamic(GetDynamic() ||
	     m_dynfRadius.GetDynamic() ||
	     m_dynfZMin.GetDynamic() ||
	     m_dynfZMax.GetDynamic() ||
	     m_dynfThetaMax.GetDynamic() ||
	     m_dyniZSlices.GetDynamic() ||
	     m_dyniThetaSlices.GetDynamic() ||
	     m_dyniEndCap.GetDynamic() ||
	     m_dyniBeginCap.GetDynamic() ||
	     m_pvTexture[0].GetDynamic() ||
	     m_pvTexture[1].GetDynamic() ||
	     m_pvTexture[2].GetDynamic() ||
	     m_pvTexture[3].GetDynamic());
}

VOID CSLIDECylinder::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  
  bRealSpaceDynamic = ( m_dynfRadius.GetDynamic() ||
			m_dynfZMin.GetDynamic() ||
			m_dynfZMax.GetDynamic() ||
			m_dynfThetaMax.GetDynamic() ||
			m_dyniEndCap.GetDynamic() ||
			m_dyniBeginCap.GetDynamic());
  bParameterSpaceDynamic = ( m_dyniZSlices.GetDynamic() ||
			     m_dyniThetaSlices.GetDynamic() );
  bTextureSpaceDynamic = ( m_pvTexture[0].GetDynamic() ||
			   m_pvTexture[1].GetDynamic() ||
			   m_pvTexture[2].GetDynamic() ||
			   m_pvTexture[3].GetDynamic() );

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || bParameterSpaceDynamic || bTextureSpaceDynamic);
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || bParameterSpaceDynamic);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic);
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic);
}

VOID CSLIDECylinder::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  m_dynfRadius.Update();
  if ( !( 0.0 < m_dynfRadius.GetValue() ) )
    {
      HandleError("Illegal radius value:  0.0 < %.3f", 
		  m_dynfRadius.GetValue());
      m_dynfRadius.SetValue(cfRadius);
    }

  m_dynfZMin.Update();
  // ZMin can take on any value
/*
  if ( !( -m_dynfRadius.GetValue() <= m_dynfZMin.GetValue() && 
	  m_dynfZMin.GetValue() <= m_dynfRadius.GetValue() ) )
    {
      HandleError("Illegal zmin value:  %.3f <= %.3f <= %.3f", 
		  -m_dynfRadius.GetValue(),
		  m_dynfZMin.GetValue(),
		  m_dynfRadius.GetValue());
      m_dynfZMin.SetValue(-m_dynfRadius.GetValue());
    }
*/

  m_dynfZMax.Update();
  // ZMax can take on any value
/*
  if ( !( -m_dynfRadius.GetValue() <= m_dynfZMax.GetValue() && 
	  m_dynfZMax.GetValue() <= m_dynfRadius.GetValue() ) )
    {
      HandleError("Illegal zmax value:  %.3f <= %.3f <= %.3f", 
		  -m_dynfRadius.GetValue(),
		  m_dynfZMax.GetValue(),
		  m_dynfRadius.GetValue());
      m_dynfZMax.SetValue(m_dynfRadius.GetValue());
    }
*/

  if ( m_dynfThetaMax.GetDynamic() )
    {
      m_dynfThetaMax.Update();
      m_dynfThetaMax.SetValue(DEGTORAD(m_dynfThetaMax.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(-360.0)) <= m_dynfThetaMax.GetValue() && 
	      m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(360.0)) ) )
	{
	  HandleError("Illegal thetamax value:  %.3f <= %.3f <= %.3f", 
		      -360.0,
		      RADTODEG(m_dynfThetaMax.GetValue()),
		      360.0);
	  m_dynfThetaMax.SetValue(cfThetaMax);
	}
    }

  m_dyniZSlices.Update();
  if ( !( 0 < m_dyniZSlices.GetValue() ) )
    {
      HandleError("Illegal zslices value:  %ld", 
		  m_dyniZSlices.GetValue());
      m_dyniZSlices.SetValue(cuiZSlices);
    }

  m_dyniThetaSlices.Update();
  if ( !( 0 < m_dyniThetaSlices.GetValue() ) )
    {
      HandleError("Illegal thetaslices value:  %ld", 
		  m_dyniThetaSlices.GetValue());
      m_dyniThetaSlices.SetValue(cuiThetaSlices);
    }

  m_dyniEndCap.Update();
  m_dyniBeginCap.Update();

  UINT_32 i;
  for ( i = 0; i < 4; i++ )
    {
      m_pvTexture[i].Update();
    }
}

VOID CSLIDECylinder::CalculateBound()
{
  CBound bound;
  Point pt;

  FLOAT pfZValues[2];
  UINT_32 uiZValue, uiZValues;
  FLOAT fRadius, fTheta, fThetaMin, fThetaMax, fThetaDelta;

  pfZValues[0] = m_dynfZMin.GetValue();
  pfZValues[1] = m_dynfZMax.GetValue();

  uiZValues = 2;

  fRadius = m_dynfRadius.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  fThetaDelta = DEGTORAD(90);
  if ( fThetaMax < fThetaMin )
    {
      fThetaDelta = -fThetaDelta;
    }

  for ( uiZValue = 0; uiZValue < uiZValues; uiZValue++ )
    {
      for ( fTheta = 0.0; 
	    ( fThetaMax < fThetaMin ) ? 
	      (fTheta > fThetaMax) : 
	      (fTheta < fThetaMax); 
	      fTheta += fThetaDelta )
	{
	  pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), pfZValues[uiZValue]);
	  bound.Insert(pt);
	}
      fTheta = fThetaMax;
      pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), pfZValues[uiZValue]);
      bound.Insert(pt);
    }
  
  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDECylinder::CalculateCDAVertices(CListInt *plistCDAVertex, 
					  const Matrix &mGroot_obj)
{
  Point pt;
  CDAVertex *pCDAVertex;

  UINT_32 i, uiThetaSlices;
  UINT_32 uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fThetaDelta;
  FLOAT fZ, fTheta;

  ASSERT( plistCDAVertex != NULL );
  
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  fZ = fZMin;
  for ( i = 0; i < 2; i++ )
    {
      fTheta = fThetaMin;
      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ);

	  pCDAVertex = AllocateCDAVertex();
	  ASSERT( pCDAVertex != NULL );
	  pCDAVertex->Init(mGroot_obj*pt);
	  plistCDAVertex->InsertLast(pCDAVertex, 0);
	  
	  fTheta += fThetaDelta;
	}
      
      fZ = fZMax;
    } 
}
#endif // __USE_CDA__

VOID CSLIDECylinder::CalculateQEMesh()
{
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ0, fZ1, fTheta, fTheta1;
  Point pt, p0, p1, p2;
  QEVertex* pqev;
  CObjectList<QEVertex> listQEVertex;

  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  ASSERT(m_pQEMesh == NULL);
  m_pQEMesh=AllocateQEMesh();
  ASSERT(m_pQEMesh != NULL);
  m_pQEMesh->Init();

  listQEVertex.Init();

  fZ0 = fZMin;
  fZ1 = fZ0 + fZDelta;
  for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
    {
      listQEVertex.RemoveAll();
      //gilBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pqev->Init(pt);
	    listQEVertex.InsertLast(pqev);
	    
	    pt[Z] = fZ0;
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pqev->Init(pt);
	    listQEVertex.InsertLast(pqev);	    
	    
	    fTheta += fThetaDelta;
	  }
      }
      //gilEnd();
      m_pQEMesh->MergeTriangleStrip(&listQEVertex);

      fZ0 = fZ1;
      fZ1 += fZDelta;
    } 

  if( m_dyniEndCap.GetValue() != 0)
    {
      listQEVertex.RemoveAll();

      pt.Set(0.0, 0.0, fZMax);
      pqev= AllocateQEVertex();
      ASSERT(pqev != NULL);
      pqev->Init(pt);
      listQEVertex.InsertLast(pqev);
      
      fTheta = fThetaMin;
      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	  pqev= AllocateQEVertex();
	  ASSERT(pqev != NULL);
	  pqev->Init(pt);
	  listQEVertex.InsertLast(pqev);
	  
	  fTheta += fThetaDelta;	  
	}      
      m_pQEMesh->MergeTriangleFan(&listQEVertex);
    }

  if( m_dyniBeginCap.GetValue() != 0)
    {
      listQEVertex.RemoveAll();

      pt.Set(0.0, 0.0, fZMin);
      pqev= AllocateQEVertex();
      ASSERT(pqev != NULL);
      pqev->Init(pt);
      listQEVertex.InsertLast(pqev);

      fTheta = fThetaMin;
      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	  pqev= AllocateQEVertex();
	  ASSERT(pqev != NULL);
	  pqev->Init(pt);
	  listQEVertex.InsertLast(pqev);
	  
	  fTheta -= fThetaDelta;	  
	}
      m_pQEMesh->MergeTriangleFan(&listQEVertex);
    }
}

VOID CSLIDECylinder::CalculateLeafQEMesh()
{
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDECylinder::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

VOID CSLIDECylinder::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  FLOAT fNormalDir;
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ0, fZ1, fTheta, fTheta1;
  Point pt, p0, p1, p2;
  Vector vNormal;

  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fZMin > fZMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fZMin > fZMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiZSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiZSlices;

  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];

  fZ0 = fZMin;
  fZ1 = fZ0 + fZDelta;
  for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;

	pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
	pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
	pvU[0] = ppvV[0][0];
	pvU[1] = ppvV[1][0];

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    vNormal.Set((FLOAT)cos(fTheta), (FLOAT)sin(fTheta), 0.0);
	    vNormal = fNormalDir * vNormal;

	    pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pt.Get());
	    
	    pt[Z] = fZ0;
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pt.Get());
	    
	    fTheta += fThetaDelta;

	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	  }
      }
      gilEnd();
      
      fZ0 = fZ1;
      fZ1 += fZDelta;

      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    } 

  if( m_dyniEndCap.GetValue() != 0)
    {
      gilBegin(GL_TRIANGLE_FAN);
      {
	p0.Set(0.0, 0.0, fZMax);

	fTheta = fThetaMin;
	fTheta1 = fTheta + fThetaDelta;

	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);
	    
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	gilVertex3fv(p0.Get());	    

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	    gilVertex3fv(p1.Get());	    
	    
	    fTheta += fThetaDelta;	  
	  }
      }
      gilEnd();
    }

  if( m_dyniBeginCap.GetValue() != 0)
    {
      gilBegin(GL_TRIANGLE_FAN);
      {
	p0.Set(0.0, 0.0, fZMin);

	fTheta = fThetaMin;
	fTheta1 = fTheta - fThetaDelta;

	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);
	    
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	gilVertex3fv(p0.Get());	    

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	    gilVertex3fv(p1.Get());	    
	    
	    fTheta -= fThetaDelta;	  
	  }
      }
      gilEnd();
    }
}

VOID CSLIDECylinder::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ0, fZ1, fTheta, fTheta1;
  Point pppt[2][2];
  Point p0, p1, p2;
  Vector vNormal;

  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiZSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiZSlices;

  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];

  fZ0 = fZMin;
  fZ1 = fZ0 + fZDelta;
  for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;

	pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
	pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
	pvU[0] = ppvV[0][0];
	pvU[1] = ppvV[1][0];

	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	pppt[1][1].Set(fRadius*(FLOAT)cos(fTheta+fThetaDelta), fRadius*(FLOAT)sin(fTheta+fThetaDelta), fZ1);

	vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	vNormal.Normalize();

	gilNormal3fv(vNormal.Get());
	gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	gilVertex3fv(pppt[1][0].Get());
	gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
	gilVertex3fv(pppt[0][0].Get());

	fTheta += fThetaDelta;

	pvU[0] += pvUDelta[0];
	pvU[1] += pvUDelta[1];

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);

	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pppt[1][1].Get());

	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pppt[0][1].Get());
	    
	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;

	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	  }
      }
      gilEnd();
      
      fZ0 = fZ1;
      fZ1 += fZDelta;

      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    } 

  if( m_dyniEndCap.GetValue() != 0)
    {
      gilBegin(GL_TRIANGLE_FAN);
      {
	p0.Set(0.0, 0.0, fZMax);
	gilVertex3fv(p0.Get());	    

	fTheta = fThetaMin;
	fTheta1 = fTheta + fThetaDelta;

	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);
	    
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	    gilVertex3fv(p1.Get());	    
	    
	    fTheta += fThetaDelta;	  
	  }
      }
      gilEnd();
    }

  if( m_dyniBeginCap.GetValue() != 0)
    {
      gilBegin(GL_TRIANGLE_FAN);
      {
	p0.Set(0.0, 0.0, fZMin);
	gilVertex3fv(p0.Get());	    

	fTheta = fThetaMin;
	fTheta1 = fTheta - fThetaDelta;

	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);
	    
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	    gilVertex3fv(p1.Get());	    
	    
	    fTheta -= fThetaDelta;	  
	  }
      }
      gilEnd();
    }
}

VOID CSLIDECylinder::RenderNormalsSmooth()
{
  FLOAT fNormalDir;
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ, fTheta, fTheta1;
  Point pt, p0, p1, p2;
  Vector vNormal;
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fZMin > fZMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fZMin > fZMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  gilBegin(GL_LINES);
  {
    fZ = fZMin;
    for ( uiZSlice = 0; uiZSlice <= uiZSlices; uiZSlice++ )
      {
	fTheta = fThetaMin;
	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    vNormal.Set((FLOAT)cos(fTheta), (FLOAT)sin(fTheta), 0.0);
	    vNormal = fNormalDir * vNormal;

	    pt.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ);
	    gilVertex3fv(pt.Get());
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());
	    
	    fTheta += fThetaDelta;
	  }
      
	fZ += fZDelta;
      } 
  }
  gilEnd();

  if( m_dyniEndCap.GetValue() != 0)
    {
      gilBegin(GL_LINES);
      {
	p0.Set(0.0, 0.0, fZMax);
	
	fTheta = fThetaMin;
	fTheta1 = fTheta + fThetaDelta;
	
	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);
	
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	gilVertex3fv(p0.Get());	    
	pt = p0 + vNormal;
	gilVertex3fv(pt.Get());	    

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	    gilVertex3fv(p1.Get());	    

	    pt = p1 + vNormal;
	    gilVertex3fv(pt.Get());	    

	    fTheta += fThetaDelta;	
	    fTheta1 += fThetaDelta;  
	  }
      }
      gilEnd();
    }

  if( m_dyniBeginCap.GetValue() != 0)
    {
      gilBegin(GL_LINES);
      {
	p0.Set(0.0, 0.0, fZMin);

	fTheta = fThetaMin;
	fTheta1 = fTheta - fThetaDelta;

	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);
	    
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	gilVertex3fv(p0.Get());	    
	pt = p0 + vNormal;
	gilVertex3fv(pt.Get());	    

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	    gilVertex3fv(p1.Get());	    

	    pt = p1 + vNormal;
	    gilVertex3fv(pt.Get());	    
	    
	    fTheta -= fThetaDelta;	  
	    fTheta1 -= fThetaDelta;
	  }
      }
      gilEnd();
    }
}

VOID CSLIDECylinder::RenderNormalsFlat()
{
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ0, fZ1, fTheta, fTheta1;
  Point pppt[2][2];
  Point pt, p0, p1, p2;
  Vector vNormal;
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  gilBegin(GL_LINES);
  {
    fZ0 = fZMin;
    fZ1 = fZ0 + fZDelta;
    for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
      {
	fTheta = fThetaMin;

	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);

	fTheta += fThetaDelta;

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);

	    pt = (pppt[1][0] + pppt[0][0] + pppt[1][1]) / 3.0;
	    gilVertex3fv(pt.Get());
	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());

	    pt = (pppt[0][1] + pppt[1][1] + pppt[0][0]) / 3.0;
	    gilVertex3fv(pt.Get());
	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());

	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;
	  }
      
	fZ0 = fZ1;
	fZ1 += fZDelta;
      } 
  }
  gilEnd();

  if( m_dyniEndCap.GetValue() != 0)
    {
      gilBegin(GL_LINES);
      {
	p0.Set(0.0, 0.0, fZMax);
	
	fTheta = fThetaMin;
	fTheta1 = fTheta + fThetaDelta;
	
	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);
	
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	    p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);

	    pt = (p0 + p1 + p2) / 3.0;
	    pt[W] = 1.0;
	    gilVertex3fv(pt.Get());	    

	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());	    

	    fTheta += fThetaDelta;	
	    fTheta1 += fThetaDelta;  
	  }
      }
      gilEnd();
    }

  if( m_dyniBeginCap.GetValue() != 0)
    {
      gilBegin(GL_LINES);
      {
	p0.Set(0.0, 0.0, fZMin);

	fTheta = fThetaMin;
	fTheta1 = fTheta - fThetaDelta;

	p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);
	    
	vNormal = (p1 - p0) * (p2 - p0);
	vNormal.Normalize();
	gilNormal3fv(vNormal.Get());

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	    p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);

	    pt = (p0 + p1 + p2) / 3.0;
	    pt[W] = 1.0;
	    gilVertex3fv(pt.Get());	    

	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());	    
	    
	    fTheta -= fThetaDelta;	  
	    fTheta1 -= fThetaDelta;
	  }
      }
      gilEnd();
    }
}

VOID CSLIDECylinder::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtCylinder], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtRadius],
	  m_dynfRadius.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtZMin],
	  m_dynfZMin.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtZMax],
	  m_dynfZMax.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtThetaMax],
	  RADTODEG(m_dynfThetaMax.GetValue()));
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtZSlices],
	  m_dyniZSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtThetaSlices],
	  m_dyniThetaSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtEndCap],
	  m_dyniEndCap.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtBeginCap],
	  m_dyniBeginCap.GetValue());
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f ) ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtTexture],
	  m_pvTexture[0](0), m_pvTexture[0](1), m_pvTexture[0](2), 
	  m_pvTexture[1](0), m_pvTexture[1](1), m_pvTexture[1](2), 
	  m_pvTexture[2](0), m_pvTexture[2](1), m_pvTexture[2](2), 
	  m_pvTexture[3](0), m_pvTexture[3](1), m_pvTexture[3](2));

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndCylinder]);
}

VOID CSLIDECylinder::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType /* eShading */,
				    SLIDELODType /* eLOD */ )
{
  RtFloat rtfRadius, rtfZMin, rtfZMax, rtfThetaMax;
  RtFloat prtvTex[4][2];

  ASSERT( fpOutputFile != NULL );

  rtfRadius = m_dynfRadius.GetValue();
  rtfZMin = m_dynfZMin.GetValue();
  rtfZMax = m_dynfZMax.GetValue();
  rtfThetaMax = RADTODEG(m_dynfThetaMax.GetValue());

  prtvTex[0][0] = m_pvTexture[0](0);
  prtvTex[0][1] = m_pvTexture[0](1);
  prtvTex[1][0] = m_pvTexture[1](0);
  prtvTex[1][1] = m_pvTexture[1](1);
  prtvTex[2][0] = m_pvTexture[3](0);
  prtvTex[2][1] = m_pvTexture[3](1);
  prtvTex[3][0] = m_pvTexture[2](0);
  prtvTex[3][1] = m_pvTexture[2](1);

  RiCylinder(rtfRadius, rtfZMin, rtfZMax, rtfThetaMax, RI_ST, (RtPointer)prtvTex, RI_NULL);

  // endcaps 

/*
  RiArchiveRecord(RI_COMMENT, "\nCylinder %.3f %.3f %.3f %.3f",
		  m_dynfRadius.GetValue(),
		  m_dynfZMin.GetValue(),
		  m_dynfZMax.GetValue(),
		  RADTODEG(m_dynfThetaMax.GetValue()));
*/
}

VOID CSLIDECylinder::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE CYLINDER: %s", m_pcGUIID);
}

VOID CSLIDECylinder::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, "   END SLIDE CYLINDER: %s", m_pcGUIID);
}

VOID CSLIDECylinder::WritePOVGeometry(FILE *fpOutputFile, 
				      CListInt *plistTransformIDs,
				      CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;

  FLOAT fNormalDir;
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ0, fZ1;
  FLOAT fTheta0, fTheta1;
  Point pt;
  CHAR *pcTransformID;
  CHAR *pcColorID;

  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();

  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();

  if ( ( (fZMin > fZMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fZMin > fZMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  if (fabs(fThetaMax) == cfThetaMax) 
    {
      fprintf(fpOutputFile, "cylinder { \n");
      fprintf(fpOutputFile, " < 0, 0, %f >, \n", fZMin);
      fprintf(fpOutputFile, " < 0, 0, %f >, \n", fZMax); 
      fprintf(fpOutputFile, " %f\n", fRadius);
      fprintf(fpOutputFile, " open\n");
    }
  else
    {
      fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
      fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

      fZ0 = fZMin;
      fZ1 = fZ0 + fZDelta;

      fprintf(fpOutputFile, "mesh { \n");

      for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
	{
	  fTheta0 = fThetaMin;
	  fTheta1 = fTheta0 + fThetaDelta;
	  for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	    {
	      fprintf(fpOutputFile, "triangle {");
	      fprintf(fpOutputFile, " <%f, %f, %f>,", fRadius*(FLOAT)sin(fTheta0), fRadius*(FLOAT)cos(fTheta0), fZ0);
	      fprintf(fpOutputFile, " <%f, %f, %f>,", fRadius*(FLOAT)sin(fTheta1), fRadius*(FLOAT)cos(fTheta1) ,fZ1);
	      fprintf(fpOutputFile, " <%f, %f, %f> ", fRadius*(FLOAT)sin(fTheta0), fRadius*(FLOAT)cos(fTheta0), fZ1);
	      fprintf(fpOutputFile, "}\n");
	      
	      fprintf(fpOutputFile, "triangle {");
	      fprintf(fpOutputFile, " <%f, %f, %f>,", fRadius*(FLOAT)sin(fTheta0), fRadius*(FLOAT)cos(fTheta0), fZ0);
	      fprintf(fpOutputFile, " <%f, %f, %f>,", fRadius*(FLOAT)sin(fTheta1), fRadius*(FLOAT)cos(fTheta1) ,fZ0);
	      fprintf(fpOutputFile, " <%f, %f, %f> ", fRadius*(FLOAT)sin(fTheta1), fRadius*(FLOAT)cos(fTheta1), fZ1);
	      fprintf(fpOutputFile, "}\n");

	      fTheta0 = fTheta1;
	      fTheta1 = fTheta1 + fThetaDelta;
	    }
	  fZ0 = fZ1;
	  fZ1 = fZ1 + fZDelta;
	}
    }
  //transform
  iter.Init(plistTransformIDs);
  for ( pcTransformID = (CHAR*)iter.PeekFirst(); 
	pcTransformID != NULL;
	pcTransformID = (CHAR*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, " transform %s\n", pcTransformID);
    }

  //color
  ASSERT(m_pCSLIDESurface != NULL);

  if (!(strcmp(GetSurfaceID(), "SLF_INHERIT")))
    {
      iter.Init(plistColors);
      pcColorID = (CHAR*)iter.PeekFirst();
      fprintf(fpOutputFile, " pigment { %s }\n", pcColorID);
    }
  else
    {
      fprintf(fpOutputFile, " pigment {");
      m_pCSLIDESurface->WritePOV(fpOutputFile);
      fprintf(fpOutputFile, " }\n");
    }
 
  fprintf(fpOutputFile, "}\n\n");
}

VOID CSLIDECylinder::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				       CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: CYLINDER\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");  
}

VOID CSLIDECylinder::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );
  
  UINT_32 uiZSlices, uiThetaSlices, uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fThetaMin, fThetaMax;
  FLOAT fZDelta, fThetaDelta;
  FLOAT fZ0, fZ1, fTheta, fTheta1;
  Point pppt[2][2];
  Point p0, p1, p2;
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  fZDelta = (fZMax - fZMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  fZ0 = fZMin;
  fZ1 = fZ0 + fZDelta;
  
  if( mTransform.Determinant() > 0 ) // Don't Mirror
    {
      for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
	{
	  fTheta = fThetaMin;
	  
	  // Must seed the for loop with the left most two points
	  pppt[0][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);
	  pppt[1][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	  
	  pppt[0][0] = mTransform * pppt[0][0];
	  pppt[1][0] = mTransform * pppt[1][0];
	  
	  fTheta += fThetaDelta;
	  
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      pppt[1][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	      pppt[0][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);
	      
	      pppt[1][1] = mTransform * pppt[1][1];
	      pppt[0][1] = mTransform * pppt[0][1];

	      WriteSTLTriangle(fpOutputFile, pppt[1][0], pppt[0][0], pppt[1][1]);
	      WriteSTLTriangle(fpOutputFile, pppt[0][1], pppt[1][1], pppt[0][0]);
	      
	      pppt[0][0] = pppt[0][1];
	      pppt[1][0] = pppt[1][1];
	      fTheta += fThetaDelta;
	    }
	  fZ0 = fZ1;
	  fZ1 += fZDelta;
	}

      if( m_dyniEndCap.GetValue() != 0)
	{
	  p0.Set(0.0, 0.0, fZMax);
	  p0 = mTransform * p0;

	  fTheta = fThetaMin;
	  fTheta1 = fTheta + fThetaDelta;      
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	      p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);

	      p1 = mTransform * p1;
	      p2 = mTransform * p2;
	      
	      WriteSTLTriangle(fpOutputFile, p0, p1, p2);
	      
	      fTheta += fThetaDelta;	  
	      fTheta1 += fThetaDelta;
	    }
	}
      
      if( m_dyniBeginCap.GetValue() != 0)
	{
	  p0.Set(0.0, 0.0, fZMin);
	  p0 = mTransform * p0;
	  
	  fTheta = fThetaMin;
	  fTheta1 = fTheta + fThetaDelta;
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	      p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);
	      
	      p1 = mTransform * p1;
	      p2 = mTransform * p2;

	      WriteSTLTriangle(fpOutputFile, p0, p2, p1);
	      
	      fTheta += fThetaDelta;	  
	      fTheta1 += fThetaDelta;  
	    }
	}
    }
  else // Mirror
    {
      for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
	{
	  fTheta = fThetaMin;
	  
	  // Must seed the for loop with the left most two points
	  pppt[0][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);
	  pppt[1][0].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	  
	  pppt[0][0] = mTransform * pppt[0][0];
	  pppt[1][0] = mTransform * pppt[1][0];
	  
	  fTheta += fThetaDelta;
	  
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      pppt[1][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ1);
	      pppt[0][1].Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZ0);
	      
	      pppt[1][1] = mTransform * pppt[1][1];
	      pppt[0][1] = mTransform * pppt[0][1];

	      WriteSTLTriangle(fpOutputFile, pppt[1][1], pppt[0][0], pppt[1][0]);
	      WriteSTLTriangle(fpOutputFile, pppt[0][0], pppt[1][1], pppt[0][1]);
	      
	      pppt[0][0] = pppt[0][1];
	      pppt[1][0] = pppt[1][1];
	      fTheta += fThetaDelta;
	    }
	  fZ0 = fZ1;
	  fZ1 += fZDelta;
	}

      if( m_dyniEndCap.GetValue() != 0)
	{
	  p0.Set(0.0, 0.0, fZMax);
	  p0 = mTransform * p0;

	  fTheta = fThetaMin;
	  fTheta1 = fTheta + fThetaDelta;      
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMax);
	      p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMax);

	      p1 = mTransform * p1;
	      p2 = mTransform * p2;
	      
	      WriteSTLTriangle(fpOutputFile, p2, p1, p0);
	      
	      fTheta += fThetaDelta;	  
	      fTheta1 += fThetaDelta;
	    }
	}
      
      if( m_dyniBeginCap.GetValue() != 0)
	{
	  p0.Set(0.0, 0.0, fZMin);
	  p0 = mTransform * p0;
	  
	  fTheta = fThetaMin;
	  fTheta1 = fTheta + fThetaDelta;
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      p1.Set(fRadius*(FLOAT)cos(fTheta), fRadius*(FLOAT)sin(fTheta), fZMin);
	      p2.Set(fRadius*(FLOAT)cos(fTheta1), fRadius*(FLOAT)sin(fTheta1), fZMin);
	      
	      p1 = mTransform * p1;
	      p2 = mTransform * p2;

	      WriteSTLTriangle(fpOutputFile, p1, p2, p0);
	      
	      fTheta += fThetaDelta;	  
	      fTheta1 += fThetaDelta;  
	    }
	}
    }
}

VOID CSLIDECylinder::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Cylinder_%s", GetGUIID());
}

VOID CSLIDECylinder::WriteDotNode(FILE * fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = diamond ];\n", GetGUIID());
}

VOID CSLIDECylinder::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

VOID CSLIDECylinder::GetParameters(Tcl_Interp *pInterp)
{
  INT_32 iCode;
  CHAR pcBuffer[1000];
  CHAR pcSurface[35];

  //pInterp = CSLIDEWorld::GetInterp();

  if( !strcmp(m_pCSLIDESurface->GetGUIID(), "SLF_INHERIT" ) )
      sprintf(pcSurface, "inherit");
  else 
      sprintf(pcSurface, m_pCSLIDESurface->GetGUIID());

  sprintf(pcBuffer, "Editor:SetParameterValues cylinder {{radius %.3f} {zmin %.3f} {zmax %.3f} {zslices %ld} {thetamax %.3f} {thetaslices %ld} {solid %s} {surface %s} {shading %s} {lod %s} {ribbegin %s} {ribend %s}}",
	  m_dynfRadius.GetValue(),
	  m_dynfZMin.GetValue(),
	  m_dynfZMax.GetValue(),
	  m_dyniZSlices.GetValue(),
	  RADTODEG(m_dynfThetaMax.GetValue()),
	  m_dyniThetaSlices.GetValue(),
	  gcppcSLIDEGUISolidReservedWords[m_dyniSolid.GetValue()],
	  pcSurface,
	  gcppcSLIDEGUIShadingReservedWords[m_dyniShading.GetValue()],
	  gcppcSLIDEGUILODReservedWords[m_dyniLOD.GetValue()],
	  m_pcRIBBegin,
	  m_pcRIBEnd);
	  
  iCode = Tcl_Eval(pInterp, pcBuffer);
  //ASSERT( iCode == TCL_OK );
  if ( iCode != TCL_OK )
    {
      printf( "%s\n", pcBuffer );
      printf("Error %s \n", pInterp->result);
    }
}

UINT_32 CSLIDECylinder::SetParameter(CSLIDEWorld *pCSLIDEWorld, 
				 SLIDEReservedWordType uiParameter, CHAR **argv)
{
  switch(uiParameter) 
    {
    case sliderwtRadius:
      m_dynfRadius.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtZMin:
      m_dynfZMin.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtZMax:
      m_dynfZMax.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtZSlices:
      m_dyniZSlices.SetValue(atoi(argv[0]));
      return 1;
      break;
    case sliderwtThetaMax:
      m_dynfThetaMax.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtThetaSlices:
      m_dyniThetaSlices.SetValue(atoi(argv[0]));
      return 1;
      break;
    default:
      return CSLIDEAtomNode::SetParameter(pCSLIDEWorld, uiParameter, argv);
      break;
    }
}
