#include "slidefog.h"
#include "slideworld.h"

// Default Values based on default values in SLIDECamera for Near and Far Z values
static const FLOAT cfNear = (FLOAT)-0.02;
static const FLOAT cfFar = (FLOAT)-100.0;

static const SLIDEFogType ceType = slidefogtSLF_LINEAR;
static const FLOAT cfDensity = -((FLOAT)0.25 / cfFar);
static const FLOAT cfStart = -cfNear;
static const FLOAT cfEnd = -cfFar;
static const VECTOR4D cvColor = { 0.0, 0.0, 0.0, 0.0 };

//////////////////////////////////////////////////////////////////////
// CSLIDEFog Class
//

CSLIDEFog::CSLIDEFog()
{
  m_dyniType.SetValue((INT_32)ceType);
  m_dynfDensity.SetValue(cfDensity);
  m_dynfStart.SetValue(cfStart);
  m_dynfEnd.SetValue(cfEnd);
  m_vColor.Set(cvColor);
}

CSLIDEFog::~CSLIDEFog()
{
}

VOID CSLIDEFog::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniType.SetValue((INT_32)ceType);
  m_dynfDensity.SetValue(cfDensity);
  m_dynfStart.SetValue(cfStart);
  m_dynfEnd.SetValue(cfEnd);
  m_vColor.Set(cvColor);
}

VOID CSLIDEFog::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEFog::Initialized()
{
  return CSLIDEPrimitive::Initialized();
}

BOOL CSLIDEFog::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  bCorrect = CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld);

  return bCorrect;
}

BOOL CSLIDEFog::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( 0.0 <= m_vColor[R] && m_vColor[R] <= 1.0 &&
	   0.0 <= m_vColor[G] && m_vColor[G] <= 1.0 &&
	   0.0 <= m_vColor[B] && m_vColor[B] <= 1.0 );
}

VOID CSLIDEFog::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CSLIDEPrimitive::Remove(pCSLIDEWorld);

  pCSLIDEWorld->RemoveFog(this);
  Uninit();
}

VOID CSLIDEFog::CheckDynamic()
{
  SetDynamic(m_dyniType.GetDynamic() ||
	     m_dynfDensity.GetDynamic() ||
	     m_dynfStart.GetDynamic() ||
	     m_dynfEnd.GetDynamic() ||
	     m_vColor.GetDynamic());
}

VOID CSLIDEFog::UpdateState()
{
  m_dyniType.Update();
  m_dynfDensity.Update();
  m_dynfStart.Update();
  m_dynfEnd.Update();
  m_vColor.Update();
}

VOID CSLIDEFog::Render()
{
  GLint iMode;

  switch ( m_dyniType.GetValue() )
    {
    case slidefogtSLF_LINEAR:
      iMode = GL_LINEAR;
      break;
    case slidefogtSLF_EXP:
      iMode = GL_EXP;
      break;
    case slidefogtSLF_EXP2:
      iMode = GL_EXP2;
      break;
    default:
      ASSERT( FALSE );
      iMode = GL_LINEAR;
      break;
    }

  gilEnable(GL_FOG);
  gilFogi(GL_FOG_MODE, iMode);
  gilFogf(GL_FOG_DENSITY, m_dynfDensity.GetValue());
  gilFogf(GL_FOG_START, m_dynfStart.GetValue());
  gilFogf(GL_FOG_END, m_dynfEnd.GetValue());
  gilFogfv(GL_FOG_COLOR, m_vColor.Get());
}

VOID CSLIDEFog::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n",
	  gcppcSLIDEReservedWords[sliderwtFog], m_pcGUIID); 

  ASSERT( 0 <= m_dyniType.GetValue() && 
	  m_dyniType.GetValue() < slidefogtSize );
  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtType], 
	  gcppcSLIDEFogReservedWords[m_dyniType.GetValue()]);
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtDensity], 
	  m_dynfDensity.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtStart], 
	  m_dynfStart.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtEnd], 
	  m_dynfEnd.GetValue());
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtColor], 
	  m_vColor[R], m_vColor[G], m_vColor[B]); 
  CSLIDEPrimitive::WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "%s\n",
	  gcppcSLIDEReservedWords[sliderwtEndFog]);
}

VOID CSLIDEFog::WriteRIB(FILE *fpOutputFile)
{
  SLIDEFogType eFog;
  RtFloat rtfMinDistance, rtfMaxDistance, rtfDistance;
  RtColor rtcBackground;

  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE FOG: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);

  rtfMinDistance = m_dynfStart.GetValue();
  rtfMaxDistance = m_dynfEnd.GetValue();

  rtfDistance = 1.0 / m_dynfDensity.GetValue();

  rtcBackground[0] = m_vColor(0);
  rtcBackground[1] = m_vColor(1);
  rtcBackground[2] = m_vColor(2);

  eFog = (SLIDEFogType)(m_dyniType.GetValue());
  switch ( eFog )
    {
    case slidefogtSLF_EXP:
    case slidefogtSLF_EXP2:
      RiAtmosphere(RI_FOG, RI_BACKGROUND, rtcBackground, RI_DISTANCE, &rtfDistance, RI_NULL); 
      break;
    case slidefogtSLF_LINEAR:
    default:
      RiAtmosphere(RI_DEPTHCUE, RI_MINDISTANCE, &rtfMinDistance, RI_MAXDISTANCE, &rtfMaxDistance, RI_BACKGROUND, rtcBackground, RI_NULL); 
      break;
    }

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE FOG: %s", m_pcGUIID);
}

VOID CSLIDEFog::WritePOV(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEFog::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEFog::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDEFog::WriteDotEdges(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );
}




