#include "slideheightfield.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "qemesh.h"
#include "qevertex.h"
#include "qewrite.h"
#include "stl.h"

static const UINT_32 cuiUSlices = 5;
static const UINT_32 cuiVSlices = 5;
static const FLOAT cfUMin = 0;
static const FLOAT cfVMin = 0;
static const FLOAT cfUMax = 1.0;
static const FLOAT cfVMax = 1.0;

//////////////////////////////////////////////////////////////////////
// CSLIDEHeightfield Class
//

//CHECKED
CSLIDEHeightfield::CSLIDEHeightfield()
{
  m_dyniUSlices.SetValue(cuiUSlices);
  m_dyniVSlices.SetValue(cuiVSlices);
  m_uiUSlices = cuiUSlices;
  m_uiVSlices = cuiVSlices;
  m_uiUSlicesPrev = cuiUSlices;

  m_dynfUMin.SetValue(cfUMin);
  m_dynfVMin.SetValue(cfVMin);
  m_dynfUMax.SetValue(cfUMax);
  m_dynfVMax.SetValue(cfVMax);
  m_fUMin = cfUMin;
  m_fVMin = cfVMin;
  m_fUMax = cfUMax;
  m_fVMax = cfVMax;

  m_ppSurfacePoint = NULL;

}

CSLIDEHeightfield::~CSLIDEHeightfield()
{
}

//CHECKED
VOID CSLIDEHeightfield::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniUSlices.SetValue(cuiUSlices);
  m_dyniVSlices.SetValue(cuiVSlices);
  m_uiUSlices = cuiUSlices;
  m_uiVSlices = cuiVSlices;
 
  m_dynfUMin.SetValue(cfUMin);
  m_dynfVMin.SetValue(cfVMin);
  m_dynfUMax.SetValue(cfUMax);
  m_dynfVMax.SetValue(cfVMax);
  m_fUMin = cfUMin;
  m_fVMin = cfVMin;
  m_fUMax = cfUMax;
  m_fVMax = cfVMax;

  ReallocateSurfacePoints();

  CalculateSurface();
}

//CHECKED
VOID CSLIDEHeightfield::Uninit()
{
  CSLIDEAtomNode::Uninit();
  DeallocateSurfacePoints();
}

BOOL CSLIDEHeightfield::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

//CHECKED
VOID CSLIDEHeightfield::DeallocateSurfacePoints()
{
  UINT_32 i;
  
  if ( m_ppSurfacePoint != NULL )
    {
      for( i = 0; i < m_uiUSlicesPrev + 1; i++ ) 
	{
	  delete[] m_ppSurfacePoint[i];
	}
      delete[] m_ppSurfacePoint;
      m_ppSurfacePoint = NULL;
    }
}

//CHECKED
VOID CSLIDEHeightfield::ReallocateSurfacePoints()
{
  UINT_32 i;

  if( m_ppSurfacePoint != NULL )
    {
      DeallocateSurfacePoints();
    }

  m_ppSurfacePoint = new Point*[m_uiUSlices+1];
  ASSERT(m_ppSurfacePoint != NULL);

  for(i = 0; i < m_uiUSlices + 1; i++ ) 
    {
      m_ppSurfacePoint[i] = new Point[m_uiVSlices+1];
      ASSERT(m_ppSurfacePoint[i] != NULL);
    }
}

//CHECKED
BOOL CSLIDEHeightfield::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  BOOL bCorrect = TRUE;

  if( (1 > m_uiUSlices) || (m_uiUSlices > 1000)  )
    {
      HandleError("uslices must be between 1 and 1000" );
      bCorrect = FALSE;
    }

  if( (1 > m_uiVSlices) || (m_uiVSlices > 1000) )
    {
      HandleError("vslices must be between 1 and 1000" );
      bCorrect = FALSE;
    }

  return bCorrect;
}

//CHECKED
BOOL CSLIDEHeightfield::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dyniUSlices.GetValue() >= 0 ) &&
	   ( m_dyniVSlices.GetValue() >= 0 ));
}

//CHECKED
VOID CSLIDEHeightfield::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  if( !GetDynamic() )
    {
      SetDynamic(m_dyniUSlices.GetDynamic() ||
		 m_dyniVSlices.GetDynamic() ||
		 m_dynfUMin.GetDynamic() ||
		 m_dynfVMin.GetDynamic() ||
		 m_dynfUMax.GetDynamic() ||
		 m_dynfVMax.GetDynamic() );
    }
}

//CHECKED
VOID CSLIDEHeightfield::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;

  BOOL bBoundEmpty;

  bRealSpaceDynamic = FALSE;
  bParameterSpaceDynamic = ( m_dyniUSlices.GetDynamic() ||
			     m_dyniVSlices.GetDynamic() ||
			     m_dynfUMin.GetDynamic() ||
			     m_dynfVMin.GetDynamic() ||
			     m_dynfUMax.GetDynamic() ||
			     m_dynfVMax.GetDynamic() );
  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     m_pCSLIDESurface->GetDynamic() );
  SetNormalsDynamic(FALSE);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic);
  SetBoundEmpty(bBoundEmpty);
  SetBoundDynamic(bRealSpaceDynamic);
}

//CHECKED
VOID CSLIDEHeightfield::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  m_dyniUSlices.Update();
  if ( (1 > m_dyniUSlices.GetValue() || (m_dyniUSlices.GetValue() > 1000)) )
    {
      HandleError("uslices must be between 1 and 1000" );
      m_dyniUSlices.SetValue(cuiUSlices);
    }
  m_dyniVSlices.Update();
  if ( (1 > m_dyniVSlices.GetValue() || (m_dyniVSlices.GetValue() > 1000)) )
    {
      HandleError("vslices must be between 1 and 1000" );
      m_dyniVSlices.SetValue(cuiVSlices);
    }

  m_dynfUMin.Update();
  m_dynfVMin.Update();
  m_dynfUMax.Update();
  m_dynfVMax.Update();

  if( ((UINT_32)m_dyniUSlices.GetValue() != m_uiUSlices) || 
      ((UINT_32)m_dyniVSlices.GetValue() != m_uiVSlices) ||
      ((FLOAT)m_dynfUMin.GetValue() != m_fUMin) ||
      ((FLOAT)m_dynfVMin.GetValue() != m_fVMin) ||
      ((FLOAT)m_dynfUMax.GetValue() != m_fUMax) ||
      ((FLOAT)m_dynfVMax.GetValue() != m_fVMax)
      )
    {
      m_uiUSlicesPrev = m_uiUSlices;
      m_uiUSlices = m_dyniUSlices.GetValue();
      m_uiVSlices = m_dyniVSlices.GetValue();
      m_fUMin = m_dynfUMin.GetValue();
      m_fVMin = m_dynfVMin.GetValue();
      m_fUMax = m_dynfUMax.GetValue();
      m_fVMax = m_dynfVMax.GetValue();
      
      ReallocateSurfacePoints();
      CalculateSurface();
    }
}

//CHECKED
VOID CSLIDEHeightfield::CalculateBound()
{
  UINT_32 i, j;
  CBound bound;
  Point pt;

  for( i = 0; i < m_uiUSlices; i++ )
    {
      for( j = 0; j < m_uiVSlices; j++ )
	{
	  pt=m_ppSurfacePoint[i][j];
	  bound.Insert(pt);
	}
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

    }
}

#ifdef __USE_CDA__
VOID CSLIDEHeightfield::CalculateCDAVertices(CListInt *plistCDAVertex, 
					      const Matrix &mGroot_obj)
{

}
#endif // __USE_CDA__

//CHECKED
VOID CSLIDEHeightfield::CalculateSurface()
{
  UINT_32 uiUSlice, uiVSlice;
  FLOAT fU, fV, fUDelta, fVDelta;
  CHAR pcBuffer[200];
  Tcl_Interp *pInterp;

  pInterp = CSLIDEWorld::GetInterp();

  fU=m_dynfUMin.GetValue();
  fV=m_dynfVMin.GetValue();
  fUDelta=(m_dynfUMax.GetValue()-m_dynfUMin.GetValue())
    /(FLOAT)m_dyniUSlices.GetValue();
  fVDelta=(m_dynfVMax.GetValue()-m_dynfVMin.GetValue())
    /(FLOAT)m_dyniVSlices.GetValue();

  for( uiUSlice = 0; uiUSlice <= m_uiUSlices; uiUSlice++ )
    {
      fV=m_dynfVMin.GetValue();
      sprintf(pcBuffer,"%f",fU);
      Tcl_SetVar(pInterp,"u",pcBuffer,TCL_GLOBAL_ONLY);
      for( uiVSlice = 0; uiVSlice <= m_uiVSlices; uiVSlice++ )
	{
	  sprintf(pcBuffer,"%f",fV);
	  Tcl_SetVar(pInterp,"v",pcBuffer,TCL_GLOBAL_ONLY);
	  m_dynvFunc.Update();
	  //???
	  m_ppSurfacePoint[uiUSlice][uiVSlice].Set
	    (m_dynvFunc[0],
	     m_dynvFunc[1],
	     m_dynvFunc[2]);
	  fV += fVDelta;
	}
      
      fU += fUDelta;
    }
}

//CHECKED
VOID CSLIDEHeightfield::CalculateQEMesh()
{
  QEVertex* pqev;
  CObjectList<QEVertex> listQEVertex;
  Point pt;

  UINT_32 i, j;
   
  ASSERT(m_pQEMesh == NULL);
  m_pQEMesh=AllocateQEMesh();
  ASSERT(m_pQEMesh != NULL);
  m_pQEMesh->Init();
  
  listQEVertex.Init();
  
  for( i = 0; i < m_uiUSlices; i++ )
    {
      listQEVertex.RemoveAll();
      //gilBegin(GL_TRIANGLE_STRIP);
      {
	for( j = 0; j < m_uiVSlices+1; j++ )
	  {
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pt=m_ppSurfacePoint[i][j];
	    pqev->Init(pt);
	    listQEVertex.InsertLast(pqev);
	    
	    pqev= AllocateQEVertex();
	    ASSERT(pqev != NULL);
	    pt=m_ppSurfacePoint[i+1][j];
	    pqev->Init(pt);
	    listQEVertex.InsertLast(pqev);
	    
	  }
      }
      m_pQEMesh->MergeTriangleStrip(&listQEVertex);
      //gilEnd();
    }
}

//TODO: exact copy of rendergeometryflat
VOID CSLIDEHeightfield::RenderGeometrySmooth()
{
  UINT_32 i, j;
  Point p1, p2, p3, p4;
  Vector vNormal;
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvU[2][2];
  Vector pvV[2];
  
  for( i = 0; i < m_uiUSlices; i++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	p1 = m_ppSurfacePoint[i][0];
	p2 = m_ppSurfacePoint[i+1][0];
	
	gilVertex3fv(p1.Get());
	gilTexCoord2f(pvV[1][0], pvV[1][1]);	
	gilVertex3fv(p2.Get());
	
	for( j = 1; j < m_uiVSlices+1; j++)	
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = m_ppSurfacePoint[i+1][j];

	    vNormal = (p2 - p1) * (p3 - p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p2 - p3) * (p4 - p3);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	    
	  }
      }
      gilEnd();
    }
}

VOID CSLIDEHeightfield::RenderGeometryFlat()
{
  UINT_32 i, j;
  Point p1, p2, p3, p4;
  Vector vNormal;
  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvU[2][2];
  Vector pvV[2];

  for( i = 0; i < m_uiUSlices; i++ )
    {
      gilBegin(GL_TRIANGLE_STRIP);
      {
	p1 = m_ppSurfacePoint[i][0];
	p2 = m_ppSurfacePoint[i+1][0];
	
	gilVertex3fv(p1.Get());

	gilVertex3fv(p2.Get());
	
	for( j = 1; j < m_uiVSlices+1; j++)	
	  {
	    p3 = m_ppSurfacePoint[i][j];
	    p4 = m_ppSurfacePoint[i+1][j];

	    //	    gilTexCoord2f((FLOAT)i/(FLOAT)m_uiUSlices, (FLOAT)j/(FLOAT)m_uiVSlices);	
	    
	    vNormal = (p2 - p1) * (p3 - p1);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p3.Get());
	    
	    vNormal = (p2 - p3) * (p4 - p3);
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(p4.Get());
	    
	    p1 = p3;
	    p2 = p4;
	    
	  }
      }
      gilEnd();
    }
}

//TODO
VOID CSLIDEHeightfield::RenderNormalsSmooth()
{
  /*
  UINT_32 i, j;
  Point pt, p1, p2, p3;
  
  gilBegin(GL_LINES);
  {
    for(i = 0; i < m_uiUSlices + 1; i++ ) 
      {
	for(j = 0; j < m_uiVSlices + 1; j++ )
	  {
	    gilVertex3fv(m_ppSurfacePoint[i][j].Get());
	    pt = m_ppSurfacePoint[i][j] + m_ppvSurfaceNormal[i][j] * m_fNormalLength;
	    gilVertex3fv(pt.Get());
	  }
      }
      }
  gilEnd();
  */
}

VOID CSLIDEHeightfield::RenderNormalsFlat()
{
/*
  if( m_dynfOffset.GetValue() != 0.0 )
    {
      RenderOffsetNormalsFlat();
      return;
    }

  UINT_32 i, j;
  Point pt, p1, p2, p3, p4;
  Vector v;
  
  gilBegin( GL_LINES );
  {
    for(i = 0; i < m_uiUSlices; i++ ) 
      {
	for(j = 0; j < m_uiVSlices; j++ )
	  {
	    p1 = m_ppSurfacePoint[i][j];
	    p2 = m_ppSurfacePoint[i+1][j];
	    p3 = m_ppSurfacePoint[i][j+1];	  
	    p4 = m_ppSurfacePoint[i+1][j+1];
	    
	    pt = (p1 + p2 + p3) / 3.0;
	    gilVertex3fv(pt.Get());
	    
	    v = (p2 - p1) * (p3 - p1);
	    v.Normalize();
	    pt = pt + v * m_fNormalLength;
	    gilVertex3fv(pt.Get());
	    
	    ///
	    
	    pt = (p2 + p3 + p4) / 3.0;
	    gilVertex3fv(pt.Get());
	    
	    v = (p2 - p3) * (p4 - p3);
	    v.Normalize();
	    pt = pt + v * m_fNormalLength;
	    gilVertex3fv(pt.Get());
	  }
      }
  }
  gilEnd();
  */
}

VOID CSLIDEHeightfield::WriteSLIDE(FILE *fpOutputFile)
{
  /*
  UINT_32 i, j;
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtHeightfield], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUOrder],
	  m_uiUOrder);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUKnots],
	  m_uiUKnots);
  fprintf(fpOutputFile, "  %s (",
	  gcppcSLIDEReservedWords[sliderwtUKnotList]);

  for( i = 0; i < m_uiUKnots; i++ )
    {
      fprintf(fpOutputFile, " %.3f", m_pfUKnot[i]);
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVOrder],
	  m_uiVOrder);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVKnots],
	  m_uiVKnots);
  fprintf(fpOutputFile, "  %s (",
	  gcppcSLIDEReservedWords[sliderwtVKnotList]);

  for( i = 0; i < m_uiVKnots; i++ )
    {
      fprintf(fpOutputFile, " %.3f", m_pfVKnot[i]);
    }
  fprintf(fpOutputFile, " )\n");


  if( m_plistControlPointIDs != NULL )
    {
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      iter.Init(m_plistControlPointIDs);
      
      for ( pcID = (CHAR*)(iter.PeekFirst()), i = 0;
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()) )
	{
	  if( i == m_uiUControlPoints )
	    {
	      fprintf(fpOutputFile, "\n\t\t    " );
	      i = 0;
	    }
	  fprintf(fpOutputFile, " %s", pcID );
	  i++;
	}
    }
  else
    {
      const FLOAT* pt;

      fprintf(fpOutputFile, "# Default Control Points:\n");
      fprintf(fpOutputFile, "# %s (\n#\t",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
	    {
	      pt = m_ppControlPoint[i][m_uiVControlPoints-1-j].Get();
	      fprintf(fpOutputFile, "(%.3f %.3f %.3f)", pt[X], pt[Y], pt[Z] );

	      if( (j != m_uiVControlPoints-1) && ( i == m_uiUControlPoints-1) )
		{
		  fprintf(fpOutputFile, "\n#\t" );
		}
	    }
	}
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUSlices],
	  m_uiUSlices);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVSlices],
	  m_uiVSlices);
  fprintf(fpOutputFile, "  %s (%.3f %.3f) (%.3f %.3f) (%.3f %.3f) (%.3f %.3f)\n",
	  gcppcSLIDEReservedWords[sliderwtTexture],
	  m_pvTexture[0][X], m_pvTexture[0][Y],
	  m_pvTexture[1][X], m_pvTexture[1][Y],
	  m_pvTexture[2][X], m_pvTexture[2][Y],
	  m_pvTexture[3][X], m_pvTexture[3][Y]);
  fprintf(fpOutputFile, "  %s %.3f %ld\n",
	  gcppcSLIDEReservedWords[sliderwtOffset],
	  m_dynfOffset.GetValue(), m_dyniOffsetConnected.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawControls],
	  m_dyniDrawControls.GetValue());
  
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndHeightfield]);
	  */
}

VOID CSLIDEHeightfield::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  /*
  UINT_32 i, j;

  if( (m_uiUOrder == 4) && (m_uiVOrder == 4) )
    {
      const FLOAT *pt;
      char *pointIds = new char[800];
      char *thisPoint = new char[50];

      RiArchiveRecord(RI_COMMENT, "\nBasis \"b-spline\" 1 \"b-spline\" 1\n");

      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
      	    {
	      pt = m_ppControlPoint[i][j].Get();
	      sprintf( thisPoint, " %.6f %.6f %.6f", pt[X], pt[Y], pt[Z] );
	      strcat( pointIds, thisPoint );
	    }
	}

      RiArchiveRecord(RI_COMMENT, "\nPatch \"bicubic\" \"P\" [%s ] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f]\n", 
      		      pointIds,
		      m_pvTexture[3][X], m_pvTexture[3][Y], 
		      m_pvTexture[2][X], m_pvTexture[2][Y],
		      m_pvTexture[0][X], m_pvTexture[0][Y], 
		      m_pvTexture[1][X], m_pvTexture[1][Y]);  

      delete[] pointIds;
      delete[] thisPoint;
    }
  else 
    {
      const FLOAT *p1, *p2, *p3, *p4, *v1, *v2, *v3, *v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvU[2][2];
      Vector pvV[2];

      pvUDelta[0] = (m_pvTexture[2] - m_pvTexture[3]) / (FLOAT)m_uiUSlices;
      pvUDelta[1] = (m_pvTexture[1] - m_pvTexture[0]) / (FLOAT)m_uiUSlices;
      
      ppvU[0][0] = m_pvTexture[3];
      ppvU[0][1] = m_pvTexture[0];
      ppvU[1][0] = ppvU[0][0] + pvUDelta[0];
      ppvU[1][1] = ppvU[0][1] + pvUDelta[1];

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  pvVDelta[0] = (ppvU[0][1] - ppvU[0][0]) / (FLOAT)m_uiVSlices;
	  pvVDelta[1] = (ppvU[1][1] - ppvU[1][0]) / (FLOAT)m_uiVSlices;
	  pvV[0] = ppvU[0][0];
	  pvV[1] = ppvU[1][0];

	  for( j = 0; j < m_uiVSlices; j++)
	    {
	      p1 = m_ppSurfacePoint[i][j].Get();
	      p2 = m_ppSurfacePoint[i+1][j].Get();
	      p3 = m_ppSurfacePoint[i][j+1].Get();
	      p4 = m_ppSurfacePoint[i+1][j+1].Get();
	      
	      if( (eShading == slideshtSLF_GOURAUD) || (eShading == slideshtSLF_PHONG) )
		{
		  v1 = m_ppvSurfaceNormal[i][j].Get();
		  v2 = m_ppvSurfaceNormal[i+1][j].Get();
		  v3 = m_ppvSurfaceNormal[i][j+1].Get();
		  v4 = m_ppvSurfaceNormal[i+1][j+1].Get();

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], p2[X], p2[Y], p2[Z], p3[X], p3[Y], p3[Z],
				  v1[X], v1[Y], v1[Z], v2[X], v2[Y], v2[Z], v3[X], v3[Y], v3[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1], 
				  (pvV[0] + pvVDelta[0])[0], (pvV[0] + pvVDelta[0])[1]);

		  pvV[0] += pvVDelta[0];

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], p2[X], p2[Y], p2[Z], p4[X], p4[Y], p4[Z],
				  v3[X], v3[Y], v3[Z], v2[X], v2[Y], v2[Z], v4[X], v4[Y], v4[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1],
				  (pvV[1] + pvVDelta[1])[0], (pvV[1] + pvVDelta[1])[1]);
		  pvV[1] += pvVDelta[1];
		}
	      else
		{
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], p2[X], p2[Y], p2[Z], p3[X], p3[Y], p3[Z],
				  pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1], 
				  (pvV[0] + pvVDelta[0])[0], (pvV[0] + pvVDelta[0])[1]);
		  pvV[0] += pvVDelta[0];

		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], p2[X], p2[Y], p2[Z], p4[X], p4[Y], p4[Z],
				   pvV[0][0], pvV[0][1], pvV[1][0], pvV[1][1],
				  (pvV[1] + pvVDelta[1])[0], (pvV[1] + pvVDelta[1])[1]);
		  pvV[1] += pvVDelta[1];
		}
	    }
	  ppvU[0][0] += pvUDelta[0];
	  ppvU[0][1] += pvUDelta[1];
	  ppvU[1][0] += pvUDelta[0];
	  ppvU[1][1] += pvUDelta[1];
	}
    }
    */
}

VOID CSLIDEHeightfield::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE Heightfield: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEHeightfield::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE Heightfield: %s", m_pcGUIID);
}

VOID CSLIDEHeightfield::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
					  CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: HEIGHTFIELD\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDEHeightfield::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  if ( mTransform.Determinant() < 0 ) // Need to Mirror triangles
    {
      UINT_32 i, j;
      Point p1, p2, p3, p4;
      
      for( i = 0; i < m_uiUSlices; i++ )
	{
	  p1 = m_ppSurfacePoint[i][0];
	  p2 = m_ppSurfacePoint[i+1][0];
	  
	  p1 = mTransform * p1;
	  p2 = mTransform * p2;
	  
	  for( j = 1; j < m_uiVSlices+1; j++)	
	    {
	      p3 = m_ppSurfacePoint[i][j];
	      p4 = m_ppSurfacePoint[i+1][j];
	      
	      p3 = mTransform * p3;
	      p4 = mTransform * p4;
	      
	      WriteSTLTriangle(fpOutputFile, p3, p2, p1);
	      WriteSTLTriangle(fpOutputFile, p2, p3, p4);

	      p1 = p3;
	      p2 = p4;
	    }
	}
    }    
  else // Don't mirror triangles
    {
      UINT_32 i, j;
      Point p1, p2, p3, p4;

      for( i = 0; i < m_uiUSlices; i++ )
	{
	  p1 = m_ppSurfacePoint[i][0];
	  p2 = m_ppSurfacePoint[i+1][0];

	  p1 = mTransform * p1;
	  p2 = mTransform * p2;
	
	  for( j = 1; j < m_uiVSlices+1; j++)	
	    {
	      p3 = m_ppSurfacePoint[i][j];
	      p4 = m_ppSurfacePoint[i+1][j];

	      p3 = mTransform * p3;
	      p4 = mTransform * p4;

	      WriteSTLTriangle(fpOutputFile, p1, p2, p3);
	      WriteSTLTriangle(fpOutputFile, p4, p3, p2);

	      p1 = p3;
	      p2 = p4;
	    }
	}
    }
}



