#include "slidenurbpatch.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"
#include "bsplinemath.h"

static const FLOAT cfNormalRatio = 1.0 / 10.0;
static const UINT_32 cuiUOrder = 4;
static const UINT_32 cuiVOrder = 4;

//////////////////////////////////////////////////////////////////////
// CSLIDENurbPatch Class
//

CSLIDENurbPatch::CSLIDENurbPatch()
{
  m_pfWDerivativeCurve = NULL;
  m_ppfCurveWDerivativeTable = NULL;
  m_uiCurveWDerivativeControlPoints = 0;
}

CSLIDENurbPatch::~CSLIDENurbPatch()
{
}

VOID CSLIDENurbPatch::DeallocateWDerivativeCurve()
{
  delete[] m_pfWDerivativeCurve;
}

VOID CSLIDENurbPatch::ReallocateWDerivativeCurve()
{
  if( m_pfWDerivativeCurve != NULL )
    {
      DeallocateWDerivativeCurve();
    }

 if( m_uiUOrder > m_uiVOrder )
    {
      m_pfWDerivativeCurve = new FLOAT[m_uiUOrder];
    }
  else
    {
      m_pfWDerivativeCurve = new FLOAT[m_uiVOrder];
    }
}

VOID CSLIDENurbPatch::DeallocateCurveWDerivativeTable()
{
  UINT_32 i;

  if ( m_ppfCurveWDerivativeTable != NULL )
    {
      for ( i = 0; i < m_uiCurveWDerivativeControlPoints; i++ )
	{
	  delete[] m_ppfCurveWDerivativeTable[i];
	}
      delete[] m_ppfCurveWDerivativeTable;
      m_ppfCurveWDerivativeTable = NULL;
      m_uiCurveWDerivativeControlPoints = 0;
    }
}

VOID CSLIDENurbPatch::ReallocateCurveWDerivativeTable()
{
  UINT_32 i, uiControlPoints;

  if( m_ppfCurveWDerivativeTable != NULL )
    {
      DeallocateCurveWDerivativeTable();
    }

  if( m_uiUControlPoints > m_uiVControlPoints )
    {
      uiControlPoints = m_uiUControlPoints;
    }
  else
    {
      uiControlPoints = m_uiVControlPoints;
    }

  ASSERT( m_uiCurveWDerivativeControlPoints == 0 );
  ASSERT( m_ppfCurveWDerivativeTable == NULL );

  m_uiCurveWDerivativeControlPoints = uiControlPoints;
  m_ppfCurveWDerivativeTable = new FLOAT*[m_uiCurveWDerivativeControlPoints];
  ASSERT( m_ppfCurveWDerivativeTable != NULL );

  for ( i = 0; i < m_uiCurveWDerivativeControlPoints; i++ )
    {
      m_ppfCurveWDerivativeTable[i] = new FLOAT[m_uiCurveDerivatives]; 
      ASSERT( m_ppfCurveWDerivativeTable[i] != NULL );
    }
}

VOID CSLIDENurbPatch::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  m_uiUOrder = cuiUOrder;
  m_uiVOrder = cuiVOrder;

  ReallocateWDerivativeCurve();
  ReallocateCurveWDerivativeTable();
  m_uiCurveWDerivativeControlPoints = 0;

  CSLIDEBSplinePatch::Init(pcFile, uiLine, uiColumn, pcID);
}

VOID CSLIDENurbPatch::Uninit()
{
  DeallocateWDerivativeCurve();
  DeallocateCurveWDerivativeTable();

  CSLIDEBSplinePatch::Uninit();
}

BOOL CSLIDENurbPatch::Initialized()
{
  if ( !CSLIDEBSplinePatch::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDENurbPatch::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEBSplinePatch::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  ReallocateWDerivativeCurve();
  ReallocateCurveWDerivativeTable();

  return TRUE;
}

BOOL CSLIDENurbPatch::Valid()
{
  if ( !CSLIDEBSplinePatch::Valid() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDENurbPatch::CheckDynamic()
{
  CSLIDEBSplinePatch::CheckDynamic();
}

VOID CSLIDENurbPatch::UpdateState()
{
  CSLIDEBSplinePatch::UpdateState();

  if( m_bSlicesChanged )
    {
      ReallocateWDerivativeCurve();
      ReallocateCurveWDerivativeTable();
    }
}

#ifdef __USE_CDA__
VOID CSLIDENurbPatch::CalculateCDAVertices(CListInt *plistCDAVertex, 
					      const Matrix &mGroot_obj)
{

}
#endif // __USE_CDA__

VOID CSLIDENurbPatch::CalculateBound()
{
  UINT_32 i, j;
  CBound bound;
  Point pt;

  for( i = 0; i < m_uiUControlPoints; i++ )
    {
      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  // Project to 3D
	  pt = m_ppControlPoint[i][j] / m_ppControlPoint[i][j][W];
	  bound.Insert(pt);
	}
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }
}

VOID CSLIDENurbPatch::PointAt(Point &pt, FLOAT fU, FLOAT fV)
{
  UINT_32 knotIndex;

  Point *pControlCurvePoint;

  pControlCurvePoint = new Point[m_uiUControlPoints];

  // First find the control points for the curve in the u-direction for tv = v
  ControlCurveAt(pControlCurvePoint, fV, TRUE);

  // Use blossoming to evaluate u on this curve
  knotIndex = FindBlossomKnotInterval(fU, TRUE);
  BlossomAt(pt, m_uiUOrder, &pControlCurvePoint[knotIndex-1], &m_pfUKnot[knotIndex], fU);

  // Project to 3D
  if( pt[W] != 0.0 )
    {
      pt = pt / pt[W];
    }
  else
    {
      printf( "Error: Can't divide by zero w-coordinate\n");
    }

  delete []pControlCurvePoint;
}

VOID CSLIDENurbPatch::ControlWeightAt()
{
  UINT_32 i, uiOrder;

  if( m_uiUOrder > m_uiVOrder )
    uiOrder = m_uiUOrder;
  else
    uiOrder = m_uiVOrder;

  for( i = 0; i < uiOrder; i++ )
    {
      m_pfWDerivativeCurve[i] = m_pDerivativeCurve[i][W];
    }
}

VOID CSLIDENurbPatch::CalculateSurface()
{
  UINT_32 uiUSlice, uiVSlice, knotIndex;
  FLOAT fU, fV, fUDelta, fVDelta;
  Point pt;
  Point *pControlCurveVertex;

  fV = GetTv();
  fUDelta = (GetTuDelta() - EPSILON) / m_uiUSlices;
  fVDelta = (GetTvDelta() - EPSILON) / m_uiVSlices;
  pControlCurveVertex = new Point[m_uiUControlPoints];

  for( uiVSlice = 0; uiVSlice <= m_uiVSlices; uiVSlice++ )
    {
      ControlCurveAt(pControlCurveVertex, fV, TRUE);
      fU = GetTu();

      for( uiUSlice = 0; uiUSlice <= m_uiUSlices; uiUSlice++ )
	{
	  knotIndex = FindBlossomKnotInterval(fU, TRUE);
	  BlossomAt(pt, m_uiUOrder, &pControlCurveVertex[knotIndex-1], 
		    &m_pfUKnot[knotIndex], fU);

	  // Project to 3D
	  if( pt[W] != 0.0 )
	    {
	      pt = pt / pt[W];
	    }
	  else
	    {
	      printf( "Error: Can't divide by zero w-coordinate\n");
	    }
	  m_ppSurfacePoint[uiUSlice][uiVSlice] = pt;

	  fU += fUDelta;
	}
      
      fV += fVDelta;
    }
  delete[] pControlCurveVertex;
}

VOID CSLIDENurbPatch::NormalAt(Vector &vN, FLOAT fU, FLOAT fV)
{
  Vector vVT, vUT;
  Point pt;

  ControlCurveAt( m_pDerivativeCurve, fV, TRUE );  // in u
  ControlWeightAt();
  CurveTangentAt( vUT, fU, fV, TRUE );

  ControlCurveAt( m_pDerivativeCurve, fU, FALSE ); // in v
  ControlWeightAt();
  CurveTangentAt( vVT, fU, fV, FALSE );    

  vN = vUT.Cross(vVT);
  vN.Normalize();
}

VOID CSLIDENurbPatch::CurveTangentAt(Vector &vT, FLOAT fU, FLOAT fV, BOOL bUDirection)
{
  UINT_32 uiDelta, uiOrder, uiControlPoints;
  Vector pvF[2];
  FLOAT pfG[2];
  FLOAT *pfKnot;
  FLOAT t;
 
  ASSERT( Valid() );

  if( bUDirection )
    {
      t = MAX(fU, GetTu());
      t = MIN(fU, GetTu() + GetTuDelta() - EPSILON);      
      uiOrder = m_uiUOrder;
      pfKnot = m_pfUKnot;
      uiControlPoints = m_uiUControlPoints;
    }
  else
    {
      t = MAX(fV, GetTv());
      t = MIN(fV, GetTv() + GetTvDelta() - EPSILON);
      uiOrder = m_uiVOrder;
      pfKnot = m_pfVKnot;
      uiControlPoints = m_uiVControlPoints;
    }

  uiDelta = FindKnotInterval(t, bUDirection);
  CalculateDeBoorTable(uiOrder, pfKnot, m_ppfCurveDeBoorTable, t, uiDelta);

  // f'(t)
  CalculateDerivativeTable(uiControlPoints, m_uiCurveDerivatives, uiOrder, pfKnot, 
			   m_pDerivativeCurve, m_ppCurveDerivativeTable);
  DerivativesAt(pvF, uiOrder, m_ppCurveDerivativeTable, m_ppfCurveDeBoorTable, 
		0, 1, t, uiDelta);

  // g'(t)
  CalculateWDerivativeTable(uiControlPoints, m_uiCurveDerivatives, uiOrder, pfKnot, 
			    m_pfWDerivativeCurve, m_ppfCurveWDerivativeTable);
  WDerivativesAt(pfG, uiOrder, m_ppfCurveWDerivativeTable, m_ppfCurveDeBoorTable, 
		 0, 1, t, uiDelta);
  
  // Dx (f(t)/g(t)) = [f'(t)g(t) - g'(t)f(t)] / [g(t)]^2
  if( pfG[0] != 0.0 )
    {
      vT = ((pvF[1] * pfG[0]) - (pvF[0] * pfG[1])) / (pfG[0]*pfG[0]);
    }

  if( vT[W] != 0.0 )
    {
      vT = vT / vT[W];
    }

  vT.Normalize();
}

VOID CSLIDENurbPatch::RenderControls()
{
  UINT_32 i, j;
  Point pt;

  /*
  GLUnurbsObj *theNurb;
  GLfloat cp[m_uiUControlPoints][m_uiVControlPoints][4];

  for( j = 0; j < m_uiVControlPoints; j++ )
    {
      for( i = 0; i < m_uiUControlPoints; i++ )
	{
	  cp[i][j][0] = (m_ppControlPoint[i][j].Get())[0];
	  cp[i][j][1] = (m_ppControlPoint[i][j].Get())[1];
	  cp[i][j][2] = (m_ppControlPoint[i][j].Get())[2];
	  cp[i][j][3] = (m_ppControlPoint[i][j].Get())[3];
	}
    }

  theNurb = gluNewNurbsRenderer();
  gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, GLU_OUTLINE_POLYGON);
  gluNurbsProperty(theNurb, GLU_SAMPLING_METHOD, GLU_DOMAIN_DISTANCE);
  gluNurbsProperty(theNurb, GLU_U_STEP, 10);
  gluNurbsProperty(theNurb, GLU_V_STEP, 10);

  PushAuxillaryGeometryAttributes();
  {
    glPushMatrix();
    gluBeginSurface(theNurb);
    {
      gilColor3f(1.0, 0.0, 0.0);
      
      gluNurbsSurface(theNurb,
		      m_uiUKnots, m_pfUKnot, m_uiVKnots, m_pfVKnot,
		      4, m_uiUControlPoints*4, &cp[0][0][0],
		      m_uiUOrder, m_uiVOrder, GL_MAP2_VERTEX_4);
    }
    gluEndSurface(theNurb);
    glPopMatrix();
  }
  PopAuxillaryGeometryAttributes();
  */


  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 1.0, 0.0);

    // Edges
    // Draw the control edges in U
    for(i = 0; i < m_uiUControlPoints; i++ ) 
      {
	gilBegin(GL_LINE_STRIP);
	{
	  for (j = 0; j < m_uiVControlPoints; j++ )
	    {
	      pt = m_ppControlPoint[i][j] / m_ppControlPoint[i][j][W];
	      gilVertex3fv(pt.Get());
	    }
	}
	gilEnd();
      }
    // Draw the control edges in V
    for (j = 0; j < m_uiVControlPoints; j++ )
      {
	gilBegin(GL_LINE_STRIP);
	{
	  for(i = 0; i < m_uiUControlPoints; i++ ) 
	    {
	      pt = m_ppControlPoint[i][j] / m_ppControlPoint[i][j][W];
	      gilVertex3fv(pt.Get());
	    }
      }
	gilEnd();
      }
    
    // Points
    gilBegin(GL_POINTS);
    {
      for(i = 0; i < m_uiUControlPoints; i++ ) 
	{
	  for (j = 0; j < m_uiVControlPoints; j++ )
	    {
	      pt = m_ppControlPoint[i][j] / m_ppControlPoint[i][j][W];
	      gilVertex3fv(pt.Get());
	    }
	}
    }
    gilEnd();
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDENurbPatch::WriteSLIDE(FILE *fpOutputFile)
{
  UINT_32 i, j;
  CListIterInt iter;
  CHAR *pcID;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtNurbPatch], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUOrder],
	  m_uiUOrder);
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVOrder],
	  m_uiVOrder);

  if( m_plistControlPointIDs != NULL )
    {
      fprintf(fpOutputFile, "  %s (",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);

      iter.Init(m_plistControlPointIDs);
      
      for ( pcID = (CHAR*)(iter.PeekFirst()), i = 0;
	    pcID != NULL;
	    pcID = (CHAR*)(iter.PeekNext()) )
	{
	  if( i == m_uiUControlPoints )
	    {
	      fprintf(fpOutputFile, "\n\t\t    " );
	      i = 0;
	    }
	  fprintf(fpOutputFile, " %s", pcID );
	  i++;
	}
    }
  else
    {
      const FLOAT* pt;

      fprintf(fpOutputFile, "# Default Control Points:\n");
      fprintf(fpOutputFile, "# %s (\n#\t",
	      gcppcSLIDEReservedWords[sliderwtControlPointList]);
      
      for( j = 0; j < m_uiVControlPoints; j++ )
	{
	  for( i = 0; i < m_uiUControlPoints; i++ )
	    {
	      pt = m_ppControlPoint[i][m_uiVControlPoints-1-j].Get();
	      fprintf(fpOutputFile, "(%.3f %.3f %.3f %.3f)", pt[X], pt[Y], pt[Z], pt[W]);

	      if( (j != m_uiVControlPoints-1) && ( i == m_uiUControlPoints-1) )
		{
		  fprintf(fpOutputFile, "\n#\t" );
		}
	    }
	}
    }
  fprintf(fpOutputFile, " )\n");

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtUSlices],
	  m_dyniUSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtVSlices],
	  m_dyniVSlices.GetValue());
  fprintf(fpOutputFile, "  %s (%.3f %.3f) (%.3f %.3f) (%.3f %.3f) (%.3f %.3f)\n",
	  gcppcSLIDEReservedWords[sliderwtTexture],
	  m_pvTexture[0][X], m_pvTexture[0][Y],
	  m_pvTexture[1][X], m_pvTexture[1][Y],
	  m_pvTexture[2][X], m_pvTexture[2][Y],
	  m_pvTexture[3][X], m_pvTexture[3][Y]);
  fprintf(fpOutputFile, "  %s %.3f %ld\n",
	  gcppcSLIDEReservedWords[sliderwtOffset],
	  m_dynfOffset.GetValue(), m_dyniOffsetConnected.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawControls],
	  m_dyniDrawControls.GetValue());

  
  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndNurbPatch]);
}

VOID CSLIDENurbPatch::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  UINT_32 i, j;
  FLOAT fUMin, fUMax, fVMin, fVMax;
  const FLOAT *pt;
  char *pcUKnots = new char[100];
  char *pcVKnots = new char[100];
  char *pcPointIds = new char[800];
  char *pcTemp = new char[50];
  
  for( i = 0; i < m_uiUKnots; i++ )
    {
      sprintf( pcTemp, " %.3f", m_pfUKnot[i] );
      strcat( pcUKnots, pcTemp );
    }

  for( j = 0; j < m_uiVKnots; j++ )
    {
      sprintf( pcTemp, " %.3f", m_pfVKnot[j] );
      strcat( pcVKnots, pcTemp );
    }

  for( j = 0; j < m_uiVControlPoints; j++ )
    {
      for( i = 0; i < m_uiUControlPoints; i++ )
	{
	  pt = m_ppControlPoint[i][j].Get();
	  sprintf( pcTemp, " %.3f %.3f %.3f %.3f", pt[X], pt[Y], pt[Z], pt[W] );
	  strcat( pcPointIds, pcTemp );
	}
    }
  
  fUMin = m_pfUKnot[0];
  fUMax = m_pfUKnot[m_uiUKnots-1];
  fVMin = m_pfVKnot[0];
  fVMax = m_pfVKnot[m_uiVKnots-1];

  RiArchiveRecord(RI_COMMENT, "\nNuPatch %ld %ld [%s] %.3f %.3f %ld %ld [%s] %.3f %.3f \"Pw\" [%s] \"st\" [%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f]\n", 
		  m_uiUControlPoints,
		  m_uiUOrder,
		  pcUKnots,
		  fUMin,
		  fUMax,
		  m_uiVControlPoints,
		  m_uiVOrder,
		  pcVKnots,
		  fVMin,
		  fVMax,
		  pcPointIds,
		  m_pvTexture[3][X], m_pvTexture[3][Y], 
		  m_pvTexture[2][X], m_pvTexture[2][Y],
		  m_pvTexture[0][X], m_pvTexture[0][Y], 
		  m_pvTexture[1][X], m_pvTexture[1][Y]);

  delete[] pcUKnots;
  delete[] pcVKnots;
  delete[] pcPointIds;
  delete[] pcTemp;
}

VOID CSLIDENurbPatch::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE NurbPatch: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDENurbPatch::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE NurbPatch: %s", m_pcGUIID);
}

