#include "slideobject.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideface.h"

#include "slidepoint.h"
#include "qemesh.h"
#include "qewrite.h"

#ifdef DT_TRIANGULATE_TEST

#include "dtvertex.h"
#include "dtedge.h"
#include "dtcontourvertex.h"
#include "dtcontouredge.h"
#include "dtcontour.h"
#include "dttriangle.h"
#endif // DT_TRIANGULATE_TEST

//////////////////////////////////////////////////////////////////////
// CSLIDEObject Class
//

CSLIDEObject::CSLIDEObject()
{
  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_plistFaceIDs = NULL;
}

CSLIDEObject::~CSLIDEObject()
{
}

VOID CSLIDEObject::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID,
			CListInt *plistFaceIDs)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniSolid.SetValue(slidestSLF_SOLID);

  m_plistFaceIDs = plistFaceIDs;
  m_listCSLIDEFace.Init();

#ifdef DT_TRIANGULATE_TEST
  m_listDTVertex.Init();
  m_listDTEdges.Init();
  m_listDTContours.Init();
  m_listDTTriangles.Init();
#endif // DT_TRIANGULATE_TEST
}

VOID CSLIDEObject::Uninit()
{
  CSLIDEBoundedNode::Uninit();

  if ( m_plistFaceIDs != NULL )
    {
      delete m_plistFaceIDs;
      m_plistFaceIDs = NULL;
    }
}

BOOL CSLIDEObject::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return (m_plistFaceIDs != NULL);
}

BOOL CSLIDEObject::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEFace *pCSLIDEFace;

  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Solid Type
  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  // Check Face References
  iter.Init(m_plistFaceIDs);
  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEFace = pCSLIDEWorld->PeekFace(pcID);
      if ( pCSLIDEFace == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtFace], pcID);
	  bCorrect = FALSE;
	}
      else
	{
	  m_listCSLIDEFace.InsertLast(pCSLIDEFace, 0);
	  pCSLIDEFace->InsertParentObject(this);
	}
    }

  return bCorrect;
}

BOOL CSLIDEObject::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniSolid.GetValue() &&
	     m_dyniSolid.GetValue() < slidestSize ) );
	   //GetFaceCount() > 0 );
}

VOID CSLIDEObject::Remove(CSLIDEWorld *pCSLIDEWorld)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  // Remove from child
  iter.Init(&m_listCSLIDEFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->RemoveParentObject(this);
    }

  // Remove from parent and world
  CSLIDENode::Remove(pCSLIDEWorld);
}

VOID CSLIDEObject::RemoveFace(CSLIDEFace *pCSLIDEFaceToRemove)
{
  CListIterInt iter;
  CHAR *pcID, *pcFaceIDToRemove;
  CSLIDEFace *pCSLIDEFace;

  pcFaceIDToRemove = pCSLIDEFaceToRemove->GetID();

  iter.Init(m_plistFaceIDs);
  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      if( pcID == pcFaceIDToRemove )
	{
	  iter.PeekNextRemoveCurr();
	  break;
	}
    }

  iter.Init(&m_listCSLIDEFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      if( (pCSLIDEFace->GetID()) == pcFaceIDToRemove )
	{
	  iter.PeekNextRemoveCurr();
	  break;
	}
    }
}

CSLIDECamera *CSLIDEObject::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDELight *CSLIDEObject::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDEObject::FindAreaLight(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

VOID CSLIDEObject::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDEObject::CheckDynamic()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  BOOL bAutoNormals;

  // Are there variables which I must call TCL to update?
  SetDynamic(m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic() ||
	     m_dyniSolid.GetDynamic());

  bAutoNormals = !(!m_dyniShading.GetDynamic() && 
		   (m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_WIRE ||
		    m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_FLAT));
      
  iter.Init(&m_listCSLIDEFace);
  
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->SetAutoNormals(bAutoNormals || pCSLIDEFace->GetAutoNormals());
    }
}

VOID CSLIDEObject::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEFace *pCSLIDEFace;
      
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;

      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Look at Faces

      iter.Init(&m_listCSLIDEFace);
      
      for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	    pCSLIDEFace != NULL;
	    pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
	{
	  bBoundEmpty = bBoundEmpty && 
	                pCSLIDEFace->GetBoundEmpty();
	  bBoundDynamic = bBoundDynamic || 
	                  pCSLIDEFace->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || 
	                     pCSLIDEFace->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic ||
	                    pCSLIDEFace->GetNormalsDynamic();
	}

      bGeometryEmpty = bBoundEmpty;
      bNormalsEmpty = bBoundEmpty;
      bBoundsEmpty = bBoundEmpty;
      bBoundsDynamic = bBoundDynamic;

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // For the OpenGL Display Lists to be really useful 
      // it is all or nothing so we enforce that here

      bGeometryDynamic = ( bGeometryDynamic ||
			   bNormalsDynamic ||
			   bBoundsDynamic );
      bNormalsDynamic = bGeometryDynamic;
      bBoundsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preproces(FLOAT)sing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDEObject::DrawGUIGraphFaceReferences(Tcl_Interp *pInterp)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  CHAR pcBuffer[200];
  INT_32 iCode;
  
  iter.Init(&m_listCSLIDEFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      sprintf(pcBuffer, "Graph:CreateReference %s %s", GetID(), pCSLIDEFace->GetID());
      iCode = Tcl_Eval(pInterp, pcBuffer);
      ASSERT( iCode == TCL_OK );
    }
}

VOID CSLIDEObject::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }
  
  if ( GetBoundsDynamic() )
    {
      CSLIDEBoundedNode::InvalidateQEMesh();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDEObject::UpdateState()
{
  CSLIDEBoundedNode::UpdateState();

  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", 
		  m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }

#ifdef DT_TRIANGULATE_TEST

#define DT_TERRAIN_TEST
#ifdef DT_TERRAIN_TEST

  CListIterInt iterCSLIDEFace, iterCSLIDEPoint;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  DTVertex *pvx;
  Point pt;

  iterCSLIDEFace.Init(&m_listCSLIDEFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iterCSLIDEFace.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iterCSLIDEFace.PeekNext() )
    {

      iterCSLIDEPoint.Init(pCSLIDEFace->GetPointList());
      for ( pCSLIDEPoint = (CSLIDEPoint*)iterCSLIDEPoint.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iterCSLIDEPoint.PeekNext() )
	{
	  pt = pCSLIDEPoint->GetPoint();
	  pt[Z] = 0.0;

	  pvx = AllocateDTVertex();
	  ASSERT( pvx != NULL );
	  pvx->Init(pt);
	  m_listDTVertex.InsertLast(pvx, 0);
	}
    }

  DTContour::IncrementalDelaunay(&m_listDTVertex, &m_listDTEdges, &m_listDTTriangles);

#else // DT_TERRAIN_TEST

  CListIterInt iterCSLIDEFace, iterCSLIDEPoint, iterDTVertex;
  DTContourIter iterDTContour;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  DTContourVertex *pcvxContour;
  DTContourEdge *pce;
  DTEdge *pe;
  DTVertex *pvx;
  Point pt;
  Vector vZ, vNormal, vAxis;
  FLOAT fSinTheta, fCosTheta, fD, fTheta;
  Matrix mRot, mRotInv, mTrans, mTransInv;
  Matrix *pm;
  Vector vTrans;
  UINT_32 i;

  vZ.Set(0.0, 0.0, 1.0);

  pm = new Matrix[2*m_listCSLIDEFace.GetLength()];
  ASSERT( pm != NULL );

  iterCSLIDEFace.Init(&m_listCSLIDEFace);

  for ( i = 0, pCSLIDEFace = (CSLIDEFace*)iterCSLIDEFace.PeekFirst(); 
	pCSLIDEFace != NULL;
	i++, pCSLIDEFace = (CSLIDEFace*)iterCSLIDEFace.PeekNext() )
    {
      vNormal = pCSLIDEFace->GetNormal();
      
      if ( vZ.Dot(vNormal) < 0.0 )
	{
	  vNormal = -vNormal;
	}

      fD = vNormal[3];
      vNormal[3] = 0.0;
      vAxis = vNormal * vZ;
      fSinTheta = vAxis.Magnitude();
      fCosTheta = vNormal.Dot(vZ);
      mRot.SetIdentity();
      if ( vAxis.Normalize() )
	{
	  fTheta = atan2(fSinTheta, fCosTheta);
	  mRot.SetRotate(vAxis, fTheta);
	  mRotInv.SetRotate(vAxis, -fTheta);
	}
      vTrans.Set(0.0, 0.0, fD);
      mTrans.SetTranslate(vTrans);
      vTrans = -vTrans;
      mTransInv.SetTranslate(vTrans);

      pm[2*i] = mTrans * mRot;
      pm[2*i + 1] = mRotInv * mTransInv;

      pcvxContour = NULL;
      iterCSLIDEPoint.Init(pCSLIDEFace->GetPointList());
      for ( pCSLIDEPoint = (CSLIDEPoint*)iterCSLIDEPoint.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iterCSLIDEPoint.PeekNext() )
	{
	  pt = pCSLIDEPoint->GetPoint();
	  pt = pm[2*i] * pt;
	  ASSERT( -EPSILON < pt[Z] && pt[Z] < EPSILON );
	  pt[Z] = 0.0;

	  pvx = AllocateDTVertex();
	  ASSERT( pvx != NULL );
	  pvx->Init(pt);
	  m_listDTVertex.InsertLast(pvx, 0);
	  DTContour::AddVertex(&pcvxContour, pvx);
	}
      iterDTContour.Init(pcvxContour);
      for ( pce = iterDTContour.PeekFirstDTContourEdge();
	    pce != NULL;
	    pce = iterDTContour.PeekNextDTContourEdge() )
	{
	  pe = AllocateDTEdge();
	  ASSERT( pe != NULL );
	  pe->Init(pce->GetVertexPrev(), pce->GetVertexNext());
	  pe->AddContourEdge(pce);
	  pce->SetEdge(pe);
	  m_listDTEdges.InsertLast(pe, 0);
	}
      m_listDTContours.InsertLast(pcvxContour, 0);
    }

  iterDTVertex.Init(&m_listDTVertex);
  for ( pvx = (DTVertex*)iterDTVertex.PeekFirst();
	pvx != NULL;
	pvx = (DTVertex*)iterDTVertex.PeekNext() )
    {
      pt = pvx->GetPoint();
      pt = pm[1] * pt;
      pvx->SetPoint(pt);
    }

  delete []pm;

  DTContour::Triangulate(&m_listDTContours, &m_listDTEdges, &m_listDTTriangles);
#endif // DT_TERRAIN_TEST

#endif // DT_TRIANGULATE_TEST
}

VOID CSLIDEObject::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  CalculateBound();
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDEObject::CalculateBound()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  CBound bound;

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->InsertIntoBound(&bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDEObject::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
    }
}
#endif // __USE_CDA__

VOID CSLIDEObject::CalculateQEMesh()
{
  CListIterInt iterFace, iterPoint;
  CSLIDEFace *pCSLIDEFace;
  CSLIDEPoint *pCSLIDEPoint;
  CObjectList<QEVertex> listQEVertex;
  QEVertex *pvx;
  Vector v;
  CSLIDESurface *pCSLIDESurface;

  ASSERT( m_pQEMesh == NULL );

  m_pQEMesh = AllocateQEMesh();
  ASSERT( m_pQEMesh != NULL );
  m_pQEMesh->Init();

  iterFace.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)iterFace.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iterFace.PeekNext() )
    {
      pCSLIDESurface = pCSLIDEFace->GetSurface();

      iterPoint.Init(pCSLIDEFace->GetPointList());
      for ( pCSLIDEPoint = (CSLIDEPoint*)(iterPoint.PeekFirst());
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)(iterPoint.PeekNext()) )
	{
	  pvx = AllocateQEVertex();
	  ASSERT( pvx != NULL );
	  pvx->Init(pCSLIDEPoint->GetPoint());
	  v = pCSLIDEPoint->GetTexture();
	  pvx->SetTexture(v);
	  listQEVertex.InsertLast(pvx);
	}

      //m_pQEMesh->MergeFacet(&listQEVertex);
      m_pQEMesh->MergeFacet(&listQEVertex, pCSLIDESurface);
      ASSERT( listQEVertex.GetLength() == 0 );
    }

  //printf("FINAL MESH:\n");
  //m_pQEMesh->Display();
}

VOID CSLIDEObject::CalculateLeafQEMesh()
{
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDEObject::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

VOID CSLIDEObject::PrerenderGeometry(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  SLIDEShadingType eShade;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  // This flag tells us whether to use Point or Face Normals
  ModifyShadingFlag(eShading);

  // This one set the shading attributes which in many cases will 
  // be SLF_INHERIT and so it will not set anything and leave up
  // to the caller above
  eShade = (SLIDEShadingType)m_dyniShading.GetValue();
  
  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	{
	  m_uiPrerenderAttributesStackDepthMax = 0;

#ifndef DT_TRIANGULATE_TEST

	  iter.Init(&m_listCSLIDEFace);
	  
	  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
		pCSLIDEFace != NULL;
		pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
	    {
	      pCSLIDEFace->Prerender(eShading);
	      m_uiPrerenderAttributesStackDepthMax = 
		MAX(m_uiPrerenderAttributesStackDepthMax,
		    pCSLIDEFace->GetPrerenderAttributesStackDepthMax());
	    }

#else // DT_TRIANGULATE_TEST

	  DTTriangle *pDTTriangle;
	  iter.Init(&m_listDTTriangles);
	  for ( pDTTriangle = (DTTriangle*)(iter.PeekFirst());
		pDTTriangle != NULL;
		pDTTriangle = (DTTriangle*)(iter.PeekNext()) )
	    {
	      pDTTriangle->Render();
	    }

#endif // DT_TRIANGULATE_TEST

	  if ( GLbAttributesMask != 0x0 )
	    {
	      m_uiPrerenderAttributesStackDepthMax++;
	    }
	}
	PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDEObject::PrerenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  iter.Init(&m_listCSLIDEFace);
  
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->RenderNormals(eShading);
    }
}

VOID CSLIDEObject::RenderGeometry(const Matrix &/*mGproj_obj*/, 
				  CBound * /*pCBoundClip*/, 
				  SLIDEShadingType eShading,
				  SLIDELODType /* eLOD */,
				  CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/ )
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  
  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
  {
#ifndef DT_TRIANGULATE_TEST

    iter.Init(&m_listCSLIDEFace);
  
    for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	  pCSLIDEFace != NULL;
	  pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
      {
	pCSLIDEFace->Render(eShading, pCSLIDEWarp);
      }

#else // DT_TRIANGULATE_TEST

    DTTriangle *pDTTriangle;
    iter.Init(&m_listDTTriangles);
    for ( pDTTriangle = (DTTriangle*)(iter.PeekFirst());
	  pDTTriangle != NULL;
	  pDTTriangle = (DTTriangle*)(iter.PeekNext()) )
      {
	pDTTriangle->Render();
      }

#endif // DT_TRIANGULATE_TEST
  }
  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDEObject::RenderGeometrySelect(const Matrix &/*mGproj_obj*/, 
					CBound * /*pCBoundClip*/, 
					SLIDEShadingType eShading,
					SLIDELODType /* eLOD */ )
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  
  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
  {
#ifndef DT_TRIANGULATE_TEST

    iter.Init(&m_listCSLIDEFace);
  
    for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	  pCSLIDEFace != NULL;
	  pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
      {
	pCSLIDEFace->RenderSelect();
      }

#else // DT_TRIANGULATE_TEST

    DTTriangle *pDTTriangle;
    iter.Init(&m_listDTTriangles);
    for ( pDTTriangle = (DTTriangle*)(iter.PeekFirst());
	  pDTTriangle != NULL;
	  pDTTriangle = (DTTriangle*)(iter.PeekNext()) )
      {
	pDTTriangle->RenderSelect();
      }

#endif // DT_TRIANGULATE_TEST
  }
  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDEObject::RenderNormals(SLIDEShadingType eShading)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);

    iter.Init(&m_listCSLIDEFace);
    
    for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	  pCSLIDEFace != NULL;
	  pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
      {
	pCSLIDEFace->RenderNormals(eShading);
      }
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDEObject::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", 
	  gcppcSLIDEReservedWords[sliderwtObject], m_pcGUIID);

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEFace->GetGUIID());      
    }

  fprintf(fpOutputFile, ")\n");

  WriteSLIDEFlags(fpOutputFile, 0);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndObject]);
}

VOID CSLIDEObject::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
  {
    WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
    /*
    if ( CheckRIBPrimitive() )
      {
	WriteRIBGeometryPrimitive(fpOutputFile, eShading, eLOD);
      }
    else
      {
	WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
      }
      */
  }
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());
}

BOOL CSLIDEObject::CheckRIBPrimitive()
{
  BOOL bResult;
  //  UINT_32 i;

  bResult = FALSE;
/*
  for ( i = 0; i < (UINT_32)slideribotSize; i++ )
    {
      bResult = bResult || (m_pcGUIID == gcppcSLIDERIBObjectReservedWords[i]);  
    }
*/
  return bResult;
}

VOID CSLIDEObject::WriteRIBGeometryPolygonal(FILE *fpOutputFile,
					     SLIDEShadingType eShading,
					     SLIDELODType /* eLOD */)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  ASSERT( fpOutputFile != NULL );

  iter.Init(&m_listCSLIDEFace);
  
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->WriteRIB(fpOutputFile, eShading);
    }
}

VOID CSLIDEObject::WriteRIBGeometryPrimitive(FILE *fpOutputFile,
					     SLIDEShadingType /* eShading */,
					     SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
/*
  RiAttributeBegin();
  {
    // RenderMan defaults to defining things in clockwise ordering 
    // while everything in SLIDE is in counter-clockwise ordering
    //RiOrientation(RI_RH);
    if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_SPHERE] )
      {
	//RiSphere(1.0, -1.0, 1.0, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nSphere 1 -1 1 360");
      }
    else if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_CONE] )
      {
	//RiCone(1.0, 1.0, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nCone 1 1 360");
      }
    else if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_DISK] )
      {
	//RiDisk(0.0, 1.0, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nDisk 0 1 360");
      }
    else if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_CYLINDER] )
      {
	//RiCylinder(1.0, -1.0, 1.0, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nCylinder 1 -1 1 360");
      }
    else if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_HYPERBOLOID] )
      {
	FLOAT fSkewAngle;
	RtPoint rtPoint1, rtPoint2;
	  
	fSkewAngle = DEGTORAD(45.0);
	
	rtPoint1[0] = (FLOAT)cos(fSkewAngle);
	rtPoint1[1] = (FLOAT)sin(fSkewAngle);
	rtPoint1[2] = -1.0;
	
	rtPoint2[0] = (FLOAT)cos(-fSkewAngle);
	rtPoint2[1] = (FLOAT)sin(-fSkewAngle);
	rtPoint2[2] = 1.0;
	
	//RiHyperboloid(rtPoint1, rtPoint2, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nHyperboloid [%.3f %.3f %.3f] [%.3f %.3f %.3f] 360", 
			rtPoint1[0], rtPoint1[1], rtPoint1[2], 
			rtPoint2[0], rtPoint2[1], rtPoint2[2]); 
      }
    else if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_PARABOLOID] )
      {
	//RiParaboloid(1.0, 0.0, 1.0, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nParaboloid 1 0 1 360");
      }
    else if ( m_pcGUIID == gcppcSLIDERIBObjectReservedWords[slideribotRI_TORUS] )
      {
	//RiTorus(1.0, 0.5, 0.0, 360.0, 360.0, RI_NULL);
	RiArchiveRecord(RI_COMMENT, "\nTorus 1 0.5 0 360 360");
      }
    else
      {
	ASSERT( FALSE );
      }
  }
  RiAttributeEnd();
  */
}

VOID CSLIDEObject::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE OBJECT: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDEObject::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE OBJECT: %s", m_pcGUIID);
}

VOID CSLIDEObject::WritePOVGeometry(FILE *fpOutputFile, 
				    CListInt *plistTransformIDs,
				    CListInt *plistColors) 
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  CHAR *pcTransformID;
  CHAR *pcColorID;

  if (m_listCSLIDEFace.GetLength() > 1)
    fprintf(fpOutputFile, "#declare %s = union { \n", GetID());
  else
    fprintf(fpOutputFile, "#declare %s = ", GetID());

  iter.Init(&m_listCSLIDEFace);
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->WritePOV(fpOutputFile, plistTransformIDs, plistColors);
    }

  if (m_listCSLIDEFace.GetLength() > 1)
    fprintf(fpOutputFile, "}\n");
  fprintf(fpOutputFile, "\n");

  fprintf(fpOutputFile, "object { \n %s \n", GetID());
  iter.Init(plistTransformIDs);
  for ( pcTransformID = (CHAR*)iter.PeekFirst(); 
	pcTransformID != NULL;
	pcTransformID = (CHAR*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, " transform %s\n", pcTransformID);
    }

  //color
  ASSERT(m_pCSLIDESurface != NULL);
  
  if (!(strcmp(GetSurfaceID(), "SLF_INHERIT")))
    {
      iter.Init(plistColors);
      pcColorID = (CHAR*)iter.PeekFirst();
      fprintf(fpOutputFile, " pigment { %s }\n", pcColorID);
    }
  else
    {
      fprintf(fpOutputFile, " pigment {");
      m_pCSLIDESurface->WritePOV(fpOutputFile);
      fprintf(fpOutputFile, " }\n");
    }

  fprintf(fpOutputFile, "}\n#undef %s\n\n", GetID());
}

VOID CSLIDEObject::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				     CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  CSLIDESurface *pCSLIDESurfacePush;


  fprintf(fpOutputFile, "Shape { # triangle mesh: OBJECT\n\n");

  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurfacePush = pCSLIDESurface;
    }
  else
    {
      pCSLIDESurfacePush = m_pCSLIDESurface;
    }

  // This color is not applied but need to print so lighting works
  pCSLIDESurfacePush->WriteWRLZ(fpOutputFile);

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform, pCSLIDESurfacePush);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDEObject::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  
  iter.Init(&m_listCSLIDEFace);
  
  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      pCSLIDEFace->WriteSTL(fpOutputFile, mTransform);
    }
}

VOID CSLIDEObject::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "object_%s", GetGUIID());
}

VOID CSLIDEObject::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDEObject::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

UINT_32 CSLIDEObject::GetFaceCount()
{
  return m_listCSLIDEFace.GetLength();
}

