#include "slidesweep.h"
#include "slidesweeppath.h"
#include "slidesweepcrosssection.h"
#include "slidecrosssection.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"

#include "qemesh.h"
#include "qevertex.h"
#include "qerender.h"
#include "qewrite.h"
#include "stl.h"

extern BOOL gbCheckSemanticsDone;

static const UINT_32 cuiMinimizeTorsion = 0;
static const UINT_32 cuiClosed = 0;
static const UINT_32 cuiSymmetry = 0;
static const FLOAT cfAzimuth = 0.0;
static const FLOAT cfTwist = 0.0;

static const UINT_32 cuiMeshType = slidemtSLF_TRIANGLES;
static const UINT_32 cuiDrawSweep = 1;
static const UINT_32 cuiDrawPath = 0;

static const FLOAT cfNormalLength = 1.0;
static const FLOAT cfNormalRatio = 1.0 / 10.0;

//////////////////////////////////////////////////////////////////////
// CSLIDESweep Class
//

CSLIDESweep::CSLIDESweep()
{
  m_fNormalLength = cfNormalLength;

  // Cross Section

  // Path
  m_pcPathID = NULL;
  m_pPath = NULL;

  m_dyniMinimizeTorsion.SetValue(cuiMinimizeTorsion);
  m_dyniClosed.SetValue(cuiClosed);
  m_dyniSymmetry.SetValue(cuiSymmetry);
  m_dynfAzimuth.SetValue(cfAzimuth);
  m_dynfTwist.SetValue(cfTwist);

  m_dyniMeshType.SetValue(cuiMeshType);
  m_dyniDrawPath.SetValue(cuiDrawPath);
  m_dyniDrawSweep.SetValue(cuiDrawSweep);
}

CSLIDESweep::~CSLIDESweep()
{
}

VOID CSLIDESweep::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_fNormalLength = cfNormalLength;

  // Cross Section
  m_listCrossSectionIDs.Init();
  m_listCSLIDESweepCrossSection.Init();

  // Path
  m_dyniMinimizeTorsion.SetValue(cuiMinimizeTorsion);
  m_dyniClosed.SetValue(cuiClosed);
  m_dyniSymmetry.SetValue(cuiSymmetry);
  m_dynfAzimuth.SetValue(cfAzimuth);
  m_dynfTwist.SetValue(cfTwist);

  m_dyniMeshType.SetValue(cuiMeshType);
  m_dyniDrawPath.SetValue(cuiDrawPath);
  m_dyniDrawSweep.SetValue(cuiDrawSweep);
}

VOID CSLIDESweep::Uninit()
{
  CSLIDEAtomNode::Uninit();
}

BOOL CSLIDESweep::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

VOID CSLIDESweep::AddCrossSectionID(CHAR *pcID, 
				    CSLIDESweepCrossSection *pCSLIDESweepCrossSection)
{
  ASSERT( pcID != NULL );

  m_listCrossSectionIDs.InsertLast(pcID, 0);  
  m_listCSLIDESweepCrossSection.InsertLast(pCSLIDESweepCrossSection, 0);
}

BOOL CSLIDESweep::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  BOOL bCorrect;
  CListIterInt iter, iter2;
  CHAR *pcID;
  CSLIDECrossSection *pCSLIDECrossSection;
  CSLIDESweepCrossSection *pCSLIDESweepCrossSection;

  bCorrect = TRUE;

  // Check Path Reference
  if( m_pcPathID != NULL )
    {
      m_pPath = (CSLIDESweepPath *)pCSLIDEWorld->PeekNode(m_pcPathID);

      if ( m_pPath == NULL )
	{
	  HandleError("Undefined sweep path \"%s\"", m_pcPathID);
	  bCorrect = FALSE;
	}
      else
	{
	  if( !gbCheckSemanticsDone )
	    {
	      // Use cross section for now, but should use path
	      m_pPath->InsertSweepParent(this);
	    }

	  m_pPath->SetMinimizeTorsion( &m_dyniMinimizeTorsion );
	  m_pPath->SetClosed( &m_dyniClosed );
	  m_pPath->SetAzimuth( &m_dynfAzimuth );
	  m_pPath->SetTwist( &m_dynfTwist );
	  m_pPath->SetSymmetry( &m_dyniSymmetry );	
	}
    }
  else
    {
      HandleError("sweep %s has no specified path", m_pcGUIID);
      bCorrect = FALSE;
    }

  // Check Cross Section References
  if( m_listCrossSectionIDs.GetLength() > 0 )
    {
      iter2.Init(&m_listCSLIDESweepCrossSection);

      iter.Init(&m_listCrossSectionIDs);
      for( pcID = (CHAR *)(iter.PeekFirst()), 
	     pCSLIDESweepCrossSection = (CSLIDESweepCrossSection*)(iter2.PeekFirst()); 
	   pcID != NULL;
	   pcID = (CHAR *)(iter.PeekNext()),
	     pCSLIDESweepCrossSection = (CSLIDESweepCrossSection*)(iter2.PeekNext()) )
	{
	  ASSERT( pCSLIDESweepCrossSection != NULL);

	  pCSLIDECrossSection = (CSLIDECrossSection *)pCSLIDEWorld->PeekNode(pcID);

	  if ( pCSLIDECrossSection == NULL )
	    {
	      HandleError("Undefined crosssection \"%s\"", pcID);
	      bCorrect = FALSE;
	    }
	  else
	    {
	      ASSERT( !strcmp( pCSLIDESweepCrossSection->GetID(), pcID) );

	      pCSLIDESweepCrossSection->SetCrossSection(pCSLIDECrossSection);
	      pCSLIDESweepCrossSection->SetSweepParent(this);
	      pCSLIDECrossSection->InsertSweepCrossSectionParent(pCSLIDESweepCrossSection);
	    }
	}
    }
  else
    {
      HandleError("sweep %s has no specified crosssection", m_pcGUIID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDESweep::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dyniDrawPath.GetValue() >= 0 ) &&
	   ( m_dyniDrawSweep.GetValue() >= 0 ) );
}

VOID CSLIDESweep::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;

  if( !GetDynamic() )
    {
      SetDynamic(m_dyniDrawSweep.GetDynamic() ||
		 m_dyniDrawPath.GetDynamic() ||
		 m_dyniMeshType.GetDynamic());
    }

  // Cross Sections
  iter.Init(&m_listCSLIDESweepCrossSection);
  for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
       pCSLIDESweepCS != NULL;
       pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
    {
      // Because SweepCS is not instanced must called checkdynamic for it
      pCSLIDESweepCS->CheckDynamic();

      if( pCSLIDESweepCS->GetDynamic() )
	{
	  SetDynamic(TRUE);
	}
    }

  // Path
  ASSERT( m_pPath != NULL );
  if( m_pPath->GetDynamic() )
    {
      SetDynamic(TRUE);
    }
}

VOID CSLIDESweep::PreprocessGraphLocal()
{
  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;

  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  // Path and crosssections may not be instanced so call preprocessgraph
  m_pPath->PreprocessGraphLocal();

  iter.Init(&m_listCSLIDESweepCrossSection);
  for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
       pCSLIDESweepCS != NULL;
       pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
    {
      pCSLIDESweepCS->PreprocessGraphLocal();
    }

  // Path and Crosssection set these appropriately
  bBoundDynamic = GetBoundDynamic();
  bGeometryDynamic = GetGeometryDynamic();
  bNormalsDynamic = GetNormalsDynamic();


  bRealSpaceDynamic = ( m_dyniDrawSweep.GetDynamic() ||
			m_dyniDrawPath.GetDynamic() ||
			m_dyniMeshType.GetDynamic());
  bParameterSpaceDynamic = FALSE;
  bTextureSpaceDynamic = FALSE;

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bTextureSpaceDynamic ||
		     bGeometryDynamic ||
		     ((m_pCSLIDESurface != NULL) && m_pCSLIDESurface->GetDynamic()) );
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || 
		    bParameterSpaceDynamic ||
		    bNormalsDynamic);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic ||
		   bBoundDynamic );
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic ||
		  bParameterSpaceDynamic ||
		  bBoundDynamic );
  SetChildrenBoundsDynamic(bRealSpaceDynamic ||
			   bParameterSpaceDynamic ||
			   bBoundDynamic);
}

VOID CSLIDESweep::Update()
{
  // For End Caps
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDESweep::UpdatePathPoints()
{
  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;

  iter.Init(&m_listCSLIDESweepCrossSection);
  for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
       pCSLIDESweepCS != NULL;
       pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
    {
      pCSLIDESweepCS->CalculateEndCaps();
    }
}

VOID CSLIDESweep::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  m_dyniMeshType.Update();
  m_dyniDrawPath.Update();
  m_dyniDrawSweep.Update();
}

VOID CSLIDESweep::CalculateBound()
{
  CBound bound;
  UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;
  Point *pPathPoint, *pCrossSectionPoint;
  Matrix *pmPathFrenet;
  Matrix mCS, m1;
  Point pt1, p1;
  Vector v, v1;

  uiSlices = m_pPath->GetSlices();
  pPathPoint = m_pPath->GetPoint();
  pmPathFrenet = m_pPath->GetFrenetMatrix();

  iter.Init(&m_listCSLIDESweepCrossSection);
  for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
       pCSLIDESweepCS != NULL;
       pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
    {
      uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
      pCrossSectionPoint = pCSLIDESweepCS->GetPoint();    
      mCS = pCSLIDESweepCS->GetTransform();

      for( uiSlice = 0; uiSlice < uiSlices+1; uiSlice++)
	{
	  pt1 = pPathPoint[uiSlice];
	  m1 = pmPathFrenet[uiSlice];        // Get Change of Basis Matrix and Warp
	  
	  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
	    {
	      p1 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
	      p1[W] = 1.0;
	      bound.Insert(p1);
	    }
	}
    }  

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweep::CalculateQEMesh()
{  
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
      UINT_32 uiBeginCap, uiEndCap;
      CListIterInt iter;
      CSLIDESweepCrossSection *pCSLIDESweepCS;
      Point *pPathPoint, *pCrossSectionPoint;
      Matrix *pmPathFrenet;
      Matrix mCS, m1, m2;
      Point pt1, pt2, p1, p2;
      QEMesh *pQEMesh, *pQEMeshFirst, *pQEMeshLast;
      CSLIDESurface *pCSLIDESurface;
      BOOL bTriangles;

      QEVertex* pqev;
      CObjectList<QEVertex> listQEVertex;

      ASSERT( m_pQEMesh == NULL );
      
      m_pQEMesh = AllocateQEMesh();
      ASSERT( m_pQEMesh != NULL );
      m_pQEMesh->Init();
      listQEVertex.Init();

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();
      
      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  mCS = pCSLIDESweepCS->GetTransform();

	  pCSLIDESurface = pCSLIDESweepCS->GetSurface();
	  
	  for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	    {
	      listQEVertex.RemoveAll();
	      pt1 = pPathPoint[uiSlice];
	      pt2 = pPathPoint[uiSlice+1];
	  
	      // Get Change of Basis Matrix and Warp
	      m1 = pmPathFrenet[uiSlice];
	      m2 = pmPathFrenet[uiSlice+1];
	      
	      //gilBegin(GL_TRIANGLE_STRIP);
	      {
		for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		  {
		    p1 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		    p1[W] = 1.0;
		    p2 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		    p2[W] = 1.0;

		    pqev = AllocateQEVertex();
		    pqev->Init(p1);
		    listQEVertex.InsertLast(pqev);
		
		    pqev = AllocateQEVertex();
		    pqev->Init(p2);
		    listQEVertex.InsertLast(pqev);
		  }
	      }
	      //m_pQEMesh->MergeTriangleStrip(&listQEVertex);
	      if( bTriangles )
		m_pQEMesh->MergeTriangleStrip(&listQEVertex, pCSLIDESurface);
	      else
		m_pQEMesh->MergeQuadStrip(&listQEVertex, pCSLIDESurface);
	      //gilEnd();
	    }

	  // End Caps
	  uiBeginCap = pCSLIDESweepCS->GetBeginCapValue();
	  uiEndCap = pCSLIDESweepCS->GetEndCapValue();
	  pQEMeshFirst = pCSLIDESweepCS->GetMeshBeginCap();
	  pQEMeshLast = pCSLIDESweepCS->GetMeshEndCap();
	  
	  if( uiBeginCap && (pQEMeshFirst != NULL) ) 
	    {
	      pQEMesh = pQEMeshFirst->Copy();
	      m_pQEMesh->MergeMesh(pQEMesh, pCSLIDESurface); // Merge and Deallocate pQEMesh
	    }

	  if( uiEndCap && (pQEMeshLast != NULL) )
	    {
	      pQEMesh = pQEMeshLast->Copy();
	      m_pQEMesh->MergeMesh(pQEMesh, pCSLIDESurface); // Merge and Deallocate pQEMesh
	    }
	}
    }
}

VOID CSLIDESweep::CalculateLeafQEMesh()
{  
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDESweep::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweep::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{  
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderGeometrySmooth();
    }
  
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
      CListIterInt iter;
      CSLIDESweepCrossSection *pCSLIDESweepCS;
      Point *pPathPoint, *pCrossSectionPoint;
      Vector *pvCrossSectionNormal;
      DynVector *pvTexture;
      Matrix *pmPathFrenet;
      Matrix mCS, m1, m2;
      Point pt1, pt2, p1, p2;
      Vector v, v1, v2;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvV[2][2];
      Vector pvU[2];
      Bool bTriangles;

      bTriangles = (m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES);

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  pvCrossSectionNormal = pCSLIDESweepCS->GetNormal();
	  pvTexture = pCSLIDESweepCS->GetTexture();
	  mCS = pCSLIDESweepCS->GetTransform();

	  pCSLIDESweepCS->PushSurface();

	  pvVDelta[0] = -(pvTexture[3] - pvTexture[0]) / (FLOAT)uiSlices;
	  pvVDelta[1] = -(pvTexture[2] - pvTexture[1]) / (FLOAT)uiSlices;
	  ppvV[0][0] = pvTexture[0];
	  ppvV[0][1] = pvTexture[1];
	  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
	  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];

	  for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	    {
	      pt1 = pPathPoint[uiSlice];
	      pt2 = pPathPoint[uiSlice+1];
	      
	      // Get Change of Basis Matrix and Warp
	      m1 = pmPathFrenet[uiSlice];
	      m2 = pmPathFrenet[uiSlice+1];
	      
	      pvUDelta[0] = -(ppvV[0][1] - ppvV[0][0]) / (FLOAT)(uiCrossSectionPoints-1);
	      pvUDelta[1] = -(ppvV[1][1] - ppvV[1][0]) / (FLOAT)(uiCrossSectionPoints-1);
	      pvU[0] = ppvV[0][0];
	      pvU[1] = ppvV[1][0];

	      if( bTriangles )
		gilBegin(GL_TRIANGLE_STRIP);
	      else
		gilBegin(GL_QUAD_STRIP);
	      {
		for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		  {
		    p1 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		    p1[W] = 1.0;
		    p2 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		    p2[W] = 1.0;
		    v1 = m1 * pvCrossSectionNormal[uiCSSlice];
		    v2 = m2 * pvCrossSectionNormal[uiCSSlice];
		    
		    gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
		    gilNormal3fv( v1.Get() );
		    gilVertex3fv( p1.Get() );
		    
		    gilTexCoord2f(pvU[1][0], pvU[1][1]);	
		    gilNormal3fv( v2.Get() );
		    gilVertex3fv( p2.Get() );
		    
		    pvU[0] += pvUDelta[0];
		    pvU[1] += pvUDelta[1];
		  }
	      }
	      gilEnd();
	      
	      ppvV[0][0] += pvVDelta[0];
	      ppvV[0][1] += pvVDelta[1];
	      ppvV[1][0] += pvVDelta[0];
	      ppvV[1][1] += pvVDelta[1];
	    }

	  pCSLIDESweepCS->RenderEndCapGeometry();

	  pCSLIDESweepCS->PopSurface();
	}
    }
}

VOID CSLIDESweep::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderGeometryFlat();
    }

  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
      CListIterInt iter;
      CSLIDESweepCrossSection *pCSLIDESweepCS;
      Point *pPathPoint, *pCrossSectionPoint;
      Vector *pvCrossSectionNormal;
      DynVector *pvTexture;
      Matrix *pmPathFrenet;
      Matrix mCS, m1, m2;
      Point pt1, pt2, p1, p2, p3, p4;
      Vector v, v1, v2, v3, v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvV[2][2];
      Vector pvU[2];
      BOOL bTriangles;

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  pvCrossSectionNormal = pCSLIDESweepCS->GetNormal();
	  pvTexture = pCSLIDESweepCS->GetTexture();
	  mCS = pCSLIDESweepCS->GetTransform();

	  pCSLIDESweepCS->PushSurface();

	  pvVDelta[0] = -(pvTexture[3] - pvTexture[0]) / (FLOAT)uiSlices;
	  pvVDelta[1] = -(pvTexture[2] - pvTexture[1]) / (FLOAT)uiSlices;
	  ppvV[0][0] = pvTexture[0];
	  ppvV[0][1] = pvTexture[1];
	  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
	  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
	  
	  for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	    {
	      pt1 = pPathPoint[uiSlice];
	      pt2 = pPathPoint[uiSlice+1];
	      
	      // Get Change of Basis Matrix and Warp
	      m1 = pmPathFrenet[uiSlice];
	      m2 = pmPathFrenet[uiSlice+1];
	      
	      pvUDelta[0] = -(ppvV[0][1] - ppvV[0][0]) / (FLOAT)(uiCrossSectionPoints-1);
	      pvUDelta[1] = -(ppvV[1][1] - ppvV[1][0]) / (FLOAT)(uiCrossSectionPoints-1);
	      pvU[0] = ppvV[0][0];
	      pvU[1] = ppvV[1][0];

	      if( bTriangles )
		gilBegin(GL_TRIANGLE_STRIP);
	      else
		gilBegin(GL_QUAD_STRIP);	      
	      {
		uiCSSlice = 0;
		p1 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		p1[W] = 1.0;
		p2 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		p2[W] = 1.0;
		p3 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice+1];
		p3[W] = 1.0;

		gilNormal3fv( ((p2-p1)*(p3-p1)).Get() );
		gilTexCoord2f(pvU[0][0], pvU[0][1]);	
		gilVertex3fv( p1.Get() );
		gilNormal3fv( ((p4-p2)*(p3-p2)).Get() );
		gilTexCoord2f(pvU[1][0], pvU[1][1]);	
		gilVertex3fv( p2.Get() );

		uiCSSlice++;
		pvU[0] += pvUDelta[0];
		pvU[1] += pvUDelta[1];
		for( ; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		  {
		    p3 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		    p3[W] = 1.0;
		    p4 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		    p4[W] = 1.0;
		    
		    gilNormal3fv( ((p2-p1)*(p3-p1)).Get() );
		    gilTexCoord2f(pvU[0][0], pvU[0][1]);	
		    gilVertex3fv( p3.Get() );

		    gilNormal3fv( ((p4-p2)*(p3-p2)).Get() );
		    gilTexCoord2f(pvU[1][0], pvU[1][1]);	
		    gilVertex3fv( p4.Get() );
			
		    p1 = p3;
		    p2 = p4;
		    pvU[0] += pvUDelta[0];
		    pvU[1] += pvUDelta[1];
		  }
	      }
	      glEnd();

	      ppvV[0][0] += pvVDelta[0];
	      ppvV[0][1] += pvVDelta[1];
	      ppvV[1][0] += pvVDelta[0];
	      ppvV[1][1] += pvVDelta[1];
	    }
	  
	  pCSLIDESweepCS->RenderEndCapGeometry();

	  pCSLIDESweepCS->PopSurface();
	}
    }
}

VOID CSLIDESweep::RenderNormalsSmooth()
{
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderNormalsSmooth();
    }

  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
      CListIterInt iter;
      CSLIDESweepCrossSection *pCSLIDESweepCS;
      Point *pPathPoint, *pCrossSectionPoint;
      Vector *pvCrossSectionNormal;
      Matrix *pmPathFrenet;
      Matrix mCS, m1;
      Vector v1;
      Point pt, pt1, p1;

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  pvCrossSectionNormal = pCSLIDESweepCS->GetNormal();
	  mCS = pCSLIDESweepCS->GetTransform();
 
	  gilBegin(GL_LINES);
	  {
	    for( uiSlice = 0; uiSlice < uiSlices+1; uiSlice++)
	      {
		pt1 = pPathPoint[uiSlice];
		m1 = pmPathFrenet[uiSlice];
		
		for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		  {
		    p1 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		    p1[W] = 1.0;
		    v1 = m1 * pvCrossSectionNormal[uiCSSlice];
		    
		    gilVertex3fv( p1.Get() );
		    pt = p1 + v1 * m_fNormalLength;
		    gilVertex3fv( pt.Get() );
		  }
	      }
	  }
	  gilEnd();
	}
    }
}

VOID CSLIDESweep::RenderNormalsFlat()
{
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderNormalsFlat();
    }

  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
      CListIterInt iter;
      CSLIDESweepCrossSection *pCSLIDESweepCS;
      Point *pPathPoint, *pCrossSectionPoint;
      Matrix *pmPathFrenet;
      Matrix mCS, m1, m2;
      Vector v, v1, v2;
      Point pt, pt1, pt2, p1, p2, p3, p4;
      BOOL bTriangles;

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  mCS = pCSLIDESweepCS->GetTransform();

	  gilBegin(GL_LINES);
	  {
	    for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	      {
		pt1 = pPathPoint[uiSlice];
		pt2 = pPathPoint[uiSlice+1];
		
		m1 = pmPathFrenet[uiSlice];
		m2 = pmPathFrenet[uiSlice+1];
		
		p1 = pt1 + m1 * mCS * pCrossSectionPoint[0];
		p1[W] = 1.0;
		p2 = pt2 + m2 * mCS * pCrossSectionPoint[0];
		p2[W] = 1.0;
		
		for( uiCSSlice = 1; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		  {
		    p3 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		    p3[W] = 1.0;
		    p4 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		    p4[W] = 1.0;
	
		    if( bTriangles )
		      {
			pt = (p1+p2+p3)/3.0;
			gilVertex3fv( pt.Get() );
			v1 = (p2-p1)*(p3-p1);
			v1.Normalize();
			pt = pt + v1 * m_fNormalLength;
			gilVertex3fv( pt.Get() );
		    
			pt = (p2+p3+p4)/3.0;
			gilVertex3fv( pt.Get() );
			v1 = (p4-p2)*(p3-p2);
			v1.Normalize();
			pt = pt + v1 * m_fNormalLength;
			gilVertex3fv( pt.Get() );
		      }
		    else // Quads
		      {
			pt = (p1+p2+p3+p4)/4.0;
			gilVertex3fv( pt.Get() );
			v1 = (p2-p1)*(p3-p1);
			v1.Normalize();
			v2 = (p4-p2)*(p3-p2);
			v2.Normalize();
			v = (v1 + v2) / 2.0;
			v.Normalize();
			pt = pt + v * m_fNormalLength;
			gilVertex3fv( pt.Get() );
		      }
		    
		    p1 = p3;
		    p2 = p4;
		  }
	      }
	  }
	  gilEnd(); 
	}
    }
}

VOID CSLIDESweep::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtSweep], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  if( m_pcPathID != NULL )
    {
      ASSERT( m_pPath != NULL );
      m_pPath->CSLIDESweepPath::WriteSLIDE(fpOutputFile);
    }

  iter.Init(&m_listCSLIDESweepCrossSection);
  for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
       pCSLIDESweepCS != NULL;
       pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
    {
      pCSLIDESweepCS->WriteSLIDEParent(fpOutputFile);
    }

  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtMeshType],
	  gcppcSLIDEMeshTypeReservedWords[m_dyniMeshType.GetValue()]);

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawPath],
	  m_dyniDrawPath.GetValue());

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawSweep],
	  m_dyniDrawSweep.GetValue());

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndSweep]);
}

VOID CSLIDESweep::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
      CListIterInt iter;
      CSLIDESweepCrossSection *pCSLIDESweepCS;
      Point *pPathPoint, *pCrossSectionPoint;
      Vector *pvCrossSectionNormal;
      DynVector *pvTexture;
      Matrix *pmPathFrenet;
      Matrix mCS, m1, m2;
      Point pt1, pt2, p1, p2, p3, p4;
      Vector v, v1, v2, v3, v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvV[2][2];
      Vector pvU[2];

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  pvCrossSectionNormal = pCSLIDESweepCS->GetNormal();
	  pvTexture = pCSLIDESweepCS->GetTexture();
	  mCS = pCSLIDESweepCS->GetTransform();

	  pvVDelta[0] = -(pvTexture[3] - pvTexture[0]) / (FLOAT)uiSlices;
	  pvVDelta[1] = -(pvTexture[2] - pvTexture[1]) / (FLOAT)uiSlices;
	  ppvV[0][0] = pvTexture[0];
	  ppvV[0][1] = pvTexture[1];
	  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
	  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
	  
	  for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	    {
	      pt1 = pPathPoint[uiSlice];
	      pt2 = pPathPoint[uiSlice+1];
	      
	      // Get Change of Basis Matrix and Warp
	      m1 = pmPathFrenet[uiSlice];
	      m2 = pmPathFrenet[uiSlice+1];
	      
	      pvUDelta[0] = -(ppvV[0][1] - ppvV[0][0]) / (FLOAT)(uiCrossSectionPoints-1);
	      pvUDelta[1] = -(ppvV[1][1] - ppvV[1][0]) / (FLOAT)(uiCrossSectionPoints-1);
	      pvU[0] = ppvV[0][0];
	      pvU[1] = ppvV[1][0];
	      
	      for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints-1; uiCSSlice++)
		{
		  p1 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		  p1[W] = 1.0;
		  p2 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		  p2[W] = 1.0;
		  p3 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice+1];
		  p3[W] = 1.0;
		  p4 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice+1];
		  p4[W] = 1.0;	      
		  
		  if( (eShading == slideshtSLF_GOURAUD) || (eShading == slideshtSLF_PHONG) )
		    {
		      v1 = m1 * pvCrossSectionNormal[uiCSSlice];
		      v2 = m2 * pvCrossSectionNormal[uiCSSlice];
		      v3 = m1 * pvCrossSectionNormal[uiCSSlice+1];
		      v4 = m2 * pvCrossSectionNormal[uiCSSlice+1];
		      
		      RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				      p1[X], p1[Y], p1[Z], 
				      p2[X], p2[Y], p2[Z], 
				      p3[X], p3[Y], p3[Z],
				      v1[X], v1[Y], v1[Z], 
				      v2[X], v2[Y], v2[Z], 
				      v3[X], v3[Y], v3[Z],
				      pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], 
				      (pvU[0] + pvUDelta[0])[0], (pvU[0] + pvUDelta[0])[1]);
		      
		      pvU[0] += pvUDelta[0];
		      
		      RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				      p3[X], p3[Y], p3[Z], 
				      p2[X], p2[Y], p2[Z], 
				      p4[X], p4[Y], p4[Z],
				      v3[X], v3[Y], v3[Z], 
				      v2[X], v2[Y], v2[Z], 
				      v4[X], v4[Y], v4[Z],
				      pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1],
				      (pvU[1] + pvUDelta[1])[0], (pvU[1] + pvUDelta[1])[1]);
		      pvU[1] += pvUDelta[1];
		    }
		  else
		    {
		      RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				      p1[X], p1[Y], p1[Z], 
				      p2[X], p2[Y], p2[Z], 
				      p3[X], p3[Y], p3[Z],
				      pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], 
				      (pvU[0] + pvUDelta[0])[0], (pvU[0] + pvUDelta[0])[1]);
		      pvU[0] += pvUDelta[0];
		      
		      RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				      p3[X], p3[Y], p3[Z], 
				      p2[X], p2[Y], p2[Z], 
				      p4[X], p4[Y], p4[Z],
				      pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], 
				      (pvU[1] + pvUDelta[1])[0], (pvU[1] + pvUDelta[1])[1]);
		      pvU[1] += pvUDelta[1];
		    }
		}
	      ppvV[0][0] += pvVDelta[0];
	      ppvV[0][1] += pvVDelta[1];
	      ppvV[1][0] += pvVDelta[0];
	      ppvV[1][1] += pvVDelta[1];
	    }

	  pCSLIDESweepCS->WriteRIBEndCaps(fpOutputFile);
	}
    }
}

VOID CSLIDESweep::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE Sweep: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDESweep::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE Sweep: %s", m_pcGUIID);
}

VOID CSLIDESweep::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				    CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  CSLIDESurface *pCSLIDESurfacePush;

  fprintf(fpOutputFile, "Shape { # triangle mesh: SWEEP\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurfacePush = pCSLIDESurface;
    }
  else
    {
      pCSLIDESurfacePush = m_pCSLIDESurface;
    }

  pCSLIDESurfacePush->WriteWRLZ(fpOutputFile);

  // Mesh
  ASSERT(m_pQEMesh != NULL);

  if( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES )
    ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform, pCSLIDESurfacePush);
  else
    ::WriteWRLZQuads(m_pQEMesh, fpOutputFile, mTransform, pCSLIDESurfacePush);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDESweep::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiSlice, uiSlices, uiCSSlice, uiCrossSectionPoints;
  CListIterInt iter;
  CSLIDESweepCrossSection *pCSLIDESweepCS;
  Point *pPathPoint, *pCrossSectionPoint;     
  Matrix *pmPathFrenet;
  Matrix mCS, m1, m2;
  Point pt1, pt2, p1, p2, p3, p4;

  uiSlices = m_pPath->GetSlices();
  pPathPoint = m_pPath->GetPoint();
  pmPathFrenet = m_pPath->GetFrenetMatrix();

  if ( mTransform.Determinant() < 0 ) // Need to Mirror triangles
    {
      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  pCSLIDESweepCS->WriteSTLEndCaps(fpOutputFile, mTransform);

	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  mCS = pCSLIDESweepCS->GetTransform();

	  for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	    {
	      pt1 = pPathPoint[uiSlice];
	      pt2 = pPathPoint[uiSlice+1];
	      
	      // Get Change of Basis Matrix and Warp
	      m1 = pmPathFrenet[uiSlice];
	      m2 = pmPathFrenet[uiSlice+1];
	      
	      p1 = pt1 + m1 * mCS * pCrossSectionPoint[0];
	      p1[W] = 1.0;
	      p2 = pt2 + m2 * mCS * pCrossSectionPoint[0];
	      p2[W] = 1.0;
	      
	      p1 = mTransform * p1;
	      p2 = mTransform * p2;
	      
	      for( uiCSSlice = 1; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		{
		  p3 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		  p3[W] = 1.0;
		  p4 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		  p4[W] = 1.0;
		  
		  p3 = mTransform * p3;
		  p4 = mTransform * p4;
		  
		  WriteSTLTriangle(fpOutputFile, p3, p2, p1);
		  WriteSTLTriangle(fpOutputFile, p2, p3, p4);
		  
		  p1 = p3;
		  p2 = p4;
		}
	    }
	}
    }
  else // Don't mirror triangles
    {
      iter.Init(&m_listCSLIDESweepCrossSection);
      for( pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekFirst()); 
	   pCSLIDESweepCS != NULL;
	   pCSLIDESweepCS = (CSLIDESweepCrossSection *)(iter.PeekNext()) )
	{
	  pCSLIDESweepCS->WriteSTLEndCaps(fpOutputFile, mTransform);

	  uiCrossSectionPoints = pCSLIDESweepCS->GetPoints();
	  pCrossSectionPoint = pCSLIDESweepCS->GetPoint();
	  mCS = pCSLIDESweepCS->GetTransform();

	  for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	    {
	      pt1 = pPathPoint[uiSlice];
	      pt2 = pPathPoint[uiSlice+1];
	      
	      // Get Change of Basis Matrix and Warp
	      m1 = pmPathFrenet[uiSlice];
	      m2 = pmPathFrenet[uiSlice+1];
	      
	      p1 = pt1 + m1 * mCS * pCrossSectionPoint[0];
	      p1[W] = 1.0;
	      p2 = pt2 + m2 * mCS * pCrossSectionPoint[0];
	      p2[W] = 1.0;
	      
	      p1 = mTransform * p1;
	      p2 = mTransform * p2;
	      
	      for( uiCSSlice = 1; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
		{
		  p3 = pt1 + m1 * mCS * pCrossSectionPoint[uiCSSlice];
		  p3[W] = 1.0;
		  p4 = pt2 + m2 * mCS * pCrossSectionPoint[uiCSSlice];
		  p4[W] = 1.0;
		  
		  p3 = mTransform * p3;
		  p4 = mTransform * p4;
		  
		  WriteSTLTriangle(fpOutputFile, p1, p2, p3);
		  WriteSTLTriangle(fpOutputFile, p4, p3, p2);
		  
		  p1 = p3;
		  p2 = p4;
		}
	    }
	}
    }
}


