#include "slidesweepmorph.h"
#include "slidesweeppath.h"
#include "slidecrosssection.h"
#include "slidecontrolpoint.h"
#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"

#include "qemesh.h"
#include "qevertex.h"
#include "qerender.h"
#include "qewrite.h"
#include "stl.h"

extern BOOL gbCheckSemanticsDone;

static const UINT_32 cuiMinimizeTorsion = 0;
static const UINT_32 cuiClosed = 0;
static const UINT_32 cuiSymmetry = 0;
static const FLOAT cfAzimuth = 0.0;
static const FLOAT cfTwist = 0.0;

static const UINT_32 cuiMeshType = slidemtSLF_TRIANGLES;
static const UINT_32 cuiDrawSweep = 1;
static const UINT_32 cuiDrawPath = 0;

static const UINT_32 cuiEndCapFirst = 0;
static const UINT_32 cuiEndCapLast = 0;

static const FLOAT cfNormalLength = 1.0;
static const FLOAT cfNormalRatio = 1.0 / 10.0;

//////////////////////////////////////////////////////////////////////
// CSLIDESweep Class
//

CSLIDESweepMorph::CSLIDESweepMorph()
{
  m_fNormalLength = cfNormalLength;

  m_bOKToUpdate = FALSE;

  // Cross Section
  m_ppCrossSectionPoint = NULL;
  m_uiCrossSectionPoints = 0;
  m_uiCrossSectionPointsPrev = m_uiCrossSectionPoints;
  m_uiPathPoints = 0;
  m_uiPathPointsPrev = m_uiPathPoints;

  // Path
  m_pcPathID = NULL;
  m_pPath = NULL;

  m_dyniMinimizeTorsion.SetValue(cuiMinimizeTorsion);
  m_dyniClosed.SetValue(cuiClosed);
  m_dyniSymmetry.SetValue(cuiSymmetry);
  m_dynfAzimuth.SetValue(cfAzimuth);
  m_dynfTwist.SetValue(cfTwist);

  m_dyniEndCapFirst.SetValue(cuiEndCapFirst);
  m_dyniEndCapLast.SetValue(cuiEndCapLast);
  m_pQEMeshEndCapFirst = NULL;
  m_pQEMeshEndCapLast = NULL;

  // General
  m_dyniMeshType.SetValue(cuiMeshType);
  m_dyniDrawPath.SetValue(cuiDrawPath);
  m_dyniDrawSweep.SetValue(cuiDrawSweep);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

CSLIDESweepMorph::~CSLIDESweepMorph()
{
}

VOID CSLIDESweepMorph::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_fNormalLength = cfNormalLength;

  // Cross Section
  m_uiCrossSectionPoints = 0;
  m_uiCrossSectionPointsPrev = m_uiCrossSectionPoints;
  m_uiPathPoints = 0;
  m_uiPathPointsPrev = m_uiPathPoints;

  // Path
  m_dyniMinimizeTorsion.SetValue(cuiMinimizeTorsion);
  m_dyniClosed.SetValue(cuiClosed);
  m_dyniSymmetry.SetValue(cuiSymmetry);
  m_dynfAzimuth.SetValue(cfAzimuth);
  m_dynfTwist.SetValue(cfTwist);

  m_dyniMeshType.SetValue(cuiMeshType);
  m_dyniDrawPath.SetValue(cuiDrawPath);
  m_dyniDrawSweep.SetValue(cuiDrawSweep);

  m_dyniEndCapFirst.SetValue(cuiEndCapFirst);
  m_dyniEndCapLast.SetValue(cuiEndCapLast);
}

VOID CSLIDESweepMorph::Uninit()
{
  CSLIDEAtomNode::Uninit();

  DeallocateCrossSectionPoints();
}

VOID CSLIDESweepMorph::DeallocateCrossSectionPoints()
{
  UINT_32 i;
  
  if ( m_ppCrossSectionPoint != NULL )
    {
      for( i = 0; i < m_uiCrossSectionPointsPrev; i++ ) 
	{
	  delete[] m_ppCrossSectionPoint[i];
	}
      delete[] m_ppCrossSectionPoint;
      m_ppCrossSectionPoint = NULL;
    }
}

VOID CSLIDESweepMorph::ReallocateCrossSectionPoints()
{
  UINT_32 i;

  if( m_ppCrossSectionPoint != NULL )
    {
      DeallocateCrossSectionPoints();
    }

  m_uiCrossSectionPointsPrev = m_uiCrossSectionPoints;
  m_ppCrossSectionPoint = new Point*[m_uiCrossSectionPoints];
  ASSERT(m_ppCrossSectionPoint != NULL);
  
  for( i = 0; i < m_uiCrossSectionPoints; i++ ) 
    {
      m_ppCrossSectionPoint[i] = new Point[m_uiPathPoints];
      ASSERT(m_ppCrossSectionPoint[i] != NULL);
    }
}

BOOL CSLIDESweepMorph::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDESweepMorph::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  BOOL bCorrect;
  CListIterInt iter, iter2;
  CListInt *plistCSLIDEControlPoints;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CSLIDECrossSection *pCSLIDECrossSection;

  bCorrect = TRUE;

  // Check Path Reference
  if( m_pcPathID != NULL )
    {
      m_pPath = (CSLIDESweepPath *)pCSLIDEWorld->PeekNode(m_pcPathID);

      if ( m_pPath == NULL )
	{
	  HandleError("Undefined sweep path \"%s\"", m_pcPathID);
	  bCorrect = FALSE;
	}
      else
	{
	  if( !gbCheckSemanticsDone )
	    {
	      // Use cross section for now, but should use path
	      m_pPath->InsertSweepMorphParent(this);
	    }

	  m_pPath->SetMinimizeTorsion( &m_dyniMinimizeTorsion );
	  m_pPath->SetClosed( &m_dyniClosed );
	  m_pPath->SetAzimuth( &m_dynfAzimuth );
	  m_pPath->SetTwist( &m_dynfTwist );
	  m_pPath->SetSymmetry( &m_dyniSymmetry );
	
	  // Check if end points have Cross Section definitions
	  plistCSLIDEControlPoints = m_pPath->GetControlPointList();
	  
	  iter.Init(plistCSLIDEControlPoints);
	  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());

	  if( pCSLIDEControlPoint->GetCrossSectionID() == NULL )
	    {
	      HandleError("sweepmorph \"%s\" does not have a crosssection specified in the first controlpoint of its path \"%s\"", m_pcGUIID, m_pPath->GetID());
	      bCorrect = FALSE;
	    }

	  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekLast());

	  if( pCSLIDEControlPoint->GetCrossSectionID() == NULL )
	    {
	      HandleError("sweepmorph \"%s\" does not have a crosssection specified in the last controlpoint of its path \"%s\"", m_pcGUIID, m_pPath->GetID());
	      bCorrect = FALSE;
	    }

	  // Insert me into child cross section
	  iter.Init(plistCSLIDEControlPoints);
	  for ( pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekFirst(); 
		pCSLIDEControlPoint != NULL;
		pCSLIDEControlPoint = (CSLIDEControlPoint*)iter.PeekNext() )
	    {
	      pCSLIDECrossSection = pCSLIDEControlPoint->GetCrossSection();

	      if( pCSLIDECrossSection != NULL )
		{
		  pCSLIDECrossSection->InsertSweepMorphParent(this);
		}
	    }
	}
    }
  else
    {
      HandleError("sweep %s has no specified path", m_pcGUIID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDESweepMorph::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( m_dyniDrawPath.GetValue() >= 0 ) &&
	   ( m_dyniDrawSweep.GetValue() >= 0 ) &&
	   ( m_dyniEndCapFirst.GetValue() >= 0) &&
	   ( m_dyniEndCapLast.GetValue() >= 0) );
}

VOID CSLIDESweepMorph::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  CListIterInt iter;

  if( !GetDynamic() )
    {
      SetDynamic(m_dyniDrawSweep.GetDynamic() ||
		 m_dyniDrawPath.GetDynamic() ||
		 m_dyniMeshType.GetDynamic() ||
		 m_dyniEndCapFirst.GetDynamic() ||
		 m_dyniEndCapLast.GetDynamic() );
    }

  // Path and Crosssection
  ASSERT( m_pPath != NULL );
  m_pPath->CheckSweepMorphCrossSectionDynamic();

  if( (m_pPath->GetDynamic()) || m_pPath->GetSweepMorphCrossSectionDynamic() )
    {
      SetDynamic(TRUE);
    }
}

VOID CSLIDESweepMorph::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  BOOL bBoundDynamic;
  BOOL bGeometryDynamic;
  BOOL bNormalsDynamic;

  // Path may not be instanced so call preprocessgraph
  m_pPath->PreprocessGraphLocal();

  // Path set these appropriately
  bBoundDynamic = GetBoundDynamic();
  bGeometryDynamic = GetGeometryDynamic();
  bNormalsDynamic = GetNormalsDynamic();


  bBoundDynamic = bBoundDynamic || m_pPath->GetBoundDynamic();
  bGeometryDynamic = bGeometryDynamic || m_pPath->GetGeometryDynamic();
  bNormalsDynamic = bNormalsDynamic || m_pPath->GetNormalsDynamic();

  bRealSpaceDynamic = ( m_dyniDrawSweep.GetDynamic() ||
			m_dyniDrawPath.GetDynamic() ||
			m_dyniMeshType.GetDynamic() );
  bParameterSpaceDynamic = ( m_dyniEndCapFirst.GetDynamic() ||
			     m_dyniEndCapLast.GetDynamic() );
  bTextureSpaceDynamic = FALSE;

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || 
		     bParameterSpaceDynamic ||
		     bTextureSpaceDynamic ||
		     bGeometryDynamic ||
		     ((m_pCSLIDESurface != NULL) && m_pCSLIDESurface->GetDynamic()) );
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || 
		    bParameterSpaceDynamic ||
		    bNormalsDynamic);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic ||
		   bParameterSpaceDynamic ||
		   bBoundDynamic );
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic ||
		  bParameterSpaceDynamic ||
		  bBoundDynamic );
  SetChildrenBoundsDynamic(bRealSpaceDynamic ||
			   bParameterSpaceDynamic ||
			   bBoundDynamic);

  // PreprocessGraph is the last function to get called if update isn't
  // Calculate Cross Section the first time
  CalculateCrossSection();
  m_bOKToUpdate = TRUE;
}

VOID CSLIDESweepMorph::Update()
{
  // For End Caps
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);

  m_bOKToUpdate = TRUE;
}

VOID CSLIDESweepMorph::UpdatePathPoints()
{
  if( m_bOKToUpdate )
    {
      CalculateCrossSection();
      CalculateEndCaps();
    }
}

VOID  CSLIDESweepMorph::UpdateCrossSectionPoints()
{
  if( m_bOKToUpdate )
    {
      CalculateCrossSection();
      CalculateEndCaps();
    }
}

VOID CSLIDESweepMorph::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  UINT_32 i;
  BOOL bEndCapFirstChanged, bEndCapLastChanged;

  bEndCapFirstChanged = FALSE;
  bEndCapLastChanged = FALSE;

  i = m_dyniEndCapFirst.GetValue();
  m_dyniEndCapFirst.Update();
  bEndCapFirstChanged = ( i != (UINT_32)m_dyniEndCapFirst.GetValue() );

  i = m_dyniEndCapLast.GetValue();
  m_dyniEndCapLast.Update();
  bEndCapLastChanged = ( i != (UINT_32)m_dyniEndCapLast.GetValue() );

  m_dyniMeshType.Update();
  m_dyniDrawPath.Update();
  m_dyniDrawSweep.Update();

  // Recalculate End Caps
  if( (m_dyniEndCapFirst.GetValue() && 
       (m_pQEMeshEndCapFirst == NULL) || bEndCapFirstChanged ) ||
      (m_dyniEndCapLast.GetValue() && 
       (m_pQEMeshEndCapLast == NULL) || bEndCapLastChanged ) )
    {
      CalculateEndCaps();      
    }
}

VOID CSLIDESweepMorph::CalculateBound()
{
  CBound bound;
  UINT_32 uiSlice, uiSlices, uiCSSlice;
  CListIterInt iter;
  Point *pPathPoint;
  Matrix *pmPathFrenet;
  Matrix m1;
  Point pt1, p1;

  uiSlices = m_pPath->GetSlices();
  pPathPoint = m_pPath->GetPoint();
  pmPathFrenet = m_pPath->GetFrenetMatrix();
  
  for( uiSlice = 0; uiSlice <= uiSlices; uiSlice++)
    {
      // Path
      pt1 = pPathPoint[uiSlice];
      m1 = pmPathFrenet[uiSlice];
      
      for( uiCSSlice = 0; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	{
	  p1 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
	  bound.Insert(p1);
	}
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);

      // Calculate normal length
      m_fNormalLength = bound.GetRadius() * cfNormalRatio;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepMorph::CalculateQEMesh()
{  
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice;
      CListIterInt iter;
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Matrix m1, m2;
      Point pt1, pt2, p1, p2;
      QEMesh *pQEMesh;
      BOOL bTriangles;

      QEVertex* pqev;
      CObjectList<QEVertex> listQEVertex;

      ASSERT( m_pQEMesh == NULL );
      
      m_pQEMesh = AllocateQEMesh();
      ASSERT( m_pQEMesh != NULL );
      m_pQEMesh->Init();
      listQEVertex.Init();

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();
      
      for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	{
	  listQEVertex.RemoveAll();
	  pt1 = pPathPoint[uiSlice];
	  pt2 = pPathPoint[uiSlice+1];
	  
	  // Get Change of Basis Matrix and Warp
	  m1 = pmPathFrenet[uiSlice];
	  m2 = pmPathFrenet[uiSlice+1];
	  
	  //gilBegin(GL_TRIANGLE_STRIP);
	  {
	    for( uiCSSlice = 0; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	      {
		p1 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		p1[W] = 1.0;
		p2 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
		p2[W] = 1.0;
		
		pqev = AllocateQEVertex();
		pqev->Init(p1);
		listQEVertex.InsertLast(pqev);
		
		pqev = AllocateQEVertex();
		pqev->Init(p2);
		listQEVertex.InsertLast(pqev);
	      }
	  }
	  if( bTriangles )
	    m_pQEMesh->MergeTriangleStrip(&listQEVertex);
	  else
	    m_pQEMesh->MergeQuadStrip(&listQEVertex);
	  //gilEnd();
	}

      // End Caps
      if( m_dyniEndCapFirst.GetValue() && (m_pQEMeshEndCapFirst != NULL) ) 
	{
	  pQEMesh = m_pQEMeshEndCapFirst->Copy();
	  m_pQEMesh->MergeMesh(pQEMesh); // Merge and Deallocate pQEMesh
	}

      if( m_dyniEndCapLast.GetValue() && (m_pQEMeshEndCapLast != NULL) )
	{
	  pQEMesh = m_pQEMeshEndCapLast->Copy();
	  m_pQEMesh->MergeMesh(pQEMesh); // Merge and Deallocate pQEMesh
	}
    }
}

VOID CSLIDESweepMorph::CalculateLeafQEMesh()
{  
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDESweepMorph::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepMorph::CalculateEndCaps()
{
  UINT_32 uiSlices;
  Point pt1, pt2;
  Point *pPathPoint;
  Matrix *pmPathFrenet;
  Matrix m1, m2;
  BOOL bAreaOK;
  CListIterInt iter;
  CSLIDECrossSection *pCSLIDECrossSectionFirst, *pCSLIDECrossSectionLast;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  CListInt *plistCSLIDEControlPoints;

  uiSlices = m_pPath->GetSlices();
  pPathPoint = m_pPath->GetPoint();
  pmPathFrenet = m_pPath->GetFrenetMatrix();

  if( (uiSlices+1 < 3) && (!m_dyniEndCapFirst.GetValue()) && (!m_dyniEndCapLast.GetValue()))
    {
      return;
    }

  plistCSLIDEControlPoints = m_pPath->GetControlPointList();
  iter.Init(plistCSLIDEControlPoints);
  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
  pCSLIDECrossSectionFirst = pCSLIDEControlPoint->GetCrossSection();
  ASSERT( pCSLIDECrossSectionFirst != NULL );
  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekLast());
  pCSLIDECrossSectionLast = pCSLIDEControlPoint->GetCrossSection();
  ASSERT( pCSLIDECrossSectionLast != NULL );

  pt1 = pPathPoint[0];
  pt2 = pPathPoint[uiSlices];
  m1 = pmPathFrenet[0];
  m2 = pmPathFrenet[uiSlices];

  if( m_dyniEndCapFirst.GetValue() )
    {
      // First End Cap 
      bAreaOK = CalculateGeneralFace(m_pQEMeshEndCapFirst, pCSLIDECrossSectionFirst, 
				     pt1, m1, FALSE);

      if( !bAreaOK )
	{
	  // Bad cross section -> zero area
	  m_pQEMeshEndCapFirst = NULL;
	  HandleWarning("Could not calculate begincap for sweepmorph \"%s\"\n");
	}
    }

  if( m_dyniEndCapLast.GetValue() )
    {
      // Last End Cap 
      bAreaOK = CalculateGeneralFace(m_pQEMeshEndCapLast, pCSLIDECrossSectionLast, 
				     pt2, m2, TRUE);

      if( !bAreaOK )
	{
	  // Bad cross section -> zero area
	  m_pQEMeshEndCapLast = NULL;
	  HandleWarning("Could not calculate endcap for sweepmorph \"%s\"\n");
	}
    }
}

BOOL CSLIDESweepMorph::CalculateGeneralFace(QEMesh *(&pQEMesh), 
					    CSLIDECrossSection *pCSLIDECrossSection,
					    Point ptSweep, Matrix mSweep, 
					    BOOL bFlipOrientation)
{
  UINT_32 uiCSSlice, uiCrossSectionPoints, uiCrossSectionEndCapPoints;
  Point ptCentroid, ptStart, ptPrev, ptCurr;
  Point *pEndCapPoints, *pCrossSectionPoint;
  Vector vNormal;

  uiCrossSectionPoints = pCSLIDECrossSection->GetPoints();
  uiCrossSectionEndCapPoints = pCSLIDECrossSection->GetEndCapPoints();
  pCrossSectionPoint = pCSLIDECrossSection->GetPoint();

  ////////////////////////////////////////////////////////////////////////
  // Transform cross section points along path

  pEndCapPoints = new Point[uiCrossSectionPoints];

  if( bFlipOrientation )
    {
      UINT_32 index;
      for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
	{
	  index = uiCrossSectionPoints-uiCSSlice-1;
	  pEndCapPoints[index] = ptSweep + mSweep * pCrossSectionPoint[uiCSSlice];
	  pEndCapPoints[index][W] = 1.0;
	}
    }
  else
    {
      for( uiCSSlice = 0; uiCSSlice < uiCrossSectionPoints; uiCSSlice++)
	{
	  pEndCapPoints[uiCSSlice] = ptSweep + mSweep * pCrossSectionPoint[uiCSSlice];
	  pEndCapPoints[uiCSSlice][W] = 1.0;
	}
    }

  ////////////////////////////////////////////////////////////////////////
  // CalculateCentroid; assuming closed so Points-1

  ptCentroid.Set(0.0, 0.0, 0.0, 0.0);
  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionEndCapPoints; uiCSSlice++)
    {
      ptCentroid += pEndCapPoints[uiCSSlice];
    }
  ptCentroid /= uiCrossSectionEndCapPoints;
  ptCentroid[W] = 1.0;

  ////////////////////////////////////////////////////////////////////////
  // CalculateNormal
  
  vNormal.Null();
  ptStart = pEndCapPoints[0];
  ptPrev = ptStart;

  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionEndCapPoints; uiCSSlice++)
    {
      ptCurr = pEndCapPoints[uiCSSlice];
      ptCurr[W] = 1.0;
      
      vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }
  ptCurr = ptStart;
  vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

  if ( !vNormal.Normalize() )
    {
      HandleWarning("Null normal because of zero area cross section");
      printf(       "      End Caps cannot be calculated.\n");

      return FALSE;
    }

  ////////////////////////////////////////////////////////////////////////

  if ( pQEMesh != NULL )
    {
      DeallocateQEMesh(pQEMesh);
      pQEMesh = NULL;
    }

  pQEMesh = AllocateQEMesh();
  ASSERT( pQEMesh != NULL );
  pQEMesh->Init();

  CListInt listContours;
  CObjectList<QEVertex> *plistQEVertex;
  QEVertex *pvx;
  CListIterInt iter;
  Point pt;
  Matrix mGobj_xy, mGxy_obj;
  
  //////////////////////////////////////////////////
  // Calculate the Gobj_xy transform (transform pts to xy plane)
  {
    Vector vZ, vAxis, vTrans;
    Point ptOrigin;
    FLOAT fSin, fCos, fTheta;
    Matrix mTrans, mTransInv, mRot, mRotInv;
    
    ptOrigin.Set(0.0, 0.0, 0.0);
    vZ.Set(0.0, 0.0, 1.0);
    
    vTrans = ptOrigin - ptCentroid;
    mTrans.SetTranslate(vTrans);
    vTrans = -vTrans;
    mTransInv.SetTranslate(vTrans);
    
    vAxis = vNormal * vZ;
    fCos = vNormal.Dot(vZ);
    fSin = vAxis.Magnitude();
    if ( !vAxis.Normalize() )
      {
	vAxis.Set(1.0, 0.0, 0.0);
	if ( fCos < 0.0 )
	  {
	    fTheta = PI;
	  }
	else
	  {
	    fTheta = 0.0;
	  }
      }
    else
      {
	fTheta = atan2(fSin, fCos);
      }
    mRot.SetRotate(vAxis, fTheta);
    mRotInv.SetRotate(vAxis, -fTheta);	
    
    mGxy_obj = mRot * mTrans;
    mGobj_xy = mTransInv * mRotInv;
  }
  // Calculate the Gobj_xy transform
  //////////////////////////////////////////////////

  listContours.Init();

  plistQEVertex = new CObjectList<QEVertex>;
  ASSERT( plistQEVertex != NULL );
  listContours.InsertLast(plistQEVertex, 0);

  for( uiCSSlice = 0; uiCSSlice < uiCrossSectionEndCapPoints; uiCSSlice++)
    {
      pvx = AllocateQEVertex();
      ASSERT( pvx != NULL );
      pt = mGxy_obj * pEndCapPoints[uiCSSlice];
      pt[2] = 0.0;
      pvx->Init(pt);

      //printf( "%f %f %f\n", pt[X], pt[Y], pt[Z]);
	  
      plistQEVertex->InsertLast(pvx);
    }

  QEMesh::SetRandomize(FALSE);
  //QEMesh::SetAccuracyType(QEMesh::atFloat);
  QEMesh::SetAccuracyType(QEMesh::atExact);

  pQEMesh->CreateTriangulatedPolygon(&listContours);
  pQEMesh->Transform(mGobj_xy, mGxy_obj);

  delete plistQEVertex;
  delete[] pEndCapPoints;

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepMorph::CalculateCrossSection()
{
  UINT_32 uiCrossSectionPoints;

  // Update Path points
  m_uiPathPoints = m_pPath->GetSlices() + 1;
  uiCrossSectionPoints = CheckCrossSectionPointsEqual();

  if( m_pPath->GetType() == sliderwtPolyline )
    {
      CalculatePolylineCrossSection(uiCrossSectionPoints);
    }
  else if( (m_pPath->GetType() == sliderwtBSplineCurve) ||
	   (m_pPath->GetType() == sliderwtBezierCurve) )
    {
      CalculateBSplineCurveCrossSection(uiCrossSectionPoints);
    }
  else
    {
      ASSERT(FALSE);
    }
}

VOID CSLIDESweepMorph::PrintWarningCrossSectionPointsNotEqual()
{
  UINT_32 uiPoints;
  CListIterInt iter;
  CListInt *plistCSLIDEControlPoints;
  CSLIDECrossSection *pCSLIDECrossSection;
  CSLIDEControlPoint *pCSLIDEControlPoint;

  printf( "Warning: sweepmorph \"%s\" has crosssections with different slices\n", GetID());
  
  plistCSLIDEControlPoints = m_pPath->GetControlPointList();
  iter.Init(plistCSLIDEControlPoints);
  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
  
  pCSLIDECrossSection = pCSLIDEControlPoint->GetCrossSection();
  ASSERT( pCSLIDECrossSection != NULL );
  
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()) ) 
    {
      pCSLIDECrossSection = pCSLIDEControlPoint->GetCrossSection();

      if( pCSLIDECrossSection != NULL )
	{
	  uiPoints = pCSLIDECrossSection->GetPoints();

	  printf( "   crosssection \"%s\" has %ld points\n", 
		  pCSLIDECrossSection->GetID(), uiPoints );
	}
    }
}

UINT_32 CSLIDESweepMorph::CheckCrossSectionPointsEqual()
{
  UINT_32 uiCrossSectionPoints, uiPoints;
  CListIterInt iter;
  CListInt *plistCSLIDEControlPoints;
  CSLIDECrossSection *pCSLIDECrossSection;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  BOOL bSlicesNotEqual;

  bSlicesNotEqual = FALSE;

  plistCSLIDEControlPoints = m_pPath->GetControlPointList();
  iter.Init(plistCSLIDEControlPoints);
  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
  
  pCSLIDECrossSection = pCSLIDEControlPoint->GetCrossSection();
  ASSERT( pCSLIDECrossSection != NULL );
  uiCrossSectionPoints = pCSLIDECrossSection->GetPoints();
  
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()) ) 
    {
      pCSLIDECrossSection = pCSLIDEControlPoint->GetCrossSection();

      if( pCSLIDECrossSection != NULL )
	{
	  uiPoints = pCSLIDECrossSection->GetPoints();

	  if( uiPoints != uiCrossSectionPoints )
	    {
	      bSlicesNotEqual = TRUE;

	      if( uiPoints < uiCrossSectionPoints )
		{
		  uiCrossSectionPoints = uiPoints;
		}
	    }
	}
    }

  if( bSlicesNotEqual )
    {
      PrintWarningCrossSectionPointsNotEqual();
    }

  return uiCrossSectionPoints;
}

VOID CSLIDESweepMorph::CalculatePolylineCrossSection(UINT_32 uiCrossSectionPoints)
{
  UINT_32 uiPathPoint, uiCSPoint, uiLength, uiCurr;
  FLOAT fInterpolate;
  CListIterInt iter;
  CListInt *plistCSLIDEControlPoints;
  Point *pCrossSectionPoint1, *pCrossSectionPoint2;
  CSLIDECrossSection *pCSLIDECrossSection1, *pCSLIDECrossSection2;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  Point pt, pt1, pt2;

  plistCSLIDEControlPoints = m_pPath->GetControlPointList();

  //////////////////////////////////////////////////////////////////
  // Get the first and second crosssection
  iter.Init(plistCSLIDEControlPoints);
  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
  
  pCSLIDECrossSection1 = pCSLIDEControlPoint->GetCrossSection();
  ASSERT( pCSLIDECrossSection1 != NULL );
  
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()), 
	  uiLength=1, uiCurr=1;
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()), 
	  uiLength++ )
    {
      pCSLIDECrossSection2 = pCSLIDEControlPoint->GetCrossSection();
      
      if( pCSLIDECrossSection2 != NULL )
	{
	  break;
	}	   
    }
  ASSERT( (pCSLIDEControlPoint != NULL) && (pCSLIDECrossSection2 != NULL) );
  
  pCrossSectionPoint1 = pCSLIDECrossSection1->GetPoint();
  pCrossSectionPoint2 = pCSLIDECrossSection2->GetPoint();
  //////////////////////////////////////////////////////////////////

  if( m_uiCrossSectionPoints != uiCrossSectionPoints )
    {
      m_uiCrossSectionPoints = uiCrossSectionPoints;
      ReallocateCrossSectionPoints();
    }

  // First Cross Section
  for( uiCSPoint = 0; uiCSPoint < m_uiCrossSectionPoints; uiCSPoint++ )
    {
      pt.Set( pCrossSectionPoint1[uiCSPoint].Get() );
      m_ppCrossSectionPoint[uiCSPoint][0] = pt;      
    }

  for( uiPathPoint = 1; uiPathPoint < m_uiPathPoints; uiPathPoint++)
    {
      ////////////////////////////////////////////////////////////////////
      // Calculate uiSlice+1 points
      if( uiCurr > uiLength )
	{
	  // Find the next Cross Section
	  pCrossSectionPoint1 = pCrossSectionPoint2;
	  
	  for ( pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()), 
		  uiLength=1;
		pCSLIDEControlPoint != NULL;
		pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()), 
		  uiLength++ )
	    {
	      pCSLIDECrossSection2 = pCSLIDEControlPoint->GetCrossSection();
	      
	      if( pCSLIDECrossSection2 != NULL )
		{
		  break;
		}	   
	    }
	  ASSERT( pCSLIDECrossSection2 != NULL );
	  uiCurr = 1;
	  
	  pCrossSectionPoint2 = pCSLIDECrossSection2->GetPoint();
	}
      ///////////////////////////////////////////////////////////////////////

      // Interpolate
      fInterpolate = (FLOAT)uiCurr/uiLength;
      uiCurr++;

      if( (1.0 - EPSILON <= fInterpolate) && (fInterpolate <= 1.0 + EPSILON) )
	{
	  // Get next user defined cross section
	  for( uiCSPoint = 0; uiCSPoint < m_uiCrossSectionPoints; uiCSPoint++ )
	    {
	      m_ppCrossSectionPoint[uiCSPoint][uiPathPoint]=pCrossSectionPoint2[uiCSPoint];
	    }
	}
      else
	{
	  // No user defined cross section -> calculate this one
	  for( uiCSPoint = 0; uiCSPoint < m_uiCrossSectionPoints; uiCSPoint++ )
	    {
	      pt1.Set( pCrossSectionPoint1[uiCSPoint].Get() );
	      pt2.Set( pCrossSectionPoint2[uiCSPoint].Get() );
	      pt = (1.0-fInterpolate)*pt1 + fInterpolate*pt2;
	      pt[W] = 1.0;
	      m_ppCrossSectionPoint[uiCSPoint][uiPathPoint] = pt;      
	    }
	}
    }
}

VOID CSLIDESweepMorph::CalculateBSplineCurveCrossSection(UINT_32 uiCrossSectionPoints)
{
  UINT_32 uiPathPoint, uiCSPoint, uiControlPoints;
  FLOAT fTu, fUSlice, fU1, fU2, fTuDelta, fUCurveDelta, fUControlCurveDelta;
  FLOAT fInterpolate;
  CListIterInt iter;
  CListInt *plistCSLIDEControlPoints;
  Point *pCrossSectionPoint1, *pCrossSectionPoint2;
  CSLIDECrossSection *pCSLIDECrossSection1, *pCSLIDECrossSection2;
  CSLIDEControlPoint *pCSLIDEControlPoint;
  Point pt, pt1, pt2;

  uiControlPoints = m_pPath->GetControlPoints();
  plistCSLIDEControlPoints = m_pPath->GetControlPointList();

  fTu = m_pPath->GetTu();
  fTuDelta = m_pPath->GetTuDelta();
  fUCurveDelta = (fTuDelta - EPSILON) / (FLOAT)(m_uiPathPoints-1);
  fUControlCurveDelta = fTuDelta / (FLOAT)(uiControlPoints-1);

  //////////////////////////////////////////////////////////////////
  // Get the first and second crosssection
  iter.Init(plistCSLIDEControlPoints);
  pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekFirst());
  
  pCSLIDECrossSection1 = pCSLIDEControlPoint->GetCrossSection();
  ASSERT( pCSLIDECrossSection1 != NULL );
  
  fUSlice = fTu;
  fU1 = fTu;
  fU2 = fTu;
  for ( pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext());
	pCSLIDEControlPoint != NULL;
	pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()) )
    {
      fU2 += fUControlCurveDelta;
      pCSLIDECrossSection2 = pCSLIDEControlPoint->GetCrossSection();
      
      if( pCSLIDECrossSection2 != NULL )
	{
	  break;
	}	   
    }
  ASSERT( (pCSLIDEControlPoint != NULL) && (pCSLIDECrossSection2 != NULL) );
  
  pCrossSectionPoint1 = pCSLIDECrossSection1->GetPoint();
  pCrossSectionPoint2 = pCSLIDECrossSection2->GetPoint();
  //////////////////////////////////////////////////////////////////

  if( (m_uiCrossSectionPoints != uiCrossSectionPoints) ||
      (m_uiPathPoints != m_uiPathPointsPrev) )
    {
      m_uiCrossSectionPoints = uiCrossSectionPoints;
      m_uiPathPointsPrev = m_uiPathPoints;
      ReallocateCrossSectionPoints();
    }

  // First Cross Section
  for( uiCSPoint = 0; uiCSPoint < m_uiCrossSectionPoints; uiCSPoint++ )
    {
      pt.Set( pCrossSectionPoint1[uiCSPoint].Get() );
      m_ppCrossSectionPoint[uiCSPoint][0] = pt;      
    }
  fUSlice += fUCurveDelta;

  for( uiPathPoint = 1; uiPathPoint < m_uiPathPoints; uiPathPoint++)
    {
      ////////////////////////////////////////////////////////////////////
      // Calculate uiSlice+1 points
      if( fUSlice > fU2 )
	{
	  // Find the next Cross Section
	  fU1 = fU2;
	  pCrossSectionPoint1 = pCrossSectionPoint2;
	  
	  for ( pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext());
		pCSLIDEControlPoint != NULL;
		pCSLIDEControlPoint = (CSLIDEControlPoint *)(iter.PeekNext()) )
	    {
	      fU2 += fUControlCurveDelta;
	      pCSLIDECrossSection2 = pCSLIDEControlPoint->GetCrossSection();
	      
	      if( pCSLIDECrossSection2 != NULL )
		{
		  break;
		}	   
	    }
	  ASSERT( pCSLIDECrossSection2 != NULL );
	  
	  pCrossSectionPoint2 = pCSLIDECrossSection2->GetPoint();
	}
      ///////////////////////////////////////////////////////////////////////

      // Interpolate
      fInterpolate = (FLOAT)(fUSlice - fU1)/(fU2 - fU1);
      fUSlice += fUCurveDelta;

      if( (1.0 - EPSILON <= fInterpolate) && (fInterpolate <= 1.0 + EPSILON) )
	{
	  // Get next user defined cross section
	  for( uiCSPoint = 0; uiCSPoint < m_uiCrossSectionPoints; uiCSPoint++ )
	    {
	      m_ppCrossSectionPoint[uiCSPoint][uiPathPoint]=pCrossSectionPoint2[uiCSPoint];
	    }
	}
      else
	{
	  // No user defined cross section -> calculate this one
	  for( uiCSPoint = 0; uiCSPoint < m_uiCrossSectionPoints; uiCSPoint++ )
	    {
	      pt1.Set( pCrossSectionPoint1[uiCSPoint].Get() );
	      pt2.Set( pCrossSectionPoint2[uiCSPoint].Get() );
	      pt = (1.0-fInterpolate)*pt1 + fInterpolate*pt2;
	      pt[W] = 1.0;
	      m_ppCrossSectionPoint[uiCSPoint][uiPathPoint] = pt;      
	    }
	}
    }
}

/////////////////////////////////////////////////////////////////////////////////////////

VOID CSLIDESweepMorph::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{  
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderGeometrySmooth();
    }
  
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice;
      CListIterInt iter;
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Matrix m1, m2;
      Point pt, pt1, pt2, p1, p2;
      Vector v, v1, v2;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvV[2][2];
      Vector pvU[2];
      BOOL bTriangles;

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      // Render
      pvVDelta[0] = -(m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiSlices;
      pvVDelta[1] = -(m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiSlices;
      ppvV[0][0] = m_pvTexture[0];
      ppvV[0][1] = m_pvTexture[1];
      ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
      ppvV[1][1] = ppvV[0][1] + pvVDelta[1];

      for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	{
	  // Path
	  pt1 = pPathPoint[uiSlice];
	  pt2 = pPathPoint[uiSlice+1];
	  
	  // Get Change of Basis Matrix and Warp
	  m1 = pmPathFrenet[uiSlice];
	  m2 = pmPathFrenet[uiSlice+1];
	  
	  pvUDelta[0] = -(ppvV[0][1] - ppvV[0][0]) / (FLOAT)(m_uiCrossSectionPoints-1);
	  pvUDelta[1] = -(ppvV[1][1] - ppvV[1][0]) / (FLOAT)(m_uiCrossSectionPoints-1);
	  pvU[0] = ppvV[0][0];
	  pvU[1] = ppvV[1][0];
	  
	  if( bTriangles )
	    gilBegin(GL_TRIANGLE_STRIP);
	  else
	    gilBegin(GL_QUAD_STRIP);
	  {
	    for( uiCSSlice = 0; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	      {
		p1 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		p1[W] = 1.0;
		p2 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
		p2[W] = 1.0;
		pt = m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		v1.Set( pt.Get() );
		pt = m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
		v2.Set( pt.Get() );
		v1.Normalize();
		v2.Normalize();
		
		gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
		gilNormal3fv( v1.Get() );
		gilVertex3fv( p1.Get() );
		
		gilTexCoord2f(pvU[1][0], pvU[1][1]);	
		gilNormal3fv( v2.Get() );
		gilVertex3fv( p2.Get() );
		
		pvU[0] += pvUDelta[0];
		pvU[1] += pvUDelta[1];
	      }
	  }
	  gilEnd();
	  
	  ppvV[0][0] += pvVDelta[0];
	  ppvV[0][1] += pvVDelta[1];
	  ppvV[1][0] += pvVDelta[0];
	  ppvV[1][1] += pvVDelta[1];
	}

      if( m_dyniEndCapFirst.GetValue() && (m_pQEMeshEndCapFirst != NULL) )
	{
	  ::RenderGeometry(m_pQEMeshEndCapFirst);
	}

      if( m_dyniEndCapFirst.GetValue() && (m_pQEMeshEndCapLast != NULL) )
	{
	  ::RenderGeometry(m_pQEMeshEndCapLast);
	}
    }
}

VOID CSLIDESweepMorph::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderGeometryFlat();
    }
  
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice;
      CListIterInt iter;
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Matrix m1, m2;
      Point pt1, pt2, p1, p2, p3, p4;
      Vector v, v1, v2, v3, v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvV[2][2];
      Vector pvU[2];
      BOOL bTriangles;

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      // Render
      pvVDelta[0] = -(m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiSlices;
      pvVDelta[1] = -(m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiSlices;
      ppvV[0][0] = m_pvTexture[0];
      ppvV[0][1] = m_pvTexture[1];
      ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
      ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
	  
      for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	{
	  // Path
	  pt1 = pPathPoint[uiSlice];
	  pt2 = pPathPoint[uiSlice+1];
	      
	  // Get Change of Basis Matrix and Warp
	  m1 = pmPathFrenet[uiSlice];
	  m2 = pmPathFrenet[uiSlice+1];
	      
	  pvUDelta[0] = -(ppvV[0][1] - ppvV[0][0]) / (FLOAT)(m_uiCrossSectionPoints-1);
	  pvUDelta[1] = -(ppvV[1][1] - ppvV[1][0]) / (FLOAT)(m_uiCrossSectionPoints-1);
	  pvU[0] = ppvV[0][0];
	  pvU[1] = ppvV[1][0];

	  if( bTriangles )
	    gilBegin(GL_TRIANGLE_STRIP);
	  else
	    gilBegin(GL_QUAD_STRIP);	      
	  {
	    uiCSSlice = 0;
	    p1 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
	    p1[W] = 1.0;
	    p2 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
	    p2[W] = 1.0;
	    p3 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice+1][uiSlice];
	    p3[W] = 1.0;

	    gilNormal3fv( ((p2-p1)*(p3-p1)).Get() );
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);	
	    gilVertex3fv( p1.Get() );
	    gilNormal3fv( ((p4-p2)*(p3-p2)).Get() );
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);	
	    gilVertex3fv( p2.Get() );
	    
	    uiCSSlice++;
	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	    for( ; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	      {
		p3 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		p3[W] = 1.0;
		p4 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
		p4[W] = 1.0;
		
		gilNormal3fv( ((p2-p1)*(p3-p1)).Get() );
		gilTexCoord2f(pvU[0][0], pvU[0][1]);	
		gilVertex3fv( p3.Get() );

		gilNormal3fv( ((p4-p2)*(p3-p2)).Get() );
		gilTexCoord2f(pvU[1][0], pvU[1][1]);	
		gilVertex3fv( p4.Get() );
		
		p1 = p3;
		p2 = p4;
		pvU[0] += pvUDelta[0];
		pvU[1] += pvUDelta[1];
	      }
	  }
	  glEnd();

	  ppvV[0][0] += pvVDelta[0];
	  ppvV[0][1] += pvVDelta[1];
	  ppvV[1][0] += pvVDelta[0];
	  ppvV[1][1] += pvVDelta[1];
	}

      if( m_dyniEndCapFirst.GetValue() && (m_pQEMeshEndCapFirst != NULL) )
	{
	  ::RenderGeometry(m_pQEMeshEndCapFirst);	  
	} 

      if( m_dyniEndCapLast.GetValue() && (m_pQEMeshEndCapLast != NULL) )
	{
	  ::RenderGeometry(m_pQEMeshEndCapLast);
	}
    }
}

VOID CSLIDESweepMorph::RenderNormalsSmooth()
{
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderNormalsSmooth();
    }

  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice;
      CListIterInt iter;
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Matrix m1;
      Vector v1;
      Point pt, pt1, p1;

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      gilBegin(GL_LINES);
      {
	for( uiSlice = 0; uiSlice < uiSlices+1; uiSlice++)
	  {
	    pt1 = pPathPoint[uiSlice];
	    m1 = pmPathFrenet[uiSlice];
	    
	    for( uiCSSlice = 0; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	      {
		p1 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		p1[W] = 1.0;
		pt = m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		v1.Set( pt.Get() );
		v1.Normalize();
		
		gilVertex3fv( p1.Get() );
		pt = p1 + v1 * m_fNormalLength;
		gilVertex3fv( pt.Get() );
	      }
	  }
      }
      gilEnd();
    }
  //RenderEndCapNormals();
}

VOID CSLIDESweepMorph::RenderNormalsFlat()
{
  if( m_dyniDrawPath.GetValue() != 0 )
    {
      m_pPath->RenderNormalsFlat();
    }

  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice;
      CListIterInt iter;
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Matrix m1, m2;
      Vector v, v1, v2;
      Point pt, pt1, pt2, p1, p2, p3, p4;
      BOOL bTriangles;

      bTriangles = ( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES );

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      gilBegin(GL_LINES);
      {
	for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	  {
	    pt1 = pPathPoint[uiSlice];
	    pt2 = pPathPoint[uiSlice+1];
	    
	    m1 = pmPathFrenet[uiSlice];
	    m2 = pmPathFrenet[uiSlice+1];
		
	    p1 = pt1 + m1 * m_ppCrossSectionPoint[0][uiSlice];
	    p1[W] = 1.0;
	    p2 = pt2 + m2 * m_ppCrossSectionPoint[0][uiSlice+1];
	    p2[W] = 1.0;
	    
	    for( uiCSSlice = 1; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	      {
		p3 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		p3[W] = 1.0;
		p4 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
		p4[W] = 1.0;

		if( bTriangles )
		  {
		    pt = (p1+p2+p3)/3.0;
		    gilVertex3fv( pt.Get() );
		    v1 = (p2-p1)*(p3-p1);
		    v1.Normalize();
		    pt = pt + v1 * m_fNormalLength;
		    gilVertex3fv( pt.Get() );
		    
		    pt = (p2+p3+p4)/3.0;
		    gilVertex3fv( pt.Get() );
		    v1 = (p4-p2)*(p3-p2);
		    v1.Normalize();
		    pt = pt + v1 * m_fNormalLength;
		    gilVertex3fv( pt.Get() );
		  }
		else
		  {
		    pt = (p1+p2+p3+p4) / 4.0;
		    gilVertex3fv( pt.Get() );
		    v1 = (p2-p1)*(p3-p1);
		    v1.Normalize();
		    v2 = (p4-p2)*(p3-p2);
		    v2.Normalize();
		    v = (v1 + v2) / 2.0;
		    v.Normalize();
		    pt = pt + v * m_fNormalLength;
		    gilVertex3fv( pt.Get() );
		  }
		
		p1 = p3;
		p2 = p4;
	      }
	  }
      }
      gilEnd(); 

      //RenderEndCapNormals();
    }
}

VOID CSLIDESweepMorph::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtSweepMorph], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  if( m_pPath != NULL )
    {
      m_pPath->CSLIDESweepPath::WriteSLIDE(fpOutputFile);
    }

  fprintf(fpOutputFile, "  %s %s\n",
	  gcppcSLIDEReservedWords[sliderwtMeshType],
	  gcppcSLIDEMeshTypeReservedWords[m_dyniMeshType.GetValue()]);

  fprintf(fpOutputFile, "    %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtBeginCap],
	  m_dyniEndCapFirst.GetValue());

  fprintf(fpOutputFile, "    %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtEndCap],
	  m_dyniEndCapLast.GetValue());

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawPath],
	  m_dyniDrawPath.GetValue());

  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtDrawSweep],
	  m_dyniDrawSweep.GetValue());

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndSweepMorph]);
}

VOID CSLIDESweepMorph::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  if( m_dyniDrawSweep.GetValue() != 0 )
    {
      UINT_32 uiSlice, uiSlices, uiCSSlice;
      CListIterInt iter;
      Point *pPathPoint;
      Matrix *pmPathFrenet;
      Matrix m1, m2;
      Point pt, pt1, pt2, p1, p2, p3, p4;
      Vector v, v1, v2, v3, v4;
      Vector pvVDelta[2];
      Vector pvUDelta[2];
      Vector ppvV[2][2];
      Vector pvU[2];

      uiSlices = m_pPath->GetSlices();
      pPathPoint = m_pPath->GetPoint();
      pmPathFrenet = m_pPath->GetFrenetMatrix();

      pvVDelta[0] = -(m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiSlices;
      pvVDelta[1] = -(m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiSlices;
      ppvV[0][0] = m_pvTexture[0];
      ppvV[0][1] = m_pvTexture[1];
      ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
      ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
	  
      for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	{
	  pt1 = pPathPoint[uiSlice];
	  pt2 = pPathPoint[uiSlice+1];
	  
	  // Get Change of Basis Matrix and Warp
	  m1 = pmPathFrenet[uiSlice];
	  m2 = pmPathFrenet[uiSlice+1];
	      
	  pvUDelta[0] = -(ppvV[0][1] - ppvV[0][0]) / (FLOAT)(m_uiCrossSectionPoints-1);
	  pvUDelta[1] = -(ppvV[1][1] - ppvV[1][0]) / (FLOAT)(m_uiCrossSectionPoints-1);
	  pvU[0] = ppvV[0][0];
	  pvU[1] = ppvV[1][0];
	      
	  for( uiCSSlice = 0; uiCSSlice < m_uiCrossSectionPoints-1; uiCSSlice++)
	    {
	      p1 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
	      p1[W] = 1.0;
	      p2 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
	      p2[W] = 1.0;
	      p3 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice+1][uiSlice];
	      p3[W] = 1.0;
	      p4 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice+1][uiSlice+1];
	      p4[W] = 1.0;	      
	      
	      if( (eShading == slideshtSLF_GOURAUD) || (eShading == slideshtSLF_PHONG) )
		{
		  pt = m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
		  v1.Set( pt.Get() );
		  pt = m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
		  v2.Set( pt.Get() );
		  pt = m1 * m_ppCrossSectionPoint[uiCSSlice+1][uiSlice];
		  v3.Set( pt.Get() );
		  pt = m2 * m_ppCrossSectionPoint[uiCSSlice+1][uiSlice+1];
		  v4.Set( pt.Get() );

		  v1.Normalize();
		  v2.Normalize();
		  v3.Normalize();
		  v4.Normalize();
		  
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], 
				  p2[X], p2[Y], p2[Z], 
				  p3[X], p3[Y], p3[Z],
				  v1[X], v1[Y], v1[Z], 
				  v2[X], v2[Y], v2[Z], 
				  v3[X], v3[Y], v3[Z],
				  pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], 
				  (pvU[0] + pvUDelta[0])[0], (pvU[0] + pvUDelta[0])[1]);
		  
		  pvU[0] += pvUDelta[0];
		  
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"N\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], 
				  p2[X], p2[Y], p2[Z], 
				  p4[X], p4[Y], p4[Z],
				  v3[X], v3[Y], v3[Z], 
				  v2[X], v2[Y], v2[Z], 
				  v4[X], v4[Y], v4[Z],
				  pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1],
				  (pvU[1] + pvUDelta[1])[0], (pvU[1] + pvUDelta[1])[1]);
		  pvU[1] += pvUDelta[1];
		}
	      else
		{
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p1[X], p1[Y], p1[Z], 
				  p2[X], p2[Y], p2[Z], 
				  p3[X], p3[Y], p3[Z],
				  pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], 
				  (pvU[0] + pvUDelta[0])[0], (pvU[0] + pvUDelta[0])[1]);
		  pvU[0] += pvUDelta[0];
		  
		  RiArchiveRecord(RI_COMMENT, "\nPolygon \"P\" [%.6f %.6f %.6f  %.6f %.6f %.6f %.6f %.6f %.6f] \"st\" [%.6f %.6f %.6f %.6f %.6f %.6f]\n",
				  p3[X], p3[Y], p3[Z], 
				  p2[X], p2[Y], p2[Z], 
				  p4[X], p4[Y], p4[Z],
				  pvU[0][0], pvU[0][1], pvU[1][0], pvU[1][1], 
				  (pvU[1] + pvUDelta[1])[0], (pvU[1] + pvUDelta[1])[1]);
		  pvU[1] += pvUDelta[1];
		}
	    }
	  ppvV[0][0] += pvVDelta[0];
	  ppvV[0][1] += pvVDelta[1];
	  ppvV[1][0] += pvVDelta[0];
	  ppvV[1][1] += pvVDelta[1];
	}
    
      if( m_dyniEndCapFirst.GetValue() && (m_pQEMeshEndCapFirst != NULL))
	{
	  Matrix m;	  
	  ::WriteRIB(m_pQEMeshEndCapFirst, fpOutputFile, m);
	} 

      if( m_dyniEndCapLast.GetValue() && (m_pQEMeshEndCapLast != NULL) )
	{
	  Matrix m;	  
	  ::WriteRIB(m_pQEMeshEndCapLast, fpOutputFile, m);
	}
    }
}

VOID CSLIDESweepMorph::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE SweepMoph: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDESweepMorph::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE SweepMorph: %s", m_pcGUIID);
}

VOID CSLIDESweepMorph::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
					 CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: SWEEPMORPH\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);

  if( m_dyniMeshType.GetValue() == slidemtSLF_TRIANGLES )
    ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform);
  else
    ::WriteWRLZQuads(m_pQEMesh, fpOutputFile, mTransform);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDESweepMorph::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiSlice, uiSlices, uiCSSlice;
  CListIterInt iter;
  Point *pPathPoint;     
  Matrix *pmPathFrenet;
  Matrix mCS, m1, m2;
  Point pt1, pt2, p1, p2, p3, p4;

  uiSlices = m_pPath->GetSlices();
  pPathPoint = m_pPath->GetPoint();
  pmPathFrenet = m_pPath->GetFrenetMatrix();

  // EndCaps
  if( m_dyniEndCapFirst.GetValue() && (m_pQEMeshEndCapFirst != NULL) )
    {
      ::WriteSTL(m_pQEMeshEndCapFirst, fpOutputFile, mTransform);      
    }

  if( m_dyniEndCapLast.GetValue()  && (m_pQEMeshEndCapLast != NULL) )
    {
      ::WriteSTL(m_pQEMeshEndCapLast, fpOutputFile, mTransform);
    }

  if ( mTransform.Determinant() < 0 ) // Need to Mirror triangles
    {
      for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	{
	  pt1 = pPathPoint[uiSlice];
	  pt2 = pPathPoint[uiSlice+1];
	  
	  // Get Change of Basis Matrix and Warp
	  m1 = pmPathFrenet[uiSlice];
	  m2 = pmPathFrenet[uiSlice+1];
	  
	  p1 = pt1 + m1 * m_ppCrossSectionPoint[0][uiSlice];
	  p1[W] = 1.0;
	  p2 = pt2 + m2 * m_ppCrossSectionPoint[0][uiSlice+1];
	  p2[W] = 1.0;
	  
	  p1 = mTransform * p1;
	  p2 = mTransform * p2;
	  
	  for( uiCSSlice = 1; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	    {
	      p3 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
	      p3[W] = 1.0;
	      p4 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
	      p4[W] = 1.0;
	      
	      p3 = mTransform * p3;
	      p4 = mTransform * p4;
	      
	      WriteSTLTriangle(fpOutputFile, p3, p2, p1);
	      WriteSTLTriangle(fpOutputFile, p2, p3, p4);
	      
	      p1 = p3;
	      p2 = p4;
	    }
	}
    }
  else // Don't mirror triangles
    {
      for( uiSlice = 0; uiSlice < uiSlices; uiSlice++)
	{
	  pt1 = pPathPoint[uiSlice];
	  pt2 = pPathPoint[uiSlice+1];
	  
	  // Get Change of Basis Matrix and Warp
	  m1 = pmPathFrenet[uiSlice];
	  m2 = pmPathFrenet[uiSlice+1];
	  
	  p1 = pt1 + m1 * m_ppCrossSectionPoint[0][uiSlice];
	  p1[W] = 1.0;
	  p2 = pt2 + m2 * m_ppCrossSectionPoint[0][uiSlice+1];
	  p2[W] = 1.0;
	  
	  p1 = mTransform * p1;
	  p2 = mTransform * p2;
	  
	  for( uiCSSlice = 1; uiCSSlice < m_uiCrossSectionPoints; uiCSSlice++)
	    {
	      p3 = pt1 + m1 * m_ppCrossSectionPoint[uiCSSlice][uiSlice];
	      p3[W] = 1.0;
	      p4 = pt2 + m2 * m_ppCrossSectionPoint[uiCSSlice][uiSlice+1];
	      p4[W] = 1.0;
	      
	      p3 = mTransform * p3;
	      p4 = mTransform * p4;
	      
	      WriteSTLTriangle(fpOutputFile, p1, p2, p3);
	      WriteSTLTriangle(fpOutputFile, p4, p3, p2);
	      
	      p1 = p3;
	      p2 = p4;
	    }
	}
    }
}


