#include "slideterrain.h"

#include "slideworld.h"
#include "slidesurface.h"

#include "slidepoint.h"

#include "qemesh.h"
#include "qerender.h"

static const SLIDETerrainType ceType = slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL;
static const SLIDEVoronoiType ceVoronoi = slidevtSLF_VORONOI_OFF;
static const SLIDEAccuracyType ceAccuracy = slideatSLF_FLOAT;
static const SLIDEOnOffType ceRandomize = slideoftSLF_OFF;

static const SLIDEOnOffType ceDrawDelaunay = slideoftSLF_ON;
static const SLIDEOnOffType ceDrawVertices = slideoftSLF_OFF;

//////////////////////////////////////////////////////////////////////
// CSLIDETerrain Class
//

CSLIDETerrain::CSLIDETerrain()
{
  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_plistPointIDs = NULL;

  m_dyniType.SetValue(ceType);
  m_dyniVoronoi.SetValue(ceVoronoi);
  m_dyniAccuracy.SetValue(ceAccuracy);
  m_dyniRandomize.SetValue(ceRandomize);

  m_dyniDrawDelaunay.SetValue(ceDrawDelaunay);
  m_dyniDrawVertices.SetValue(ceDrawVertices);

  m_pmeVoronoi = NULL;
}

CSLIDETerrain::~CSLIDETerrain()
{
}

VOID CSLIDETerrain::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID,
			 CListInt *plistPointIDs)
{
  CSLIDEBoundedNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dyniSolid.SetValue(slidestSLF_SOLID);

  m_plistPointIDs = plistPointIDs;
  m_listCSLIDEPoint.Init();

  m_dyniType.SetValue(ceType);
  m_dyniVoronoi.SetValue(ceVoronoi);
  m_dyniAccuracy.SetValue(ceAccuracy);
  m_dyniRandomize.SetValue(ceRandomize);

  m_dyniDrawDelaunay.SetValue(ceDrawDelaunay);
  m_dyniDrawVertices.SetValue(ceDrawVertices);

  m_pmeVoronoi = NULL;
}

VOID CSLIDETerrain::Uninit()
{
  CSLIDEBoundedNode::Uninit();

  if ( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }
}

BOOL CSLIDETerrain::Initialized()
{
  if ( !CSLIDEBoundedNode::Initialized() )
    {
      return FALSE;
    }

  return (m_plistPointIDs != NULL);
}

BOOL CSLIDETerrain::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;

  if ( !CSLIDEBoundedNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Solid Type

  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  // Check Terrain Type
  m_dyniType.Update();
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slidetertSize ) )
    {
      HandleError("Illegal Terrain Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(ceType);
    }

  // Check Accuracy Type
  m_dyniAccuracy.Update();
  if ( !( 0 <= m_dyniAccuracy.GetValue() &&
	  m_dyniAccuracy.GetValue() < slideatSize ) )
    {
      HandleError("Illegal Terrain Accuracy flag value:  %ld", 
		  m_dyniAccuracy.GetValue());
      m_dyniAccuracy.SetValue(ceAccuracy);
    }

  // Check Randomize
  m_dyniRandomize.Update();
  if ( !( 0 <= m_dyniRandomize.GetValue() &&
	  m_dyniRandomize.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Random flag value:  %ld", 
		  m_dyniRandomize.GetValue());
      m_dyniRandomize.SetValue(ceRandomize);
    }

  // Check Draw Delaunay
  m_dyniDrawDelaunay.Update();
  if ( !( 0 <= m_dyniDrawDelaunay.GetValue() &&
	  m_dyniDrawDelaunay.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Delaunay flag value:  %ld", 
		  m_dyniDrawDelaunay.GetValue());
      m_dyniDrawDelaunay.SetValue(ceDrawDelaunay);
    }

  // Check Draw Vertices
  m_dyniDrawVertices.Update();
  if ( !( 0 <= m_dyniDrawVertices.GetValue() &&
	  m_dyniDrawVertices.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Vertices flag value:  %ld", 
		  m_dyniDrawVertices.GetValue());
      m_dyniDrawVertices.SetValue(ceDrawVertices);
    }

  // Check Voronoi
  m_dyniVoronoi.Update();
  if ( !( 0 <= m_dyniVoronoi.GetValue() &&
	  m_dyniVoronoi.GetValue() < slidevtSize ) )
    {
      HandleError("Illegal Terrain Voronoi flag value:  %ld", 
		  m_dyniVoronoi.GetValue());
      m_dyniVoronoi.SetValue(ceVoronoi);
    }

  // Check Point References

  iter.Init(m_plistPointIDs);

  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
      if ( pCSLIDEPoint == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint], pcID);
	  bCorrect = FALSE;
	}
      m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
    }

  return bCorrect;
}

BOOL CSLIDETerrain::Valid()
{
  if ( !CSLIDEBoundedNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniSolid.GetValue() &&
	     m_dyniSolid.GetValue() < slidestSize ) );
	   //GetPointCount() > 0 );
}

CSLIDECamera *CSLIDETerrain::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtTerrain], GetID());

  return NULL;
}

CSLIDELight *CSLIDETerrain::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtTerrain], GetID());

  return NULL;
}

CSLIDEAreaLight *CSLIDETerrain::FindAreaLight(CListIterInt *piterIDs, 
					     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtAreaLight],
	      gcppcSLIDEReservedWords[sliderwtTerrain], GetID());

  return NULL;
}

VOID CSLIDETerrain::Preprocess()
{
  CSLIDEBoundedNode::Preprocess();

  CheckDynamic();
  UpdateState();
}

VOID CSLIDETerrain::CheckDynamic()
{
  CListIterInt iter;
  //CSLIDEPoint *pCSLIDEPoint;
  //BOOL bAutoNormals;

  // Are there variables which I must call TCL to update?
  SetDynamic(m_dyniShading.GetDynamic() ||
	     m_dyniLOD.GetDynamic() ||
	     m_dyniSolid.GetDynamic() ||
	     m_dyniType.GetDynamic() ||
	     m_dyniVoronoi.GetDynamic() ||
	     m_dyniAccuracy.GetDynamic() ||
	     m_dyniRandomize.GetDynamic() ||
	     m_dyniDrawDelaunay.GetDynamic() ||
	     m_dyniDrawVertices.GetDynamic());
/*
  bAutoNormals = !(!m_dyniShading.GetDynamic() && 
		   (m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_WIRE ||
		    m_dyniShading.GetValue() == 
		    (INT_32)slideshtSLF_FLAT));
      
  iter.Init(&m_listCSLIDEPoint);
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->SetAutoNormals(bAutoNormals || pCSLIDEPoint->GetAutoNormals());
    }
*/
}

VOID CSLIDETerrain::PreprocessGraph()
{
  if ( !GetVisited() )
    {
      CListIterInt iter;
      CSLIDEPoint *pCSLIDEPoint;
      
      BOOL bGeometryEmpty;
      BOOL bGeometryDynamic;
      BOOL bNormalsEmpty;
      BOOL bNormalsDynamic;
      BOOL bBoundsEmpty;
      BOOL bBoundsDynamic;

      BOOL bBoundEmpty;
      BOOL bBoundDynamic;

      BOOL bChildrenBoundsDynamic;

      bGeometryEmpty = TRUE;
      bGeometryDynamic = FALSE;
      bNormalsEmpty = TRUE;
      bNormalsDynamic = FALSE;
      bBoundsEmpty = TRUE;
      bBoundsDynamic = FALSE;

      bBoundEmpty = TRUE;
      bBoundDynamic = FALSE;

      //////////////////////////////////////////////////
      // Look at Points

      iter.Init(&m_listCSLIDEPoint);
      
      for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
	{
	  bBoundEmpty = FALSE;
	  bBoundDynamic = bBoundDynamic || 
	                  pCSLIDEPoint->GetBoundDynamic();
	  bGeometryDynamic = bGeometryDynamic || 
	                     pCSLIDEPoint->GetGeometryDynamic();
	  bNormalsDynamic = bNormalsDynamic ||
	                    pCSLIDEPoint->GetNormalDynamic();
	}

      bGeometryEmpty = bBoundEmpty;
      bNormalsEmpty = bBoundEmpty;
      bBoundsEmpty = bBoundEmpty;
      bBoundsDynamic = bBoundDynamic;

      //////////////////////////////////////////////////
      // Look at Local flags

      // Update whether we are Empty based on our LOD flag
      if ( !m_dyniLOD.GetDynamic() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      // This node is always off
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	      bBoundsEmpty = TRUE;
	      bBoundEmpty = TRUE;
	    }
	  else if ( m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
	    {
	      // This node is always just a bounding box
	      bGeometryEmpty = TRUE;
	      bNormalsEmpty = TRUE;
	    }
	}

      // Check what is really dynamic

      bChildrenBoundsDynamic = bBoundsDynamic;

      bGeometryDynamic = ( bGeometryDynamic ||
			   m_pCSLIDESurface->GetDynamic() ||
			   m_dyniShading.GetDynamic() ||
			   m_dyniLOD.GetDynamic() ||
			   m_dyniSolid.GetDynamic() );
      bNormalsDynamic = ( bNormalsDynamic ||
			  m_dyniShading.GetDynamic() ||
			  m_dyniLOD.GetDynamic() );
      bBoundsDynamic = ( bBoundsDynamic || 
			 m_dyniLOD.GetDynamic() );
      bBoundDynamic = ( bBoundDynamic || 
			m_dyniLOD.GetDynamic() );

      // If anything is now Empty then it will always be empty so
      // it is not dynamic

      bChildrenBoundsDynamic = ( bChildrenBoundsDynamic && !bBoundsEmpty );

      bGeometryDynamic = ( bGeometryDynamic && !bGeometryEmpty );
      bNormalsDynamic = ( bNormalsDynamic && !bNormalsEmpty );
      bBoundsDynamic = ( bBoundsDynamic && !bBoundsEmpty );
      bBoundDynamic = ( bBoundDynamic && !bBoundEmpty );
      
      //////////////////////////////////////////////////
      // For the OpenGL Display Lists to be really useful 
      // it is all or nothing so we enforce that here

      bGeometryDynamic = ( bGeometryDynamic ||
			   bNormalsDynamic ||
			   bBoundsDynamic );
      bNormalsDynamic = bGeometryDynamic;
      bBoundsDynamic = bGeometryDynamic;

      //////////////////////////////////////////////////
      // Now save these correct flag values

      SetChildrenBoundsDynamic(bChildrenBoundsDynamic);

      SetGeometryEmpty(bGeometryEmpty);
      SetGeometryDynamic(bGeometryDynamic);
      SetNormalsEmpty(bNormalsEmpty);
      SetNormalsDynamic(bNormalsDynamic);
      SetBoundsEmpty(bBoundsEmpty);
      SetBoundsDynamic(bBoundsDynamic);
      SetBoundEmpty(bBoundEmpty);
      SetBoundDynamic(bBoundDynamic);
      
      //////////////////////////////////////////////////
      // Calculate the Bounding Box (if Empty this will do nothing)
      if ( GetBoundEmpty() )
	{
	  SetBoundValid(TRUE);
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);

      //////////////////////////////////////////////////
      // Done Preproces(FLOAT)sing this Node in the Graph
      SetVisited(TRUE);
    }
}

VOID CSLIDETerrain::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() ||
       GetChildrenBoundsDynamic() )
    {
      if ( GetChildrenBoundsDynamic() )
	{
	  CSLIDEBoundedNode::InvalidateQEMesh();
	}

      UpdateState();
    }
  
  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }

  SetVisited(FALSE);
}

VOID CSLIDETerrain::UpdateState()
{
  UINT_32 uiType, uiVoronoi, uiAccuracy, uiRandomize;

  CSLIDEBoundedNode::UpdateState();

  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", 
		  m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }

  uiType = m_dyniType.GetValue();
  m_dyniType.Update();
  if ( !( 0 <= m_dyniType.GetValue() &&
	  m_dyniType.GetValue() < slidetertSize ) )
    {
      HandleError("Illegal Terrain Type flag value:  %ld", 
		  m_dyniType.GetValue());
      m_dyniType.SetValue(ceType);
    }

  uiVoronoi = m_dyniVoronoi.GetValue();
  m_dyniVoronoi.Update();
  if ( !( 0 <= m_dyniVoronoi.GetValue() &&
	  m_dyniVoronoi.GetValue() < slidevtSize ) )
    {
      HandleError("Illegal Terrain Voronoi flag value:  %ld", 
		  m_dyniVoronoi.GetValue());
      m_dyniVoronoi.SetValue(ceVoronoi);
    }

  uiAccuracy = m_dyniAccuracy.GetValue();
  m_dyniAccuracy.Update();
  if ( !( 0 <= m_dyniAccuracy.GetValue() &&
	  m_dyniAccuracy.GetValue() < slideatSize ) )
    {
      HandleError("Illegal Terrain Accuracy flag value:  %ld", 
		  m_dyniAccuracy.GetValue());
      m_dyniAccuracy.SetValue(ceAccuracy);
    }

  uiRandomize = m_dyniRandomize.GetValue();
  m_dyniRandomize.Update();
  if ( !( 0 <= m_dyniRandomize.GetValue() &&
	  m_dyniRandomize.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Random flag value:  %ld", 
		  m_dyniRandomize.GetValue());
      m_dyniRandomize.SetValue(ceRandomize);
    }

  m_dyniDrawDelaunay.Update();
  if ( !( 0 <= m_dyniDrawDelaunay.GetValue() &&
	  m_dyniDrawDelaunay.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Delaunay flag value:  %ld", 
		  m_dyniDrawDelaunay.GetValue());
      m_dyniDrawDelaunay.SetValue(ceDrawDelaunay);
    }

  m_dyniDrawVertices.Update();
  if ( !( 0 <= m_dyniDrawVertices.GetValue() &&
	  m_dyniDrawVertices.GetValue() < slideoftSize ) )
    {
      HandleError("Illegal Terrain Draw Vertices flag value:  %ld", 
		  m_dyniDrawVertices.GetValue());
      m_dyniDrawVertices.SetValue(ceDrawVertices);
    }

  if ( uiType != (UINT_32)m_dyniType.GetValue() || 
       uiAccuracy != (UINT_32)m_dyniAccuracy.GetValue() || 
       uiRandomize != (UINT_32)m_dyniRandomize.GetValue() ||
       m_pQEMesh == NULL )
    {
      if ( m_pQEMesh != NULL )
	{
	  DeallocateQEMesh(m_pQEMesh);
	  m_pQEMesh = NULL;
	}
      if ( m_pmeVoronoi != NULL )
	{
	  DeallocateQEMesh(m_pmeVoronoi);
	  m_pmeVoronoi = NULL;
	}

      m_pQEMesh = AllocateQEMesh();
      ASSERT( m_pQEMesh != NULL );
      m_pQEMesh->Init();

      CObjectList<QEVertex> listQEVertex;
      QEVertex *pvx;
      CListIterInt iter;
      CSLIDEPoint *pCSLIDEPoint;

      iter.Init(&m_listCSLIDEPoint);
      for ( pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekFirst());
	    pCSLIDEPoint != NULL;
	    pCSLIDEPoint = (CSLIDEPoint*)(iter.PeekNext()) )
	{
	  pvx = AllocateQEVertex();
	  ASSERT( pvx != NULL );
	  pvx->Init(pCSLIDEPoint->GetPoint());

	  listQEVertex.InsertLast(pvx);
	}

      if ( m_dyniRandomize.GetValue() )
	{
	  QEMesh::SetRandomize(TRUE);
	}
      else
	{
	  QEMesh::SetRandomize(FALSE);
	}

      switch ( m_dyniAccuracy.GetValue() )
	{
	case slideatSLF_FLOAT_NAIVE:
	  QEMesh::SetAccuracyType(QEMesh::atFloatNaive);
	  break;
	case slideatSLF_FLOAT:
	  QEMesh::SetAccuracyType(QEMesh::atFloat);
	  break;
	case slideatSLF_EXACT:
	  QEMesh::SetAccuracyType(QEMesh::atExact);
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      switch ( m_dyniType.GetValue() )
	{
	case slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL: 
	  m_pQEMesh->CreateTerrain(QEMesh::dtIncremental, &listQEVertex);
	  break;
	case slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL_SLOW: 
	  m_pQEMesh->CreateTerrain(QEMesh::dtIncrementalSlow, &listQEVertex);
	  break;
	case slidetertSLF_TERRAIN_DELAUNAY_INCREMENTAL_NO_BOUND: 
	  m_pQEMesh->CreateTerrain(QEMesh::dtIncrementalNoBound, &listQEVertex);
	  break;
	case slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER: 
	  m_pQEMesh->CreateTerrain(QEMesh::dtDivideAndConquer, &listQEVertex);
	  break;
	case slidetertSLF_TERRAIN_DELAUNAY_DIVIDE_AND_CONQUER_ALTERNATING: 
	  m_pQEMesh->CreateTerrain(QEMesh::dtDivideAndConquerAlternating, &listQEVertex);
	  break;
	case slidetertSLF_TERRAIN_DELAUNAY_CONVEX_HULL: 
	  m_pQEMesh->CreateTerrain(QEMesh::dtConvexHull, &listQEVertex);
	  break;
	default:
	  ASSERT(FALSE);
	  break;
	} 

      for ( pvx = listQEVertex.RemoveFirst();
	    pvx != NULL;
	    pvx = listQEVertex.RemoveFirst() )
	{
	  DeallocateQEVertex(pvx);
	}
    }

  //////////////////////////////////////////////////
  // Voronoi Diagram
  //
  if ( uiVoronoi != (UINT_32)m_dyniVoronoi.GetValue() )
    {
      if ( m_pmeVoronoi != NULL )
	{
	  DeallocateQEMesh(m_pmeVoronoi);
	  m_pmeVoronoi = NULL;
	}
    }
  if ( (m_dyniVoronoi.GetValue() != slidevtSLF_VORONOI_OFF) &&
       (m_pmeVoronoi == NULL) )
    {
      if ( m_dyniRandomize.GetValue() )
	{
	  QEMesh::SetRandomize(TRUE);
	}
      else
	{
	  QEMesh::SetRandomize(FALSE);
	}

      switch ( m_dyniAccuracy.GetValue() )
	{
	case slideatSLF_FLOAT_NAIVE:
	  QEMesh::SetAccuracyType(QEMesh::atFloatNaive);
	  break;
	case slideatSLF_FLOAT:
	  QEMesh::SetAccuracyType(QEMesh::atFloat);
	  break;
	case slideatSLF_EXACT:
	  QEMesh::SetAccuracyType(QEMesh::atExact);
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}

      switch ( m_dyniVoronoi.GetValue() )
	{
	case slidevtSLF_VORONOI_DELAUNAY: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  m_pmeVoronoi->CreateVoronoi(QEMesh::vtDelaunay, m_pQEMesh);
	  break;
	case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  m_pmeVoronoi->CreateVoronoi(QEMesh::vtDivideAndConquer, m_pQEMesh);
	  break;
	case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
	  m_pmeVoronoi = AllocateQEMesh();
	  ASSERT( m_pmeVoronoi != NULL );
	  m_pmeVoronoi->Init();

	  m_pmeVoronoi->CreateVoronoi(QEMesh::vtDivideAndConquerAlternating, m_pQEMesh);
	  break;
	case slidevtSLF_VORONOI_CONES: 
	case slidevtSLF_VORONOI_PLANES: 
	  break;
	default:
	  ASSERT(FALSE);
	  break;
	} 
    }
}

VOID CSLIDETerrain::UpdateGraph()
{
  if ( !GetVisited() )
    {
      if ( !GetBoundValid() )
	{
	  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	    {
	      if ( !m_bound.GetEmpty() )
		{
		  SetBoundChanged(TRUE);
		  m_bound.Null();
		}
	      return;
	    }
	  
	  CalculateBound();
	  SetBoundValid(TRUE);
	}

      SetVisited(TRUE);
    }
}

VOID CSLIDETerrain::CalculateBound()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  CBound bound;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->InsertIntoBound(&bound);
    }

  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDETerrain::CalculateCDAVertices(CListInt *plistCDAVertex, 
					const Matrix &mGroot_obj)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->CalculateCDAVertices(plistCDAVertex, mGroot_obj);
    }
}
#endif // __USE_CDA__

VOID CSLIDETerrain::PrerenderGeometry(SLIDEShadingType eShading)
{
  SLIDEShadingType eShade;
  
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  // This flag tells us whether to use Point or Point Normals
  ModifyShadingFlag(eShading);

  // This one set the shading attributes which in many cases will 
  // be SLF_INHERIT and so it will not set anything and leave up
  // to the caller above
  eShade = (SLIDEShadingType)m_dyniShading.GetValue();
  
  GLbitfield GLbAttributesMask;
  
  GLbAttributesMask = CalculateAttributesMask(eShade);
  
  PushAttributes(GLbAttributesMask);
  {
    PushShadingAttributes(eShade);
    {
      m_pCSLIDESurface->Push();
      {
	PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
	{
	  m_uiPrerenderAttributesStackDepthMax = 0;

	  if ( m_dyniDrawDelaunay.GetValue() )
	    {
	      ::RenderGeometry(m_pQEMesh);
	    }

	  PushAuxillaryGeometryAttributes();
	  {
	    switch ( m_dyniVoronoi.GetValue() )
	      {
	      case slidevtSLF_VORONOI_OFF: 
		break;
	      case slidevtSLF_VORONOI_DELAUNAY:
		::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtDelaunay);
		break;
	      case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
		::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtDivideAndConquer);
		break;
	      case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
		::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtDivideAndConquerAlternating);
		break;
	      case slidevtSLF_VORONOI_CONES: 
		::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtCones);
		break;
	      case slidevtSLF_VORONOI_PLANES: 
		::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtPlanes);
		break;
	      default:
		ASSERT(FALSE);
		break;
	      } 

	    if ( m_dyniDrawVertices.GetValue() )
	      {
		gilPointSize(3.0);
		gilColor3f(1.0, 1.0, 0.0);
		gilBegin(GL_POINTS);
		{
		  QEMeshIter iterQEMesh;
		  QEVertex *pvx;
		  Point pt;
		  
		  iterQEMesh.Init(m_pQEMesh);
		  for ( pvx = iterQEMesh.PeekFirstVertex();
			pvx != NULL;
			pvx = iterQEMesh.PeekNextVertex() )
		    {
		      pt = pvx->GetPoint();
		      gilVertex3fv(pt.Get());	
		    }
		}
		gilEnd();
	      }
	  }
	  PopAuxillaryGeometryAttributes();
	  
	  if ( GLbAttributesMask != 0x0 )
	    {
	      m_uiPrerenderAttributesStackDepthMax++;
	    }
	}
	PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
      }
      m_pCSLIDESurface->Pop();
    }
    PopShadingAttributes(eShade);
  }
  PopAttributes(GLbAttributesMask);
}

VOID CSLIDETerrain::PrerenderNormals(SLIDEShadingType eShading)
{
  if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF ||
       m_dyniLOD.GetValue() == slidelodtSLF_BOUND )
    {
      return;
    }

  ModifyShadingFlag(eShading);

  if ( m_dyniDrawDelaunay.GetValue() )
    {
      ::RenderNormals(m_pQEMesh);
    }
}

VOID CSLIDETerrain::RenderGeometry(const Matrix &/*mGproj_obj*/, 
				   CBound * /*pCBoundClip*/, 
				   SLIDEShadingType eShading,
				   SLIDELODType /*eLOD*/,
				   CSLIDEWormwarp * /*pCSLIDEWarp *//*= NULL*/)
{
  PushSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
  {
    if ( m_dyniDrawDelaunay.GetValue() )
      {
	::RenderGeometry(m_pQEMesh);
      }

    PushAuxillaryGeometryAttributes();
    {
      switch ( m_dyniVoronoi.GetValue() )
	{
	case slidevtSLF_VORONOI_OFF: 
	  break;
	case slidevtSLF_VORONOI_DELAUNAY:
	  ::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtDelaunay);
	  break;
	case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER: 
	  ::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtDivideAndConquer);
	  break;
	case slidevtSLF_VORONOI_DIVIDE_AND_CONQUER_ALTERNATING: 
	  ::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtDivideAndConquerAlternating);
	  break;
	case slidevtSLF_VORONOI_CONES: 
	  ::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtCones);
	  break;
	case slidevtSLF_VORONOI_PLANES: 
	  ::RenderVoronoi(m_pQEMesh, m_pmeVoronoi, QE::vtPlanes);
	  break;
	default:
	  ASSERT(FALSE);
	  break;
	} 

      if ( m_dyniDrawVertices.GetValue() )
	{
	  gilPointSize(3.0);
	  gilColor3f(1.0, 1.0, 0.0);
	  gilBegin(GL_POINTS);
	  {
	    QEMeshIter iterQEMesh;
	    QEVertex *pvx;
	    Point pt;
	    
	    iterQEMesh.Init(m_pQEMesh);
	    for ( pvx = iterQEMesh.PeekFirstVertex();
		  pvx != NULL;
		  pvx = iterQEMesh.PeekNextVertex() )
	      {
		pt = pvx->GetPoint();
		gilVertex3fv(pt.Get());	
	      }
	  }
	  gilEnd();
	}
    }
    PopAuxillaryGeometryAttributes();
  }
  PopSolidAttributes((SLIDESolidType)m_dyniSolid.GetValue());
}

VOID CSLIDETerrain::RenderNormals(SLIDEShadingType eShading)
{
  PushAuxillaryGeometryAttributes();
  {
    gilColor3f(1.0, 0.0, 1.0);

    if ( m_dyniDrawDelaunay.GetValue() )
      {
	::RenderNormals(m_pQEMesh);
      }
  }
  PopAuxillaryGeometryAttributes();
}

VOID CSLIDETerrain::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", 
	  gcppcSLIDEReservedWords[sliderwtTerrain], m_pcGUIID);

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetGUIID());      
    }

  fprintf(fpOutputFile, ")\n");

  WriteSLIDEFlags(fpOutputFile, 0);

  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndTerrain]);
}

VOID CSLIDETerrain::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType eShading,
				    SLIDELODType eLOD)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBPushSolidAttributes(fpOutputFile,
			      (SLIDESolidType)m_dyniSolid.GetValue());
  {
    WriteRIBGeometryPolygonal(fpOutputFile, eShading, eLOD);
  }
  WriteRIBPopSolidAttributes(fpOutputFile, 
			     (SLIDESolidType)m_dyniSolid.GetValue());
}

BOOL CSLIDETerrain::CheckRIBPrimitive()
{
  BOOL bResult;
  //UINT_32 i;

  bResult = FALSE;
/*
  for ( i = 0; i < (UINT_32)slideribotSize; i++ )
    {
      bResult = bResult || (m_pcGUIID == gcppcSLIDERIBTerrainReservedWords[i]);  
    }
*/
  return bResult;
}

VOID CSLIDETerrain::WriteRIBGeometryPolygonal(FILE *fpOutputFile,
					     SLIDEShadingType eShading,
					     SLIDELODType /* eLOD */)
{
  //  CListIterInt iter;
  //CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );
/*
  iter.Init(&m_listCSLIDEPoint);
  
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->WriteRIB(fpOutputFile, eShading);
    }
*/
}

VOID CSLIDETerrain::WriteRIBGeometryPrimitive(FILE *fpOutputFile,
					     SLIDEShadingType /* eShading */,
					     SLIDELODType /* eLOD */)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDETerrain::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE TERRAIN: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDETerrain::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE TERRAIN: %s", m_pcGUIID);
}

VOID CSLIDETerrain::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &/*mTransform*/,
				      CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDETerrain::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &/*mTransform*/)
{
  ASSERT( fpOutputFile != NULL );
}

VOID CSLIDETerrain::WriteDotID(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "terrain_%s", GetGUIID());
}

VOID CSLIDETerrain::WriteDotNode(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "  ");
  WriteDotID(fpOutputFile);
  fprintf(fpOutputFile, " [ label = \"%s\", shape = ellipse, peripheries = 2", GetGUIID());
  if ( GetBoundDynamic() )
    {
      fprintf(fpOutputFile, ", color = red, fontcolor = red");
    }
  else if ( GetGeometryDynamic() )
    {
      fprintf(fpOutputFile, ", color = blue, fontcolor = blue");
    }
  fprintf(fpOutputFile, " ];\n");
}

VOID CSLIDETerrain::WriteDotEdges(FILE * /*fpOutputFile*/)
{
}

UINT_32 CSLIDETerrain::GetPointCount()
{
  return m_listCSLIDEPoint.GetLength();
}

