#include "slidetorus.h"

#include "slideworld.h"
#include "slidesurface.h"

#include "qemesh.h"
#include "qevertex.h"
#include "qewrite.h"
#include "stl.h"

static const FLOAT cfMajorRadius = 1.0;
static const FLOAT cfMinorRadius = 0.5;
static const FLOAT cfPhiMin = 0.0;
static const FLOAT cfPhiMax = DEGTORAD(360.0);
static const FLOAT cfThetaMax = DEGTORAD(360.0);
static const UINT_32 cuiPhiSlices = 8;
static const UINT_32 cuiThetaSlices = 8;

//////////////////////////////////////////////////////////////////////
// CSLIDETorus Class
//

CSLIDETorus::CSLIDETorus()
{
  m_dynfMajorRadius.SetValue(cfMajorRadius);
  m_dynfMinorRadius.SetValue(cfMinorRadius);
  m_dynfPhiMin.SetValue(cfPhiMin);
  m_dynfPhiMax.SetValue(cfPhiMax);
  m_dynfThetaMax.SetValue(cfThetaMax);
  m_dyniPhiSlices.SetValue(cuiPhiSlices);
  m_dyniThetaSlices.SetValue(cuiThetaSlices);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

CSLIDETorus::~CSLIDETorus()
{
}

VOID CSLIDETorus::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID)
{
  CSLIDEAtomNode::Init(pcFile, uiLine, uiColumn, pcID);

  m_dynfMajorRadius.SetValue(cfMajorRadius);
  m_dynfMinorRadius.SetValue(cfMinorRadius);
  m_dynfPhiMin.SetValue(cfPhiMin);
  m_dynfPhiMax.SetValue(cfPhiMax);
  m_dynfThetaMax.SetValue(cfThetaMax);
  m_dyniPhiSlices.SetValue(cuiPhiSlices);
  m_dyniThetaSlices.SetValue(cuiThetaSlices);

  m_pvTexture[0].Set(0.0, 0.0, 0.0);
  m_pvTexture[1].Set(1.0, 0.0, 0.0);
  m_pvTexture[2].Set(1.0, 1.0, 0.0);
  m_pvTexture[3].Set(0.0, 1.0, 0.0);
}

VOID CSLIDETorus::Uninit()
{
  CSLIDEAtomNode::Uninit();
}

BOOL CSLIDETorus::Initialized()
{
  if ( !CSLIDEAtomNode::Initialized() )
    {
      return FALSE;
    }

  return TRUE;
}

BOOL CSLIDETorus::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;

  if ( !CSLIDEAtomNode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  if ( !m_dynfMajorRadius.GetDynamic() && !( 0.0 < m_dynfMajorRadius.GetValue() ) )
    {
      HandleError("Illegal majorradius value:  0.0 < %.3f", 
		  m_dynfMajorRadius.GetValue());
      m_dynfMajorRadius.SetValue(cfMajorRadius);
      bCorrect = FALSE;
    }

  if ( !m_dynfMinorRadius.GetDynamic() && !( 0.0 < m_dynfMinorRadius.GetValue() ) )
    {
      HandleError("Illegal minorradius value:  0.0 < %.3f", 
		  m_dynfMinorRadius.GetValue());
      m_dynfMinorRadius.SetValue(cfMinorRadius);
      bCorrect = FALSE;
    }

  if ( !m_dynfPhiMin.GetDynamic() && 
       !( (FLOAT)(DEGTORAD(-720.0)) <= m_dynfPhiMin.GetValue() && 
	  m_dynfPhiMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
    {
      HandleError("Illegal phimin value:  %.3f <= %.3f <= %.3f", 
		  -720.0,
		  RADTODEG(m_dynfPhiMin.GetValue()),
		  720.0);
      m_dynfPhiMin.SetValue(cfPhiMin);
      bCorrect = FALSE;
    }

  if ( !m_dynfPhiMax.GetDynamic() &&
       !( (FLOAT)(DEGTORAD(-720.0)) <= m_dynfPhiMax.GetValue() && 
	  m_dynfPhiMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
    {
      HandleError("Illegal phimax value:  %.3f <= %.3f <= %.3f", 
		  -720.0,
		  RADTODEG(m_dynfPhiMax.GetValue()),
		  720.0);
      m_dynfPhiMax.SetValue(cfPhiMax);
      bCorrect = FALSE;
    }

  if ( !m_dynfThetaMax.GetDynamic() &&
       !( (FLOAT)(DEGTORAD(-360.0)) <= m_dynfThetaMax.GetValue() && 
	  m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(360.0)) ) )
    {
      HandleError("Illegal thetamax value:  %.3f <= %.3f <= %.3f", 
		  -360.0,
		  RADTODEG(m_dynfThetaMax.GetValue()),
		  360.0);
      m_dynfThetaMax.SetValue(cfThetaMax);
      bCorrect = FALSE;
    }

  if ( !m_dyniPhiSlices.GetDynamic() && 
       !( 0 < m_dyniPhiSlices.GetValue() ) )
    {
      HandleError("Illegal Phislices value:  %ld", 
		  m_dyniPhiSlices.GetValue());
      m_dyniPhiSlices.SetValue(cuiPhiSlices);
      bCorrect = FALSE;
    }

  if ( !m_dyniThetaSlices.GetDynamic() &&
       !( 0 < m_dyniThetaSlices.GetValue() ) )
    {
      HandleError("Illegal thetaslices value:  %ld", 
		  m_dyniThetaSlices.GetValue());
      m_dyniThetaSlices.SetValue(cuiThetaSlices);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDETorus::Valid()
{
  if ( !CSLIDEAtomNode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0.0 < m_dynfMajorRadius.GetValue() ) &&
	   ( 0.0 < m_dynfMinorRadius.GetValue() ) &&
	   ( (FLOAT)(DEGTORAD(-720.0)) <= m_dynfPhiMin.GetValue() && 
	     m_dynfPhiMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) &&
	   ( (FLOAT)(DEGTORAD(-720.0)) <= m_dynfPhiMax.GetValue() && 
	     m_dynfPhiMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) &&
	   ( (FLOAT)(DEGTORAD(-360.0)) <= m_dynfThetaMax.GetValue() && 
	     m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(360.0)) ) &&
	   ( 0 < m_dyniPhiSlices.GetValue() ) &&
	   ( 0 < m_dyniThetaSlices.GetValue() ) );
}

VOID CSLIDETorus::GetSide(CListInt *plistPoint, CListInt *plistNormal, UINT_32 uiSide, 
			  const Matrix &mTransform, const Matrix &mInverse, BOOL bReverse)
{
  ASSERT(uiSide < 4);

  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi, fZ, fR, fTheta, fZNormal, fRNormal, fNormalDir;
  Point pt;
  Point *ppt;
  Vector vNormal;
  Vector *pvNormal;

  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  if( (uiSide == 0) || (uiSide == 2) )
    {
      if( uiSide == 0 ) 
	{
	  fTheta = fThetaMin;
	}
      else
	{
	  fTheta = fThetaMax;
	}

      if( bReverse == 0 )
	{
	  fPhi = fPhiMin;
	}
      else
	{
	  fPhi = fPhiMax;
	  fPhiDelta *= -1.0;
	}

      for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
	{
	  fZ = fMinorRadius*(FLOAT)sin(fPhi);
	  fR = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi);
	  fZNormal = (FLOAT)sin(fPhi);
	  fRNormal = (FLOAT)cos(fPhi);
	  
	  pt.Set(fR*(FLOAT)cos(fTheta), fR*(FLOAT)sin(fTheta), fZ);
	  pt = mTransform * pt;	  
	  ppt = new Point;
	  ppt->Set(pt.Get());	  
	  plistPoint->InsertLast(ppt, 0);

	  vNormal.Set(fRNormal*(FLOAT)cos(fTheta), fRNormal*(FLOAT)sin(fTheta), fZNormal);
	  vNormal = fNormalDir * vNormal;
	  vNormal.Normalize();
	  pvNormal = new Vector;
	  pvNormal->Set(vNormal.Get());
	  plistNormal->InsertLast(pvNormal, 0);

	  fPhi += fPhiDelta;
	}
    }
  else // (uiSide == 1) || (uiSide == 3)  )
    {
      if( uiSide == 1 )
	{
	  fPhi = fPhiMax;
	}
      else
	{
	  fPhi = fPhiMin;
	}
      fZ = fMinorRadius*(FLOAT)sin(fPhi);
      fR = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi);
      fZNormal = (FLOAT)sin(fPhi);
      fRNormal = (FLOAT)cos(fPhi);

      if( bReverse == 0 )
	{
	  fTheta= fThetaMin;
	}
      else
	{
	  fTheta= fThetaMax;
	  fThetaDelta *= -1.0;
	}

      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  pt.Set(fR*(FLOAT)cos(fTheta), fR*(FLOAT)sin(fTheta), fZ);
	  pt = mTransform * pt;
	  ppt = new Point;
	  ppt->Set(pt.Get());
	  plistPoint->InsertLast(ppt, 0);

	  vNormal.Set(fRNormal*(FLOAT)cos(fTheta), fRNormal*(FLOAT)sin(fTheta), fZNormal);
	  vNormal = fNormalDir * vNormal;
	  vNormal.Normalize();
	  pvNormal = new Vector;
	  pvNormal->Set(vNormal.Get());
	  plistNormal->InsertLast(pvNormal, 0);

	  fTheta += fThetaDelta;
	}
    }
}

VOID CSLIDETorus::CheckDynamic()
{
  CSLIDEAtomNode::CheckDynamic();

  // Are there variables which I must call TCL to update?
  SetDynamic(GetDynamic() ||
	     m_dynfMajorRadius.GetDynamic() ||
	     m_dynfMinorRadius.GetDynamic() ||
	     m_dynfPhiMin.GetDynamic() ||
	     m_dynfPhiMax.GetDynamic() ||
	     m_dynfThetaMax.GetDynamic() ||
	     m_dyniPhiSlices.GetDynamic() ||
	     m_dyniThetaSlices.GetDynamic() ||
	     m_pvTexture[0].GetDynamic() ||
	     m_pvTexture[1].GetDynamic() ||
	     m_pvTexture[2].GetDynamic() ||
	     m_pvTexture[3].GetDynamic());
}

VOID CSLIDETorus::PreprocessGraphLocal()
{
  BOOL bRealSpaceDynamic;
  BOOL bParameterSpaceDynamic;
  BOOL bTextureSpaceDynamic;
  
  bRealSpaceDynamic = ( m_dynfMajorRadius.GetDynamic() ||
			m_dynfMinorRadius.GetDynamic() ||
			m_dynfPhiMin.GetDynamic() ||
			m_dynfPhiMax.GetDynamic() ||
			m_dynfThetaMax.GetDynamic() );
  bParameterSpaceDynamic = ( m_dyniPhiSlices.GetDynamic() ||
			     m_dyniThetaSlices.GetDynamic() );
  bTextureSpaceDynamic = ( m_pvTexture[0].GetDynamic() ||
			   m_pvTexture[1].GetDynamic() ||
			   m_pvTexture[2].GetDynamic() ||
			   m_pvTexture[3].GetDynamic() );

  SetGeometryEmpty(FALSE);
  SetGeometryDynamic(bRealSpaceDynamic || bParameterSpaceDynamic || bTextureSpaceDynamic);
  SetNormalsEmpty(FALSE);
  SetNormalsDynamic(bRealSpaceDynamic || bParameterSpaceDynamic);
  SetBoundsEmpty(FALSE);
  SetBoundsDynamic(bRealSpaceDynamic);
  SetBoundEmpty(FALSE);
  SetBoundDynamic(bRealSpaceDynamic);
}

VOID CSLIDETorus::UpdateState()
{
  CSLIDEAtomNode::UpdateState();

  m_dynfMajorRadius.Update();
  if ( !( 0.0 < m_dynfMajorRadius.GetValue() ) )
    {
      HandleError("Illegal majorradius value:  0.0 < %.3f", 
		  m_dynfMajorRadius.GetValue());
      m_dynfMajorRadius.SetValue(cfMajorRadius);
    }

  m_dynfMinorRadius.Update();
  if ( !( 0.0 < m_dynfMinorRadius.GetValue() ) )
    {
      HandleError("Illegal minorradius value:  0.0 < %.3f", 
		  m_dynfMinorRadius.GetValue());
      m_dynfMinorRadius.SetValue(cfMinorRadius);
    }

  if ( m_dynfPhiMin.GetDynamic() )
    {
      m_dynfPhiMin.Update();
      m_dynfPhiMin.SetValue(DEGTORAD(m_dynfPhiMin.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(-720.0)) <= m_dynfPhiMin.GetValue() && 
	      m_dynfPhiMin.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
	{
	  HandleError("Illegal phimin value:  %.3f <= %.3f <= %.3f", 
		      -720.0,
		      RADTODEG(m_dynfPhiMin.GetValue()),
		      720.0);
	  m_dynfPhiMin.SetValue(cfPhiMin);
	}
    }

  if ( m_dynfPhiMax.GetDynamic() )
    {
      m_dynfPhiMax.Update();
      m_dynfPhiMax.SetValue(DEGTORAD(m_dynfPhiMax.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(-720.0)) <= m_dynfPhiMax.GetValue() && 
	      m_dynfPhiMax.GetValue() <= (FLOAT)(DEGTORAD(720.0)) ) )
	{
	  HandleError("Illegal phimax value:  %.3f <= %.3f <= %.3f", 
		      -720.0,
		      RADTODEG(m_dynfPhiMax.GetValue()),
		      720.0);
	  m_dynfPhiMax.SetValue(cfPhiMax);
	}
    }

  if ( m_dynfThetaMax.GetDynamic() )
    {
      m_dynfThetaMax.Update();
      m_dynfThetaMax.SetValue(DEGTORAD(m_dynfThetaMax.GetValue()));

      if ( !( (FLOAT)(DEGTORAD(-360.0)) <= m_dynfThetaMax.GetValue() && 
	      m_dynfThetaMax.GetValue() <= (FLOAT)(DEGTORAD(360.0)) ) )
	{
	  HandleError("Illegal thetamax value:  %.3f <= %.3f <= %.3f", 
		      -360.0,
		      RADTODEG(m_dynfThetaMax.GetValue()),
		      360.0);
	  m_dynfThetaMax.SetValue(cfThetaMax);
	}
    }

  m_dyniPhiSlices.Update();
  if ( !( 0 < m_dyniPhiSlices.GetValue() ) )
    {
      HandleError("Illegal phislices value:  %ld", 
		  m_dyniPhiSlices.GetValue());
      m_dyniPhiSlices.SetValue(cuiPhiSlices);
    }

  m_dyniThetaSlices.Update();
  if ( !( 0 < m_dyniThetaSlices.GetValue() ) )
    {
      HandleError("Illegal thetaslices value:  %ld", 
		  m_dyniThetaSlices.GetValue());
      m_dyniThetaSlices.SetValue(cuiThetaSlices);
    }

  UINT_32 i;

  for ( i = 0; i < 4; i++ )
    {
      m_pvTexture[i].Update();
    }
}

VOID CSLIDETorus::CalculateBound()
{
  CBound bound;
  Point pt;

  FLOAT pfPhiValues[6];
  UINT_32 uiPhiValue, uiPhiValues;
  FLOAT fMajorRadius, fMinorRadius;
  FLOAT fTheta, fThetaMin, fThetaMax, fThetaDelta;
  FLOAT fPhi, fPhiMin, fPhiMax, fPhiDelta;
  FLOAT fR, fZ;

  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();

  // Figure out which Phi values are important to the calculation
  uiPhiValues = 2;
  pfPhiValues[0] = m_dynfPhiMin.GetValue();
  pfPhiValues[1] = m_dynfPhiMax.GetValue();

  fPhiDelta = DEGTORAD(90);
  for ( fPhi = DEGTORAD(-720); 
	fPhi < pfPhiValues[1] && fPhi < pfPhiValues[0];
	fPhi += fPhiDelta )
    {
    }
  for ( ; 
	uiPhiValues < 6 &&
	(fPhi < pfPhiValues[1] || fPhi < pfPhiValues[0]);
	fPhi += fPhiDelta)
    {
      ASSERT( uiPhiValues < 6 );

      pfPhiValues[uiPhiValues] = fPhi;
      uiPhiValues++;
    }

  fThetaDelta = DEGTORAD(90);
  if ( fThetaMax < fThetaMin )
    {
      fThetaDelta = -fThetaDelta;
    }

  for ( uiPhiValue = 0; uiPhiValue < uiPhiValues; uiPhiValue++ )
    {
      fR = fMajorRadius + fMinorRadius*(FLOAT)cos(pfPhiValues[uiPhiValue]);
      fZ = fMinorRadius*(FLOAT)sin(pfPhiValues[uiPhiValue]);
      for ( fTheta = 0.0; 
	    ( fThetaMax < fThetaMin ) ? 
	      (fTheta > fThetaMax) : 
	      (fTheta < fThetaMax); 
	      fTheta += fThetaDelta )
	{
	  pt.Set(fR*(FLOAT)cos(fTheta), fR*(FLOAT)sin(fTheta), fZ);
	  bound.Insert(pt);
	}
      fTheta = fThetaMax;
      pt.Set(fR*(FLOAT)cos(fTheta), fR*(FLOAT)sin(fTheta), fZ);
      bound.Insert(pt);
    }
  
  if ( bound != m_bound )
    {
      m_bound = bound;
      SetBoundChanged(TRUE);
    }
}

#ifdef __USE_CDA__
VOID CSLIDETorus::CalculateCDAVertices(CListInt *plistCDAVertex, 
				       const Matrix &mGroot_obj)
{
  Point pt;
  CDAVertex *pCDAVertex;

  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi, fZ, fR, fTheta;

  ASSERT( plistCDAVertex != NULL );
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  fPhi = fPhiMin;
  for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
    {	
      fZ = fMinorRadius*(FLOAT)sin(fPhi);
      fR = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi);
      
      fTheta = fThetaMin;
      for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	{
	  pt.Set(fR*(FLOAT)cos(fTheta), fR*(FLOAT)sin(fTheta), fZ);

	  pCDAVertex = AllocateCDAVertex();
	  ASSERT( pCDAVertex != NULL );
	  pCDAVertex->Init(mGroot_obj*pt);
	  plistCDAVertex->InsertLast(pCDAVertex, 0);
	  
	  fTheta += fThetaDelta;
	}
      
      fPhi += fPhiDelta;
    }
}
#endif // __USE_CDA__

VOID CSLIDETorus::RenderGeometrySmooth(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  FLOAT fNormalDir;
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  FLOAT fRNormal0, fZNormal0, fRNormal1, fZNormal1;
  Point pt;
  Vector vNormal;

  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiPhiSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiPhiSlices;

  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
  
  fPhi0 = fPhiMin;
  fPhi1 = fPhi0 + fPhiDelta;
  for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
    {
      fZNormal1 = (FLOAT)sin(fPhi1);
      fRNormal1 = (FLOAT)cos(fPhi1);
      fZNormal0 = (FLOAT)sin(fPhi0);
      fRNormal0 = (FLOAT)cos(fPhi0);

      fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
      fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
      fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
      fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;

	pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
	pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
	pvU[0] = ppvV[0][0];
	pvU[1] = ppvV[1][0];

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    vNormal.Set(fRNormal1*(FLOAT)cos(fTheta), fRNormal1*(FLOAT)sin(fTheta), fZNormal1);
	    vNormal = fNormalDir * vNormal;
	    pt.Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pt.Get());
	    
	    vNormal.Set(fRNormal0*(FLOAT)cos(fTheta), fRNormal0*(FLOAT)sin(fTheta), fZNormal0);
	    vNormal = fNormalDir * vNormal;
	    pt.Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pt.Get());
	    
	    fTheta += fThetaDelta;

	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	  }
      }
      gilEnd();
      
      fPhi0 = fPhi1;
      fPhi1 += fPhiDelta;

      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    }
}

VOID CSLIDETorus::CalculateQEMesh()
{
  //FLOAT fNormalDir;
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  FLOAT fRNormal0, fZNormal0, fRNormal1, fZNormal1;
  Point pt;
  
  QEVertex* pqev;
  CObjectList<QEVertex> listQEVertex;

  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  ASSERT( m_pQEMesh == NULL );
  
  m_pQEMesh = AllocateQEMesh();
  ASSERT( m_pQEMesh != NULL );
  m_pQEMesh->Init();
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  listQEVertex.Init();

  fPhi0 = fPhiMin;
  fPhi1 = fPhi0 + fPhiDelta;
  for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
    {
      listQEVertex.RemoveAll();
      fZNormal1 = (FLOAT)sin(fPhi1);
      fRNormal1 = (FLOAT)cos(fPhi1);
      fZNormal0 = (FLOAT)sin(fPhi0);
      fRNormal0 = (FLOAT)cos(fPhi0);

      fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
      fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
      fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
      fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
      
      //gilBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;

	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    pt.Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	
	    pqev = AllocateQEVertex();
	    pqev->Init(pt);
	    listQEVertex.InsertLast(pqev);
	    pt.Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	    pqev = AllocateQEVertex();
	    pqev->Init(pt);
	    listQEVertex.InsertLast(pqev);
	    
	    fTheta += fThetaDelta;
	  }
      }
      m_pQEMesh->MergeTriangleStrip(&listQEVertex);
      //gilEnd();
      
      fPhi0 = fPhi1;
      fPhi1 += fPhiDelta;
    }
}

VOID CSLIDETorus::CalculateLeafQEMesh()
{
  if( m_pQEMesh == NULL )
    {
      CalculateQEMesh();
    }
}

VOID CSLIDETorus::DeallocateLeafQEMesh()
{
  if( m_pQEMesh != NULL )
    {
      DeallocateQEMesh(m_pQEMesh);
      m_pQEMesh = NULL;
    }
}

VOID CSLIDETorus::RenderGeometryFlat(CSLIDEWormwarp *pCSLIDEWarp /*= NULL*/)
{
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  Point pppt[2][2];
  Vector vNormal;

  Vector pvVDelta[2];
  Vector pvUDelta[2];
  Vector ppvV[2][2];
  Vector pvU[2];
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  pvVDelta[0] = (m_pvTexture[3] - m_pvTexture[0]) / (FLOAT)uiPhiSlices;
  pvVDelta[1] = (m_pvTexture[2] - m_pvTexture[1]) / (FLOAT)uiPhiSlices;

  ppvV[0][0] = m_pvTexture[0];
  ppvV[0][1] = m_pvTexture[1];
  ppvV[1][0] = ppvV[0][0] + pvVDelta[0];
  ppvV[1][1] = ppvV[0][1] + pvVDelta[1];
  
  fPhi0 = fPhiMin;
  fPhi1 = fPhi0 + fPhiDelta;
  for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
    {
      fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
      fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
      fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
      fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
      
      gilBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;

	pvUDelta[0] = (ppvV[0][1] - ppvV[0][0]) / (FLOAT)uiThetaSlices;
	pvUDelta[1] = (ppvV[1][1] - ppvV[1][0]) / (FLOAT)uiThetaSlices;
	pvU[0] = ppvV[0][0];
	pvU[1] = ppvV[1][0];

	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	pppt[1][1].Set(fR1*(FLOAT)cos(fTheta+fThetaDelta), fR1*(FLOAT)sin(fTheta+fThetaDelta), fZ1);

	vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	vNormal.Normalize();

	gilNormal3fv(vNormal.Get());
	gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	gilVertex3fv(pppt[1][0].Get());
	gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
	gilVertex3fv(pppt[0][0].Get());

	fTheta += fThetaDelta;

	pvU[0] += pvUDelta[0];
	pvU[1] += pvUDelta[1];

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);

	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    gilTexCoord2f(pvU[1][0], pvU[1][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pppt[1][1].Get());

	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    gilTexCoord2f(pvU[0][0], pvU[0][1]);	    
	    gilNormal3fv(vNormal.Get());
	    gilVertex3fv(pppt[0][1].Get());
	    
	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;

	    pvU[0] += pvUDelta[0];
	    pvU[1] += pvUDelta[1];
	  }
      }
      gilEnd();
      
      fPhi0 = fPhi1;
      fPhi1 += fPhiDelta;

      ppvV[0][0] += pvVDelta[0];
      ppvV[0][1] += pvVDelta[1];
      ppvV[1][0] += pvVDelta[0];
      ppvV[1][1] += pvVDelta[1];
    } 
}

VOID CSLIDETorus::RenderNormalsSmooth()
{
  FLOAT fNormalDir;
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi, fZ, fR, fTheta;
  FLOAT fRNormal, fZNormal;
  Point pt;
  Vector vNormal;
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  gilBegin(GL_LINES);
  {
    fPhi = fPhiMin;
    for ( uiPhiSlice = 0; uiPhiSlice <= uiPhiSlices; uiPhiSlice++ )
      {
	fZNormal = (FLOAT)sin(fPhi);
	fRNormal = (FLOAT)cos(fPhi);
	
	fZ = fMinorRadius*(FLOAT)sin(fPhi);
	fR = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi);

	fTheta = fThetaMin;
	for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	  {
	    vNormal.Set(fRNormal*(FLOAT)cos(fTheta), fRNormal*(FLOAT)sin(fTheta), fZNormal);
	    vNormal = fNormalDir * vNormal;
	    pt.Set(fR*(FLOAT)cos(fTheta), fR*(FLOAT)sin(fTheta), fZ);
	    gilVertex3fv(pt.Get());
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());
	    
	    fTheta += fThetaDelta;
	  }
      
      fPhi += fPhiDelta;
    }
  }
  gilEnd();
}

VOID CSLIDETorus::RenderNormalsFlat()
{
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  Point pppt[2][2];
  Point pt;
  Vector vNormal;
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  gilBegin(GL_LINES);
  {
    fPhi0 = fPhiMin;
    fPhi1 = fPhi0 + fPhiDelta;
    for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
      {
	fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
	fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
	fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
	fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
	
	fTheta = fThetaMin;

	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);

	fTheta += fThetaDelta;

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);

	    pt = (pppt[1][0] + pppt[0][0] + pppt[1][1]) / 3.0;
	    gilVertex3fv(pt.Get());
	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());

	    pt = (pppt[0][1] + pppt[1][1] + pppt[0][0]) / 3.0;
	    gilVertex3fv(pt.Get());
	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());
	    
	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;
	  }

	fPhi0 = fPhi1;
	fPhi1 += fPhiDelta;
      } 
  }
  gilEnd();
      
  /*
  UINT_32 uiZSlices, uiThetaSlices;
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fRadius, fZMin, fZMax, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta, fVDelta, fUDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  Point pppt[2][2];
  Point pt;
  Vector vNormal;
  
  uiZSlices = (UINT_32)m_dyniZSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fRadius = m_dynfRadius.GetValue();
  fZMin = m_dynfZMin.GetValue();
  fZMax = m_dynfZMax.GetValue();
  fPhiMin = a(FLOAT)sin(MIN(MAX(fZMin/fRadius, -1.0), 1.0));
  fPhiMax = a(FLOAT)sin(MIN(MAX(fZMax/fRadius, -1.0), 1.0));
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  gilBegin(GL_LINES);
  {
    fPhi0 = fPhiMin;
    fPhi1 = fPhi0 + fPhiDelta;
    for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
      {
	fZ1 = fRadius*(FLOAT)sin(fPhi1);
	fR1 = fRadius*(FLOAT)cos(fPhi1);
	fZ0 = fRadius*(FLOAT)sin(fPhi0);
	fR0 = fRadius*(FLOAT)cos(fPhi0);
      
	fTheta = fThetaMin;

	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);

	fTheta += fThetaDelta;

	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);

	    pt = (pppt[1][0] + pppt[0][0] + pppt[1][1]) / 3.0;
	    gilVertex3fv(pt.Get());
	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());

	    pt = (pppt[0][1] + pppt[1][1] + pppt[0][0]) / 3.0;
	    gilVertex3fv(pt.Get());
	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    pt = pt + vNormal;
	    gilVertex3fv(pt.Get());

	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;
	  }
      
	fPhi0 = fPhi1;
	fPhi1 += fPhiDelta;
      } 
  }
  gilEnd();
  */
}

VOID CSLIDETorus::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n", 
	  gcppcSLIDEReservedWords[sliderwtTorus], m_pcGUIID);

  WriteSLIDEFlags(fpOutputFile, 0);

  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtMajorRadius],
	  m_dynfMajorRadius.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtMinorRadius],
	  m_dynfMinorRadius.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtPhiMin],
	  RADTODEG(m_dynfPhiMin.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtPhiMax],
	  RADTODEG(m_dynfPhiMax.GetValue()));
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtThetaMax],
	  RADTODEG(m_dynfThetaMax.GetValue()));
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtPhiSlices],
	  m_dyniPhiSlices.GetValue());
  fprintf(fpOutputFile, "  %s %ld\n",
	  gcppcSLIDEReservedWords[sliderwtThetaSlices],
	  m_dyniThetaSlices.GetValue());

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndTorus]);
}

VOID CSLIDETorus::WriteRIBGeometry(FILE *fpOutputFile,
				    const Matrix &/*mGproj_obj*/, 
				    CBound * /*pCBoundClip*/, 
				    SLIDEShadingType /* eShading */,
				    SLIDELODType /* eLOD*/ )
{
  RtFloat rtfMajorRadius, rtfMinorRadius, rtfPhiMin, rtfPhiMax, rtfThetaMax;
  RtFloat prtvTex[4][2];

  ASSERT( fpOutputFile != NULL );

  rtfMajorRadius = m_dynfMajorRadius.GetValue();
  rtfMinorRadius = m_dynfMinorRadius.GetValue();
  rtfPhiMin = RADTODEG(m_dynfPhiMin.GetValue());
  rtfPhiMax = RADTODEG(m_dynfPhiMax.GetValue());
  rtfThetaMax = RADTODEG(m_dynfThetaMax.GetValue());

  prtvTex[0][0] = m_pvTexture[0](0);
  prtvTex[0][1] = m_pvTexture[0](1);
  prtvTex[1][0] = m_pvTexture[1](0);
  prtvTex[1][1] = m_pvTexture[1](1);
  prtvTex[2][0] = m_pvTexture[3](0);
  prtvTex[2][1] = m_pvTexture[3](1);
  prtvTex[3][0] = m_pvTexture[2](0);
  prtvTex[3][1] = m_pvTexture[2](1);

  RiTorus(rtfMajorRadius, rtfMinorRadius, rtfPhiMin, rtfPhiMax, rtfThetaMax, RI_ST, (RtPointer)prtvTex, RI_NULL);
/*
  RiArchiveRecord(RI_COMMENT, "\nTorus %.3f %.3f %.3f %.3f %.3f",
		  m_dynfMajorRadius.GetValue(),
		  m_dynfMinorRadius.GetValue(),
		  RADTODEG(m_dynfPhiMin.GetValue()),
		  RADTODEG(m_dynfPhiMax.GetValue()),
		  RADTODEG(m_dynfThetaMax.GetValue()));
*/
}

VOID CSLIDETorus::WriteRIBPushSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  RiArchiveRecord(RI_COMMENT, " BEGIN SLIDE TORUS: %s", m_pcGUIID);
  WriteRIBBegin(fpOutputFile);
}

VOID CSLIDETorus::WriteRIBPopSLIDEName(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  WriteRIBEnd(fpOutputFile);
  RiArchiveRecord(RI_COMMENT, "   END SLIDE TORUS: %s", m_pcGUIID);
}

VOID CSLIDETorus::WritePOVGeometry(FILE *fpOutputFile, 
				   CListInt *plistTransformIDs,
				   CListInt *plistColors)  
{

  ASSERT( fpOutputFile != NULL );

  CListIterInt iter;

  FLOAT fNormalDir;
  UINT_32 uiPhiSlices, uiThetaSlices;
  UINT_32 uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1;
  FLOAT fTheta0, fTheta1;
  //FLOAT fRNormal0, fZNormal0, fRNormal1, fZNormal1;
  CHAR * pcTransformID;
  CHAR *pcColorID;
  Point pt;
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();
  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  
  if ( ( (fPhiMin > fPhiMax) && !(fThetaMin > fThetaMax) ) ||
       ( !(fPhiMin > fPhiMax) && (fThetaMin > fThetaMax) ) )
    {
      fNormalDir = -1.0;
    }
  else
    {
      fNormalDir = 1.0;
    }

  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;

  if ( (fabs(fThetaMax) == cfThetaMax) && (fabs(fPhiMin-fPhiMax) >= cfThetaMax) ) 
    {
      fprintf(fpOutputFile, "torus {\n");
      fprintf(fpOutputFile, " %f, %f \n", fMajorRadius, fMinorRadius);
    }
  else
    {
      fPhi0 = fPhiMin;
      fPhi1 = fPhi0 + fPhiDelta;
      
      fprintf(fpOutputFile, "mesh { \n");

      for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
	{
	  fTheta0 = fThetaMin;
	  fTheta1 = fTheta0 + fThetaDelta;

	  fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
	  fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
	  fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
	  fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
	  for ( uiThetaSlice = 0; uiThetaSlice <= uiThetaSlices; uiThetaSlice++ )
	    {
	      fprintf(fpOutputFile, "triangle { ");
	      fprintf(fpOutputFile, "<%f, %f, %f>,", fR0*(FLOAT)sin(fTheta0), fR0*(FLOAT)cos(fTheta0), fZ0);
	      fprintf(fpOutputFile, "<%f, %f, %f>,", fR1*(FLOAT)sin(fTheta0), fR1*(FLOAT)cos(fTheta0), fZ1);
	      fprintf(fpOutputFile, "<%f, %f, %f> ", fR0*(FLOAT)sin(fTheta1), fR0*(FLOAT)cos(fTheta1), fZ0);
	      fprintf(fpOutputFile, "}\n");

	      fprintf(fpOutputFile, "triangle { ");
	      fprintf(fpOutputFile, "<%f, %f, %f>,", fR0*(FLOAT)sin(fTheta1), fR0*(FLOAT)cos(fTheta1), fZ0);
	      fprintf(fpOutputFile, "<%f, %f, %f>,", fR1*(FLOAT)sin(fTheta0), fR1*(FLOAT)cos(fTheta0), fZ1);
	      fprintf(fpOutputFile, "<%f, %f, %f> ", fR1*(FLOAT)sin(fTheta1), fR1*(FLOAT)cos(fTheta1), fZ1);
	      fprintf(fpOutputFile, "}\n");

	      fTheta0 = fTheta1;
	      fTheta1 = fTheta1 + fThetaDelta;
	    }
	  fPhi0 = fPhi1;
	  fPhi1 = fPhi1 + fPhiDelta;
	}
    }
  iter.Init(plistTransformIDs);
  for ( pcTransformID = (CHAR*)iter.PeekFirst(); 
	pcTransformID != NULL;
	pcTransformID = (CHAR*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, " transform %s\n", pcTransformID);
    }

  //color
  ASSERT(m_pCSLIDESurface != NULL);

  if (!(strcmp(GetSurfaceID(), "SLF_INHERIT")))
    {
      iter.Init(plistColors);
      pcColorID = (CHAR*)iter.PeekFirst();
      fprintf(fpOutputFile, " pigment { %s }\n", pcColorID);
    }
  else
    {
      fprintf(fpOutputFile, " pigment {");
      m_pCSLIDESurface->WritePOV(fpOutputFile);
      fprintf(fpOutputFile, " }\n");
    }

  fprintf(fpOutputFile, "}\n\n");
}

VOID CSLIDETorus::WriteWRLZGeometry(FILE *fpOutputFile, const Matrix &mTransform,
				    CSLIDESurface *pCSLIDESurface)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "Shape { # triangle mesh: TORUS\n\n");

  // Color Inheritance
  if( m_pCSLIDESurface == CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }
  else
    {
      ASSERT( m_pCSLIDESurface != NULL );
      m_pCSLIDESurface->WriteWRLZ(fpOutputFile);
    }

  // Mesh
  ASSERT(m_pQEMesh != NULL);
  ::WriteWRLZ(m_pQEMesh, fpOutputFile, mTransform);

  fprintf(fpOutputFile, "} # end triangle mesh\n\n");
}

VOID CSLIDETorus::WriteSTLGeometry(FILE *fpOutputFile, const Matrix &mTransform)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 uiPhiSlices, uiThetaSlices, uiPhiSlice, uiThetaSlice;
  FLOAT fMajorRadius, fMinorRadius, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  Point pppt[2][2];
  
  uiPhiSlices = (UINT_32)m_dyniPhiSlices.GetValue();
  uiThetaSlices = (UINT_32)m_dyniThetaSlices.GetValue();  
  fMajorRadius = m_dynfMajorRadius.GetValue();
  fMinorRadius = m_dynfMinorRadius.GetValue();
  fPhiMin = m_dynfPhiMin.GetValue();
  fPhiMax = m_dynfPhiMax.GetValue();
  fThetaMin = 0.0;
  fThetaMax = m_dynfThetaMax.GetValue();
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiPhiSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  fPhi0 = fPhiMin;
  fPhi1 = fPhi0 + fPhiDelta;
      
  if( mTransform.Determinant() > 0 ) // Don't mirror
    {
      for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
	{
	  fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
	  fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
	  fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
	  fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
	  
	  fTheta = fThetaMin;
	  
	  // Must seed the for loop with the left most two points
	  pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	  pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	  
	  pppt[0][0] = mTransform * pppt[0][0];
	  pppt[1][0] = mTransform * pppt[1][0];
	  
	  fTheta += fThetaDelta;
	  
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	      pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	      
	      pppt[1][1] = mTransform * pppt[1][1];
	      pppt[0][1] = mTransform * pppt[0][1];

	      WriteSTLTriangle(fpOutputFile, pppt[1][0], pppt[0][0], pppt[1][1]);
	      WriteSTLTriangle(fpOutputFile, pppt[0][1], pppt[1][1], pppt[0][0]);
	      
	      pppt[0][0] = pppt[0][1];
	      pppt[1][0] = pppt[1][1];
	      fTheta += fThetaDelta;
	    }
	  fPhi0 = fPhi1;
	  fPhi1 += fPhiDelta;
	} 
    }
  else // Mirror
    {
      for ( uiPhiSlice = 0; uiPhiSlice < uiPhiSlices; uiPhiSlice++ )
	{
	  fZ1 = fMinorRadius*(FLOAT)sin(fPhi1);
	  fR1 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi1);
	  fZ0 = fMinorRadius*(FLOAT)sin(fPhi0);
	  fR0 = fMajorRadius + fMinorRadius*(FLOAT)cos(fPhi0);
	  
	  fTheta = fThetaMin;
	  
	  // Must seed the for loop with the left most two points
	  pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	  pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	  
	  pppt[0][0] = mTransform * pppt[0][0];
	  pppt[1][0] = mTransform * pppt[1][0];
	  
	  fTheta += fThetaDelta;
	  
	  for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	    {
	      pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	      pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	      
	      pppt[1][1] = mTransform * pppt[1][1];
	      pppt[0][1] = mTransform * pppt[0][1];

	      WriteSTLTriangle(fpOutputFile, pppt[1][1], pppt[0][0], pppt[1][0]);
	      WriteSTLTriangle(fpOutputFile, pppt[0][0], pppt[1][1], pppt[0][1]);
	      
	      pppt[0][0] = pppt[0][1];
	      pppt[1][0] = pppt[1][1];
	      fTheta += fThetaDelta;
	    }
	  fPhi0 = fPhi1;
	  fPhi1 += fPhiDelta;
	} 
    }
}

VOID CSLIDETorus::GetParameters(Tcl_Interp *pInterp)
{
  INT_32 iCode;
  CHAR pcBuffer[1000];
  CHAR pcSurface[35];

  //pInterp = CSLIDEWorld::GetInterp();

  if( !strcmp(m_pCSLIDESurface->GetGUIID(), "SLF_INHERIT" ) )
      sprintf(pcSurface, "inherit");
  else 
      sprintf(pcSurface, m_pCSLIDESurface->GetGUIID());

  sprintf(pcBuffer, "Editor:SetParameterValues torus {{majorradius %.3f} {minorradius %.3f} {thetamax %.3f} {thetaslices %ld} {phimin %.3f} {phimax %.3f} {phislices %ld} {solid %s} {surface %s} {shading %s} {lod %s} {ribbegin %s} {ribend %s}}",
	  m_dynfMajorRadius.GetValue(),
	  m_dynfMinorRadius.GetValue(),
	  RADTODEG(m_dynfThetaMax.GetValue()),
	  m_dyniThetaSlices.GetValue(),
	  RADTODEG(m_dynfPhiMin.GetValue()),
	  RADTODEG(m_dynfPhiMax.GetValue()),
	  m_dyniPhiSlices.GetValue(),
	  gcppcSLIDEGUISolidReservedWords[m_dyniSolid.GetValue()],
	  pcSurface,
	  gcppcSLIDEGUIShadingReservedWords[m_dyniShading.GetValue()],
	  gcppcSLIDEGUILODReservedWords[m_dyniLOD.GetValue()],
	  m_pcRIBBegin,
	  m_pcRIBEnd);
	  
  iCode = Tcl_Eval(pInterp, pcBuffer);
  //ASSERT( iCode == TCL_OK );
  if ( iCode != TCL_OK )
    {
      printf( "%s\n", pcBuffer );
      printf("Error %s \n", pInterp->result);
    }
}

UINT_32 CSLIDETorus::SetParameter(CSLIDEWorld *pCSLIDEWorld, 
				 SLIDEReservedWordType uiParameter, CHAR **argv)
{
  switch(uiParameter) 
    {
    case sliderwtMajorRadius:
      m_dynfMajorRadius.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtMinorRadius:
      m_dynfMinorRadius.SetValue(atof(argv[0]));
      return 1;
      break;
    case sliderwtThetaMax:
      m_dynfThetaMax.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtThetaSlices:
      m_dyniThetaSlices.SetValue(atoi(argv[0]));
      return 1;
      break;
    case sliderwtPhiMin:
      m_dynfPhiMin.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtPhiMax:
      m_dynfPhiMax.SetValue(DEGTORAD(atoi(argv[0])));
      return 1;
      break;
    case sliderwtPhiSlices:
      m_dyniPhiSlices.SetValue(atoi(argv[0]));
      return 1;
      break;
    default:
      return CSLIDEAtomNode::SetParameter(pCSLIDEWorld, uiParameter, argv);
      break;
    }
}
