#include "hashtable.h"
#include "qemesh.h"

#include "qenode.h"
#include "qeedgecollapse.h"

#include "qevertex.h"
#include "qewedge.h"
#include "qeedge.h"
#include "qefacet.h"

#include "predicates.h"

//////////////////////////////////////////////////////////////////////
// QEMesh Class Subdivision Methods
//

//////////////////////////////////////////////////
// START: Sharp Loop Tables
//

//#define DEBUGGING_SHARP_LOOP_MATH
static const UINT_32 scpuiSharpLoopSharpTypeIndex[QE::svtSize] = {2, 2, 0, 0, 1, 1};
static const FLOAT scppppfSharpLoopLimitTable[2][3][3][7] = 
{
  {
    {
      {   4.0/6.0,      0.0,  1.0/6.0,      0.0,      0.0,  1.0/6.0,      0.0 }, // R_RR
      { 33.0/48.0,      0.0, 8.0/48.0,      0.0,      0.0, 7.0/48.0,      0.0 }, // R_RN
      { 31.0/48.0,      0.0, 8.0/48.0,      0.0, 1.0/48.0, 7.0/48.0, 1.0/48.0 }  // R_RD
    },
    {
      { 33.0/48.0,      0.0, 7.0/48.0,      0.0,      0.0, 8.0/48.0,      0.0 }, // R_NR
      { 34.0/48.0,      0.0, 7.0/48.0,      0.0,      0.0, 7.0/48.0,      0.0 }, // R_NN
      { 32.0/48.0,      0.0, 7.0/48.0,      0.0, 1.0/48.0, 7.0/48.0, 1.0/48.0 }  // R_ND
    },
    {
      { 31.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0,      0.0, 8.0/48.0,      0.0 }, // R_DR
      { 32.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0,      0.0, 7.0/48.0,      0.0 }, // R_DN
      { 30.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0 }  // R_DD
    }
  },
  {
    {
      {   3.0/5.0,      0.0,  1.0/5.0,      0.0,      0.0,  1.0/5.0,      0.0 }, // N_RR
      { 25.0/40.0,      0.0, 8.0/40.0,      0.0,      0.0, 7.0/40.0,      0.0 }, // N_RN
      { 24.0/40.0,      0.0, 8.0/40.0,      0.0, 1.0/40.0, 6.0/40.0, 1.0/40.0 }  // N_RD
    },
    {
      { 25.0/40.0,      0.0, 7.0/40.0,      0.0,      0.0, 8.0/40.0,      0.0 }, // N_NR
      { 26.0/40.0,      0.0, 7.0/40.0,      0.0,      0.0, 7.0/40.0,      0.0 }, // N_NN
      { 25.0/40.0,      0.0, 7.0/40.0,      0.0, 1.0/40.0, 6.0/40.0, 1.0/40.0 }  // N_ND
    },
    {
      { 24.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0,      0.0, 8.0/40.0,      0.0 }, // N_DR
      { 25.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0,      0.0, 7.0/40.0,      0.0 }, // N_DN
      { 24.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0 }  // N_DD
    }
  }
};
static const FLOAT scppppfSharpLoopDeltaTable[2][3][3][7] = 
{
  {
    {
      {  2.0/6.0,      0.0,  -1.0/6.0,      0.0,       0.0   -1.0/6.0,       0.0 }, // R_RR
      { 3.0/48.0,      0.0, -2.0/48.0,      0.0,       0.0  -1.0/48.0,       0.0 }, // R_RN
      { 5.0/48.0,      0.0, -2.0/48.0,      0.0, -1.0/48.0, -1.0/48.0, -1.0/48.0 }  // R_RD
    },
    {
      { 3.0/48.0,      0.0, -1.0/48.0,      0.0,       0.0, -2.0/48.0,       0.0 }, // R_NR
      { 2.0/48.0,      0.0, -1.0/48.0,      0.0,       0.0, -1.0/48.0,       0.0 }, // R_NN
      { 4.0/48.0,      0.0, -1.0/48.0,      0.0, -1.0/48.0, -1.0/48.0, -1.0/48.0 }  // R_ND
    },
    {
      { 5.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0,     0.0, -2.0/48.0,       0.0 },  // R_DR
      { 4.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0,     0.0, -1.0/48.0,       0.0 },  // R_DN
      { 6.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0 } // R_DD
    }
  },
  {
    {
      {  2.0/5.0,      0.0,  -1.0/5.0,      0.0,       0.0   -1.0/5.0,       0.0 }, // N_RR
      { 5.0/40.0,      0.0, -3.0/40.0,      0.0,       0.0  -2.0/40.0,       0.0 }, // N_RN
      { 6.0/40.0,      0.0, -3.0/40.0,      0.0, -1.0/40.0, -1.0/40.0, -1.0/40.0 }  // N_RD
    },
    {
      { 5.0/40.0,      0.0, -2.0/40.0,      0.0,       0.0, -3.0/40.0,       0.0 }, // N_NR
      { 4.0/40.0,      0.0, -2.0/40.0,      0.0,       0.0, -2.0/40.0,       0.0 }, // N_NN
      { 5.0/40.0,      0.0, -2.0/40.0,      0.0, -1.0/40.0, -1.0/40.0, -1.0/40.0 }  // N_ND
    },
    {
      { 6.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0,     0.0, -3.0/40.0,       0.0 },  // N_DR
      { 5.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0,     0.0, -2.0/40.0,       0.0 },  // N_DN
      { 6.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0 } // N_DD
    }
  }
};
static const FLOAT scppfSharpLoopTangentTable[3][5] = 
{
  { -2.0, 1.0, 1.0, 0.0, 0.0 },
  { -1.0, 0.0, 1.0, 0.0, 0.0 },
  { -2.0, -1.0, 2.0, 2.0, -1.0 }
};

//
// END  : Sharp Loop Tables
//////////////////////////////////////////////////

//////////////////////////////////////////////////
// START: Sharp Catmull-Clark Tables
//

//#define DEBUGGING_SHARP_CATMULL_CLARK_MATH
static const UINT_32 scpuiSharpCatmullClarkSharpTypeIndex[QE::svtSize] = {2, 2, 0, 0, 0, 0};
static const FLOAT scppppfSharpCatmullClarkLimitTable[2][3][3][7] = 
{
  {
    {
      {   4.0/6.0,      0.0,  1.0/6.0,      0.0,      0.0,  1.0/6.0,      0.0 }, // R_RR
      { 33.0/48.0,      0.0, 8.0/48.0,      0.0,      0.0, 7.0/48.0,      0.0 }, // R_RN
      { 31.0/48.0,      0.0, 8.0/48.0,      0.0, 1.0/48.0, 7.0/48.0, 1.0/48.0 }  // R_RD
    },
    {
      { 33.0/48.0,      0.0, 7.0/48.0,      0.0,      0.0, 8.0/48.0,      0.0 }, // R_NR
      { 34.0/48.0,      0.0, 7.0/48.0,      0.0,      0.0, 7.0/48.0,      0.0 }, // R_NN
      { 32.0/48.0,      0.0, 7.0/48.0,      0.0, 1.0/48.0, 7.0/48.0, 1.0/48.0 }  // R_ND
    },
    {
      { 31.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0,      0.0, 8.0/48.0,      0.0 }, // R_DR
      { 32.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0,      0.0, 7.0/48.0,      0.0 }, // R_DN
      { 30.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0, 1.0/48.0, 7.0/48.0, 1.0/48.0 }  // R_DD
    }
  },
  {
    {
      {   3.0/5.0,      0.0,  1.0/5.0,      0.0,      0.0,  1.0/5.0,      0.0 }, // N_RR
      { 25.0/40.0,      0.0, 8.0/40.0,      0.0,      0.0, 7.0/40.0,      0.0 }, // N_RN
      { 24.0/40.0,      0.0, 8.0/40.0,      0.0, 1.0/40.0, 6.0/40.0, 1.0/40.0 }  // N_RD
    },
    {
      { 25.0/40.0,      0.0, 7.0/40.0,      0.0,      0.0, 8.0/40.0,      0.0 }, // N_NR
      { 26.0/40.0,      0.0, 7.0/40.0,      0.0,      0.0, 7.0/40.0,      0.0 }, // N_NN
      { 25.0/40.0,      0.0, 7.0/40.0,      0.0, 1.0/40.0, 6.0/40.0, 1.0/40.0 }  // N_ND
    },
    {
      { 24.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0,      0.0, 8.0/40.0,      0.0 }, // N_DR
      { 25.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0,      0.0, 7.0/40.0,      0.0 }, // N_DN
      { 24.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0, 1.0/40.0, 6.0/40.0, 1.0/40.0 }  // N_DD
    }
  }
};
static const FLOAT scppppfSharpCatmullClarkDeltaTable[2][3][3][7] = 
{
  {
    {
      {  2.0/6.0,      0.0,  -1.0/6.0,      0.0,       0.0   -1.0/6.0,       0.0 }, // R_RR
      { 3.0/48.0,      0.0, -2.0/48.0,      0.0,       0.0  -1.0/48.0,       0.0 }, // R_RN
      { 5.0/48.0,      0.0, -2.0/48.0,      0.0, -1.0/48.0, -1.0/48.0, -1.0/48.0 }  // R_RD
    },
    {
      { 3.0/48.0,      0.0, -1.0/48.0,      0.0,       0.0, -2.0/48.0,       0.0 }, // R_NR
      { 2.0/48.0,      0.0, -1.0/48.0,      0.0,       0.0, -1.0/48.0,       0.0 }, // R_NN
      { 4.0/48.0,      0.0, -1.0/48.0,      0.0, -1.0/48.0, -1.0/48.0, -1.0/48.0 }  // R_ND
    },
    {
      { 5.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0,     0.0, -2.0/48.0,       0.0 },  // R_DR
      { 4.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0,     0.0, -1.0/48.0,       0.0 },  // R_DN
      { 6.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0, -1.0/48.0 } // R_DD
    }
  },
  {
    {
      {  2.0/5.0,      0.0,  -1.0/5.0,      0.0,       0.0   -1.0/5.0,       0.0 }, // N_RR
      { 5.0/40.0,      0.0, -3.0/40.0,      0.0,       0.0  -2.0/40.0,       0.0 }, // N_RN
      { 6.0/40.0,      0.0, -3.0/40.0,      0.0, -1.0/40.0, -1.0/40.0, -1.0/40.0 }  // N_RD
    },
    {
      { 5.0/40.0,      0.0, -2.0/40.0,      0.0,       0.0, -3.0/40.0,       0.0 }, // N_NR
      { 4.0/40.0,      0.0, -2.0/40.0,      0.0,       0.0, -2.0/40.0,       0.0 }, // N_NN
      { 5.0/40.0,      0.0, -2.0/40.0,      0.0, -1.0/40.0, -1.0/40.0, -1.0/40.0 }  // N_ND
    },
    {
      { 6.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0,     0.0, -3.0/40.0,       0.0 },  // N_DR
      { 5.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0,     0.0, -2.0/40.0,       0.0 },  // N_DN
      { 6.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0, -1.0/40.0 } // N_DD
    }
  }
};
static const FLOAT scppfSharpCatmullClarkTangentTable[3][5] = 
{
  { -2.0, 1.0, 1.0, 0.0, 0.0 },
  { -1.0, 0.0, 1.0, 0.0, 0.0 },
  { -2.0, -1.0, 2.0, 2.0, -1.0 }
};

//
// END  : Sharp Catmull-Clark Tables
//////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// QEMesh Class Subdivision Methods
//

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsLoop()
{
  CalculateSubdivisionLimitPositionsAndNormals(stLoop);
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsSharpLoop()
{
  CalculateSubdivisionLimitPositionsAndNormals(stSharpLoop);
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsCatmullClark()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;
  Point pt;
  Vector vNormal;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexLimitSubdivisionCatmullClark(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = CalculateVertexNormalSubdivisionCatmullClark(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
    }
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsCatmullClarkC2()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;
  Point pt;
  Vector vNormal;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexLimitSubdivisionCatmullClarkC2(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = CalculateVertexNormalSubdivisionCatmullClarkC2(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
    }
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsSharpCatmullClark()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;
  Point pt;
  Vector vNormal;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexLimitSubdivisionSharpCatmullClark(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = CalculateVertexNormalSubdivisionSharpCatmullClark(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
    }
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsDooSabin()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;
  Point pt;
  Vector vNormal;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexLimitSubdivisionDooSabin(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = CalculateVertexNormalSubdivisionDooSabin(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
    }
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormalsDefault()
{
  CalculateSubdivisionLimitPositionsAndNormals(stButterfly);
}

VOID QEMesh::CalculateSubdivisionLimitPositionsAndNormals(SubdivisionType eSubdivisionType)
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;
  Point pt;
  Vector vNormal;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexLimitSubdivisionCornerRounding(eSubdivisionType, pvx);
      pvx->SetLimitPosition(pt);
      vNormal = CalculateVertexNormalSubdivisionCornerRounding(eSubdivisionType, pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
    }
}

VOID QEMesh::CalculateSubdivisionSharpVertexTypesSharpLoop()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetSharpVertexType(CalculateSubdivisionSharpVertexTypeSharpLoop(pvx));
    }
}

QEMesh::SharpVertexType QEMesh::CalculateSubdivisionSharpVertexTypeSharpLoop(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCrease;
  UINT_32 uiSharpEdges, uiN;
  SharpVertexType eSharp;

  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return svtCorner;
    }

  iterQEVertex.Init(pvx);
  uiSharpEdges = 0;
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      if ( pe->GetKnotInterval() == INFINITY )
	{
	  uiSharpEdges++;
	  peCrease = pe;
	}
    }
  
  switch ( uiSharpEdges )
    {
    case 0:
      eSharp = svtSmooth;
      break;
    case 1:
      eSharp = svtDart;
      break;
    case 2:
      if ( ( peCrease->GetEdgeNext(1)->GetKnotInterval() == INFINITY &&
	     peCrease->GetFacet() == GetNullFacet() ) ||
	   ( peCrease->GetEdgePrev(1)->GetKnotInterval() == INFINITY &&
	     peCrease->GetEdgePrev(1)->GetFacet() == GetNullFacet() ) )
	{
	  eSharp = svtBoundaryCrease;
	}
      else if ( uiN == 6 && 
	   peCrease->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval() == INFINITY )
	{
	  eSharp = svtRegularCrease;
	}
      else
	{
	  eSharp = svtNonregularCrease;
	}
      break;
    default:
      eSharp = svtCorner;
      break;
    }

  return eSharp;
}

VOID QEMesh::CalculateSubdivisionBoundaryEdgesSharpnessesSharpCatmullClark()
{
  QEMeshIter iterQEMesh; 
  QEEdge *pe;

  iterQEMesh.Init(this);
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      pe->SetKnotInterval(INFINITY);
      pe->GetEdge()->SetKnotInterval(INFINITY);
    }  
}

VOID QEMesh::CalculateSubdivisionSharpVertexTypesSharpCatmullClark()
{
  QEMeshIter iterQEMesh; 
  QEVertex *pvx;

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetSharpVertexType(CalculateSubdivisionSharpVertexTypeSharpCatmullClark(pvx));
    }
}

QEMesh::SharpVertexType QEMesh::CalculateSubdivisionSharpVertexTypeSharpCatmullClark(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  QEEdge *peCrease;
  UINT_32 uiSharpEdges, uiN;
  SharpVertexType eSharp;

  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return svtCorner;
    }

  iterQEVertex.Init(pvx);
  uiSharpEdges = 0;
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      if ( pe->GetKnotInterval() == INFINITY )
	{
	  uiSharpEdges++;
	  peCrease = pe;
	}
    }
  
  switch ( uiSharpEdges )
    {
    case 0:
      eSharp = svtSmooth;
      break;
    case 1:
      eSharp = svtDart;
      break;
    case 2:
      if ( ( peCrease->GetEdgeNext(1)->GetKnotInterval() == INFINITY &&
	     peCrease->GetFacet() == GetNullFacet() ) ||
	   ( peCrease->GetEdgePrev(1)->GetKnotInterval() == INFINITY &&
	     peCrease->GetEdgePrev(1)->GetFacet() == GetNullFacet() ) )
	{
	  eSharp = svtBoundaryCrease;
	}
      else
	{
	  eSharp = svtRegularCrease;
	}
      break;
    default:
      eSharp = svtCorner;
      break;
    }

  return eSharp;
}

///////////////////////////////////////////////////
// QEMesh::SubdivideCornerCutting                //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /    \          //
//           |                 e1      e0        //
//           |               /            \      //
//  e2------vV------e0  => v2      fV      v0    //
//           |               \            /      //
//           |                 e2      e3        //
//           |                   \    /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
QEMesh *QEMesh::SubdivideCornerCutting()
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  // Maximum number of sides on the Facet that this code will handle
  const UINT_32 cuiSidesMax = 20;
  const UINT_32 cuiNewVerticesMax = cuiSidesMax;
  const UINT_32 cuiNewEdgesMax = cuiSidesMax;
  QEVertex *ppvx[cuiNewVerticesMax];
  QEEdge *ppe[cuiNewEdgesMax];
  UINT_32 uiSides;

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEVertex *pvx;
  QEFacet *pfaNew;
  QEEdge *pe;

  UINT_32 i, j;

  Point pt;
  //Point ptFacet;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      memset(ppvx, 0, sizeof(ppvx));
      memset(ppe, 0, sizeof(ppe));

      iterQEVertex.Init(pvx);

      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0, uiSides = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++, uiSides++ )
	{
	  ASSERT( uiSides < cuiSidesMax );

	  if ( pe->GetEdge()->GetForwardEdge() != NULL )
	    {
	      ppvx[i] = pe->GetEdge()->GetForwardEdge()->GetVertex();
	    }
	}

      ASSERT( 2 < uiSides && uiSides <= cuiSidesMax );

      // Create any vertices necessary and all of the edges
      pfaNew = pme->CreateFacet();
      ASSERT( pfaNew != NULL );

      // Create any vertices necessary and all of the edges
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	  if ( ppvx[i] == NULL )
	    {
	      pt = (pe->GetVertex()->GetPoint() + 
	          pe->GetEdge()->GetVertex()->GetPoint()) / 2.0;
	      pt[3] = 1.0;
	      
	      ppvx[i] = pme->CreateVertex(pt);
	      ASSERT( ppvx[i] != NULL );
	    }

	  ppe[i] = pme->CreateEdge();
	  ASSERT( ppe[i] != NULL );
	  ppe[i]->SetFacet(pfaNew);

	  pe->SetForwardEdge(ppe[i]);
	}

      // Link the corner cap together and to adjacent caps if they exist
      j = uiSides-1;
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	  pfaNew = pe->GetFacet()->GetForwardFacet();
	  if ( pfaNew == NULL )
	    {
	      pfaNew = pme->CreateFacet();
	      ASSERT( pfaNew != NULL );
	      pe->GetFacet()->SetForwardFacet(pfaNew);
	    }
	  ppe[i]->GetEdge()->SetFacet(pfaNew);
	  
	  ppe[i]->SetVertex(ppvx[i]);
	  ppe[j]->GetEdge()->SetVertex(ppvx[i]);

	  Splice(1, ppe[i], ppe[j]->GetEdge());
	  if ( pe->GetEdge()->GetForwardEdge() != NULL )
	    {
	      Splice(1, ppe[j]->GetEdge(), pe->GetEdge()->GetForwardEdge()->GetEdgePrev(1));
	    }

	  j = i;
	}
    }

  pme->Preprocess();

  return pme;
}

QEMesh *QEMesh::SubdivideCornerRoundingTesselate()
{
  return SubdivideCornerRounding(stTesselate);
}

QEMesh *QEMesh::SubdivideCornerRounding()
{
  return SubdivideCornerRounding(stCornerRounding);
}

QEMesh *QEMesh::SubdivideCornerRoundingLoop()
{
  return SubdivideCornerRounding(stLoop);
}

QEMesh *QEMesh::SubdivideCornerRoundingSharpLoop()
{
  return SubdivideCornerRoundingSharp(stSharpLoop);
}

QEMesh *QEMesh::SubdivideCornerRoundingButterfly()
{
  return SubdivideCornerRounding(stButterfly);
}

QEMesh *QEMesh::SubdivideCornerRoundingZorin()
{
  return SubdivideCornerRounding(stZorin);
}

///////////////////////////////////////////////////
// QEMesh::SubdivideCornerRounding               //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0  => v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
QEMesh *QEMesh::SubdivideCornerRounding(SubdivisionType eSubdivisionType)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  // Maximum number of sides on the Facet that this code will handle
  const UINT_32 cuiSidesMax = 20;
  const UINT_32 cuiNewVerticesMax = cuiSidesMax;
  const UINT_32 cuiNewEdgesMax = 2*cuiSidesMax;
  QEVertex *ppvx[cuiNewVerticesMax];
  QEEdge *ppe[cuiNewEdgesMax];
  UINT_32 uiSides;

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEVertex *pvx, *pvxVertex;
  QEFacet *pfaNew;
  QEEdge *pe;

  UINT_32 i, j;

  Point pt;
  //Point ptFacet;
  //Vector vNormal;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      memset(ppvx, 0, sizeof(ppvx));
      memset(ppe, 0, sizeof(ppe));

      // Create the Vertex Vertex
      pt = CalculateVertexSubdivisionCornerRounding(eSubdivisionType, pvx);
      pvxVertex = pme->CreateVertex(pt);
      ASSERT( pvxVertex != NULL );

      iterQEVertex.Init(pvx);

      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0, uiSides = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++, uiSides++ )
	{
	  ASSERT( uiSides < cuiSidesMax );

	  if ( pe->GetEdge()->GetForwardEdge() != NULL )
	    {
	      ppvx[i] = pe->GetEdge()->GetForwardEdge()->GetEdge()->GetVertex();
	    }
	}

      ASSERT( 2 < uiSides && uiSides <= cuiSidesMax );

      // Create any vertices necessary and all of the edges
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	  if ( ppvx[i] == NULL )
	    {
	      pt = CalculateEdgeSubdivisionCornerRounding(eSubdivisionType, pe);
	      pt[3] = 1.0;
	      
	      ppvx[i] = pme->CreateVertex(pt);
	      ASSERT( ppvx[i] != NULL );
	    }

	  ppe[i] = pme->CreateEdge();
	  ASSERT( ppe[i] != NULL );
	  ppe[uiSides+i] = pme->CreateEdge();
	  ASSERT( ppe[uiSides+i] != NULL );

	  pe->SetForwardEdge(ppe[i]);
	}

      // Link the corner cap together and to adjacent caps if they exist
      j = uiSides-1;
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );

	  ppe[i]->GetEdge()->SetFacet(pfaNew);
	  ppe[j]->SetFacet(pfaNew);
	  ppe[uiSides+j]->SetFacet(pfaNew);
	  pfaNew = pe->GetFacet()->GetForwardFacet();
	  if ( pfaNew == NULL )
	    {
	      pfaNew = pme->CreateFacet();
	      ASSERT( pfaNew != NULL );
	      pe->GetFacet()->SetForwardFacet(pfaNew);
	    }
	  ppe[uiSides+i]->GetEdge()->SetFacet(pfaNew);
	  
	  ppe[i]->SetVertex(pvxVertex);
	  ppe[i]->GetEdge()->SetVertex(ppvx[i]);
	  ppe[uiSides+i]->SetVertex(ppvx[i]);
	  ppe[uiSides+j]->GetEdge()->SetVertex(ppvx[i]);

	  if ( i != 0 )
	    {
	      Splice(1, ppe[j], ppe[i]);
	    }
	  Splice(1, ppe[i]->GetEdge(), ppe[uiSides+i]);
	  Splice(1, ppe[i]->GetEdge(), ppe[uiSides+j]->GetEdge());
	  if ( pe->GetEdge()->GetForwardEdge() != NULL )
	    {
	      Splice(1, ppe[uiSides+j]->GetEdge(), pe->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgeNext(1));
	    }

	  j = i;
	}
    }

  pme->Preprocess();

  pme->CalculateSubdivisionLimitPositionsAndNormalsLoop();

  return pme;
}

//#define BOUNDARY_EDGES_NOT_OK
#ifndef BOUNDARY_EDGES_NOT_OK

///////////////////////////////////////////////////
// QEMesh::SubdivideCornerRoundingSharp          //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0  => v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
QEMesh *QEMesh::SubdivideCornerRoundingSharp(SubdivisionType eSubdivisionType)
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxNew;
  QEFacet *pfa, *pfaNew, *pfaCenter;
  QEEdge *pe, *peNew;
  Point pt;
  FLOAT fSharpness;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      // Create the Vertex Vertex
      pt = CalculateVertexSubdivisionCornerRounding(eSubdivisionType, pvx);
      pvxNew = pme->CreateVertex(pt);
      ASSERT( pvxNew != NULL );
      pvxNew->SetSharpVertexType(pvx->GetSharpVertexType());
      pvx->SetForwardVertex(pvxNew);

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  ASSERT( pe->GetForwardEdge() == NULL );

	  ASSERT( pe->GetKnotInterval() == pe->GetEdge()->GetKnotInterval() );
	  fSharpness = pe->GetKnotInterval();
	  if ( fSharpness != INFINITY )
	    {
	      fSharpness = MAX(0.0, fSharpness - 1.0);
	    }

	  peNew = pme->CreateEdge();
	  peNew->SetFacet(GetNullFacet());
	  peNew->GetEdge()->SetFacet(GetNullFacet());
	  peNew->SetVertex(pvx->GetForwardVertex());
	  peNew->SetKnotInterval(fSharpness);
	  peNew->GetEdge()->SetKnotInterval(fSharpness);
	  pe->SetForwardEdge(peNew);

	  if ( pe->GetEdgePrev(1)->GetForwardEdge() != NULL )
	    {
	      Splice(1, pe->GetEdgePrev(1)->GetForwardEdge(), pe->GetForwardEdge());
	    }

	  if ( pe->GetEdge()->GetForwardEdge() == NULL )
	    {
	      pt = CalculateEdgeSubdivisionCornerRounding(eSubdivisionType, pe);
	      pt[3] = 1.0;
	      pvxNew = pme->CreateVertex(pt);
	      ASSERT( pvxNew != NULL );
	      if ( pe->GetFacet() == GetNullFacet() ||
		   pe->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  pvxNew->SetSharpVertexType(svtBoundaryCrease);
		}
	      else if ( pe->GetKnotInterval() == INFINITY )
		{
		  pvxNew->SetSharpVertexType(svtRegularCrease);
		}
	      else
		{
		  pvxNew->SetSharpVertexType(svtSmooth);
		}
	      pe->GetForwardEdge()->GetEdge()->SetVertex(pvxNew);
	    }
	  else
	    {
	      peNew->GetEdge()->SetVertex(pe->GetEdge()->GetForwardEdge()->GetEdge()->GetVertex());	      
	      Splice(1, peNew->GetEdge(), pe->GetEdge()->GetForwardEdge()->GetEdge());
	    }
	}
    }

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      // Create the facet-facet
      pfaCenter = pme->CreateFacet();

      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  ASSERT( pe->GetForwardEdge() != NULL );

	  // Cut off the corner
	  pfaNew = pme->CreateFacet();

	  peNew = pme->CreateEdge();
	  peNew->SetKnotInterval(0.0);
	  peNew->GetEdge()->SetKnotInterval(0.0);
	  peNew->SetFacet(pfaNew);
	  peNew->GetEdge()->SetFacet(pfaCenter);

	  pe->GetForwardEdge()->SetFacet(pfaNew);
	  pe->GetForwardEdge()->GetEdgePrev(0)->SetFacet(pfaNew);

	  peNew->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetForwardEdge()->GetEdgePrev(0)->GetVertex());

	  Splice(1, peNew, pe->GetForwardEdge()->GetEdge()->GetEdgePrev(1));
	  Splice(1, peNew->GetEdge(), pe->GetForwardEdge()->GetEdgePrev(0));
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      ASSERT( pe->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge()->GetFacet() != pme->GetNullFacet() );

      pe->GetForwardEdge()->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge());

      pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge()->GetEdgeNext(0));
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge()->GetEdgeNext(0));
    }

  pme->Preprocess();
  pme->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();

  return pme;
}

#else // BOUNDARY_EDGES_NOT_OK

///////////////////////////////////////////////////
// QEMesh::SubdivideCornerRoundingSharp          //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0  => v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
QEMesh *QEMesh::SubdivideCornerRoundingSharp(SubdivisionType eSubdivisionType)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  // Maximum number of sides on the Facet that this code will handle
  const UINT_32 cuiSidesMax = 20;
  const UINT_32 cuiNewVerticesMax = cuiSidesMax;
  const UINT_32 cuiNewEdgesMax = 2*cuiSidesMax;
  QEVertex *ppvx[cuiNewVerticesMax];
  QEEdge *ppe[cuiNewEdgesMax];
  UINT_32 uiSides;

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEVertex *pvx, *pvxVertex;
  QEFacet *pfaNew;
  QEEdge *pe;

  UINT_32 i, j;

  Point pt;
  //Point ptFacet;
  //Vector vNormal;

  FLOAT fSharpness;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      memset(ppvx, 0, sizeof(ppvx));
      memset(ppe, 0, sizeof(ppe));

      // Create the Vertex Vertex
      pt = CalculateVertexSubdivisionCornerRounding(eSubdivisionType, pvx);
      pvxVertex = pme->CreateVertex(pt);
      ASSERT( pvxVertex != NULL );
      pvxVertex->SetSharpVertexType(pvx->GetSharpVertexType());

      iterQEVertex.Init(pvx);

      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0, uiSides = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++, uiSides++ )
	{
	  ASSERT( uiSides < cuiSidesMax );

	  if ( pe->GetEdge()->GetForwardEdge() != NULL )
	    {
	      ppvx[i] = pe->GetEdge()->GetForwardEdge()->GetEdge()->GetVertex();
	    }
	}

      ASSERT( 2 < uiSides && uiSides <= cuiSidesMax );

      // Create any vertices necessary and all of the edges
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	  if ( ppvx[i] == NULL )
	    {
	      pt = CalculateEdgeSubdivisionCornerRounding(eSubdivisionType, pe);
	      pt[3] = 1.0;
	      
	      ppvx[i] = pme->CreateVertex(pt);
	      ASSERT( ppvx[i] != NULL );
	      if ( pe->GetFacet() == GetNullFacet() ||
		   pe->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  pvxNew->SetSharpVertexType(svtBoundaryCrease);
		}
	      else if ( pe->GetKnotInterval() == INFINITY )
		{
		  ppvx[i]->SetSharpVertexType(svtRegularCrease);
		}
	      else
		{
		  ppvx[i]->SetSharpVertexType(svtSmooth);
		}
	    }

	  ASSERT( pe->GetKnotInterval() == pe->GetEdge()->GetKnotInterval() );
	  fSharpness = pe->GetKnotInterval();
	  if ( fSharpness != INFINITY )
	    {
	      fSharpness = MAX(0.0, fSharpness - 1.0);
	    }

	  ppe[i] = pme->CreateEdge();
	  ASSERT( ppe[i] != NULL );
	  ppe[i]->SetKnotInterval(fSharpness);
	  ppe[i]->GetEdge()->SetKnotInterval(fSharpness);
	  ppe[uiSides+i] = pme->CreateEdge();
	  ASSERT( ppe[uiSides+i] != NULL );
	  ppe[uiSides+i]->SetKnotInterval(0.0);
	  ppe[uiSides+i]->GetEdge()->SetKnotInterval(0.0);

	  pe->SetForwardEdge(ppe[i]);
	}

      // Link the corner cap together and to adjacent caps if they exist
      j = uiSides-1;
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );

	  ppe[i]->GetEdge()->SetFacet(pfaNew);
	  ppe[j]->SetFacet(pfaNew);
	  ppe[uiSides+j]->SetFacet(pfaNew);
	  pfaNew = pe->GetFacet()->GetForwardFacet();
	  if ( pfaNew == NULL )
	    {
	      pfaNew = pme->CreateFacet();
	      ASSERT( pfaNew != NULL );
	      pe->GetFacet()->SetForwardFacet(pfaNew);
	    }
	  ppe[uiSides+i]->GetEdge()->SetFacet(pfaNew);
	  
	  ppe[i]->SetVertex(pvxVertex);
	  ppe[i]->GetEdge()->SetVertex(ppvx[i]);
	  ppe[uiSides+i]->SetVertex(ppvx[i]);
	  ppe[uiSides+j]->GetEdge()->SetVertex(ppvx[i]);

	  if ( i != 0 )
	    {
	      Splice(1, ppe[j], ppe[i]);
	    }
	  Splice(1, ppe[i]->GetEdge(), ppe[uiSides+i]);
	  Splice(1, ppe[i]->GetEdge(), ppe[uiSides+j]->GetEdge());
	  if ( pe->GetEdge()->GetForwardEdge() != NULL )
	    {
	      Splice(1, ppe[uiSides+j]->GetEdge(), pe->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgeNext(1));
	    }

	  j = i;
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      ASSERT( pe->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge()->GetFacet() != pme->GetNullFacet() );

      pe->GetForwardEdge()->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge());

      pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge()->GetEdgeNext(0));
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge()->GetEdgeNext(0));
    }

  pme->Preprocess();
  pme->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();

  return pme;
}

#endif // BOUNDARY_EDGES_NOT_OK

Point QEMesh::CalculateVertexSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateVertexSubdivisionTesselate(pvx);
      break;
    case stLoop:
      pt = CalculateVertexSubdivisionLoop(pvx);
      break;
    case stSharpLoop:
      pt = CalculateVertexSubdivisionSharpLoop(pvx);
      break;
    case stCornerRounding:
      // Experimental
      pt = CalculateVertexSubdivisionCornerRounding(pvx);
      break;
    case stButterfly:
      pt = CalculateVertexSubdivisionCornerRoundingButterfly(pvx);
      break;
    case stZorin:
      pt = CalculateVertexSubdivisionCornerRoundingZorin(pvx);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stLoop:
      pt = CalculateVertexLimitSubdivisionLoop(pvx);
      break;
    case stSharpLoop:
      pt = CalculateVertexLimitSubdivisionSharpLoop(pvx);
      break;
    case stTesselate:
    case stCornerRounding:
    case stButterfly:
    case stZorin:
      pt = pvx->GetPoint();
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Vector vNormal;

  switch ( eSubdivisionType )
    {
    case stLoop:
      vNormal = CalculateVertexNormalSubdivisionLoop(pvx);
      break;
    case stSharpLoop:
      vNormal = CalculateVertexNormalSubdivisionSharpLoop(pvx);
      break;
    case stTesselate:
    case stCornerRounding:
    case stButterfly:
    case stZorin:
      vNormal.Set(0.0, 0.0, 1.0);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionCornerRounding(SubdivisionType eSubdivisionType, QEEdge *pe)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateEdgeSubdivisionTesselate(pe);
      break;
    case stLoop:
      pt = CalculateEdgeSubdivisionLoop(pe);
      break;
    case stCornerRounding:
      // Experimental
      pt = CalculateEdgeSubdivisionCornerRounding(pe);
      break;
    case stSharpLoop:
      pt = CalculateEdgeSubdivisionSharpLoop(pe);
      break;
    case stButterfly:
      pt = CalculateEdgeSubdivisionCornerRoundingButterfly(pe);
      break;
    case stZorin:
      pt = CalculateEdgeSubdivisionCornerRoundingZorin(pe);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCornerRoundingButterfly(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionCornerRoundingButterfly(QEEdge *pe)
{
  Point pt, ppt[10];
  UINT_32 uiSide;
  FLOAT fT;

  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  for ( uiSide = 0; uiSide < 2; uiSide++ )
    {
      ppt[uiSide] = pe->GetVertex()->GetPoint();

      ppt[uiSide+2] = pe->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();

      ppt[uiSide+4] = pe->GetEdge()->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();

      ppt[uiSide+6] = pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();
      ppt[uiSide+8] = pe->GetEdge()->GetEdgePrev(1)->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();

      pe = pe->GetEdge();
    }

  fT = 1.0 / 8.0;
  //fT = 0.0;

  pt = ((ppt[0] + ppt[1]) + fT * ( (ppt[2] + ppt[3] + ppt[4] + ppt[5]) - (ppt[6] + ppt[7] + ppt[8] + ppt[9]))) / 2.0;
  pt[3] = 1.0;
  
  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCornerRoundingZorin(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionCornerRoundingZorin(QEEdge *pe)
{
  Point pt, ptOrigin;
  QEVertexIter iterQEVertex;
  QEEdge *peCurr;
  QEVertex *ppvx[2];
  Point ppt[2];
  Vector pv[2], v;
  UINT_32 i, j, puiNeighbors[2];
  FLOAT fS;
  
  for ( i = 0; i < 2; i++ )
    {
      ppvx[i] = pe->GetVertex();

      // Count the number of neighbors around the vertex
      iterQEVertex.Init(ppvx[i]);
      puiNeighbors[i] = 0;
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  puiNeighbors[i]++;
	}

      // Calculate the weight point
      ptOrigin = ppvx[i]->GetPoint();
      pv[i].Null();
      switch ( puiNeighbors[i] )
	{
	case 0:
	case 1:
	case 2:
	  ASSERT( FALSE );
	  break;
	case 3:
	  {
	    FLOAT pfWeight[3] = { 5.0 / 12.0, -1.0 / 12.0, -1.0 / 12.0 };

	    peCurr = pe;
	    j = 0;
	    do
	      {
		ASSERT( j < 3 );
		pt = peCurr->GetEdge()->GetVertex()->GetPoint();

		v = pt - ptOrigin;
		pv[i] = pv[i] + pfWeight[j] * v;
		
		// Advance to the next edge
		peCurr = peCurr->GetEdgeNext(1);

		j++;
	      }
	    while ( peCurr != pe );
	  }
	  break;
	case 4:
	  {
	    FLOAT pfWeight[4] = { 3.0 / 8.0, 0.0, -1.0 / 8.0, 0.0 };

	    peCurr = pe;
	    j = 0;
	    do
	      {
		ASSERT( j < 4 );
		pt = peCurr->GetEdge()->GetVertex()->GetPoint();
		
		v = pt - ptOrigin;
		pv[i] = pv[i] + pfWeight[j] * v;
		
		// Advance to the next edge
		peCurr = peCurr->GetEdgeNext(1);

		j++;
	      }
	    while ( peCurr != pe );
	  }
          break;	  
	case 6:
          break;	  
	default:
	  {
	    peCurr = pe;
	    j = 0;
	    do
	      {
		pt = peCurr->GetEdge()->GetVertex()->GetPoint();
		fS = (0.25 + cos(2.0 * PI * j / puiNeighbors[i]) + 0.5 * + cos(4.0 * PI * j / puiNeighbors[i])) / puiNeighbors[i];
		
		v = pt - ptOrigin;
		pv[i] = pv[i] + fS * v;
		
		// Advance to the next edge
		peCurr = peCurr->GetEdgeNext(1);

		j++;
	      }
	    while ( peCurr != pe );
	  }
	  break;
	}

      ppt[i] = ptOrigin + pv[i];

      pe = pe->GetEdge();
    }

  if ( puiNeighbors[0] == 6 && puiNeighbors[1] == 6 )
    {
      pt = CalculateEdgeSubdivisionCornerRoundingButterfly(pe);
    }
  else if ( puiNeighbors[1] == 6 )
    {
      pt = ppt[0];
    }
  else if ( puiNeighbors[0] == 6 )
    {
      pt = ppt[1];
    }
  else
    {
      pt = (ppt[0] + ppt[1]) / 2.0;
    }
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCornerRounding(QEVertex *pvx)
{
  /*
  // Normal Loop Rule
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fAlpha;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
    }
  
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  pt = (pt + (fAlpha * pvx->GetPoint())) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;
  */

  // Eigenvalue based Loop Rule
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fAlpha;
  //const UINT_32 cuiMax = 15;
  //const FLOAT cpfBeta[cuiMax] = {0.0, 0.0, 0.5625, 0.5000, 0.4523, 0.3750, 0.3056, 0.2500, 0.2066, 0.1727, 0.1461, 0.1250, 0.1080, 0.0941, 0.0827};

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
    }
  
  /*
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  if ( uiNeighbors <= cuiMax )
    {
      FLOAT fANew;
      fANew = cpfBeta[uiNeighbors-1];
      fA = fANew;
    }
  */
  if ( uiNeighbors < 5 )
    {
      fA = 0.5;
    }
  else
    {
      fA = (1.0 - cos(4.0*PI/(FLOAT)uiNeighbors))/4.0;
    }
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  pt = (pt + (fAlpha * pvx->GetPoint())) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;

/*
  // Generalized Loop Rule
  Point pt, ptF;
  QEVertexIter iter;
  QEEdge *pe, *peCurr;
  UINT_32 uiNeighbors, uiSides;
  FLOAT fBeta, fF;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  fF = 0.0;
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      uiSides = 0;
      ptF.Null();
      peCurr = pe;
      do
	{
	  uiSides++;
	  ptF += peCurr->GetVertex()->GetPoint();
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
      fBeta = cos(PI/(FLOAT)uiSides);
      pt += (fBeta*ptF);
      fF += fBeta;
    }
  
  fF = 1.0 + fF / (FLOAT)uiNeighbors;
  fBeta = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fBeta = fBeta * fBeta;
  fBeta = (5.0 / 8.0) - (fBeta / 64.0);
  //fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fBeta) / fBeta;
  //pt = (pt + (fAlpha * pvx->GetPoint())) / (fAlpha + (FLOAT)uiNeighbors);
  pt = ((1.0 - fBeta * fF) * pvx->GetPoint()) + ((fBeta / (FLOAT)uiNeighbors) * pt);
  pt[3] = 1.0;

  return pt;
  */
}

Point QEMesh::CalculateEdgeSubdivisionCornerRounding(QEEdge *pe)
{
  // Normal Loop Rule
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN;

  uiN = 0;
  pt = pe->GetVertex()->GetPoint();
  uiN++;
  pt += pe->GetEdge()->GetVertex()->GetPoint();
  uiN++;

  iterQEFacet.Init(pe->GetFacet());
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      pt += pvx->GetPoint();
      uiN++;
    }

  iterQEFacet.Init(pe->GetEdge()->GetFacet());
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      pt += pvx->GetPoint();
      uiN++;
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  /*
  // Generalized Loop Rule  
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt, ptCentroid;
  UINT_32 uiN;
  FLOAT fDen;

  pt = pe->GetVertex()->GetPoint();
  pt += pe->GetEdge()->GetVertex()->GetPoint();
  fDen = 2.0;

  ptCentroid.Null();
  uiN = 0;
  iterQEFacet.Init(pe->GetFacet());
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      ptCentroid += pvx->GetPoint();
      uiN++;
    }
  fDen += (FLOAT)uiN;
  pt += ptCentroid;

  ptCentroid.Null();
  uiN = 0;
  iterQEFacet.Init(pe->GetEdge()->GetFacet());
  for ( pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex() )
    {
      ptCentroid += pvx->GetPoint();
      uiN++;
    }
  fDen += (FLOAT)uiN;
  pt += ptCentroid;

  pt /= fDen;
  pt[3] = 1.0;
  */

  /*
  // Put it back to this one
  Point pt;
  Point ppt[4];
  QEEdge *peCurr;
  FLOAT pfWeights[4];
  UINT_32 uiN0, uiN1;

  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

  uiN0 = 0;
  ppt[2].Null();
  peCurr = pe;
  do
    {
      ppt[2] += peCurr->GetVertex()->GetPoint();
      uiN0++;
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe );
  ppt[2] /= (FLOAT)uiN0;

  uiN1 = 0;
  ppt[3].Null();
  peCurr = pe->GetEdge();
  do
    {
      ppt[3] += peCurr->GetVertex()->GetPoint();
      uiN1++;
      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe->GetEdge() );
  ppt[3] /= (FLOAT)uiN1;

  pfWeights[0] = 1.0;
  pfWeights[1] = 1.0;
  // Length Based
  pfWeights[2] = 1.0 / (2.0 * cos(PI / (FLOAT)uiN0));
  pfWeights[3] = 1.0 / (2.0 * cos(PI / (FLOAT)uiN1));

  // Area Based
  //pfWeights[2] = cos(PI / (FLOAT)uiN0);
  //pfWeights[2] = 1.0 / (4.0 * pfWeights[2] * pfWeights[2]);
  //pfWeights[3] = cos(PI / (FLOAT)uiN1);
  //pfWeights[3] = 1.0 / (4.0 * pfWeights[3] * pfWeights[3]);

  pt = ( (pfWeights[0]*ppt[0]) + (pfWeights[1]*ppt[1]) + (pfWeights[2]*ppt[2]) + (pfWeights[3]*ppt[3]) ) / (pfWeights[0] + pfWeights[1] + pfWeights[2] + pfWeights[3]);
  pt[3] = 1.0;
*/

  return pt;
}

///////////////////////////////////////////////////
// QEMesh::SubdivideTriangulate                  //
//                                               //
//       Old Quad             Triangulated       //
//                                               //
//  v3------e2------v2     v3------e2------v2    //
//   |               |      | \           / |    //
//   |               |      |  e7      e6   |    //
//   |               |      |     \   /     |    //
//  e3      fF      e1  => e3      vF      e1    //
//   |               |      |     /   \     |    //
//   |               |      |  e4      e5   |    //
//   |               |      | /           \ |    //
//  v0------e0------v1     v0------e0------v1    //
//                                               //
QEMesh *QEMesh::SubdivideTriangulate()
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *peCurr, *peNext, *peNew;
  QEVertex *pvxNew;

  Point pt;
  UINT_32 uiSides;

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      iterQEFacet.Init(pfa);
      uiSides = 0;
      pt.Null();
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  pt = pt + pe->GetVertex()->GetPoint();
	  uiSides++;
	}
      ASSERT( uiSides > 2 );
      if ( uiSides > 3 )
	{
	  pt /= (FLOAT)uiSides;
	  pt[3] = 1.0;
	  pvxNew = pme->CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  
	  pe = pfa->GetEdge();
	  ASSERT( pe != NULL );
	  peCurr = pe;
	  peNext = peCurr->GetEdgeNext(0);

	  peNew = pme->CreateEdge();
	  ASSERT( peNew != NULL );

	  peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);

	  peNew->SetFacet(peCurr->GetFacet());

	  Splice(1, peNew, peCurr->GetEdge()->GetEdgePrev(1));

	  peCurr = peNext;

	  do
	    {
	      peNext = peCurr->GetEdgeNext(0);

	      pfaNew = pme->CreateFacet();
	      ASSERT( pfaNew != NULL );

	      peNew = pme->CreateEdge();
	      ASSERT( peNew != NULL );

	      peNew->SetVertex(peCurr->GetEdge()->GetVertex());
	      peNew->GetEdge()->SetVertex(pvxNew);
	      
	      peNew->SetFacet(pfaNew);
	      peCurr->GetEdgePrev(0)->SetFacet(pfaNew);
	      peCurr->SetFacet(pfaNew);
	      
	      Splice(1, peNew->GetEdge(), peCurr->GetEdgePrev(0));
	      Splice(1, peNew, peCurr->GetEdge()->GetEdgePrev(1));

	      peCurr = peNext;
	    }
	  while ( pe != peCurr );
	  pe->GetEdgePrev(0)->SetFacet(pe->GetFacet());
	}
    }

  pme->Preprocess();

  return pme;
}

QEMesh *QEMesh::SubdivideTriangleTesselate()
{
  return SubdivideTriangle(stTesselate);
}

QEMesh *QEMesh::SubdivideTriangleLoop()
{
  return SubdivideTriangle(stLoop);
}

QEMesh *QEMesh::SubdivideTriangleLoopEdgeRule()
{
  return SubdivideTriangle(stLoopEdgeRule);
}

QEMesh *QEMesh::SubdivideTriangleButterfly()
{
  return SubdivideTriangle(stButterfly);
}

QEMesh *QEMesh::SubdivideTriangleZorin()
{
  return SubdivideTriangle(stZorin);
}

QEMesh *QEMesh::SubdivideTriangleSphere()
{
  return SubdivideTriangle(stSphere);
}

///////////////////////////////////////////////////
// QEMesh::SubdivideTriangle                     //
//                                               //
//      Old Triangle       Subdivided Triangles  //
//                                               //
//          v2                     v4            //
//         /  \                   /  \           //
//        /    \                 e4  e3          //
//       /      \               /      \         //
//      e2      e1      =>     v5--e8--v3        //
//     /          \           /  \    /  \       //
//    /            \         e5  e6  e7  e2      //
//   /              \       /      \/      \     //
//  v0------e0------v1     v0--e0--v1--e1--v2    //
//                                               //
QEMesh *QEMesh::SubdivideTriangle(SubdivisionType eSubdivisionType)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaNew, *pfaCenter;
  QEVertex *ppvx[6], *pvx;
  QEEdge *ppe[9], *pe;
  Point pt;
  UINT_32 i, j, k;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      memset(ppvx, 0, sizeof(ppvx));
      memset(ppe, 0, sizeof(ppe));

      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge(), i += 2 )
	{
	  ppe[i] = pe->GetForwardEdge();
	  ppe[i+1] = pe->GetEdge()->GetForwardEdge();
	  ASSERT( (ppe[i] == NULL && ppe[i+1] == NULL) ||
		  (ppe[i] != NULL && ppe[i+1] != NULL) );
	  if ( ppe[i] != NULL )
	    {
	      ppvx[i] = ppe[i]->GetVertex();
	      ppvx[i+1] = ppe[i]->GetEdge()->GetVertex();
	      if ( i == 4 )
		{
		  j = 0;
		}
	      else
		{
		  j = i + 2;
		}
	      ppvx[j] = ppe[i+1]->GetVertex();
	    }
	}

      for ( pe = iterQEFacet.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge(), i+= 2 )
	{
	  ASSERT( (ppe[i] == NULL && ppe[i+1] == NULL) ||
		  (ppe[i] != NULL && ppe[i+1] != NULL) );
	  if ( ppe[i] == NULL )
	    {
	      if ( ppvx[i] == NULL )
		{
		  pvx = pe->GetVertex();
		  ppvx[i] = pvx->GetForwardVertex();
		  if ( ppvx[i] == NULL )
		    {
		      pt = CalculateVertexSubdivisionTriangle(eSubdivisionType, pvx);

		      ppvx[i] = pme->CreateVertex(pt);
		      ASSERT( ppvx[i] != NULL );
		      pvx->SetForwardVertex(ppvx[i]);
		      pvx->SetNormal(ppvx[i]->GetNormal());
		    }
		}

	      if ( i == 4 )
		{
		  j = 0;
		}
	      else
		{
		  j = i + 2;
		}
	      if ( ppvx[j] == NULL )
		{
		  pvx = pe->GetEdge()->GetVertex();
		  ppvx[j] = pvx->GetForwardVertex();
		  if ( ppvx[j] == NULL )
		    {
		      pt = CalculateVertexSubdivisionTriangle(eSubdivisionType, pvx);

		      ppvx[j] = pme->CreateVertex(pt);
		      ASSERT( ppvx[j] != NULL );
		      pvx->SetForwardVertex(ppvx[j]);
		      pvx->SetNormal(ppvx[j]->GetNormal());
		    }
		}

	      // Create the midpoint
	      pt = CalculateEdgeSubdivisionTriangle(eSubdivisionType, pe);
	      ASSERT( ppvx[i+1] == NULL );
	      ppvx[i+1] = pme->CreateVertex(pt);
	      ASSERT( ppvx[i+1] != NULL );

	      {
		// Better Normals
		Vector v;
		v = ppvx[i]->GetNormal() + ppvx[j]->GetNormal();
		v.Normalize();
		ppvx[i+1]->SetNormal(v);
	      }

	      // Create the two subdivided edges
	      ppe[i] = pme->CreateEdge();
	      ASSERT( ppe[i] != NULL );
	      pe->SetForwardEdge(ppe[i]);

	      ppe[i+1] = pme->CreateEdge();
	      ASSERT( ppe[i+1] != NULL );
	      pe->GetEdge()->SetForwardEdge(ppe[i+1]);

	      ppe[i]->SetVertex(ppvx[i]);
	      ppe[i]->GetEdge()->SetVertex(ppvx[i+1]);
	      ppe[i+1]->GetEdge()->SetVertex(ppvx[i+1]);
	      ppe[i+1]->SetVertex(ppvx[j]);

	      Splice(1, ppe[i]->GetEdge(), ppe[i+1]->GetEdge());
	    }
	}

      pfaCenter = pme->CreateFacet();
      ASSERT( pfaCenter != NULL );

      j = 5;
      for ( k = 0, i = 0; k < 3; k++, i += 2 )
	{
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );
	  
	  if ( ppe[j]->GetEdgePrev(1) != ppe[i] )
	    {
	      Splice(1, ppe[i], ppe[j]->GetEdgePrev(1));
	    }

	  ppe[6+k] = pme->CreateEdge();
	  ASSERT( ppe[6+k] != NULL );
	  ppe[6+k]->SetVertex(ppe[j]->GetEdge()->GetVertex());
	  ppe[6+k]->GetEdge()->SetVertex(ppe[i]->GetEdge()->GetVertex());

	  Splice(1, ppe[6+k], ppe[j]->GetEdge());
	  Splice(1, ppe[6+k]->GetEdge(), ppe[i]->GetEdge()->GetEdgePrev(1));

	  ppe[i]->SetFacet(pfaNew);
	  ppe[j]->GetEdge()->SetFacet(pfaNew);
	  ppe[6+k]->SetFacet(pfaCenter);
	  ppe[6+k]->GetEdge()->SetFacet(pfaNew);

	  j = i + 1;
	}      
    }

  pme->Preprocess();

  return pme;
}

Point QEMesh::CalculateVertexSubdivisionTriangle(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateVertexSubdivisionTesselate(pvx);
      break;
    case stLoop:
      pt = CalculateVertexSubdivisionLoop(pvx);
      break;
    case stLoopEdgeRule:
      pt = CalculateVertexSubdivisionLoopEdgeRule(pvx);
      break;
    case stButterfly:
      pt = CalculateVertexSubdivisionButterfly(pvx);
      break;
    case stZorin:
      pt = CalculateVertexSubdivisionZorin(pvx);
      break;
    case stSphere:
      //pt = CalculateVertexSubdivisionLoopExperimental(pvx);
      pt = CalculateVertexSubdivisionHermite(pvx);
      //pt = CalculateVertexSubdivisionSphere(pvx);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionTriangle(SubdivisionType eSubdivisionType, QEEdge *pe)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateEdgeSubdivisionTesselate(pe);
      break;
    case stLoop:
      pt = CalculateEdgeSubdivisionLoop(pe);
      break;
    case stLoopEdgeRule:
      pt = CalculateEdgeSubdivisionLoopEdgeRule(pe);
      break;
    case stButterfly:
      pt = CalculateEdgeSubdivisionButterfly(pe);
      break;
    case stZorin:
      pt = CalculateEdgeSubdivisionZorin(pe);
      break;
    case stSphere:
      //pt = CalculateEdgeSubdivisionLoopExperimental(pe);
      pt = CalculateEdgeSubdivisionHermite(pe);
      //pt = CalculateEdgeSubdivisionSphere(pe);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionTesselate(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN;

  pt.Null();

  iterQEFacet.Init(pfa);
  for ( pvx = iterQEFacet.PeekFirstVertex(), uiN = 0;
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex(), uiN++ )
    {
      pt += pvx->GetPoint();
    }

  pt /= uiN;

  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionTesselate(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionTesselate(QEEdge *pe)
{
  Point pt;

  pt = (pe->GetVertex()->GetPoint() + pe->GetEdge()->GetVertex()->GetPoint()) / 2.0;
  pt[3] = 1.0;
  
  return pt;
}

Point QEMesh::CalculateVertexSubdivisionLoop(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fAlpha;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
    }
  
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  pt = (pt + (fAlpha * pvx->GetPoint())) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionLoop(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fEpsilon;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
    }
  
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiNeighbors) / (8.0 * fA);
  pt = (pt + (fEpsilon * pvx->GetPoint())) / (fEpsilon + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionLoop(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, v;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fTheta, fThetaDelta;

  iter.Init(pvx);

  // Loop Vertex-Normal Subdivision
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
    }

  vT0.Null();
  vT1.Null();
  fTheta = 0.0;
  fThetaDelta = (2.0 * PI) / (FLOAT)uiNeighbors;
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      vT0 += cos(fTheta) * v;
      fTheta += fThetaDelta;
      vT1 += cos(fTheta) * v;
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;

  vNormal = vT1 * vT0;
  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionLoop(QEEdge *pe)
{
  /*  
  Point pt;
  Point ppt[4];

  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint();
  ppt[3] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();

  pt = ((3.0 * ppt[0]) + (3.0 * ppt[1]) + ppt[2] + ppt[3]) / 8.0;
  pt[3] = 1.0;
  */
  Point pt;
  Point ppt[4];
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint() + 
	    pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
  ppt[3] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
	    pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();

  pt = ((6.0 * ppt[0]) + (6.0 * ppt[1]) + ppt[2] + ppt[3]) / 16.0;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSharpLoop(QEVertex *pvx)
{
  Point pt, ptCorner, ptCrease, ptSmooth;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 i, uiNeighbors, uiSharpEdges, uiSemiSharpEdges;
  FLOAT fA, fAlpha;
  FLOAT fSharpness, pfSemiSharpnessesMax[3], fSwap, fT;

  // Sharp Loop Vertex-Vertex Subdivision
  ptCorner = pvx->GetPoint();
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptCorner;
    }

  iter.Init(pvx);

  uiSharpEdges = 0;
  uiSemiSharpEdges = 0;
  memset(pfSemiSharpnessesMax, 0, sizeof(pfSemiSharpnessesMax));
  ptSmooth.Null();
  ptCrease.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pe->GetEdge()->GetVertex()->GetPoint();
      ptSmooth += pt;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ptCrease += pt;
	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      uiSharpEdges++;
	    }
	  else
	    {
	      // Semi-sharp for this step
	      uiSemiSharpEdges++;
	      if ( fSharpness > pfSemiSharpnessesMax[2] )
		{
		  pfSemiSharpnessesMax[2] = fSharpness;
		  for ( i = 2; i > 0; i-- )
		    {
		      if ( pfSemiSharpnessesMax[i] > pfSemiSharpnessesMax[i-1] )
			{
			  fSwap = pfSemiSharpnessesMax[i];
			  pfSemiSharpnessesMax[i] = pfSemiSharpnessesMax[i-1];
			  pfSemiSharpnessesMax[i-1] = fSwap;
			}
		    }
		}
	    }
	}
    }

  //////////////////////////////
  // Corner rule
  if ( uiSharpEdges >= 3 )
    {
      return ptCorner;
    }

  //////////////////////////////
  // Crease rule
  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      ptCrease = ((6.0 * ptCorner) + ptCrease) / 8.0;
      ptCrease[3] = 1.0;
    }

  if ( uiSharpEdges == 2 && uiSemiSharpEdges == 0 )
    {
      return ptCrease;
    }

  //////////////////////////////
  // Smooth rule
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  ptSmooth = (ptSmooth + (fAlpha * ptCorner)) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  if ( uiSharpEdges + uiSemiSharpEdges <= 1 )
    {
      return ptSmooth;
    }

  ASSERT( uiSharpEdges < 3 );
  i = MIN(uiSemiSharpEdges, 3 - uiSharpEdges);
  fT = pfSemiSharpnessesMax[i-1];

  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
      pt[3] = 1.0;
      return pt;
    }

  ASSERT( uiSharpEdges + uiSemiSharpEdges >= 3 );
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCorner);
  pt[3] = 1.0;
  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionSharpLoop(QEVertex *pvx)
{
  Point pt, ptCorner, ptSmooth, ptCrease;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *ppeCrease[2];
  UINT_32 uiNeighbors, uiCreases;
  FLOAT fA, fEpsilon;
  Point ppptCrease[2][3];
  SharpVertexType psvtCrease[2];
  const FLOAT *pfCoeffTable;
  UINT_32 i, j, k, uiStart, uiEnd, uiCenter;

  iterQEVertex.Init(pvx);

  // Sharp Loop Vertex-Vertex Subdivision
  ptCorner = pvx->GetPoint();
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptCorner;
    }

  ptSmooth.Null();
  uiCreases = 0;
  for ( uiNeighbors = 0, pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iterQEVertex.PeekNextEdge() )
    {
      pt = pe->GetEdge()->GetVertex()->GetPoint();
      ptSmooth += pt;

      if ( pe->GetKnotInterval() != 0.0 )
	{
	  if ( uiCreases < 2 )
	    {
	      ppeCrease[uiCreases] = pe;
	      psvtCrease[uiCreases] = pe->GetEdge()->GetVertex()->GetSharpVertexType();
	      ppptCrease[uiCreases][0] = pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();
	      ppptCrease[uiCreases][1] = pe->GetEdge()->GetVertex()->GetPoint();
	      ppptCrease[uiCreases][2] = pe->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();
	    }
	  uiCreases++;
	}
    }
  
  if ( uiCreases >= 3 )
    {
      return ptCorner;
    }

  if ( uiCreases == 2 )
    {
      if ( ( ppeCrease[0]->GetEdgeNext(1) == ppeCrease[1] &&
	     ppeCrease[0]->GetFacet() == GetNullFacet() ) ||
	   ( ppeCrease[1]->GetEdgeNext(1) == ppeCrease[0] &&
	     ppeCrease[1]->GetFacet() == GetNullFacet() ) )
	{
	  uiCenter = scpuiSharpLoopSharpTypeIndex[svtBoundaryCrease];
	}
      else if ( uiNeighbors == 6 && 
		ppeCrease[0]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1) == ppeCrease[1] )
	{
	  uiCenter = scpuiSharpLoopSharpTypeIndex[svtRegularCrease];
	}
      else
	{
	  uiCenter = scpuiSharpLoopSharpTypeIndex[svtNonregularCrease];
	}
      uiStart = scpuiSharpLoopSharpTypeIndex[psvtCrease[0]];
      uiEnd = scpuiSharpLoopSharpTypeIndex[psvtCrease[1]];

      pfCoeffTable = scppppfSharpLoopLimitTable[uiCenter][uiStart][uiEnd];
      i = 0;
      ptCrease = pfCoeffTable[i] * ptCorner;
#ifdef DEBUGGING_SHARP_LOOP_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	  fprintf(stderr, "  %f * ", pfCoeffTable[i]); 
	  ptCorner.Display();
	} 
#endif // DEBUGGING_SHARP_LOOP_MATH
      for ( j = 0; j < 2; j++ )
	{
	  for ( k = 0; k < 3; k++ )
	    {
	      i++;
	      ptCrease += pfCoeffTable[i] * ppptCrease[j][k];
#ifdef DEBUGGING_SHARP_LOOP_MATH
	      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
		   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )		   
		{
		  fprintf(stderr, "  %f * ", pfCoeffTable[i]);
		  ppptCrease[j][k].Display();
		} 
#endif // DEBUGGING_SHARP_LOOP_MATH
	    }
	}
      ptCrease[3] = 1.0;
#ifdef DEBUGGING_SHARP_LOOP_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "  Limit = "); 
	  ptCorner.Display();
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	} 
#endif // DEBUGGING_SHARP_LOOP_MATH

      return ptCrease;
    }

  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiNeighbors) / (8.0 * fA);
  ptSmooth = (ptSmooth + (fEpsilon * ptCorner)) / (fEpsilon + (FLOAT)uiNeighbors);
  ptSmooth[3] = 1.0;

  return ptSmooth;
}

Vector QEMesh::CalculateVertexNormalSubdivisionSharpLoop(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, vNormalMax;
  //Vector v;
  Vector vCenter;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCrease;
  UINT_32 uiNeighbors, uiCreases, i, j, k, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  FLOAT fSharpness, fMag, fMagMax;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];
  BOOL pbCrease[cuiNeighborsMax];
  UINT_32 puiSegmentLength[cuiNeighborsMax];

  iterQEVertex.Init(pvx);

  // Search for a crease edge (if one exists it will be in the mesh)
  peCrease = NULL;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      uiNeighbors++;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness == INFINITY )
	{
	  peCrease = pe;
	}
    }
  ASSERT( uiNeighbors < cuiNeighborsMax );

  // Calculate all the neighbor vertices
  if ( peCrease == NULL )
    {
      peCrease = iterQEVertex.PeekFirstEdge();
    }
  uiCreases = 0;
  memset(pbCrease, 0, sizeof(pbCrease));
  memset(puiSegmentLength, 0, sizeof(puiSegmentLength));
  i = 0;
  pe = peCrease;
  do
    {
      fSharpness = pe->GetKnotInterval();

      ASSERT( i < uiNeighbors );
      pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
      if ( fSharpness == INFINITY )
	{
	  uiCreases++;
	  pbCrease[i] = TRUE;
	}
      i++;
      pe = pe->GetEdgeNext(1);
    }
  while( pe != peCrease );

  for ( i = 0; i < uiNeighbors; i++ )
    {      
      pv[i][3] = 1.0;
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;
  if ( uiCreases < 2 )
    {
      vT0.Null();
      vT1.Null();
      fTheta = 0.0;
      fThetaDelta = 2.0 * PI / (FLOAT)uiNeighbors;
      for ( i = 0; i < uiNeighbors; i++ )
	{
	  vT1 += cos(fTheta) * pv[i];
	  fTheta += fThetaDelta;
	  vT0 += cos(fTheta) * pv[i];
	}
      vT0[3] = 0.0;
      vT1[3] = 0.0;
      vNormal = vT0 * vT1;
    }
  else
    {
      // Copy the first edge to the end
      pv[uiNeighbors] = pv[0]; 
      pbCrease[uiNeighbors] = pbCrease[0];
      j = 0;
      for ( i = 0; i < uiNeighbors; i++ )
	{      
	  if ( pbCrease[i] )
	    {
	      j++;
	    }
	  ASSERT( j > 0 && j <= uiCreases );
	  puiSegmentLength[j-1]++;
	}

      bRegularCrease = ( uiNeighbors == 6 && uiCreases == 2 &&
			 puiSegmentLength[0] == 3 && puiSegmentLength[1] == 3 ); 

      // Calculate the normal for the individual segments and take the one with max mag 
      vNormalMax.Null();
      fMagMax = 0.0;
      k = 0;
      for ( i = 0; i < uiCreases; k += puiSegmentLength[i], i++ )
	{
	  vT0.Null();
	  vT1.Null();
	  if ( bRegularCrease )
	    {
	      vT0 = pv[k];
	      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
	      for ( j = 0; j <= puiSegmentLength[i]; j++ )
		{
		  vT1 += scppfSharpLoopTangentTable[2][j+1] * pv[k+j];
		}
	      vT0 -= pv[k+puiSegmentLength[i]];	      
	    }
	  else
	    {
	      switch ( puiSegmentLength[i] )
		{
		case 0:
		  ASSERT( FALSE );
		  break;
		case 1:
		case 2:
		  uiIndex = puiSegmentLength[i] - 1;
		  vT0 = pv[k];
		  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
		  for ( j = 0; j <= puiSegmentLength[i]; j++ )
		    {
		      vT1 += scppfSharpLoopTangentTable[uiIndex][j+1] * pv[k+j];
		    }
		  vT0 -= pv[k+puiSegmentLength[i]];
		  break;
		default:
		  fThetaDelta = PI / (FLOAT)puiSegmentLength[i];
		  fA = sin(fThetaDelta);
		  vT1 = pv[k];
		  vT0 = (fA * pv[k]);
		  vT1 -= pv[k+puiSegmentLength[i]];
		  vT0 += (fA * pv[k+puiSegmentLength[i]]);
		  fA = 2.0 * (cos(fThetaDelta) - 1.0);
		  for ( j = 1, fTheta = fThetaDelta; 
			j < puiSegmentLength[i];
			j++, fTheta += fThetaDelta )
		    {
		      vT0 += ((fA * sin(fTheta)) * pv[k+j]);
		    }
		  break;
		}
	    }
	  vT0[3] = 0.0;
	  vT1[3] = 0.0;
	  vNormal = vT0 * vT1;
	  fMag = vNormal.Magnitude();
	  if ( fMag > fMagMax )
	    {
	      fMagMax = fMag;
	      vNormalMax = vNormal;
	    }
	}
      vNormal = vNormalMax;
    }

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Vector QEMesh::CalculateVertexNormalSubdivisionSharpLoop(QEEdge *peCrease)
{
  Vector vT0, vT1, vNormal;
  Vector vCenter;
  QEVertex *pvx;
  QEEdge *pe;
  UINT_32 uiSegmentLength, i, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];

  pvx = peCrease->GetVertex();
  ASSERT( pvx->GetSharpVertexType() == svtRegularCrease ||
	  pvx->GetSharpVertexType() == svtBoundaryCrease ||
	  pvx->GetSharpVertexType() == svtNonregularCrease ||
	  pvx->GetSharpVertexType() == svtCorner );
  ASSERT( peCrease->GetFacet() != GetNullFacet() );
  ASSERT( peCrease->GetKnotInterval() == INFINITY );

  // Search for a crease edge (if one exists it will be in the mesh)
  uiSegmentLength = 2;
  for ( pe = peCrease->GetEdgeNext(1);
	pe != peCrease && pe->GetKnotInterval() != INFINITY;
	pe = pe->GetEdgeNext(1) )
    {
      uiSegmentLength++;
    }
  ASSERT( uiSegmentLength < cuiNeighborsMax );

  pe = peCrease;
  for ( i = 0; i < uiSegmentLength; i++ )
    {
      ASSERT( i < uiSegmentLength );
      pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
      pv[i][3] = 1.0;
      pe = pe->GetEdgeNext(1);
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;

  bRegularCrease = (pvx->GetSharpVertexType() == svtRegularCrease);
  
  // Calculate the normal for the individual segments and take the one with max magnitude
  vT0.Null();
  vT1.Null();
  if ( bRegularCrease )
    {
      vT0 = pv[0];
      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
      for ( i = 0; i < uiSegmentLength; i++ )
	{
	  vT1 += scppfSharpLoopTangentTable[2][i+1] * pv[i];
	}
      vT0 -= pv[uiSegmentLength - 1];	      
    }
  else
    {
      switch ( uiSegmentLength )
	{
	case 0:
	case 1:
	  ASSERT( FALSE );
	  break;
	case 2:
	case 3:
	  uiIndex = uiSegmentLength - 2;
	  vT0 = pv[0];
	  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
	  for ( i = 0; i < uiSegmentLength; i++ )
	    {
	      vT1 += scppfSharpLoopTangentTable[uiIndex][i+1] * pv[i];
	    }
	  vT0 -= pv[uiSegmentLength - 1];
	  break;
	default:
	  fThetaDelta = PI / (FLOAT)(uiSegmentLength - 1);
	  fA = sin(fThetaDelta);
	  vT1 = pv[0];
	  vT0 = (fA * pv[0]);
	  vT1 -= pv[uiSegmentLength - 1];
	  vT0 += (fA * pv[uiSegmentLength - 1]);
	  fA = 2.0 * (cos(fThetaDelta) - 1.0);
	  for ( i = 1, fTheta = fThetaDelta; 
		i < uiSegmentLength - 1;
		i++, fTheta += fThetaDelta )
	    {
	      vT0 += ((fA * sin(fTheta)) * pv[i+1]);
	    }
	  break;
	}
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;
  vNormal = vT0 * vT1;

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionSharpLoop(QEEdge *pe)
{
  QEEdge *ppe[2], *peCurr, *peSwap;
  BOOL pbCrease[2], pbRegularCrease[2], bCrease;
  Point pt, ptSmooth, ptCrease;
  Point ppt[4];
  UINT_32 i, uiN, uiSharpEdges;
  FLOAT fSharpness, fT;

  fSharpness = pe->GetKnotInterval();
  bCrease = FALSE;
  if ( fSharpness != 0.0 )
    {
      ppe[0] = pe;
      ppe[1] = pe->GetEdge();
      pbCrease[0] = FALSE;
      pbCrease[1] = FALSE;
      pbRegularCrease[0] = FALSE;
      pbRegularCrease[1] = FALSE;
      for ( i = 0; i < 2; i++ )
	{
	  uiSharpEdges = 1;
	  uiN = 1;
	  for ( peCurr = ppe[i]->GetEdgeNext(1); 
		peCurr != ppe[i]; 
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      uiN++;
	      if ( peCurr->GetKnotInterval() != 0.0 )
		{
		  uiSharpEdges++;
		}
	    }

	  if ( uiSharpEdges < 2 )
	    {
	      break;
	    }

	  pbCrease[i] = TRUE;
	  if ( uiSharpEdges == 2 && uiN == 6 &&
	       ppe[i]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval() != 0.0 )
	    {
	      pbRegularCrease[i] = TRUE;
	    }	  
	}

      bCrease = pbCrease[0] && pbCrease[1];
      if ( bCrease )
	{
	  if ( ( pbRegularCrease[0] && pbRegularCrease[1] ) ||
	       ( !pbRegularCrease[0] && !pbRegularCrease[1] ) )
	    {
	      ptCrease = ( ppe[0]->GetVertex()->GetPoint() + ppe[1]->GetVertex()->GetPoint() ) / 2.0;
	      ptCrease[3] = 1.0;
	    }
	  else
	    {
	      if ( pbRegularCrease[1] )
		{
		  peSwap = ppe[0];
		  ppe[0] = ppe[1];
		  ppe[1] = peSwap;
		}
	      ptCrease = ( (5.0 * ppe[0]->GetVertex()->GetPoint()) + (3.0 * ppe[1]->GetVertex()->GetPoint()) ) / 8.0;
	      ptCrease[3] = 1.0;
	      
	    }

	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      return ptCrease;
	    }
	}
    }

  // Midpoint smooth rule
  /*
  // Sharp Loop Edge-Vertex Subdivision
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint();
  ppt[3] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();

  pt = ((3.0 * ppt[0]) + (3.0 * ppt[1]) + ppt[2] + ppt[3]) / 8.0;
  pt[3] = 1.0;
  */
  // Sharp Generalized Loop Edge-Vertex Subdivision
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint() + 
	    pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
  ppt[3] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
	    pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();

  ptSmooth = ((6.0 * ppt[0]) + (6.0 * ppt[1]) + ppt[2] + ppt[3]) / 16.0;
  ptSmooth[3] = 1.0;

  if ( !bCrease )
    {
      return ptSmooth;
    } 

  fT = fSharpness;
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
  pt[3] = 1.0;
  return pt;
}

//////////////////////////////////////////////////////////////////////
// Hermite interpolation

Point QEMesh::CalculateVertexSubdivisionHermite(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionHermite(QEEdge *pe)
{
  QEEdge *ppe[2], *peCurr;
  Point pt;
  Point ppt[2];
  //Vector pvN[2];
  Vector vE;
  Vector pvT[2];
  FLOAT fMag;
  UINT_32 i, j, uiN;

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  for ( i = 0; i < 2; i++ )
    {
      ppt[i] = ppe[i]->GetVertex()->GetPoint();

      // Compute tangent vector
      peCurr = ppe[i];
      uiN = 0;
      do
	{
	  uiN++;
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != ppe[i] );
      
      peCurr = ppe[i];
      pvT[i].Null();
      j = 0;
      do
	{
	  pvT[i] += cos(2.0*PI*j/uiN)*(peCurr->GetEdge()->GetVertex()->GetPoint() - ppt[i]);
	  j++;
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != ppe[i] );

      pvT[i].Normalize();
    }

  vE = ppt[1] - ppt[0];
  fMag = vE.Magnitude();

  pt = (ppt[0] + ppt[1]) / 2.0;
  pt += (pvT[0] + pvT[1]) * ( (2.0*fMag/3.0) / 8.0 );
  pt[3] = 1.0;

  return pt;
}

//////////////////////////////////////////////////////////////////////
// Loop's bounded curvature scheme with modified edge rules

Point QEMesh::CalculateVertexSubdivisionLoopEdgeRule(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fAlpha, fBeta;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
    }
  
  FLOAT fLambda1, fLambda0;

  switch ( uiNeighbors )
    {
    case 0:
    case 1:
    case 2:
      ASSERT( FALSE );
      break;
    case 3:
      fAlpha = 7.0/16.0;
      break;
    case 4:
      fAlpha = 1.0/2.0;
      break;
    case 5:
      fAlpha = (81.0 - 5.0*sqrt(5.0))/128.0;
      break;
    default:
      fLambda1 = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors))/8.0;
      fLambda0 = fLambda1*(4.0 + fLambda1*(5.0*fLambda1 - 8.0)) / (2.0*(1.0 - fLambda1));
      fAlpha = 1 - fLambda0 + fLambda1*fLambda1;
     break;
    }

  fBeta = (1.0 - fAlpha)/(FLOAT)uiNeighbors;

  pt = (fAlpha * pvx->GetPoint()) + (fBeta * pt);
  pt[3] = 1.0;

  return pt;
}

static Point CalculateEdgeSubdivisionLoopEdgeRuleIrregular(QEEdge *pe)
{
  QEEdge *peCurr;
  Point pt, ptCurr;
  UINT_32 uiN, i;
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  uiN = 1;
  for ( peCurr = pe->GetEdgeNext(1);
	peCurr != pe;
	peCurr = peCurr->GetEdgeNext(1) )
    {
      uiN++;
    }

  FLOAT fLambda1, fLambda0;
  FLOAT fGamma, fU, fA, fB;

  fLambda1 = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiN))/8.0;
  fA = 2.0*fLambda1*fLambda1*fLambda1/((FLOAT)uiN*(1.0 - fLambda1));
  fB = (1.0/fLambda1) - (3.0/2.0);

  fLambda0 = 0.0;
  pt.Null();
  peCurr = pe;
  i = 0;
  do
    {
      fU = cos(2.0*PI*(FLOAT)i/(FLOAT)uiN);

      switch ( uiN )
	{
	case 0:
	case 1:
	case 2:
	  ASSERT( FALSE );
	  break;
	case 3:
	  fGamma = (5.0/4.0 + fU)/6.0;
	  break;
	case 4:
	  fGamma = (4.0 + 3*fU)/8.0;
	  fGamma = fGamma*fGamma/2.0;
	  break;
	case 5:
	  fGamma = ((5.0 - sqrt(5.0))/5.0) + fU;
	  fGamma = ((3.0 + sqrt(5.0))/32.0)*fGamma*fGamma;
	  break;
	default:  
	  fGamma = fU + fB;
	  fGamma = fA*(1.0 + fU)*fGamma*fGamma;
	  break;
	}      

      ptCurr = peCurr->GetEdge()->GetVertex()->GetPoint();
      pt += (fGamma*ptCurr);
      fLambda0 += fGamma;

      peCurr = peCurr->GetEdgeNext(1);
      i++;
    }
  while ( peCurr != pe );

  ptCurr = pe->GetVertex()->GetPoint();
  pt += (1.0 - fLambda0)*ptCurr;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionLoopEdgeRule(QEEdge *pe)
{
  Point pt, ppt[2];
  QEEdge *ppe[2], *peCurr;
  UINT_32 puiN[2], i;
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  for ( i = 0; i < 2; i++ )
    {
      puiN[i] = 1;
      for ( peCurr = ppe[i]->GetEdgeNext(1);
	    peCurr != ppe[i];
	    peCurr = peCurr->GetEdgeNext(1) )
	{
	  puiN[i]++;
	}
    }

  if ( puiN[0] == 6 )
    {
      pt = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[1]);
    }
  else if ( puiN[1] == 6 )
    {
      pt = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[0]);
    }
  else
    {
      ppt[0] = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[0]);
      ppt[1] = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[1]);
      pt = (ppt[0] + ppt[1]) / 2.0;
    }

  pt[3] = 1.0;
  return pt;
}

Point QEMesh::CalculateVertexSubdivisionLoopExperimental(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fAlpha, fBeta;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
    }
  
  FLOAT fLambda1, fLambda0;

  switch ( uiNeighbors )
    {
    case 0:
    case 1:
    case 2:
      ASSERT( FALSE );
      break;
    case 3:
      fAlpha = 7.0/16.0;
      break;
    case 4:
      fAlpha = 1.0/2.0;
      break;
    case 5:
      fAlpha = (81.0 - 5.0*sqrt(5.0))/128.0;
      break;
    default:
      //fLambda1 = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors))/8.0;
      //fLambda0 = fLambda1*(4.0 + fLambda1*(5.0*fLambda1 - 8.0)) / (2.0*(1.0 - fLambda1));
      //fAlpha = 1 - fLambda0 + fLambda1*fLambda1;
      fAlpha = 5.0 / 8.0;
     break;
    }

  fBeta = (1.0 - fAlpha)/(FLOAT)uiNeighbors;

  pt = (fAlpha * pvx->GetPoint()) + (fBeta * pt);
  pt[3] = 1.0;

  return pt;
}

static Point CalculateEdgeSubdivisionLoopExperimentalIrregular(QEEdge *pe)
{
  QEEdge *peCurr;
  Point pt, ptCurr;
  UINT_32 uiN, i;
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  uiN = 1;
  for ( peCurr = pe->GetEdgeNext(1);
	peCurr != pe;
	peCurr = peCurr->GetEdgeNext(1) )
    {
      uiN++;
    }

  FLOAT fLambda1, fLambda0;
  FLOAT fGamma, fU, fA, fB;

  fLambda1 = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiN))/8.0;
  fA = 2.0*fLambda1*fLambda1*fLambda1/((FLOAT)uiN*(1.0 - fLambda1));
  fB = (1.0/fLambda1) - (3.0/2.0);
  if ( uiN > 5 )
    {
      fLambda1 = (1.0/2.0);
      fA = 1.0/(2.0*(FLOAT)uiN);
      fB = (1.0/2.0);      
    }

  fLambda0 = 0.0;
  pt.Null();
  peCurr = pe;
  i = 0;
  do
    {
      fU = cos(2.0*PI*(FLOAT)i/(FLOAT)uiN);

      switch ( uiN )
	{
	case 0:
	case 1:
	case 2:
	  ASSERT( FALSE );
	  break;
	case 3:
	  fGamma = (5.0/4.0 + fU)/6.0;
	  break;
	case 4:
	  fGamma = (4.0 + 3*fU)/8.0;
	  fGamma = fGamma*fGamma/2.0;
	  break;
	case 5:
	  fGamma = ((5.0 - sqrt(5.0))/5.0) + fU;
	  fGamma = ((3.0 + sqrt(5.0))/32.0)*fGamma*fGamma;
	  break;
	default:  
	  fGamma = fU + fB;
	  fGamma = fA*(1.0 + fU)*fGamma*fGamma;
	  break;
	}      

      ptCurr = peCurr->GetEdge()->GetVertex()->GetPoint();
      pt += (fGamma*ptCurr);
      fLambda0 += fGamma;

      peCurr = peCurr->GetEdgeNext(1);
      i++;
    }
  while ( peCurr != pe );

  ptCurr = pe->GetVertex()->GetPoint();
  pt += (1.0 - fLambda0)*ptCurr;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionLoopExperimental(QEEdge *pe)
{
  Point pt, ppt[2];
  QEEdge *ppe[2], *peCurr;
  UINT_32 puiN[2], i;
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  for ( i = 0; i < 2; i++ )
    {
      puiN[i] = 1;
      for ( peCurr = ppe[i]->GetEdgeNext(1);
	    peCurr != ppe[i];
	    peCurr = peCurr->GetEdgeNext(1) )
	{
	  puiN[i]++;
	}
    }

  if ( puiN[0] == 6 )
    {
      pt = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[1]);
    }
  else if ( puiN[1] == 6 )
    {
      pt = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[0]);
    }
  else
    {
      ppt[0] = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[0]);
      ppt[1] = CalculateEdgeSubdivisionLoopEdgeRuleIrregular(ppe[1]);
      pt = (ppt[0] + ppt[1]) / 2.0;
    }

  pt[3] = 1.0;
  return pt;
}

//////////////////////////////////////////////////////////////////////
// Butterfly scheme averaging rules

Point QEMesh::CalculateVertexSubdivisionButterfly(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionButterfly(QEEdge *pe)
{
  Point pt, ppt[8];
  UINT_32 uiSide;
  FLOAT fT;

  pt = CalculateEdgeSubdivisionTesselate(pe);
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  for ( uiSide = 0; uiSide < 2; uiSide++ )
    {
      ppt[uiSide] = pe->GetVertex()->GetPoint();

      ppt[uiSide+2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint();

      ppt[uiSide+4] = pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();
      ppt[uiSide+6] = pe->GetEdge()->GetEdgePrev(1)->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();

      pe = pe->GetEdge();
    }

  fT = 1.0 / 8.0;
  //fT = 0.0;

  pt = ((ppt[0] + ppt[1]) + fT * ( 2.0 * (ppt[2] + ppt[3]) - (ppt[4] + ppt[5] + ppt[6] + ppt[7]))) / 2.0;
  pt[3] = 1.0;
  
  return pt;
}

Point QEMesh::CalculateVertexSubdivisionZorin(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionZorin(QEEdge *pe)
{
  Point pt, ptOrigin;
  QEVertexIter iterQEVertex;
  QEEdge *peCurr;
  QEVertex *ppvx[2];
  Point ppt[2];
  Vector pv[2], v;
  UINT_32 i, j, puiNeighbors[2];
  FLOAT fS;
  
  for ( i = 0; i < 2; i++ )
    {
      ppvx[i] = pe->GetVertex();

      // Count the number of neighbors around the vertex
      iterQEVertex.Init(ppvx[i]);
      puiNeighbors[i] = 0;
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  puiNeighbors[i]++;
	}

      // Calculate the weight point
      ptOrigin = ppvx[i]->GetPoint();
      pv[i].Null();
      switch ( puiNeighbors[i] )
	{
	case 0:
	case 1:
	case 2:
	  ASSERT( FALSE );
	  break;
	case 3:
	  {
	    FLOAT pfWeight[3] = { 5.0 / 12.0, -1.0 / 12.0, -1.0 / 12.0 };

	    peCurr = pe;
	    j = 0;
	    do
	      {
		ASSERT( j < 3 );
		pt = peCurr->GetEdge()->GetVertex()->GetPoint();

		v = pt - ptOrigin;
		pv[i] = pv[i] + pfWeight[j] * v;
		
		// Advance to the next edge
		peCurr = peCurr->GetEdgeNext(1);

		j++;
	      }
	    while ( peCurr != pe );
	  }
	  break;
	case 4:
	  {
	    FLOAT pfWeight[4] = { 3.0 / 8.0, 0.0, -1.0 / 8.0, 0.0 };

	    peCurr = pe;
	    j = 0;
	    do
	      {
		ASSERT( j < 4 );
		pt = peCurr->GetEdge()->GetVertex()->GetPoint();
		
		v = pt - ptOrigin;
		pv[i] = pv[i] + pfWeight[j] * v;
		
		// Advance to the next edge
		peCurr = peCurr->GetEdgeNext(1);

		j++;
	      }
	    while ( peCurr != pe );
	  }
          break;	  
	case 6:
          break;	  
	default:
	  {
	    peCurr = pe;
	    j = 0;
	    do
	      {
		pt = peCurr->GetEdge()->GetVertex()->GetPoint();
		fS = (0.25 + cos(2.0 * PI * j / puiNeighbors[i]) + 0.5 * + cos(4.0 * PI * j / puiNeighbors[i])) / puiNeighbors[i];
		
		v = pt - ptOrigin;
		pv[i] = pv[i] + fS * v;
		
		// Advance to the next edge
		peCurr = peCurr->GetEdgeNext(1);

		j++;
	      }
	    while ( peCurr != pe );
	  }
	  break;
	}

      ppt[i] = ptOrigin + pv[i];

      pe = pe->GetEdge();
    }

  if ( puiNeighbors[0] == 6 && puiNeighbors[1] == 6 )
    {
      pt = CalculateEdgeSubdivisionButterfly(pe);
    }
  else if ( puiNeighbors[1] == 6 )
    {
      pt = ppt[0];
    }
  else if ( puiNeighbors[0] == 6 )
    {
      pt = ppt[1];
    }
  else
    {
      pt = (ppt[0] + ppt[1]) / 2.0;
    }
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSphere(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionSphere(QEEdge *pe)
{
  QEVertexIter iterQEVertex;
  QEVertex *ppvx[2];
  Vector pv[2], pvEdge[2], v;
  Point pt;
  FLOAT fWeight;
  //FLOAT fTheta, fCos, fSin, fMag, fOffset;
  UINT_32 i;

  ppvx[0] = pe->GetVertex();
  ppvx[1] = pe->GetEdge()->GetVertex();

  for ( i = 0; i < 2; i++ )
    {
      pv[i] = ppvx[i]->GetNormal();
      if ( pv[i].Magnitude() == 0.0 )
	{
	  iterQEVertex.Init(ppvx[i]);
	  
	  pv[i].Null();
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      pvEdge[0] = pe->GetEdge()->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      pvEdge[0].Normalize();
	      pvEdge[1] = pe->GetEdgePrev(0)->GetVertex()->GetPoint() - pe->GetVertex()->GetPoint();
	      pvEdge[1].Normalize();
	      fWeight = pvEdge[0].Dot(pvEdge[1]);
	      fWeight = acos(fWeight);
	      
	      pv[i] += (fWeight * pe->GetFacet()->GetNormal());
	    }
	  pv[i][3] = 0.0;
	  pv[i].Normalize();

	  ppvx[i]->SetNormal(pv[i]);
	}
    }

  pt = (ppvx[1]->GetPoint() + ppvx[0]->GetPoint()) / 2.0;
  pt[3] = 1.0;
/*
  v = ppvx[1]->GetPoint() - ppvx[0]->GetPoint();
  fMag = v.Magnitude();
  fMag /= 2.0;


  fTheta = pv[0].Dot(pv[1]);
  if ( fTheta > 1.0 )
    {
      fTheta = 1.0;
    }
  if ( fTheta < -1.0 )
    {
      fTheta = -1.0;
    }
  fTheta = acos(fTheta);
  fTheta /= 2.0;

  fCos = cos(fTheta);
  fSin = sin(fTheta);

  if ( fSin != 0.0 )
    {
      fOffset = fMag * ( (1.0 - fCos) / fSin );
			
      v = pv[0] + pv[1];
      v.Normalize();
      v *= fOffset;
      
      pt += v;
    }
    */

  v = pt;
  v[3] = 0.0;
  v.Normalize();
  pt = v;
  pt[3] = 1.0;

  return pt;
}

///////////////////////////////////////////////////
// QEMesh::SubdivideSqrt4                        //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1              f0    //
//           |               \            /      //
//           |                 e1      e0        //
//           |                   \    /          //
//  e2------vV------e0  =>         vV            //
//           |                   /    \          //
//           |                 e2      e3        //
//           |               /            \      //
//  f2      e3      f3     f2              f3    //
//                                               //
QEMesh *QEMesh::SubdivideSqrt4()
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertex *pvxNew;
  QEFacet *pfaNew;
  QEEdge *pe, *peCurr, *peNew;

  Point pt;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      peNew = pme->CreateEdge();
      ASSERT( peNew != NULL );

      pe->SetForwardEdge(peNew);

      // The facet vertex
      if ( pe->GetFacet()->GetForwardVertex() != NULL )
	{
	  for ( peCurr = pe->GetEdgePrev(0);
		peCurr->GetForwardEdge() == NULL;
		peCurr = peCurr->GetEdgePrev(0) )
	    {
	      ASSERT( peCurr != pe );
	    }

	  Splice(1, peNew, peCurr->GetForwardEdge());

	  peNew->SetVertex(pe->GetFacet()->GetForwardVertex());
	}
      else
	{
	  pt = CalculateFacetSubdivisionSqrt4(pe->GetFacet());
	  pvxNew = pme->CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pe->GetFacet()->SetForwardVertex(pvxNew);

	  peNew->SetVertex(pvxNew);
	}
      
      // The vertex vertex
      if ( pe->GetVertex()->GetForwardVertex() != NULL )
	{
	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr->GetForwardEdge() == NULL;
		peCurr = peCurr->GetEdgePrev(1) )
	    {
	      ASSERT( peCurr != pe );
	    }

	  Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetEdge());

	  peNew->GetEdge()->SetVertex(pe->GetVertex()->GetForwardVertex());
	}
      else
	{
	  pt = CalculateVertexSubdivisionSqrt4(pe->GetVertex());
	  pvxNew = pme->CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pe->GetVertex()->SetForwardVertex(pvxNew);

	  peNew->GetEdge()->SetVertex(pvxNew);
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetForwardEdge() != NULL )
	{
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );
	  peCurr = pe->GetForwardEdge();
	  do
	    {
	      peCurr->SetFacet(pfaNew);
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( peCurr != pe->GetForwardEdge() );

	  pe->SetForwardEdge(NULL);
	  pe->GetEdge()->SetForwardEdge(NULL);
	}
    }
  pme->Preprocess();

  return pme;
}

Point QEMesh::CalculateFacetSubdivisionSqrt4(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN;

  pt.Null();
  iterQEFacet.Init(pfa);
  for ( uiN = 0, pvx = iterQEFacet.PeekFirstVertex();
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      pt += pvx->GetPoint();
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSqrt4(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  FLOAT fAlpha;
  UINT_32 uiN;

  pt.Null();
  iterQEVertex.Init(pvx);
  for ( uiN = 0, pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	uiN++, pe = iterQEVertex.PeekNextEdge() )
    {
      pt += pe->GetEdge()->GetVertex()->GetPoint();
    }

  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;

  pt = ( ((1.0 - fAlpha)*pvx->GetPoint()) + (fAlpha/(FLOAT)uiN)*pt );
  pt[3] = 1.0;

  return pt;
}

/////////////////////////////////
// QEMesh::SubdivideDooSabin   //
//                             //
// Old Edge    Subdivided Edge //
//                             //
//    v0         v2--e1--v1    //
//     |          |       |    //
//     |          |       |    //
//     |          |       |    //
//     e    =>   e2      e0    //
//     |          |       |    //
//     |          |       |    //
//     |          |       |    //
//    v1         v3--e3--v0    //
//                             //
QEMesh *QEMesh::SubdivideDooSabin()
{
  //ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  QEMeshIter iterQEMesh;
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEMesh *pme;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *ppe[4], *peCurr, *peStart, *peNew;
  QEVertex *pvx, *pvxNew;
  Point pt, ptV, ptSum;
  FLOAT fWeight, fWeights;
  UINT_32 i, j, uiN;
  BOOL bBoundaryVertex;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  ClearForwardPointers();

  iterQEMesh.Init(this);

  // Edges
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      ASSERT( pe->GetFacet() != m_pfaNULL );
      if ( pe->GetForwardEdge() == NULL )
	{
	  if ( pe->GetEdge()->GetFacet() == m_pfaNULL )
	    {
	      ppe[0] = pme->CreateEdge();
	      pme->m_plistQEEdge->Remove(ppe[0]);
	      pme->m_plistQEEdgeBoundary->InsertLast(ppe[0]);

	      pe->SetForwardEdge(ppe[0]->GetEdge());
	      ASSERT( pe->GetEdge()->GetForwardEdge() == NULL );

	      fWeight = pe->GetKnotInterval() / 2.0;
	      ppe[0]->GetEdge()->SetKnotInterval(fWeight);
	      fWeight = pe->GetEdge()->GetKnotInterval() / 2.0;
	      ppe[0]->SetKnotInterval(fWeight);

	      //////////////////////////////
	      // v0
	      peCurr = pe->GetEdgeNext(0);
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  peCurr = peCurr->GetForwardEdge();
		  pvxNew = peCurr->GetVertex();
		  pme->Splice(1, ppe[0]->GetEdgePrev(1), peCurr);
		}
	      else
		{
		  pt.Null();
		  pvxNew = pme->CreateVertex(pt);
		}
	      ppe[0]->SetVertex(pvxNew);

	      //////////////////////////////
	      // v1
	      peCurr = pe->GetEdgePrev(0);
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  peCurr = peCurr->GetForwardEdge()->GetEdge()->GetEdgePrev(1);
		  pvxNew = peCurr->GetVertex();
		  pme->Splice(1, ppe[0]->GetEdge()->GetEdgePrev(1), peCurr);
		}
	      else
		{
		  pt.Null();
		  pvxNew = pme->CreateVertex(pt);
		}
	      ppe[0]->GetEdge()->SetVertex(pvxNew);
	    }
	  else
	    {
	      pfaNew = pme->CreateFacet();

	      for ( i = 0; i < 4; i++ )
		{
		  ppe[i] = pme->CreateEdge();
		}
	      /*
	      if ( pe->GetEdgeNext(1)->GetFacet() == m_pfaNULL &&
		   pe->GetEdgePrev(1)->GetEdge()->GetFacet() == m_pfaNULL )
		{
		  pme->m_plistQEEdge->Remove(ppe[1]->GetEdge());
		  pme->m_plistQEEdgeBoundary->InsertLast(ppe[1]->GetEdge());
		}
	      if ( pe->GetEdge()->GetEdgeNext(1)->GetFacet() == m_pfaNULL &&
		   pe->GetEdge()->GetEdgePrev(1)->GetEdge()->GetFacet() == m_pfaNULL )
		{
		  pme->m_plistQEEdge->Remove(ppe[3]->GetEdge());
		  pme->m_plistQEEdgeBoundary->InsertLast(ppe[3]->GetEdge());
		}
		*/

	      pe->SetForwardEdge(ppe[0]->GetEdge());
	      pe->GetEdge()->SetForwardEdge(ppe[2]->GetEdge());

	      j = 3;
	      for ( i = 0; i < 4; i++ )
		{
		  ppe[i]->SetFacet(pfaNew);
		  pme->Splice(1, ppe[j]->GetEdge(), ppe[i]);
		  j = i;
		}

	      fWeight = pe->GetKnotInterval() / 2.0;
	      ppe[0]->GetEdge()->SetKnotInterval(fWeight);
	      ppe[2]->SetKnotInterval(fWeight);
	      fWeight = pe->GetEdge()->GetKnotInterval() / 2.0;
	      ppe[0]->SetKnotInterval(fWeight);
	      ppe[2]->GetEdge()->SetKnotInterval(fWeight);

	      // Constant Refinement Matrix
	      //fWeight = pe->GetKnotInterval() / 2.0;
	      // More pleasing shape
	      fWeight = (pe->GetEdgePrev(1)->GetKnotInterval() + 
			 pe->GetEdgeNext(1)->GetKnotInterval()) / 4.0;
	      ppe[1]->SetKnotInterval(fWeight);
	      ppe[1]->GetEdge()->SetKnotInterval(fWeight);

	      // Constant Refinement Matrix
	      //fWeight = pe->GetEdge()->GetKnotInterval() / 2.0;
	      // More pleasing shape
	      fWeight = (pe->GetEdge()->GetEdgePrev(1)->GetKnotInterval() + 
			 pe->GetEdge()->GetEdgeNext(1)->GetKnotInterval()) / 4.0;
	      ppe[3]->SetKnotInterval(fWeight);
	      ppe[3]->GetEdge()->SetKnotInterval(fWeight);

	      //////////////////////////////
	      // v0
	      peCurr = pe->GetEdgeNext(0);
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  peCurr = peCurr->GetForwardEdge();
		  pvxNew = peCurr->GetVertex();
		  pme->Splice(1, ppe[0]->GetEdgePrev(1), peCurr);
		}
	      else
		{
		  pt.Null();
		  pvxNew = pme->CreateVertex(pt);
		}
	      ppe[0]->GetEdgeNext(1)->SetVertex(pvxNew);
	      ppe[0]->SetVertex(pvxNew);

	      //////////////////////////////
	      // v1
	      peCurr = pe->GetEdgePrev(0);
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  peCurr = peCurr->GetForwardEdge()->GetEdge()->GetEdgePrev(1);
		  pvxNew = peCurr->GetVertex();
		  pme->Splice(1, ppe[1]->GetEdgePrev(1), peCurr);
		}
	      else
		{
		  pt.Null();
		  pvxNew = pme->CreateVertex(pt);
		}
	      ppe[1]->GetEdgeNext(1)->SetVertex(pvxNew);
	      ppe[1]->SetVertex(pvxNew);

	      //////////////////////////////
	      // v2
	      peCurr = pe->GetEdge()->GetEdgeNext(0);
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  peCurr = peCurr->GetForwardEdge();
		  pvxNew = peCurr->GetVertex();
		  pme->Splice(1, ppe[2]->GetEdgePrev(1), peCurr);
		}
	      else
		{
		  pt.Null();
		  pvxNew = pme->CreateVertex(pt);
		}
	      ppe[2]->GetEdgeNext(1)->SetVertex(pvxNew);
	      ppe[2]->SetVertex(pvxNew);

	      //////////////////////////////
	      // v3
	      peCurr = pe->GetEdge()->GetEdgePrev(0);
	      if ( peCurr->GetForwardEdge() != NULL )
		{
		  peCurr = peCurr->GetForwardEdge()->GetEdge()->GetEdgePrev(1);
		  pvxNew = peCurr->GetVertex();
		  pme->Splice(1, ppe[3]->GetEdgePrev(1), peCurr);
		}
	      else
		{
		  pt.Null();
		  pvxNew = pme->CreateVertex(pt);
		}
	      ppe[3]->GetEdgeNext(1)->SetVertex(pvxNew);
	      ppe[3]->SetVertex(pvxNew);
	    }
	}
    }

  // Vertices
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      bBoundaryVertex = FALSE;

      pe = pvx->GetEdge();
      peCurr = pe;
      do
	{
	  if ( peCurr->GetFacet() == m_pfaNULL )
	    {
	      bBoundaryVertex = TRUE;
	      break;
	    }
	  peCurr = peCurr->GetEdgePrev(1);
	}
      while ( peCurr != pe );

      if ( bBoundaryVertex )
	{
	  pe = peCurr;
	  peStart = peCurr;

	  i = 0;
	  do
	    {
	      i++;
	      peCurr = peCurr->GetEdgePrev(1);
	    }
	  while ( peCurr->GetFacet() != m_pfaNULL );
	  ASSERT( peCurr == pe );

	  switch ( i )
	    {
	    case 1:
	      {
		ASSERT(FALSE);
	      }
	      break;
	    case 2:
	      {
	      }
	      break;
	    case 3:
	      {
		pme->m_plistQEEdge->Remove(peStart->GetEdgePrev(1)->GetForwardEdge()->GetEdgePrev(1)->GetEdge());
		pme->m_plistQEEdgeBoundary->InsertLast(peStart->GetEdgePrev(1)->GetForwardEdge()->GetEdgePrev(1)->GetEdge());
	      }
	      break;
	    default:
	      {
		// Only create facets for vertices with full disks

		peNew = peStart->GetEdgePrev(1);
		while ( peNew != peCurr->GetEdgeNext(1) )
		  {
		    pme->m_plistQEEdge->Remove(peNew->GetForwardEdge()->GetEdgePrev(1)->GetEdge());
		    pme->m_plistQEEdgeBoundary->InsertLast(peNew->GetForwardEdge()->GetEdgePrev(1)->GetEdge());
		    peNew = peNew->GetEdgePrev(1);
		  }
	      }
	      break;
	      /*
	    default:
	      {
		// Create a facet for each half disk

		peNew = pme->CreateEdge();
		pme->m_plistQEEdge->Remove(peNew->GetEdge());
		pme->m_plistQEEdgeBoundary->InsertLast(peNew->GetEdge());
		
		fWeight = (peStart->GetKnotInterval() + peCurr->GetEdgeNext(1)->GetKnotInterval()) / 4.0;
		peNew->SetKnotInterval(fWeight);
		peNew->GetEdge()->SetKnotInterval(fWeight);

		peNew->SetVertex(peStart->GetEdge()->GetForwardEdge()->GetEdge()->GetVertex());
		peNew->GetEdge()->SetVertex(peCurr->GetEdgeNext(1)->GetForwardEdge()->GetEdgePrev(1)->GetVertex());

		Splice(1, peNew, peStart->GetEdge()->GetForwardEdge()->GetEdge());
		Splice(1, peNew->GetEdge(), peCurr->GetEdgeNext(1)->GetForwardEdge()->GetEdgePrev(1));

		pfaNew = pme->CreateFacet();
		peStart = peNew;
		do
		  {
		    peNew->SetFacet(pfaNew);
		    peNew = peNew->GetEdgeNext(0);
		  }
		while ( peNew != peStart );

		peStart = peCurr;
	      }
	      break;
	      */
	    }
	}
      else
	{
	  pfaNew = pme->CreateFacet();
	  
	  iterQEVertex.Init(pvx);
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      pe->GetForwardEdge()->GetEdgePrev(1)->GetEdge()->SetFacet(pfaNew);
	    }
	}
    }

  // Facets
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfaNew = pme->CreateFacet();

      iterQEFacet.Init(pfa);
      fWeights = 0.0;
      uiN = 0;
      ptV.Null();
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  uiN++;
	  fWeight = pe->GetEdge()->GetKnotInterval() * pe->GetEdgePrev(0)->GetKnotInterval();
	  ptV += (fWeight * pe->GetVertex()->GetPoint());
	  fWeights += fWeight;
	}
      ptV = ptV / fWeights;

      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  pe->GetForwardEdge()->SetFacet(pfaNew);

	  i = 0;
	  ptSum.Null();
	  peCurr = pe;
	  do
	    {
	      fWeight = 1.0 + 2.0 * cos( 2.0 * PI * i / (FLOAT)uiN );
	      ptSum += (fWeight * peCurr->GetVertex()->GetPoint());

	      i++;
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( peCurr != pe );

	  fWeight = ( ( pe->GetEdgeNext(0)->GetKnotInterval() * 
			pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetKnotInterval() ) +
		      ( pe->GetEdgePrev(0)->GetEdgePrev(0)->GetEdge()->GetKnotInterval() * 
			pe->GetEdgePrev(0)->GetEdgePrev(0)->GetEdgePrev(0)->GetKnotInterval() ) );

	  pt = ( ( ( ptV + pe->GetVertex()->GetPoint() ) / 2.0 ) + 
		 ( ( fWeight / (8.0 * fWeights) ) * 
		   ( ptSum - ((FLOAT)uiN * pe->GetVertex()->GetPoint()) ) ) );

	  pt[3] = 1.0;
	  pe->GetForwardEdge()->GetVertex()->SetPoint(pt);
	}
    }

  pme->Preprocess();

  pme->CalculateSubdivisionLimitPositionsAndNormalsDooSabin();

  return pme;
}

///////////////////////////////////////////////////
// QEMesh::SubdivideCatmullClark                 //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     v4--e3--v3--e2--v2    //
//           |              |       |       |    //
//           |             e4      e9      e1    //
//           |              |       |       |    //
//  e2------vV------e0  => v5--e10-vV--e8--v1    //
//           |              |       |       |    //
//           |             e5      e11     e0    //
//           |              |       |       |    //
//  f2      e3      f3     v6--e6--v7--e7--v0    //
//                                               //
QEMesh *QEMesh::SubdivideCatmullClark()
{
  //ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  // Maximum number of sides on the Facet that this code will handle
  const UINT_32 cuiSidesMax = 20;
  const UINT_32 cuiNewVerticesMax = 2*cuiSidesMax;
  const UINT_32 cuiNewEdgesMax = 3*cuiSidesMax;
  QEVertex *ppvx[cuiNewVerticesMax], *pvxVertex;
  QEEdge *ppe[cuiNewEdgesMax];
  UINT_32 uiSides;

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *pePrev;
  QEVertex *pvx, *pvxPrev;
  FLOAT fKnotInterval;

  UINT_32 i, j, k;
  BOOL bBoundaryVertex;

  Point pt;
  //Point ptFacet;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      memset(ppvx, 0, sizeof(ppvx));
      memset(ppe, 0, sizeof(ppe));

      iterQEVertex.Init(pvx);

      bBoundaryVertex = FALSE;
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe->GetFacet() == m_pfaNULL )
	    {
	      bBoundaryVertex = TRUE;
	      break;
	    }
	}
      if ( bBoundaryVertex )
	{
	  continue;
	}

      pvxPrev = NULL;
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0, uiSides = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i += 2, uiSides++ )
	{
	  ASSERT( uiSides < cuiSidesMax );

	  if ( pvxPrev != NULL )
	    {
	      ppvx[i] = pvxPrev;
	      pvxPrev = NULL;
	    }

	  ppe[i] = pe->GetEdge()->GetForwardEdge();
	  ppe[i+1] = pe->GetForwardEdge();
	  ASSERT( (ppe[i] == NULL && ppe[i+1] == NULL) ||
		  (ppe[i] != NULL && ppe[i+1] != NULL) );
	  if ( ppe[i] != NULL )
	    {
	      ppvx[i] = ppe[i]->GetVertex();
	      ppvx[i+1] = ppe[i]->GetEdge()->GetVertex();
	      pvxPrev = ppe[i+1]->GetVertex();
	    }
	}
      if ( pvxPrev != NULL )
	{
	  ppvx[0] = pvxPrev;
	}

      ASSERT( 2 < uiSides && uiSides <= cuiSidesMax );

      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i += 2 )
	{
	  ASSERT( (ppe[i] == NULL && ppe[i+1] == NULL) ||
		  (ppe[i] != NULL && ppe[i+1] != NULL) );
	  if ( ppe[i] == NULL )
	    {
	      if ( ppvx[i] == NULL )
		{
		  pfa = pe->GetEdge()->GetFacet();
		  ASSERT( pfa != m_pfaNULL );
		  ppvx[i] = pfa->GetForwardVertex();
		  if ( ppvx[i] == NULL )
		    {
		      pt = CalculateFacetSubdivisionCatmullClark(pfa);

		      ppvx[i] = pme->CreateVertex(pt);
		      ASSERT( ppvx[i] != NULL );
		      pfa->SetForwardVertex(ppvx[i]);
		    }
		}

	      if ( i == 2*(uiSides-1) )
		{
		  j = 0;
		}
	      else
		{
		  j = i + 2;
		}
	      if ( ppvx[j] == NULL )
		{
		  ASSERT( j != 0 );
		  pfa = pe->GetFacet();
		  ASSERT( pfa != NULL );
		  ppvx[j] = pfa->GetForwardVertex();
		  if ( ppvx[j] == NULL )
		    {
		      pt = CalculateFacetSubdivisionCatmullClark(pfa);

		      ppvx[j] = pme->CreateVertex(pt);
		      ASSERT( ppvx[j] != NULL );
		      pfa->SetForwardVertex(ppvx[j]);
		    }
		}

	      // Create the midpoint
	      pt = CalculateEdgeSubdivisionCatmullClark(pe);
	      ASSERT( ppvx[i+1] == NULL );
	      ppvx[i+1] = pme->CreateVertex(pt);
	      ASSERT( ppvx[i+1] != NULL );

	      // Create the two subdivided edges
	      ppe[i] = pme->CreateEdge();
	      ASSERT( ppe[i] != NULL );
	      pe->GetEdge()->SetForwardEdge(ppe[i]);
	      if ( pe->GetEdge()->GetEdgePrev(0)->GetEdge()->GetFacet() == m_pfaNULL )
		{
		  pme->m_plistQEEdge->Remove(ppe[i]->GetEdge());
		  pme->m_plistQEEdgeBoundary->InsertLast(ppe[i]->GetEdge());
		}

	      ppe[i+1] = pme->CreateEdge();
	      ASSERT( ppe[i+1] != NULL );
	      pe->SetForwardEdge(ppe[i+1]);
	      if ( pe->GetEdgeNext(0)->GetEdge()->GetFacet() == m_pfaNULL )
		{
		  pme->m_plistQEEdge->Remove(ppe[i+1]);
		  pme->m_plistQEEdgeBoundary->InsertLast(ppe[i+1]);
		}

	      //////////////////////////////
	      // Jordan Check This!!!!
	      fKnotInterval = (pe->GetEdge()->GetEdgePrev(0)->GetKnotInterval() + 
			       pe->GetEdge()->GetEdgeNext(0)->GetKnotInterval()) / 4.0;
	      ppe[i]->SetKnotInterval(fKnotInterval);
	      ppe[i]->GetEdge()->SetKnotInterval(fKnotInterval);
	      fKnotInterval = (pe->GetEdgePrev(0)->GetKnotInterval() + 
			       pe->GetEdgeNext(0)->GetKnotInterval()) / 4.0;
	      ppe[i+1]->SetKnotInterval(fKnotInterval);
	      ppe[i+1]->GetEdge()->SetKnotInterval(fKnotInterval);
	      // Jordan Check This!!!!
	      //////////////////////////////
	      
	      //////////////////////////////
	      // Facet Centric Way
	      //fKnotInterval = pe->GetKnotInterval() / 2.0;
	      //ppe[i]->SetKnotInterval(fKnotInterval);
	      //ppe[i]->GetEdge()->SetKnotInterval(fKnotInterval);
	      //ppe[i+1]->SetKnotInterval(fKnotInterval);
	      //ppe[i+1]->GetEdge()->SetKnotInterval(fKnotInterval);
	      // Facet Centric Way
	      //////////////////////////////

	      ppe[i]->SetVertex(ppvx[i]);
	      ppe[i]->GetEdge()->SetVertex(ppvx[i+1]);
	      ppe[i+1]->GetEdge()->SetVertex(ppvx[i+1]);
	      ppe[i+1]->SetVertex(ppvx[j]);

	      Splice(1, ppe[i]->GetEdge(), ppe[i+1]->GetEdge());
	    }
	}

      // Create the Vertex Vertex
      pt = CalculateVertexSubdivisionCatmullClark(pvx);
      pvxVertex = pme->CreateVertex(pt);
      ASSERT( pvxVertex != NULL );

      // Create the internal edges
      pePrev = NULL;
      k = 2*(uiSides-1) + 1;
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0, j = 2*uiSides;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i += 2, j++ )
	{
	  ASSERT( i < 2*uiSides );

	  // Splice the corner edges together
	  Splice(1, ppe[k]->GetEdgePrev(1), ppe[i]);

	  // Create the internal edge and splice it in
	  ppe[j] = pme->CreateEdge();
	  ASSERT( ppe[j] != NULL );

	  //////////////////////////////
	  // Jordan Check This!!!!
	  fKnotInterval = pe->GetKnotInterval() / 2.0;
	  ppe[j]->SetKnotInterval(fKnotInterval);
	  ppe[j]->GetEdge()->SetKnotInterval(fKnotInterval);
	  // Jordan Check This!!!!
	  //////////////////////////////

	  //////////////////////////////
	  // Facet Centric Way
	  //fKnotInterval = (pe->GetEdgePrev(0)->GetKnotInterval() + 
	  //		   pe->GetEdgeNext(0)->GetKnotInterval()) / 4.0;
	  //ppe[j]->SetKnotInterval(fKnotInterval);
	  //ppe[j]->GetEdge()->SetKnotInterval(fKnotInterval);
	  // Facet Centric Way
	  //////////////////////////////

	  ppe[j]->SetVertex(ppvx[i+1]);
	  ppe[j]->GetEdge()->SetVertex(pvxVertex);

	  Splice(1, ppe[j], ppe[i+1]->GetEdge());
	  if ( pePrev != NULL )
	    {
	      Splice(1, pePrev->GetEdge(), ppe[j]->GetEdge());
	    }

	  pePrev = ppe[j];
	  k = i + 1;
	}
      ASSERT( i == 2*uiSides );

      //////////////////////////////////////////////////
      // Create the Facets

      for ( i = 0; i < (2*uiSides); i += 2)
	{	  
	  // Create the new facet
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );

	  pe = ppe[i];
	  do
	    {
	      pe->SetFacet(pfaNew);

	      pe = pe->GetEdgeNext(0);
	    }
	  while ( pe != ppe[i] );
	}
    }

  pme->Preprocess();

  pme->CalculateSubdivisionLimitPositionsAndNormalsCatmullClark();

  return pme;
}

///////////////////////////////////////////////////
// QEMesh::SubdivideSharpCatmullClark            //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     v4--e3--v3--e2--v2    //
//           |              |       |       |    //
//           |             e4      e9      e1    //
//           |              |       |       |    //
//  e2------vV------e0  => v5--e10-vV--e8--v1    //
//           |              |       |       |    //
//           |             e5      e11     e0    //
//           |              |       |       |    //
//  f2      e3      f3     v6--e6--v7--e7--v0    //
//                                               //
QEMesh *QEMesh::SubdivideSharpCatmullClark()
{
  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEVertexIter iterQEVertex;
  QEFacet *pfaNew;
  QEEdge *pe, *peNew, *peCurr;
  QEVertex *pvx, *pvxNew;
  FLOAT fSharpness;
  Point pt;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = CalculateVertexSubdivisionSharpCatmullClark(pvx);
      pvxNew = pme->CreateVertex(pt);
      pvxNew->SetSharpVertexType(pvx->GetSharpVertexType());
      pvx->SetForwardVertex(pvxNew);

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  ASSERT( pe->GetForwardEdge() == NULL );

	  ASSERT( pe->GetKnotInterval() == pe->GetEdge()->GetKnotInterval() );
	  fSharpness = pe->GetKnotInterval();
	  if ( fSharpness != INFINITY )
	    {
	      fSharpness = MAX(0.0, fSharpness - 1.0);
	    }

	  peNew = pme->CreateEdge();
	  peNew->SetVertex(pvx->GetForwardVertex());
	  peNew->SetKnotInterval(fSharpness);
	  peNew->GetEdge()->SetKnotInterval(fSharpness);
	  pe->SetForwardEdge(peNew);

	  if ( pe->GetEdgePrev(1)->GetForwardEdge() != NULL )
	    {
	      Splice(1, pe->GetEdgePrev(1)->GetForwardEdge(), pe->GetForwardEdge());
	    }

	  if ( pe->GetEdge()->GetForwardEdge() == NULL )
	    {
	      pt = CalculateEdgeSubdivisionSharpCatmullClark(pe);
	      pvxNew = pme->CreateVertex(pt);
	      if ( pe->GetFacet() == GetNullFacet() ||
		   pe->GetEdge()->GetFacet() == GetNullFacet() )
		{
		  pvxNew->SetSharpVertexType(svtBoundaryCrease);
		}
	      else if ( pe->GetKnotInterval() == INFINITY )
		{
		  pvxNew->SetSharpVertexType(svtRegularCrease);
		}
	      else
		{
		  pvxNew->SetSharpVertexType(svtSmooth);
		}
	      pe->GetForwardEdge()->GetEdge()->SetVertex(pvxNew);

	      if ( pe->GetFacet() != pme->GetNullFacet() )
		{
		  peNew = pme->CreateEdge();
		  peNew->SetKnotInterval(0.0);
		  peNew->GetEdge()->SetKnotInterval(0.0);
		  peNew->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());
		  Splice(1, peNew, pe->GetForwardEdge()->GetEdge());

		  peCurr = pe->GetEdgePrev(0);
		  if ( peCurr->GetEdge()->GetForwardEdge() != NULL )
		    {
		      peNew->GetEdge()->SetVertex(peCurr->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgePrev(0)->GetVertex());
		      Splice(1, peNew->GetEdge(), peCurr->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgePrev(0));
		    }
		  else
		    {
		      for ( peCurr = pe->GetEdgePrev(0);
			    peCurr != pe;
			    peCurr = peCurr->GetEdgePrev(0) )
			{
			  if ( peCurr->GetForwardEdge() != NULL )
			    {
			      peNew->GetEdge()->SetVertex(peCurr->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetVertex());
			      Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetEdgeNext(0)->GetEdge());
			      break;
			    }
			}
		      if ( peCurr == pe )
			{
			  pt = CalculateFacetSubdivisionSharpCatmullClark(pe->GetFacet());
			  pvxNew = pme->CreateVertex(pt);
			  pvxNew->SetSharpVertexType(svtSmooth);
			  peNew->GetEdge()->SetVertex(pvxNew);
			}
		    }
		}

	      if ( pe->GetEdge()->GetFacet() != pme->GetNullFacet() )
		{
		  peNew = pme->CreateEdge();
		  peNew->SetKnotInterval(0.0);
		  peNew->GetEdge()->SetKnotInterval(0.0);
		  peNew->SetVertex(pe->GetForwardEdge()->GetEdge()->GetVertex());
		  Splice(1, peNew, pe->GetForwardEdge()->GetEdge());

		  peCurr = pe->GetEdge()->GetEdgeNext(0);
		  if ( peCurr->GetForwardEdge() != NULL )
		    {
		      peNew->GetEdge()->SetVertex(peCurr->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetVertex());
		      Splice(1, peNew->GetEdge(), peCurr->GetForwardEdge()->GetEdgeNext(0)->GetEdge()->GetEdgePrev(1));
		    }
		  else
		    {
		      for ( peCurr = pe->GetEdge()->GetEdgeNext(0);
			    peCurr != pe->GetEdge();
			    peCurr = peCurr->GetEdgeNext(0) )
			{
			  if ( peCurr->GetEdge()->GetForwardEdge() != NULL )
			    {
			      peNew->GetEdge()->SetVertex(peCurr->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgePrev(0)->GetVertex());
			      Splice(1, peNew->GetEdge(), peCurr->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgePrev(0)->GetEdgePrev(1));
			      break;
			    }
			}
		      if ( peCurr == pe->GetEdge() )
			{
			  pt = CalculateFacetSubdivisionSharpCatmullClark(pe->GetEdge()->GetFacet());
			  pvxNew = pme->CreateVertex(pt);
			  pvxNew->SetSharpVertexType(svtSmooth);
			  peNew->GetEdge()->SetVertex(pvxNew);
			}
		    }
		}
	    }
	  else
	    {
	      peNew->GetEdge()->SetVertex(pe->GetEdge()->GetForwardEdge()->GetEdge()->GetVertex());
	      if ( pe->GetFacet() != pme->GetNullFacet() )
		{
		  Splice(1, pe->GetEdge()->GetForwardEdge()->GetEdge()->GetEdgeNext(1), peNew->GetEdge());
		}
	      else
		{
		  Splice(1, pe->GetEdge()->GetForwardEdge()->GetEdge(), peNew->GetEdge());
		}
	    }
	}

      // Create the facets
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  if ( pe->GetFacet() != pme->GetNullFacet() )
	    {
	      pfaNew = pme->CreateFacet();
	      peCurr = pe->GetForwardEdge();
	      do
		{
		  peCurr->SetFacet(pfaNew);

		  peCurr = peCurr->GetEdgeNext(0);
		}
	      while ( peCurr != pe->GetForwardEdge() );
	      
	    }
	}
    }

  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      ASSERT( pe->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge() != NULL );
      ASSERT( pe->GetEdge()->GetForwardEdge()->GetFacet() != pme->GetNullFacet() );

      pe->GetForwardEdge()->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge());
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge());

      pe->GetForwardEdge()->GetEdgeNext(0)->SetFacet(pme->GetNullFacet());
      pme->m_plistQEEdge->Remove(pe->GetForwardEdge()->GetEdgeNext(0));
      pme->m_plistQEEdgeBoundary->InsertLast(pe->GetForwardEdge()->GetEdgeNext(0));
    }

  pme->Preprocess();
  pme->CalculateSubdivisionLimitPositionsAndNormalsSharpCatmullClark();

  return pme;
}

QEMesh *QEMesh::SubdivideCatmullClarkFacetCentric()
{
  return SubdivideCatmullClarkFacetCentric(stCatmullClark);
}

QEMesh *QEMesh::SubdivideCatmullClarkTesselate()
{
  return SubdivideCatmullClarkFacetCentric(stTesselate);
}

QEMesh *QEMesh::SubdivideCatmullClarkInterpolate()
{
  return SubdivideCatmullClarkFacetCentric(stButterfly);
}

FLOAT gfCatmullClarkC2GenFactor = 0.0;

QEMesh *QEMesh::SubdivideCatmullClarkC2(INT_32 iGen)
{
  QEMesh *pme;

  if ( iGen < 0 )
    {
      gfCatmullClarkC2GenFactor = 0.0;
    }
  else
    {
      gfCatmullClarkC2GenFactor = pow(2.0, -iGen);
    }

  pme = SubdivideCatmullClarkFacetCentric(stCatmullClarkC2);
  pme->CalculateSubdivisionLimitPositionsAndNormalsCatmullClarkC2();

  return pme;
}

///////////////////////////////////////////////////
// QEMesh::SubdivideCatmullClarkFacetCentric     //
//                                               //
//       Old Quad          Subdivided Facets     //
//                                               //
//  v3------e2------v2     v6--e5--v5--e4--v4    //
//   |               |      |       |       |    //
//   |               |     e6      e10     e3    //
//   |               |      |       |       |    //
//  e3              e1  => v7--e11-vF--e9--v3    //
//   |               |      |       |       |    //
//   |               |     e7      e8      e2    //
//   |               |      |       |       |    //
//  v0------e0------v1     v0--e0--v1--e1--v2    //
//                                               //
QEMesh *QEMesh::SubdivideCatmullClarkFacetCentric(SubdivisionType eSubdivisionType)
{
  ASSERT( m_plistQEEdgeBoundary->GetLength() == 0 );

  // Maximum number of sides on the Facet that this code will handle
  const UINT_32 cuiSidesMax = 20;
  const UINT_32 cuiNewVerticesMax = 2*cuiSidesMax;
  const UINT_32 cuiNewEdgesMax = 3*cuiSidesMax;
  QEVertex *ppvx[cuiNewVerticesMax], *pvxFacet;
  QEEdge *ppe[cuiNewEdgesMax];
  UINT_32 uiSides;

  QEMeshIter iterQEMesh;
  QEMesh *pme;
  QEFacetIter iterQEFacet;
  QEFacet *pfa, *pfaNew;
  QEEdge *pe, *pePrev;
  QEVertex *pvx, *pvxPrev;
  FLOAT fKnotInterval;

  UINT_32 i, j, k;

  Point pt, ptFacet;

  pme = AllocateQEMesh();
  ASSERT( pme != NULL );
  pme->Init();

  // Clear the Subdivision forwarding pointers
  ClearForwardPointers();

  iterQEMesh.Init(this);
  
  // Calculate the valence of the all the vertices
  /*
  QEVertexIter iterQEVertex;

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), i++ )
	{
	}
      pvx->SetValence(i);
    }
  */

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      memset(ppvx, 0, sizeof(ppvx));
      memset(ppe, 0, sizeof(ppe));

      iterQEFacet.Init(pfa);
      pvxPrev = NULL;
      for ( pe = iterQEFacet.PeekFirstEdge(), i = 0, uiSides = 0;
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge(), i += 2, uiSides++ )
	{
	  ASSERT( uiSides < cuiSidesMax );

	  if ( pvxPrev != NULL )
	    {
	      ppvx[i] = pvxPrev;
	      pvxPrev = NULL;
	    }

	  ppe[i] = pe->GetForwardEdge();
	  ppe[i+1] = pe->GetEdge()->GetForwardEdge();
	  ASSERT( (ppe[i] == NULL && ppe[i+1] == NULL) ||
		  (ppe[i] != NULL && ppe[i+1] != NULL) );
	  if ( ppe[i] != NULL )
	    {
	      ppvx[i] = ppe[i]->GetVertex();
	      ppvx[i+1] = ppe[i]->GetEdge()->GetVertex();
	      pvxPrev = ppe[i+1]->GetVertex();
	    }
	}
      if ( pvxPrev != NULL )
	{
	  ppvx[0] = pvxPrev;
	}

      ASSERT( 2 < uiSides && uiSides <= cuiSidesMax );

      for ( pe = iterQEFacet.PeekFirstEdge(), i = 0;
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge(), i += 2 )
	{
	  ASSERT( (ppe[i] == NULL && ppe[i+1] == NULL) ||
		  (ppe[i] != NULL && ppe[i+1] != NULL) );
	  if ( ppe[i] == NULL )
	    {
	      if ( ppvx[i] == NULL )
		{
		  pvx = pe->GetVertex();
		  ppvx[i] = pvx->GetForwardVertex();
		  if ( ppvx[i] == NULL )
		    {
		      pt = CalculateVertexSubdivisionCatmullClark(eSubdivisionType, pvx);

		      ppvx[i] = pme->CreateVertex(pt);
		      ASSERT( ppvx[i] != NULL );
		      pvx->SetForwardVertex(ppvx[i]);
		    }
		}

	      if ( i == 2*(uiSides-1) )
		{
		  j = 0;
		}
	      else
		{
		  j = i + 2;
		}
	      if ( ppvx[j] == NULL )
		{
		  ASSERT( j != 0 );
		  pvx = pe->GetEdge()->GetVertex();
		  ppvx[j] = pvx->GetForwardVertex();
		  if ( ppvx[j] == NULL )
		    {
		      pt = CalculateVertexSubdivisionCatmullClark(eSubdivisionType, pvx);

		      ppvx[j] = pme->CreateVertex(pt);
		      ASSERT( ppvx[j] != NULL );
		      pvx->SetForwardVertex(ppvx[j]);
		    }
		}

	      // Create the midpoint
	      pt = CalculateEdgeSubdivisionCatmullClark(eSubdivisionType, pe);
	      ASSERT( ppvx[i+1] == NULL );
	      ppvx[i+1] = pme->CreateVertex(pt);
	      ASSERT( ppvx[i+1] != NULL );

	      // Create the two subdivided edges
	      ppe[i] = pme->CreateEdge();
	      ASSERT( ppe[i] != NULL );
	      pe->SetForwardEdge(ppe[i]);

	      ppe[i+1] = pme->CreateEdge();
	      ASSERT( ppe[i+1] != NULL );
	      pe->GetEdge()->SetForwardEdge(ppe[i+1]);

	      fKnotInterval = pe->GetKnotInterval() / 2.0;
	      ppe[i]->SetKnotInterval(fKnotInterval);
	      ppe[i]->GetEdge()->SetKnotInterval(fKnotInterval);
	      ppe[i+1]->SetKnotInterval(fKnotInterval);
	      ppe[i+1]->GetEdge()->SetKnotInterval(fKnotInterval);

	      ppe[i]->SetVertex(ppvx[i]);
	      ppe[i]->GetEdge()->SetVertex(ppvx[i+1]);
	      ppe[i+1]->GetEdge()->SetVertex(ppvx[i+1]);
	      ppe[i+1]->SetVertex(ppvx[j]);

	      Splice(1, ppe[i]->GetEdge(), ppe[i+1]->GetEdge());
	    }
	}

      // Create the Facet Vertex
      ptFacet = CalculateFacetSubdivisionCatmullClark(eSubdivisionType, pfa);
      //ptFacet = pDTFacet->GetForwardPoint();
      pvxFacet = pme->CreateVertex(ptFacet);
      ASSERT( pvxFacet != NULL );

      // Create the internal edges
      pePrev = NULL;
      k = 2*(uiSides-1) + 1;
      for ( pe = iterQEFacet.PeekFirstEdge(), i = 0, j = 2*uiSides;
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge(), i += 2, j++ )
	{
	  ASSERT( i < 2*uiSides );

	  // Splice the corner edges together
	  Splice(1, ppe[k]->GetEdgePrev(1), ppe[i]);

	  // Create the internal edge and splice it in
	  ppe[j] = pme->CreateEdge();
	  ASSERT( ppe[j] != NULL );

	  fKnotInterval = (pe->GetEdgePrev(0)->GetKnotInterval() + 
			   pe->GetEdgeNext(0)->GetKnotInterval()) / 4.0;
	  ppe[j]->SetKnotInterval(fKnotInterval);
	  ppe[j]->GetEdge()->SetKnotInterval(fKnotInterval);

	  ppe[j]->SetVertex(ppvx[i+1]);
	  ppe[j]->GetEdge()->SetVertex(pvxFacet);

	  Splice(1, ppe[j], ppe[i+1]->GetEdge());
	  if ( pePrev != NULL )
	    {
	      Splice(1, pePrev->GetEdge(), ppe[j]->GetEdge());
	    }

	  pePrev = ppe[j];
	  k = i + 1;
	}
      ASSERT( i == 2*uiSides );

      //////////////////////////////////////////////////
      // Create the Facets

      for ( i = 0; i < (2*uiSides); i += 2)
	{	  
	  // Create the new facet
	  pfaNew = pme->CreateFacet();
	  ASSERT( pfaNew != NULL );

	  pe = ppe[i];
	  do
	    {
	      pe->SetFacet(pfaNew);

	      pe = pe->GetEdgeNext(0);
	    }
	  while ( pe != ppe[i] );
	}
    }

  pme->Preprocess();

  return pme;
}

Point QEMesh::CalculateFacetSubdivisionCatmullClark(SubdivisionType eSubdivisionType, QEFacet *pfa)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateFacetSubdivisionCatmullClark(pfa);
      break;
    case stCatmullClark:
      pt = CalculateFacetSubdivisionCatmullClark(pfa);
      break;
    case stCatmullClarkC2:
      pt = CalculateFacetSubdivisionCatmullClarkC2(pfa);
      break;
    case stButterfly:
      //pt = CalculateFacetSubdivisionCatmullClark(pfa);
      pt = CalculateFacetSubdivisionCatmullClarkButterfly(pfa);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point CalculateVertexSubdivisionCatmullClarkNew(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  Point pt, ptV, ptE, ptF;
  FLOAT fA, fB, fC;
  UINT_32 uiN;

  const UINT_32 cuiNMax = 10;
  const FLOAT cpfB[cuiNMax+1] = { 0.0, 0.0000, 0.0000, 0.4381, 0.3750, 0.3149, 0.2758, 0.2562, 0.2500, 0.2518, 0.2580 };
  const FLOAT cpfC[cuiNMax+1] = { 0.0, 0.0000, 0.0000, 0.1087, 0.0625, 0.0269, -0.0042, -0.0341, -0.0625, -0.0889, -0.1130 };

  pt.Null(); 

  ptV = pvx->GetPoint();
  ptE.Null();
  ptF.Null();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      peCurr = pe->GetEdgeNext(0);
      ptE += peCurr->GetVertex()->GetPoint();
      for ( peCurr = peCurr->GetEdgeNext(0);
	    peCurr != pe->GetEdgePrev(0);
	    peCurr = peCurr->GetEdgeNext(0) )
	{
	  ptF += peCurr->GetVertex()->GetPoint();
	}
    }
  ASSERT( uiN > 2 );

  if ( uiN <= cuiNMax )
    {
      fB = cpfB[uiN];
      fC = cpfC[uiN];
      fA = 1.0 - fB - fC;
      fB /= (FLOAT)(uiN);
      fC /= (FLOAT)(uiN);

      pt = fA*ptV + fB*ptE + fC*ptF;
      pt[3] = 1.0;
    }
  else
    {
      pt = ptV;
    }

  return pt;
}

Point CalculateVertexSubdivisionCatmullClarkNewSquared(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  Point pt, ptV, ptE, ptF;
  FLOAT fA, fB, fC;
  UINT_32 uiN;

  const UINT_32 cuiNMax = 10;
  const FLOAT cpfB[cuiNMax+1] = { 0.0, 0.0000, 0.0000, 0.4717, 0.3750, 0.1853, -0.0518, -0.2884, -0.5000, -0.6803, -0.8310 };
  const FLOAT cpfC[cuiNMax+1] = { 0.0, 0.0000, 0.0000, 0.1027, 0.0625, 0.1089, 0.2235, 0.3631, 0.5000, 0.6227, 0.7285 };

  pt.Null(); 

  ptV = pvx->GetPoint();
  ptE.Null();
  ptF.Null();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      peCurr = pe->GetEdgeNext(0);
      ptE += peCurr->GetVertex()->GetPoint();
      for ( peCurr = peCurr->GetEdgeNext(0);
	    peCurr != pe->GetEdgePrev(0);
	    peCurr = peCurr->GetEdgeNext(0) )
	{
	  ptF += peCurr->GetVertex()->GetPoint();
	}
    }
  ASSERT( uiN > 2 );

  if ( uiN <= cuiNMax )
    {
      fB = cpfB[uiN];
      fC = cpfC[uiN];
      fA = 1.0 - fB - fC;
      fB /= (FLOAT)(uiN);
      fC /= (FLOAT)(uiN);

      pt = fA*ptV + fB*ptE + fC*ptF;
      pt[3] = 1.0;
    }
  else
    {
      pt = ptV;
    }

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCatmullClark(SubdivisionType eSubdivisionType, QEVertex *pvx)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateVertexSubdivisionCatmullClarkNew(pvx);
      //pt = CalculateVertexSubdivisionTesselate(pvx);
      break;
    case stCatmullClark:
      pt = CalculateVertexSubdivisionCatmullClark(pvx);
      break;
    case stCatmullClarkC2:
      pt = CalculateVertexSubdivisionCatmullClarkC2(pvx);
      break;
    case stButterfly:
      //pt = CalculateVertexSubdivisionCatmullClarkNewSquared(pvx);
      pt = CalculateVertexSubdivisionCatmullClarkButterfly(pvx);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionCatmullClark(SubdivisionType eSubdivisionType, QEEdge *pe)
{
  Point pt;

  switch ( eSubdivisionType )
    {
    case stTesselate:
      pt = CalculateEdgeSubdivisionCatmullClark(pe);
      //pt = CalculateEdgeSubdivisionTesselate(pe);
      break;
    case stCatmullClark:
      pt = CalculateEdgeSubdivisionCatmullClark(pe);
      break;
    case stCatmullClarkC2:
      pt = CalculateEdgeSubdivisionCatmullClarkC2(pe);
      break;
    case stButterfly:
      //pt = CalculateEdgeSubdivisionCatmullClark(pe);
      pt = CalculateEdgeSubdivisionCatmullClarkButterfly(pe);
      break;
    default:
      ASSERT(FALSE);
      break;
    }

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionCatmullClark(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN;

  iterQEFacet.Init(pfa);
  for ( pvx = iterQEFacet.PeekFirstVertex(), uiN = 0;
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex(), uiN++ )
    {
      pt += pvx->GetPoint();
    }
  pt /= (FLOAT)uiN;

  //pt = pfa->CalculatePointCatmullClark();

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCatmullClark(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt, ptM, ptF;
  FLOAT pfM[2], fM;
  FLOAT pfF[2], fF;
  FLOAT fDen, fC;
  UINT_32 uiN;

  pt.Null(); 
  fDen = 0.0;

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      ptF = CalculateFacetSubdivisionCatmullClark(pe->GetFacet());
      ptM = pe->CalculateMidpointCatmullClark();

      pfM[0] = pe->GetEdgeNext(1)->GetKnotInterval();
      pfM[0] += pe->GetEdgePrev(1)->GetKnotInterval();

      pfM[1] = pe->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval();
      pfM[1] += pe->GetEdgePrev(1)->GetEdgePrev(1)->GetKnotInterval();

      fM = pfM[0]*pfM[1]/2.0;

      pfF[0] = pe->GetEdgeNext(1)->GetKnotInterval();
      pfF[1] = pe->GetKnotInterval();

      fF = pfF[0]*pfF[1];

      pt += (fM * ptM) + (fF * ptF);
      fDen += fM + fF;
    }
  ASSERT( uiN > 2 );

  fDen *= (FLOAT)uiN;
  if ( fDen == 0.0 )
    {
      fC = 1.0;
    }
  else
    {
      fC = (FLOAT)(uiN - 3) / (FLOAT)uiN;
    }

  pt = (fC * pvx->GetPoint()) + ((3.0 / fDen) * pt);
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionCatmullClark(QEVertex *pvx)
{
  Point pt, ptE, ptF;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fDen;

  iter.Init(pvx);

  // Catmull-Clark Vertex-Vertex Subdivision
  pt = pvx->GetPoint();
  ptE.Null();
  ptF.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      ptF += CalculateFacetSubdivisionCatmullClark(pe->GetFacet());
      ptE += CalculateEdgeSubdivisionCatmullClark(pe);
    }
  
  fDen = (FLOAT)(uiNeighbors * (uiNeighbors + 5));
  pt = (((FLOAT)(uiNeighbors*uiNeighbors)*pt) + (4.0*ptE) + ptF) / fDen; 
  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCatmullClark(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, vE, vF;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fTheta, fThetaNext, fThetaDelta, fA, fE, fF;

  iter.Init(pvx);

  // Catmull-Clark Vertex-Normal Subdivision
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
    }

  fA = ( 1.0 + cos((2.0 * PI)/(FLOAT)uiNeighbors) + 
	 cos(PI/(FLOAT)uiNeighbors)*sqrt(2.0*(9.0 + cos((2.0 * PI)/(FLOAT)uiNeighbors))) );

  vT0.Null();
  vT1.Null();
  fTheta = 0.0;
  fThetaDelta = (2.0 * PI) / (FLOAT)uiNeighbors;
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      vE = CalculateEdgeSubdivisionCatmullClark(pe);
      vF = CalculateFacetSubdivisionCatmullClark(pe->GetFacet());

      fThetaNext = fTheta + fThetaDelta;
      fE = fA * cos(fTheta);
      fF = cos(fTheta) + cos(fThetaNext);
      vT0 += (fE*vE + fF*vF);

      fTheta = fThetaNext;

      fThetaNext = fTheta + fThetaDelta;
      fE = fA * cos(fTheta);
      fF = cos(fTheta) + cos(fThetaNext);
      vT1 += (fE*vE + fF*vF);
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;

  vNormal = vT1 * vT0;
  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionCatmullClark(QEEdge *pe)
{
  Point pt, ptM, pptF[2];
  QEEdge *ppe[2];
  FLOAT pfA[2], fDen;

  ptM = pe->CalculateMidpointCatmullClark();

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  pfA[0] = ppe[0]->GetEdge()->GetEdgeNext(1)->GetKnotInterval();
  pfA[0] += ppe[0]->GetEdgePrev(1)->GetKnotInterval();
  pfA[1] = ppe[1]->GetEdge()->GetEdgeNext(1)->GetKnotInterval();
  pfA[1] += ppe[1]->GetEdgePrev(1)->GetKnotInterval();

  fDen = 2.0 * ( pfA[0] + pfA[1] );
  pfA[0] /= fDen;
  pfA[1] /= fDen;

  pptF[0] = CalculateFacetSubdivisionCatmullClark(ppe[0]->GetFacet());
  pptF[1] = CalculateFacetSubdivisionCatmullClark(ppe[1]->GetFacet());

  pt = (( 1.0 - pfA[0] - pfA[1] ) * ptM) + (pfA[0] * pptF[0]) + (pfA[1] * pptF[1]);
  pt[3] = 1.0;

  return pt;
}

//////////////////////////////////////////////////////////////////////
// Jordan's new C2 Catmull-Clark guided by Loop for Val 3 & 6

Point QEMesh::CalculateFacetSubdivisionCatmullClarkC2(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe;
  QEVertex *pvx;
  Point pt, ptLoop, ptE;
  UINT_32 uiN, uiLoopVertices;

  uiLoopVertices = 0;

  pt.Null();
  ptLoop.Null();

  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge(), uiN++ )
    {
      pvx = pe->GetVertex();
      pt += pvx->GetPoint();

      ASSERT( pvx->GetValence() > 2 );
      switch ( pvx->GetValence() )
	{
	  // Regular Loop cases
	case 3:
	  uiLoopVertices++;
	  ptE = ( 6.0 * pe->GetVertex()->GetPoint() +
		  3.0 * (pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
		  2.0 * (pe->GetEdge()->GetVertex()->GetPoint() +
			 pe->GetEdgePrev(0)->GetVertex()->GetPoint()) ) / 16.0;
	  ptLoop += ptE;
	  break;
	case 6:
	  uiLoopVertices++;
	  ptE = ( 2.0 * pe->GetVertex()->GetPoint() +
		  1.0 * (pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
		  6.0 * (pe->GetEdge()->GetVertex()->GetPoint() +
			 pe->GetEdgePrev(0)->GetVertex()->GetPoint()) ) / 16.0;
	  ptLoop += ptE;
	  break;
	default:
	  break;
	}
    }

  pt /= (FLOAT)uiN;
  if ( uiLoopVertices > 0 )
    {
      ptLoop /= (FLOAT)uiLoopVertices;
      pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
    }

  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCatmullClarkC2(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  Point pt, ptE, ptF, ptLoop;
  UINT_32 uiN, i, uiLoopVertices;

  //fprintf(stderr, "Vertex %p -- Val == %ld\n", pvx, pvx->GetValence());

  pt.Null(); 
  ptLoop.Null(); 

  // Catmull-Clark Rules
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      ptF.Null();
      iterQEFacet.Init(pe->GetFacet());
      for ( peCurr = iterQEFacet.PeekFirstEdge(), i = 0;
	    peCurr != NULL;
	    peCurr = iterQEFacet.PeekNextEdge(), i++ )
	{
	  ptF += peCurr->GetVertex()->GetPoint();
	}
      ptF /= i;
      ptE = pe->GetEdge()->GetVertex()->GetPoint();
      
      pt += ptE + ptF;
	}
  ASSERT( uiN > 2 );
  
  pt += ((FLOAT)((uiN - 2) * uiN) * pvx->GetPoint());
  
  pt /= (uiN * uiN);
      
  // Loop rules
  ASSERT( pvx->GetValence() > 2 );
  switch ( pvx->GetValence() )
    {
      // Regular Loop cases
    case 3:
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptLoop += ( 4.0 * pe->GetEdge()->GetVertex()->GetPoint() +
		      2.0 * (pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			     pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint()));
	}
      ASSERT( uiN == 3 );
      ptLoop += (40.0 * pvx->GetPoint());
      ptLoop /= 64.0;

      pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
      break;

    case 6:
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptLoop += ( 4.0 * pe->GetEdge()->GetVertex()->GetPoint() );
	}
      ASSERT( uiN == 6 );
      ptLoop += (40.0 * pvx->GetPoint());
      ptLoop /= 64.0;

      pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
      break;

      // Deal with valence 3 and 6 vertices in the 1-ring
    case 4:
      uiLoopVertices = 0;
      ptLoop.Null();
      pt.Null();

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptF.Null();
	  iterQEFacet.Init(pe->GetFacet());
	  for ( peCurr = iterQEFacet.PeekFirstEdge(), i = 0;
		peCurr != NULL;
		peCurr = iterQEFacet.PeekNextEdge(), i++ )
	    {
	      ptF += peCurr->GetVertex()->GetPoint();
	    }
	  ptF /= i;
	  ptE = pe->GetEdge()->GetVertex()->GetPoint();
	  
	  pt += ptE + ptF;

	  if ( pe->GetEdge()->GetVertex()->GetValence() == 3 )
	    {
	      uiLoopVertices++;
	      ptLoop += (4.0 * pe->GetEdge()->GetVertex()->GetPoint() +
			 2.0 * (pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				pe->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 2.0 * (pe->GetEdgeNext(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				pe->GetEdgeNext(1)->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 4.0 * (pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint()) +
			 2.0 * (pe->GetEdgePrev(1)->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 2.0 * (pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
				pe->GetEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()));
	    }
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetValence() == 3 )
	    {
	      uiLoopVertices++;
	      ptLoop += (4.0 * pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 4.0 * pe->GetEdge()->GetVertex()->GetPoint() +
			 4.0 * pe->GetEdgePrev(0)->GetVertex()->GetPoint() +
			 4.0 * pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
			 4.0 * pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint() +
			 2.0 * (pe->GetEdgePrev(1)->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()));
	    }
	  if ( pe->GetEdge()->GetVertex()->GetValence() == 6 )
	    {
	      uiLoopVertices++;
	      ptLoop += (4.0 * pe->GetEdge()->GetVertex()->GetPoint() +
			 4.0 * pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 4.0 * pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
			 2.0 * (pe->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 2.0 * (pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
				pe->GetEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 4.0 * pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint());
	    }
	}
      ASSERT( uiN > 2 );

      pt += ((FLOAT)((uiN - 2) * uiN) * pvx->GetPoint());
      pt /= (uiN * uiN);
      if ( uiLoopVertices > 0 )
	{
	  ptLoop += 40.0 * uiLoopVertices * pvx->GetPoint();
	  ptLoop /= (64.0 * uiLoopVertices);

	  pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
	}
      break;

    default:
      break;
    }

  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionCatmullClarkC2(QEVertex *pvx)
{
  /*
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEEdge *pe;
  Point pt, ptE, ptF, ptLoop;
  UINT_32 uiN, uiLoopVertices;

  //fprintf(stderr, "Vertex %p -- Val == %ld\n", pvx, pvx->GetValence());

  pt.Null(); 
  ptLoop.Null(); 

  // Catmull-Clark Rules
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge(), uiN++ )
    {
      pt += (8.0 * pe->GetEdge()->GetVertex()->GetPoint() +
	     1.0 * (pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
		    pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()));
    }
  ASSERT( uiN > 2 );
  
  pt += (2 * uiN * uiN) * pvx->GetPoint();
  pt /= (2 * uiN * (uiN + 5));
      
  // Loop rules
  ASSERT( pvx->GetValence() > 2 );
  switch ( pvx->GetValence() )
    {
      // Regular Loop cases
    case 3:
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptLoop += ( 2.0 * pe->GetEdge()->GetVertex()->GetPoint() +
		      1.0 * (pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			     pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint()));
	}
      ASSERT( uiN == 3 );
      ptLoop += (12.0 * pvx->GetPoint());
      ptLoop /= 24.0;

      pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
      break;

    case 6:
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptLoop += ( 2.0 * pe->GetEdge()->GetVertex()->GetPoint() );
	}
      ASSERT( uiN == 6 );
      ptLoop += (12.0 * pvx->GetPoint());
      ptLoop /= 24.0;

      pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
      break;

      // Deal with valence 3 and 6 vertices in the 1-ring
    case 4:
      uiLoopVertices = 0;
      ptLoop.Null();
      pt.Null();

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  pt += (8.0 * pe->GetEdge()->GetVertex()->GetPoint() +
		 1.0 * (pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
			pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()));

	  if ( pe->GetEdge()->GetVertex()->GetValence() == 3 )
	    {
	      uiLoopVertices++;
	      ptLoop += (268.0 * pvx->GetPoint() +
			 68.0 * pe->GetEdge()->GetVertex()->GetPoint() +
			 2.0 * pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 31.0 * (pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				 pe->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 16.0 * (pe->GetEdgeNext(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				 pe->GetEdgeNext(1)->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 48.0 * (pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint()) +
			 16.0 * (pe->GetEdgePrev(1)->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				 pe->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 31.0 * (pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
				 pe->GetEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 31.0 * pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint()) / 576.0;
	    }
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetValence() == 3 )
	    {
	      uiLoopVertices++;
	      ptLoop += (274.0 * pvx->GetPoint() +
			 38.0 * pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 50.0 * pe->GetEdgePrev(0)->GetVertex()->GetPoint() +
			 4.0 * (pe->GetEdgeNext(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				pe->GetEdgeNext(1)->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 60.0 * pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint() +
			 14.0 * (pe->GetEdgePrev(1)->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 60.0 * pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
			 4.0 * (pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 50.0 * pe->GetEdge()->GetVertex()->GetPoint()) / 576.0;
	    }
	  if ( pe->GetEdge()->GetVertex()->GetValence() == 6 )
	    {
	      uiLoopVertices++;
	      ptLoop += (276.0 * pvx->GetPoint() +
			 52.0 * pe->GetEdge()->GetVertex()->GetPoint() +
			 34.0 * pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 25.0 * (pe->GetEdgePrev(0)->GetVertex()->GetPoint() +
				 pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 6.0 * (pe->GetEdgeNext(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				pe->GetEdgeNext(1)->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 56.0 * pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint() +
			 6.0 * (pe->GetEdgePrev(1)->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 25.0 * (pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
				 pe->GetEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +

			 34.0 * pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint()) / 576.0;
	    }
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetValence() == 6 )
	    {
	      uiLoopVertices++;
	      ptLoop += (254.0 * pvx->GetPoint() +
			 14.0 * pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 66.0 * pe->GetEdgePrev(0)->GetVertex()->GetPoint() +
			 16.0 * pe->GetEdgeNext(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
			 32.0 * (pe->GetEdgeNext(1)->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
				 pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint()) +
			 8.0 * (pe->GetEdgePrev(1)->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				pe->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 32.0 * (pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint() +
				 pe->GetEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
			 16.0 * (pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
				 pe->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
			 66.0 * pe->GetEdge()->GetVertex()->GetPoint()) / 576.0;
	    }
	}
      ASSERT( uiN > 2 );

      pt += (2 * uiN * uiN) * pvx->GetPoint();
      pt /= (2 * uiN * (uiN + 5));
      if ( uiLoopVertices > 0 )
	{
	  ptLoop /= uiLoopVertices;

	  pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
	}
      break;

    default:
      break;
    }

  pt[3] = 1.0;

  return pt;
  */

  Point pt, ptE, ptF;
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  UINT_32 uiN;
  FLOAT fDen;

  iterQEVertex.Init(pvx);
  pt = CalculateVertexSubdivisionCatmullClarkC2(pvx);
  ptE.Null();
  ptF.Null();

  ASSERT( pvx->GetValence() > 2 );
  switch ( pvx->GetValence() )
    {
      // Regular Loop cases
    case 3:
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptF += CalculateFacetSubdivisionCatmullClarkC2(pe->GetFacet());
	  ptE += CalculateEdgeSubdivisionCatmullClarkC2(pe);
	}
      ASSERT( uiN == 3 );

      pt = (6.0 * pt + ptE + ptF);
      pt /= 12.0;
      break;

    case 6:
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptE += CalculateEdgeSubdivisionCatmullClarkC2(pe);
	}
      ASSERT( uiN == 6 );

      pt = (6.0 * pt + ptE);
      pt /= 12.0;
      break;

    default:
      // Catmull-Clark Vertex-Vertex Subdivision
      for ( pe = iterQEVertex.PeekFirstEdge(), uiN = 0;
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge(), uiN++ )
	{
	  ptF += CalculateFacetSubdivisionCatmullClarkC2(pe->GetFacet());
	  ptE += CalculateEdgeSubdivisionCatmullClarkC2(pe);
	}
  
      fDen = (FLOAT)(uiN * (uiN + 5));
      pt = (((FLOAT)(uiN*uiN)*pt) + (4.0*ptE) + ptF) / fDen; 
      break;
    }

  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCatmullClarkC2(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, vE, vF;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiN;
  FLOAT fTheta, fThetaNext, fThetaDelta, fA, fE, fF;

  iter.Init(pvx);

  // Catmull-Clark Vertex-Normal Subdivision
  uiN = pvx->GetValence();
  fA = ( 1.0 + cos((2.0 * PI)/(FLOAT)uiN) + 
	 cos(PI/(FLOAT)uiN)*sqrt(2.0*(9.0 + cos((2.0 * PI)/(FLOAT)uiN))) );

  vT0.Null();
  vT1.Null();
  fTheta = 0.0;
  fThetaDelta = (2.0 * PI) / (FLOAT)uiN;
  for ( pe = iter.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iter.PeekNextEdge(), uiN++ )
    {
      vE = CalculateEdgeSubdivisionCatmullClarkC2(pe);
      vF = CalculateFacetSubdivisionCatmullClarkC2(pe->GetFacet());

      fThetaNext = fTheta + fThetaDelta;
      switch( pvx->GetValence() )
	{
	case 3:
	  fE = cos(fTheta);
	  fF = cos(fThetaNext);
	  break;
	case 6:
	  fE = cos(fTheta);
	  fF = 0.0;
	  break;
	default:
	  fE = fA * cos(fTheta);
	  fF = cos(fTheta) + cos(fThetaNext);
	  break;
	}
      vT0 += (fE*vE + fF*vF);

      fTheta = fThetaNext;

      fThetaNext = fTheta + fThetaDelta;
      switch( pvx->GetValence() )
	{
	case 3:
	  fE = cos(fTheta);
	  fF = cos(fThetaNext);
	  break;
	case 6:
	  fE = cos(fTheta);
	  fF = 0.0;
	  break;
	default:
	  fE = fA * cos(fTheta);
	  fF = cos(fTheta) + cos(fThetaNext);
	  break;
	}
      vT1 += (fE*vE + fF*vF);
    }
  ASSERT( uiN == pvx->GetValence() );
  vT0[3] = 0.0;
  vT1[3] = 0.0;

  vNormal = vT1 * vT0;
  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionCatmullClarkC2(QEEdge *pe)
{
  Point pt, ptE, ptF, ptLoop;
  QEFacetIter iterQEFacet;
  QEEdge *ppe[2], *peCurr;
  UINT_32 i, uiN, uiLoopVertices;

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  uiLoopVertices = 0;
  ptLoop.Null();
  pt.Null();
  for ( i = 0; i< 2; i++ )
    {
      ptF.Null();
      iterQEFacet.Init(ppe[i]->GetFacet());
      for ( peCurr = iterQEFacet.PeekFirstEdge(), uiN = 0;
	    peCurr != NULL;
	    peCurr = iterQEFacet.PeekNextEdge(), uiN++ )
	{
	  ptF += peCurr->GetVertex()->GetPoint();
	}
      ptF /= uiN;

      ptE = ppe[i]->GetVertex()->GetPoint();
      pt += ptE + ptF;

      if ( ppe[i]->GetVertex()->GetValence() == 6 )
	{
	  uiLoopVertices++;
	  ptLoop += (6.0 * ppe[i]->GetVertex()->GetPoint() +
		     6.0 * ppe[i]->GetEdge()->GetVertex()->GetPoint() +
		     2.0 * ppe[i]->GetEdgePrev(0)->GetVertex()->GetPoint() +
		     2.0 * ppe[i]->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint());
	}
      if ( ppe[i]->GetVertex()->GetValence() == 3 )
	{
	  uiLoopVertices++;
	  ptLoop += (6.0 * ppe[i]->GetVertex()->GetPoint() +
		     6.0 * ppe[i]->GetEdge()->GetVertex()->GetPoint() +
		     1.0 * (ppe[i]->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint() +
			    ppe[i]->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
		     1.0 * (ppe[i]->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint() +
			    ppe[i]->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()));
	}
      if ( ppe[i]->GetEdgeNext(0)->GetEdge()->GetVertex()->GetValence() == 3 )
	{
	  uiLoopVertices++;
	  ptLoop += (3.0 * (ppe[i]->GetVertex()->GetPoint() +
			    ppe[i]->GetEdgeNext(0)->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint()) +
		     6.0 * ppe[i]->GetEdge()->GetVertex()->GetPoint() +
		     2.0 * ppe[i]->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint() +
		     2.0 * ppe[i]->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint());
	}
      if ( ppe[i]->GetEdgePrev(0)->GetVertex()->GetValence() == 3 )
	{
	  uiLoopVertices++;
	  ptLoop += (6.0 * ppe[i]->GetVertex()->GetPoint() +
		     3.0 * (ppe[i]->GetEdge()->GetVertex()->GetPoint() +
			    ppe[i]->GetEdgePrev(0)->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint()) +
		     2.0 * ppe[i]->GetEdgePrev(0)->GetVertex()->GetPoint() +
		     2.0 * ppe[i]->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint());
	}
    }

  pt /= 4.0;
  if ( uiLoopVertices > 0 )
    {
      ptLoop /= (16.0 * uiLoopVertices);

      pt = gfCatmullClarkC2GenFactor*pt + (1.0 - gfCatmullClarkC2GenFactor)*ptLoop;
    }
  pt[3] = 1.0;

  return pt;
}

//////////////////////////////////////////////////////////////////////
// Interpolatory Subdivision on Open Quadrilateral Nets with Abitrary Topology
//   Leif Kobbelt 1995
//   0 < w < 2(sqrt(5) - 1)  default w = 1.0
const FLOAT cfCatmullClarkW = 1.0;

Point QEMesh::CalculateFacetSubdivisionCatmullClarkButterfly(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  Point pt, ptE, ptF;
  UINT_32 uiN, i;

  ptE.Null();
  ptF.Null();  
  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge(), uiN = 0;
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge(), uiN++ )
    {
      ptE += CalculateEdgeSubdivisionCatmullClarkButterfly(pe);

      pt.Null();
      for ( peCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0), i = 0;
	    peCurr != pe->GetEdge()->GetEdgePrev(0);
	    peCurr = peCurr->GetEdgeNext(0), i++ )
	{
	  pt += CalculateEdgeSubdivisionCatmullClarkButterfly(peCurr);
	}
      if ( i == 0 )
	{
	  pt = ( CalculateEdgeSubdivisionCatmullClarkButterfly(pe->GetEdge()->GetEdgeNext(0)) +
		 CalculateEdgeSubdivisionCatmullClarkButterfly(pe->GetEdge()->GetEdgePrev(0)) ) / 2.0;
	}
      else
	{
	  pt /= i;
	}
      ptF += pt;
    }
  ptE /= uiN;
  ptF /= uiN;

  pt = ( (8.0 + cfCatmullClarkW)*ptE + (-cfCatmullClarkW)*ptF ) / (8.0);
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCatmullClarkButterfly(QEVertex *pvx)
{
  Point pt;

  pt = pvx->GetPoint();

  return pt;
}

Point QEMesh::CalculateVirtualEdgeSubdivisionCatmullClarkButterfly(QEEdge *pe)
{
  Point pt, ptE, ptC, ptEV, ptCV;
  QEEdge *peCurr, *peC, *ppe[2];
  UINT_32 i, j, k, uiN;

  ptE.Null();
  ptC.Null();
  uiN = 0;
  peCurr = pe;
  do
    {
      pt.Null();
      for ( peC = peCurr->GetEdgeNext(0), i = 0; 
	    peC != peCurr->GetEdgePrev(0);
	    peC = peC->GetEdgeNext(0), i++ )
	{
	  pt += peC->GetVertex()->GetPoint();
	}
      if ( i == 0 )
	{
	  pt = ( pe->GetEdge()->GetVertex()->GetPoint() + pe->GetEdgePrev(0)->GetVertex()->GetPoint() ) / 2.0;
	}
      else
	{
	  pt /= i;
	}
      ptC += pt;
      ptE += peCurr->GetEdge()->GetVertex()->GetPoint();

      peCurr = peCurr->GetEdgeNext(1);
      uiN++;
    }
  while ( peCurr != pe );
  ptE /= uiN;
  ptC /= uiN;

  ptEV = pe->GetEdge()->GetVertex()->GetPoint();
  ptEV += pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();
  ptEV += pe->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();

  ppe[0] = pe->GetEdgePrev(1)->GetEdgePrev(1);
  ppe[1] = pe;
  ptCV.Null();
  for ( k = 0; k < 2; k++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  pt.Null();
	  for ( peC = ppe[j]->GetEdgeNext(0), i = 0;
		peC != ppe[j]->GetEdgePrev(0);
		peC = peC->GetEdgeNext(0), i++ )
	    {
	      pt += peC->GetVertex()->GetPoint();
	    }
	  if ( i == 0 )
	    {
	      pt = ( ppe[j]->GetEdge()->GetVertex()->GetPoint() + ppe[j]->GetEdgePrev(0)->GetVertex()->GetPoint() ) / 2.0;
	    }
	  else
	    {
	      pt /= i;
	    }
	  ptCV += pt;

	  ppe[j] = ppe[j]->GetEdgeNext(1);
	}
    }

  pt = (4.0*ptE - ptEV) - (cfCatmullClarkW/(4.0 + cfCatmullClarkW))*(4.0*ptC - ptCV);
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionCatmullClarkButterfly(QEEdge *pe)
{
  Point pt, ppt[4];

  ppt[0] = CalculateVirtualEdgeSubdivisionCatmullClarkButterfly(pe);
  ppt[1] = pe->GetVertex()->GetPoint();
  ppt[2] = pe->GetEdge()->GetVertex()->GetPoint();
  ppt[3] = CalculateVirtualEdgeSubdivisionCatmullClarkButterfly(pe->GetEdge());

  pt = ((8.0 + cfCatmullClarkW)*(ppt[1] + ppt[2]) + 
	(-cfCatmullClarkW)*(ppt[0] + ppt[3]))/16.0; 
  pt[3] = 1.0;
  
  return pt;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

Point QEMesh::CalculateFacetSubdivisionSharpCatmullClark(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx;
  Point pt;
  UINT_32 uiN;

  pt.Null();

  iterQEFacet.Init(pfa);
  for ( pvx = iterQEFacet.PeekFirstVertex(), uiN = 0;
	pvx != NULL;
	pvx = iterQEFacet.PeekNextVertex(), uiN++ )
    {
      pt += pvx->GetPoint();
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionSharpCatmullClark(QEVertex *pvx)
{
  Point pt, ptCorner, ptCrease, ptSmooth;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 i, uiNeighbors, uiSharpEdges, uiSemiSharpEdges;
  FLOAT fSharpness, pfSemiSharpnessesMax[3], fSwap, fT;

  iter.Init(pvx);

  // Sharp Catmull-Clark Vertex-Vertex Subdivision
  ptCorner = pvx->GetPoint();
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptCorner;
    }

  uiSharpEdges = 0;
  uiSemiSharpEdges = 0;
  memset(pfSemiSharpnessesMax, 0, sizeof(pfSemiSharpnessesMax));
  ptSmooth.Null();
  ptCrease.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      pt = pe->GetEdge()->GetVertex()->GetPoint();
      ptSmooth += pt;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ptCrease += pt;
	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      uiSharpEdges++;
	    }
	  else
	    {
	      // Semi-sharp for this step
	      uiSemiSharpEdges++;
	      if ( fSharpness > pfSemiSharpnessesMax[2] )
		{
		  pfSemiSharpnessesMax[2] = fSharpness;
		  for ( i = 2; i > 0; i-- )
		    {
		      if ( pfSemiSharpnessesMax[i] > pfSemiSharpnessesMax[i-1] )
			{
			  fSwap = pfSemiSharpnessesMax[i];
			  pfSemiSharpnessesMax[i] = pfSemiSharpnessesMax[i-1];
			  pfSemiSharpnessesMax[i-1] = fSwap;
			}
		    }
		}
	    }
	}
    }

  //////////////////////////////
  // Corner rule
  if ( uiSharpEdges >= 3 )
    {
      return ptCorner;
    }

  //////////////////////////////
  // Crease rule
  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      ptCrease = ((6.0 * ptCorner) + ptCrease) / 8.0;
      ptCrease[3] = 1.0;
    }

  if ( uiSharpEdges == 2 && uiSemiSharpEdges == 0 )
    {
      return ptCrease;
    }

  //////////////////////////////
  // Smooth rule
  ASSERT( uiNeighbors >= 3 );
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetFacet() != GetNullFacet() );
      pt = CalculateFacetSubdivisionSharpCatmullClark(pe->GetFacet());
      ptSmooth += pt;
    }
  ptSmooth = ((((FLOAT)(uiNeighbors*(uiNeighbors - 2))) * ptCorner) + ptSmooth) / (FLOAT)(uiNeighbors*uiNeighbors);
  pt[3] = 1.0;

  if ( uiSharpEdges + uiSemiSharpEdges <= 1 )
    {
      return ptSmooth;
    }

  ASSERT( uiSharpEdges < 3 );
  i = MIN(uiSemiSharpEdges, 3 - uiSharpEdges);
  fT = pfSemiSharpnessesMax[i-1];

  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
      pt[3] = 1.0;
      return pt;
    }

  ASSERT( uiSharpEdges + uiSemiSharpEdges >= 3 );
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCorner);
  pt[3] = 1.0;
  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionSharpCatmullClark(QEVertex *pvx)
{
  Point pt, ptCorner, ptSmooth, ptCrease, ptV, ptE, ptF;
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  //QEEdge *ppeCrease[2];
  UINT_32 uiNeighbors, uiCreases;
  FLOAT fDen;
  Point ppptCrease[2][3];
  SharpVertexType psvtCrease[2];
  const FLOAT *pfCoeffTable;
  UINT_32 i, j, k, uiStart, uiEnd, uiCenter;

  iterQEVertex.Init(pvx);

  // Sharp Catmull-Clark Vertex-Vertex Subdivision
  ptCorner = pvx->GetPoint();
  if ( pvx->GetSharpVertexType() == svtCorner )
    {
      return ptCorner;
    }

  ptSmooth.Null();
  uiCreases = 0;
  for ( uiNeighbors = 0, pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iterQEVertex.PeekNextEdge() )
    {
      pt = pe->GetEdge()->GetVertex()->GetPoint();
      ptSmooth += pt;

      if ( pe->GetKnotInterval() != 0.0 )
	{
	  if ( uiCreases < 2 )
	    {
	      //ppeCrease[uiCreases] = pe;
	      psvtCrease[uiCreases] = pe->GetEdge()->GetVertex()->GetSharpVertexType();
	      ppptCrease[uiCreases][0] = pe->GetEdgePrev(1)->GetEdge()->GetVertex()->GetPoint();
	      ppptCrease[uiCreases][1] = pe->GetEdge()->GetVertex()->GetPoint();
	      ppptCrease[uiCreases][2] = pe->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();
	    }
	  uiCreases++;
	}
    }
  
  if ( uiCreases >= 3 )
    {
      return ptCorner;
    }

  if ( uiCreases == 2 )
    {
      uiCenter = scpuiSharpCatmullClarkSharpTypeIndex[svtRegularCrease];
      uiStart = scpuiSharpCatmullClarkSharpTypeIndex[psvtCrease[0]];
      uiEnd = scpuiSharpCatmullClarkSharpTypeIndex[psvtCrease[1]];

      pfCoeffTable = scppppfSharpCatmullClarkLimitTable[uiCenter][uiStart][uiEnd];
      i = 0;
      ptCrease = pfCoeffTable[i] * ptCorner;
#ifdef DEBUGGING_SHARP_CATMULL_CLARK_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	  fprintf(stderr, "  %f * ", pfCoeffTable[i]); 
	  ptCorner.Display();
	} 
#endif // DEBUGGING_SHARP_CATMULL_CLARK_MATH
      for ( j = 0; j < 2; j++ )
	{
	  for ( k = 0; k < 3; k++ )
	    {
	      i++;
	      ptCrease += pfCoeffTable[i] * ppptCrease[j][k];
#ifdef DEBUGGING_SHARP_CATMULL_CLARK_MATH
	      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
		   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )		   
		{
		  fprintf(stderr, "  %f * ", pfCoeffTable[i]);
		  ppptCrease[j][k].Display();
		} 
#endif // DEBUGGING_SHARP_CATMULL_CLARK_MATH
	    }
	}
      ptCrease[3] = 1.0;
#ifdef DEBUGGING_SHARP_CATMULL_CLARK_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "  Limit = "); 
	  ptCorner.Display();
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	} 
#endif // DEBUGGING_SHARP_CATMULL_CLARK_MATH

      return ptCrease;
    }

  // Catmull-Clark Vertex-Vertex Subdivision
  if ( !(uiNeighbors > 2) )
    {
      fprintf(stderr, "Error: smooth interior vertex with valence %ld < 3.\n", uiNeighbors);
    }
  ASSERT( uiNeighbors > 2 );
  ptV = ptCorner;
  ptE = ptSmooth;
  ptF.Null();
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptF += CalculateFacetSubdivisionSharpCatmullClark(pe->GetFacet());
    }
  
  fDen = (FLOAT)(uiNeighbors * (uiNeighbors + 5));
  pt = (((FLOAT)(uiNeighbors*uiNeighbors)*ptV) + (4.0*ptE) + ptF) / fDen; 
  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionSharpCatmullClark(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, vE, vF;
  Vector vCenter;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCrease;
  UINT_32 uiNeighbors, uiCreases;
  FLOAT fTheta, fThetaNext, fThetaDelta, fA, fE, fF;
  FLOAT fSharpness;

  iterQEVertex.Init(pvx);

  // Search for a crease edge (if one exists it will be in the mesh)
  peCrease = NULL;
  uiNeighbors = 0;
  uiCreases = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      uiNeighbors++;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness == INFINITY )
	{
	  uiCreases++;
	  peCrease = pe;
	}
    }
  ASSERT( uiNeighbors > 1 );

  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;
  if ( uiCreases < 2 )
    {
      ASSERT( uiNeighbors > 2 );

      fA = ( 1.0 + cos((2.0 * PI)/(FLOAT)uiNeighbors) + 
	     cos(PI/(FLOAT)uiNeighbors)*sqrt(2.0*(9.0 + cos((2.0 * PI)/(FLOAT)uiNeighbors))) );
      
      vT0.Null();
      vT1.Null();
      fTheta = 0.0;
      fThetaDelta = (2.0 * PI) / (FLOAT)uiNeighbors;
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  vE = pe->GetEdge()->GetVertex()->GetPoint();
	  ASSERT( pe->GetFacet() != GetNullFacet() );
	  vF = CalculateFacetSubdivisionSharpCatmullClark(pe->GetFacet());

	  fThetaNext = fTheta + fThetaDelta;
	  fE = fA * cos(fTheta);
	  fF = cos(fTheta) + cos(fThetaNext);
	  vT0 += (fE*vE + fF*vF);
	  
	  fTheta = fThetaNext;
	  
	  fThetaNext = fTheta + fThetaDelta;
	  fE = fA * cos(fTheta);
	  fF = cos(fTheta) + cos(fThetaNext);
	  vT1 += (fE*vE + fF*vF);
	}
      vT0[3] = 0.0;
      vT1[3] = 0.0;
      
      vNormal = vT1 * vT0;
    }
  else
    {
      if ( peCrease->GetFacet() == GetNullFacet() )
	{
	  peCrease = peCrease->GetEdgeNext(1);
	  ASSERT( peCrease->GetKnotInterval() == INFINITY );
	}
      vNormal = CalculateVertexNormalSubdivisionSharpCatmullClark(peCrease);
    }

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Vector QEMesh::CalculateVertexNormalSubdivisionSharpCatmullClark(QEEdge *peCrease)
{
  Vector vT0, vT1, vNormal;
  Vector vCenter;
  QEVertex *pvx;
  QEEdge *pe;
  UINT_32 uiSegmentLength, i, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];

  pvx = peCrease->GetVertex();
  ASSERT( pvx->GetSharpVertexType() == svtRegularCrease ||
	  pvx->GetSharpVertexType() == svtBoundaryCrease ||
	  pvx->GetSharpVertexType() == svtCorner );
  ASSERT( peCrease->GetFacet() != GetNullFacet() );
  ASSERT( peCrease->GetKnotInterval() == INFINITY );

  // Search for a crease edge (if one exists it will be in the mesh)
  uiSegmentLength = 2;
  for ( pe = peCrease->GetEdgeNext(1);
	pe != peCrease && pe->GetKnotInterval() != INFINITY;
	pe = pe->GetEdgeNext(1) )
    {
      uiSegmentLength++;
    }
  ASSERT( uiSegmentLength < cuiNeighborsMax );

  pe = peCrease;
  for ( i = 0; i < uiSegmentLength; i++ )
    {
      ASSERT( i < uiSegmentLength );
      pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
      pv[i][3] = 1.0;
      pe = pe->GetEdgeNext(1);
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;

  bRegularCrease = (pvx->GetSharpVertexType() == svtRegularCrease);
  
  // Calculate the normal for the individual segments and take the one with max magnitude
  vT0.Null();
  vT1.Null();
  if ( bRegularCrease )
    {
      vT0 = pv[0];
      vT1 = scppfSharpCatmullClarkTangentTable[2][0] * vCenter;
      for ( i = 0; i < uiSegmentLength; i++ )
	{
	  vT1 += scppfSharpCatmullClarkTangentTable[2][i+1] * pv[i];
	}
      vT0 -= pv[uiSegmentLength - 1];	      
    }
  else
    {
      switch ( uiSegmentLength )
	{
	case 0:
	case 1:
	  ASSERT( FALSE );
	  break;
	case 2:
	case 3:
	  uiIndex = uiSegmentLength - 2;
	  vT0 = pv[0];
	  vT1 = scppfSharpCatmullClarkTangentTable[uiIndex][0] * vCenter;
	  for ( i = 0; i < uiSegmentLength; i++ )
	    {
	      vT1 += scppfSharpCatmullClarkTangentTable[uiIndex][i+1] * pv[i];
	    }
	  vT0 -= pv[uiSegmentLength - 1];
	  break;
	default:
	  fThetaDelta = PI / (FLOAT)(uiSegmentLength - 1);
	  fA = sin(fThetaDelta);
	  vT1 = pv[0];
	  vT0 = (fA * pv[0]);
	  vT1 -= pv[uiSegmentLength - 1];
	  vT0 += (fA * pv[uiSegmentLength - 1]);
	  fA = 2.0 * (cos(fThetaDelta) - 1.0);
	  for ( i = 1, fTheta = fThetaDelta; 
		i < uiSegmentLength - 1;
		i++, fTheta += fThetaDelta )
	    {
	      vT0 += ((fA * sin(fTheta)) * pv[i+1]);
	    }
	  break;
	}
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;
  vNormal = vT0 * vT1;

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionSharpCatmullClark(QEEdge *pe)
{
  QEEdge *ppe[2], *peCurr;
  BOOL pbCrease[2], bCrease;
  Point pt, ptSmooth, ptCrease;
  Point ppt[4];
  UINT_32 i, uiSharpEdges;
  FLOAT fSharpness, fT;

  fSharpness = pe->GetKnotInterval();
  bCrease = FALSE;
  if ( fSharpness != 0.0 )
    {
      ppe[0] = pe;
      ppe[1] = pe->GetEdge();
      pbCrease[0] = FALSE;
      pbCrease[1] = FALSE;
      for ( i = 0; i < 2; i++ )
	{
	  uiSharpEdges = 1;
	  for ( peCurr = ppe[i]->GetEdgeNext(1); 
		peCurr != ppe[i]; 
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      if ( peCurr->GetKnotInterval() != 0.0 )
		{
		  uiSharpEdges++;
		}
	    }

	  if ( uiSharpEdges < 2 )
	    {
	      break;
	    }

	  pbCrease[i] = TRUE;
	}

      bCrease = pbCrease[0] && pbCrease[1];
      if ( bCrease )
	{
	  ptCrease = ( ppe[0]->GetVertex()->GetPoint() + ppe[1]->GetVertex()->GetPoint() ) / 2.0;
	  ptCrease[3] = 1.0;

	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      return ptCrease;
	    }
	}
    }

  // Midpoint smooth rule

  // Sharp Catmull-Clark Edge-Vertex Subdivision
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ASSERT( pe->GetFacet() != GetNullFacet() );
  ppt[2] = CalculateFacetSubdivisionSharpCatmullClark(pe->GetFacet());
  ppt[3] = CalculateFacetSubdivisionSharpCatmullClark(pe->GetEdge()->GetFacet());

  ptSmooth = (ppt[0] + ppt[1] + ppt[2] + ppt[3]) / 4.0;
  ptSmooth[3] = 1.0;

  if ( !bCrease )
    {
      return ptSmooth;
    } 

  fT = fSharpness;
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
  pt[3] = 1.0;
  return pt;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

Point QEMesh::CalculateVertexFacetSubdivisionDooSabin(QEEdge *pe)
{
  ASSERT( pe != NULL );
  ASSERT( pe->GetFacet() != GetNullFacet() );

  QEEdge *peCurr;
  Point pt, ptF, ptW;
  UINT_32 i, uiNeighbors;
  FLOAT fTheta, fThetaDelta;

  uiNeighbors = 0;
  ptF.Null();
  peCurr = pe;
  do
    {
      uiNeighbors++;
      ptF += peCurr->GetVertex()->GetPoint();

      peCurr = peCurr->GetEdgeNext(0);
    }
  while ( peCurr != pe );
  
  fTheta = 0.0;
  fThetaDelta = 2.0 * PI / (FLOAT)uiNeighbors;
  ptW.Null();
  for ( i = 0, peCurr = pe;
	i < uiNeighbors;
	i++, peCurr = peCurr->GetEdgeNext(0) )
    {
      pt = peCurr->GetVertex()->GetPoint();
      ptW += (2.0 * cos(fTheta)) * pt;

      fTheta += fThetaDelta;
    }

  pt = pe->GetVertex()->GetPoint();

  pt = (((FLOAT)uiNeighbors*pt) + ptW + (3.0*ptF)) / (4.0*uiNeighbors);
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionDooSabin(QEVertex *pvx)
{
  ASSERT( pvx != NULL );
  
  QEVertexIter iter;
  QEEdge *pe;
  Point pt;
  UINT_32 uiNeighbors;

  pt.Null();
  uiNeighbors = 0;
  iter.Init(pvx);
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      pt += CalculateVertexFacetSubdivisionDooSabin(pe);
      uiNeighbors++;
    }
  pt /= (FLOAT)uiNeighbors;
  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionDooSabin(QEVertex *pvx)
{
  ASSERT( pvx != NULL );

  Vector vT0, vT1, vNormal, v;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fTheta, fThetaDelta;

  iter.Init(pvx);

  // Doo-Sabin Vertex-Normal Subdivision
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
    }

  vT0.Null();
  vT1.Null();
  fTheta = 0.0;
  fThetaDelta = (2.0 * PI) / (FLOAT)uiNeighbors;
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      v = CalculateVertexFacetSubdivisionDooSabin(pe);
      vT0 += cos(fTheta) * v;
      fTheta += fThetaDelta;
      vT1 += cos(fTheta) * v;
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;

  vNormal = vT1 * vT0;
  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// Selective Refinement Subdivision

//////////////////////////////////////////////////////////////////////
// Selective Refinement Subdivision Loop

QEMesh *QEMesh::SubdivideCornerRoundingSelective(QEMesh::ErrorMetricType eErrorMetric,
						 FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvx;
  QEEdge *pe;
  Point pt;
  Vector v;
  FLOAT fError;
  UINT_32 uiN;

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideCornerRoundingSelective(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);
      pvx->SetHeapEntryReverse(NULL);

      pvx->SetGeneration(0);
      pvx->SetBirth(0);
      pt.Null();
      uiN = 0;
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  uiN++;
	  pt += pe->GetEdge()->GetVertex()->GetPoint();
	}
      pt /= (FLOAT)uiN;
      v = pvx->GetPoint() - pt;
      //v = pvx->GetPoint();
      pvx->SetTexture(v);
      pvx->SetForwardVertex(pvx);
    }

  pme->Preprocess();

  pme->ReSubdivideCornerRoundingSelective(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::ReSubdivideCornerRoundingSelective(QEMesh::ErrorMetricType /*eErrorMetric*/,
						FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peStart;
  FLOAT fError;
  Point pt;
  Vector v, vNormal;

  //fprintf(stderr, "START: ReSubdivideCornerRoundingSelective\n");
  //DisplayEuler();

  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  else
		    {
		      pvxCurr = NULL;
		    }
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      if ( pvx->GetHeapEntryReverse() == NULL )
	{
	  // This is the first time we are subdividing at this vertex
	  pt = CalculateVertexLimitSubdivisionCornerRoundingSelective(pvx);
	  pvx->SetLimitPosition(pt);
	  vNormal = CalculateVertexNormalSubdivisionCornerRoundingSelective(pvx);
	  vNormal.CompletePlaneEquation(pt);
	  pvx->SetNormal(vNormal);
	  v = CalculateGenerationVectorSubdivisionCornerRoundingSelective(pvx);
	  pvx->SetTexture(v);
	}

      SubdivideCornerRoundingSelective(pvx);

      //////////////////////////////
      // Update the current vertex
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      fError = CalculateVertexErrorSubdivideCornerRoundingSelective(pvx);
      ASSERT( fError <= phe->GetKey() + EPSILON );
      m_pheapSelective->Update(phe, fError);

      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      //m_plistSelective->InsertFirst(pvx, 0);
      phe = pvx->GetHeapEntryReverse();
      if ( phe == NULL )
	{
	  phe = m_pheapSelectiveReverse->Insert(fError, pvx);
	  pvx->SetHeapEntryReverse(phe);
	}      
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      //////////////////////////////
      // Update the area of influence around the vertex
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Update the neighbor vertices
	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();

	      fError = CalculateVertexErrorSubdivideCornerRoundingSelective(pvxCurr);
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      m_pheapSelective->Update(phe, fError);

	      phe = pvxCurr->GetHeapEntryReverse();
	      if ( phe != NULL )
		{
		  m_pheapSelectiveReverse->Update(phe, fError);
		}
	    }

	  // Insert the new vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSelective(pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  if ( phe == NULL )
	    {   
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      m_pheapSelective->Update(phe, fError);
	    }

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	  pe->GetEdgeNext(0)->GetEdge()->GetFacet()->Preprocess();
	}
    }

  //DisplayEuler();
  //fprintf(stderr, "END  : ReSubdivideCornerRoundingSelective\n");
}

VOID QEMesh::UnSubdivideCornerRoundingSelective(QEMesh::ErrorMetricType /*eErrorMetric*/,
						FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  FLOAT fError;

  //fprintf(stderr, "START: UnSubdivideCornerRoundingSelective\n");
  //DisplayEuler();

  for ( pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError);
	( pvx != NULL && 
	  ( !(m_plistQEFacet->GetLength() < uiFacetsMax) || !(fError > fTolerance) ) );
	pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  iterQEVertex.Init(pvx);
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      peCurr = pe->GetEdge();
	      if ( peCurr->GetVertex()->GetGeneration() > pvx->GetGeneration() ) 
		{
		  // Search for the neighboring vertex that can be unsubdivided
		  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
		  pvxCurr = peCurr->GetVertex();
		  break;
		}
	      else
		{
		  ASSERT( peCurr->GetVertex()->GetGeneration() == pvx->GetGeneration() ); 
		}
	    }
	}
      ASSERT( pvx != NULL );
      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      ASSERT( pvx->GetHeapEntryReverse() != NULL );

      UnSubdivideCornerRoundingSelective(pvx);

      // Update the current vertex
      fError = CalculateVertexErrorSubdivideCornerRoundingSelective(pvx);
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Update(phe, fError);

      phe = pvx->GetHeapEntryReverse();
      ASSERT( phe != NULL );
      if ( pvx->GetGeneration() == pvx->GetBirth() )
	{
	  m_pheapSelectiveReverse->Remove(phe);
	  pvx->SetHeapEntryReverse(NULL);
	}
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSelective(pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Update(phe, fError);

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	}
    }

  //DisplayEuler();
  //fprintf(stderr, "END  : UnSubdivideCornerRoundingSelective\n");
}

QEEdge *QEMesh::FindSubdivideCornerRoundingSelectiveCenterVertex(QEEdge *pe)
{
  QEEdge *peCurr;

  // Search for the neighboring vertex that can be unsubdivided
  for ( peCurr = pe;
	peCurr->GetVertex()->GetGeneration() > peCurr->GetEdge()->GetVertex()->GetGeneration();
	peCurr = peCurr->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge() )
    {
      ASSERT( peCurr->GetEdgeNext(1)->GetEdgeNext(1) == peCurr->GetEdgePrev(1)->GetEdgePrev(1) );
      ASSERT( peCurr->GetVertex()->GetGeneration() <= peCurr->GetEdgeNext(1)->GetEdge()->GetVertex()->GetGeneration() );
      ASSERT( peCurr->GetVertex()->GetGeneration() <= peCurr->GetEdgePrev(1)->GetEdge()->GetVertex()->GetGeneration() );
    }
  ASSERT( peCurr->GetVertex()->GetGeneration() == peCurr->GetEdge()->GetVertex()->GetGeneration() );

  return peCurr;
}

Point QEMesh::CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(QEVertex *pvx, UINT_32 uiGeneration)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  FLOAT fBeta, fA, fB;
  UINT_32 uiN;

  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      uiN++;
    }

  fBeta = 3.0 + 2.0 * cos(2.0 * PI / (FLOAT)uiN);
  fBeta = ( 40.0 - fBeta * fBeta ) / 8.0;
  fA = fBeta / ( 3.0 + fBeta );
  fB = ( 5.0 - fBeta ) / 8.0;

  ASSERT( uiGeneration >= pvx->GetBirth() );
  pt = pvx->GetTexture();
  pt = pvx->GetLimitPosition() + ((fA * pow(fB, uiGeneration - pvx->GetBirth())) * pt);

  return pt;
}


///////////////////////////////////////////////////
// QEMesh::SubdivideCornerRoundingSelective      //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0  => v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
VOID QEMesh::SubdivideCornerRoundingSelective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEVertex *pvxNew;
  QEEdge *pe, *peNew;
  QEFacet *pfaNew;
  Point ptNew, pt;
  //Point ptCenter;
  Vector v;
  //Vector vNormal;

  //v = pvx->GetPoint();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Create midpoints and edges, but do not connect them
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  pt = CalculateEdgeSubdivisionCornerRoundingSelective(pe);
	  pvxNew = CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pvxNew->SetGeneration(pvx->GetGeneration()+1);
	  pvxNew->SetBirth(pvxNew->GetGeneration());
	  v = pt;
	  pvxNew->SetTexture(v);
	  pvxNew->SetForwardVertex(pvx);
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );

	  peNew->SetVertex(pvxNew);
	  pe->SetForwardEdge(peNew);
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Connect the midpoints in dividing edges in half
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = pe->GetForwardEdge();
	  ASSERT( peNew != NULL );
	  pe->SetForwardEdge(NULL);
	  pvxNew = peNew->GetVertex();
	  ASSERT( pvxNew != NULL );
	  
	  peNew->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  
	  peNew->SetVertex(pvxNew);
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	  pe->GetEdge()->SetVertex(pvxNew);
	  
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	  Splice(1, pe->GetEdge(), peNew);
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Cut the corners creating new triangles
      pfaNew = CreateFacet();
      ASSERT( pfaNew != NULL );
      peNew = CreateEdge();
      ASSERT( peNew != NULL );

      peNew->SetFacet(pfaNew);
      peNew->GetEdge()->SetFacet(pe->GetFacet());
      pe->SetFacet(pfaNew);
      pe->GetEdgePrev(0)->SetFacet(pfaNew);

      peNew->SetVertex(pe->GetEdge()->GetVertex());
      peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());

      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew);
      Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());
    }

  ptNew = CalculateVertexSubdivisionCornerRoundingSelective(pvx);
  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()+1);
}

///////////////////////////////////////////////////
// QEMesh::UnSubdivideCornerRoundingSelective    //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0 <=  v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
VOID QEMesh::UnSubdivideCornerRoundingSelective(QEVertex *pvx)
{
  ASSERT( pvx != NULL && pvx->GetGeneration() > 0 );

  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvxCurr;
  QEEdge *pe, *peCurr, *peStart;
  QEFacet *pfa, *pfaOld;
  Point ptNew, pt;
  UINT_32 uiN;
  FLOAT fAlpha;

  iterQEVertex.Init(pvx);
  uiN = 0;
  pt.Null();
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Calculate the previous position of the vertex
      uiN++;
      pt += pe->GetEdge()->GetVertex()->GetPoint();
    }

  fAlpha = 3.0 + 2.0 * cos(2.0 * PI / (FLOAT)uiN);
  fAlpha = 40.0 / (fAlpha * fAlpha);

  ptNew = (fAlpha * pvx->GetPoint()) + ((1.0 - fAlpha) / (FLOAT)uiN) * pt;

  iterQEVertex.Init(pvx);
  peStart = NULL;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Clean up the facets
      pfaOld = pe->GetFacet();
      pfa = pe->GetEdgeNext(0)->GetEdge()->GetFacet();
      pe->SetFacet(pfa);
      pe->GetEdgePrev(0)->SetFacet(pfa);

      DestroyFacet(pfaOld);

      // Deal with the edge vertex
      pe = pe->GetEdge();

      ASSERT( pe->GetVertex()->GetGeneration() == pvx->GetGeneration() );
      peCurr = pe->GetEdgeNext(1);
      Splice(1, peCurr, pe);
      if ( peStart == NULL )
	{
	  peStart = peCurr;
	}
      else
	{
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}
      peCurr = pe->GetEdgePrev(1)->GetEdgePrev(1);
      Splice(1, peCurr, peCurr->GetEdgeNext(1));
      pe->SetVertex(pe->GetVertex());

      if ( pe->GetEdgeNext(1) == peCurr )
	{
	  // Remove the edge-vertex
	  Splice(1, peCurr, pe);
	  pe->SetVertex(peCurr->GetEdge()->GetVertex());
	  Splice(1, peCurr->GetEdge(), pe);
	  Splice(1, peCurr->GetEdge(), peCurr->GetEdge()->GetEdgePrev(1));

	  pvxCurr = peCurr->GetVertex();
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Remove(phe);
	  ASSERT( pvxCurr->GetHeapEntryReverse() == NULL );
	  DestroyVertex(pvxCurr);
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}      
      else
	{
	  // Change the edge-vertex's forwarding pointer to point at the neighbor
	  ASSERT( pe->GetEdgeNext(1)->GetEdgeNext(1) == pe->GetEdgePrev(1)->GetEdgePrev(1) );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge());
	  if ( !( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		  pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() ) )
	    {
	      ASSERT( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		      pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() );
	    }
	  pe->GetVertex()->SetForwardVertex(peCurr->GetVertex());
	}
    }
  ASSERT( peStart != NULL );
  DestroyEdge(peStart->GetEdge());
  DestroyEdge(peStart);

  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()-1);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideCornerRoundingSelective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

//////////////////////////////////////////////////////////////////////
// Selective Refinement Subdivision Loop Limit Positions

//#define SELECTIVE_LOOP_FORWARD_ONLY

QEMesh *QEMesh::SubdivideCornerRoundingSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
						      FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvx;
  QEEdge *pe;
  Point pt;
  Vector v;
  FLOAT fError;
  UINT_32 uiN;


  pme = Copy();
  ASSERT( pme != NULL );
  pme->CalculateSubdivisionLimitPositionsAndNormalsLoop();

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);
      pvx->SetHeapEntryReverse(NULL);

      pvx->SetGeneration(0);
      pvx->SetBirth(0);
      pt.Null();
      uiN = 0;
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  uiN++;
	  pt += pe->GetEdge()->GetVertex()->GetPoint();
	}
      pt /= (FLOAT)uiN;
      v = pvx->GetPoint() - pt;
      pvx->SetTexture(v);
      pvx->SetForwardVertex(pvx);
    }

  pme->Preprocess();

  pme->ReSubdivideCornerRoundingSelectiveLimit(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::ReSubdivideCornerRoundingSelectiveLimit(QEMesh::ErrorMetricType /*eErrorMetric*/,
						     FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peStart, *peOuter;
  FLOAT fError;
  //Vector v;

  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  else if ( pvxCurr->GetGeneration() == pvx->GetGeneration() )
		    {
		      peOuter = peCurr;
		      do
			{
			  pvxCurr = peOuter->GetEdge()->GetVertex();
			  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
			    {
			      break;
			    }
			  pvxCurr = NULL;
			  peOuter = peOuter->GetEdgeNext(1);
			}
		      while ( peOuter != peCurr );
		    }
		  else
		    {
		      pvxCurr = NULL;
		    }
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideCornerRoundingSelectiveLimit(pvx);

      // Update the current vertex
      fError = CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvx);
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      ASSERT( fError <= phe->GetKey() + EPSILON );
      m_pheapSelective->Update(phe, fError);

      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      phe = pvx->GetHeapEntryReverse();
      if ( phe == NULL )
	{
	  phe = m_pheapSelectiveReverse->Insert(fError, pvx);
	  pvx->SetHeapEntryReverse(phe);
	}      
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      //////////////////////////////
      // Update the area of influence around the vertex
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
#ifdef NOT_USING_LIMIT_POSITION_ERROR
	  // Update the neighbor vertices
	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();

	      fError = CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvxCurr);
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      m_pheapSelective->Update(phe, fError);

	      phe = pvxCurr->GetHeapEntryReverse();
	      if ( phe != NULL )
		{
		  m_pheapSelectiveReverse->Update(phe, fError);
		}
	    }
#endif // NOT_USING_LIMIT_POSITION_ERROR

	  // Insert the new vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  if ( phe == NULL )
	    {   
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      m_pheapSelective->Update(phe, fError);
	    }

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	  pe->GetEdgeNext(0)->GetEdge()->GetFacet()->Preprocess();
	}
    }
}

VOID QEMesh::UnSubdivideCornerRoundingSelectiveLimit(QEMesh::ErrorMetricType /*eErrorMetric*/,
						     FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  FLOAT fError;

  //fprintf(stderr, "START: UnSubdivideCornerRoundingSelectiveLimit\n");
  //DisplayEuler();

  for ( pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError);
	( pvx != NULL && 
	  ( !(m_plistQEFacet->GetLength() < uiFacetsMax) || !(fError > fTolerance) ) );
	pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  iterQEVertex.Init(pvx);
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      peCurr = pe->GetEdge();
	      if ( peCurr->GetVertex()->GetGeneration() > pvx->GetGeneration() ) 
		{
		  // Search for the neighboring vertex that can be unsubdivided
		  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
		  pvxCurr = peCurr->GetVertex();
		  break;
		}
	      else
		{
		  ASSERT( peCurr->GetVertex()->GetGeneration() == pvx->GetGeneration() ); 
		}
	    }
	}
      ASSERT( pvx != NULL );
      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      ASSERT( pvx->GetHeapEntryReverse() != NULL );

      UnSubdivideCornerRoundingSelectiveLimit(pvx);

      // Update the current vertex
      fError = CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvx);
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Update(phe, fError);

      phe = pvx->GetHeapEntryReverse();
      ASSERT( phe != NULL );
      if ( pvx->GetGeneration() == pvx->GetBirth() )
	{
	  m_pheapSelectiveReverse->Remove(phe);
	  pvx->SetHeapEntryReverse(NULL);
	}
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Update(phe, fError);

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	}
    }

  //DisplayEuler();
  //fprintf(stderr, "END  : UnSubdivideCornerRoundingSelectiveLimit\n");
}

///////////////////////////////////////////////////
// QEMesh::SubdivideCornerRoundingSelectiveLimit //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1      v1      f0    //
//           |                   /  | \          //
//           |                 e1  e5  e0        //
//           |               /      |     \      //
//  e2------vV------e0  => v2--e6--vV--e4--v0    //
//           |               \      |     /      //
//           |                 e2  e7  e3        //
//           |                   \  | /          //
//  f2      e3      f3     f2      v3      f3    //
//                                               //
VOID QEMesh::SubdivideCornerRoundingSelectiveLimit(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEVertex *pvxNew, *pvxCurr;
  QEEdge *pe, *peNew;
  QEFacet *pfaNew;
  Point ptNew, pt;
  //Point ptCenter;
  Vector v, vNormal;

  //v = pvx->GetPoint();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Create midpoints and edges, but do not connect them
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  pt = CalculateEdgeSubdivisionCornerRoundingSelective(pe);
	  pvxNew = CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pvxNew->SetGeneration(pvx->GetGeneration()+1);
	  pvxNew->SetBirth(pvxNew->GetGeneration());
	  //v = pt;
	  //pvxNew->SetTexture(v);
	  pvxNew->SetForwardVertex(pvx);
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );

	  peNew->SetVertex(pvxNew);
	  pe->SetForwardEdge(peNew);
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Connect the midpoints in dividing edges in half
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = pe->GetForwardEdge();
	  ASSERT( peNew != NULL );
	  pe->SetForwardEdge(NULL);
	  pvxNew = peNew->GetVertex();
	  ASSERT( pvxNew != NULL );
	  
	  peNew->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  
	  peNew->SetVertex(pvxNew);
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	  pe->GetEdge()->SetVertex(pvxNew);
	  
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	  Splice(1, pe->GetEdge(), peNew);
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Cut the corners creating new triangles
      pfaNew = CreateFacet();
      ASSERT( pfaNew != NULL );
      peNew = CreateEdge();
      ASSERT( peNew != NULL );

      peNew->SetFacet(pfaNew);
      peNew->GetEdge()->SetFacet(pe->GetFacet());
      pe->SetFacet(pfaNew);
      pe->GetEdgePrev(0)->SetFacet(pfaNew);

      peNew->SetVertex(pe->GetEdge()->GetVertex());
      peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());

      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew);
      Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());
    }

  ptNew = CalculateVertexSubdivisionCornerRoundingSelective(pvx);
  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()+1);

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Calculate Limit Positions and Normals
      pvxCurr = pe->GetEdge()->GetVertex();
      if ( pvxCurr->GetHeapEntry() == NULL )
	{   
	  pt = CalculateVertexLimitSubdivisionCornerRoundingSelective(pvxCurr);
	  pvxCurr->SetLimitPosition(pt);
	  vNormal = CalculateVertexNormalSubdivisionCornerRoundingSelective(pvxCurr);
	  vNormal.CompletePlaneEquation(pt);
	  pvxCurr->SetNormal(vNormal);
	  v = CalculateGenerationVectorSubdivisionCornerRoundingSelective(pvxCurr);
	  pvxCurr->SetTexture(v);
	}
    }
}

/////////////////////////////////////////////////////
// QEMesh::UnSubdivideCornerRoundingSelectiveLimit //
//                                                 //
//   Old Vertex Disk       Subdivided Facets       //
//                                                 //
//  f1      e1      f0     f1      v1      f0      //
//           |                   /  | \            //
//           |                 e1  e5  e0          //
//           |               /      |     \        //
//  e2------vV------e0 <=  v2--e6--vV--e4--v0      //
//           |               \      |     /        //
//           |                 e2  e7  e3          //
//           |                   \  | /            //
//  f2      e3      f3     f2      v3      f3      //
//                                                 //
VOID QEMesh::UnSubdivideCornerRoundingSelectiveLimit(QEVertex *pvx)
{
  ASSERT( pvx != NULL && pvx->GetGeneration() > 0 );

  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvxCurr;
  QEEdge *pe, *peCurr, *peStart;
  QEFacet *pfa, *pfaOld;
  Point ptNew, pt;
  UINT_32 uiN;
  FLOAT fAlpha;

  iterQEVertex.Init(pvx);
  uiN = 0;
  pt.Null();
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Calculate the previous position of the vertex
      uiN++;
      pt += pe->GetEdge()->GetVertex()->GetPoint();
    }

  fAlpha = 3.0 + 2.0 * cos(2.0 * PI / (FLOAT)uiN);
  fAlpha = 40.0 / (fAlpha * fAlpha);

  ptNew = (fAlpha * pvx->GetPoint()) + ((1.0 - fAlpha) / (FLOAT)uiN) * pt;

  iterQEVertex.Init(pvx);
  peStart = NULL;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Clean up the facets
      pfaOld = pe->GetFacet();
      pfa = pe->GetEdgeNext(0)->GetEdge()->GetFacet();
      pe->SetFacet(pfa);
      pe->GetEdgePrev(0)->SetFacet(pfa);

      DestroyFacet(pfaOld);

      // Deal with the edge vertex
      pe = pe->GetEdge();

      ASSERT( pe->GetVertex()->GetGeneration() == pvx->GetGeneration() );
      peCurr = pe->GetEdgeNext(1);
      Splice(1, peCurr, pe);
      if ( peStart == NULL )
	{
	  peStart = peCurr;
	}
      else
	{
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}
      peCurr = pe->GetEdgePrev(1)->GetEdgePrev(1);
      Splice(1, peCurr, peCurr->GetEdgeNext(1));
      pe->SetVertex(pe->GetVertex());

      if ( pe->GetEdgeNext(1) == peCurr )
	{
	  // Remove the edge-vertex
	  Splice(1, peCurr, pe);
	  pe->SetVertex(peCurr->GetEdge()->GetVertex());
	  Splice(1, peCurr->GetEdge(), pe);
	  Splice(1, peCurr->GetEdge(), peCurr->GetEdge()->GetEdgePrev(1));

	  pvxCurr = peCurr->GetVertex();
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Remove(phe);
	  ASSERT( pvxCurr->GetHeapEntryReverse() == NULL );
	  DestroyVertex(pvxCurr);
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}      
      else
	{
	  // Change the edge-vertex's forwarding pointer to point at the neighbor
	  ASSERT( pe->GetEdgeNext(1)->GetEdgeNext(1) == pe->GetEdgePrev(1)->GetEdgePrev(1) );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge());
	  if ( !( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		  pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() ) )
	    {
	      ASSERT( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		      pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() );
	    }
	  pe->GetVertex()->SetForwardVertex(peCurr->GetVertex());
	}
    }
  ASSERT( peStart != NULL );
  DestroyEdge(peStart->GetEdge());
  DestroyEdge(peStart);

  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()-1);
}

//////////////////////////////////////////////////////////////////////
// Selective Refinement Subdivision Sharp Loop Limit Positions

#define SELECTIVE_SHARP_LOOP_FORWARD_ONLY

QEMesh *QEMesh::SubdivideCornerRoundingSharpSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
							   FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx;
  Point pt;
  Vector v, vNormal;
  FLOAT fError;
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  UINT_32 uiN;
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

  eErrorMetric = emtBestFitPlane;

  pme = Copy();
  ASSERT( pme != NULL );
  pme->CalculateSubdivisionSharpVertexTypesSharpLoop();
  v.Null();
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetGeneration(0);
      pvx->SetBirth(0);
      pvx->SetTexture(v);
    }
  pme->PreSubdivideCornerRoundingSharpSelectiveLimit();
  //pme->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = pme->CalculateVertexLimitSubdivisionCornerRoundingSharpSelective(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = pme->CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      v = pme->CalculateGenerationVectorSubdivisionCornerRoundingSharpSelective(pvx);
      pvx->SetTexture(v);
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

      fError = pme->CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);
      pvx->SetHeapEntryReverse(NULL);

      //pvx->SetGeneration(0);
      pvx->SetBirth(pvx->GetGeneration());

#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      //v.Null();
      //pvx->SetTexture(v);
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      pt.Null();
      uiN = 0;
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  uiN++;
	  pt += pe->GetEdge()->GetVertex()->GetPoint();
	}
      pt /= (FLOAT)uiN;
      v = pvx->GetPoint() - pt;
      pvx->SetTexture(v);
      pvx->SetForwardVertex(pvx);
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  pme->Preprocess();

  pme->ReSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::PreSubdivideCornerRoundingSharpSelectiveLimit()
{
  CListInt listQEEdge;
  QEMeshIter iterQEMesh;
  QEVertex *pvx, *pvxCurr, *ppvx[2];
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peStart, *peOuter;
  FLOAT fSharpness;
  //Vector v;
  UINT_32 i, uiVertices;

  listQEEdge.Init();
  iterQEMesh.Init(this);

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 && fSharpness != INFINITY )
	{
	  listQEEdge.InsertLast(pe, 0);
	}
    }
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 && fSharpness != INFINITY )
	{
	  listQEEdge.InsertLast(pe, 0);
	}
    }

  for ( pe = (QEEdge*)(listQEEdge.RemoveFirst());
	pe != NULL;
	pe = (QEEdge*)(listQEEdge.RemoveFirst()) )
    {
      fSharpness = pe->GetKnotInterval();
      ASSERT( fSharpness != INFINITY );
      if ( fSharpness == 0.0 )
	{
	  continue;
	}

      ppvx[0] = pe->GetVertex();
      ppvx[1] = pe->GetEdge()->GetVertex();
      uiVertices = 1;
      if ( ppvx[0]->GetGeneration() == ppvx[1]->GetGeneration() )
	{
	  uiVertices = 2;
	}
      else if ( ppvx[0]->GetGeneration() > ppvx[1]->GetGeneration() )
	{
	  ppvx[0] = ppvx[1];
	}

      // Force the endpoints to be subdivided
      for ( i = 0; i < uiVertices; i++ )
	{
	  /*
	  fprintf(stderr, "FORCING %ld: Gen = %ld: %p = ", 
		  i, ppvx[i]->GetGeneration(), ppvx[i]);
	  ppvx[i]->GetPoint().Display();
	  */
	  pvx = NULL;
	  while ( pvx != ppvx[i] )
	    {	  
	      pvx = ppvx[i];
	      pvxCurr = pvx;
	      while ( pvxCurr != NULL )
		{
		  pvx = pvxCurr;
		  pvxCurr = NULL;

		  peStart = pvx->GetEdge();

		  // Check the immediate neighbors
		  pe = peStart;
		  do
		    {
		      pvxCurr = pe->GetEdge()->GetVertex();
		      if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
			{
			  break;
			}
		      pvxCurr = NULL;
		      pe = pe->GetEdgeNext(1);
		    }
		  while ( pe != peStart );
		  if ( pvxCurr != NULL )
		    {
		      continue;
		    }

		  // Check the vertices on the same facet
		  pe = peStart;
		  do
		    {
		      ASSERT( pvxCurr == NULL );
		      for ( peCurr = pe->GetEdgeNext(0)->GetEdgeNext(0);
			    peCurr != pe->GetEdgePrev(0);
			    peCurr = peCurr->GetEdgeNext(0) )
			{
			  pvxCurr = peCurr->GetEdge()->GetVertex();
			  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
			    {
			      break;
			    }
			  pvxCurr = NULL;
			}
		      pe = pe->GetEdgeNext(1);
		    }
		  while ( pe != peStart && pvxCurr == NULL );
		  if ( pvxCurr != NULL )
		    {
		      continue;
		    }

		  // Check the vertices that are neighbors to one ring neighborhood
		  pe = peStart;
		  do
		    {
		      for ( peCurr = pe->GetEdgeNext(0);
			    peCurr != pe;
			    peCurr = peCurr->GetEdgeNext(0) )
			{
			  pvxCurr = peCurr->GetVertex();
			  ASSERT( pvxCurr->GetGeneration() >= pvx->GetGeneration() );
			  if ( pvxCurr->GetGeneration() == pvx->GetGeneration() )
			    {
			      peOuter = peCurr;
			      do
				{
				  pvxCurr = peOuter->GetEdge()->GetVertex();
				  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
				    {
				      break;
				    }
				  pvxCurr = NULL;
				  peOuter = peOuter->GetEdgeNext(1);
				}
			      while ( peOuter != peCurr );
			    }
			  else
			    {
			      pvxCurr = NULL;
			    }
			}
		      pe = pe->GetEdgeNext(1);
		    }
		  while ( pe != peStart && pvxCurr == NULL );
		  fSharpness = 3.0;
		}

	      /*
	      fprintf(stderr, "  Presubdivide: Gen = %ld: %p = ", 
		      pvx->GetGeneration(), pvx);
	      pvx->GetPoint().Display();
	      */
	      SubdivideCornerRoundingSharpSelectiveLimit(pvx, TRUE);

	      //////////////////////////////
	      // Update the area of influence around the vertex
	      iterQEVertex.Init(pvx);
	      for ( pe = iterQEVertex.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEVertex.PeekNextEdge() )
		{
		  fSharpness = pe->GetKnotInterval();
		  if ( fSharpness != 0.0 && fSharpness != INFINITY )
		    {
		      listQEEdge.InsertLast(pe, 0);
		      listQEEdge.InsertLast(pe->GetEdge(), 0);
		    }
		}
	    }
	}
    }
}

VOID QEMesh::ReSubdivideCornerRoundingSharpSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
						     FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peStart, *peOuter;
  FLOAT fError;
  //Vector v;

  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  else if ( pvxCurr->GetGeneration() == pvx->GetGeneration() )
		    {
		      peOuter = peCurr;
		      do
			{
			  pvxCurr = peOuter->GetEdge()->GetVertex();
			  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
			    {
			      break;
			    }
			  pvxCurr = NULL;
			  peOuter = peOuter->GetEdgeNext(1);
			}
		      while ( peOuter != peCurr );
		    }
		  else
		    {
		      pvxCurr = NULL;
		    }
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideCornerRoundingSharpSelectiveLimit(pvx);

      // Update the current vertex
      fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvx);
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      //ASSERT( fError <= phe->GetKey() + EPSILON );
      if ( !(fError <= phe->GetKey() + EPSILON) )
	{
	  fprintf(stderr, "ResubdivideSharpLoop: %f > %f\n", fError, phe->GetKey() + EPSILON);
	}
      m_pheapSelective->Update(phe, fError);

      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      phe = pvx->GetHeapEntryReverse();
      if ( phe == NULL )
	{
	  phe = m_pheapSelectiveReverse->Insert(fError, pvx);
	  pvx->SetHeapEntryReverse(phe);
	}      
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      //////////////////////////////
      // Update the area of influence around the vertex
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
#ifdef NOT_USING_LIMIT_POSITION_ERROR
	  // Update the neighbor vertices
	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();

	      fError = CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(pvxCurr);
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      m_pheapSelective->Update(phe, fError);

	      phe = pvxCurr->GetHeapEntryReverse();
	      if ( phe != NULL )
		{
		  m_pheapSelectiveReverse->Update(phe, fError);
		}
	    }
#endif // NOT_USING_LIMIT_POSITION_ERROR

	  // Insert the new vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  if ( phe == NULL )
	    {   
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      m_pheapSelective->Update(phe, fError);
	    }

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	  pe->GetEdgeNext(0)->GetEdge()->GetFacet()->Preprocess();
	}
    }
}

VOID QEMesh::UnSubdivideCornerRoundingSharpSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
							  FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  FLOAT fError;

  //fprintf(stderr, "START: UnSubdivideCornerRoundingSharpSelectiveLimit\n");
  //DisplayEuler();

  for ( pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError);
	( pvx != NULL && 
	  ( !(m_plistQEFacet->GetLength() < uiFacetsMax) || !(fError > fTolerance) ) );
	pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  iterQEVertex.Init(pvx);
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      peCurr = pe->GetEdge();
	      if ( peCurr->GetVertex()->GetGeneration() > pvx->GetGeneration() ) 
		{
		  // Search for the neighboring vertex that can be unsubdivided
		  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
		  pvxCurr = peCurr->GetVertex();
		  break;
		}
	      else
		{
		  ASSERT( peCurr->GetVertex()->GetGeneration() == pvx->GetGeneration() ); 
		}
	    }
	}
      ASSERT( pvx != NULL );
      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      ASSERT( pvx->GetHeapEntryReverse() != NULL );

      UnSubdivideCornerRoundingSharpSelectiveLimit(pvx);

      // Update the current vertex
      fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvx);
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Update(phe, fError);

      phe = pvx->GetHeapEntryReverse();
      ASSERT( phe != NULL );
      if ( pvx->GetGeneration() == pvx->GetBirth() )
	{
	  m_pheapSelectiveReverse->Remove(phe);
	  pvx->SetHeapEntryReverse(NULL);
	}
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Update(phe, fError);

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	}
    }

  //DisplayEuler();
  //fprintf(stderr, "END  : UnSubdivideCornerRoundingSharpSelectiveLimit\n");
}

////////////////////////////////////////////////////////
// QEMesh::SubdivideCornerRoundingSharpSelectiveLimit //
//                                                    //
//   Old Vertex Disk       Subdivided Facets          //
//                                                    //
//  f1      e1      f0     f1      v1      f0         //
//           |                   /  | \               //
//           |                 e1  e5  e0             //
//           |               /      |     \           //
//  e2------vV------e0  => v2--e6--vV--e4--v0         //
//           |               \      |     /           //
//           |                 e2  e7  e3             //
//           |                   \  | /               //
//  f2      e3      f3     f2      v3      f3         //
//                                                    //
VOID QEMesh::SubdivideCornerRoundingSharpSelectiveLimit(QEVertex *pvx, BOOL bPresubdivide/* = FALSE*/)
{
  QEVertexIter iterQEVertex;
  QEVertex *pvxNew, *pvxCurr;
  QEEdge *pe, *peNew, *peStart;
  QEFacet *pfaNew;
  Point ptNew, pt;
  //Point ptCenter;
  Vector v, vNormal;
  FLOAT fSharpness;

#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
  v = pvx->GetPoint();
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Create midpoints and edges, but do not connect them
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  pt = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe);
	  pvxNew = CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pvxNew->SetGeneration(pvx->GetGeneration()+1);
	  pvxNew->SetBirth(pvxNew->GetGeneration());
	  if ( pe->GetKnotInterval() == INFINITY )
	    {
	      pvxNew->SetSharpVertexType(svtRegularCrease);
	    }
	  else
	    {
	      pvxNew->SetSharpVertexType(svtSmooth);
	    }
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pvxNew->SetTexture(v);
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pvxNew->SetForwardVertex(pvx);
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );
	  fSharpness = pe->GetKnotInterval();
	  peNew->SetKnotInterval(fSharpness);
	  peNew->GetEdge()->SetKnotInterval(fSharpness);

	  peNew->SetVertex(pvxNew);
	  pe->SetForwardEdge(peNew);
	}
    }

  peStart = iterQEVertex.PeekFirstEdge();
  pe = peStart;
  do
    {
      // Connect the midpoints in dividing edges in half
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = pe->GetForwardEdge();
	  ASSERT( peNew != NULL );
	  pe->SetForwardEdge(NULL);
	  pvxNew = peNew->GetVertex();
	  ASSERT( pvxNew != NULL );
	  
	  peNew = peNew->GetEdge();

	  peNew->SetFacet(pe->GetFacet());
	  pe->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  pe->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  
	  peNew->SetVertex(pe->GetVertex());
	  peNew->GetEdge()->SetVertex(pvxNew);
	  pe->SetVertex(pvxNew);
	  
	  Splice(1, pe, peNew);
	  Splice(1, pe->GetEdgePrev(1), pe);
	  Splice(1, pe, peNew->GetEdge());

	  if ( peStart == pe )
	    {
	      peStart = peNew;
	    }
	  pe = peNew;
	}

      pe = pe->GetEdgeNext(1);
    }
  while ( pe != peStart );

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Cut the corners creating new triangles
      pfaNew = CreateFacet();
      ASSERT( pfaNew != NULL );
      peNew = CreateEdge();
      ASSERT( peNew != NULL );
      peNew->SetKnotInterval(0.0);
      peNew->GetEdge()->SetKnotInterval(0.0);

      peNew->SetFacet(pfaNew);
      peNew->GetEdge()->SetFacet(pe->GetFacet());
      pe->SetFacet(pfaNew);
      pe->GetEdgePrev(0)->SetFacet(pfaNew);

      peNew->SetVertex(pe->GetEdge()->GetVertex());
      peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());

      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew);
      Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());
    }

  ptNew = CalculateVertexSubdivisionCornerRoundingSharpSelective(pvx);
  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()+1);

  if ( bPresubdivide )
    {
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  fSharpness = pe->GetKnotInterval();
	  if ( fSharpness != 0.0 && fSharpness != INFINITY )
	    {
	      fSharpness = MAX(0.0, fSharpness - 1.0);
	      pe->SetKnotInterval(fSharpness);
	      pe->GetEdge()->SetKnotInterval(fSharpness);
	    }
	}
    }
  else
    {
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Calculate Limit Positions and Normals
	  pvxCurr = pe->GetEdge()->GetVertex();
	  if ( pvxCurr->GetHeapEntry() == NULL )
	    {   
	      pt = CalculateVertexLimitSubdivisionCornerRoundingSharpSelective(pvxCurr);
	      pvxCurr->SetLimitPosition(pt);
	      vNormal = CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(pvxCurr);
	      vNormal.CompletePlaneEquation(pt);
	      pvxCurr->SetNormal(vNormal);
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	      v = CalculateGenerationVectorSubdivisionCornerRoundingSharpSelective(pvxCurr);
	      pvxCurr->SetTexture(v);
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	    }
	}
    }
}

//////////////////////////////////////////////////////////
// QEMesh::UnSubdivideCornerRoundingSharpSelectiveLimit //
//                                                      //
//   Old Vertex Disk       Subdivided Facets            //
//                                                      //
//  f1      e1      f0     f1      v1      f0           //
//           |                   /  | \                 //
//           |                 e1  e5  e0               //
//           |               /      |     \             //
//  e2------vV------e0 <=  v2--e6--vV--e4--v0           //
//           |               \      |     /             //
//           |                 e2  e7  e3               //
//           |                   \  | /                 //
//  f2      e3      f3     f2      v3      f3           //
//                                                      //
VOID QEMesh::UnSubdivideCornerRoundingSharpSelectiveLimit(QEVertex *pvx)
{
  ASSERT( pvx != NULL && pvx->GetGeneration() > 0 );

  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvxCurr;
  QEEdge *pe, *peCurr, *peStart;
  QEFacet *pfa, *pfaOld;
  Point ptNew, pt;
  UINT_32 uiN;
  FLOAT fAlpha;

  iterQEVertex.Init(pvx);
  uiN = 0;
  pt.Null();
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Calculate the previous position of the vertex
      uiN++;
      pt += pe->GetEdge()->GetVertex()->GetPoint();
    }

  fAlpha = 3.0 + 2.0 * cos(2.0 * PI / (FLOAT)uiN);
  fAlpha = 40.0 / (fAlpha * fAlpha);

  ptNew = (fAlpha * pvx->GetPoint()) + ((1.0 - fAlpha) / (FLOAT)uiN) * pt;

  iterQEVertex.Init(pvx);
  peStart = NULL;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Clean up the facets
      pfaOld = pe->GetFacet();
      pfa = pe->GetEdgeNext(0)->GetEdge()->GetFacet();
      pe->SetFacet(pfa);
      pe->GetEdgePrev(0)->SetFacet(pfa);

      DestroyFacet(pfaOld);

      // Deal with the edge vertex
      pe = pe->GetEdge();

      ASSERT( pe->GetVertex()->GetGeneration() == pvx->GetGeneration() );
      peCurr = pe->GetEdgeNext(1);
      Splice(1, peCurr, pe);
      if ( peStart == NULL )
	{
	  peStart = peCurr;
	}
      else
	{
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}
      peCurr = pe->GetEdgePrev(1)->GetEdgePrev(1);
      Splice(1, peCurr, peCurr->GetEdgeNext(1));
      pe->SetVertex(pe->GetVertex());

      if ( pe->GetEdgeNext(1) == peCurr )
	{
	  // Remove the edge-vertex
	  Splice(1, peCurr, pe);
	  pe->SetVertex(peCurr->GetEdge()->GetVertex());
	  Splice(1, peCurr->GetEdge(), pe);
	  Splice(1, peCurr->GetEdge(), peCurr->GetEdge()->GetEdgePrev(1));

	  pvxCurr = peCurr->GetVertex();
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Remove(phe);
	  ASSERT( pvxCurr->GetHeapEntryReverse() == NULL );
	  DestroyVertex(pvxCurr);
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}      
      else
	{
	  // Change the edge-vertex's forwarding pointer to point at the neighbor
	  ASSERT( pe->GetEdgeNext(1)->GetEdgeNext(1) == pe->GetEdgePrev(1)->GetEdgePrev(1) );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge());
	  if ( !( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		  pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() ) )
	    {
	      ASSERT( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		      pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() );
	    }
	  pe->GetVertex()->SetForwardVertex(peCurr->GetVertex());
	}
    }
  ASSERT( peStart != NULL );
  DestroyEdge(peStart->GetEdge());
  DestroyEdge(peStart);

  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()-1);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideCornerRoundingSelectiveLimit(QEVertex *pvx)
{
  Vector v;

  v = pvx->GetPoint() - pvx->GetLimitPosition();
  v[3] = 0.0;

  return v.Magnitude();

#ifdef WE_DONT_HAVE_LIMIT_POSITIONS
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
#endif // WE_DONT_HAVE_LIMIT_POSITIONS
}

FLOAT QEMesh::CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(ErrorMetricType eErrorMetric, QEVertex *pvx)
{
  FLOAT fError;

  fError = 0.0;
  switch ( eErrorMetric )
    {
    case emtBestFitPlane:
      fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimitTangentPlane(pvx);
      break;
    case emtLimitDistance:
      fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimitLimitDistance(pvx);
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return fError;
}

FLOAT QEMesh::CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimitLimitDistance(QEVertex *pvx)
{
  Vector v;

  v = pvx->GetPoint() - pvx->GetLimitPosition();
  v[3] = 0.0;

  return v.Magnitude();
}

FLOAT QEMesh::CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimitTangentPlane(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal;
  //Vector v;
  Point ptCenterLimit;
  Vector vCenter;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCrease;
  UINT_32 uiNeighbors, uiCreases, i, j, k, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  FLOAT fSharpness, fError, fErrorMax;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];
  BOOL pbCrease[cuiNeighborsMax];
  UINT_32 puiSegmentLength[cuiNeighborsMax];

  iterQEVertex.Init(pvx);

  // Search for a crease edge (if one exists it will be in the mesh)
  peCrease = NULL;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else
	{
	  uiNeighbors += 3;
	  // This can add at most one crease edge
	}

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ASSERT( fSharpness == INFINITY );
	  peCrease = pe;
	}
    }
  ASSERT( uiNeighbors < cuiNeighborsMax );

  // Calculate all the neighbor vertices
  if ( peCrease == NULL )
    {
      peCrease = iterQEVertex.PeekFirstEdge();
    }
  uiCreases = 0;
  memset(pbCrease, 0, sizeof(pbCrease));
  memset(puiSegmentLength, 0, sizeof(puiSegmentLength));
  i = 0;
  pe = peCrease;
  do
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );

      fSharpness = pe->GetKnotInterval();
      ASSERT( fSharpness == 0.0 || fSharpness == INFINITY );

      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiNeighbors );
	  pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pv[i] = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  ASSERT( i < uiNeighbors );
	  pv[i] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(peCurr->GetVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  // This a smooth edge
	  j = i;
	  if ( i == 0 )
	    {
	      j = uiNeighbors - 1;
	    }
	  else
	    {
	      i++;
	    }
	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      ASSERT( j < uiNeighbors );
	      pv[j] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      ASSERT( j < uiNeighbors );
	      pv[j] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetEdgeNext(1));
	    }

	  ASSERT( i < uiNeighbors );
	  pv[i] = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetVertex());
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;

	  // This a smooth edge
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      ASSERT( i < uiNeighbors );
	      pv[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      ASSERT( i < uiNeighbors );
	      pv[i] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdgeNext(0));
	    }
	  i++;
	}
      pe = pe->GetEdgeNext(1);
    }
  while( pe != peCrease );

  for ( i = 0; i < uiNeighbors; i++ )
    {      
      pv[i][3] = 1.0;
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;
  //ptCenterLimit = pvx->GetLimitPosition();
  ptCenterLimit = pvx->GetPoint();
  if ( uiCreases < 2 )
    {
      fErrorMax = 0.0;

      vT0.Null();
      vT1.Null();
      fTheta = 0.0;
      fThetaDelta = 2.0 * PI / (FLOAT)uiNeighbors;
      for ( i = 0; i < uiNeighbors; i++ )
	{
	  vT1 += cos(fTheta) * pv[i];
	  fTheta += fThetaDelta;
	  vT0 += cos(fTheta) * pv[i];
	}
      vT0[3] = 0.0;
      vT1[3] = 0.0;
      vNormal = vT0 * vT1;
      if ( !vNormal.Normalize() )
	{
	  vNormal.Set(0.0, 0.0, 1.0);
	}
      vNormal.CompletePlaneEquation(ptCenterLimit);

      fError = vNormal.Dot(vCenter);
      fErrorMax = MAX(fErrorMax, ABS(fError));
      for ( i = 0; i < uiNeighbors; i++ )
	{
	  fError = vNormal.Dot(pv[i]);
	  fErrorMax = MAX(fErrorMax, ABS(fError));
	}
    }
  else
    {
      // Copy the first edge to the end
      pv[uiNeighbors] = pv[0]; 
      pbCrease[uiNeighbors] = pbCrease[0];
      j = 0;
      for ( i = 0; i < uiNeighbors; i++ )
	{      
	  if ( pbCrease[i] )
	    {
	      j++;
	    }
	  ASSERT( j > 0 && j <= uiCreases );
	  puiSegmentLength[j-1]++;
	}

      bRegularCrease = ( uiNeighbors == 6 && uiCreases == 2 &&
			 puiSegmentLength[0] == 3 && puiSegmentLength[1] == 3 ); 

      // Calculate the error for the individual segments and take the max
      fErrorMax = 0.0;
      k = 0;
      for ( i = 0; i < uiCreases; i++ )
	{
	  vT0.Null();
	  vT1.Null();
	  if ( bRegularCrease )
	    {
	      vT0 = pv[k];
	      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
	      for ( j = 0; j <= puiSegmentLength[i]; j++ )
		{
		  vT1 += scppfSharpLoopTangentTable[2][j+1] * pv[k+j];
		}
	      vT0 -= pv[k+puiSegmentLength[i]];	      
	    }
	  else
	    {
	      switch ( puiSegmentLength[i] )
		{
		case 0:
		  ASSERT( FALSE );
		  break;
		case 1:
		case 2:
		  uiIndex = puiSegmentLength[i] - 1;
		  vT0 = pv[k];
		  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
		  for ( j = 0; j <= puiSegmentLength[i]; j++ )
		    {
		      vT1 += scppfSharpLoopTangentTable[uiIndex][j+1] * pv[k+j];
		    }
		  vT0 -= pv[k+puiSegmentLength[i]];
		  break;
		default:
		  fThetaDelta = PI / (FLOAT)puiSegmentLength[i];
		  fA = sin(fThetaDelta);
		  vT1 = pv[k];
		  vT0 = (fA * pv[k]);
		  vT1 -= pv[k+puiSegmentLength[i]];
		  vT0 += (fA * pv[k+puiSegmentLength[i]]);
		  fA = 2.0 * (cos(fThetaDelta) - 1.0);
		  for ( j = 1, fTheta = fThetaDelta; 
			j < puiSegmentLength[i];
			j++, fTheta += fThetaDelta )
		    {
		      vT0 += ((fA * sin(fTheta)) * pv[k+j]);
		    }
		  break;
		}
	    }
	  vT0[3] = 0.0;
	  vT1[3] = 0.0;
	  vNormal = vT0 * vT1;
	  if ( !vNormal.Normalize() )
	    {
	      vNormal.Set(0.0, 0.0, 1.0);
	    }
	  vNormal.CompletePlaneEquation(ptCenterLimit);

	  fError = vNormal.Dot(vCenter);
	  fErrorMax = MAX(fErrorMax, ABS(fError));
	  for ( j = 0; j <= puiSegmentLength[i]; j++, k++ )
	    {
	      fError = vNormal.Dot(pv[k]);
	      fErrorMax = MAX(fErrorMax, ABS(fError));
	    }
	  k--;
	}
    }

  return fErrorMax;
}

//////////////////////////////////////////////////
// Loop Selective
//

Point QEMesh::CalculateVertexSubdivisionCornerRoundingSelective(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fAlpha;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetGeneration() >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pt + pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();
	}
      else
	{
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
	  pt = pt + pe->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  pt = pt + CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
	}
    }
  
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  pt = (pt + (fAlpha * pvx->GetPoint())) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexHelperSubdivisionCornerRoundingSelective(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fAlpha;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
	  pt = pt + pe->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  pt = pt + CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	  pt = pt + CalculateEdgeSubdivisionCornerRoundingSelective(pe);
	}
    }
  
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  pt = (pt + (fAlpha * pvx->GetPoint())) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateGenerationVectorSubdivisionCornerRoundingSelective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  Vector v;
  UINT_32 uiNeighbors;

  iterQEVertex.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
	  pt = pt + pe->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  pt = pt + CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
	  uiNeighbors++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	  pt = pt + CalculateVertexHelperSubdivisionCornerRoundingSelective(pe->GetEdge()->GetVertex());

	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pt = pt + pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pt = pt + CalculateEdgeSubdivisionCornerRoundingSelective(pe->GetEdgeNext(0));
	    }

	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pt = pt + pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pt = pt + CalculateEdgeSubdivisionCornerRoundingSelective(pe->GetEdge()->GetEdgeNext(1));
	    }

	  uiNeighbors += 3;
	}
    }
  pt /= (FLOAT)uiNeighbors;

  v = pvx->GetPoint() - pt;

  return v;
}

Point QEMesh::CalculateVertexLimitSubdivisionCornerRoundingSelective(QEVertex *pvx)
{
  Point pt;
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fA, fEpsilon;

  iterQEVertex.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  pt.Null();
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pt + pe->GetEdge()->GetVertex()->GetPoint();
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
	  pt = pt + pe->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  pt = pt + CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
	  uiNeighbors++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	  pt = pt + CalculateVertexHelperSubdivisionCornerRoundingSelective(pe->GetEdge()->GetVertex());

	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pt = pt + pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pt = pt + CalculateEdgeSubdivisionCornerRoundingSelective(pe->GetEdgeNext(0));
	    }

	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pt = pt + pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pt = pt + CalculateEdgeSubdivisionCornerRoundingSelective(pe->GetEdge()->GetEdgeNext(1));
	    }

	  uiNeighbors += 3;
	}
    }
  
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiNeighbors) / (8.0 * fA);
  pt = (pt + (fEpsilon * pvx->GetPoint())) / (fEpsilon + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  return pt;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCornerRoundingSelective(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, v;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiNeighbors;
  FLOAT fTheta, fThetaDelta;

  iter.Init(pvx);

  // Loop Vertex-Normal Subdivision
  uiNeighbors = 0;
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else
	{
	  uiNeighbors += 3;
	}
    }

  vT0.Null();
  vT1.Null();
  fTheta = 0.0;
  fThetaDelta = (2.0 * PI) / (FLOAT)uiNeighbors;
  for ( pe = iter.PeekFirstEdge();
	pe != NULL;
	pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  v = pe->GetEdge()->GetVertex()->GetPoint();
	  vT0 += cos(fTheta) * v;
	  fTheta += fThetaDelta;
	  vT1 += cos(fTheta) * v;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
	  v = pe->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  v = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
	  vT0 += cos(fTheta) * v;
	  fTheta += fThetaDelta;
	  vT1 += cos(fTheta) * v;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      v = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      v = CalculateEdgeSubdivisionCornerRoundingSelective(pe->GetEdge()->GetEdgeNext(1));
	    }
	  vT0 += cos(fTheta) * v;
	  fTheta += fThetaDelta;
	  vT1 += cos(fTheta) * v;

	  v = CalculateVertexHelperSubdivisionCornerRoundingSelective(pe->GetEdge()->GetVertex());
	  vT0 += cos(fTheta) * v;
	  fTheta += fThetaDelta;
	  vT1 += cos(fTheta) * v;

	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      v = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      v = CalculateEdgeSubdivisionCornerRoundingSelective(pe->GetEdgeNext(0));
	    }
	  vT0 += cos(fTheta) * v;
	  fTheta += fThetaDelta;
	  vT1 += cos(fTheta) * v;
	}
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;

  vNormal = vT1 * vT0;
  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionCornerRoundingSelective(QEEdge *pe)
{
  Point pt;
  Point ppt[6];
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ASSERT( pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() );

  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

  if ( pe->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
      ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdgePrev(0)->GetVertex()->GetForwardVertex() != NULL );
      ppt[2] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdgePrev(0)->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[3] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
      ppt[3] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
      ppt[3] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[4] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
      ppt[4] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetForwardVertex() != NULL );
      ppt[4] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[5] = pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_LOOP_FORWARD_ONLY
      ppt[5] = pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex()->GetTexture();
#else // SELECTIVE_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
      ppt[5] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_LOOP_FORWARD_ONLY
    }

  pt = ((6.0 * ppt[0]) + (6.0 * ppt[1]) + ppt[2] + ppt[3] + ppt[4] + ppt[5]) / 16.0;
  pt[3] = 1.0;
  
  return pt;
}

//////////////////////////////////////////////////
// Sharp Loop Selective
//

Point QEMesh::CalculateVertexSubdivisionCornerRoundingSharpSelective(QEVertex *pvx)
{
  Point pt, ptCorner, ptCrease, ptSmooth;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 i, uiNeighbors, uiSharpEdges, uiSemiSharpEdges;
  FLOAT fA, fAlpha;
  FLOAT fSharpness, pfSemiSharpnessesMax[3], fSwap, fT;
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
  QEEdge *peCurr;
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

  iter.Init(pvx);

  // Sharp Loop Vertex-Vertex Subdivision
  uiSharpEdges = 0;
  uiSemiSharpEdges = 0;
  memset(pfSemiSharpnessesMax, 0, sizeof(pfSemiSharpnessesMax));
  ptSmooth.Null();
  ptCrease.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetGeneration() >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();
	}
      else
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pt = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  pt = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	}

      ptSmooth += pt;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ptCrease += pt;
	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      uiSharpEdges++;
	    }
	  else
	    {
	      // Semi-sharp for this step
	      uiSemiSharpEdges++;
	      if ( fSharpness > pfSemiSharpnessesMax[2] )
		{
		  pfSemiSharpnessesMax[2] = fSharpness;
		  for ( i = 2; i > 0; i-- )
		    {
		      if ( pfSemiSharpnessesMax[i] > pfSemiSharpnessesMax[i-1] )
			{
			  fSwap = pfSemiSharpnessesMax[i];
			  pfSemiSharpnessesMax[i] = pfSemiSharpnessesMax[i-1];
			  pfSemiSharpnessesMax[i-1] = fSwap;
			}
		    }
		}
	    }
	}
    }

  //////////////////////////////
  // Corner rule
  ptCorner = pvx->GetPoint();

  if ( uiSharpEdges >= 3 )
    {
      return ptCorner;
    }

  //////////////////////////////
  // Crease rule
  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      ptCrease = ((6.0 * ptCorner) + ptCrease) / 8.0;
      ptCrease[3] = 1.0;
    }

  if ( uiSharpEdges == 2 && uiSemiSharpEdges == 0 )
    {
      return ptCrease;
    }

  //////////////////////////////
  // Smooth rule
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  ptSmooth = (ptSmooth + (fAlpha * ptCorner)) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  if ( uiSharpEdges + uiSemiSharpEdges <= 1 )
    {
      return ptSmooth;
    }

  ASSERT( uiSharpEdges < 3 );
  i = MIN(uiSemiSharpEdges, 3 - uiSharpEdges);
  fT = pfSemiSharpnessesMax[i-1];

  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
      pt[3] = 1.0;
      return pt;
    }

  ASSERT( uiSharpEdges + uiSemiSharpEdges >= 3 );
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCorner);
  pt[3] = 1.0;
  return pt;
}

Point QEMesh::CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(QEVertex *pvx)
{
  Point pt, ptCorner, ptCrease, ptSmooth;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 i, uiNeighbors, uiSharpEdges, uiSemiSharpEdges;
  FLOAT fA, fAlpha;
  FLOAT fSharpness, pfSemiSharpnessesMax[3], fSwap, fT;
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
  QEEdge *peCurr;
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

  iter.Init(pvx);

  // Sharp Loop Vertex-Vertex Subdivision
  uiSharpEdges = 0;
  uiSemiSharpEdges = 0;
  memset(pfSemiSharpnessesMax, 0, sizeof(pfSemiSharpnessesMax));
  ptSmooth.Null();
  ptCrease.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pe->GetEdge()->GetVertex()->GetPoint();
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pt = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  pt = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	  pt = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe);
	}

      ptSmooth += pt;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ptCrease += pt;
	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      uiSharpEdges++;
	    }
	  else
	    {
	      // Semi-sharp for this step
	      uiSemiSharpEdges++;
	      if ( fSharpness > pfSemiSharpnessesMax[2] )
		{
		  pfSemiSharpnessesMax[2] = fSharpness;
		  for ( i = 2; i > 0; i-- )
		    {
		      if ( pfSemiSharpnessesMax[i] > pfSemiSharpnessesMax[i-1] )
			{
			  fSwap = pfSemiSharpnessesMax[i];
			  pfSemiSharpnessesMax[i] = pfSemiSharpnessesMax[i-1];
			  pfSemiSharpnessesMax[i-1] = fSwap;
			}
		    }
		}
	    }
	}
    }

  //////////////////////////////
  // Corner rule
  ptCorner = pvx->GetPoint();

  if ( uiSharpEdges >= 3 )
    {
      return ptCorner;
    }

  //////////////////////////////
  // Crease rule
  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      ptCrease = ((6.0 * ptCorner) + ptCrease) / 8.0;
      ptCrease[3] = 1.0;
    }

  if ( uiSharpEdges == 2 && uiSemiSharpEdges == 0 )
    {
      return ptCrease;
    }

  //////////////////////////////
  // Smooth rule
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  ptSmooth = (ptSmooth + (fAlpha * ptCorner)) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  if ( uiSharpEdges + uiSemiSharpEdges <= 1 )
    {
      return ptSmooth;
    }

  ASSERT( uiSharpEdges < 3 );
  i = MIN(uiSemiSharpEdges, 3 - uiSharpEdges);
  fT = pfSemiSharpnessesMax[i-1];

  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
      pt[3] = 1.0;
      return pt;
    }

  ASSERT( uiSharpEdges + uiSemiSharpEdges >= 3 );
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCorner);
  pt[3] = 1.0;
  return pt;
}

Vector QEMesh::CalculateGenerationVectorSubdivisionCornerRoundingSharpSelective(QEVertex *pvx)
{
  Vector v;
  //Point pt;
  Point ptCorner, ptSmooth;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *ppeCrease[3];
  QEVertex *pppvxCrease[2][3], *ppvxCrease[3];
  Point ppptCrease[2][3], pptCrease[3], *pppt[2];
  UINT_32 uiNeighbors, uiCreases;
  SharpVertexType peSharpType[2], eSharpType;
  UINT_32 i;

  iterQEVertex.Init(pvx);

  // Sharp Loop Vertex-Vertex Subdivision
  ptSmooth.Null();
  memset(pppvxCrease, 0, sizeof(pppvxCrease));
  uiCreases = 0;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
        pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );

      memset(ppvxCrease, 0, sizeof(ppvxCrease));

      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ppvxCrease[1] = pe->GetEdge()->GetVertex();
	  pptCrease[1] = ppvxCrease[1]->GetPoint();
	  ptSmooth += pptCrease[1]; 

	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ppvxCrease[1] = peCurr->GetVertex();
	  pptCrease[1] = pe->GetEdge()->GetVertex()->GetTexture();
	  ptSmooth += pptCrease[1]; 
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ppvxCrease[1] = peCurr->GetVertex();
	  ASSERT( ppvxCrease[1] == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  pptCrease[1] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(ppvxCrease[1], pe->GetVertex()->GetGeneration());
	  ptSmooth += pptCrease[1]; 
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

	  uiNeighbors++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  ppvxCrease[1] = pe->GetEdge()->GetVertex();
	  pptCrease[1] = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(ppvxCrease[1]);
	  ptSmooth += pptCrease[1];

	  // This a smooth edge
	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pptCrease[0] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	      ptSmooth += pptCrease[0];
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pptCrease[0] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetEdgeNext(1));
	      ptSmooth += pptCrease[0];
	    }

	  // This a smooth edge
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pptCrease[2] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	      ptSmooth += pptCrease[2];
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pptCrease[2] += CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdgeNext(0));
	      ptSmooth += pptCrease[2];
	    }

	  uiNeighbors += 3;
	}

      if ( pe->GetKnotInterval() != 0.0 )
	{
	  if ( uiCreases < 2 )
	    {
	      if ( pe->GetEdge()->GetVertex()->GetGeneration() >= pe->GetVertex()->GetGeneration() )
		{
		  pppvxCrease[uiCreases][1] = ppvxCrease[1];
		  ppptCrease[uiCreases][1] = pptCrease[1];

		  ppeCrease[0] = pe->GetEdgePrev(1);
		  ppeCrease[2] = pe->GetEdgeNext(1);
		  for ( i = 0; i < 3; i += 2 )
		    {
		      pe = ppeCrease[i];
		      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
			{
			  pppvxCrease[uiCreases][i] = pe->GetEdge()->GetVertex();
			  ppptCrease[uiCreases][i] = pppvxCrease[uiCreases][i]->GetPoint();
			}
		      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
			{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
			  ppptCrease[uiCreases][i] = pe->GetEdge()->GetVertex()->GetTexture();
			  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
			  pppvxCrease[uiCreases][i] = peCurr->GetVertex();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
			  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
			  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
			  pppvxCrease[uiCreases][i] = peCurr->GetVertex();
			  ASSERT( pppvxCrease[uiCreases][i] == pe->GetEdge()->GetVertex()->GetForwardVertex() );
			  ppptCrease[uiCreases][i] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pppvxCrease[uiCreases][i], pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
			}
		      else
			{
			  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
			  
			  pppvxCrease[uiCreases][i] = pe->GetEdge()->GetVertex();
			  ppptCrease[uiCreases][i] = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(pppvxCrease[uiCreases][i]);
			}
		    }
		}
	      else
		{
		  pppvxCrease[uiCreases][0] = NULL;
		  ppptCrease[uiCreases][0] = pptCrease[0];
		  pppvxCrease[uiCreases][1] = ppvxCrease[1];
		  ppptCrease[uiCreases][1] = pptCrease[1];
		  pppvxCrease[uiCreases][2] = NULL;
		  ppptCrease[uiCreases][2] = pptCrease[2];
		}
	    }
	  uiCreases++;
	}
    }
  
  ptCorner = pvx->GetPoint();
  if ( uiCreases >= 3 )
    {
      v = ptCorner;
      return v;
    }

  if ( uiCreases == 2 )
    {
      pppt[0] = ppptCrease[0];
      pppt[1] = ppptCrease[1];

      peSharpType[0] = pppvxCrease[0][1]->GetSharpVertexType();
      peSharpType[1] = pppvxCrease[1][1]->GetSharpVertexType();
      if ( peSharpType[0] != svtRegularCrease &&
	   ( peSharpType[1] == svtRegularCrease ||
	     ( !(peSharpType[0] == svtDart || peSharpType[0] == svtSmooth) &&
	       (peSharpType[1] == svtDart || peSharpType[1] == svtSmooth) ) ) )
	{
	  // Swap
	  pppt[0] = ppptCrease[1];
	  pppt[1] = ppptCrease[0];

	  eSharpType = peSharpType[0];
	  peSharpType[0] = peSharpType[1];
	  peSharpType[1] = eSharpType;
	}

      switch ( peSharpType[0] )
	{
	case svtRegularCrease:
	  {
	    switch ( peSharpType[1] )
	      {
	      case svtRegularCrease:
		v = ( ( (2.0 * ptCorner) - 
			( pppt[0][1] +
			  pppt[1][1] ) ) / 6.0 );
		break;
	      case svtDart:
	      case svtSmooth:
		v = ( ( (5.0 * ptCorner) - 
			( (2.0 * pppt[0][1]) +
			  pppt[1][0] + 
			  pppt[1][1] +
			  pppt[1][2] ) ) / 48.0 );
		break;
	      case svtNonregularCrease:
	      case svtCorner:
		v = ( ( (3.0 * ptCorner) -
			( (2.0 * pppt[0][1]) +
			  pppt[1][1] ) ) / 48.0 );
		break;
	      default:
		ASSERT( FALSE );
		break;
	      }
	  }
 	  break;
	case svtDart:
	case svtSmooth:
	  {
	    switch ( peSharpType[1] )
	      {
	      case svtDart:
	      case svtSmooth:
		v = ( ( (6.0 * ptCorner) - 
			( pppt[0][0] + 
			  pppt[0][1] + 
			  pppt[0][2] + 
			  pppt[1][0] + 
			  pppt[1][1] +
			  pppt[1][2] ) ) / 48.0 );
		break;
	      case svtNonregularCrease:
	      case svtCorner:
		v = ( ( (4.0 * ptCorner) - 
			( pppt[0][0] +
			  pppt[0][1] + 
			  pppt[0][2] +
			  pppt[1][1] ) ) / 48.0 );
		break;
	      default:
		ASSERT( FALSE );
		break;
	      }
	  }
	  break;
	case svtNonregularCrease:
	case svtCorner:
          ASSERT( peSharpType[1] == svtNonregularCrease || peSharpType[1] == svtCorner );
	  v = ( ( (2.0 * ptCorner) -
		  ( pppt[0][1] +
		    pppt[1][1] ) ) / 48.0 );
	  break;
	default:
	  ASSERT( FALSE );
	  break;
	}
      v[3] = 0.0;

      return v;
    }

  v = ptCorner - (ptSmooth / (FLOAT)uiNeighbors); 
  v[3] = 0.0;

  return v;
}

Point QEMesh::CalculateVertexLimitSubdivisionCornerRoundingSharpSelective(QEVertex *pvx)
{
  Point pt, ptCorner, ptSmooth, ptCrease;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *ppeCrease[3], *ppe[3];
  UINT_32 uiNeighbors, uiCreases, uiIndex;
  FLOAT fA, fEpsilon;

  BOOL ppbCreasePointReady[3][3];
  Point ppptCrease[3][3];
  SharpVertexType psvtCrease[3];
  const FLOAT *pfCoeffTable;
  UINT_32 i, j, k, uiStart, uiEnd, uiCenter;

  iterQEVertex.Init(pvx);

  // Sharp Loop Vertex-Vertex Subdivision
  ptSmooth.Null();
  memset(ppbCreasePointReady, 0, sizeof(ppbCreasePointReady));
  uiCreases = 0;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
        pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );

      uiIndex = MIN(uiCreases, 2);
      ppeCrease[uiIndex] = pe;

      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pe->GetEdge()->GetVertex()->GetPoint();

	  ptSmooth += pt; 
	  ppptCrease[uiIndex][1] = pt;
	  ppbCreasePointReady[uiIndex][1] = TRUE;
	  psvtCrease[uiIndex] = pe->GetEdge()->GetVertex()->GetSharpVertexType();

	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  psvtCrease[uiIndex] = peCurr->GetVertex()->GetSharpVertexType();

#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pt = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  pt = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(peCurr->GetVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

	  ptSmooth += pt;
	  ppptCrease[uiIndex][1] = pt;
	  ppbCreasePointReady[uiIndex][1] = TRUE;

	  uiNeighbors++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  // This a smooth edge
	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pt = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pt = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetEdgeNext(1));
	    }
	  ptSmooth += pt;
	  ppptCrease[uiIndex][0] = pt;
	  ppbCreasePointReady[uiIndex][0] = TRUE;

	  pt = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetVertex());
	  ptSmooth += pt;
	  ppptCrease[uiIndex][1] = pt;
	  ppbCreasePointReady[uiIndex][1] = TRUE;
	  psvtCrease[uiIndex] = pe->GetEdge()->GetVertex()->GetSharpVertexType();

	  // This a smooth edge
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      pt = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      pt = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdgeNext(0));
	    }
	  ptSmooth += pt;
	  ppptCrease[uiIndex][2] = pt;
	  ppbCreasePointReady[uiIndex][2] = TRUE;

	  uiNeighbors += 3;
	}

      if ( pe->GetKnotInterval() != 0.0 )
	{
	  ASSERT( pe->GetKnotInterval() == INFINITY );
	  uiCreases++;
	}
    }
  
  ptCorner = pvx->GetPoint();
  if ( uiCreases >= 3 )
    {
      return ptCorner;
    }

  if ( uiCreases == 2 )
    {
      ASSERT( ( pvx->GetSharpVertexType() == svtRegularCrease && uiNeighbors == 6 ) ||
	      pvx->GetSharpVertexType() == svtNonregularCrease );
      uiCenter = scpuiSharpLoopSharpTypeIndex[pvx->GetSharpVertexType()];
      uiStart = scpuiSharpLoopSharpTypeIndex[psvtCrease[0]];
      uiEnd = scpuiSharpLoopSharpTypeIndex[psvtCrease[1]];
      pfCoeffTable = scppppfSharpLoopLimitTable[uiCenter][uiStart][uiEnd];

      for ( i = 0; i < 2; i++ )
	{
	  if ( !( ppbCreasePointReady[i][0] && ppbCreasePointReady[i][2] ) )
	    {
	      ppe[0] = ppeCrease[i]->GetEdgePrev(1);
	      ppe[2] = ppeCrease[i]->GetEdgeNext(1);
	      for ( j = 0; j < 3; j += 2 )
		{
		  pe = ppe[j];
		  
		  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pe->GetVertex()->GetGeneration() );
		  if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
		    {
		      ppptCrease[i][j] = pe->GetEdge()->GetVertex()->GetPoint();
		    }
		  else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
		    {
		      peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
		      pt = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
		      ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
		      ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
		      pt = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(peCurr->GetVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
		      ppptCrease[i][j] = pt;
		    }
		  else
		    {
		      ASSERT( FALSE );
		    }
		}	      
	    }
	}

      i = 0;
      ptCrease = pfCoeffTable[i] * ptCorner;
#ifdef DEBUGGING_SHARP_LOOP_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	  fprintf(stderr, "  %f * ", pfCoeffTable[i]); 
	  ptCorner.Display();
	} 
#endif // DEBUGGING_SHARP_LOOP_MATH
      for ( j = 0; j < 2; j++ )
	{
	  for ( k = 0; k < 3; k++ )
	    {
	      i++;
	      ptCrease += pfCoeffTable[i] * ppptCrease[j][k];
#ifdef DEBUGGING_SHARP_LOOP_MATH
	      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
		   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )		   
		{
		  fprintf(stderr, "  %f * ", pfCoeffTable[i]);
		  ppptCrease[j][k].Display();
		} 
#endif // DEBUGGING_SHARP_LOOP_MATH
	    }
	}
      ptCrease[3] = 1.0;
#ifdef DEBUGGING_SHARP_LOOP_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "  Limit = "); 
	  ptCorner.Display();
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	} 
#endif // DEBUGGING_SHARP_LOOP_MATH

      return ptCrease;
    }

  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiNeighbors) / (8.0 * fA);
  ptSmooth = (ptSmooth + (fEpsilon * ptCorner)) / (fEpsilon + (FLOAT)uiNeighbors);
  ptSmooth[3] = 1.0;

  return ptSmooth;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, vNormalMax;
  //Vector v;
  Vector vCenter;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCrease;
  UINT_32 uiNeighbors, uiCreases, i, j, k, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  FLOAT fSharpness, fMag, fMagMax;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];
  BOOL pbCrease[cuiNeighborsMax];
  UINT_32 puiSegmentLength[cuiNeighborsMax];

  iterQEVertex.Init(pvx);

  // Search for a crease edge (if one exists it will be in the mesh)
  peCrease = NULL;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else
	{
	  uiNeighbors += 3;
	  // This can add at most one crease edge
	}

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ASSERT( fSharpness == INFINITY );
	  peCrease = pe;
	}
    }
  ASSERT( uiNeighbors < cuiNeighborsMax );

  // Calculate all the neighbor vertices
  if ( peCrease == NULL )
    {
      peCrease = iterQEVertex.PeekFirstEdge();
    }
  uiCreases = 0;
  memset(pbCrease, 0, sizeof(pbCrease));
  memset(puiSegmentLength, 0, sizeof(puiSegmentLength));
  i = 0;
  pe = peCrease;
  do
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );

      fSharpness = pe->GetKnotInterval();
      ASSERT( fSharpness == 0.0 || fSharpness == INFINITY );

      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiNeighbors );
	  pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pv[i] = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  ASSERT( i < uiNeighbors );
	  pv[i] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(peCurr->GetVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  // This a smooth edge
	  j = i;
	  if ( i == 0 )
	    {
	      j = uiNeighbors - 1;
	    }
	  else
	    {
	      i++;
	    }
	  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      ASSERT( j < uiNeighbors );
	      pv[j] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      ASSERT( j < uiNeighbors );
	      pv[j] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetEdgeNext(1));
	    }

	  ASSERT( i < uiNeighbors );
	  pv[i] = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetVertex());
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;

	  // This a smooth edge
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	    {
	      ASSERT( i < uiNeighbors );
	      pv[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	    }
	  else
	    {
	      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
	      ASSERT( i < uiNeighbors );
	      pv[i] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdgeNext(0));
	    }
	  i++;
	}
      pe = pe->GetEdgeNext(1);
    }
  while( pe != peCrease );

  for ( i = 0; i < uiNeighbors; i++ )
    {      
      pv[i][3] = 1.0;
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;
  if ( uiCreases < 2 )
    {
      vT0.Null();
      vT1.Null();
      fTheta = 0.0;
      fThetaDelta = 2.0 * PI / (FLOAT)uiNeighbors;
      for ( i = 0; i < uiNeighbors; i++ )
	{
	  vT1 += cos(fTheta) * pv[i];
	  fTheta += fThetaDelta;
	  vT0 += cos(fTheta) * pv[i];
	}
      vT0[3] = 0.0;
      vT1[3] = 0.0;
      vNormal = vT0 * vT1;
    }
  else
    {
      // Copy the first edge to the end
      pv[uiNeighbors] = pv[0]; 
      pbCrease[uiNeighbors] = pbCrease[0];
      j = 0;
      for ( i = 0; i < uiNeighbors; i++ )
	{      
	  if ( pbCrease[i] )
	    {
	      j++;
	    }
	  ASSERT( j > 0 && j <= uiCreases );
	  puiSegmentLength[j-1]++;
	}

      bRegularCrease = ( uiNeighbors == 6 && uiCreases == 2 &&
			 puiSegmentLength[0] == 3 && puiSegmentLength[1] == 3 ); 

      // Calculate the error for the individual segments and take the max
      vNormalMax.Null();
      fMagMax = 0.0;
      k = 0;
      for ( i = 0; i < uiCreases; k += puiSegmentLength[i], i++ )
	{
	  vT0.Null();
	  vT1.Null();
	  if ( bRegularCrease )
	    {
	      vT0 = pv[k];
	      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
	      for ( j = 0; j <= puiSegmentLength[i]; j++ )
		{
		  vT1 += scppfSharpLoopTangentTable[2][j+1] * pv[k+j];
		}
	      vT0 -= pv[k+puiSegmentLength[i]];	      
	    }
	  else
	    {
	      switch ( puiSegmentLength[i] )
		{
		case 0:
		  ASSERT( FALSE );
		  break;
		case 1:
		case 2:
		  uiIndex = puiSegmentLength[i] - 1;
		  vT0 = pv[k];
		  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
		  for ( j = 0; j <= puiSegmentLength[i]; j++ )
		    {
		      vT1 += scppfSharpLoopTangentTable[uiIndex][j+1] * pv[k+j];
		    }
		  vT0 -= pv[k+puiSegmentLength[i]];
		  break;
		default:
		  fThetaDelta = PI / (FLOAT)puiSegmentLength[i];
		  fA = sin(fThetaDelta);
		  vT1 = pv[k];
		  vT0 = (fA * pv[k]);
		  vT1 -= pv[k+puiSegmentLength[i]];
		  vT0 += (fA * pv[k+puiSegmentLength[i]]);
		  fA = 2.0 * (cos(fThetaDelta) - 1.0);
		  for ( j = 1, fTheta = fThetaDelta; 
			j < puiSegmentLength[i];
			j++, fTheta += fThetaDelta )
		    {
		      vT0 += ((fA * sin(fTheta)) * pv[k+j]);
		    }
		  break;
		}
	    }
	  vT0[3] = 0.0;
	  vT1[3] = 0.0;
	  vNormal = vT0 * vT1;
	  fMag = vNormal.Magnitude();
	  if ( fMag > fMagMax )
	    {
	      fMagMax = fMag;
	      vNormalMax = vNormal;
	    }
	}
      vNormal = vNormalMax;
    }

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCornerRoundingSharpSelective(QEEdge *peCrease)
{
  Vector vT0, vT1, vNormal;
  Vector vCenter;
  QEVertex *pvx;
  QEEdge *pe;
  UINT_32 uiSegmentLength, i, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  FLOAT fSharpness;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];

  pvx = peCrease->GetVertex();
  ASSERT( pvx->GetSharpVertexType() == svtRegularCrease ||
	  pvx->GetSharpVertexType() == svtNonregularCrease ||
	  pvx->GetSharpVertexType() == svtCorner );
  ASSERT( peCrease->GetKnotInterval() == INFINITY );

  // Search for a crease edge (if one exists it will be in the mesh)
  pe = peCrease;
  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
  if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      uiSegmentLength = 1;
    }
  else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
    {
      uiSegmentLength = 1;
    }
  else
    {
      uiSegmentLength = 2;
      // This can add at most one crease edge
    }
  for ( pe = pe->GetEdgeNext(1);
	pe != peCrease;
	pe = pe->GetEdgeNext(1) )
    {
      fSharpness = pe->GetKnotInterval();

      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiSegmentLength++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiSegmentLength++;
	}
      else
	{
	  // This can add at most one crease edge
	  if ( fSharpness != 0.0 )
	    {
	      ASSERT( fSharpness == INFINITY );
	      // This is the ending crease edge
	      uiSegmentLength += 2;
	    }
	  else
	    {
	      uiSegmentLength += 3;
	    }
	}

      if ( fSharpness != 0.0 )
	{
	  ASSERT( fSharpness == INFINITY );
	  break;
	}
    }
  ASSERT( uiSegmentLength < cuiNeighborsMax );

  pe = peCrease;
  for ( i = 0; i < uiSegmentLength; pe = pe->GetEdgeNext(1) )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiSegmentLength );
	  pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
	  pv[i][3] = 1.0;
	  i++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( i < uiSegmentLength );
	  pv[i] = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  ASSERT( i < uiSegmentLength );
	  pv[i] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(peCurr->GetVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pv[i][3] = 1.0;
	  i++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  // This a smooth edge
	  if ( i != 0 )
	    {
	      if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
		{
		  ASSERT( i < uiSegmentLength );
		  pv[i] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
		}
	      else
		{
		  ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
		  ASSERT( i < uiSegmentLength );
		  pv[i] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetEdgeNext(1));
		}
	      i++;
	    }

	  ASSERT( i < uiSegmentLength );
	  pv[i] = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetVertex());
	  i++;

	  // This a smooth edge
	  if ( i < uiSegmentLength )
	    {
	      if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
		{
		  ASSERT( i < uiSegmentLength );
		  pv[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
		}
	      else
		{
		  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
		  ASSERT( i < uiSegmentLength );
		  pv[i] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdgeNext(0));
		}
	      i++;
	    }
	}
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;

  bRegularCrease = (pvx->GetSharpVertexType() == svtRegularCrease);
  
  // Calculate the error for the individual segments and take the max
  vT0.Null();
  vT1.Null();
  if ( bRegularCrease )
    {
      vT0 = pv[0];
      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
      for ( i = 0; i < uiSegmentLength; i++ )
	{
	  vT1 += scppfSharpLoopTangentTable[2][i+1] * pv[i];
	}
      vT0 -= pv[uiSegmentLength - 1];	      
    }
  else
    {
      switch ( uiSegmentLength )
	{
	case 0:
	case 1:
	  ASSERT( FALSE );
	  break;
	case 2:
	case 3:
	  uiIndex = uiSegmentLength - 2;
	  vT0 = pv[0];
	  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
	  for ( i = 0; i < uiSegmentLength; i++ )
	    {
	      vT1 += scppfSharpLoopTangentTable[uiIndex][i+1] * pv[i];
	    }
	  vT0 -= pv[uiSegmentLength - 1];
	  break;
	default:
	  fThetaDelta = PI / (FLOAT)(uiSegmentLength - 1);
	  fA = sin(fThetaDelta);
	  vT1 = pv[0];
	  vT0 = (fA * pv[0]);
	  vT1 -= pv[uiSegmentLength - 1];
	  vT0 += (fA * pv[uiSegmentLength - 1]);
	  fA = 2.0 * (cos(fThetaDelta) - 1.0);
	  for ( i = 1, fTheta = fThetaDelta; 
		i < uiSegmentLength - 1;
		i++, fTheta += fThetaDelta )
	    {
	      vT0 += ((fA * sin(fTheta)) * pv[i+1]);
	    }
	  break;
	}
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;
  vNormal = vT0 * vT1;

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

Point QEMesh::CalculateEdgeSubdivisionCornerRoundingSharpSelective(QEEdge *pe)
{
  QEEdge *ppe[2], *peCurr, *peSwap;
  BOOL pbCrease[2], pbRegularCrease[2], bCrease;
  Point pt, ptSmooth, ptCrease;
  Point ppt[6];
  UINT_32 i, uiN, uiSharpEdges;
  FLOAT fSharpness, fT;

  ASSERT( pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() );

  fSharpness = pe->GetKnotInterval();
  bCrease = FALSE;
  if ( fSharpness != 0.0 )
    {
      ppe[0] = pe;
      ppe[1] = pe->GetEdge();
      pbCrease[0] = FALSE;
      pbCrease[1] = FALSE;
      pbRegularCrease[0] = FALSE;
      pbRegularCrease[1] = FALSE;
      for ( i = 0; i < 2; i++ )
	{
	  uiSharpEdges = 1;
	  uiN = 1;
	  for ( peCurr = ppe[i]->GetEdgeNext(1); 
		peCurr != ppe[i]; 
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      uiN++;
	      if ( peCurr->GetKnotInterval() != 0.0 )
		{
		  uiSharpEdges++;
		}
	    }

	  if ( uiSharpEdges < 2 )
	    {
	      break;
	    }

	  pbCrease[i] = TRUE;
	  if ( uiSharpEdges == 2 && uiN == 6 &&
	       ppe[i]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval() != 0.0 )
	    {
	      pbRegularCrease[i] = TRUE;
	    }	  
	}

      bCrease = pbCrease[0] && pbCrease[1];
      if ( bCrease )
	{
	  if ( ( pbRegularCrease[0] && pbRegularCrease[1] ) ||
	       ( !pbRegularCrease[0] && !pbRegularCrease[1] ) )
	    {
	      ptCrease = ( ppe[0]->GetVertex()->GetPoint() + ppe[1]->GetVertex()->GetPoint() ) / 2.0;
	      ptCrease[3] = 1.0;
	    }
	  else
	    {
	      if ( pbRegularCrease[1] )
		{
		  peSwap = ppe[0];
		  ppe[0] = ppe[1];
		  ppe[1] = peSwap;
		}
	      ptCrease = ( (5.0 * ppe[0]->GetVertex()->GetPoint()) + (3.0 * ppe[1]->GetVertex()->GetPoint()) ) / 8.0;
	      ptCrease[3] = 1.0;
	      
	    }

	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      return ptCrease;
	    }
	}
    }

  // Midpoint smooth rule
  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();

  if ( pe->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[2] = pe->GetEdgePrev(0)->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdgePrev(0)->GetVertex()->GetForwardVertex() != NULL );
      peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdgePrev(0));
      ASSERT( peCurr->GetVertex() == pe->GetEdgePrev(0)->GetVertex()->GetForwardVertex() );
      ppt[2] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdgePrev(0)->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[3] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[3] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
      peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdgeNext(0)->GetEdge());
      ASSERT( peCurr->GetVertex() == pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex() );
      ppt[3] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[4] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[4] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetForwardVertex() != NULL );
      peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge()->GetEdgePrev(0));
      ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetForwardVertex() );
      ppt[4] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[5] = pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
    }
  else
    {
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[5] = pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
      peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge()->GetEdgeNext(0)->GetEdge());
      ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex() );
      ppt[5] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex()->GetForwardVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  ptSmooth = ((6.0 * ppt[0]) + (6.0 * ppt[1]) + ppt[2] + ppt[3] + ppt[4] + ppt[5]) / 16.0;
  ptSmooth[3] = 1.0;
  
  if ( !bCrease )
    {
      return ptSmooth;
    } 

  fT = fSharpness;
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
  pt[3] = 1.0;
  return pt;
}

//////////////////////////////////////////////////
//  Catmull-Clark Selective
//

QEMesh *QEMesh::SubdivideCatmullClarkSelective(QEMesh::ErrorMetricType eErrorMetric,
					       FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertex *pvx;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;
  Vector v;

  v.Null();

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideCatmullClarkSelective(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);

      pvx->SetGeneration(0);
      pvx->SetTexture(v);
    }

  pme->Preprocess();

  pme->ReSubdivideCatmullClarkSelective(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::ReSubdivideCatmullClarkSelective(QEMesh::ErrorMetricType /*eErrorMetric*/,
					      FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr, *ppvx[5];
  QEVertexIter iterQEVertex;
  QEEdge *pe,  *peCurr;
  QEEdge *peStart;
  FLOAT fError;
  //Vector v;
  UINT_32 uiN, uiSize, i;

  uiN = 0;
  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      // Unnecessarily Strict Rule that all vertices around the facets which are around
      // the central vertex must be the same generation or younger
      uiN++;
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  pvxCurr = NULL;
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideCatmullClarkSelective(pvx);

      // Update the current vertex
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      fError = CalculateVertexErrorSubdivideCatmullClarkSelective(pvx);
      m_pheapSelective->Update(phe, fError);

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  i = 0;
	  peCurr = pe->GetEdge();
	  ppvx[i] = peCurr->GetVertex();
	  i++;
	  peCurr = peCurr->GetEdgePrev(1);
	  ppvx[i] = peCurr->GetEdgePrev(1)->GetEdge()->GetVertex();
	  i++;
	  peCurr = peCurr->GetEdge();
	  ppvx[i] = peCurr->GetVertex();
	  i++;
	  if ( peCurr->GetEdgeNext(1) != peCurr->GetEdgePrev(1) )
	    {
	      peCurr = peCurr->GetEdgeNext(1);
	      ppvx[i] = peCurr->GetEdge()->GetVertex();
	      i++;
	      peCurr = peCurr->GetEdgeNext(1);
	      ppvx[i] = peCurr->GetEdge()->GetVertex();
	      i++;	      
	    }

	  uiSize = i;
	  for ( i = 0; i < uiSize; i++ )
	    {
	      // Insert the new vertices
	      pvxCurr = ppvx[i];
	      if ( pvxCurr->GetHeapEntry() == NULL )
		{   
		  fError = CalculateVertexErrorSubdivideCatmullClarkSelective(pvxCurr);
		  phe = m_pheapSelective->Insert(fError, pvxCurr);
		  pvxCurr->SetHeapEntry(phe);
		}
	      else
		{
		  phe = pvxCurr->GetHeapEntry();
		  ASSERT( phe != NULL );
		  fError = CalculateVertexErrorSubdivideCatmullClarkSelective(pvxCurr);
		  m_pheapSelective->Update(phe, fError);
		}
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  peStart = pe->GetEdgeNext(0)->GetEdgeNext(0);
	  peCurr = peStart;
	  do
	    {
	      peCurr->GetFacet()->Preprocess();

	      peCurr = peCurr->GetEdgeNext(1);
	    }
	  while ( peCurr != peStart );
	}
    }
  /*  
  if ( uiN > 0 )
    {
      Preprocess();
    }
    */
}

///////////////////////////////////////////////////
// QEMesh::SubdivideCatmullClark                 //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     v4--e3--v3--e2--v2    //
//           |              |       |       |    //
//           |             e4      e9      e1    //
//           |              |       |       |    //
//  e2------vV------e0  => v5--e10-vV--e8--v1    //
//           |              |       |       |    //
//           |             e5      e11     e0    //
//           |              |       |       |    //
//  f2      e3      f3     v6--e6--v7--e7--v0    //
//                                               //
VOID QEMesh::SubdivideCatmullClarkSelective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEVertex *pvxNew;
  QEEdge *pe, *peNew, *peCurr;
  QEFacet *pfaNew;
  //Point ptNew;
  Point pt;
  Vector vCenter, vNull;

  vCenter = pvx->GetPoint();
  vNull.Null();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Create the cross edges and vertices if necessary
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );
	  pe->GetEdge()->SetForwardEdge(peNew);

	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );
	  pe->SetForwardEdge(peNew);

	  pt.Null();
	  pvxNew = CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pvxNew->SetGeneration(pvx->GetGeneration()+1);
	  pvxNew->SetTexture(vCenter);

	  pe->GetEdge()->GetForwardEdge()->GetEdge()->SetVertex(pvxNew);
	  pe->GetForwardEdge()->SetVertex(pvxNew);
	  Splice(1, pe->GetEdge()->GetForwardEdge()->GetEdge(), pe->GetForwardEdge());

	  pvxNew = NULL;
	  for ( peCurr = pe->GetEdgeNext(0);
		peCurr != pe;
		peCurr = peCurr->GetEdgeNext(0) )
	    {
	      if ( pvx->GetGeneration() < peCurr->GetVertex()->GetGeneration() )
		{
		  peCurr = peCurr->GetEdge();
		  pvxNew = peCurr->GetVertex();
		  pe->GetForwardEdge()->GetEdge()->SetVertex(pvxNew);
		  peCurr->SetVertex(pvxNew);
		  break;
		}
	    }
	  if ( pvxNew == NULL )
	    {
	      pt = CalculateFacetSubdivisionCatmullClarkSelective(pe->GetFacet());
	      pvxNew = CreateVertex(pt);
	      ASSERT( pvxNew != NULL );
	      pvxNew->SetGeneration(pvx->GetGeneration()+1);
	      pvxNew->SetTexture(vNull);

	      pe->GetForwardEdge()->GetEdge()->SetVertex(pvxNew);	      
	    }
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Calculate the new midpoint positions
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  pt = CalculateEdgeSubdivisionCatmullClarkSelective(pe);
	  pe->GetForwardEdge()->GetVertex()->SetPoint(pt);
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Connect the cross edges
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );
	  
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	  peNew->SetVertex(pe->GetForwardEdge()->GetVertex());
	  pe->GetEdge()->SetVertex(pe->GetForwardEdge()->GetVertex());

	  peNew->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  pe->GetForwardEdge()->SetFacet(pe->GetFacet());
	  pe->GetForwardEdge()->GetEdge()->SetFacet(pe->GetFacet());
	  pe->GetEdge()->GetForwardEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  pe->GetEdge()->GetForwardEdge()->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());

	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	  Splice(1, pe->GetEdge(), pe->GetEdge()->GetEdgePrev(1));

	  Splice(1, pe->GetForwardEdge()->GetEdgePrev(1), peNew);
	  Splice(1, pe->GetForwardEdge(), pe->GetEdge());
	}
    }

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Connect at the corners and create new facets where necessary
      if ( pe->GetForwardEdge() != NULL && 
	   pe->GetEdgePrev(0)->GetForwardEdge() != NULL )
	{
	  if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetEdge() == pe->GetEdgeNext(0)->GetEdge() )
	    {
	      // New facet vertex - 1 new facet

	      Splice(1, pe->GetEdgeNext(0)->GetEdge(), pe->GetEdgePrev(0)->GetEdgePrev(0));

	      pe->GetEdgePrev(0)->GetEdgePrev(0)->SetVertex(pe->GetEdgeNext(0)->GetEdge()->GetVertex());
	      
	      // Deal with the facets
	      pe->GetEdgeNext(0)->GetEdge()->SetFacet(pe->GetEdgeNext(0)->GetEdge()->GetFacet());

	      pfaNew = CreateFacet();
	      ASSERT( pfaNew != NULL );
	      peCurr = pe;
	      do
		{
		  peCurr->SetFacet(pfaNew);
		  peCurr = peCurr->GetEdgeNext(0);
		}
	      while ( pe != peCurr );
	    }
	  else
	    {
	      // 2 new facets

	      Splice(1, pe->GetEdgeNext(0)->GetEdge(), pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetEdge()->GetEdgePrev(1));
	      Splice(1, pe->GetEdgeNext(0)->GetEdge()->GetEdgePrev(1), pe->GetEdgePrev(0)->GetEdgePrev(0));

	      pe->GetEdgePrev(0)->GetEdgePrev(0)->SetVertex(pe->GetEdgeNext(0)->GetEdge()->GetVertex());
	      
	      // Deal with the facets
	      pe->GetEdgeNext(0)->GetEdge()->SetFacet(pe->GetEdgeNext(0)->GetEdge()->GetFacet());

	      pfaNew = CreateFacet();
	      ASSERT( pfaNew != NULL );
	      peCurr = pe;
	      do
		{
		  peCurr->SetFacet(pfaNew);
		  peCurr = peCurr->GetEdgeNext(0);
		}
	      while ( pe != peCurr );

	      pfaNew = CreateFacet();
	      ASSERT( pfaNew != NULL );
	      peCurr = pe->GetEdgePrev(0)->GetEdgeNext(1);
	      do
		{
		  peCurr->SetFacet(pfaNew);
		  peCurr = peCurr->GetEdgeNext(0);
		}
	      while ( pe->GetEdgePrev(0)->GetEdgeNext(1) != peCurr );
	    }
	}
      else if ( pe->GetForwardEdge() != NULL )
	{
	  // 1 new facet

	  Splice(1, pe->GetEdgeNext(0)->GetEdge(), pe->GetEdgePrev(0)->GetEdgePrev(0));
	  
	  // Deal with the facets
	  pe->GetEdgeNext(0)->GetEdge()->SetFacet(pe->GetEdgeNext(0)->GetEdge()->GetFacet());
	  
	  pfaNew = CreateFacet();
	  ASSERT( pfaNew != NULL );
	  peCurr = pe;
	  do
	    {
	      peCurr->SetFacet(pfaNew);
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( pe != peCurr );
	}
      else if ( pe->GetEdgePrev(0)->GetForwardEdge() != NULL )
	{
	  // 1 new facet

	  Splice(1, pe->GetEdgeNext(0)->GetEdge()->GetEdgePrev(1), pe->GetEdgePrev(0)->GetEdgePrev(0));
	  
	  pe->GetEdgePrev(0)->GetEdgePrev(0)->SetVertex(pe->GetEdgeNext(0)->GetEdge()->GetVertex());

	  // Deal with the facets
	  pe->GetEdgePrev(0)->GetEdgePrev(0)->GetEdge()->SetFacet(pe->GetEdgePrev(0)->GetEdgePrev(0)->GetEdge()->GetFacet());
	  
	  pfaNew = CreateFacet();
	  ASSERT( pfaNew != NULL );
	  peCurr = pe;
	  do
	    {
	      peCurr->SetFacet(pfaNew);
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( pe != peCurr );
	}
      else
	{
	  // Don't do anything
	}

      pe->SetForwardEdge(NULL);
      pe->GetEdgePrev(0)->SetForwardEdge(NULL);
    }

  pvx->SetTexture(vNull);
  pt = CalculateVertexSubdivisionCatmullClarkSelective(pvx);
  pvx->SetPoint(pt);
  pvx->SetGeneration(pvx->GetGeneration()+1);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideCatmullClarkSelective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  ASSERT( uiN > 1 );
  if ( uiN < 3 )
    {
      //fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return -1.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

Point QEMesh::CalculateFacetSubdivisionCatmullClarkSelective(QEFacet *pfa)
{
  Point pt;
  QEFacetIter iter;
  QEVertex *pvx;
  UINT_32 uiNeighbors, uiGeneration;

  uiGeneration = pfa->GetEdge()->GetVertex()->GetGeneration();

  iter.Init(pfa);
  pt.Null();
  for ( uiNeighbors = 0, pvx = iter.PeekFirstVertex();
	pvx != NULL;
	uiNeighbors++, pvx = iter.PeekNextVertex() )
    {
      ASSERT( pvx->GetGeneration() == uiGeneration );
      pt += pvx->GetPoint();
    }

  pt /= (FLOAT)uiNeighbors;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateVertexSubdivisionCatmullClarkSelective(QEVertex *pvx)
{
  Point pt;
  Point ppt[2];
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 uiN;

  iter.Init(pvx);

  // Loop Vertex-Vertex Subdivision
  ppt[0].Null();
  ppt[1].Null();
  for ( uiN = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiN++, pe = iter.PeekNextEdge() )
    {
      // Facets
      ppt[0] += pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
      // Vertices
      if ( pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  ppt[1] += pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->GetVertex()->GetPoint();	  
	}
      else
	{
	  ppt[1] += pe->GetEdge()->GetVertex()->GetTexture();	  
	}
    }
  
  pt = pvx->GetPoint();
  pt = (((FLOAT)(uiN - 2)*pt) + ((ppt[0] + ppt[1])/(FLOAT)uiN)) / (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateEdgeSubdivisionCatmullClarkSelective(QEEdge *pe)
{
  Point pt;
  Point ppt[4];
  
  // Move the midpoint
  // Loop Edge-Vertex Subdivision
  ASSERT( pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() );
  ASSERT( pe->GetForwardEdge() != NULL );

  ppt[0] = pe->GetVertex()->GetPoint();
  ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
  ppt[2] = pe->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
  if ( pe->GetEdgePrev(1)->GetForwardEdge() == NULL )
    {
      ppt[3] = pe->GetEdgePrev(1)->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
    }
  else
    {
      ppt[3] = pe->GetEdgePrev(1)->GetForwardEdge()->GetEdge()->GetVertex()->GetPoint();
    }

  pt = (ppt[0] + ppt[1] + ppt[2] + ppt[3]) / 4.0;
  pt[3] = 1.0;
  
  return pt;
}

#ifdef LOOP_FACET_CENTRIC
//////////////////////////////////////////////////////////////////////
// Selective Refinement Subdivision Facet Centered Sharp Loop Limit Positions

#define SELECTIVE_SHARP_LOOP_FORWARD_ONLY

QEMesh *QEMesh::SubdivideCornerRoundingFacetSharpSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
							   FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx;
  QEFacet *pfa;
  Point pt;
  Vector v, vNormal;
  FLOAT fError;
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  UINT_32 uiN;
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

  eErrorMetric = emtBestFitPlane;

  pme = Copy();
  ASSERT( pme != NULL );
  pme->CalculateSubdivisionSharpVertexTypesSharpLoop();
  v.Null();
  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pvx->SetGeneration(0);
      pvx->SetBirth(0);
      pvx->SetTexture(v);
    }
  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      pfa->SetGeneration(0);
      pfa->SetBirth(0);
    }
  //pme->PreSubdivideCornerRoundingFacetSharpSelectiveLimit();
  //pme->CalculateSubdivisionLimitPositionsAndNormalsSharpLoop();

  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      pt = pme->CalculateVertexLimitSubdivisionCornerRoundingFacetSharpSelective(pvx);
      pvx->SetLimitPosition(pt);
      vNormal = pme->CalculateVertexNormalSubdivisionCornerRoundingFacetSharpSelective(pvx);
      vNormal.CompletePlaneEquation(pt);
      pvx->SetNormal(vNormal);
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      v = pme->CalculateGenerationVectorSubdivisionCornerRoundingFacetSharpSelective(pvx);
      pvx->SetTexture(v);
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

      //pvx->SetGeneration(0);
      pvx->SetBirth(pvx->GetGeneration());
    }

  for ( pfa = iterQEMesh.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEMesh.PeekNextFacet() )
    {
      fError = pme->CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, pfa);
      phe = pme->m_pheapSelective->Insert(fError, pfa);
      pfa->SetHeapEntry(phe);
      pfa->SetHeapEntryReverse(NULL);

      //pfa->SetGeneration(0);
      pfa->SetBirth(pfa->GetGeneration());
    }

  pme->Preprocess();

  pme->ReSubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::PreSubdivideCornerRoundingFacetSharpSelectiveLimit()
{
  CListInt listQEEdge;
  QEMeshIter iterQEMesh;
  QEFacet *pfa, *pfaCurr, *ppfa[2];
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr;
  FLOAT fSharpness;
  //Vector v;
  UINT_32 i, uiFacets;

  listQEEdge.Init();
  iterQEMesh.Init(this);

  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 && fSharpness != INFINITY )
	{
	  listQEEdge.InsertLast(pe, 0);
	}
    }
  for ( pe = iterQEMesh.PeekFirstEdgeBoundary();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdgeBoundary() )
    {
      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 && fSharpness != INFINITY )
	{
	  listQEEdge.InsertLast(pe, 0);
	}
    }

  for ( pe = (QEEdge*)(listQEEdge.RemoveFirst());
	pe != NULL;
	pe = (QEEdge*)(listQEEdge.RemoveFirst()) )
    {
      fSharpness = pe->GetKnotInterval();
      ASSERT( fSharpness != INFINITY );
      if ( fSharpness == 0.0 )
	{
	  continue;
	}

      ppfa[0] = pe->GetFacet();
      ppfa[1] = pe->GetEdge()->GetFacet();
      uiFacets = 1;
      if ( ppfa[0]->GetGeneration() == ppfa[1]->GetGeneration() )
	{
	  uiFacets = 2;
	}
      else if ( ppfa[0]->GetGeneration() > ppfa[1]->GetGeneration() )
	{
	  ppfa[0] = ppfa[1];
	}

      // Force the edge's neighboring facets to be subdivided
      for ( i = 0; i < uiFacets; i++ )
	{
	  /*
	  fprintf(stderr, "FORCING %ld: Gen = %ld: f = %p = ", 
		  i, ppfa[i]->GetGeneration(), ppfa[i]);
	  ppfa[i]->GetPoint().Display();
	  */
	  pfa = NULL;
	  while ( pfa != ppfa[i] )
	    {	  
	      pfa = ppfa[i];
	      pfaCurr = pfa;
	      while ( pfaCurr != NULL )
		{
		  pfa = pfaCurr;
		  pfaCurr = NULL;

		  iterQEFacet.Init(pfa);
		  for ( pe = iterQEFacet.PeekFirstEdge();
			pe != NULL && pfaCurr == NULL;
			pe = iterQEFacet.PeekNextEdge() )
		    {
		      ASSERT( pe->GetVertex()->GetGeneration() >= pfa->GetGeneration() );
		      if ( pe->GetVertex()->GetGeneration() == pfa->GetGeneration() )
			{
			  for ( peCurr = pe->GetEdgePrev(1);
				peCurr != pe;
				peCurr = peCurr->GetEdgePrev(1) )
			    {
			      pfaCurr = peCurr->GetFacet();
			      if ( pfaCurr->GetGeneration() < pfa->GetGeneration() )
				{
				  break;
				}
			      pfaCurr = NULL;
			    }
			}
		    }
		}

	      /*
	      fprintf(stderr, "  Presubdivide: Gen = %ld: f = %p = ", 
		      pfa->GetGeneration(), pfa);
	      pfa->GetPoint().Display();
	      */
	      SubdivideCornerRoundingFacetSharpSelectiveLimit(pfa, TRUE);

	      //////////////////////////////
	      // Update the area of influence around the facet
	      iterQEFacet.Init(pfa);
	      for ( pe = iterQEFacet.PeekFirstEdge();
		    pe != NULL;
		    pe = iterQEFacet.PeekNextEdge() )
		{
		  fSharpness = pe->GetEdgePrev(1)->GetKnotInterval();
		  if ( fSharpness != 0.0 && fSharpness != INFINITY )
		    {
		      listQEEdge.InsertLast(pe->GetEdgePrev(1), 0);
		      listQEEdge.InsertLast(pe->GetEdgePrev(1)->GetEdge(), 0);
		    }
		  fSharpness = pe->GetEdge()->GetEdgePrev(0)->GetKnotInterval();
		  if ( fSharpness != 0.0 && fSharpness != INFINITY )
		    {
		      listQEEdge.InsertLast(pe->GetEdge()->GetEdgePrev(0), 0);
		      listQEEdge.InsertLast(pe->GetEdge()->GetEdgePrev(0)->GetEdge(), 0);
		    }
		}
	    }
	}
    }
}

VOID QEMesh::ReSubdivideCornerRoundingFacetSharpSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
							  FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEFacet *pfa, *pfaCurr;
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr, *peStart;
  FLOAT fError;
  //Vector v;

  for ( pfa = (QEFacet*)(m_pheapSelective->PeekTop(&fError));
	pfa != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pfa = (QEFacet*)(m_pheapSelective->PeekTop(&fError)) )
    {
      pfaCurr = pfa;
      while ( pfaCurr != NULL )
	{
	  pfa = pfaCurr;
	  pfaCurr = NULL;

	  iterQEFacet.Init(pfa);
	  for ( pe = iterQEFacet.PeekFirstEdge();
		pe != NULL && pfaCurr == NULL;
		pe = iterQEFacet.PeekNextEdge() )
	    {
	      ASSERT( pe->GetVertex()->GetGeneration() >= pfa->GetGeneration() );
	      if ( pe->GetVertex()->GetGeneration() == pfa->GetGeneration() )
		{
		  for ( peCurr = pe->GetEdgePrev(1);
			peCurr != pe;
			peCurr = peCurr->GetEdgePrev(1) )
		    {
		      pfaCurr = peCurr->GetFacet();
		      if ( pfaCurr->GetGeneration() < pfa->GetGeneration() )
			{
			  break;
			}
		      pfaCurr = NULL;
		    }
		}
	    }
	}

      SubdivideCornerRoundingFacetSharpSelectiveLimit(pfa);

      // Update the current vertex
      fError = CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, pfa);
      phe = pfa->GetHeapEntry();
      ASSERT( phe != NULL );
      //ASSERT( fError <= phe->GetKey() + EPSILON );
      if ( !(fError <= phe->GetKey() + EPSILON) )
	{
	  fprintf(stderr, "ResubdivideSharpLoopFacet: %f > %f\n", fError, phe->GetKey() + EPSILON);
	}
      m_pheapSelective->Update(phe, fError);

      ASSERT( pfa->GetGeneration() > pfa->GetBirth() );
      phe = pfa->GetHeapEntryReverse();
      if ( phe == NULL )
	{
	  phe = m_pheapSelectiveReverse->Insert(fError, pfa);
	  pfa->SetHeapEntryReverse(phe);
	}      
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}


      //////////////////////////////
      // Update the area of influence around the facet
      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  // Update the neighbor facets
	  peStart = pe->GetEdge()->GetEdgePrev(0);
	  peCurr = peStart;
	  do
	    {
	      pfaCurr = peCurr->GetFacet();

	      fError = CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimit(eErrorMetric, pfaCurr);
	      phe = pfaCurr->GetHeapEntry();
	      if ( phe == NULL )
		{   
		  phe = m_pheapSelective->Insert(fError, pfaCurr);
		  pfaCurr->SetHeapEntry(phe);
		}
	      else
		{
		  m_pheapSelective->Update(phe, fError);
		}
	      
	      phe = pfaCurr->GetHeapEntryReverse();
	      if ( phe != NULL )
		{
		  m_pheapSelectiveReverse->Update(phe, fError);
		}

	      peCurr = peCurr->GetEdgeNext(1);
	    }
	  while ( peCurr != peStart );
	}

      // Update the facet normals
      iterQEFacet.Init(pfa);
      for ( pe = iterQEFacet.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEFacet.PeekNextEdge() )
	{
	  // Update the neighbor facets
	  peStart = pe->GetEdge()->GetEdgePrev(0);
	  peCurr = peStart;
	  do
	    {
	      peCurr->GetFacet()->Preprocess();

	      peCurr = peCurr->GetEdgeNext(1);
	    }
	  while ( peCurr != peStart );
	}
    }
}

// JORDAN -- THIS HAS NOT BEEN IMPLEMENTED YET
VOID QEMesh::UnSubdivideCornerRoundingFacetSharpSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
							  FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr;
  FLOAT fError;

  //fprintf(stderr, "START: UnSubdivideCornerRoundingSharpSelectiveLimit\n");
  //DisplayEuler();

  for ( pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError);
	( pvx != NULL && 
	  ( !(m_plistQEFacet->GetLength() < uiFacetsMax) || !(fError > fTolerance) ) );
	pvx = (QEVertex*)(m_pheapSelectiveReverse->PeekTop(&fError)),
	  m_pheapSelective->PeekTop(&fError) )
    {
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  iterQEVertex.Init(pvx);
	  for ( pe = iterQEVertex.PeekFirstEdge();
		pe != NULL;
		pe = iterQEVertex.PeekNextEdge() )
	    {
	      peCurr = pe->GetEdge();
	      if ( peCurr->GetVertex()->GetGeneration() > pvx->GetGeneration() ) 
		{
		  // Search for the neighboring vertex that can be unsubdivided
		  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
		  pvxCurr = peCurr->GetVertex();
		  break;
		}
	      else
		{
		  ASSERT( peCurr->GetVertex()->GetGeneration() == pvx->GetGeneration() ); 
		}
	    }
	}
      ASSERT( pvx != NULL );
      ASSERT( pvx->GetGeneration() > pvx->GetBirth() );
      ASSERT( pvx->GetHeapEntryReverse() != NULL );

      UnSubdivideCornerRoundingSharpSelectiveLimit(pvx);

      // Update the current vertex
      fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvx);
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Update(phe, fError);

      phe = pvx->GetHeapEntryReverse();
      ASSERT( phe != NULL );
      if ( pvx->GetGeneration() == pvx->GetBirth() )
	{
	  m_pheapSelectiveReverse->Remove(phe);
	  pvx->SetHeapEntryReverse(NULL);
	}
      else
	{
	  m_pheapSelectiveReverse->Update(phe, fError);
	}

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  fError = CalculateVertexErrorSubdivideCornerRoundingSharpSelectiveLimit(eErrorMetric, pvxCurr);
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Update(phe, fError);

	  phe = pvxCurr->GetHeapEntryReverse();
	  if ( phe != NULL )
	    {
	      m_pheapSelectiveReverse->Update(phe, fError);
	    }
	}

      // Update the facet normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	}
    }

  //DisplayEuler();
  //fprintf(stderr, "END  : UnSubdivideCornerRoundingSharpSelectiveLimit\n");
}

////////////////////////////////////////////////////////
// QEMesh::SubdivideCornerRoundingFacetSharpSelectiveLimit //
//                                                    //
//   Old Vertex Disk       Subdivided Facets          //
//                                                    //
//  f1      e1      f0     f1      v1      f0         //
//           |                   /  | \               //
//           |                 e1  e5  e0             //
//           |               /      |     \           //
//  e2------vV------e0  => v2--e6--vV--e4--v0         //
//           |               \      |     /           //
//           |                 e2  e7  e3             //
//           |                   \  | /               //
//  f2      e3      f3     f2      v3      f3         //
//                                                    //
VOID QEMesh::SubdivideCornerRoundingFacetSharpSelectiveLimit(QEFacet *pfa, BOOL bPresubdivide/* = FALSE*/)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvxNew, *pvxCurr;
  QEEdge *pe, *peNew, *peStart;
  QEFacet *pfaNew;
  Point pt;
  //Point ptNew, ptCenter;
  Vector v, vNormal;
  FLOAT fSharpness;

  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      // Create midpoints and edges, but do not connect them
      ASSERT( pe->GetEdge()->GetFacet()->GetGeneration() >= pfa->GetGeneration() );
      if ( pe->GetEdge()->GetFacet()->GetGeneration() == pfa->GetGeneration() )
	{
	  pvxCurr = pe->GetVertex();
	  if ( pvxCurr->GetGeneration() == pfa->GetGeneration() )
	    {
	      v = pvxCurr->GetPoint();
	      pvxCurr->SetTexture(v);
	      pt = CalculateVertexSubdivisionCornerRoundingFacetSharpSelective(pvxCurr);
	      pvxCurr->SetPoint(pt);
	      pvxCurr->SetGeneration(pvxCurr->GetGeneration()+1);
	    }
	  pvxCurr = NULL;

	  pt = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe);
	  pvxNew = CreateVertex(pt);
	  ASSERT( pvxNew != NULL );
	  pvxNew->SetGeneration(pfa->GetGeneration()+1);
	  pvxNew->SetBirth(pvxNew->GetGeneration());
	  if ( pe->GetKnotInterval() == INFINITY )
	    {
	      pvxNew->SetSharpVertexType(svtRegularCrease);
	    }
	  else
	    {
	      pvxNew->SetSharpVertexType(svtSmooth);
	    }
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  v = pt;
	  pvxNew->SetTexture(v);
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );

	  fSharpness = pe->GetKnotInterval();
	  peNew->SetKnotInterval(fSharpness);
	  peNew->GetEdge()->SetKnotInterval(fSharpness);

	  peNew->SetVertex(pvxNew);
	  pe->SetForwardEdge(peNew);
	}
      else
	{
	  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() > pfa->GetGeneration() &&
		  pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() );
	  pe = iterQEFacet.PeekNextEdge();
	  ASSERT( pe != NULL );
	}
    }

  peStart = iterQEFacet.PeekFirstEdge();
  pe = peStart;
  do
    {
      // Connect the midpoints in dividing edges in half
      ASSERT( pe->GetEdge()->GetFacet()->GetGeneration() >= pfa->GetGeneration() );
      if ( pe->GetEdge()->GetFacet()->GetGeneration() == pfa->GetGeneration() )
	{
	  peNew = pe->GetForwardEdge();
	  ASSERT( peNew != NULL );
	  pe->SetForwardEdge(NULL);
	  pvxNew = peNew->GetVertex();
	  ASSERT( pvxNew != NULL );
	  
	  peNew->SetFacet(pe->GetFacet());
	  pe->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pe->GetEdge()->GetFacet());
	  
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());
	  peNew->SetVertex(pvxNew);
	  pe->GetEdge()->SetVertex(pvxNew);
	  
	  Splice(1, pe->GetEdge(), peNew->GetEdge());
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), pe->GetEdge());
	  Splice(1, pe->GetEdge(), peNew);

	  pe = peNew;
	}
      else
	{
	  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() > pfa->GetGeneration() &&
		  pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() );
	  pe = pe->GetEdgeNext(0);
	}

      pe = pe->GetEdgeNext(0);
    }
  while ( pe != peStart );

  pe = peStart;
  do
    {
      // Cut the corners creating new triangles
      pfaNew = CreateFacet();
      ASSERT( pfaNew != NULL );
      pfaNew->SetGeneration(pfa->GetGeneration()+1);
      pfaNew->SetBirth(pfaNew->GetGeneration());

      peNew = CreateEdge();
      ASSERT( peNew != NULL );
      peNew->SetKnotInterval(0.0);
      peNew->GetEdge()->SetKnotInterval(0.0);

      peNew->SetFacet(pfaNew);
      peNew->GetEdge()->SetFacet(pe->GetFacet());
      pe->SetFacet(pfaNew);
      pe->GetEdgePrev(0)->SetFacet(pfaNew);

      peNew->SetVertex(pe->GetEdge()->GetVertex());
      peNew->GetEdge()->SetVertex(pe->GetEdgePrev(0)->GetVertex());

      Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew);
      Splice(1, pe->GetEdgePrev(0), peNew->GetEdge());

      pe = pe->GetEdge()->GetEdgePrev(1)->GetEdgePrev(1)->GetEdgeNext(0);
    }
  while ( pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge() != peStart );

  pfa->SetGeneration(pfa->GetGeneration()+1);

  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      ASSERT( pe->GetEdgePrev(1)->GetEdge()->GetFacet()->GetGeneration() + 1 >= pfa->GetGeneration() );
      if ( pe->GetEdgePrev(1)->GetEdge()->GetFacet()->GetGeneration() < pfa->GetGeneration() )
	{
	  ASSERT( pe->GetEdgePrev(1)->GetEdge()->GetFacet() == pe->GetEdgeNext(1)->GetEdgeNext(1)->GetFacet() );
	  if ( bPresubdivide )
	    {
	      fSharpness = pe->GetEdgePrev(1)->GetKnotInterval();
	      if ( fSharpness != 0.0 && fSharpness != INFINITY )
		{
		  fSharpness = MAX(0.0, fSharpness - 1.0);
		  pe->GetEdgePrev(1)->SetKnotInterval(fSharpness);
		  pe->GetEdgePrev(1)->GetEdge()->SetKnotInterval(fSharpness);
		  pe->GetEdgeNext(1)->GetEdgeNext(1)->SetKnotInterval(fSharpness);
		  pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge()->SetKnotInterval(fSharpness);
		}
	    }
	  else
	    {
	      // Calculate Limit Positions and Normals
	      pvxCurr = pe->GetVertex();
	      if ( pvxCurr->GetHeapEntry() == NULL )
		{   
		  pt = CalculateVertexLimitSubdivisionCornerRoundingFacetSharpSelective(pvxCurr);
		  pvxCurr->SetLimitPosition(pt);
		  vNormal = CalculateVertexNormalSubdivisionCornerRoundingFacetSharpSelective(pvxCurr);
		  vNormal.CompletePlaneEquation(pt);
		  pvxCurr->SetNormal(vNormal);
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
		  v = CalculateGenerationVectorSubdivisionCornerRoundingSharpSelective(pvxCurr);
		  pvxCurr->SetTexture(v);
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
		}
	    }
	}
    }
}

// JORDAN -- THIS IS NOT IMPLEMENTED
//////////////////////////////////////////////////////////
// QEMesh::UnSubdivideCornerRoundingFacetSharpSelectiveLimit //
//                                                      //
//   Old Vertex Disk       Subdivided Facets            //
//                                                      //
//  f1      e1      f0     f1      v1      f0           //
//           |                   /  | \                 //
//           |                 e1  e5  e0               //
//           |               /      |     \             //
//  e2------vV------e0 <=  v2--e6--vV--e4--v0           //
//           |               \      |     /             //
//           |                 e2  e7  e3               //
//           |                   \  | /                 //
//  f2      e3      f3     f2      v3      f3           //
//                                                      //
VOID QEMesh::UnSubdivideCornerRoundingFacetSharpSelectiveLimit(QEFacet *pfa)
{
  ASSERT( FALSE );
  ASSERT( pfa != NULL && pfa->GetGeneration() > 0 );

  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvxCurr;
  QEVertex *pvx;
  QEEdge *pe, *peCurr, *peStart;
  QEFacet *pfaOld;
  Point ptNew, pt;
  UINT_32 uiN;
  FLOAT fAlpha;

  pvx = NULL;

  iterQEVertex.Init(pvx);
  uiN = 0;
  pt.Null();
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Calculate the previous position of the vertex
      uiN++;
      pt += pe->GetEdge()->GetVertex()->GetPoint();
    }

  fAlpha = 3.0 + 2.0 * cos(2.0 * PI / (FLOAT)uiN);
  fAlpha = 40.0 / (fAlpha * fAlpha);

  ptNew = (fAlpha * pvx->GetPoint()) + ((1.0 - fAlpha) / (FLOAT)uiN) * pt;

  iterQEVertex.Init(pvx);
  peStart = NULL;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Clean up the facets
      pfaOld = pe->GetFacet();
      pfa = pe->GetEdgeNext(0)->GetEdge()->GetFacet();
      pe->SetFacet(pfa);
      pe->GetEdgePrev(0)->SetFacet(pfa);

      DestroyFacet(pfaOld);

      // Deal with the edge vertex
      pe = pe->GetEdge();

      ASSERT( pe->GetVertex()->GetGeneration() == pvx->GetGeneration() );
      peCurr = pe->GetEdgeNext(1);
      Splice(1, peCurr, pe);
      if ( peStart == NULL )
	{
	  peStart = peCurr;
	}
      else
	{
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}
      peCurr = pe->GetEdgePrev(1)->GetEdgePrev(1);
      Splice(1, peCurr, peCurr->GetEdgeNext(1));
      pe->SetVertex(pe->GetVertex());

      if ( pe->GetEdgeNext(1) == peCurr )
	{
	  // Remove the edge-vertex
	  Splice(1, peCurr, pe);
	  pe->SetVertex(peCurr->GetEdge()->GetVertex());
	  Splice(1, peCurr->GetEdge(), pe);
	  Splice(1, peCurr->GetEdge(), peCurr->GetEdge()->GetEdgePrev(1));

	  pvxCurr = peCurr->GetVertex();
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  m_pheapSelective->Remove(phe);
	  ASSERT( pvxCurr->GetHeapEntryReverse() == NULL );
	  DestroyVertex(pvxCurr);
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}      
      else
	{
	  // Change the edge-vertex's forwarding pointer to point at the neighbor
	  ASSERT( pe->GetEdgeNext(1)->GetEdgeNext(1) == pe->GetEdgePrev(1)->GetEdgePrev(1) );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdgeNext(1)->GetEdgeNext(1)->GetEdge());
	  if ( !( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		  pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() ) )
	    {
	      ASSERT( pe->GetVertex()->GetForwardVertex() == pe->GetEdge()->GetVertex() ||
		      pe->GetVertex()->GetForwardVertex() == peCurr->GetVertex() );
	    }
	  pe->GetVertex()->SetForwardVertex(peCurr->GetVertex());
	}
    }
  ASSERT( peStart != NULL );
  DestroyEdge(peStart->GetEdge());
  DestroyEdge(peStart);

  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()-1);
}

FLOAT QEMesh::CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimit(ErrorMetricType eErrorMetric, QEFacet *pfa)
{
  FLOAT fError;

  fError = 0.0;
  switch ( eErrorMetric )
    {
    case emtBestFitPlane:
    case emtLimitDistance:
      fError = CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimitMidpoint(pfa);
      break;
    default:
      ASSERT( FALSE );
      break;
    }

  return fError;
}

FLOAT QEMesh::CalculateFacetErrorSubdivideCornerRoundingFacetSharpSelectiveLimitMidpoint(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEEdge *pe;
  Point ptCenter;
  Vector pvCurr[3];
  Vector pvLimit[3];
  //Vector v;
  Vector vNormal, vNormalMax;
  UINT_32 i, uiVertices;
  FLOAT fError, fErrorMax, fMag, fMagMax;

  ptCenter.Null();
  iterQEFacet.Init(pfa);
  for ( uiVertices = 0, pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	uiVertices++, pe = iterQEFacet.PeekNextEdge() )
    {
      ASSERT( uiVertices < 3 );
      if ( pe->GetVertex()->GetGeneration() == pfa->GetGeneration() )
	{
	  pvCurr[uiVertices] = pe->GetVertex()->GetPoint();
	}
      else
	{
	  pvCurr[uiVertices] = pe->GetVertex()->GetTexture();
	}
      pvCurr[uiVertices][3] = 1.0;
      pvLimit[uiVertices] = pe->GetVertex()->GetLimitPosition();
      pvLimit[uiVertices][3] = 1.0;
      ptCenter += pvLimit[uiVertices];

      if ( pe->GetEdge()->GetFacet()->GetGeneration() > pfa->GetGeneration() )
	{
	  ASSERT( pe->GetEdge()->GetFacet()->GetGeneration() == pfa->GetGeneration() + 1 );
	  pe = iterQEFacet.PeekNextEdge();
	  ASSERT( pe != NULL );
	}
    }
  ASSERT( uiVertices == 3 );
  ptCenter /= (FLOAT)uiVertices;
  ptCenter[3] = 1.0;

  vNormalMax = (pvLimit[1] - pvLimit[0]) * (pvLimit[2] - pvLimit[0]);
  fMagMax = vNormalMax.Magnitude();

  vNormal = (pvLimit[2] - pvLimit[1]) * (pvLimit[0] - pvLimit[1]);
  fMag = vNormal.Magnitude();
  if ( fMag > fMagMax )
    {
      fMagMax = fMag;
      vNormalMax = vNormal;
    }

  vNormal = (pvLimit[0] - pvLimit[2]) * (pvLimit[1] - pvLimit[2]);
  fMag = vNormal.Magnitude();
  if ( fMag > fMagMax )
    {
      fMagMax = fMag;
      vNormalMax = vNormal;
    }

  if ( !vNormalMax.Normalize() )
    {
      vNormalMax.Set(0.0, 0.0, 1.0);
    }
  vNormalMax.CompletePlaneEquation(ptCenter);

  fErrorMax = 0.0;
  for ( i = 0; i < uiVertices; i++ )
    {
      fError = vNormalMax.Dot(pvCurr[i]);
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

Point QEMesh::CalculateVertexSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx)
{
  Point pt, ptCorner, ptCrease, ptSmooth;
  QEVertexIter iter;
  QEEdge *pe;
  UINT_32 i, uiNeighbors, uiSharpEdges, uiSemiSharpEdges;
  FLOAT fA, fAlpha;
  FLOAT fSharpness, pfSemiSharpnessesMax[3], fSwap, fT;
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
  QEEdge *peCurr;
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY

  iter.Init(pvx);

  // Sharp Loop Vertex-Vertex Subdivision
  uiSharpEdges = 0;
  uiSemiSharpEdges = 0;
  memset(pfSemiSharpnessesMax, 0, sizeof(pfSemiSharpnessesMax));
  ptSmooth.Null();
  ptCrease.Null();
  for ( uiNeighbors = 0, pe = iter.PeekFirstEdge();
	pe != NULL;
	uiNeighbors++, pe = iter.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  pt = pe->GetEdge()->GetVertex()->GetPoint();
	}
      else
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pt = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	}

      ptSmooth += pt;

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ptCrease += pt;
	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      uiSharpEdges++;
	    }
	  else
	    {
	      // Semi-sharp for this step
	      uiSemiSharpEdges++;
	      if ( fSharpness > pfSemiSharpnessesMax[2] )
		{
		  pfSemiSharpnessesMax[2] = fSharpness;
		  for ( i = 2; i > 0; i-- )
		    {
		      if ( pfSemiSharpnessesMax[i] > pfSemiSharpnessesMax[i-1] )
			{
			  fSwap = pfSemiSharpnessesMax[i];
			  pfSemiSharpnessesMax[i] = pfSemiSharpnessesMax[i-1];
			  pfSemiSharpnessesMax[i-1] = fSwap;
			}
		    }
		}
	    }
	}
    }

  //////////////////////////////
  // Corner rule
  ptCorner = pvx->GetPoint();

  if ( uiSharpEdges >= 3 )
    {
      return ptCorner;
    }

  //////////////////////////////
  // Crease rule
  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      ptCrease = ((6.0 * ptCorner) + ptCrease) / 8.0;
      ptCrease[3] = 1.0;
    }

  if ( uiSharpEdges == 2 && uiSemiSharpEdges == 0 )
    {
      return ptCrease;
    }

  //////////////////////////////
  // Smooth rule
  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fAlpha = ((FLOAT)uiNeighbors) * (1.0 - fA) / fA;
  ptSmooth = (ptSmooth + (fAlpha * ptCorner)) / (fAlpha + (FLOAT)uiNeighbors);
  pt[3] = 1.0;

  if ( uiSharpEdges + uiSemiSharpEdges <= 1 )
    {
      return ptSmooth;
    }

  ASSERT( uiSharpEdges < 3 );
  i = MIN(uiSemiSharpEdges, 3 - uiSharpEdges);
  fT = pfSemiSharpnessesMax[i-1];

  if ( uiSharpEdges + uiSemiSharpEdges == 2 )
    {
      pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
      pt[3] = 1.0;
      return pt;
    }

  ASSERT( uiSharpEdges + uiSemiSharpEdges >= 3 );
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCorner);
  pt[3] = 1.0;
  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx)
{
  /*
  Point pt, ptCorner, ptSmooth, ptCrease;
  QEEdge *pe, *peCurr, *ppeCrease[3], *ppe[3];
  UINT_32 uiNeighbors, uiCreases, uiIndex;
  FLOAT fA, fEpsilon;

  BOOL ppbCreasePointReady[3][3];
  */

  Point ptCorner, ptSmooth, ptCrease;
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr, *peCrease;
  FLOAT fA, fEpsilon, fSharpness;
  UINT_32 uiNeighbors, uiCreases;
  BOOL bEdgeMidpoint;

  Point ppptCrease[2][3];
  SharpVertexType psvtCrease[2];
  const FLOAT *pfCoeffTable;
  UINT_32 i, j, k, uiStart, uiEnd, uiCenter;

  const UINT_32 cuiNeighborsMax = 20;
  Point ppt[cuiNeighborsMax];
  SharpVertexType psvt[cuiNeighborsMax];
  BOOL pbCrease[cuiNeighborsMax];
  UINT_32 puiSegmentLength[cuiNeighborsMax];

  iterQEVertex.Init(pvx);

  // Search for a crease edge (if one exists it will be in the mesh)
  peCrease = NULL;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else
	{
	  ASSERT( FALSE );
	  uiNeighbors++;
	}

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ASSERT( fSharpness == INFINITY );
	  peCrease = pe;
	}

      if ( pe->GetVertex()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdgePrev(0)->GetVertex()->GetGeneration() &&
	   pe->GetEdge()->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetEdgeNext(1)->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() )

	{
	  // This might be an edge midpoint which is half formed
	  // These two points have no effect on the crease stuff
	  bEdgeMidpoint = TRUE;
	  iterQEFacet.Init(pe->GetFacet());
	  for ( peCurr = iterQEFacet.PeekFirstEdge();
		peCurr != NULL;
		peCurr = iterQEFacet.PeekNextEdge() )
	    {
	      if ( peCurr == pe )
		{
		  bEdgeMidpoint = FALSE;
		  break;
		}
	      if ( peCurr->GetEdge()->GetFacet()->GetGeneration() > peCurr->GetFacet()->GetGeneration() )
		{
		  peCurr = iterQEFacet.PeekNextEdge();
		  ASSERT( peCurr != NULL );		  
		}
	    }
	  if ( bEdgeMidpoint )
	    {
	      uiNeighbors += 2;
	    }
	}
    }
  ASSERT( uiNeighbors < cuiNeighborsMax );

  // Calculate all the neighbor vertices
  if ( peCrease == NULL )
    {
      peCrease = iterQEVertex.PeekFirstEdge();
    }
  uiCreases = 0;
  memset(pbCrease, 0, sizeof(pbCrease));
  memset(puiSegmentLength, 0, sizeof(puiSegmentLength));
  i = 0;
  pe = peCrease;
  do
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );

      fSharpness = pe->GetKnotInterval();
      ASSERT( fSharpness == 0.0 || fSharpness == INFINITY );

      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiNeighbors );
	  ppt[i] = pe->GetEdge()->GetVertex()->GetPoint();
	  psvt[i] = pe->GetEdge()->GetVertex()->GetSharpVertexType();
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiNeighbors );
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ppt[i] = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  psvt[i] = pe->GetEdge()->GetVertex()->GetSharpVertexType();
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else
	{
	  ASSERT( FALSE );
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  ASSERT( i < uiNeighbors );
	  ppt[i] = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe);
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	      psvt[i] = svtRegularCrease;
	    }
	  else
	    {
	      psvt[i] = svtSmooth;
	    }
	  i++;
	}

      if ( pe->GetVertex()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdgePrev(0)->GetVertex()->GetGeneration() &&
	   pe->GetEdge()->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetEdgeNext(1)->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() )
	
	{
	  // This might be an edge midpoint which is half formed
	  // These two points have no effect on the crease stuff
	  bEdgeMidpoint = TRUE;
	  iterQEFacet.Init(pe->GetFacet());
	  for ( peCurr = iterQEFacet.PeekFirstEdge();
		peCurr != NULL;
		peCurr = iterQEFacet.PeekNextEdge() )
	    {
	      if ( peCurr == pe )
		{
		  bEdgeMidpoint = FALSE;
		  break;
		}
	      if ( peCurr->GetEdge()->GetFacet()->GetGeneration() > peCurr->GetFacet()->GetGeneration() )
		{
		  peCurr = iterQEFacet.PeekNextEdge();
		  ASSERT( peCurr != NULL );		  
		}
	    }
	  if ( bEdgeMidpoint )
	    {
	      // These two uncreated edges are both smooth
	      ASSERT( i < uiNeighbors );
	      if ( pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetFacet()->GetGeneration() )
		{
		  ppt[i] = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe->GetEdgeNext(0));
		  fSharpness = pe->GetEdgeNext(0)->GetKnotInterval();
		  if ( fSharpness != 0.0 )
		    {
		      psvt[i] = svtRegularCrease;
		    }
		  else
		    {
		      psvt[i] = svtSmooth;
		    }
		}
	      else
		{
		  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() );
		  ppt[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
		  psvt[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetSharpVertexType();
		}
	      i++;

	      ASSERT( i < uiNeighbors );
	      if ( pe->GetEdgePrev(0)->GetEdgePrev(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetFacet()->GetGeneration() )
		{
		  ppt[i] = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe->GetEdgePrev(0)->GetEdgePrev(0));
		  fSharpness = pe->GetEdgePrev(0)->GetEdgePrev(0)->GetKnotInterval();
		  if ( fSharpness != 0.0 )
		    {
		      psvt[i] = svtRegularCrease;
		    }
		  else
		    {
		      psvt[i] = svtSmooth;
		    }
		}
	      else
		{
		  ASSERT( pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() );
		  ppt[i] = pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint();
		  psvt[i] = pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetSharpVertexType();
		}
	      i++;
	    }
	}

      pe = pe->GetEdgeNext(1);
    }
  while( pe != peCrease );

  //////////////////////////////////////////////////
  // We have all of the neighbors so calculate the limit position

  ptCorner = pvx->GetPoint();
  if ( uiCreases >= 3 )
    {
      return ptCorner;
    }

  if ( uiCreases == 2 )
    {
      ASSERT( ( pvx->GetSharpVertexType() == svtRegularCrease && uiNeighbors == 6 &&
		pbCrease[0] && pbCrease[3] ) ||
	      pvx->GetSharpVertexType() == svtNonregularCrease );

      for ( i = 1; i < uiNeighbors; i++ )
	{
	  if ( pbCrease[i] )
	    {
	      break;
	    }
	}
      ASSERT( i < uiNeighbors );

      psvtCrease[0] = psvt[0];
      psvtCrease[1] = psvt[i];

      ppptCrease[0][0] = ppt[uiNeighbors-1];
      ppptCrease[0][1] = ppt[0];
      ppptCrease[0][2] = ppt[1];

      ppptCrease[1][0] = ppt[i-1];
      ppptCrease[1][1] = ppt[i];
      ppptCrease[1][2] = ppt[(i+1 == uiNeighbors) ? (0) : (i+1)];

      uiCenter = scpuiSharpLoopSharpTypeIndex[pvx->GetSharpVertexType()];
      uiStart = scpuiSharpLoopSharpTypeIndex[psvtCrease[0]];
      uiEnd = scpuiSharpLoopSharpTypeIndex[psvtCrease[1]];
      pfCoeffTable = scppppfSharpLoopLimitTable[uiCenter][uiStart][uiEnd];

      i = 0;
      ptCrease = pfCoeffTable[i] * ptCorner;
#ifdef DEBUGGING_SHARP_LOOP_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	  fprintf(stderr, "  %f * ", pfCoeffTable[i]); 
	  ptCorner.Display();
	} 
#endif // DEBUGGING_SHARP_LOOP_MATH
      for ( j = 0; j < 2; j++ )
	{
	  for ( k = 0; k < 3; k++ )
	    {
	      i++;
	      ptCrease += pfCoeffTable[i] * ppptCrease[j][k];
#ifdef DEBUGGING_SHARP_LOOP_MATH
	      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
		   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )		   
		{
		  fprintf(stderr, "  %f * ", pfCoeffTable[i]);
		  ppptCrease[j][k].Display();
		} 
#endif // DEBUGGING_SHARP_LOOP_MATH
	    }
	}
      ptCrease[3] = 1.0;
#ifdef DEBUGGING_SHARP_LOOP_MATH
      if ( (psvtCrease[0] == svtDart && psvtCrease[1] == svtCorner) ||
	   (psvtCrease[1] == svtDart && psvtCrease[0] == svtCorner) )
	{
	  fprintf(stderr, "  Limit = "); 
	  ptCorner.Display();
	  fprintf(stderr, "START: Computing limit position: %p (%ld %ld)\n", 
		  pvx, psvtCrease[0], psvtCrease[1]);
	} 
#endif // DEBUGGING_SHARP_LOOP_MATH

      return ptCrease;
    }

  ptSmooth.Null();
  for ( i = 0; i < uiNeighbors; i++ )
    {
      ptSmooth += ppt[i];
    }

  fA = (3.0 + 2.0*cos(2.0*PI/(FLOAT)uiNeighbors));
  fA = fA * fA;
  fA = (5.0 / 8.0) - (fA / 64.0);
  fEpsilon = (3.0 * (FLOAT)uiNeighbors) / (8.0 * fA);
  ptSmooth = (ptSmooth + (fEpsilon * ptCorner)) / (fEpsilon + (FLOAT)uiNeighbors);
  ptSmooth[3] = 1.0;

  return ptSmooth;
}

Vector QEMesh::CalculateVertexNormalSubdivisionCornerRoundingFacetSharpSelective(QEVertex *pvx)
{
  Vector vT0, vT1, vNormal, vNormalMax;
  //Vector v;
  Vector vCenter;
  QEVertexIter iterQEVertex;
  QEFacetIter iterQEFacet;
  QEEdge *pe, *peCurr, *peCrease;
  UINT_32 uiNeighbors, uiCreases, i, j, k, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  FLOAT fSharpness, fMag, fMagMax;
  BOOL bRegularCrease, bEdgeMidpoint;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];
  BOOL pbCrease[cuiNeighborsMax];
  UINT_32 puiSegmentLength[cuiNeighborsMax];

  iterQEVertex.Init(pvx);

  // Search for a crease edge (if one exists it will be in the mesh)
  peCrease = NULL;
  uiNeighbors = 0;
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiNeighbors++;
	}
      else
	{
	  ASSERT( FALSE );
	  uiNeighbors++;
	}

      fSharpness = pe->GetKnotInterval();
      if ( fSharpness != 0.0 )
	{
	  ASSERT( fSharpness == INFINITY );
	  peCrease = pe;
	}

      if ( pe->GetVertex()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdgePrev(0)->GetVertex()->GetGeneration() &&
	   pe->GetEdge()->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetEdgeNext(1)->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() )

	{
	  // This might be an edge midpoint which is half formed
	  // These two points have no effect on the crease stuff
	  bEdgeMidpoint = TRUE;
	  iterQEFacet.Init(pe->GetFacet());
	  for ( peCurr = iterQEFacet.PeekFirstEdge();
		peCurr != NULL;
		peCurr = iterQEFacet.PeekNextEdge() )
	    {
	      if ( peCurr == pe )
		{
		  bEdgeMidpoint = FALSE;
		  break;
		}
	      if ( peCurr->GetEdge()->GetFacet()->GetGeneration() > peCurr->GetFacet()->GetGeneration() )
		{
		  peCurr = iterQEFacet.PeekNextEdge();
		  ASSERT( peCurr != NULL );		  
		}
	    }
	  if ( bEdgeMidpoint )
	    {
	      uiNeighbors += 2;
	    }
	}
    }
  ASSERT( uiNeighbors < cuiNeighborsMax );

  // Calculate all the neighbor vertices
  if ( peCrease == NULL )
    {
      peCrease = iterQEVertex.PeekFirstEdge();
    }
  uiCreases = 0;
  memset(pbCrease, 0, sizeof(pbCrease));
  memset(puiSegmentLength, 0, sizeof(puiSegmentLength));
  i = 0;
  pe = peCrease;
  do
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );

      fSharpness = pe->GetKnotInterval();
      ASSERT( fSharpness == 0.0 || fSharpness == INFINITY );

      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiNeighbors );
	  pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiNeighbors );
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pv[i] = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}
      else
	{
	  ASSERT( FALSE );
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  ASSERT( i < uiNeighbors );
	  pv[i] = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe);
	  if ( fSharpness != 0.0 )
	    {
	      uiCreases++;
	      pbCrease[i] = TRUE;
	    }
	  i++;
	}

      if ( pe->GetVertex()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() &&
	   pe->GetVertex()->GetGeneration() == pe->GetEdgePrev(0)->GetVertex()->GetGeneration() &&
	   pe->GetEdge()->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() &&
	   pe->GetEdgeNext(1)->GetFacet()->GetGeneration() > pe->GetFacet()->GetGeneration() )
	
	{
	  // This might be an edge midpoint which is half formed
	  // These two points have no effect on the crease stuff
	  bEdgeMidpoint = TRUE;
	  iterQEFacet.Init(pe->GetFacet());
	  for ( peCurr = iterQEFacet.PeekFirstEdge();
		peCurr != NULL;
		peCurr = iterQEFacet.PeekNextEdge() )
	    {
	      if ( peCurr == pe )
		{
		  bEdgeMidpoint = FALSE;
		  break;
		}
	      if ( peCurr->GetEdge()->GetFacet()->GetGeneration() > peCurr->GetFacet()->GetGeneration() )
		{
		  peCurr = iterQEFacet.PeekNextEdge();
		  ASSERT( peCurr != NULL );		  
		}
	    }
	  if ( bEdgeMidpoint )
	    {
	      // These two uncreated edges are both smooth
	      ASSERT( i < uiNeighbors );
	      if ( pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetFacet()->GetGeneration() )
		{
		  pv[i] = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe->GetEdgeNext(0));
		}
	      else
		{
		  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() );
		  pv[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
		}
	      i++;

	      ASSERT( i < uiNeighbors );
	      if ( pe->GetEdgePrev(0)->GetEdgePrev(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetFacet()->GetGeneration() )
		{
		  pv[i] = CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(pe->GetEdgePrev(0)->GetEdgePrev(0));
		}
	      else
		{
		  ASSERT( pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() );
		  pv[i] = pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex()->GetPoint();
		}
	      i++;
	    }
	}

      pe = pe->GetEdgeNext(1);
    }
  while( pe != peCrease );

  for ( i = 0; i < uiNeighbors; i++ )
    {      
      pv[i][3] = 1.0;
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;
  if ( uiCreases < 2 )
    {
      vT0.Null();
      vT1.Null();
      fTheta = 0.0;
      fThetaDelta = 2.0 * PI / (FLOAT)uiNeighbors;
      for ( i = 0; i < uiNeighbors; i++ )
	{
	  vT1 += cos(fTheta) * pv[i];
	  fTheta += fThetaDelta;
	  vT0 += cos(fTheta) * pv[i];
	}
      vT0[3] = 0.0;
      vT1[3] = 0.0;
      vNormal = vT0 * vT1;
    }
  else
    {
      // Copy the first edge to the end
      pv[uiNeighbors] = pv[0]; 
      pbCrease[uiNeighbors] = pbCrease[0];
      j = 0;
      for ( i = 0; i < uiNeighbors; i++ )
	{      
	  if ( pbCrease[i] )
	    {
	      j++;
	    }
	  ASSERT( j > 0 && j <= uiCreases );
	  puiSegmentLength[j-1]++;
	}

      bRegularCrease = ( uiNeighbors == 6 && uiCreases == 2 &&
			 puiSegmentLength[0] == 3 && puiSegmentLength[1] == 3 ); 

      // Calculate the error for the individual segments and take the max
      vNormalMax.Null();
      fMagMax = 0.0;
      k = 0;
      for ( i = 0; i < uiCreases; k += puiSegmentLength[i], i++ )
	{
	  vT0.Null();
	  vT1.Null();
	  if ( bRegularCrease )
	    {
	      vT0 = pv[k];
	      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
	      for ( j = 0; j <= puiSegmentLength[i]; j++ )
		{
		  vT1 += scppfSharpLoopTangentTable[2][j+1] * pv[k+j];
		}
	      vT0 -= pv[k+puiSegmentLength[i]];	      
	    }
	  else
	    {
	      switch ( puiSegmentLength[i] )
		{
		case 0:
		  ASSERT( FALSE );
		  break;
		case 1:
		case 2:
		  uiIndex = puiSegmentLength[i] - 1;
		  vT0 = pv[k];
		  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
		  for ( j = 0; j <= puiSegmentLength[i]; j++ )
		    {
		      vT1 += scppfSharpLoopTangentTable[uiIndex][j+1] * pv[k+j];
		    }
		  vT0 -= pv[k+puiSegmentLength[i]];
		  break;
		default:
		  fThetaDelta = PI / (FLOAT)puiSegmentLength[i];
		  fA = sin(fThetaDelta);
		  vT1 = pv[k];
		  vT0 = (fA * pv[k]);
		  vT1 -= pv[k+puiSegmentLength[i]];
		  vT0 += (fA * pv[k+puiSegmentLength[i]]);
		  fA = 2.0 * (cos(fThetaDelta) - 1.0);
		  for ( j = 1, fTheta = fThetaDelta; 
			j < puiSegmentLength[i];
			j++, fTheta += fThetaDelta )
		    {
		      vT0 += ((fA * sin(fTheta)) * pv[k+j]);
		    }
		  break;
		}
	    }
	  vT0[3] = 0.0;
	  vT1[3] = 0.0;
	  vNormal = vT0 * vT1;
	  fMag = vNormal.Magnitude();
	  if ( fMag > fMagMax )
	    {
	      fMagMax = fMag;
	      vNormalMax = vNormal;
	    }
	}
      vNormal = vNormalMax;
    }

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

/*
Vector QEMesh::CalculateVertexNormalSubdivisionCornerRoundingFacetSharpSelective(QEEdge *peCrease)
{
  Vector vT0, vT1, vNormal;
  Vector vCenter;
  QEVertex *pvx;
  QEEdge *pe;
  UINT_32 uiSegmentLength, i, uiIndex;
  FLOAT fTheta, fThetaDelta, fA;
  FLOAT fSharpness;
  BOOL bRegularCrease;

  const UINT_32 cuiNeighborsMax = 20;
  Vector pv[cuiNeighborsMax];

  pvx = peCrease->GetVertex();
  ASSERT( pvx->GetSharpVertexType() == svtRegularCrease ||
	  pvx->GetSharpVertexType() == svtNonregularCrease ||
	  pvx->GetSharpVertexType() == svtCorner );
  ASSERT( peCrease->GetKnotInterval() == INFINITY );

  // Search for a crease edge (if one exists it will be in the mesh)
  pe = peCrease;
  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
  if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      uiSegmentLength = 1;
    }
  else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
    {
      uiSegmentLength = 1;
    }
  else
    {
      uiSegmentLength = 2;
      // This can add at most one crease edge
    }
  for ( pe = pe->GetEdgeNext(1);
	pe != peCrease;
	pe = pe->GetEdgeNext(1) )
    {
      fSharpness = pe->GetKnotInterval();

      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  uiSegmentLength++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
	  uiSegmentLength++;
	}
      else
	{
	  // This can add at most one crease edge
	  if ( fSharpness != 0.0 )
	    {
	      ASSERT( fSharpness == INFINITY );
	      // This is the ending crease edge
	      uiSegmentLength += 2;
	    }
	  else
	    {
	      uiSegmentLength += 3;
	    }
	}

      if ( fSharpness != 0.0 )
	{
	  ASSERT( fSharpness == INFINITY );
	  break;
	}
    }
  ASSERT( uiSegmentLength < cuiNeighborsMax );

  pe = peCrease;
  for ( i = 0; i < uiSegmentLength; pe = pe->GetEdgeNext(1) )
    {
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 >= pe->GetVertex()->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
	{
	  ASSERT( i < uiSegmentLength );
	  pv[i] = pe->GetEdge()->GetVertex()->GetPoint();
	  pv[i][3] = 1.0;
	  i++;
	}
      else if ( pe->GetEdge()->GetVertex()->GetGeneration() > pe->GetVertex()->GetGeneration() )
	{
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( i < uiSegmentLength );
	  pv[i] = pe->GetEdge()->GetVertex()->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  ASSERT( pe->GetEdge()->GetVertex()->GetForwardVertex() != NULL );
	  peCurr = FindSubdivideCornerRoundingSelectiveCenterVertex(pe->GetEdge());
	  ASSERT( peCurr->GetVertex() == pe->GetEdge()->GetVertex()->GetForwardVertex() );
	  ASSERT( i < uiSegmentLength );
	  pv[i] = CalculateVertexSubdivisionCornerRoundingSelectiveGeneration(peCurr->GetVertex(), pe->GetVertex()->GetGeneration());
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
	  pv[i][3] = 1.0;
	  i++;
	}
      else
	{
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );

	  // This a smooth edge
	  if ( i != 0 )
	    {
	      if ( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
		{
		  ASSERT( i < uiSegmentLength );
		  pv[i] = pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetPoint();
		}
	      else
		{
		  ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
		  ASSERT( i < uiSegmentLength );
		  pv[i] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetEdgeNext(1));
		}
	      i++;
	    }

	  ASSERT( i < uiSegmentLength );
	  pv[i] = CalculateVertexHelperSubdivisionCornerRoundingSharpSelective(pe->GetEdge()->GetVertex());
	  i++;

	  // This a smooth edge
	  if ( i < uiSegmentLength )
	    {
	      if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pe->GetVertex()->GetGeneration() )
		{
		  ASSERT( i < uiSegmentLength );
		  pv[i] = pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
		}
	      else
		{
		  ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
		  ASSERT( i < uiSegmentLength );
		  pv[i] = CalculateEdgeSubdivisionCornerRoundingSharpSelective(pe->GetEdgeNext(0));
		}
	      i++;
	    }
	}
    }
  
  vCenter = pvx->GetPoint();
  vCenter[3] = 1.0;

  bRegularCrease = (pvx->GetSharpVertexType() == svtRegularCrease);
  
  // Calculate the error for the individual segments and take the max
  vT0.Null();
  vT1.Null();
  if ( bRegularCrease )
    {
      vT0 = pv[0];
      vT1 = scppfSharpLoopTangentTable[2][0] * vCenter;
      for ( i = 0; i < uiSegmentLength; i++ )
	{
	  vT1 += scppfSharpLoopTangentTable[2][i+1] * pv[i];
	}
      vT0 -= pv[uiSegmentLength - 1];	      
    }
  else
    {
      switch ( uiSegmentLength )
	{
	case 0:
	case 1:
	  ASSERT( FALSE );
	  break;
	case 2:
	case 3:
	  uiIndex = uiSegmentLength - 2;
	  vT0 = pv[0];
	  vT1 = scppfSharpLoopTangentTable[uiIndex][0] * vCenter;
	  for ( i = 0; i < uiSegmentLength; i++ )
	    {
	      vT1 += scppfSharpLoopTangentTable[uiIndex][i+1] * pv[i];
	    }
	  vT0 -= pv[uiSegmentLength - 1];
	  break;
	default:
	  fThetaDelta = PI / (FLOAT)(uiSegmentLength - 1);
	  fA = sin(fThetaDelta);
	  vT1 = pv[0];
	  vT0 = (fA * pv[0]);
	  vT1 -= pv[uiSegmentLength - 1];
	  vT0 += (fA * pv[uiSegmentLength - 1]);
	  fA = 2.0 * (cos(fThetaDelta) - 1.0);
	  for ( i = 1, fTheta = fThetaDelta; 
		i < uiSegmentLength - 1;
		i++, fTheta += fThetaDelta )
	    {
	      vT0 += ((fA * sin(fTheta)) * pv[i+1]);
	    }
	  break;
	}
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;
  vNormal = vT0 * vT1;

  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}
*/
Point QEMesh::CalculateEdgeSubdivisionCornerRoundingFacetSharpSelective(QEEdge *pe)
{
  QEVertex *pvxCurr;
  QEEdge *ppe[2], *peCurr;
  BOOL pbCrease[2], pbRegularCrease[2], bCrease;
  Point pt, ptSmooth, ptCrease;
  Point ppt[6];
  UINT_32 i, uiN, uiSharpEdges;
  FLOAT fSharpness, fT;

  ASSERT( pe->GetFacet()->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() );

  // Calculate Edge Endpoints
  if ( pe->GetFacet()->GetGeneration() == pe->GetVertex()->GetGeneration() )
    {
      ppt[0] = pe->GetVertex()->GetPoint();
    }
  else
    {
      ASSERT( pe->GetFacet()->GetGeneration() + 1 == pe->GetVertex()->GetGeneration() );
      ppt[0] = pe->GetVertex()->GetTexture();
    }
  if ( pe->GetFacet()->GetGeneration() == pe->GetEdge()->GetVertex()->GetGeneration() )
    {
      ppt[1] = pe->GetEdge()->GetVertex()->GetPoint();
    }
  else
    {
      ASSERT( pe->GetFacet()->GetGeneration() + 1 == pe->GetEdge()->GetVertex()->GetGeneration() );
      ppt[1] = pe->GetEdge()->GetVertex()->GetTexture();
    }

  fSharpness = pe->GetKnotInterval();
  bCrease = FALSE;
  if ( fSharpness != 0.0 )
    {
      ppe[0] = pe;
      ppe[1] = pe->GetEdge();
      pbCrease[0] = FALSE;
      pbCrease[1] = FALSE;
      pbRegularCrease[0] = FALSE;
      pbRegularCrease[1] = FALSE;
      for ( i = 0; i < 2; i++ )
	{
	  uiSharpEdges = 1;
	  uiN = 1;
	  for ( peCurr = ppe[i]->GetEdgeNext(1); 
		peCurr != ppe[i]; 
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      uiN++;
	      if ( peCurr->GetKnotInterval() != 0.0 )
		{
		  uiSharpEdges++;
		}
	    }

	  if ( uiSharpEdges < 2 )
	    {
	      break;
	    }

	  pbCrease[i] = TRUE;
	  if ( uiSharpEdges == 2 && uiN == 6 &&
	       ppe[i]->GetEdgeNext(1)->GetEdgeNext(1)->GetEdgeNext(1)->GetKnotInterval() != 0.0 )
	    {
	      pbRegularCrease[i] = TRUE;
	    }	  
	}

      bCrease = pbCrease[0] && pbCrease[1];
      if ( bCrease )
	{
	  if ( ( pbRegularCrease[0] && pbRegularCrease[1] ) ||
	       ( !pbRegularCrease[0] && !pbRegularCrease[1] ) )
	    {
	      ptCrease = ( ppt[0] + ppt[1] ) / 2.0;
	      ptCrease[3] = 1.0;
	    }
	  else
	    {
	      if ( pbRegularCrease[0] )
		{
		  ptCrease = (( 5.0 * ppt[0]) + (3.0 * ppt[1]) ) / 8.0;
		}
	      else
		{
		  ptCrease = (( 5.0 * ppt[1]) + (3.0 * ppt[0]) ) / 8.0;
		}
	      ptCrease[3] = 1.0;	      
	    }

	  if ( fSharpness == INFINITY || fSharpness >= 1.0 )
	    {
	      // Infinitely sharp for this step
	      return ptCrease;
	    }
	}
    }

  // Midpoint smooth rule

  if ( pe->GetEdgePrev(0)->GetEdge()->GetFacet()->GetGeneration() <= pe->GetFacet()->GetGeneration() )
    {
      pvxCurr = pe->GetEdgePrev(0)->GetVertex();
    }
  else
    {
      ASSERT( pe->GetEdgePrev(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetFacet()->GetGeneration() + 1 );
      pvxCurr = pe->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex();
    }
  if ( pvxCurr->GetGeneration() == pe->GetFacet()->GetGeneration() )
    {
      ppt[2] = pvxCurr->GetPoint();
    }
  else
    {
      ASSERT( pvxCurr->GetGeneration() == pe->GetFacet()->GetGeneration() + 1 );
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[2] = pvxCurr->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() <= pe->GetFacet()->GetGeneration() )
    {
      pvxCurr = pe->GetEdgeNext(0)->GetEdge()->GetVertex();
    }
  else
    {
      ASSERT( pe->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetFacet()->GetGeneration() + 1 );
      pvxCurr = pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex();
    }
  if ( pvxCurr->GetGeneration() == pe->GetFacet()->GetGeneration() )
    {
      ppt[3] = pvxCurr->GetPoint();
    }
  else
    {
      ASSERT( pvxCurr->GetGeneration() == pe->GetFacet()->GetGeneration() + 1 );
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[3] = pvxCurr->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdge()->GetEdgePrev(0)->GetEdge()->GetFacet()->GetGeneration() <= pe->GetEdge()->GetFacet()->GetGeneration() )
    {
      pvxCurr = pe->GetEdge()->GetEdgePrev(0)->GetVertex();
    }
  else
    {
      ASSERT( pe->GetEdge()->GetEdgePrev(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() + 1 );
      pvxCurr = pe->GetEdge()->GetEdgePrev(0)->GetEdgePrev(0)->GetVertex();
    }
  if ( pvxCurr->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() )
    {
      ppt[4] = pvxCurr->GetPoint();
    }
  else
    {
      ASSERT( pvxCurr->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() + 1 );
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[4] = pvxCurr->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  if ( pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() <= pe->GetEdge()->GetFacet()->GetGeneration() )
    {
      pvxCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetVertex();
    }
  else
    {
      ASSERT( pe->GetEdge()->GetEdgeNext(0)->GetEdge()->GetFacet()->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() + 1 );
      pvxCurr = pe->GetEdge()->GetEdgeNext(0)->GetEdgeNext(0)->GetEdge()->GetVertex();
    }
  if ( pvxCurr->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() )
    {
      ppt[5] = pvxCurr->GetPoint();
    }
  else
    {
      ASSERT( pvxCurr->GetGeneration() == pe->GetEdge()->GetFacet()->GetGeneration() + 1 );
#ifdef SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ppt[5] = pvxCurr->GetTexture();
#else // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
      ASSERT( FALSE );
#endif // SELECTIVE_SHARP_LOOP_FORWARD_ONLY
    }

  ptSmooth = ((6.0 * ppt[0]) + (6.0 * ppt[1]) + ppt[2] + ppt[3] + ppt[4] + ppt[5]) / 16.0;
  ptSmooth[3] = 1.0;
  
  if ( !bCrease )
    {
      return ptSmooth;
    } 

  fT = fSharpness;
  pt = ((1.0 - fT) * ptSmooth) + (fT * ptCrease);
  pt[3] = 1.0;
  return pt;
}

#endif // LOOP_FACET_CENTRIC

//////////////////////////////////////////////////////////////////////
// Selective Subdivision Doo-Sabin

QEMesh *QEMesh::SubdivideDooSabinSelective(QEMesh::ErrorMetricType eErrorMetric,
					   FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertex *pvx;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;
  Vector v;

  v.Null();

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideDooSabinSelective(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);

      pvx->SetGeneration(0);
      pvx->SetTexture(v);
    }

  pme->Preprocess();

  pme->ReSubdivideDooSabinSelective(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::ReSubdivideDooSabinSelective(QEMesh::ErrorMetricType /*eErrorMetric*/,
					  FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  //QEVertexIter iterQEVertex;
  QEEdge *pe,  *peCurr;
  QEEdge *peStart;
  FLOAT fError;
  //Vector v;

  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      // Unnecessarily Strict Rule that all vertices around the facets which are around
      // the central vertex must be the same generation or younger
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  pvxCurr = NULL;
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideDooSabinSelective(pvx);

      m_plistSelective->InsertFirst(pvx, 0);

      peStart = pvx->GetEdge();

      // Remove the current vertex
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Remove(phe);
      pvx->SetHeapEntry(NULL);
      //DestroyVertex(pvx);

      // Update the priority queue
      pe = peStart;
      do
	{
	  for ( peCurr = pe->GetEdgePrev(1);
		peCurr != pe->GetEdge();
		peCurr = peCurr->GetEdgeNext(0) )
	    {
	      // Insert the new vertices
	      pvxCurr = peCurr->GetVertex();
	      if ( pvxCurr->GetHeapEntry() == NULL )
		{   
		  fError = CalculateVertexErrorSubdivideDooSabinSelective(pvxCurr);
		  phe = m_pheapSelective->Insert(fError, pvxCurr);
		  pvxCurr->SetHeapEntry(phe);
		}
	      else
		{
		  phe = pvxCurr->GetHeapEntry();
		  ASSERT( phe != NULL );
		  fError = CalculateVertexErrorSubdivideDooSabinSelective(pvxCurr);
		  m_pheapSelective->Update(phe, fError);
		}
	    }
	  pe = pe->GetEdgeNext(0);
	}
      while ( pe != peStart );

      // Update the face normals
      pe = peStart;
      pe->GetFacet()->Preprocess();
      do
	{
	  pe->GetEdgePrev(1)->GetFacet()->Preprocess();
	  pe->GetEdgePrev(1)->GetEdgePrev(1)->GetFacet()->Preprocess();

	  pe = pe->GetEdgeNext(0);
	}
      while ( pe != peStart );
    }
}

VOID QEMesh::UnSubdivideDooSabinSelective(QEMesh::ErrorMetricType /*eErrorMetric*/,
					  FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex;
  QEVertex *pvx, *pvxCurr;
  QEEdge *pe;
  FLOAT fError;

  //fprintf(stderr, "START: UnSubdivideDooSabinSelective\n");
  //DisplayEuler();

  for ( pvx = (QEVertex*)(m_plistSelective->PeekFirst());
	pvx != NULL && 
	  !( //pvx->GetHeapEntry()->GetKey() > fTolerance && 
	     m_plistQEFacet->GetLength() < uiFacetsMax );
	pvx = (QEVertex*)(m_plistSelective->PeekFirst()) )
    {
      m_plistSelective->RemoveFirst();

      UnSubdivideDooSabinSelective(pvx);

      // Reinsert the current vertex
      ASSERT( pvx->GetHeapEntry() == NULL );
      fError = CalculateVertexErrorSubdivideDooSabinSelective(pvx);
      phe = m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);

      // Update the priority queue
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  // Insert the new vertices
	  pvxCurr = pe->GetVertex();
	  phe = pvxCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  fError = CalculateVertexErrorSubdivideDooSabinSelective(pvxCurr);
	  m_pheapSelective->Update(phe, fError);
	}

      // Update the face normals
      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{
	  pe->GetFacet()->Preprocess();
	}
    }
}

///////////////////////////////////////////////////
// QEMesh::SubdivideDooSabinSelective            //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     v4--e3--v3--e2--v2    //
//           |              |       |       |    //
//           |             e4      e9      e1    //
//           |              |       |       |    //
//  e2------vV------e0  => v5--e10-vV--e8--v1    //
//           |              |       |       |    //
//           |             e5      e11     e0    //
//           |              |       |       |    //
//  f2      e3      f3     v6--e6--v7--e7--v0    //
//                                               //
VOID QEMesh::SubdivideDooSabinSelective(QEVertex *pvx)
{
  //QEVertexIter iterQEVertex;
  QEVertex *pvxNew;
  QEEdge *pe, *peNew, *peStart;
  QEFacet *pfaNew;
  //Point ptNew;
  Point pt;
  Vector vCenter;

  vCenter = pvx->GetPoint();

  peStart = pvx->GetEdge();
  pe = peStart;
  do
    {
      // Double edges connecting vertices of the same generation
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );
	  pfaNew = CreateFacet();
	  ASSERT( pfaNew != NULL );

	  peNew->SetVertex(pe->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());

	  peNew->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pfaNew);
	  pe->SetFacet(pfaNew);

	  Splice(1, pe, peNew);
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	}

      pe = pe->GetEdgeNext(1)->GetEdgeNext(1);
    }
  while ( pe != peStart );

  pfaNew = CreateFacet();
  ASSERT( pfaNew != NULL );
  pvx->SetForwardFacet(pfaNew);
  pfaNew->SetForwardVertex(pvx);

  pe = peStart;
  do
    {
      // Create ring edges and vertices

      peNew = CreateEdge();
      ASSERT( peNew != NULL );

      pt = CalculateEdgeSubdivisionDooSabinSelective(pe->GetEdgeNext(1));
      pvxNew = CreateVertex(pt);
      ASSERT( pvxNew != NULL );
      pvxNew->SetGeneration(pvx->GetGeneration()+1);
      pvxNew->SetTexture(vCenter);

      peNew->SetVertex(pvxNew);
      peNew->SetFacet(pe->GetFacet());
      peNew->GetEdge()->SetFacet(pfaNew);

      pe->SetForwardEdge(peNew->GetEdge());
      pe->GetEdgeNext(1)->SetForwardEdge(peNew);

      pe = pe->GetEdgeNext(1)->GetEdgeNext(1);
    }
  while ( pe != peStart );

  for ( pe = peStart->GetEdgeNext(1);
	peStart->GetEdgeNext(1) != peStart->GetEdgePrev(1);
	pe = peStart->GetEdgeNext(1) )
    {
      // Unhook the vertex and hook up the facet contour

      Splice(1, pe->GetEdgePrev(1), pe->GetEdgeNext(1));

      pe->SetVertex(pe->GetForwardEdge()->GetVertex());
      pe->GetEdgeNext(1)->SetVertex(pe->GetForwardEdge()->GetVertex());
      pe->GetEdgeNext(1)->GetForwardEdge()->SetVertex(pe->GetForwardEdge()->GetVertex());

      Splice(1, pe->GetForwardEdge(), pe->GetEdgeNext(1)->GetForwardEdge());
      Splice(1, pe->GetForwardEdge(), pe->GetEdgePrev(1));

      pe->SetForwardEdge(NULL);
      pe->GetEdgeNext(1)->SetForwardEdge(NULL);
    }

  // Make the last connection
  pe = peStart->GetEdgeNext(1);

  pe->SetVertex(pe->GetForwardEdge()->GetVertex());
  pe->GetEdgeNext(1)->SetVertex(pe->GetForwardEdge()->GetVertex());
  pe->GetEdgeNext(1)->GetForwardEdge()->SetVertex(pe->GetForwardEdge()->GetVertex());
  
  Splice(1, pe->GetForwardEdge(), pe->GetEdgeNext(1)->GetForwardEdge());
  Splice(1, pe->GetForwardEdge(), pe->GetEdgePrev(1));
  
  pe->SetForwardEdge(NULL);
  pe->GetEdgeNext(1)->SetForwardEdge(NULL);

  // Make the defunct vertex point to an edge around the new facet
  pvx->SetEdge(peStart->GetEdgeNext(1));
}

///////////////////////////////////////////////////
// QEMesh::UnSubdivideDooSabinSelective          //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     v4--e3--v3--e2--v2    //
//           |              |       |       |    //
//           |             e4      e9      e1    //
//           |              |       |       |    //
//  e2------vV------e0  => v5--e10-vV--e8--v1    //
//           |              |       |       |    //
//           |             e5      e11     e0    //
//           |              |       |       |    //
//  f2      e3      f3     v6--e6--v7--e7--v0    //
//                                               //
VOID QEMesh::UnSubdivideDooSabinSelective(QEVertex *pvx)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEFacetIter iterQEFacet;
  QEFacet *pfa;
  QEEdge *pe, *peStart, *peCurr, *peNext, *pePrev;
  QEVertex *pvxCurr;

  pfa = pvx->GetForwardFacet();
  ASSERT( pfa != NULL );
  pvx->SetForwardFacet(NULL);
  ASSERT( pfa->GetForwardVertex() == pvx );
  pfa->SetForwardVertex(NULL);
  iterQEFacet.Init(pfa);
  for ( pvxCurr = iterQEFacet.PeekFirstVertex();
	pvxCurr != NULL;
	pvxCurr = iterQEFacet.PeekNextVertex() )
    {
      ASSERT( pvxCurr->GetGeneration() == pvx->GetGeneration() + 1 );
    }

  //////////////////////////////////////////////////
  // Disconnect the facet and reconnect the vertex
  //
  pvx->SetEdge(NULL);
  pePrev = NULL;

  iterQEFacet.Init(pfa);
  for ( pe = iterQEFacet.PeekFirstEdge();
	pe != NULL;
	pe = iterQEFacet.PeekNextEdge() )
    {
      pe = pe->GetEdge();

      if ( pe->GetEdgeNext(0)->GetEdgeNext(0)->GetEdgeNext(0) == pe )
	{
	  // Collapse and destroy the edge-facet

	  DestroyFacet(pe->GetFacet());
	  peCurr = pe->GetEdgePrev(0);
	  pe->GetEdgeNext(0)->SetFacet(peCurr->GetEdge()->GetFacet());
	  pe->GetEdgeNext(0)->GetEdge()->SetVertex(pe->GetEdgeNext(0)->GetEdge()->GetVertex());

	  Splice(1, peCurr->GetEdgePrev(1), peCurr);
	  Splice(1, peCurr->GetEdge()->GetEdgePrev(1), peCurr->GetEdge());
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);

	  peCurr = pe->GetEdgeNext(0);
	  Splice(1, peCurr->GetEdgePrev(1), peCurr);

	  if ( pePrev != NULL )
	    {
	      Splice(1, peCurr, pePrev);
	    }
	  peCurr->SetVertex(pvx);
	  pePrev = peCurr;
	}
      else
	{
	  // Collapse one side of the edge-facet

	  peCurr = pe->GetEdgeNext(0);
	  peCurr->SetFacet(peCurr->GetFacet());
	  Splice(1, peCurr->GetEdgePrev(1), peCurr);

	  if ( pePrev != NULL )
	    {
	      Splice(1, peCurr, pePrev);
	    }
	  peCurr->SetVertex(pvx);
	  pePrev = peCurr;

	  peCurr = pe->GetEdgeNext(1);
	  Splice(1, peCurr->GetEdgePrev(1), peCurr);

	  ASSERT( pePrev != NULL );
	  Splice(1, peCurr, pePrev);
	  peCurr->SetVertex(pvx);
	  pePrev = peCurr;
	}
    }

  //////////////////////////////////////////////////
  // Destroy the facet and its borders
  //
  peStart = pfa->GetEdge();
  ASSERT( peStart != NULL );
  peCurr = peStart;
  do
    {
      peNext = peCurr->GetEdgeNext(0);

      pvxCurr = peNext->GetVertex();
      phe = pvxCurr->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Remove(phe);
      DestroyVertex(pvxCurr);
      
      Splice(1, peNext, peNext->GetEdgePrev(1));

      if ( peCurr != peStart )
	{
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);
	}

      peCurr = peNext;
    }
  while ( peCurr != peStart );

  DestroyEdge(peStart->GetEdge());
  DestroyEdge(peStart);

  DestroyFacet(pfa);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideDooSabinSelective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  ASSERT( uiN > 1 );
  if ( uiN < 3 )
    {
      //fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return -1.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

//////////////////////////////////////////////////////////////////////
// Selective Subdivision Doo-Sabin Limit Surface

QEMesh *QEMesh::SubdivideDooSabinSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
						FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertex *pvx;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;
  Vector v;

  v.Null();

  pme = Copy();
  ASSERT( pme != NULL );
  pme->CalculateSubdivisionLimitPositionsAndNormalsDooSabin();

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideDooSabinSelective(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);

      pvx->SetGeneration(0);
      pvx->SetTexture(v);
    }

  pme->Preprocess();

  pme->ReSubdivideDooSabinSelectiveLimit(eErrorMetric, fTolerance, uiFacetsMax);

  return pme;
}

VOID QEMesh::ReSubdivideDooSabinSelectiveLimit(QEMesh::ErrorMetricType eErrorMetric,
					       FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  //QEVertexIter iterQEVertex;
  QEEdge *pe,  *peCurr;
  QEEdge *peStart;
  FLOAT fError;
  //Vector v;
  UINT_32 uiN;

  eErrorMetric = emtLimitDistance;

  uiN = 0;
  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      // Unnecessarily Strict Rule that all vertices around the facets which are around
      // the central vertex must be the same generation or younger
      uiN++;
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  pvxCurr = NULL;
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideDooSabinSelectiveLimit(pvx);

      peStart = pvx->GetEdge();

      // Remove the current vertex
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      m_pheapSelective->Remove(phe);
      DestroyVertex(pvx);

      // Update the priority queue
      pe = peStart;
      do
	{
	  switch ( eErrorMetric )
	    {
	    case emtLimitDistance:
	      {
		// Insert the new vertices
		pvxCurr = pe->GetVertex();
		ASSERT( pvxCurr->GetHeapEntry() == NULL );
		fError = CalculateVertexErrorSubdivideDooSabinSelectiveLimit(pvxCurr);
		phe = m_pheapSelective->Insert(fError, pvxCurr);
		pvxCurr->SetHeapEntry(phe);
	      }
	    break;
	    case emtBestFitPlane:
	    default:
	      {
		for ( peCurr = pe->GetEdgePrev(1);
		      peCurr != pe->GetEdge();
		      peCurr = peCurr->GetEdgeNext(0) )
		  {
		    // Insert the new vertices
		    pvxCurr = peCurr->GetVertex();
		    if ( pvxCurr->GetHeapEntry() == NULL )
		      {   
			fError = CalculateVertexErrorSubdivideDooSabinSelectiveLimit(pvxCurr);
			phe = m_pheapSelective->Insert(fError, pvxCurr);
			pvxCurr->SetHeapEntry(phe);
		      }
		    else
		      {
			phe = pvxCurr->GetHeapEntry();
			ASSERT( phe != NULL );
			fError = CalculateVertexErrorSubdivideDooSabinSelectiveLimit(pvxCurr);
			m_pheapSelective->Update(phe, fError);
		      }
		  }
	      }
	    break;
	    }
	  pe = pe->GetEdgeNext(0);
	}
      while ( pe != peStart );

      // Update the face normals
      pe = peStart;
      pe->GetFacet()->Preprocess();
      do
	{
	  pe->GetEdgePrev(1)->GetFacet()->Preprocess();
	  pe->GetEdgePrev(1)->GetEdgePrev(1)->GetFacet()->Preprocess();

	  pe = pe->GetEdgeNext(0);
	}
      while ( pe != peStart );
    }
  
  /*
  if ( uiN > 0 )
    {
      Preprocess();
    }
    */
}

///////////////////////////////////////////////////
// QEMesh::SubdivideDooSabinSelectiveLimit       //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     v4--e3--v3--e2--v2    //
//           |              |       |       |    //
//           |             e4      e9      e1    //
//           |              |       |       |    //
//  e2------vV------e0  => v5--e10-vV--e8--v1    //
//           |              |       |       |    //
//           |             e5      e11     e0    //
//           |              |       |       |    //
//  f2      e3      f3     v6--e6--v7--e7--v0    //
//                                               //
VOID QEMesh::SubdivideDooSabinSelectiveLimit(QEVertex *pvx)
{
  //QEVertexIter iterQEVertex;
  QEVertex *pvxNew;
  QEEdge *pe, *peNew, *peStart;
  QEFacet *pfaNew;
  //Point ptNew;
  Point pt;
  Vector v, vCenter;

  vCenter = pvx->GetPoint();

  peStart = pvx->GetEdge();
  pe = peStart;
  do
    {
      // Double edges connecting vertices of the same generation
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  peNew = CreateEdge();
	  ASSERT( peNew != NULL );
	  pfaNew = CreateFacet();
	  ASSERT( pfaNew != NULL );

	  peNew->SetVertex(pe->GetVertex());
	  peNew->GetEdge()->SetVertex(pe->GetEdge()->GetVertex());

	  peNew->SetFacet(pe->GetFacet());
	  peNew->GetEdge()->SetFacet(pfaNew);
	  pe->SetFacet(pfaNew);

	  Splice(1, pe, peNew);
	  Splice(1, pe->GetEdge()->GetEdgePrev(1), peNew->GetEdge());
	}

      pe = pe->GetEdgeNext(1)->GetEdgeNext(1);
    }
  while ( pe != peStart );

  pfaNew = CreateFacet();
  ASSERT( pfaNew != NULL );

  pe = peStart;
  do
    {
      // Create ring edges and vertices

      peNew = CreateEdge();
      ASSERT( peNew != NULL );

      pt = CalculateEdgeSubdivisionDooSabinSelective(pe->GetEdgeNext(1));
      pvxNew = CreateVertex(pt);
      ASSERT( pvxNew != NULL );
      pvxNew->SetGeneration(pvx->GetGeneration()+1);
      pvxNew->SetTexture(vCenter);

      peNew->SetVertex(pvxNew);
      peNew->SetFacet(pe->GetFacet());
      peNew->GetEdge()->SetFacet(pfaNew);

      pe->SetForwardEdge(peNew->GetEdge());
      pe->GetEdgeNext(1)->SetForwardEdge(peNew);

      pe = pe->GetEdgeNext(1)->GetEdgeNext(1);
    }
  while ( pe != peStart );

  for ( pe = peStart->GetEdgeNext(1);
	peStart->GetEdgeNext(1) != peStart->GetEdgePrev(1);
	pe = peStart->GetEdgeNext(1) )
    {
      // Unhook the vertex and hook up the facet contour

      Splice(1, pe->GetEdgePrev(1), pe->GetEdgeNext(1));

      pe->SetVertex(pe->GetForwardEdge()->GetVertex());
      pe->GetEdgeNext(1)->SetVertex(pe->GetForwardEdge()->GetVertex());
      pe->GetEdgeNext(1)->GetForwardEdge()->SetVertex(pe->GetForwardEdge()->GetVertex());

      Splice(1, pe->GetForwardEdge(), pe->GetEdgeNext(1)->GetForwardEdge());
      Splice(1, pe->GetForwardEdge(), pe->GetEdgePrev(1));

      pe->SetForwardEdge(NULL);
      pe->GetEdgeNext(1)->SetForwardEdge(NULL);
    }

  // Make the last connection
  pe = peStart->GetEdgeNext(1);

  pe->SetVertex(pe->GetForwardEdge()->GetVertex());
  pe->GetEdgeNext(1)->SetVertex(pe->GetForwardEdge()->GetVertex());
  pe->GetEdgeNext(1)->GetForwardEdge()->SetVertex(pe->GetForwardEdge()->GetVertex());
  
  Splice(1, pe->GetForwardEdge(), pe->GetEdgeNext(1)->GetForwardEdge());
  Splice(1, pe->GetForwardEdge(), pe->GetEdgePrev(1));
  
  pe->SetForwardEdge(NULL);
  pe->GetEdgeNext(1)->SetForwardEdge(NULL);

  //////////////////////////////
  // Calculate Limit position and normals
  peStart = peStart->GetEdgeNext(1);

  pe = peStart;
  do
    {
      pt = CalculateVertexLimitSubdivisionDooSabinSelective(pe);
      pe->GetVertex()->SetLimitPosition(pt);
      v = CalculateVertexNormalSubdivisionDooSabinSelective(pe);
      pe->GetVertex()->SetNormal(v);

      pe = pe->GetEdgeNext(0);
    }
  while ( pe != peStart);

  // Make the defunct vertex point to an edge around the new facet
  pvx->SetEdge(peStart);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideDooSabinSelectiveLimit(QEVertex *pvx)
{
  Vector v;

  v = pvx->GetPoint() - pvx->GetLimitPosition();
  v[3] = 0.0;

  return v.Magnitude();
}

Point QEMesh::CalculateEdgeSubdivisionDooSabinSelective(QEEdge *pe)
{
  Point pt, ptV, ptSum;
  QEEdge *peCurr;
  UINT_32 uiN, i;
  FLOAT fWeight;

  uiN = 0;
  ptV.Null();
  peCurr = pe;
  do
    {
      ASSERT( pe->GetVertex()->GetGeneration() <= peCurr->GetVertex()->GetGeneration() );
      if ( pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() )
	{
	  ptV += peCurr->GetVertex()->GetPoint();
	}
      else
	{
	  ptV += peCurr->GetVertex()->GetTexture();
	}

      peCurr = peCurr->GetEdgeNext(0);
      uiN++;
    }
  while ( peCurr != pe );
  ptV /= (FLOAT)uiN;
  
  i = 0;
  ptSum.Null();
  peCurr = pe;
  do
    {
      fWeight = 1.0 + 2.0 * cos( 2.0 * PI * i / (FLOAT)uiN );
      ASSERT( pe->GetVertex()->GetGeneration() <= peCurr->GetVertex()->GetGeneration() );
      if ( pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() )
	{
	  ptSum += (fWeight * peCurr->GetVertex()->GetPoint());
	}
      else
	{
	  ptSum += (fWeight * peCurr->GetVertex()->GetTexture());
	}

      peCurr = peCurr->GetEdgeNext(0);
      i++;
    }
  while ( peCurr != pe );

  pt = ( ( ( ptV + pe->GetVertex()->GetPoint() ) / 2.0 ) + 
	 ( ( ptSum - ((FLOAT)uiN * pe->GetVertex()->GetPoint()) ) / (4.0*(FLOAT)uiN) ) );
  pt[3] = 1.0;
  
  return pt;
}

Point QEMesh::CalculateVertexLimitSubdivisionDooSabinSelective(QEEdge *peVertex)
{
  QEEdge *pe, *peCurr;
  Point ppt[4], pt;
  UINT_32 i, uiN;
  FLOAT fAlpha, fTheta, fThetaDelta;

  i = 0;
  pe = peVertex;
  do
    {
      ASSERT( i < 4 );

      ppt[i].Null();

      uiN = 0;
      peCurr = pe;
      do
	{
	  ASSERT( pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() ||
		  pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() + 1 );
	  if ( (i & 0x1) &&
	       pe->GetVertex()->GetGeneration() > peCurr->GetVertex()->GetGeneration() )
	    {
	      uiN++;
	    } 
	  uiN++;
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
      ASSERT( ( i == 0 || i == 2 ) ||
	      ( ( i == 1 || i == 3 ) && ( uiN == 4 ) ) );

      ppt[i] = pe->GetVertex()->GetPoint();
      ppt[i] *= (FLOAT)uiN;
      fTheta = 0.0;
      fThetaDelta = (2.0 * PI) / (FLOAT)uiN;
      peCurr = pe;
      do
	{
	  fAlpha = 3.0 + 2.0 * cos(fTheta);

	  ASSERT( pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() ||
		  pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() + 1 );
	  if ( pe->GetVertex()->GetGeneration() > peCurr->GetVertex()->GetGeneration() )
	    {
	      if ( i & 0x1 )
		{
		  // Edge Facet
		  pt = CalculateEdgeSubdivisionDooSabinSelective(peCurr->GetEdgeNext(1));
		  ppt[i] += fAlpha * pt;

		  fTheta += fThetaDelta;
		  fAlpha = 3.0 + 2.0 * cos(fTheta);
		  pt = CalculateEdgeSubdivisionDooSabinSelective(peCurr->GetEdgePrev(1));
		  ppt[i] += fAlpha * pt;
		}
	      else
		{
		  // Vertex or Facet Facet
		  pt = CalculateEdgeSubdivisionDooSabinSelective(peCurr);
		  ppt[i] += fAlpha * pt;
		}
	    }
	  else
	    {
	      pt = peCurr->GetVertex()->GetPoint();
	      ppt[i] += fAlpha * pt;
	    }
	  
	  fTheta += fThetaDelta;
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
      ppt[i] /= (4.0*(FLOAT)uiN);
      ppt[i][3] = 1.0;

      i++;
      pe = pe->GetEdgeNext(1);
    }
  while ( pe != peVertex );
  ASSERT( i == 4 );

  //////////////////////////////////////////////////
  // Calculate the limit position
  Point ptLimit;

  ptLimit.Null();
  for ( i = 0; i < 4; i++ )
    {
      ptLimit += ppt[i];
    }
  ptLimit /= 4.0;
  ptLimit[3] = 1.0;

  return ptLimit;
}

Vector QEMesh::CalculateVertexNormalSubdivisionDooSabinSelective(QEEdge *peVertex)
{
  QEEdge *pe, *peCurr;
  Point ppt[4], pt;
  UINT_32 i, uiN;
  FLOAT fAlpha, fTheta, fThetaDelta;

  i = 0;
  pe = peVertex;
  do
    {
      ASSERT( i < 4 );

      ppt[i].Null();

      uiN = 0;
      peCurr = pe;
      do
	{
	  ASSERT( pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() ||
		  pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() + 1 );
	  if ( (i & 0x1) &&
	       pe->GetVertex()->GetGeneration() > peCurr->GetVertex()->GetGeneration() )
	    {
	      uiN++;
	    } 
	  uiN++;
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
      ASSERT( ( i == 0 || i == 2 ) ||
	      ( ( i == 1 || i == 3 ) && ( uiN == 4 ) ) );

      ppt[i] = pe->GetVertex()->GetPoint();
      ppt[i] *= (FLOAT)uiN;
      fTheta = 0.0;
      fThetaDelta = (2.0 * PI) / (FLOAT)uiN;
      peCurr = pe;
      do
	{
	  fAlpha = 3.0 + 2.0 * cos(fTheta);

	  ASSERT( pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() ||
		  pe->GetVertex()->GetGeneration() == peCurr->GetVertex()->GetGeneration() + 1 );
	  if ( pe->GetVertex()->GetGeneration() > peCurr->GetVertex()->GetGeneration() )
	    {
	      if ( i & 0x1 )
		{
		  // Edge Facet
		  pt = CalculateEdgeSubdivisionDooSabinSelective(peCurr->GetEdgeNext(1));
		  ppt[i] += fAlpha * pt;

		  fTheta += fThetaDelta;
		  fAlpha = 3.0 + 2.0 * cos(fTheta);
		  pt = CalculateEdgeSubdivisionDooSabinSelective(peCurr->GetEdgePrev(1));
		  ppt[i] += fAlpha * pt;
		}
	      else
		{
		  // Vertex or Facet Facet
		  pt = CalculateEdgeSubdivisionDooSabinSelective(peCurr);
		  ppt[i] += fAlpha * pt;
		}
	    }
	  else
	    {
	      pt = peCurr->GetVertex()->GetPoint();
	      ppt[i] += fAlpha * pt;
	    }
	  
	  fTheta += fThetaDelta;
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != pe );
      ppt[i] /= (4.0*(FLOAT)uiN);
      ppt[i][3] = 1.0;

      i++;
      pe = pe->GetEdgeNext(1);
    }
  while ( pe != peVertex );
  ASSERT( i == 4 );

  //////////////////////////////////////////////////
  // Calculate the normal
  Vector vNormal, vT0, vT1, v;

  uiN = 4;
  vT0.Null();
  vT1.Null();
  fTheta = 0.0;
  fThetaDelta = (2.0 * PI) / (FLOAT)uiN;
  for ( i = 0; i < uiN; i++ )
    {
      v = ppt[i];
      vT0 += cos(fTheta) * v;
      fTheta += fThetaDelta;
      vT1 += cos(fTheta) * v;
    }
  vT0[3] = 0.0;
  vT1[3] = 0.0;

  vNormal = vT1 * vT0;
  if ( !vNormal.Normalize() )
    {
      vNormal.Set(0.0, 0.0, 1.0);
    }

  return vNormal;
}

//////////////////////////////////////////////////////////////////////
// Selective Subdivision Sqrt(4)

QEMesh *QEMesh::SubdivideSqrt4Selective(QEMesh::ErrorMetricType eErrorMetric,
					FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEVertex *pvx;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;
  Vector v;

  pme = Copy();
  ASSERT( pme != NULL );

  iterQEMesh.Init(pme);
  for ( pvx = iterQEMesh.PeekFirstVertex();
	pvx != NULL;
	pvx = iterQEMesh.PeekNextVertex() )
    {
      fError = pme->CalculateVertexErrorSubdivideSqrt4Selective(pvx);
      phe = pme->m_pheapSelective->Insert(fError, pvx);
      pvx->SetHeapEntry(phe);

      pvx->SetGeneration(0);
      v.Null();
      pvx->SetTexture(v);
    }

  pme->ReSubdivideSqrt4Selective(eErrorMetric, fTolerance, uiFacetsMax);

  pme->Preprocess();

  return pme;
}

VOID QEMesh::ReSubdivideSqrt4Selective(QEMesh::ErrorMetricType /*eErrorMetric*/,
				       FLOAT fTolerance, UINT_32 uiFacetsMax)
{
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertex *pvx, *pvxCurr;
  QEVertexIter iterQEVertex;
  QEEdge *pe, *peCurr, *peStart;
  FLOAT fError;
  //Vector v;
  UINT_32 uiN;

  uiN = 0;
  for ( pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError));
	pvx != NULL && fError > fTolerance && m_plistQEFacet->GetLength() < uiFacetsMax;
	pvx = (QEVertex*)(m_pheapSelective->PeekTop(&fError)) )
    {
      uiN++;
      pvxCurr = pvx;
      while ( pvxCurr != NULL )
	{
	  pvx = pvxCurr;
	  pvxCurr = NULL;

	  peStart = pvx->GetEdge();
	  pe = peStart;
	  do
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  pvxCurr = peCurr->GetVertex();
		  if ( pvxCurr->GetGeneration() < pvx->GetGeneration() )
		    {
		      break;
		    }
		  pvxCurr = NULL;
		}
	      pe = pe->GetEdgeNext(1);
	    }
	  while ( pe != peStart && pvxCurr == NULL );
	}

      SubdivideSqrt4Selective(pvx);

      // Update the current vertex
      phe = pvx->GetHeapEntry();
      ASSERT( phe != NULL );
      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvx);
      m_pheapSelective->Update(phe, fError);

      iterQEVertex.Init(pvx);
      for ( pe = iterQEVertex.PeekFirstEdge();
	    pe != NULL;
	    pe = iterQEVertex.PeekNextEdge() )
	{

	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }

	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdge()->GetVertex();
	  if ( pvxCurr->GetHeapEntry() == NULL )
	    {   
	      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvxCurr);
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }


	  for ( peCurr = pe->GetEdge()->GetEdgeNext(1)->GetEdgeNext(1);
		peCurr != pe->GetEdge()->GetEdgePrev(1);
		peCurr = peCurr->GetEdgeNext(1) )
	    {
	      pvxCurr = peCurr->GetEdge()->GetVertex();
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }

	  // Update the neighbor vertices
	  pvxCurr = pe->GetEdgeNext(0)->GetEdge()->GetVertex();
	  if ( pvxCurr->GetHeapEntry() == NULL )
	    {   
	      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvxCurr);
	      phe = m_pheapSelective->Insert(fError, pvxCurr);
	      pvxCurr->SetHeapEntry(phe);
	    }
	  else
	    {
	      phe = pvxCurr->GetHeapEntry();
	      ASSERT( phe != NULL );
	      fError = CalculateVertexErrorSubdivideSqrt4Selective(pvxCurr);
	      m_pheapSelective->Update(phe, fError);
	    }
	}
    }
  
  if ( uiN > 0 )
    {
      Preprocess();
    }
}

///////////////////////////////////////////////////
// QEMesh::SubdivideSqrt4Selective               //
//                                               //
//   Old Vertex Disk       Subdivided Facets     //
//                                               //
//  f1      e1      f0     f1              f0    //
//           |               \            /      //
//           |                 e1      e0        //
//           |                   \    /          //
//  e2------vV------e0  =>         vV            //
//           |                   /    \          //
//           |                 e2      e3        //
//           |               /            \      //
//  f2      e3      f3     f2              f3    //
//                                               //
//                                               //
VOID QEMesh::SubdivideSqrt4Selective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEVertex *pvxNew;
  QEEdge *pe, *peCurr, *peNext, *peSpoke, *peNew, *peStart;
  QEFacet faNULL; // Local Temporary NULL facet
  QEFacet *pfaNew;
  Point ptNew, pt;
  //Point ptCenter;
  Vector v;

  v.Null();

  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      // Create facet centers and connect them to the central vertex
      ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  break;
	}
    }

  if ( pe != NULL )
    {
      peStart = pe;
      do
	{
	  peNext = pe->GetEdgeNext(1);

	  // Create facet centers and connect them to the central vertex
	  ASSERT( pe->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
	  if ( pe->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	    {
	      for ( peCurr = pe->GetEdgeNext(0);
		    peCurr != pe;
		    peCurr = peCurr->GetEdgeNext(0) )
		{
		  if ( peCurr->GetVertex()->GetGeneration() > pvx->GetGeneration() )
		    {
		      break;
		    }
		}

	      if ( peCurr == pe )
		{
		  // Case 1
		  pt = CalculateFacetSubdivisionSqrt4Selective(pe->GetFacet());
		  pvxNew = CreateVertex(pt);
		  ASSERT( pvxNew != NULL );
		  pvxNew->SetGeneration(pvx->GetGeneration()+1);
		  pvxNew->SetTexture(v);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );

		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(pvxNew);
		  peSpoke->SetFacet(&faNULL);
		  peSpoke->GetEdge()->SetFacet(&faNULL);

		  Splice(1, peSpoke, pe);

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );

		  peNew->SetVertex(peSpoke->GetEdgePrev(0)->GetVertex());
		  peNew->GetEdge()->SetVertex(pvxNew);
		  peNew->GetEdge()->SetFacet(&faNULL);
		  peNew->SetFacet(pe->GetFacet());

		  Splice(1, peNew, peSpoke->GetEdgePrev(0));
		  Splice(1, peNew->GetEdge(), peSpoke->GetEdge());

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );

		  peNew->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(pvxNew);
		  peNew->SetFacet(&faNULL);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());

		  Splice(1, peNew, pe->GetEdgeNext(0));
		  Splice(1, peNew->GetEdge(), peSpoke->GetEdge());

		  peSpoke->GetEdgePrev(0)->SetFacet(&faNULL);
		  pe->SetFacet(&faNULL);
		}
	      else if ( peCurr == pe->GetEdgePrev(0) )
		{
		  if ( peCurr == pe->GetEdgeNext(0)->GetEdgeNext(0) )
		    {
		      // Case 2
		    }
		  else
		    {
		      // Case 3
		      peNew = CreateEdge();
		      ASSERT( peNew != NULL );

		      peNew->SetVertex(pe->GetEdge()->GetVertex());
		      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		      peNew->SetFacet(&faNULL);
		      peNew->GetEdge()->SetFacet(pe->GetFacet());

		      Splice(1, peNew, pe->GetEdgeNext(0));
		      Splice(1, peNew->GetEdge(), peCurr);

		      peCurr->SetFacet(&faNULL);
		      pe->SetFacet(&faNULL);
		    }
		}
	      else if ( peCurr == pe->GetEdgePrev(0)->GetEdgePrev(0) &&
			peCurr == pe->GetEdgeNext(0)->GetEdgeNext(0) )
		{
		  // Case 4
		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(pe->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(&faNULL);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());
	      
		  Splice(1, peNew, pe);
		  Splice(1, peNew->GetEdge(), peCurr);

		  peCurr->SetFacet(&faNULL);
		  peCurr->GetEdgeNext(0)->SetFacet(&faNULL);
		}
	      else if ( peCurr == pe->GetEdgeNext(0)->GetEdgeNext(0) )
		{
		  // Case 5
		  pe->GetEdgeNext(0)->SetFacet(&faNULL);
		  pe->SetFacet(&faNULL);
		  pe->GetEdgePrev(0)->SetFacet(&faNULL);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );
	      
		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(peCurr->GetVertex());
		  peSpoke->SetFacet(&faNULL);
		  peSpoke->GetEdge()->SetFacet(&faNULL);
	      
		  Splice(1, peSpoke, pe);
		  Splice(1, peSpoke->GetEdge(), peCurr);

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(peSpoke->GetEdgePrev(0)->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(peCurr->GetFacet());
		  peCurr->SetFacet(peCurr->GetFacet());
		  peNew->GetEdge()->SetFacet(&faNULL);
	      
		  Splice(1, peNew, pe);
		  Splice(1, peNew->GetEdge(), peCurr);
		}
	      else if ( peCurr == pe->GetEdgePrev(0)->GetEdgePrev(0) )
		{
		  // Case 6
		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(&faNULL);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());
	      
		  Splice(1, peNew, pe->GetEdgeNext(0));
		  Splice(1, peNew->GetEdge(), peCurr);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );
	      
		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(peCurr->GetVertex());
		  peSpoke->SetFacet(&faNULL);
		  peSpoke->GetEdge()->SetFacet(&faNULL);
	      
		  Splice(1, peSpoke, pe);
		  Splice(1, peSpoke->GetEdge(), peCurr);

		  peCurr->SetFacet(&faNULL);
		  peCurr->GetEdgeNext(0)->SetFacet(&faNULL);
		  pe->SetFacet(&faNULL);
		}
	      else
		{
		  // Case 7
		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(pe->GetEdge()->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(&faNULL);
		  peNew->GetEdge()->SetFacet(pe->GetFacet());
	      
		  Splice(1, peNew, pe->GetEdgeNext(0));
		  Splice(1, peNew->GetEdge(), peCurr);

		  pe->GetEdgePrev(0)->SetFacet(&faNULL);
		  pe->SetFacet(&faNULL);

		  peSpoke = CreateEdge();
		  ASSERT( peSpoke != NULL );
	      
		  peSpoke->SetVertex(pe->GetVertex());
		  peSpoke->GetEdge()->SetVertex(peCurr->GetVertex());
		  peSpoke->SetFacet(&faNULL);
		  peSpoke->GetEdge()->SetFacet(&faNULL);
	      
		  Splice(1, peSpoke, pe);
		  Splice(1, peSpoke->GetEdge(), peCurr);

		  peNew = CreateEdge();
		  ASSERT( peNew != NULL );
	      
		  peNew->SetVertex(peSpoke->GetEdgePrev(0)->GetVertex());
		  peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		  peNew->SetFacet(&faNULL);
		  peNew->GetEdge()->SetFacet(&faNULL);
	      
		  Splice(1, peNew, peSpoke->GetEdgePrev(0));
		  Splice(1, peNew->GetEdge(), peCurr);

		  pfaNew = CreateFacet();
		  ASSERT( pfaNew != NULL );
		  for ( peNew = peCurr->GetEdgeNext(0);
			peNew != peCurr;
			peNew = peNew->GetEdgeNext(0) )
		    {
		      peNew->SetFacet(pfaNew);
		    }
		  peCurr->SetFacet(pfaNew);
		}
	    }
	  else
	    {
	      ASSERT( peNext->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
	      if ( peNext->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
		{
		  if ( pe->GetEdgeNext(0) != pe->GetEdgePrev(0)->GetEdgePrev(0) )
		    {
		      // case 8
		      peCurr = pe->GetEdgeNext(0);

		      peNew = CreateEdge();
		      ASSERT( peNew != NULL );
	      
		      peNew->SetVertex(pe->GetEdgePrev(0)->GetVertex());
		      peNew->GetEdge()->SetVertex(peCurr->GetVertex());
		      peNew->SetFacet(pe->GetFacet());
		      peNew->GetEdge()->SetFacet(&faNULL);
	      
		      Splice(1, peNew, pe->GetEdgePrev(0));
		      Splice(1, peNew->GetEdge(), peCurr);

		      pe->SetFacet(&faNULL);
		      pe->GetEdgePrev(0)->SetFacet(&faNULL);		      
		    }
		  else
		    {
		      // case 9
		    }
		}
	      else
		{
		  //case 10
		}
	    }

	  pe = peNext;
	}
      while ( pe != peStart );
    }

  for ( peStart = pvx->GetEdge();
	peStart->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration();
	peStart = peStart->GetEdgeNext(1) )
    {
    }
  pe = peStart;
  do
    {
      peCurr = pe->GetEdgeNext(1);
      // Remove the old edges and heal the quads
      ASSERT( peCurr->GetEdge()->GetVertex()->GetGeneration() >= pvx->GetGeneration() );
      if ( peCurr->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  if ( peCurr->GetFacet() == &faNULL &&
	       peCurr->GetEdge()->GetFacet() == &faNULL )
	    {
	      pfaNew = CreateFacet();
	      ASSERT( pfaNew != NULL );
	    }
	  else if ( peCurr->GetFacet() != &faNULL &&
		    peCurr->GetEdge()->GetFacet() != &faNULL )
	    {
	      pfaNew = peCurr->GetFacet();
	      DestroyFacet(peCurr->GetEdge()->GetFacet());
	    }
	  else if ( peCurr->GetFacet() != &faNULL )
	    {
	      pfaNew = peCurr->GetFacet();
	    }
	  else
	    {
	      pfaNew = peCurr->GetEdge()->GetFacet();
	    }

	  // Remove the edge
	  Splice(1, peCurr->GetEdge(), peCurr->GetEdge()->GetEdgePrev(1));
	  Splice(1, peCurr, peCurr->GetEdgePrev(1));
	  DestroyEdge(peCurr->GetEdge());
	  DestroyEdge(peCurr);

	  peCurr = pe;
	  do
	    {
	      peCurr->SetFacet(pfaNew);
	      peCurr = peCurr->GetEdgeNext(0);
	    }
	  while ( peCurr != pe );
	}

      pe = pe->GetEdgeNext(1);
    }
  while ( pe != peStart );

  v = pvx->GetPoint();
  pvx->SetTexture(v);
  ptNew = CalculateVertexSubdivisionSqrt4Selective(pvx);
  pvx->SetPoint(ptNew);
  pvx->SetGeneration(pvx->GetGeneration()+1);
}

FLOAT QEMesh::CalculateVertexErrorSubdivideSqrt4Selective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

Point QEMesh::CalculateVertexSubdivisionSqrt4Selective(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Point pt;
  FLOAT fAlpha;
  UINT_32 uiN;

  pt.Null();
  iterQEVertex.Init(pvx);
  for ( uiN = 0, pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	uiN++, pe = iterQEVertex.PeekNextEdge() )
    {
      if ( pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetGeneration() == pvx->GetGeneration() )
	{
	  pt += pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetPoint();
	}
      else
	{
	  pt += pe->GetEdgeNext(0)->GetEdge()->GetVertex()->GetTexture();
	}
    }

  fAlpha = (4.0 - 2.0*cos(2.0*PI/(FLOAT)uiN)) / 9.0;

  pt = ( ((1.0 - fAlpha)*pvx->GetPoint()) + (fAlpha/(FLOAT)uiN)*pt );
  pt[3] = 1.0;

  return pt;
}

Point QEMesh::CalculateFacetSubdivisionSqrt4Selective(QEFacet *pfa)
{
  QEFacetIter iterQEFacet;
  QEVertex *pvx, *pvxStart;
  Point pt;
  UINT_32 uiN;

  pt.Null();
  iterQEFacet.Init(pfa);
  pvxStart = iterQEFacet.PeekFirstVertex();
  for ( uiN = 0, pvx = pvxStart;
	pvx != NULL;
	uiN++, pvx = iterQEFacet.PeekNextVertex() )
    {
      ASSERT( pvxStart->GetGeneration() == pvx->GetGeneration() );
      pt += pvx->GetPoint();
    }

  pt /= (FLOAT)uiN;
  pt[3] = 1.0;

  return pt;
}

//////////////////////////////////////////////////////////////////////
// Progressive Meshes

QEMesh *QEMesh::CreateProgressiveMesh(ErrorMetricType eErrorMetric, 
				      UINT_32 uiFacetsMin/* = 4*/)
{
  QEMesh *pme;
  QEMeshIter iterQEMesh;
  QEEdge *pe;
  CHeapEntry<VOID*, FLOAT> *phe;
  FLOAT fError;

  pme = Copy();
  ASSERT( pme != NULL );

  pme->m_pheapSelective->SetBiggestFirst(FALSE);

  //pme->Display();
  iterQEMesh.Init(pme);
  for ( pe = iterQEMesh.PeekFirstEdge();
	pe != NULL;
	pe = iterQEMesh.PeekNextEdge() )
    {
      if ( pe->GetHeapEntry() == NULL )
	{
	  fError = pme->CalculateEdgeErrorProgressiveMesh(eErrorMetric, pe);
	  phe = pme->m_pheapSelective->Insert(fError, pe);
	  pe->SetHeapEntry(phe);
	  pe->GetEdge()->SetHeapEntry(phe);
	}
    }

  for ( pe = (QEEdge*)(pme->m_pheapSelective->PeekTop(&fError));
	pe != NULL && fError < INFINITY && pme->m_plistQEFacet->GetLength() > uiFacetsMin;
	pe = (QEEdge*)(pme->m_pheapSelective->PeekTop(&fError)) )
    {
      pme->CreateProgressiveCollapseEdge(eErrorMetric, pe);
      //pme->Display();
    }

  pme->m_pecCurr = pme->m_plistQEEdgeCollapse->PeekLast();

  return pme;
}

VOID QEMesh::AdaptProgressiveMesh(UINT_32 uiFacetsMax)
{
  QEEdgeCollapse *pec;

  if ( m_plistQEFacet->GetLength() < uiFacetsMax )
    {
      // Vertex Split
      if ( m_pecCurr->GetEdgeExists() )
	{
	  pec = m_plistQEEdgeCollapse->PeekPrev(m_pecCurr);
	  if ( pec == NULL )
	    {
	      return;
	    }
	  m_pecCurr = pec;
	}
      while ( m_plistQEFacet->GetLength() < uiFacetsMax )
	{
	  ProgressiveSplitVertex(m_pecCurr);
	  //Display();
	  pec = m_plistQEEdgeCollapse->PeekPrev(m_pecCurr);
	  if ( pec == NULL )
	    {
	      return;
	    }
	  m_pecCurr = pec;
	}
    }
  else if ( m_plistQEFacet->GetLength() > uiFacetsMax )
    {
      // EdgeCollapse
      if ( !m_pecCurr->GetEdgeExists() )
	{
	  pec = m_plistQEEdgeCollapse->PeekNext(m_pecCurr);
	  if ( pec == NULL )
	    {
	      return;
	    }
	  m_pecCurr = pec;
	}
      while ( m_plistQEFacet->GetLength() > uiFacetsMax )
	{
	  ProgressiveCollapseEdge(m_pecCurr);
	  //Display();
	  pec = m_plistQEEdgeCollapse->PeekNext(m_pecCurr);
	  if ( pec == NULL )
	    {
	      break;
	    }
	  m_pecCurr = pec;
	}
    }
}

FLOAT QEMesh::CalculateEdgeErrorProgressiveMesh(ErrorMetricType eErrorMetric,
						QEEdge *pe)
{
  FLOAT pfError[2], fError;
  FLOAT pfCoefs[2] = {1.0, 2.0};
  QEFacetIter iterQEFacet;
  QEVertexIter iterQEVertex;
  QEEdge *ppe[2], *peCurr;
  UINT_32 puiN[2], uiN, i;
  Vector v;

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();

  v = ppe[0]->GetVertex()->GetPoint() - ppe[1]->GetVertex()->GetPoint();
  pfError[0] = v.Magnitude();

  pfError[1] = CalculateVertexErrorProgressiveMesh(eErrorMetric, ppe[0]->GetVertex());
  pfError[1] += CalculateVertexErrorProgressiveMesh(eErrorMetric, ppe[1]->GetVertex());

  fError = (pfCoefs[0]*pfError[0] + pfCoefs[1]*pfError[1]) / (pfCoefs[0] + pfCoefs[1]);

  for ( i = 0; i < 2; i++ )
    {
      uiN = 0;
      iterQEFacet.Init(ppe[i]->GetFacet());
      for ( peCurr = iterQEFacet.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEFacet.PeekNextEdge() )
	{
	  uiN++;
	}
      if ( uiN < 4 )
	{
	  ASSERT( uiN > 2 );
	  uiN = 0;
	  iterQEVertex.Init(ppe[i]->GetEdgePrev(0)->GetVertex());
	  for ( peCurr = iterQEVertex.PeekFirstEdge();
		peCurr != NULL;
		peCurr = iterQEVertex.PeekNextEdge() )
	    {
	      uiN++;
	    }
	  if ( uiN < 4 )
	    {
	      fError = INFINITY;
	      return fError;
	    }
	}

      puiN[i] = 0;
      iterQEVertex.Init(ppe[i]->GetVertex());
      for ( peCurr = iterQEVertex.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertex.PeekNextEdge() )
	{
	  puiN[i]++;
	}
      ASSERT( puiN[i] > 2 );
    }

  if ( puiN[0] < 4 && puiN[1] < 4 )
    {
      fError = INFINITY;
    }

  return fError;
}

FLOAT QEMesh::CalculateVertexErrorProgressiveMesh(ErrorMetricType eErrorMetric, 
						  QEVertex *pvx)
{
  FLOAT fError;

  fError = 0.0;
  switch ( eErrorMetric )
    {
    default:
      {
	fError = CalculateVertexErrorProgressiveMeshBestFitPlane(pvx);
      }
      break;
    }

  return fError;
}

FLOAT QEMesh::CalculateVertexErrorProgressiveMeshBestFitPlane(QEVertex *pvx)
{
  QEVertexIter iterQEVertex;
  QEEdge *pe;
  Vector vPlane, v;
  FLOAT fError, fErrorMax;
  UINT_32 uiN;
  Point ptCentroid, ptStart, ptCurr, ptPrev;

  vPlane.Null();

  ptCentroid = pvx->GetPoint();
  uiN = 0;
  iterQEVertex.Init(pvx);
  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCentroid += pe->GetEdge()->GetVertex()->GetPoint();
      uiN++;
    }
  if ( uiN < 3 )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Incomplete Vertex disk for %p\n", pvx);
      return 0.0;
    }
  ptCentroid /= (FLOAT)(uiN+1);
  ptCentroid[3] = 1.0;

  pe = iterQEVertex.PeekFirstEdge();
  ASSERT( pe != NULL );

  ptStart = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;
  ptPrev = ptStart;

  for ( pe = iterQEVertex.PeekNextEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      ptCurr = pe->GetEdge()->GetVertex()->GetPoint() - ptCentroid;

      vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  vPlane[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  vPlane[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  vPlane[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !vPlane.Normalize() )
    {
      fprintf(stderr, "QEMesh::SubdivideSelective: Warning: Null normal because of zero area vertex disk\n");
      return 0.0;
    }

  // Calculate the D coefficient of the Tangent Plane defined by the normal
  vPlane.CompletePlaneEquation(ptCentroid);

  v = pvx->GetPoint();
  v[3] = 1.0;
  fErrorMax = fabs(vPlane.Dot(v));

  for ( pe = iterQEVertex.PeekFirstEdge();
	pe != NULL;
	pe = iterQEVertex.PeekNextEdge() )
    {
      v = pe->GetEdge()->GetVertex()->GetPoint();
      v[3] = 1.0;
      fError = fabs(vPlane.Dot(v));
      
      fErrorMax = MAX(fErrorMax, fError);
    }

  return fErrorMax;
}

VOID QEMesh::CreateProgressiveCollapseEdge(ErrorMetricType eErrorMetric, QEEdge *pe)
{
  QEEdgeCollapse *pec;
  CHeapEntry<VOID*, FLOAT> *phe;
  QEVertexIter iterQEVertex, iterQEVertexInner;
  QEVertex *pvx;
  QEEdge *ppe[2], *peCurr;
  Point pt;
  FLOAT pfError[2], fError;
  UINT_32 i, uiN;

  ppe[0] = pe;
  ppe[1] = pe->GetEdge();
  
  pfError[0] = CalculateVertexErrorProgressiveMesh(eErrorMetric, ppe[0]->GetVertex());
  pfError[1] = CalculateVertexErrorProgressiveMesh(eErrorMetric, ppe[1]->GetVertex());
  if ( pfError[0] == 0.0 && pfError[1] == 0.0 )
    {
      pt = (ppe[0]->GetVertex()->GetPoint() + ppe[1]->GetVertex()->GetPoint()) / 2.0;
    }
  else
    {
      pt = ((pfError[0]*ppe[0]->GetVertex()->GetPoint()) + (pfError[1]*ppe[1]->GetVertex()->GetPoint())) / (pfError[0]+pfError[1]);
    }
  pt[3] = 1.0;
  pvx = CreateVertex(pt);
  ASSERT( pvx != NULL );
  RemoveVertex(pvx);

  pec = CreateEdgeCollapse();
  ASSERT( pec != NULL );

  pec->SetEdge(pe);
  m_pheapSelective->Remove(pe->GetHeapEntry());
  pe->SetHeapEntry(NULL);
  pe->GetEdge()->SetHeapEntry(NULL);
  pec->SetVertex(pvx);
  for ( i = 0; i < 2; i++ )
    {
      peCurr = ppe[i];
      uiN = 0;
      do
	{
	  uiN++;
	  peCurr = peCurr->GetEdgeNext(0);
	}
      while ( peCurr != ppe[i] );

      pec->SetNeighborEdge(i, ppe[i]->GetEdgeNext(0));
      switch ( uiN )
	{
	case 1:
	case 2:
	  {
	    ASSERT( FALSE );
	  }
	  break;
	case 3:
	  {
	    m_pheapSelective->Remove(ppe[i]->GetEdgePrev(0)->GetHeapEntry());
	    ppe[i]->GetEdgePrev(0)->SetHeapEntry(NULL);
	    ppe[i]->GetEdgePrev(0)->GetEdge()->SetHeapEntry(NULL);
	    pec->SetRemovedEdge(i, ppe[i]->GetEdgePrev(0));
	    pec->SetRemovedFacet(i, ppe[i]->GetFacet());
	  }
	  break;
	default:
	  {
	    pec->SetRemovedEdge(i, NULL);
	    pec->SetRemovedFacet(i, NULL);
	  }
	  break;
	}
    }

  // Change the mesh
  ProgressiveCollapseEdge(pec);

  iterQEVertex.Init(pec->GetVertex());
  for ( peCurr = iterQEVertex.PeekFirstEdge();
	peCurr != NULL;
	peCurr = iterQEVertex.PeekNextEdge() )
    {
      iterQEVertexInner.Init(peCurr->GetEdge()->GetVertex());
      for ( peCurr = iterQEVertexInner.PeekFirstEdge();
	    peCurr != NULL;
	    peCurr = iterQEVertexInner.PeekNextEdge() )
	{
	  // Update edge ranking of peCurr
	  phe = peCurr->GetHeapEntry();
	  ASSERT( phe != NULL );
	  fError = CalculateEdgeErrorProgressiveMesh(eErrorMetric, peCurr);
	  m_pheapSelective->Update(phe, fError);
	}
    }
}

VOID QEMesh::ProgressiveCollapseEdge(QEEdgeCollapse *pec)
{
  QEVertexIter iterQEVertex;
  QEFacet *pfa;
  QEEdge *ppe[2], *peCurr;
  UINT_32 i;

  ASSERT( pec->GetEdgeExists() );

  ppe[0] = pec->GetEdge();
  ppe[1] = pec->GetEdge()->GetEdge();
  /*
  printf("CollapseEdge: %p\n", pec);
  printf("  Vertex:   %p\n", pec->GetVertex());
  printf("  Vertices: %p %p\n", ppe[0]->GetVertex(), ppe[1]->GetVertex());
  printf("  Edge:     %p %p\n", ppe[0], ppe[1]);
  printf("  Neighbor: %p %p\n", pec->GetNeighborEdge(0), pec->GetNeighborEdge(1));
  printf("  Removed : %p %p\n", pec->GetRemovedEdge(0), pec->GetRemovedEdge(1));
  printf("  Facet   : %p %p\n", pec->GetRemovedFacet(0), pec->GetRemovedFacet(1));
  */
  for ( i = 0; i < 2; i++ )
    {
      ppe[i]->SetVertex(ppe[i]->GetVertex());
      pec->GetNeighborEdge(i)->SetFacet(pec->GetNeighborEdge(i)->GetFacet());
    }

  // Detach the edge (and subsequently the two end point vertices
  ASSERT( ppe[0]->GetEdgePrev(1) == pec->GetNeighborEdge(1));
  Splice(1, ppe[0], pec->GetNeighborEdge(1));
  ASSERT( ppe[1]->GetEdgePrev(1) == pec->GetNeighborEdge(0));
  Splice(1, ppe[1], pec->GetNeighborEdge(0));
  RemoveVertex(ppe[1]->GetVertex());
  RemoveEdge(ppe[1]);
  RemoveVertex(ppe[0]->GetVertex());
  RemoveEdge(ppe[0]);

  // Merge the two end point vertex cylces together
  Splice(1, pec->GetNeighborEdge(1), pec->GetNeighborEdge(0));
  AddVertex(pec->GetVertex());
  for ( peCurr = pec->GetNeighborEdge(1)->GetEdgeNext(1);
	peCurr != pec->GetNeighborEdge(1);
	peCurr = peCurr->GetEdgeNext(1) )
    {
      peCurr->SetVertex(pec->GetVertex());
    }
  pec->GetNeighborEdge(1)->SetVertex(pec->GetVertex());

  for ( i = 0; i < 2; i++ )
    {
      if ( pec->GetRemovedFacet(i) != NULL )
	{
	  pec->GetNeighborEdge(i)->SetFacet(pec->GetRemovedEdge(i)->GetEdge()->GetFacet());
	  pec->GetNeighborEdge(i)->SetVertex(pec->GetNeighborEdge(i)->GetVertex());
	  pec->GetNeighborEdge(i)->GetEdge()->SetVertex(pec->GetNeighborEdge(i)->GetEdge()->GetVertex());
	  RemoveFacet(pec->GetRemovedFacet(i));
	  Splice(1, pec->GetRemovedEdge(i), pec->GetRemovedEdge(i)->GetEdgePrev(1));
	  Splice(1, pec->GetRemovedEdge(i)->GetEdge(), pec->GetRemovedEdge(i)->GetEdge()->GetEdgePrev(1));
	  RemoveEdge(pec->GetRemovedEdge(i)->GetEdge());
	  RemoveEdge(pec->GetRemovedEdge(i));
	}
    }

  pec->SetEdgeExists(FALSE);

  // Update facet normals
  iterQEVertex.Init(pec->GetVertex());
  for ( pfa = iterQEVertex.PeekFirstFacet();
	pfa != NULL;
	pfa = iterQEVertex.PeekNextFacet() )
    {
      pfa->Preprocess();
    }
}

VOID QEMesh::ProgressiveSplitVertex(QEEdgeCollapse *pec)
{
  QEVertexIter iterQEVertex;
  QEFacet *pfa;
  QEEdge *ppe[2], *peCurr;
  UINT_32 i;

  ASSERT( !pec->GetEdgeExists() );

  ppe[0] = pec->GetEdge();
  ppe[1] = pec->GetEdge()->GetEdge();
  /*
  printf("SplitVertex: %p\n", pec);
  printf("  Vertex:   %p\n", pec->GetVertex());
  printf("  Vertices: %p %p\n", ppe[0]->GetVertex(), ppe[1]->GetVertex());
  printf("  Edge:     %p %p\n", ppe[0], ppe[1]);
  printf("  Neighbor: %p %p\n", pec->GetNeighborEdge(0), pec->GetNeighborEdge(1));
  printf("  Removed : %p %p\n", pec->GetRemovedEdge(0), pec->GetRemovedEdge(1));
  printf("  Facet   : %p %p\n", pec->GetRemovedFacet(0), pec->GetRemovedFacet(1));
  */
  for ( i = 0; i < 2; i++ )
    {
      if ( pec->GetRemovedFacet(i) != NULL )
	{
	  pec->GetRemovedEdge(i)->SetVertex(pec->GetNeighborEdge(i)->GetEdge()->GetVertex());
	  pec->GetRemovedEdge(i)->GetEdge()->SetVertex(pec->GetNeighborEdge(i)->GetVertex());
	  pec->GetRemovedEdge(i)->GetEdge()->SetFacet(pec->GetNeighborEdge(i)->GetFacet());
	  pec->GetRemovedEdge(i)->SetFacet(pec->GetRemovedFacet(i));
	  pec->GetNeighborEdge(i)->SetFacet(pec->GetRemovedFacet(i));
	  AddFacet(pec->GetRemovedFacet(i));
	  Splice(1, pec->GetRemovedEdge(i), pec->GetNeighborEdge(i)->GetEdge()->GetEdgePrev(1));
	  Splice(1, pec->GetRemovedEdge(i)->GetEdge(), pec->GetNeighborEdge(i));
	  AddEdge(pec->GetRemovedEdge(i)->GetEdge());
	  AddEdge(pec->GetRemovedEdge(i));
	}
    }

  RemoveVertex(pec->GetVertex());

  // Split apart the two end point vertex cylces
  Splice(1, pec->GetNeighborEdge(1), pec->GetNeighborEdge(0));

  // Attach the edge (and subsequently the two end point vertices
  Splice(1, ppe[0], pec->GetNeighborEdge(1));
  Splice(1, ppe[1], pec->GetNeighborEdge(0));

  ASSERT( ppe[0]->GetEdgePrev(1) == pec->GetNeighborEdge(1));
  ASSERT( ppe[1]->GetEdgePrev(1) == pec->GetNeighborEdge(0));

  for ( i = 0; i < 2; i++ )
    {
      AddEdge(ppe[i]);
      AddVertex(ppe[i]->GetVertex());
      ppe[i]->SetFacet(pec->GetNeighborEdge(i)->GetFacet());

      peCurr = ppe[i];
      do 
	{
	  peCurr->SetVertex(ppe[i]->GetVertex());
	  peCurr = peCurr->GetEdgeNext(1);
	}
      while ( peCurr != ppe[i] );
    }

  pec->SetEdgeExists(TRUE);

  // Update facet normals
  for ( i = 0; i < 2; i++ )
    {
      iterQEVertex.Init(ppe[i]->GetVertex());
      for ( pfa = iterQEVertex.PeekFirstFacet();
	    pfa != NULL;
	    pfa = iterQEVertex.PeekNextFacet() )
	{
	  pfa->Preprocess();
	}
    }
}

